/*
 * Copyright (c) 2016-2020
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: blake2.h 3111 2020-05-21 18:16:34Z brachman $
 */

/*
 * Partially adapted from the Argon2 distribution:
 *   https://github.com/p-h-c/phc-winner-argon2
 * 2-Jun-2016 (Commit d7266c4)
 * License: CC0 (Creative Commons)
 */

#ifndef PORTABLE_BLAKE2_H
#define PORTABLE_BLAKE2_H

#include <stddef.h>
#include <stdint.h>
#include <limits.h>

#include "blake2-impl.h"

enum Blake2b_constant {
  BLAKE2b_BLOCKBYTES    = 128,
  BLAKE2b_OUTBYTES      =  64,
  BLAKE2b_KEYBYTES      =  64,
  BLAKE2b_SALTBYTES     =  16,
  BLAKE2b_PERSONALBYTES =  16
};

#pragma pack(push, 1)
typedef struct Blake2b_param {
  uint8_t digest_length;                   /*  1 */
  uint8_t key_length;                      /*  2 */
  uint8_t fanout;                          /*  3 */
  uint8_t depth;                           /*  4 */
  uint32_t leaf_length;                    /*  8 */
  uint64_t node_offset;                    /* 16 */
  uint8_t node_depth;                      /* 17 */
  uint8_t inner_length;                    /* 18 */
  uint8_t reserved[14];                    /* 32 */
  uint8_t salt[BLAKE2b_SALTBYTES];         /* 48 */
  uint8_t personal[BLAKE2b_PERSONALBYTES]; /* 64 */
} Blake2b_param;
#pragma pack(pop)

typedef struct Blake2b_state {
  uint64_t h[8];
  uint64_t t[2];
  uint64_t f[2];
  uint8_t buf[BLAKE2b_BLOCKBYTES];
  unsigned int buflen;
  unsigned int outlen;
  uint8_t last_node;
} Blake2b_state;

/*
 * Ensure param structs have not been wrongly padded.
 * Poor man's static_assert.
 */
enum {
  BLAKE2_SIZE_CHECK_0 = 1 / !!(CHAR_BIT == 8),
  BLAKE2_SIZE_CHECK_2 =
  1 / !!(sizeof(Blake2b_param) == sizeof(uint64_t) * CHAR_BIT)
};

enum Blake2s_constant {
  BLAKE2s_OUTBYTES = 32,
  BLAKE2s_KEYBYTES = 32
};

typedef struct Blake2s_state {
  uint8_t b[64];
  uint32_t h[8];
  uint32_t t[2];
  size_t c;
  size_t outlen;
} Blake2s_state;

#if defined(__cplusplus)
extern "C" {
#endif

/* Streaming API */
extern int blake2b_init(Blake2b_state *S, size_t outlen);
extern int blake2b_init_key(Blake2b_state *S, size_t outlen, const void *key,
							size_t keylen);
extern int blake2b_init_param(Blake2b_state *S, const Blake2b_param *P);
extern int blake2b_update(Blake2b_state *S, const void *in, size_t inlen);
extern int blake2b_final(Blake2b_state *S, void *out, size_t outlen);

/* Simple API */
extern int blake2b(void *out, size_t outlen, const void *in, size_t inlen,
				   const void *key, size_t keylen);

/* Argon2 Team - Begin Code */
extern int blake2b_long(void *out, size_t outlen, const void *in, size_t inlen);
/* Argon2 Team - End Code */

extern int do_blake2b_tests(void);

extern int blake2s_init(Blake2s_state *S, size_t outlen);
extern int blake2s_init_key(Blake2s_state *S, size_t outlen,
							const void *key, size_t keylen);
extern void blake2s_update(Blake2s_state *S,
						   const void *in, size_t inlen);
extern void blake2s_final(Blake2s_state *S, void *out);
extern int blake2s(void *out, size_t outlen,
				   const void *key, size_t keylen, const void *in, size_t inlen);
extern int do_blake2s_tests(void);

#if defined(__cplusplus)
}
#endif

#endif
