/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: dacs_malloc.h 3321 2024-06-19 23:39:55Z brachman $
 */

#ifndef _DACS_MALLOC_H_
#define _DACS_MALLOC_H_

#include "log.h"

#include <stdlib.h>

/*
 * If enabled at build time, any file that is compiled with this file included
 * (other than dacs_malloc.c) will have these functions interposed so that
 * the functionality can be obtained from the system library, as usual,
 * or another library that has been configured for this purpose.
 */
#ifdef ENABLE_DACS_MALLOC
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
  /* Valid selectors are positive integers. */
  DACS_MALLOC_SYSTEM   = 1,	/* Use the system library (base). */
  DACS_MALLOC_PHK      = 2,	/* Use the phkmalloc library (base). */
  DACS_MALLOC_DL       = 3,	/* Use the dlmalloc library (base). */
  /*
   * Use the frame library, which wraps one of the base libraries.
   */
  DACS_MALLOC_FRAME    = 4,
  DACS_MALLOC_DEFAULT  = DACS_MALLOC_SYSTEM,
  DACS_MALLOC_INVALID  = -1
} DACS_malloc_selector;

typedef struct DACS_malloc_switch DACS_malloc_switch;

/*
 * XXX Only the original, standardized API functions are handled, plus CHECK_PTR.
 */
struct DACS_malloc_switch {
  char *name;
  DACS_malloc_selector selector;
  void *(*malloc_func)(size_t size);
  void *(*realloc_func)(void *ptr, size_t size);
  void (*free_func)(void *ptr);
  void *(*calloc_func)(size_t nmemb, size_t memb_size);
  int (*check_ptr_func)(const void *);
  DACS_malloc_switch *next;
};

/*
 * Define OMIT_SAFE_MALLOC before including this file if these functions
 * are not required, or if you don't want to drag in the logging functions.
 */
#ifndef OMIT_SAFE_MALLOC
/*
 * The safe_*() functions are statically included because they use
 * log_msg(), which is a macro that interpolates information from the
 * context in which it is used (e.g., the file name and line number).
 */
static inline MAYBE_UNUSED void *
safe_inline_malloc(size_t size, const char *log_module_name)
{
  void *p;

  if ((p = malloc(size)) == NULL) {
	log_msg((LOG_ALERT_LEVEL, "malloc failed, exiting"));
	exit(1);
  }

  return(p);
}
#define safe_malloc(SIZE)		safe_inline_malloc(SIZE, log_module_name)

static inline MAYBE_UNUSED void
safe_free(void *ptr)
{
  free(ptr);
}

static inline MAYBE_UNUSED void *
safe_inline_calloc(size_t number, size_t size, const char *log_module_name)
{
  void *p;

  if ((p = calloc(number, size)) == NULL) {
	log_msg((LOG_ALERT_LEVEL, "calloc failed, exiting"));
	exit(1);
  }

  return(p);
}
#define safe_calloc(NUMBER, SIZE)					\
  safe_inline_calloc(NUMBER, SIZE, log_module_name)

static inline MAYBE_UNUSED void *
safe_inline_realloc(void *ptr, size_t size, const char *log_module_name)
{
  void *p;

  if ((p = realloc(ptr, size)) == NULL) {
	log_msg((LOG_ALERT_LEVEL, "realloc failed, exiting"));
	exit(1);
  }
  return(p);
}
#define safe_realloc(PTR, SIZE)						\
  safe_inline_realloc(PTR, SIZE, log_module_name)

static inline MAYBE_UNUSED char *
safe_inline_strdup(const char *str, const char *log_module_name)
{
  size_t len;
  void *p;

  len = strlen(str) + 1;
  p = safe_malloc(len);
  memcpy(p, str, len);
  return((char *) p);
}
#define safe_strdup(STR)	safe_inline_strdup(STR, log_module_name)

/*
 * Use the caps macros to call a layer that wraps the vanilla versions so that
 * we can catch and log invocations that fail. This makes the calling code
 * much cleaner, because there is no repeated code for handling a failure.
 * It also makes it much easier to change the formatting of log messages,
 * add instrumentation, and to insert a breakpoint for debugging.
 */
#define MALLOC(X)		safe_malloc(X)
#define CALLOC(X, Y)	safe_calloc(X, Y)
#define FREE(X)			safe_free(X)
#define REALLOC(X, Y)	safe_realloc(X, Y)
#define REALLOCF(X, Y)	safe_realloc(X, Y)
#define STRDUP(X)		safe_strdup(X)
#endif

/*
 * Macros for conveniently allocating memory of a given type.
 * The _NS suffixed versions are "not safe"; i.e., unwrapped and no
 * error logging is done.
 */

/* Return a pointer to space allocated for one OBJ. */
#define ALLOC(OBJ)			((OBJ *) MALLOC(sizeof(OBJ)))
#define ALLOC_NS(OBJ)		((OBJ *) malloc(sizeof(OBJ)))

/* Return a pointer to space allocated for a vector of N objects, each an OBJ. */
#define ALLOC_N(OBJ, N)		((OBJ *) MALLOC(sizeof(OBJ) * (N)))
#define ALLOC_N_NS(OBJ, N)	((OBJ *) malloc(sizeof(OBJ) * (N)))

#define REALLOC_N(OBJ, PTR, N)		((OBJ *) REALLOC(PTR, sizeof(OBJ) * (N)))
#define REALLOC_N_NS(OBJ, PTR, N)	((OBJ *) realloc(PTR, sizeof(OBJ) * (N)))

extern DACS_malloc_switch *dacs_malloc_add_library(DACS_malloc_selector which,
										   const char *name,
										   void *(*malloc_cb)(size_t),
										   void *(*calloc_cb)(size_t, size_t),
										   void *(*realloc_cb)(void *, size_t),
										   void (*free_cb)(void *),
										   int (*check_ptr_cb)(const void *));

extern int dacs_malloc_init(void);

extern void *dacs_malloc(size_t size);
extern void *dacs_realloc(void *ptr, size_t size);
extern void *dacs_calloc(size_t nmemb, size_t memb_size);
extern void dacs_cfree(void *ptr);
extern void dacs_free(void *ptr);
extern DACS_malloc_switch *dacs_malloc_check_ptr(const void *ptr);

extern int dacs_debug_malloc(int state);
extern int dacs_set_malloc(DACS_malloc_selector which);
extern int dacs_set_base_malloc(DACS_malloc_selector which);
extern int dacs_get_malloc(DACS_malloc_switch **which_base,
						   DACS_malloc_switch **which);

#ifdef __cplusplus
}
#endif

#endif
