/*
 * Copyright (c) 2003-2025
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: ds.h 3372 2025-04-01 00:15:31Z brachman $
 */

/*****************************************************************************
 * COPYRIGHT AND PERMISSION NOTICE
 * 
 * Copyright (c) 2001-2003 The Queen in Right of Canada
 * 
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, and/or sell
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, provided that the above copyright notice(s) and this
 * permission notice appear in all copies of the Software and that both the
 * above copyright notice(s) and this permission notice appear in supporting
 * documentation.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE 
 * BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 * 
 * Except as contained in this notice, the name of a copyright holder shall not
 * be used in advertising or otherwise to promote the sale, use or other
 * dealings in this Software without prior written authorization of the
 * copyright holder.
 ***************************************************************************/

#ifndef _DS_H_
#define _DS_H_

/*
 * Package for safely allocating and using dynamically constructed character
 * strings and dynamically constructed vectors.
 */

#include "dacs_config.h"
#include "range.h"

#include <regex.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>

#define HAVE_DS

#ifndef MAYBE_UNUSED
#define MAYBE_UNUSED	__attribute__ ((__unused__))
#endif

/* Multi-source input descriptor */
typedef struct Dsio {
  FILE *fp;                 /* If non-NULL, the input stream */
  char *buf;                /* If no input stream, the input buffer... */
  char *ptr;                /* ... and a pointer into the input buffer */
  unsigned long length;     /* The length of the input, if known, in bytes */
  unsigned long rem;        /* Bytes remaining in input, if known */
  int have_length;          /* Non-zero if length is unavailable */
  int got_peeked_char;      /* Non-zero if there's a peeked character */
  char peeked_char;         /* The peeked character */
} Dsio;

/*
 * A dynamic string descriptor.
 *
 * The descriptor and the ds_* API provide operations on dynamically
 * allocated buffers that grow as necessary. These are most often used
 * for character strings, but buffers are just a sequence of unsigned bytes.
 *
 * xxx it might be interesting to add a "virtual head pointer" and a
 * "virtual tail pointer" that could be used in certain contexts to
 * do insert and delete operations using basic functions; e.g.,
 * with the virtual head pointer set to the second character, doing
 * ds_prepend() would insert characters immediately after the second character
 * and with the virtual tail pointer set to the second character, doing
 * ds_append() would insert characters between the second and third characters.
 */
typedef struct Ds {
  unsigned int alloc_flag : 1;	/* Set if this struct was malloc-ed */
  unsigned int clear_flag : 1;	/* Set to zero freed memory */
  unsigned int exact_flag : 1;	/* Allocate the exact amount required */
  unsigned int delnl_flag : 1;	/* Delete newlines read from stream */
  unsigned int crnl_flag  : 1;  /* Make \r\n equivalent to \n from stream */
  unsigned int escnl_flag : 1;	/* Allow lines to be continued */
  void *buf;					/* Pointer to a malloc-ed buffer, or NULL */
  size_t nalloc;				/* Allocated size of the buffer */
  size_t len;					/* Bytes put into the buffer, not */
  								/* necessarily the strlen */
  size_t len_limit;				/* Maximum value of len (0 means no limit) */
  void *(*malloc_func)(size_t size);
  void (*free_func)(void *ptr);
  Dsio *io;
} Ds;

enum ds_prompt_flag {
  DS_PROMPT_ECHO    = 0,
  DS_PROMPT_NOECHO  = 1,
  DS_PROMPT_HIDE    = 2,	/* Not implemented. */
  DS_PROMPT_MAYECHO = 3		/* Not implemented. */
};

/*
 * Return a pointer to the dynamically allocated buffer.
 * Be careful though; if the buffer is reallocated, this pointer could become
 * invalid.
 */
static inline char *
ds_buf(Ds *ds)
{

  return((char *) ds->buf);
}

static inline unsigned char *
ds_ucbuf(Ds *ds)
{

  return((unsigned char *) ds->buf);
}

/*
 * Return the number of bytes in the buffer that are in use (which is less
 * than or equal to the number of bytes allocated for the buffer).
 * Remember that if the buffer contains a null-terminated string, this
 * will return the strlen(3) of the buffer plus one.
 */
static inline size_t
ds_len(Ds *ds)
{

  return(ds->len);
}

/*
 * Return the number of bytes allocated for the buffer (which is greater than
 * or equal to the number of bytes "in use").
 */
static inline size_t
ds_size(Ds *ds)
{

  return(ds->nalloc);
}

/*
 * Return the number of bytes remaining for use.
 */
static inline size_t
ds_rem(Ds *ds)
{

  return(ds->nalloc - ds->len);
}

/*
 * Return 0 if DS1 and DS2 are not identical strings (possibly binary),
 * otherwise return 1.
 */
static inline int
ds_eq(Ds *ds1, Ds *ds2)
{

  if (ds1 == ds2)
	return(1);
  if (ds1->len != ds2->len)
	return(0);
  return(memcmp((const void *) ds1->buf, (const void *) ds2->buf, ds1->len)
		 == 0);
}

/* XXX Should be integrated with Ds */
extern char *ds_errmsg;

#ifdef __cplusplus
extern "C" {
#endif

extern Ds *ds_alloc(void);
extern Ds *ds_alloc_size(size_t nbytes);
extern Ds *ds_init(Ds *ds);
extern Ds *ds_init_size(Ds *ds, size_t nbytes);
extern Ds *ds_dup(Ds *dsx);
extern Ds *ds_reinit(Ds *ds);
extern Ds *ds_reinit_size(Ds *ds, size_t nbytes);
extern Ds *ds_reset(Ds *ds);
extern Ds *ds_reset_buf(Ds *ds);
extern char *ds_fill(Ds *ds, size_t nbytes);
extern void ds_free(Ds *ds);
extern Ds *ds_trim(Ds *ds);
extern Ds *ds_grow(Ds *ds, size_t new_size);
extern int ds_appendc(Ds *ds, int ch);
extern char *ds_append(Ds *ds, char *str);
extern char *ds_appendn(Ds *ds, char *str, size_t len);
extern char *ds_concat(Ds *ds, char *str);
extern char *ds_concatc(Ds *ds, int ch);
extern char *ds_concatn(Ds *ds, char *str, size_t len);
extern char *ds_prepend(Ds *ds, char *str);
extern int ds_prependc(Ds *ds, int ch);
extern Ds *ds_set(Ds *ds, char *str);
extern Ds *ds_setn(Ds *dsx, unsigned char *s, size_t slen);
extern char *ds_copyb(Ds *ds, void *src, size_t len, unsigned int offset);
extern Ds *ds_copyspn(Ds *dsx, unsigned char *s, char *end_chars);
extern char *ds_insertb(Ds *ds, void *src, size_t len, unsigned int offset);
extern Ds *ds_range(Ds *dsx, char *str, char *range_spec, Range_syntax *ors);
extern size_t ds_chop(Ds *ds, size_t chop);
extern int ds_zapnull(Ds *ds);

extern int ds_getc(Ds *ds, FILE *fp, int *chp);
extern char *ds_gets(Ds *ds, FILE *fp);
extern char *ds_agets(Ds *ds, FILE *fp);
extern Ds *ds_load_file(Ds *ds, char *pathname);
extern Ds *ds_getf(FILE *fp);
extern Ds *ds_agetf(Ds *ds, FILE *fp);
extern char *ds_prompt(Ds *ds, char *prompt, unsigned int flag);
extern Ds *ds_subst(Ds *ds, unsigned char ch_src, unsigned char ch_rpl);

#ifndef NOTDEF
extern int ds_asprintf(Ds *ds, const char *fmt, ...)
  __attribute__ ((format (printf, 2, 0)));
extern int ds_psprintf(Ds *ds, const char *fmt, ...)
  __attribute__ ((format (printf, 2, 0)));
extern int ds_sprintf(Ds *ds, unsigned int offset, const char *fmt, ...)
  __attribute__ ((format (printf, 3, 0)));
extern char *ds_xprintf(const char *fmt, ...)
  __attribute__ ((format (printf, 1, 2)));
extern int ds_vasprintf(Ds *ds, const char *fmt, va_list ap)
  __attribute__ ((format (printf, 2, 0)));
extern char *ds_vxprintf(const char *fmt, va_list ap)
  __attribute__ ((format (printf, 1, 0)));
#else
extern int ds_asprintf(Ds *ds, const char *fmt, ...);
extern int ds_psprintf(Ds *ds, const char *fmt, ...);
extern int ds_sprintf(Ds *ds, unsigned int offset, const char *fmt, ...);
extern int ds_vasprintf(Ds *ds, const char *fmt, va_list ap);
extern char *ds_vxprintf(const char *fmt, va_list ap);
extern char *ds_xprintf(const char *fmt, ...);
#endif


extern FILE *ds_fopen_secure(const char *path, const char *mode, size_t size);
extern int ds_fclose(FILE *fp);
extern void ds_fclean(FILE *fp);

extern Ds *ds_dsappend(Ds *dsx, int no_nul, unsigned int n, ...);
extern Ds *ds_dsvaappend(Ds *dsx, int no_nul, unsigned int n, va_list ap);

extern Dsio *dsio_set(Ds *ds, FILE *fp, char *buf, unsigned long len,
					  int have_length);
extern int dsio_eof(Ds *ds);
extern int dsio_nextc(Ds *ds, int *ch_ptr);
extern int dsio_peekc(Ds *ds, int *ch_ptr);
extern char *dsio_agets(Ds *ds);
extern char *dsio_gets(Ds *ds);
extern char *dsio_agets_buf(Ds *ds, char *buf);
extern int dsio_load(Ds *ds);
extern char *dsio_load_str(Ds *ds);
extern int dsio_free(Ds *ds);

#ifdef HAVE_READLINE
extern char *ds_readline(Ds *ds, char *prompt, char *cprompt);
#endif
#ifdef __cplusplus
}
#endif

extern int ds_default_clear_flag;
extern int ds_default_delnl_flag;
extern size_t ds_default_len_limit;
extern void *(*ds_default_malloc_func)(size_t size);
extern void (*ds_default_free_func)(void *ptr);

/*
 * Although Dsvec is a related but really separate idea, it is currently
 * included in this library.
 * The data structure and supporting functions comprise a means of constructing
 * and dynamically extending a vector of objects (such as pointers or
 * structures).
 */
typedef struct Dsvec {
  unsigned int alloc_flag : 1;	/* Set if this struct was malloc-ed */
  unsigned int clear_flag : 1;	/* Set to zero freed memory */
  unsigned int exact_flag : 1;	/* Allocate the exact amount required */
  void **ptr;					/* Pointer to a malloc-ed vector, or NULL */
  size_t el_size;				/* Size of each element, in bytes */
  unsigned int nalloc;			/* Number of elements allocated */
  unsigned int nused;			/* Number of elements used */
  unsigned int nused_limit;		/* Max value of nused (0 means no limit) */
  void *(*malloc_func)(size_t size);
  void (*free_func)(void *ptr);
} Dsvec;

/* Macro for accessing pointer at offset IND, which is of type TYPE. */
#define dsvec_ptr(DSV, IND, TYPE)	((TYPE) dsvec_ptr_index(DSV, IND))

/* Macro for accessing object at offset IND, which is of type TYPE. */
#define dsvec_obj(DSV, IND, TYPE)	(&((TYPE) dsvec_base(DSV))[IND])

#define dsvec_new_obj(DSV, TYPE)	(dsvec_add_obj(DSV) == -1 \
		? (TYPE) NULL : dsvec_obj(DSV, DSV->nused - 1, TYPE))

#define dsvec_base_type(DSV, TYPE)	(TYPE) dsvec_base(DSV)

static inline void **
dsvec_base(Dsvec *dsv)
{

  if (dsv == NULL)
	return(NULL);
  return(dsv->ptr);
}

/*
 * Return 0 if DSV is NULL, otherwise the number of elements in DSV
 * (which may be zero).
 */
static inline unsigned int
dsvec_len(Dsvec *dsv)
{

  if (dsv == NULL)
	return(0);
  return(dsv->nused);
}

/*
 * Return 1 if DSV is NULL or has no elements, 0 otherwise.
 *
 * Although using (dsvec_len(Dsvec) == 0) is equivalent, this is a more
 * explicit way to test whether a Dsvec has no elements,
 * since in some cases it has not been initialized and in others it has
 * been initialized but no elements have been added or all elements have
 * been removed.
 */
static inline int
dsvec_is_empty(Dsvec *dsv)
{

  return((dsv == NULL || dsv->nused == 0) ? 1 : 0);
}

/*
 * Return 0 if DSV is NULL, otherwise the size of each element of DSV.
 */
static inline unsigned int
dsvec_size(Dsvec *dsv)
{

  if (dsv == NULL)
	return(0);
  return(dsv->el_size);
}

typedef struct Mkargv {
  /* Non-zero: include the start/end quote character with the element. */
  int keepq;
  /* Non-zero: include whitespace that precedes an element with the element. */
  int keepws;
  /* Any character in this string is a field separator character. */
  char *ifs;
  /* Any character in this string is a starting quote character. */
  char *startq;
  /*
   * Each character is an ending quote character, but only when it matches the
   * corresponding starting quote character.
   */
  char *endq;
} Mkargv;

/* For dsvec_select() */
typedef struct Select_arg {
  char *regex;
  regex_t *preg;
  int regex_flags;
  int anchored;
  char *errbuf;
  size_t errbuf_size;
} Select_arg;

/*
 * Mode selector bits for the dsvec_complete() default comparison function
 * flag argument.
 */
#define COMPLETE_COMPAR_CASE	0
#define COMPLETE_COMPAR_NOCASE	1

typedef int (*Dsvec_search_compar)(void *, void *, void *,
								   unsigned int, size_t *);

/* For dsvec_complete(). */
typedef struct Completions {
  int start;		/* If non-negative, the index of the first completion. */
  int end;			/* The index of the last completion. */
#ifdef NOTDEF
  Dsvec *dsv;
#endif
  int prefix_cnt;	/* Number of initial common completion characters. */
  int skip_cnt;		/* Number of completion characters after prefix. */
} Completions;

/* For dsvec_sortx(). */
typedef enum Sortx_flag {
  SORTX_DEFAULT      = 0x000,
  SORTX_QSORT        = 0x000,	/* Bits 0:3 are a function family selector. */
  SORTX_HEAPSORT     = 0x001,
  SORTX_MERGESORT    = 0x002,
  SORTX_CUSTOMSORT   = 0x003,
  /* Values 4 to 15 are currently unused but reserved. */
  SORTX_FUNC_MASK    = 0x00f,	/* To identify the requested sort algorithm. */

  SORTX_BLOCK_COMPAR = 0x010,	/* "_b" style. */
  SORTX_ARG_COMPAR   = 0x020,	/* "_r" style. */
  SORTX_COMPAR_MASK  = 0x0f0,	/* To identify the API style. */
  SORTX_INVERT       = 0x100,	/* Invert value of comparison function. */
  SORTX_COPY         = 0x200	/* Do not change input, return a sorted copy. */
} SortxFlag;

/*
 * Within DACS, the reentrant qsort_r() variant and its comparison function
 * always use the original BSD-style function signatures rather than
 * the (annoyingly incompatible) glibc API.
 */
typedef struct Sortx_args {
  int flags;
  void *arg;		/* Optional argument to the "_r" sort style. */
  int rc;			/* Value, if any, returned by the sort function. */
  /*
   * Optional (custom) sort function to call: one of three styles as determined
   * by the flags.  The comparison function configured as an argument is
   * assumed to match the signature of this sort function.
   */
  union {
	int (*sort)(void *base, size_t nmemb, size_t size,
				int (*compar)(const void *, const void *));
	int (*sort_r)(void *base, size_t nmemb, size_t size, void *thunk,
				  int (*compar)(void *argptr, const void *, const void *));
#ifdef __BLOCK__
	int (*sort_b)(void *base, size_t nmemb, size_t size,
				  int (^compar)(const void *, const void *));
#endif
  } f;

  /*
   * Required comparison function to call: one of three styles as determined
   * by the flags.
   */
  union Sortx_compar {
	int (*compar)(const void *, const void *);
	int (*compar_r)(void *argptr, const void *, const void *);
#ifdef __BLOCK__
	int (^compar_b)(const void *, const void *);
#endif
  } c;
} Sortx_args;

/* XXX Should be integrated with Dsvec */
extern char *dsvec_errmsg;

extern int dsvec_default_initial_nalloc;
extern int dsvec_default_clear_flag;
extern int dsvec_default_nused_limit;
extern void *(*dsvec_default_malloc_func)(size_t size);
extern void (*dsvec_default_free_func)(void *ptr);

#ifdef __cplusplus
extern "C" {
#endif
extern Dsvec *dsvec_init(Dsvec *odsv, size_t size);
extern void dsvec_reinit(Dsvec *dsv);
extern Dsvec *dsvec_alloc(size_t size);
extern Dsvec *dsvec_init_size(Dsvec *odsv, size_t size, int n);
extern Dsvec *dsvec_alloc_size(size_t size, int n);
extern void dsvec_free(Dsvec *dsv);
extern Dsvec *dsvec_grow(Dsvec *dsv, unsigned int new_nelements);
extern void *dsvec_ptr_index(Dsvec *dsv, unsigned int ind);
extern void *dsvec_ptr_rindex(Dsvec *dsv, unsigned int rind);
extern int dsvec_append(Dsvec *primary, Dsvec *secondary);
extern int dsvec_prepend(Dsvec *primary, Dsvec *secondary);
extern int dsvec_set(Dsvec *odsv, void *base, size_t size, int n);
extern int dsvec_add_obj(Dsvec *dsv);
extern Dsvec *dsvec_list(Dsvec *odsv, char **list);
extern Dsvec *dsvec_csv(char *str, char *ifs, char *startq, char *endq,
						int include_ws);

extern void dsvec_sort(Dsvec *dsv, int (*compar)(const void *, const void *));
extern Dsvec *dsvec_sortx(Dsvec *dsv, Sortx_args *args);
extern Dsvec *dsvec_uniq(Dsvec *dsv, Sortx_args *args);

extern int dsvec_compar_string(const void *ap, const void *bp);
extern int dsvec_compar_string_nocase(const void *ap, const void *bp);
extern int dsvec_find(Dsvec *dsv, void *el,
					  int (*compar)(const void *, const void *));
extern Dsvec *dsvec_select(Dsvec *dsv, void *el,
						   void *(*compar_init)(const void *),
						   void *compar_init_arg,
						   int (*compar)(const void *, const void *, void *));
extern void * dsvec_select_init(const void *arg);
extern void *dsvec_select_initx(const void *arg);
extern int dsvec_select_compar(const void *a, const void *b, void *compar_arg);
extern void *dsvec_assoc(Dsvec *dsv, void *el,
						 void *(*compar_init)(const void *),
						 void *compar_init_arg,
						 int (*compar)(const void *, const void *, void *));

extern int dsvec_delete_ptr_index(Dsvec *dsv, unsigned int ind);
extern Dsvec *dsvec_load(Ds *ds, Dsvec *dsv);
extern Dsvec *ds_mkargv(Dsvec *dsv, char *str, Mkargv *conf);
extern Dsvec *ds_mkargvx(Dsvec *dsv, char *str, Mkargv *conf, int nulls_flag);
extern Dsvec *ds_mkargv_add(Dsvec *dsv, char *el);
extern Dsvec *ds_mkargv_addv(Dsvec *dsv, int argc, char **argv);

extern int dsvec_add_ptr(Dsvec *dsv, void *ptr);
extern Dsvec *dsvec_copy(Dsvec *odsv, Dsvec *old_dsv);
extern Dsvec *dsvec_subset(Dsvec *odsv, Dsvec *old_dsv, unsigned int start,
						   int len);
extern int dsvec_delete(Dsvec *dsv, unsigned int ind, unsigned int nelements);
extern int dsvec_delete_ptr(Dsvec *dsv, void *ptr);
extern int dsvec_insert(Dsvec *primary, unsigned int ind, Dsvec *secondary);
extern int dsvec_insert_ptr(Dsvec *dsv, unsigned int ind, void *ptr);
extern int dsvec_prepend_ptr(Dsvec *dsv, void *ptr);
extern Dsvec *dsvec_range(Dsvec *odsv, Dsvec *primary, char *range_spec,
						  Range_syntax *rs);
extern int dsvec_replace(Dsvec *primary, unsigned int ind,
						 unsigned int nelements, Dsvec *secondary);
extern int dsvec_replace_ptr(Dsvec *dsv, void *ptr, unsigned int ind);
extern int dsvec_rotate(Dsvec *dsv, int npos);
extern int dsvec_shuffle(Dsvec *dsv);

extern int dsvec_shift(Dsvec *dsv, int count);
extern Dsvec *dsvec_slice(Dsvec *odsv, Dsvec *primary, unsigned int ind,
						  unsigned int len);
extern Dsvec *dsvec_lines(Dsvec *dsv, char *buf);
extern Dsvec *dsvec_load_lines(Dsvec *dsv, char *pathname);
extern int dsvec_streq(Dsvec *dsv1, Dsvec *dsv2);
extern int dsvec_strneq(Dsvec *dsv1, Dsvec *dsv2, size_t len);
extern Dsvec *dsvec_strlist(Dsvec *odsv, const char *str_spec);
extern void dsvec_str_emit(FILE *ofp, Dsvec *dsv);

extern int dsvec_is_subset(Dsvec *dsv, Dsvec *dsv_subset,
						   int (*compar)(const void *, const void *));
extern int dsvec_is_equiv(Dsvec *dsv_a, Dsvec *dsv_b,
						  int (*compar)(const void *, const void *));
extern int dsvec_is_identical(Dsvec *dsv_a, Dsvec *dsv_b,
							  int (*compar)(const void *, const void *));

extern int dsvec_binary_search(Dsvec *dsv, void *ww, Dsvec_search_compar compar,
							   void *conf, unsigned int flags);
extern int dsvec_complete(Dsvec *dsv, char *word, Dsvec_search_compar compar,
						  void *conf, unsigned int flags,
						  Completions *completions);

typedef int (*Dsvec_search_callback)(void *arg, char *test_path);
extern char *dsvec_searchpath(Dsvec *odsv, char *search_path, char *filename,
							  Dsvec_search_callback callback_func,
							  void *callback_arg, int *which);

extern int dsvec_selftest(FILE *out);

#ifdef SIMPLE_GLOB
/* XXX Not ready for prime time. */
extern Dsvec *dsvec_glob(Dsvec *odsv, char *globpat, char *home);
#else
extern Dsvec *dsvec_glob(Dsvec *odsv, char *globpat, int flags, char *home,
						 int *err);
#endif
extern Dsvec *dsvec_fts(Dsvec *odsv, char *root, int options, char **errmsg);

extern int dsvec_mkargv(Dsvec *dsv, char ***argvp, void * (*copy_func)(void *));

#ifdef __cplusplus
}
#endif

static MAYBE_UNUSED Dsvec *
dsvec_set_init(Dsvec *odsv, void *base, size_t size)
{
  Dsvec *dsv;

  if (base == NULL || size == 0)
	return(NULL);

  if (odsv != NULL) {
	if (odsv->el_size != size)
	  return(NULL);
	dsv = odsv;
  }
  else
	dsv = dsvec_init(odsv, size);

  return(dsv);
}

/*
 * Create or append N elements of type OBJ_TYPE to a vector, starting at the
 * address BASE.  If appending, the elements must be of the same size as the
 * existing one.
 * If N is NULL, then a NULL element marks the end of the copied list; the
 * NULL element is NOT copied.
 * DSV is set to point to the list and on entry must either be NULL or a
 * valid vector.
 */
#define dsvec_set(DSV, BASE, OBJ_TYPE, N) \
  { \
    int i, st; \
    Dsvec *xdsv; \
    OBJ_TYPE *ptr; \
    \
	st = ((xdsv = dsvec_set_init(DSV, BASE, sizeof(OBJ_TYPE))) == NULL) ? -1 : 0; \
    ptr = BASE; \
    for (i = 0; st == 0 && ((N == 0) || (i < N)); i++) { \
      if (N == 0 && ptr[i] == NULL) \
        break; \
      if (dsvec_add_ptr(xdsv, ptr[i]) == -1) \
		 st = -1; \
    } \
    DSV = (st == -1) ? NULL : xdsv; \
  }

#endif
