/*
 * Copyright (c) 2003-2025
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: str.h 3368 2025-04-01 00:13:26Z brachman $
 */

/*****************************************************************************
 * COPYRIGHT AND PERMISSION NOTICE
 * 
 * Copyright (c) 2001-2003 The Queen in Right of Canada
 * 
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, and/or sell
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, provided that the above copyright notice(s) and this
 * permission notice appear in all copies of the Software and that both the
 * above copyright notice(s) and this permission notice appear in supporting
 * documentation.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE 
 * BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 * 
 * Except as contained in this notice, the name of a copyright holder shall not
 * be used in advertising or otherwise to promote the sale, use or other
 * dealings in this Software without prior written authorization of the
 * copyright holder.
 ***************************************************************************/

#ifndef _STR_H_
#define _STR_H_

#ifndef MAYBE_UNUSED
#define MAYBE_UNUSED	__attribute__ ((__unused__))
#endif

#include "dacs_config.h"

#include <ctype.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <regex.h>
#include <time.h>
#include <strings.h>
#include <stdio.h>

#include "mkargv.h"
#include "ds.h"

static inline MAYBE_UNUSED int
streq(const char *p, const char *q)
{

	return(strcmp(p, q) == 0);
}

static inline MAYBE_UNUSED int
strneq(const char *p, const char *q, size_t n)
{

	return(strncmp(p, q, n) == 0);
}

static inline MAYBE_UNUSED int
strcaseeq(const char *p, const char *q)
{

	return(strcasecmp(p, q) == 0);
}

static inline MAYBE_UNUSED int
strncaseeq(const char *p, const char *q, size_t n)
{

	return(strncasecmp(p, q, n) == 0);
}

/*
 * Erase a string (such as a plaintext password)
 * Make sure the function isn't optimized away because the string is not
 * read after it is written.
 * See msdn.microsoft.com/library/default.asp?url=/library/en-us/dncode/html/secure10102002.asp
 */
static MAYBE_UNUSED void
strzap(char *str)
{
  volatile char *p;

  if (str != NULL) {
	for (p = (volatile char *) str; *p != '\0'; p++)
	  *p = '\0';
  }
}

/*
 * If CH appears in STR, return its offset from the start of STR; otherwise,
 * return -1.
 * Since strchr(3) considers the terminating null character to be part
 * of the string, CH may be '\0'.
 */
static MAYBE_UNUSED int
strindex(char *str, int ch)
{
  char *p;

  if ((p = strchr(str, ch)) == NULL)
	return(-1);

  return(p - str);
}

#ifndef HAVE_STRNLEN
/*
 * Return the length of P, which may not be null-terminated, up
 * to a maximum of MAXLEN characters.
 */
static inline MAYBE_UNUSED size_t
strnlen(const char *p, size_t maxlen)
{
  char *endp;

  if ((endp = memchr(p, '\0', maxlen)) == NULL)
	return(maxlen);

  return((size_t) (endp - p));
}
#endif

#ifndef HAVE_STRNDUP
/*
 * Return a copy of P, up to a maximum of MAXLEN bytes.
 * That is, strlen() of the returned string will be the lessor of
 * strlen(p) and MAXLEN.
 * The returned string is always null-terminated.
 */
static inline MAYBE_UNUSED char *
strndup(const char *p, size_t maxlen)
{
  char *str;
  size_t len;

  len = strnlen(p, maxlen);

  str = (char *) malloc(len + 1);
  memcpy(str, p, len);
  str[len] = '\0';

  return(str);
}
#endif

static inline MAYBE_UNUSED int
is_octdigit(int ch)
{

  return(ch >= '0' && ch <= '7');
}

/* Assumes CH is known to be an octal digit. */
static inline MAYBE_UNUSED int
octdigit_val(int ch)
{

  if (ch >= '0' && ch <= '9')
	return(ch - '0');

  return(-1);
}

static inline MAYBE_UNUSED int
is_hexdigit(int ch)
{

  return((ch >= '0' && ch <= '9')
		 || (ch >= 'a' && ch <= 'f')
		 || (ch >= 'A' && ch <= 'F'));
}

/* Assumes CH is known to be a hex digit. */
static inline MAYBE_UNUSED int
hexdigit_val(int ch)
{

  if (ch >= '0' && ch <= '9')
	return(ch - '0');
  if (ch >= 'a' && ch <= 'f')
	return(ch - 'a' + 10);
  if (ch >= 'A' && ch <= 'F')
	return(ch - 'A' + 10);

  return(-1);
}

/* For strbtohex() */
#define STRBTOHEX_DEFAULT	0
#define STRBTOHEX_UPPER		(1 << 0)	/* Use upper case digits */
#define STRBTOHEX_NOZ		(1 << 1)	/* Suppress leading zero */

/* For strtr() */
typedef struct Strtr_io {
  int (*char_in)(void *);
  void *in_arg;
  int (*char_out)(int, void *);
  void *out_arg;
  char *errmsg;
} Strtr_io;

/* For strpfields() and strffields() */
typedef struct Strfields {
  Mkargv *conf;
  int argc;
  char **argv;
} Strfields;

#include <math.h>
#include <inttypes.h>

/*
 * Commonly-used C types for conversion by strnum() etc.  Extensible.
 */
typedef enum {
#define STRNUM_I_TYPE			int
  STRNUM_I         = 1, STRNUM_INT   = 1,
  STRNUM_IGZ       = 2, STRNUM_INTGZ = 2,		/* Greater than Zero */
  STRNUM_INN       = 3, STRNUM_INTNN = 3,		/* Not Negative */
#define STRNUM_UI_TYPE			unsigned int
  STRNUM_UI        = 4, STRNUM_UINT  = 4,
  STRNUM_UINZ      = 5, STRNUM_UINTZ = 5,		/* Not Zero */
#define STRNUM_L_TYPE			long
  STRNUM_L         = 6, STRNUM_LONG  = 6,
#define STRNUM_UL_TYPE			unsigned long
  STRNUM_UL        = 7, STRNUM_ULONG = 7,
#define STRNUM_TIME_T_TYPE		time_t
  STRNUM_TIME_T    = 8,
#define STRNUM_MODE_T_TYPE		mode_t
  STRNUM_MODE_T    = 9,
#define STRNUM_PID_T_TYPE		pid_t
  STRNUM_PID_T     = 10,
#define STRNUM_IN_PORT_T_TYPE	in_port_t
  STRNUM_IN_PORT_T = 11,
#define STRNUM_STR_TYPE			char *
  STRNUM_STR       = 12,
#define STRNUM_SIZE_T_TYPE		size_t
  STRNUM_SIZE_T    = 13,
#define STRNUM_SSIZE_T_TYPE		ssize_t
  STRNUM_SSIZE_T   = 14,
#define STRNUM_LL_TYPE			long long
  STRNUM_LL        = 15, STRNUM_LONGLONG = 15,
#define STRNUM_ULL_TYPE			unsigned long long
  STRNUM_ULL       = 16, STRNUM_ULONGLONG = 16,
#define STRNUM_IMAX_T_TYPE		intmax_t
  STRNUM_IMAX_T    = 17, STRNUM_INTMAX_T  = 17,
#define STRNUM_UIMAX_T_TYPE		uintmax_t
  STRNUM_UIMAX_T   = 18, STRNUM_UINTMAX_T = 18,
#define STRNUM_I32_T_TYPE		int32_t
  STRNUM_I32_T     = 19, STRNUM_INT32_T   = 19,
#define STRNUM_I64_T_TYPE		int64_t
  STRNUM_I64_T     = 20, STRNUM_INT64_T   = 20,
#define STRNUM_UI32_T_TYPE		uint32_t
  STRNUM_UI32_T    = 21, STRNUM_UINT32_T  = 21,
#define STRNUM_UI64_T_TYPE		uint64_t
  STRNUM_UI64_T    = 22, STRNUM_UINT64_T  = 22,
#define STRNUM_FLOAT_TYPE		float
  STRNUM_F         = 23, STRNUM_FLOAT = 23,
#define STRNUM_DOUBLE_TYPE		double
  STRNUM_D         = 24, STRNUM_DOUBLE = 24,
#define STRNUM_LDOUBLE_TYPE		long double
  STRNUM_LD        = 25, STRNUM_LDOUBLE = 25, STRNUM_LONGDOUBLE = 25,
  STRNUM_ERR       = -1
} Strnum;

typedef enum {
  STRNUM_CHECK  = 0,	/* Discard the value. */
  STRNUM_STORE  = 1,	/* Store the value in given memory. */
  STRNUM_ALLOC  = 2,	/* Copy the value into allocated memory. */
  STRNUM_NOCOPY = 3		/* Set the value pointer to given memory. */
} Strnum_mode;

#ifdef NOTDEF
/*
 * We might generalize strnum_b() to replace its BASE argument with a
 * pointer to a set of configuration parameters.
 * This could be used to validate the input string in various ways via a
 * callback function (e.g., check that it falls into a given range or is a
 * certain length) or filter/convert/translate the input string
 * (e.g., uppercase to lowercase, enum string name to enum integer).
 * Callbacks could be invoked before and/or after the TYPE conversion.
 *
 * Not sure this is even a good idea because it would go way beyond
 * straightforward string-to-number conversions into a general
 * string -> validate/filter -> convert -> validate/filter -> output API.
 * The main use case is parameter validating/converting.
 */
typedef struct Strnum_conf {
  int base;
  int (*strnum_callback)(const char *str, Strnum type, Strnum_mode mode,
						 int base, void *value, char **endp, void *arg);
  void *arg;
} Strnum_conf;
#endif

/* Concatenation. */
#define STRNUM_VAR_NAME(NAME)	NAME ## _var

typedef struct Strnum_var {
  Strnum type;
  char *str;
  union {
	STRNUM_I_TYPE			STRNUM_VAR_NAME(STRNUM_I);
	STRNUM_UI_TYPE			STRNUM_VAR_NAME(STRNUM_UI);
	STRNUM_L_TYPE			STRNUM_VAR_NAME(STRNUM_L);
	STRNUM_UL_TYPE			STRNUM_VAR_NAME(STRNUM_UL);
	STRNUM_TIME_T_TYPE		STRNUM_VAR_NAME(STRNUM_TIME_T);
	STRNUM_MODE_T_TYPE		STRNUM_VAR_NAME(STRNUM_MODE_T);
	STRNUM_PID_T_TYPE		STRNUM_VAR_NAME(STRNUM_PID_T);
	STRNUM_IN_PORT_T_TYPE	STRNUM_VAR_NAME(STRNUM_IN_PORT_T);
	STRNUM_STR_TYPE			STRNUM_VAR_NAME(STRNUM_STR);
	STRNUM_SIZE_T_TYPE		STRNUM_VAR_NAME(STRNUM_SIZE_T);
	STRNUM_SSIZE_T_TYPE		STRNUM_VAR_NAME(STRNUM_SSIZE_T);
	STRNUM_LL_TYPE			STRNUM_VAR_NAME(STRNUM_LL);
	STRNUM_ULL_TYPE			STRNUM_VAR_NAME(STRNUM_ULL);
	STRNUM_IMAX_T_TYPE		STRNUM_VAR_NAME(STRNUM_IMAX_T);
	STRNUM_UIMAX_T_TYPE		STRNUM_VAR_NAME(STRNUM_UIMAX_T);
	STRNUM_I32_T_TYPE		STRNUM_VAR_NAME(STRNUM_I32_T);
	STRNUM_I64_T_TYPE		STRNUM_VAR_NAME(STRNUM_I64_T);
	STRNUM_UI32_T_TYPE		STRNUM_VAR_NAME(STRNUM_UI32_T);
	STRNUM_UI64_T_TYPE		STRNUM_VAR_NAME(STRNUM_UI64_T);
	STRNUM_FLOAT_TYPE		STRNUM_VAR_NAME(STRNUM_FLOAT);
	STRNUM_DOUBLE_TYPE		STRNUM_VAR_NAME(STRNUM_DOUBLE);
	STRNUM_LDOUBLE_TYPE		STRNUM_VAR_NAME(STRNUM_LDOUBLE);
  } u;
} Strnum_var;

typedef struct Strnum_tab {
  Strnum type;
  char *name;
  int size;				/* Size of data type, in bytes. */
  char *c_name;			/* The canonical C name. */
} Strnum_tab;

typedef enum {
  STROPT_OK             =  1,
  STROPT_DUP_ERR        = -1,	/* Key appears more than once. */
  STROPT_TYPE_ERR       = -2,	/* Error converting value. */
  STROPT_NOTFOUND_ERR   = -3,	/* Key was not found. */
  STROPT_CONVERSION_ERR = -4,	/* Conversion callback failed. */
  STROPT_PARSE_ERR      = -5,	/* Option format invalid, parse failed. */
  STROPT_UNKNOWN_ERR    = -6,	/* Internal error/bug. */
  STROPT_PARAM_ERR      = -7	/* Invalid parameter. */
} Stropt_rc;

/* Configuration flags. */
typedef enum Stropt_flags {
  STROPT_DEFAULT        = 0,		/* Defaults: must be zero. */
  STROPT_NOCASE         = 0x001,	/* Case insensitive name (key) matching. */
  STROPT_KEY_UNIQUE     = 0x000,	/* Default: Key may only be seen once. */
  STROPT_KEY_MULTI      = 0x002,	/* Key can be multi-valued. */
  STROPT_KEY_KEEP_FIRST = 0x004,	/* Keep first key instance, ignore others. */
  STROPT_KEY_KEEP_LAST  = 0x006,	/* Keep last key instance, ignore others. */
  STROPT_KEY_CATCH      = 0x008,	/* Handle all non-matching names. */
  STROPT_KEY_IGNORE     = 0x00a,	/* Do no conversions. */
  STROPT_KEY_MODE_MASK  = 0x00e,	/* Mask for selecting key mode. */
  STROPT_VALUE_NOCOPY   = 0x000,	/* Default: Do not keep the value string. */
  STROPT_VALUE_SET_PTR  = 0x010,	/* Set a pointer to given value string. */
  STROPT_VALUE_COPY     = 0x020,	/* Set a pointer to copied value string. */
  STROPT_VALUE_MASK     = 0x030,	/* Mask for selecting value mode. */
  STROPT_REQUIRED       = 0x080		/* Name (key) must be configured. */
} Stropt_flags;

typedef struct Stropt_tab Stropt_tab;

typedef Stropt_rc (*Stropt_pre_callback)(Stropt_tab *m, char *name,
										 char **value);
typedef Stropt_rc (*Stropt_callback)(Stropt_tab *m, char *name, char *value,
									 void **v);

/*
 * This is similar to X's XrmOptionDescRec and XtResource facility.
 * It is handy for visible, extensible,
 * more secure, and less error prone handling of runtime parameters.
 */
struct Stropt_tab {
  char *name;						/* The option name (key). */
  Strnum type;						/* Type of string conversion. */
  int flags; 		 				/* Various Stropt_flags. */
  Stropt_pre_callback pre_callback;	/* Pre-conversion callback. */
  Stropt_callback post_callback;	/* Post-conversion callback. */
  Stropt_callback callback_func;	/* Use this instead of strnum(). */
  void *callback_arg;				/* Passed to the callback functions. */
  int count;						/* Number of occurrences and error flag. */
  void *value;						/* Pointer to the converted value. */
  void *value_str;					/* Optionally, the original value string. */
};

/* For strpack() and strunpack() */
typedef struct Strpack {
  Strnum type;
  void *ptr;
} Strpack;

/* For strsubst() */
typedef struct Strsubst {
  char *src_text;
  char *subst_text;
} Strsubst;

/* For is_strclass() and is_strclass_char(). */
enum {
  STRCLASS_ERROR         = 0,
  STRCLASS_LC            = (1 << 0),	/* Lower case. */
  STRCLASS_UC            = (1 << 1),	/* Upper case. */
  STRCLASS_ALPHA         = (STRCLASS_LC|STRCLASS_UC),
  STRCLASS_DECIMAL       = (1 << 2),
  STRCLASS_DIGITS        = STRCLASS_DECIMAL,
  STRCLASS_ALNUM         = (STRCLASS_ALPHA|STRCLASS_DECIMAL),
  STRCLASS_OCTAL         = (1 << 3),
  STRCLASS_HEX_UC        = (1 << 4),
  STRCLASS_HEX_LC        = (1 << 5),
  STRCLASS_HEX           = (STRCLASS_DECIMAL|STRCLASS_HEX_UC|STRCLASS_HEX_LC),
  STRCLASS_PUNCT         = (1 << 6),
  STRCLASS_HOSTNAME      = (1 << 7),	/* RFC 1123 and others. */
  STRCLASS_SPEC          = (1 << 8),
  STRCLASS_SIGNED_DIGITS = (1 << 9),
  STRCLASS_SPACE         = (1 << 10),
  STRCLASS_EOL           = (1 << 11),
  STRCLASS_PRINTABLE     = (STRCLASS_ALNUM|STRCLASS_PUNCT|STRCLASS_SPACE),
  STRCLASS_NUL           = (1 << 12),
  STRCLASS_STOPCH        = (1 << 13),
  STRCLASS_GENNUM        = (1 << 14),
  STRCLASS_NUMERIC_DATE  = (1 << 15)
};

#define IS_STRCLASS(A, C)	((A & C) == C)

typedef struct Strtable {
  int ncols;
  int nrows;
  int align;
  int column_sep;
  Dsvec **columns;
} Strtable;

typedef struct Strtable_row_sort {
  int rownum;
  char *colval;
} Strtable_row_sort;

/*
 * The following are for strcomma(), used to prettyprint numbers by inserting
 * commas as necessary.
 */

/* In case you want to use a period or some other character instead. */
#define STRCOMMA_CHAR		','

typedef enum {
  STRCOMMA_DEFAULT     = 0,		/* Do not fabricate a plus sign. */
  STRCOMMA_WITHSIGN    = 1,		/* Always include a sign. */
  STRCOMMA_WITHOUTSIGN = 2		/* Always suppress a sign. */
} Strcomma_flag;

typedef struct Strregex_conf {
  char *regex;
  int cflags;
  regex_t *preg_in;			/* Compiled version of regex/cflags. */
  unsigned int nmatch;
  int errcode;
} Strregex_conf;

/*
 * Base64 encoding algorithms.
 */
typedef enum {
  STRBA64_DACS64 = 0,		/* DACS base64 algorithm. */
  STRBA64_B64    = 1,		/* RFC 4648 base64. */
  STRBA64_B64URL = 2		/* RFC 4648 base64url. */
} Strba64_alg;

/*
 * Base64 decoding algorithms.
 */
typedef enum {
  STRA64B_DACS64 = 0,
  STRA64B_B64    = 1,
  STRA64B_B64URL = 2
} Stra64b_alg;

#ifdef __cplusplus
extern "C" {
#endif

extern int strnum(char *str, Strnum type, void *value);
extern int strnumx(const char *str, Strnum type, void *value, char **endp);
extern int strnum_b(const char *str, Strnum type, Strnum_mode mode,
					int base, void *value, char **endp);
extern Strnum_tab *strnum_ind(int ind);
extern Strnum_tab *strnum_lookup_name(char *name);
extern Strnum_tab *strnum_lookup_type(Strnum type);
extern Strnum strnum_type(char *name);
extern size_t strnum_size(Strnum type);

extern Strnum_var *strnum_var(Strnum_var *ovar, Strnum type, char *fmt,
							  void *val);

extern Stropt_rc stropt(Stropt_tab *conv_tab, char *name,
						   char *value, Stropt_tab **which);
extern int stropt_argv(char **argv, Stropt_tab *conv_tab, int *ind);
extern Stropt_rc stropt_check(Stropt_tab *conv_tab, Stropt_tab **which);
extern Stropt_tab *stropt_lookup(Stropt_tab *conv_tab, char *name);
extern void stropt_reset(Stropt_tab *conv_tab);

extern char *strsuffix(char *p, size_t len, const char *suffix);
extern char *strcasesuffix(const char *p, size_t len, const char *suffix);
extern char *strprefix(const char *p, const char *prefix);
extern char *strcaseprefix(const char *p, const char *prefix);
extern int strncaseprefix(char *prefix, char *word, size_t match_len);
extern Ds *strfold(char *str);
extern int is_alpha_string(const char *str);
extern int is_digit_string(const char *str);
extern int is_signed_digit_string(const char *str);
extern int is_strclass(char *str, int allowable, ...);
extern int is_strclass_char(int ch, int allowable, char *spec);

extern unsigned char *stra32b(const char *inp, unsigned char **outp,
							  unsigned int *nbytes);
extern unsigned char *stra32b_x(const char *inp, unsigned char **outp,
								unsigned int *nbytes,
								int ignore_key_length);
extern unsigned char *strdec32(const char *inp);
extern unsigned int strba32(const unsigned char *inp, unsigned int nbytes,
							int suppress_padding, char **outp);
extern char *strbenc32(const unsigned char *str, unsigned int len);
extern char *strbenc32_np(const unsigned char *str, unsigned int len);
extern char *strenc32(char *str);
extern Ds *ds_strba32(Ds *ods, const unsigned char *inp, unsigned int nbytes,
					  int suppress_padding);

extern Ds *ds_strba64_x(Ds *ods, Strba64_alg alg, const unsigned char *inp,
						unsigned int nbytes);
extern Ds *ds_stra64b_x(Ds *ods, Stra64b_alg alg, const char *inp, int end_char);

extern Ds *ds_strba64(Ds *ods, const unsigned char *inp, unsigned int nbytes);
extern Ds *ds_stra64b(Ds *ods, const char *inp);
extern unsigned int strba64(const unsigned char *inp,
							unsigned int nbytes, char **outp);
extern unsigned char *stra64b(const char *inp,
							  unsigned char **outp, unsigned int *nbytes);
extern unsigned char *strdec64(const char *inp);
extern char *strbenc64(const unsigned char *str, unsigned int len);
extern char *strenc64(char *str);

extern Ds *ds_strba85(Ds *ods, const unsigned char *inp, unsigned int nbytes);
extern unsigned int strba85(const unsigned char *inp, unsigned int nbytes,
							char **outp);
extern unsigned char *stra85b(const char *inp, unsigned char **outp,
							  unsigned int *nbytes);

#ifndef HAVE_STRSEP
extern char *strsep(char **stringp, const char *delim);
#endif
extern int strchrcount(const char *p, int c);
extern int strchrscount(const char *p, const char *charset);
extern int strprintable(char *str, size_t nbytes, int allow_sp_ws);
extern int strcasecspn(const char *s1, const char *s2);

extern char *dacs_strstr(const char *str, const char *substr);
#ifndef HAVE_STRSTR
#define strstr dacs_strstr
#endif

extern char *dacs_strcasestr(const char *str, const char *substr);
#ifndef HAVE_STRCASESTR
#define strcasestr dacs_strcasestr
#endif

extern char *strrstr(const char *str, const char *substr);
extern char *strrcasestr(const char *str, const char *substr);

extern char *strqs(const char *str, const char *substr);
extern char *strcaseqs(const char *str, const char *substr);
extern char *strqsx(const char *str, const char *endp, const char *substr,
					unsigned int *td_one, size_t plen, int nocase);
extern unsigned int *strqs_init(int *td_one, const char *pattern,
								size_t pattern_len, int nocase);
extern char *strpack(Strpack *desc, const char *delimit);
extern int strunpack(const char *str, const char *delimit, int limit,
					 Strpack *desc);

extern int memzeq(const void *ptr, size_t len);
extern int memeq(void *pp, void *qq, size_t len);
extern int memcaseeq(const void *b1, const void *b2, size_t len);
#define HAVE_MEMZAP
extern void memzap(void *s, size_t len);
extern void memzapb(void *s, int val, size_t len);
extern void *memdupn(void *, size_t len);

extern void envzap(char *name);
extern char *env_parse(char *env_string, char **env_name, char **env_val);
extern char *strcpyn(char *dst, const char *src, size_t n);
extern char *strtolower(const char *);
extern char *strlowerbuf(char *buf, char *str);
extern char *strtoupper(const char *);
extern char *strcapitalize(Ds *ods, char *str);
extern char *strquote(char *str, char *qchars);
extern char *strdequote(char *p);
extern char *strextract(char *str, char *start_q, char *end_q);
extern char *strtrim(char *str, char *trim_chars, int limit);
extern Dsvec *strsplit(const char *str, const char *delim, int limit);
extern Dsvec *strsplitx(Dsvec *odsv, const char *str, const char *delimit,
						int limit);
extern Dsvec *strsplitd(const char *str, const char *delimit, int limit,
						int inc_delimit);
extern Dsvec *strsplit_nocopy(char *str, const char *delim, int limit);
extern Dsvec *strsplit_vec(Dsvec *dsv, const char *delimit, int limit);
extern Dsvec *strsplit_re(const char *str, const char *regex, int limit,
						  int inc_delimit, char **errmsg);
extern char *strjoin_subset(Dsvec *dsv, unsigned int start, int len,
							char *sepstr);
extern char *strjoin(Dsvec *dsv, char *sepstr);
extern Dsvec *strpathparse(const char *path, int *is_abs, int *trailing);

extern Ds *struncescape(char *str);
extern char *strcescape(char *str, size_t slen);
extern Strfields *strpfields(const char *str, Strfields *fields);
extern char *strffields(const char *fmt, Strfields *spec, ...);
extern char *strregexsub(char *src, char *regex, char *repl, int cflags,
						 int repeat, char **errmsg);
extern int strmatch(char *src, char *regex);
extern int strregex(char *src, char *regex, int cflags, char **match,
					char **errmsg);
extern int strregex_x(char *src, Strregex_conf *conf, Dsvec *matches);
extern Strregex_conf *strregex_init(Strregex_conf *conf, char *regex_str,
									int cflags, unsigned int nmatch);
extern int strregexflags(char *str);

extern char *strbasename(const char *pathname, const char *suffix);
extern char *strdirname(const char *pathname);
extern char *strextname(const char *pathname);

extern unsigned char *strhextob(char *str, unsigned int *lenp);
extern char *strbtohex(unsigned char *buf, size_t len, int flags);
extern Ds *ds_strba16(Ds *ods, unsigned char *buf, size_t len, int flags);
extern unsigned char *stra16b(char *str, unsigned char **outp,
							  unsigned int *lenp);

extern Ds *strsubst(char *str, Strsubst *tab);
extern char *strunique(char *str);
extern char *strelide(char *str, char *del);
extern Ds *strelidews(Ds *nds, Ds *ods);
extern char *strchop(char *str, char *del);

extern char *strchrdup(const char *s, int stop_ch, char **endp);
extern char *strcspndup(const char *s, const char *charset, char **endp);
extern char *strclassdup(const char *s, int allowable, char **endp);

extern char *strfmt(const char *, const char *, size_t);

extern Ds *strbtoutf16le(unsigned char *buf, size_t nbytes);

extern Strtable *strtable_init(int ncols, int column_sep, int align);
extern void strtable_add_row(Strtable *st, ...);
extern Dsvec *strtable_sort(Strtable *st, int col,
							int (*compar)(const void *, const void *));
extern void strtable_format(FILE *fp_out, Strtable *st, Dsvec *dsv_sorted);
extern void strtable_free(Strtable *st);

extern char *strcomma(Strnum type, void *arg, Strcomma_flag flag);
extern char *strfilemode(mode_t mode, char *outp);

extern char *strui2b(unsigned char *val, size_t nbytes, int big_endian,
					 char *outbuf, char **startp);

/* In ds.c */
extern Dsvec *dsvec_csv_unpack(char *str, char *ifs, Strnum type);
extern char *dsvec_csv_pack(Dsvec *dsv, char *ifs, char *fmt, Strnum type);

extern DACS_ui32 strhash32(unsigned char *str, size_t slen);
extern DACS_ui64 strhash64(unsigned char *str, size_t slen);

extern int strhash(char *alg, unsigned char *str, size_t slen, void *valuep);

extern int hexpair2int(char *pair);

#ifndef HAVE_ISBLANK
extern int isblank(int);
#endif
  
/* In strtr.c */
extern int strtr_io(char *string1, char *string2, int cflag, int dflag,
					int sflag, Strtr_io *io);
extern int strtr(char *input, char *string1, char *string2, int cflag,
				 int dflag, int sflag, char **tr_string);
extern int strtr_char(int ch, const char *string1, int cflag);

#ifdef __cplusplus
}
#endif

#endif
