/*
 * The BLAKE2s cryptographic hash function and self tests.
 * Designed for 8- to 32-bit platforms and produces digests of any size
 * between 1 and 32 bytes.
 * Refer to RFC 7693 and https://blake2.net
 *
 * Copyright (c) 2017-2023
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

#ifndef MAYBE_UNUSED
#define MAYBE_UNUSED    __attribute__ ((__unused__))
#endif

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2003-2023\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: blake2s.c 3251 2023-06-28 22:59:07Z brachman $";
#endif

/*
 * An implementation based on RFC 7693
 *   https://tools.ietf.org/html/rfc7693
 */

#include <stdint.h>
#include <string.h>
#include <stdio.h>

#include "blake2.h"
#include "blake2-impl.h"

static const uint32_t blake2s_IV[8] = {
  UINT32_C(0x6a09e667), UINT32_C(0xbb67ae85),
  UINT32_C(0x3c6ef372), UINT32_C(0xa54ff53a),
  UINT32_C(0x510e527f), UINT32_C(0x9b05688c),
  UINT32_C(0x1f83d9ab), UINT32_C(0x5be0cd19)
};

/*
 * Initialize context S with optional KEY (0 <= KEYLEN <= 32 bytes)
 * OUTLEN (1 <= OUTLEN <= 32) is the requested digest size in bytes.
 * Return 0 if ok, -1 if there is an error.
*/
int
blake2s_init_key(Blake2s_state *S, size_t outlen, const void *key, size_t keylen)
{
  size_t i;

  if (S == NULL)
	return(-1);

  if (outlen == 0 || outlen > BLAKE2s_OUTBYTES || keylen > BLAKE2s_KEYBYTES)
	return(-1);

  for (i = 0; i < 8; i++)
	S->h[i] = blake2s_IV[i];

  S->h[0] ^= 0x01010000 ^ (keylen << 8) ^ outlen;

  S->t[0] = 0;
  S->t[1] = 0;
  S->c = 0;
  S->outlen = outlen;

  for (i = keylen; i < 64; i++)
	S->b[i] = 0;

  if (keylen > 0) {
	blake2s_update(S, key, keylen);
	S->c = 64;
  }

  return(0);
}

int
blake2s_init(Blake2s_state *S, size_t outlen)
{
  int st;

  st = blake2s_init_key(S, outlen, NULL, 0);

  return(st);
}

static const uint8_t blake2s_sigma[10][16] = {
  {  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15 },
  { 14, 10,  4,  8,  9, 15, 13,  6,  1, 12,  0,  2, 11,  7,  5,  3 },
  { 11,  8, 12,  0,  5,  2, 15, 13, 10, 14,  3,  6,  7,  1,  9,  4 },
  {  7,  9,  3,  1, 13, 12, 11, 14,  2,  6,  5, 10,  4,  0, 15,  8 },
  {  9,  0,  5,  7,  2,  4, 10, 15, 14,  1, 11, 12,  6,  8,  3, 13 },
  {  2, 12,  6, 10,  0, 11,  8,  3,  4, 13,  7,  5, 15, 14,  1,  9 },
  { 12,  5,  1, 15, 14, 13,  4, 10,  0,  7,  6,  3,  9,  2,  8, 11 },
  { 13, 11,  7, 14, 12,  1,  3,  9,  5,  0, 15,  4,  8,  6,  2, 10 },
  {  6, 15, 14,  9, 11,  3,  0,  8, 12,  2, 13,  7,  1,  4, 10,  5 },
  { 10,  2,  8,  4,  7,  6,  1,  5, 15, 11,  9, 14,  3, 12, 13,  0 }
};

#define B2S_G(A, B, C, D, X, Y)		   \
  { v[A] = v[A] + v[B] + X;			   \
	v[D] = rotr32(v[D] ^ v[A], 16);	   \
	v[C] = v[C] + v[D];				   \
	v[B] = rotr32(v[B] ^ v[C], 12);	   \
	v[A] = v[A] + v[B] + Y;			   \
	v[D] = rotr32(v[D] ^ v[A], 8);	   \
	v[C] = v[C] + v[D];				   \
	v[B] = rotr32(v[B] ^ v[C], 7);	   \
  }

static void
blake2s_compress(Blake2s_state *S, int is_last)
{
  int i;
  uint32_t m[16], v[16];

  for (i = 0; i < 8; i++) {
	v[i] = S->h[i];
	v[i + 8] = blake2s_IV[i];
  }

  v[12] ^= S->t[0];
  v[13] ^= S->t[1];

  if (is_last)
	v[14] = ~v[14];

  for (i = 0; i < 16; i++)
	m[i] = load32(&S->b[4 * i]);

  for (i = 0; i < 10; i++) {
	B2S_G( 0, 4,  8, 12, m[blake2s_sigma[i][ 0]], m[blake2s_sigma[i][ 1]]);
	B2S_G( 1, 5,  9, 13, m[blake2s_sigma[i][ 2]], m[blake2s_sigma[i][ 3]]);
	B2S_G( 2, 6, 10, 14, m[blake2s_sigma[i][ 4]], m[blake2s_sigma[i][ 5]]);
	B2S_G( 3, 7, 11, 15, m[blake2s_sigma[i][ 6]], m[blake2s_sigma[i][ 7]]);
	B2S_G( 0, 5, 10, 15, m[blake2s_sigma[i][ 8]], m[blake2s_sigma[i][ 9]]);
	B2S_G( 1, 6, 11, 12, m[blake2s_sigma[i][10]], m[blake2s_sigma[i][11]]);
	B2S_G( 2, 7,  8, 13, m[blake2s_sigma[i][12]], m[blake2s_sigma[i][13]]);
	B2S_G( 3, 4,  9, 14, m[blake2s_sigma[i][14]], m[blake2s_sigma[i][15]]);
  }

  for (i = 0; i < 8; i++)
	S->h[i] ^= v[i] ^ v[i + 8];

#undef B2S_G
}

/*
 * Add INLEN more bytes from IN to the hash function with context (state) S.
 */
void
blake2s_update(Blake2s_state *S, const void *in, size_t inlen)
{
  int i;

  for (i = 0; i < inlen; i++) {
	if (S->c == 64) {
	  S->t[0] += S->c;
	  if (S->t[0] < S->c)
		S->t[1]++;
	  blake2s_compress(S, 0);
	  S->c = 0;
	}
	S->b[S->c++] = ((const uint8_t *) in)[i];
  }
}

/*
 * Generate the message digest with size given to blake2_init().
 * The result is copied into OUT, which is assumed to be large enough.
 */
void
blake2s_final(Blake2s_state *S, void *out)
{
  int i;

  S->t[0] += S->c;
  if (S->t[0] < S->c)
	S->t[1]++;

  while (S->c < 64)
	S->b[S->c++] = 0;

  /* Final block. */
  blake2s_compress(S, 1);

  /* Store result as little endian. */
  for (i = 0; i < S->outlen; i++) {
	((uint8_t *) out)[i]
	  = (S->h[i >> 2] >> (8 * (i & 3))) & 0xff;
  }
}

/*
 * Simple API for common usage.
 * Compute the BLAKE2s hash of the first INLEN bytes of IN using KEY of
 * KEYLEN bytes, putting the OUTLEN byte result in OUT.
 * Note that KEY (0 <= KEYLEN <= BLAKE2s_KEYBYTES bytes) and IN are optional,
 * and (1 <= OUTLEN <= BLAKE2s_OUTBYTES).
 *
 * Return 0 if ok, -1 if an error occurs.
 */
int
blake2s(void *out, size_t outlen, const void *in, size_t inlen,
		const void *key, size_t keylen)
{
  Blake2s_state S;
  int ret;

  ret = -1;

  if (out == NULL || outlen == 0 || outlen > BLAKE2s_OUTBYTES)
	goto fail;

  if (keylen > BLAKE2s_KEYBYTES)
	goto fail;

  if ((in == NULL && inlen != 0) || (key == NULL && keylen != 0))
	goto fail;

  if (blake2s_init_key(&S, outlen, key, keylen) == -1)
	goto fail;

  blake2s_update(&S, in, inlen);
  blake2s_final(&S, out);
  ret = 0;

fail:
  burn(&S, sizeof(S));
  return(ret);
}

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

/*
 * RFC 7693
 * The BLAKE2 Cryptographic Hash and Message Authentication Code (MAC)
 */

/*
 * Test vectors from the BLAKE2 official implementations
 * https://github.com/BLAKE2/BLAKE2/blob/master/testvectors/blake2s-kat.txt
 * and elsewhere
 */

/*
 * If the first character of IN or KEY is a space, the value (sans space)
 * is the plaintext string to use, otherwise it is the hex-encoded byte
 * sequence to use.
 */
typedef struct Blake2s_test {
  char *name;
  char *in;
  int in_seq_count;
  char *key;
  int key_seq_count;
  char *hval;
} Blake2s_test;

static Blake2s_test blake2s_test[] = {
  /* https://github.com/BLAKE2/BLAKE2/blob/master/testvectors/blake2s-kat.txt */
  { "blake2s-kat.txt[1]", NULL, 0, NULL, 32,
	"48a8997da407876b3d79c0d92325ad3b89cbb754d86ab71aee047ad345fd2c49" },
  { "blake2s-kat.txt[2]", NULL, 1, NULL, 32,
	"40d15fee7c328830166ac3f918650f807e7e01e177258cdc0a39b11f598066f1" },
  { "blake2s-kat.txt[3]", NULL, 2, NULL, 32,
	"6bb71300644cd3991b26ccd4d274acd1adeab8b1d7914546c1198bbe9fc9d803" },
  { "blake2s-kat.txt[4]", NULL, 3, NULL, 32,
	"1d220dbe2ee134661fdf6d9e74b41704710556f2f6e5a091b227697445dbea6b" },
  { "blake2s-kat.txt[5]", NULL, 4, NULL, 32,
	"f6c3fbadb4cc687a0064a5be6e791bec63b868ad62fba61b3757ef9ca52e05b2" },
  { "blake2s-kat.txt[6]", NULL, 5, NULL, 32,
	"49c1f21188dfd769aea0e911dd6b41f14dab109d2b85977aa3088b5c707e8598" },
  { "blake2s-kat.txt[7]", NULL, 6, NULL, 32,
	"fdd8993dcd43f696d44f3cea0ff35345234ec8ee083eb3cada017c7f78c17143" },
  /* XXX there are lots more to add... */

  /*
   * From a variety of sources...
   * From Python pyblake2 0.9 documentation.
   * https://pythonhosted.org/pyblake2/examples.html
   */
  { "pyblake2[1]", NULL, 0, NULL, 0, "1bf21a98c78a1c376ae9" },
  { "pyblake2[2]", NULL, 0, NULL, 0, "567004bf96e4a25773ebf4" },
  { "RFC7693[1]", "abc", 0, NULL, 0,
	"508c5e8c327c14e2e1a72ba34eeb452f37458b209ed63a294d999b4c86675982" },
  { "Wikipedia[1]", NULL, 0, NULL, 0,
	"1fa1291e65248b37b3433475b2a0dd63d54a11ecc4e3e034e7bc1ef4" },
  { "Wikipedia[2]", NULL, 0, NULL, 0,
	"69217a3079908094e11121d042354a7c1f55b6482ca1a51e1b250dfd1ed0eef9" },

  { NULL, NULL, 0, NULL, 0, NULL }
};

static int
import_hex(const char *str, uint8_t **valp, size_t *lenp, char **errp)
{
  int i;
  const char *p, *which;
  uint8_t *bin, *bp;
  size_t binlen, len;
  static char *hex_chars = "0123456789abcdef";

  if (str == NULL) {
	*valp = NULL;
	*lenp = 0;
	return(0);
  }

  len = strlen(str);

  if (len & 01) {
	*errp = "odd length input string";
	return(-1);
  }

  if (len == 0) {
	*valp = NULL;
	*lenp = 0;
	return(0);
  }

  binlen = len / 2;
  bp = bin = (uint8_t *) malloc(binlen);

  p = str;
  for (i = 0; i < binlen; i++) {
	which = strchr(hex_chars, tolower((int) *p));
	if (which == NULL) {
	  *errp = "invalid hex character";
	  return(-1);
	}
	bp[i] = (which - hex_chars) * 16;

	p++;
	which = strchr(hex_chars, tolower((int) *p));
	if (which == NULL) {
	  *errp = "invalid hex character";
	  return(-1);
	}
	bp[i] += which - hex_chars;
	p++;
  }

  *valp = bin;
  *lenp = binlen;
  return(0);
}

static char *
to_hex(const uint8_t *value, size_t nbytes)
{
  int i;
  char *p;
  const uint8_t *v;
  static char buf[512];

  if (((nbytes * 2) + 1) > sizeof(buf))
	return(NULL);

  v = value;
  p = buf;
  for (i = 0; i < nbytes; i++)
	p += sprintf(p, "%.2x", *v++ & 0xff);

  return(buf);
}

/*
 * Initialize OUT with LEN bytes from the Fibonacci generator.
 * RFC 7693, Appendix E.
 */
static void
selftest_seq(uint8_t *out, size_t len, uint32_t seed)
{
  int i;
  uint32_t a , b, t;

  a = 0xdead4bad * seed;
  b = 1;

  /* Fill the buffer. */
  for (i = 0; i < len; i++) {
	t = a + b;
	a = b;
	b = t;
	out[i] = (t >> 24) & 0xff;
  }
}

/*
 * RFC 7693's RECOMMENDED BLAKE2s validation self-test.
 * Return 0 if OK, -1 otherwise.
 */
static int
do_blake2s_selftest(void)
{
  int i;
  uint8_t md[BLAKE2s_OUTBYTES];
  Blake2s_state S;
  static const size_t b2s_md_len[4] = { 16, 20, 28, 32 };
  static const size_t b2s_in_len[6] = { 0, 3, 64, 65, 255, 1024 };
  static const uint8_t blake2s_res[BLAKE2s_OUTBYTES] = {
    0x6a, 0x41, 0x1f, 0x08, 0xce, 0x25, 0xad, 0xcd,
	0xfb, 0x02, 0xab, 0xa6, 0x41, 0x45, 0x1c, 0xec,
	0x53, 0xc5, 0x98, 0xb2, 0x4f, 0x4f, 0xc7, 0x87,
	0xfb, 0xdc, 0x88, 0x79, 0x7f, 0x4c, 0x1d, 0xfe
  };

  fprintf(stderr, "BLAKE2s RFC 7693 Self-Test Module...\n");

  /* Initialize for a 256-bit hash. */
  if (blake2s_init(&S, BLAKE2s_OUTBYTES) == -1) {
	fprintf(stderr, "blake2s_init() failed\n");
	return(-1);
  }

  for (i = 0; i < 4; i++) {
	int j;
	uint8_t in[1024], *inp, key[BLAKE2s_KEYBYTES], *keyp;
	size_t outlen, inlen;

	outlen = b2s_md_len[i];
	for (j = 0; j < 6; j++) {
	  inlen = b2s_in_len[j];

	  /* Unkeyed hash. */
	  if (inlen) {
		selftest_seq(in, inlen, inlen);
		inp = in;
	  }
	  else
		inp = NULL;
	  if (blake2s(md, outlen, inp, inlen, NULL, 0) == -1) {
		fprintf(stderr, "Unkeyed hash failed!\n");
		return(-1);
	  }

	  /* Hash the hash. */
	  blake2s_update(&S, md, outlen);

	  /* Keyed hash. */
	  if (outlen) {
		selftest_seq(key, outlen, outlen);
		keyp = key;
	  }
	  else
		keyp = NULL;
	  if (blake2s(md, outlen, inp, inlen, keyp, outlen)  == -1) {
		fprintf(stderr, "Keyed hash failed!\n");
		return(-1);
	  }

	  /* Hash the hash. */
	  blake2s_update(&S, md, outlen);
	}
  }

  /* Hashing done, get the resulting message digest and check it. */
  blake2s_final(&S, md);

  if (memcmp(md, blake2s_res, BLAKE2s_OUTBYTES) != 0) {
	fprintf(stderr, "Failed!\n");
	fprintf(stderr, "Expected: %s\n", to_hex(blake2s_res, BLAKE2s_OUTBYTES));
	fprintf(stderr, "Got:      %s\n", to_hex(md, BLAKE2s_OUTBYTES));
	return(-1);
  }

  fprintf(stderr, "BLAKE2s RFC 7693 Self-Test Module... ok\n");
  return(0);
}

/*
 * Generate the simple key sequence used with the official test vectors.
 */
static void
set_selftest_seq(uint8_t *buf, size_t len)
{
  int i;

  for (i = 0; i < len; i++) {
	buf[i] = i;
  }
}

int
do_blake2s_test_vectors(void)
{
  int st, tnum;
  uint8_t outbuf[BLAKE2s_OUTBYTES];
  uint8_t keybuf[1024], *keyp;
  uint8_t inbuf[1024], *inp;
  size_t hlen, inlen, keylen;
  Blake2s_test *bt;

  tnum = 1;
  for (bt = &blake2s_test[0]; bt->name != NULL; bt++) {
	char *errmsg;
	uint8_t *ehval;
	size_t ehlen;

	if (bt->key == NULL) {
	  if (bt->key_seq_count) {
		set_selftest_seq(keybuf, bt->key_seq_count);
		keyp = keybuf;
	  }
	  else
		keyp = NULL;
	  keylen = bt->key_seq_count;
	}
	else {
	  keyp = (uint8_t *) bt->key;
	  keylen = strlen(bt->key);
	}

	if (bt->in == NULL) {
	  if (bt->in_seq_count) {
		set_selftest_seq(inbuf, bt->in_seq_count);
		inp = inbuf;
	  }
	  else
		inp = NULL;
	  inlen = bt->in_seq_count;
	}
	else {
	  inp = (uint8_t *) bt->in;
	  inlen = strlen(bt->in);
	}

	fprintf(stderr, "%s: ", bt->name);

	hlen = strlen(bt->hval) / 2;
	st = blake2s(outbuf, hlen, inp, inlen, keyp, keylen);
	if (st == -1) {
	  fprintf(stderr, "failed: test %d\n", tnum);
	  return(-1);
	}
	fprintf(stderr, "%s...", to_hex(outbuf, hlen));
	import_hex(bt->hval, &ehval, &ehlen, &errmsg);
	if (memcmp(outbuf, ehval, ehlen) != 0) {
	  fprintf(stderr, "failed: test %d\n", tnum);
	  fprintf(stderr, "Expected: %s\n", bt->hval);
	  return(-1);
	}
	fprintf(stderr, "ok\n");
	tnum++;
  }
  fprintf(stderr, "BLAKE2s test vectors... ok\n");
  return(0);
}

int
do_blake2s_tests(void)
{

  fprintf(stderr, "BLAKE2s selftests\n");

  fprintf(stderr, "1. Official test vectors...\n");
  if (do_blake2s_test_vectors() == -1)
	return(-1);

  fprintf(stderr, "2. RFC 7693, Appendix E. selftest...\n");
  if (do_blake2s_selftest() == -1)
	return(-1);

  return(0);
}

#ifdef PROG

extern unsigned char *crypto_hmac(const char *digest_name,
								  unsigned char *key, unsigned int klen,
								  unsigned char *value, unsigned int vlen,
								  unsigned char **hmac, unsigned int *hlen);
int
main(int argc, char **argv)
{
  char *hexstr;
  unsigned int hlen;
  unsigned char *h;

  /* A quick and dirty HMAC test. */
  h = crypto_hmac("blake2s", NULL, 0, "abc", 3, NULL, &hlen);
  if (h == NULL)
	exit(1);
  hexstr = to_hex(h, hlen);
  fprintf(stderr, "hmac=\"%s\"\n", hexstr);
  if (hlen != BLAKE2s_OUTBYTES
	  ||  strcmp(hexstr,
		 "4a79a31b10a8c34b7c89735b8088e6b810b4bad4f36d671b65ca2ea0914952bf")) {
	fprintf(stderr, "Failed!\n");
	exit(1);
  }

  if (do_blake2s_tests() == -1) {
	fprintf(stderr, "Failed\n");
	exit(1);
  }
  
  exit(0);
}
#endif
