/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * This is a simple, extensible compression layer, modeled after the
 * cryptographic digest API in crypto.c, to access any of several compression
 * algorithms in a uniform way.
 *
 * To build as a command:
 *   make dacscompress
 *
 * The API includes two "styles" for compressing and decompressing.
 * For compression:
 * 1) All-at-once buffer compression: compress_pack() and compress_pack_x();
 *    output to automatically allocated buffers
 * 2) Chunked compression: compress_pack_open(), compress_pack_update(),
 *    compress_pack_finish(), compress_pack_close(); output to a user-supplied
 *    buffer
 *
 * Stream compression is implemented in terms of chunked compression and
 * is provided via compress_fpack() and compress_fpack_x(), and
 * compress_funpack() and compress_funpack_x().
 *
 * As usual in our code, the "_x" function name suffix variety provide extended
 * functionality and/or configurability.
 *
 * Decompression works similarly, the main difference being that the correct
 * library can (usually) be automatically selected by inspecting the input.
 *
 * The initial goal is to provide basic compression/decompression
 * functionality for several libraries rather than a complete, high-performance
 * and full-functioned API.  Consequently, many features of the supported
 * compression libraries are not available, such as multithreading,
 * dictionary access, reuse of contexts, frames, etc.
 *
 * XXX In this initial implementation, configuration and use of the compression
 * libraries is very basic and undoubtedly suboptimal.
 * XXX need better reuse of compression/decompression configuration and contexts
 * XXX memory leaks are likely
 * XXX add an optional mode that compresses an input using each available
 * algorithm to find the one that works best OR "learn" which algorithm
 * usually compresses best for a given class of input and apply a table-driven
 * heuristic to select an algorithm at run time (mainly for one-shot encryption
 * but could also buffer and process the initial N bytes of a stream).
 *
 * A small assortment of modern algorithms ought to be sufficient, but new
 * libraries and improved configurability should be easy to integrate.
 * Supporting several compression libraries is more about application
 * portability (being able to build and run on many platforms without much
 * additional administrator effort) than about seeking optimal performance.
 * We are mainly interested in payloads of between a few hundred bytes and a
 * few megabytes, and are more focussed on higher compression factors than
 * lower CPU or memory requirements.  All of these algorithms meet our
 * requirements well enough.
 *
 * To validate the supported algorithms, you can use 'dacscompress -t'
 * or 'dacscompress -ta <alg>'.
 * For <alg>:      Use:
 *    bz2  ..   bzip2/bunzip2 [.bz2]
 *    lzma ..   xz/unxz       [.xz]
 *    zlib ..   gzip/gunzip   [.gz]
 *    zstd ..   zstd/unzstd   [.zst]
 * Not available yet:
 *    brotli .. brotli        [.br]
 *
 * Why was this implemented?
 * 1. Security is improved by compressing before encrypting; that should be
 *    carefully weighed against the time/space overhead of using compression.
 * 2. DACS can emit data that is stored client-side (e.g., cookies, information
 *    embedded in a web page), and minimizing the footprint of that data is
 *    desirable and perhaps necessary.
 * 3. Random applications that use the DACS library may use this functionality
 *    (e.g., pv, the password vault).
 * 4. It exercises other DACS components, which helps to find bugs,
 *    portability issues, etc.
 * 5. Although much better benchmark programs are available, a testbed for
 *    comparing the available compression libraries is provided when this is
 *    built as a program (if the symbol PROG is defined).
 *
 * Note: to avoid repetitiveness and to follow naming used by third party APIs,
 * we will sometimes use "pack" for "compress" and "unpack" for "decompress".
 */

/*
 * Discover new compression innovations Brotli and Zstandard
 * The Guardian, 01-Dec-2016
 * https://www.theguardian.com/info/developer-blog/2016/dec/01/discover-new-compression-iinovations-brotli-and-zstandard
 */

/*
 * For benchmarks, see:
 *   https://github.com/inikep/lzbench
 */

/*
 * XXX at present, decompression relies on the application and/or underlying
 * compression library to be responsible for efficient buffer management.
 *
 * Also, either the application or libmagic(3)
 * are expected to know the compression method that was used.
 * This layer could emit a (compression type, uncompressed length) header
 * and/or trailer to improve decompression efficiency, but there is a
 * way to ask for a block of memory to be reserved just before the start
 * of the compressed data.
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2024\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: compress.c 3321 2024-06-19 23:39:55Z brachman $";
#endif

#include "local.h"

#include "dacs_crypto.h"
#include "compress.h"

static int debug_flag = 0;

static MAYBE_UNUSED char *log_module_name = "compress";

/*
 * Most compression libraries allow the three standard dynamic memory
 * allocation functions they use to be selected at runtime.  Different
 * compression libraries may use different calling signatures for these
 * functions, however, so for each library we implement stubs that call the
 * following three functions, which inspect their configuration to call the
 * actual allocation/deallocation function and do some common bookkeeping tasks
 * (they are also handy places to put breakpoints when debugging).
 *
 * For example, zlib_alloc() calls custom_alloc() which might call malloc(3);
 * zlib_dealloc() calls custom_dealloc() which might call free(3); and
 * zlib_realloc() calls custom_realloc() which might call realloc(3).
 * But an application's particular "handle" for a compression operation
 * might be configured to call "myalloc()" instead of malloc(3) and "myfree()"
 * instead of free(3), for example.
 *
 * Each compression library that supports a custom memory allocator is
 * configured by its *_compress_open() and *_decompress_open() functions to
 * call custom_alloc(), custom_dealloc(), and custom_realloc().
 *
 * These three functions should only be called by compress_* functions.
 *
 * custom_alloc() calls default_alloc() (malloc()) by default, or the alloc()
 * function the caller has set in the configuration.
 * custom_dealloc() calls default_dealloc() (free()) by default, or the dealloc()
 * function the caller has set in the configuration.
 * custom_realloc() calls realloc() by default, the
 * realloc() function the caller has set in the configuration, or
 * else the configured dealloc() and alloc() functions.
 *
 * The three functions should always be compatible, lest memory obtained
 * through one library be deallocated or reallocated by a very different one.
 * If the defaults need to be overrideen, the user should probably always set
 * all three functions in the same function call.
 */
static void *
custom_alloc(Compress_conf *conf, size_t need)
{
  void *ptr;

  if (conf->alloc == NULL)
	ptr = malloc(need);
  else
	ptr = conf->alloc(need);

  if (ptr == NULL) {
	DACS_ABORT_UTIL(("Allocation error, need=%lu", need));
	/*NOTREACHED*/
  }

  conf->result.buffer = ptr;
  conf->result.buflen = need;

  conf->result.total_alloc += need;
  conf->result.alloc_count++;
  if (debug_flag)
	memset(ptr, (int) 0xbb, need);

  return(ptr);
}

static void
custom_dealloc(Compress_conf *conf, void *ptr)
{

  if (conf->dealloc == NULL)
	free(ptr);
  else
	conf->dealloc(ptr);
  conf->result.dealloc_count++;

  if (ptr == conf->result.buffer) {
	conf->result.buffer = NULL;
	conf->result.buflen = 0;
  }
}

/*
 * In ISO C90, if PTR is NULL realloc(3) behaves like malloc(3).
 * In some implementations of realloc(3), (PTR != NULL && NEED == 0) is
 * equivalent to free(PTR).
 * The actual behaviour here, for better or worse, depends on the underlying
 * implmentations.
 * XXX behaviour of the odd cases may need to be unified
 */
static void *
custom_realloc(Compress_conf *conf, void *ptr, size_t last_alloc, size_t need)
{
  void *new_ptr;

  if (conf->realloc != NULL)
	new_ptr = (conf->realloc)(ptr, need);
  else if (conf->alloc == NULL)
	new_ptr = realloc(ptr, need);
  else {
	/* Do our own realloc() here. */
	new_ptr = conf->alloc(need);
	if (new_ptr != NULL && ptr != NULL && last_alloc) {
	  memcpy(new_ptr, ptr, last_alloc);
	  if (conf->dealloc == NULL)
		free(ptr);
	  else
		conf->dealloc(ptr);
	}
  }

 if (new_ptr == NULL) {
   DACS_ABORT_UTIL(("Allocation error, last_alloc=%lu, need=%lu",
					last_alloc, need));
	/*NOTREACHED*/
  }

 conf->result.buffer = new_ptr;
 conf->result.buflen = need;

 /* Note that we do not increment ALLOC_COUNT and DEALLOC_COUNT. */
 conf->result.realloc_count++;
 conf->result.total_alloc += (need - last_alloc);

  return(new_ptr);
}

/************************************************************************/

#ifndef MEM_ALIGNMENT_TYPE
typedef union {
  int a;
  short b;
  char c[1];
  long d;
  long long ll;
  float f;
  double g;
  long double h;
  void *i;
} MEM_ALIGNMENT_TYPE;
#endif

enum {
  ALLOC_ALIGNMENT = (sizeof(MEM_ALIGNMENT_TYPE)),
  ALLOC_ALIGN_MASK = (ALLOC_ALIGNMENT - 1)
};

/*
 * Return the number of bytes to allocate for an object of NEED bytes so that
 * the address after the allocation is aligned to an address evenly divisible
 * by ALIGNMENT, where ALIGNMENT is a power of 2.
 * Example: align_alloc(18, 8) returns 24, which means that if ADDR=malloc(24)
 * is done, (ADDR + 24) will be aligned on an 8-byte boundary and 6 bytes
 * of the 24-byte allocation is unused.
 */
static size_t
align_alloc(size_t need, size_t alignment)
{
  size_t aligned_size;

  aligned_size = (need + alignment - 1) & (~ (alignment - 1));

  return(aligned_size);
}

static size_t
max_align_alloc(size_t need)
{

  return(align_alloc(need, ALLOC_ALIGNMENT));
}

/*
 * Helper function to compute pointers and sizes when compressing to account
 * for a header (additional buffer space) that appears immediately before
 * the start of the output of a compression function.  The header is allocated
 * for application-specific purposes.
 */
static size_t
adjust_for_output_header(Compress_conf *conf, unsigned char *outbuf,
						 size_t outlen, size_t *allocated)
{
  unsigned char *ptr;
  size_t avail_out, header_need, total_need;

  if (conf->header_alloc)
	header_need = conf->header_alloc;
  else
	header_need = 0;

  total_need = header_need + outlen;

  if (outbuf == NULL) {
	ptr = custom_alloc(conf, total_need);
	memset(ptr, 0, total_need);
	if (allocated != NULL)
	  *allocated = total_need;
  }
  else {
	if (total_need > outlen)
	  return(0);
	ptr = outbuf;
	if (allocated != NULL)
	  *allocated = 0;
  }

  conf->result.header_start = ptr;
  conf->result.data_start = ptr + header_need;
  avail_out = total_need - header_need;

  conf->header_status = COMPRESS_HAVE_HEADER;

  return(avail_out);
}
				  
/*
 * Helper function to compute pointers and sizes when decompressing to account
 * for a header (additional buffer space) that appears immediately before
 * the start of the output of a compression function.  The header is allocated
 * for application-specific purposes.
 */
static size_t
adjust_for_input_header(Compress_conf *conf, unsigned char *inbuf,
						size_t inlen)
{
  size_t avail_in;

  conf->result.header_start = inbuf;
  conf->result.data_start = inbuf + conf->header_alloc;
  avail_in = inlen - conf->header_alloc;

  return(avail_in);
}
				  
/************************************************************************/
#ifdef HAVE_LIBLZ4
/* https://github.com/lz4/lz4 */
/* https://en.wikipedia.org/wiki/LZ4_(compression_algorithm) */

#include <lz4.h>

/*
 * XXX not needed yet
 */
#endif

/************************************************************************/
/*
 * XXX Brotli compression does not work, so do not use until it is fixed.
 * Refer to run_test3() where brotli_compress()/brotli_decompress()
 * do not work with some inputs.
 * I don't know if this is a usage issue, a bug in this code, or a bug in
 * the brotli implementation or "documentation".
 * With all due respect, the brotli C library API needs some work.
 */
#undef HAVE_LIBBROTLI

#ifdef HAVE_LIBBROTLI
/*
 * Brötli means ‘small bread’ in Swiss German.
 *
 * https://github.com/google/brotli
 * https://en.wikipedia.org/wiki/Brotli
 * https://tools.ietf.org/html/rfc7932
 * https://opensource.googleblog.com/2015/09/introducing-brotli-new-compression.html
 *
 *
 * XXX
 * Using only default parameters:
 * a) for smaller files, it usually seems to squeeze a few percent better
 * than zstd/gzip/bzip2/xz:
 * -rw-r--r--  1 sammy    wheel  127034 Mar 28 16:48 out.c
 * -rw-r--r--  1 sammy    wheel   30975 Mar 28 16:50 out.zstd
 * -rw-r--r--  1 sammy    wheel   28874 Mar 28 16:51 out.gz
 * -rw-r--r--  1 sammy    wheel   25264 Mar 28 16:48 out.xz
 * -rw-r--r--  1 sammy    wheel   24331 Mar 28 16:51 out.bz2
 * -rw-r--r--  1 sammy    wheel   23856 Mar 28 16:48 out.br
 *
 * b) but for larger files, meh...
 * -rw-r--r--   1 sammy   wheel  35737600 Mar 28 10:41 brotli-1.0.7.tar
 * -rw-r--r--   1 sammy   wheel  24191046 Mar 28 16:55 brotli-1.0.7.tzstd
 * -rw-r--r--   1 sammy   wheel  23679142 Mar 28 16:56 brotli-1.0.7.tgz
 * -rw-r--r--   1 sammy   wheel  23059039 Mar 28 16:56 brotli-1.0.7.tbz
 * -rw-r--r--   1 sammy   wheel  20709297 Mar 28 16:42 brotli-1.0.7.tbr
 * -rw-r--r--   1 sammy   wheel  19999836 Mar 28 16:39 brotli-1.0.7.txz
 * and compression was much slower than for any of the others.
 */

#include <brotli/encode.h>
#include <brotli/decode.h>

static const char *
brotli_error_string(Compress_op op, size_t error_code)
{
  const char *msg;

  if (op == COMPRESS_UNPACK) {
	BrotliDecoderErrorCode c;

	c = (BrotliDecoderErrorCode) error_code;
	msg = BrotliDecoderErrorString(c);
  }
  else if (op == COMPRESS_PACK) {

	/* XXX there is no BrotliEncoderErrorString()... */
	msg = "Unknown encoding error";
  }
  else
	msg = NULL;

  return(msg);
}

static const char *
brotli_version_string(void)
{
  static char *version_string = NULL;

  if (version_string == NULL) {
	int major, minor, patch;
	uint32_t brotli_version;

	brotli_version = BrotliEncoderVersion();
	major = brotli_version >> 24;
	minor = (brotli_version >> 12) & 0xFFF;
	patch = brotli_version & 0xFFF;

	version_string = ds_xprintf("%d.%d.%d", major, minor, patch);
  }

  return(version_string);
}

/*
 * Memory allocation callback.
 */
static void *
brotli_alloc(void *opaque, size_t need)
{
  void *ptr;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  ptr = custom_alloc(conf, need);

  return(ptr);
}

/*
 * Memory deallocation callback.
 */
static void
brotli_dealloc(void *opaque, void *ptr)
{
  Compress_conf *conf;

  if (ptr == NULL)
	return;

  conf = (Compress_conf *) opaque;
  custom_dealloc(conf, ptr);
}

/*
 * Memory reallocation callback.
 */
static void *
brotli_realloc(void *opaque, void *ptr, size_t last_alloc, size_t need)
{
  void *new_ptr;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  new_ptr = custom_realloc(conf, ptr, last_alloc, need);

  return(new_ptr);
}

/*
 * Initialize for Brotli compression with configuration CONF.
 * Return a new context or NULL if an error occurs.
 */
static void *
brotli_compress_open(Compress_conf *conf)
{
  BrotliEncoderState *s;

  s = BrotliEncoderCreateInstance(brotli_alloc, brotli_dealloc, conf);

  return(s);
}

typedef struct Brotli_io {
  size_t avail_in;
  const uint8_t *next_in;
  size_t avail_out;
  uint8_t *next_out;
  size_t total_out;
} Brotli_io;

/*
 * Add INLEN bytes from IN to the Brotli compression stream,
 * according to context CONF, putting up to OUTLEN bytes into OUT.
 *
 * Return the number of bytes read from IN, or -1 if an
 * error occurred.  The actual number of output bytes is put in BUFLEN.
 */
static int
brotli_compress_update(Compress_conf *conf, void *in, size_t inlen,
					   unsigned char *out, size_t outlen, size_t *nwritten)
{
  int rc;
  Brotli_io io;

  io.next_in = in;
  io.avail_in = inlen;
  if (conf->header_status == COMPRESS_NEED_HEADER) {
	if ((io.avail_out = adjust_for_output_header(conf, out, outlen, NULL)) == 0)
	  return(-1);

	io.next_out = conf->result.data_start;
	/* Below, BUFCOUNT will be set to a size that includes this header. */
  }
  else {
	io.next_out = out;
	io.avail_out = outlen;
  }
  io.total_out = 0;

  if (nwritten != NULL)
	*nwritten = 0;

  rc = BrotliEncoderCompressStream(conf->subctx, BROTLI_OPERATION_PROCESS,
								   &io.avail_in, &io.next_in,
								   &io.avail_out, &io.next_out, &io.total_out);
  if (rc == 0) {
	conf->result.error_code = (long) rc;
	conf->result.error_mesg = brotli_error_string(COMPRESS_PACK, rc);
	return(-1);
  }

  if (nwritten != NULL)
	*nwritten = outlen - io.avail_out;

  return(inlen - io.avail_in);
}

/*
 * There is no more input. Try to finish compressing through Brotli,
 * putting up to OUTLEN bytes into OUT.
 *
 * Set NWRITTEN to the number of bytes copied into OUT and return 0 if
 * compression is complete, 1 if this function must be called again, or -1
 * if an error occurs.
 */
static int
brotli_compress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					   size_t *nwritten)
{
  int rc;
  Brotli_io io;

  io.avail_in = 0;
  io.next_in = NULL;
  io.next_out = out;
  io.avail_out = outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  if (conf->is_finished)
	return(0);

  rc = BrotliEncoderCompressStream(conf->subctx, BROTLI_OPERATION_FINISH,
								   &io.avail_in, &io.next_in,
								   &io.avail_out, &io.next_out, &io.total_out);
  if (rc == 0) {
	conf->result.error_code = (long) rc;
	conf->result.error_mesg = brotli_error_string(COMPRESS_PACK, rc);
	return(-1);
  }

  if (BrotliEncoderIsFinished(conf->subctx)) {
	rc = 0;
	conf->is_finished = 1;
  }
  else
	rc = 1;

  if (nwritten != NULL)
	*nwritten = outlen - io.avail_out;

  return(rc);
}

static void
brotli_compress_close(Compress_conf *conf)
{

  if (conf->subctx != NULL)
	BrotliEncoderDestroyInstance(conf->subctx);
}

/*
 * Compress INLEN bytes from IN, setting OUT to the allocated buffer containing
 * OUTLEN compressed bytes.
 * Adjust the output if an application header was requested.
 */
static int
brotli_compress(Compress_conf *conf, void *in, size_t inlen,
				unsigned char **out, size_t *nwritten)
{
  unsigned char *dest;
  size_t destLen, encoded_size, rc, nalloc, need;
  BROTLI_BOOL st;
  BrotliEncoderState *s;
  Brotli_io io;

  if (in == NULL)
	return(-1);

  if ((destLen = BrotliEncoderMaxCompressedSize(inlen)) == 0)
	return(-1);

  s = BrotliEncoderCreateInstance(brotli_alloc, brotli_dealloc, conf);
  if (s == NULL)
	return(-1);

  io.avail_in = inlen;
  io.next_in = in;

  if (conf->header_status == COMPRESS_NEED_HEADER) {
	need = destLen;
	if ((io.avail_out = adjust_for_output_header(conf, NULL, need, NULL)) == 0)
	  return(-1);

	dest = conf->result.header_start;
	io.next_out = conf->result.data_start;
  }
  else {
	dest = brotli_alloc(conf, destLen);
	io.next_out = dest;
	io.avail_out = destLen;
  }

  if (nwritten != NULL)
	*nwritten = 0;

  st = BrotliEncoderCompress(BROTLI_MAX_QUALITY, BROTLI_MAX_WINDOW_BITS,
							 BROTLI_MODE_GENERIC, io.avail_in, io.next_in,
							 &encoded_size, io.next_out);

  BrotliEncoderDestroyInstance(s);

  if (st != BROTLI_TRUE)
	return(-1);

  if (out != NULL)
	*out = dest;
  else
	brotli_dealloc(conf, dest);

  if (nwritten != NULL)
	*nwritten = encoded_size + conf->header_alloc;

  return(0);
}

/*
 * Initialize for Brotli decompression with configuration CONF.
 * Return a new context or NULL if an error occurs.
 */
static void *
brotli_decompress_open(Compress_conf *conf)
{
  BrotliDecoderState *s;

  s = BrotliDecoderCreateInstance(brotli_alloc, brotli_dealloc, conf);

  return(s);
}

/*
 * Add INLEN bytes from IN to the Brotli decompression stream,
 * according to context CONF, putting up to OUTLEN bytes into OUT.
 *
 * Return the number of bytes read from IN, or -1 if an error occurred.
 * Set BUFLEN to the number of bytes copied into OUT.
 */
static int
brotli_decompress_update(Compress_conf *conf, void *in, size_t inlen,
						 unsigned char *out, size_t outlen, size_t *nwritten)
{
  int nread, try_more;
  Brotli_io io;
  BrotliDecoderState *s;
  BrotliDecoderResult result;

  s = (BrotliDecoderState *) conf->subctx;
  if (conf->header_status == COMPRESS_HAVE_HEADER) {
    io.avail_in = adjust_for_input_header(conf, in, inlen);
    io.next_in = conf->result.data_start;
  }
  else {
    io.avail_in = inlen;
    io.next_in = in;
  }

  io.next_out = out;
  io.avail_out = outlen;
  io.total_out = 0;

  if (nwritten != NULL)
    *nwritten = 0;

  try_more = 1;
  while (try_more) {
    result = BrotliDecoderDecompressStream(s, &io.avail_in, &io.next_in,
										   &io.avail_out, &io.next_out,
										   &io.total_out);
	nread = (int) (inlen - io.avail_in);
	switch (result) {
	case BROTLI_DECODER_RESULT_NEEDS_MORE_INPUT:
	  try_more = 0;
	  break;

    case BROTLI_DECODER_RESULT_NEEDS_MORE_OUTPUT:
	  if (io.avail_out == 0)
		try_more = 0;
	  break;

    case BROTLI_DECODER_RESULT_SUCCESS:
	  try_more = 0;
	  break;

	default:
	  nread = -1;
	  try_more = 0;
	  break;
    }
  }

  if (nwritten != NULL)
	*nwritten = io.total_out;

  return(nread);
}

static int
brotli_decompress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
						 size_t *nwritten)
{
  int rc, try_more;
  Brotli_io io;
  BrotliDecoderResult result;
  BrotliDecoderState *s;

  io.avail_in = 0;
  io.next_in = NULL;
  io.next_out = out;
  io.avail_out = outlen;

  if (nwritten != NULL)
    *nwritten = 0;

  if (conf->is_finished)
    return(0);

  s = (BrotliDecoderState *) conf->subctx;
  try_more = 1;
  while (try_more) {
    result = BrotliDecoderDecompressStream(s, &io.avail_in, &io.next_in,
										   &io.avail_out, &io.next_out,
										   &io.total_out);

	switch (result) {
	case BROTLI_DECODER_RESULT_NEEDS_MORE_INPUT:
	  rc = -1;
	  try_more = 0;
	  break;

	case BROTLI_DECODER_RESULT_NEEDS_MORE_OUTPUT:
	  if (io.avail_out == 0) {
		rc = 1;
		try_more = 0;
	  }
	  break;

    case BROTLI_DECODER_RESULT_SUCCESS:
	  rc = 0;
	  try_more = 0;
	  conf->is_finished = 1;
	  break;

	default:
	  rc = -1;
	  try_more = 0;
	  break;
    }
  }

  if (nwritten != NULL)
	*nwritten = outlen - io.avail_out;

  return(rc);
}

static void
brotli_decompress_close(Compress_conf *conf)
{

  if (conf->subctx != NULL)
	BrotliDecoderDestroyInstance(conf->subctx);
}

/*
 * Decompress INLEN bytes from IN, setting OUT to the allocated buffer
 * containing OUTLEN decompressed bytes.
 */
static int
brotli_decompress(Compress_conf *conf, void *in, size_t inlen,
				  unsigned char **out, size_t *outlen)
{
  int nread, try_more;
  unsigned char *dest;
  size_t grow_len, hint, nalloc, need, nused;
  Brotli_io io;
  BrotliDecoderState *s;
  BrotliDecoderResult result;

  if (in == NULL || inlen == 0)
	return(-1);

  s = BrotliDecoderCreateInstance(brotli_alloc, brotli_dealloc, conf);
  if (s == NULL)
	return(-1);

  /* Any adjustments for a header were done by compress_unpack_x(). */
  io.avail_in = inlen;
  io.next_in = in;

  if (conf->expected_output_size)
	need = grow_len = conf->expected_output_size;
  else {
	grow_len = 0;
	need = inlen;
  }

  nalloc = need;
  dest = brotli_alloc(conf, nalloc);
  io.next_out = dest;
  io.avail_out = nalloc;
  io.total_out = 0;

  while (1) {
	char *errstr;
	size_t last_alloc;

	/*
	 * XXX I can't get BrotliDecoderDecompress() or this to work...
	 * they are both returning BROTLI_DECODER_RESULT_ERROR
	 */
    result = BrotliDecoderDecompressStream(s, &io.avail_in, &io.next_in,
										   &io.avail_out, &io.next_out,
										   &io.total_out);
	if (result == BROTLI_DECODER_RESULT_SUCCESS)
	  break;

	if (result == BROTLI_DECODER_RESULT_NEEDS_MORE_INPUT) {
	err:
	  if (dest != NULL)
		free(dest);
	  BrotliDecoderDestroyInstance(s);
	  
	  return(-1);
	}

	if (result == BROTLI_DECODER_RESULT_ERROR) {
	  const char *err_msg;
	  BrotliDecoderErrorCode err_code;

	  err_code = BrotliDecoderGetErrorCode(s);
	  err_msg = BrotliDecoderErrorString(err_code);
	  if (err_msg != NULL)
		fprintf(stderr, "brotli_decompress: err_msg=\"%s\"\n", err_msg);
	  else
		fprintf(stderr, "brotli_decompress: unknown error err_code=%d\n",
				err_code);
	  goto err;
	}

	if (result != BROTLI_DECODER_RESULT_NEEDS_MORE_OUTPUT)
	  goto err;

	/* Allocate more memory. */
	last_alloc = nalloc;
	if (grow_len)
	  nalloc += grow_len;
	else
	  nalloc *= 2;

	dest = brotli_realloc(conf, dest, last_alloc, nalloc);
	io.next_out = dest + io.total_out;
	io.avail_out = nalloc - last_alloc;
  }

  BrotliDecoderDestroyInstance(s);

  if (out != NULL)
	*out = dest;

  if (outlen != NULL)
	*outlen = io.total_out;

  return(0);
}
#endif

/************************************************************************/
#ifdef HAVE_LIBZSTD

/*
 * https://facebook.github.io/zstd/zstd_manual.html
 * https://github.com/facebook/zstd
 * https://tools.ietf.org/html/rfc8478
 */

/* Required to reference "advanced experimental functions"... */
#define ZSTD_STATIC_LINKING_ONLY

#include <zstd.h>

static const char *
zstd_error_string(size_t error_code)
{
  const char *mesg;
  
  mesg = ZSTD_getErrorName(error_code);

  return(mesg);
}

/*
 * Memory allocation callback.
 */
static void *
zstd_alloc(void *opaque, size_t need)
{
  void *ptr;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  ptr = custom_alloc(conf, need);

  return(ptr);
}

/*
 * Memory deallocation callback.
 */
static void
zstd_dealloc(void *opaque, void *ptr)
{
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  custom_dealloc(conf, ptr);
}

/*
 * Memory reallocation callback.
 */
static void *
zstd_realloc(void *opaque, void *ptr, size_t last_alloc, size_t need)
{
  void *new_ptr;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  new_ptr = custom_realloc(conf, ptr, last_alloc, need);

  return(new_ptr);
}

/*
 * Initialize for Zstandard compression with configuration CONF.
 * Return a new context or NULL if an error occurs.
 */
static void *
zstd_compress_open(Compress_conf *conf)
{
  size_t rc;
  ZSTD_CCtx *zctx;
  ZSTD_customMem customMem;

  customMem.customAlloc = zstd_alloc;
  customMem.customFree = zstd_dealloc;
  customMem.opaque = conf;
  if ((zctx = ZSTD_createCCtx_advanced(customMem)) == NULL)
	return(NULL);

  rc = ZSTD_CCtx_setParameter(zctx, ZSTD_c_strategy, ZSTD_btultra2);
  if (ZSTD_isError(rc)) {
	conf->result.error_code = (long) rc;
	conf->result.error_mesg = zstd_error_string(rc);
	return(NULL);
  }

  return(zctx);
}

/*
 * Add INLEN bytes from IN to the Zstandard compression stream,
 * according to context CONF, putting up to OUTLEN bytes into OUT.
 *
 * Return the number of bytes read from IN, or -1 if an
 * error occurred.  The actual number of output bytes is put in BUFLEN.
 */
static int
zstd_compress_update(Compress_conf *conf, void *in, size_t inlen,
					 unsigned char *out, size_t outlen, size_t *nwritten)
{
  size_t rc;
  ZSTD_outBuffer output;
  ZSTD_inBuffer input;

  if (conf->header_status == COMPRESS_NEED_HEADER) {
	if ((output.size = adjust_for_output_header(conf, out, outlen, NULL)) == 0)
	  return(-1);

	output.dst = conf->result.data_start;
	output.pos = 0;
	/* Below, BUFCOUNT will be set to a size that includes this header. */
  }
  else {
	output.dst = out;
	output.size = outlen;
	output.pos = 0;
  }

  input.src = in;
  input.size = inlen;
  input.pos = 0;

  if (nwritten != NULL)
	*nwritten = 0;

  while (1) {
	rc = ZSTD_compressStream2(conf->subctx, &output, &input, ZSTD_e_continue);
	if (ZSTD_isError(rc)) {
	  conf->result.error_code = (long) rc;
	  conf->result.error_mesg = zstd_error_string(rc);
	  return(-1);
	}

	/*
	 * The compressor has processed all of the input buffer and needs more.
	 * Return to have the caller empty the output buffer, if necessary,
	 * and call here again with new input.
	 */
	if (input.pos == input.size) {
	  rc = input.pos;
	  break;
	}

	/*
	 * The compressor has filled the output buffer but not processed all of
	 * the input buffer.
	 */
	if (output.pos == output.size) {
	  rc = input.pos;
	  break;
	}
  }

  if (nwritten != NULL)
	*nwritten = output.pos;

  return((int) rc);
}

/*
 * There is no more input. Try to finish compressing through Zstandard,
 * putting up to OUTLEN bytes into OUT.
 *
 * Set OUTLEN to the number of bytes copied into OUT and return 0 if
 * compression is complete, 1 if this function must be called again, or -1
 * if an error occurs.
 */
static int
zstd_compress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					 size_t *bufcount)
{
  size_t rc;
  ZSTD_outBuffer output;
  ZSTD_inBuffer input;

  output.dst = out;
  output.size = outlen;
  output.pos = 0;

  input.src = NULL;
  input.size = 0;
  input.pos = 0;

  if (bufcount != NULL)
	*bufcount = 0;

  if (conf->is_finished)
	return(0);

  rc = ZSTD_compressStream2(conf->subctx, &output, &input, ZSTD_e_end);
  if (ZSTD_isError(rc)) {
	conf->result.error_code = (long) rc;
	conf->result.error_mesg = zstd_error_string(rc);
	return(-1);
  }

  if (bufcount != NULL)
	*bufcount = output.pos;

  if (rc == 0) {
	conf->is_finished = 1;
	return(0);
  }

  return(1);
}

static void
zstd_compress_close(Compress_conf *conf)
{

  if (conf->subctx != NULL)
	ZSTD_freeCCtx(conf->subctx);
}

/*
 * Compress INLEN bytes from IN, setting OUT to the allocated buffer containing
 * OUTLEN compressed bytes.
 * Adjust the output if an application header was requested.
 */
static int
zstd_compress(Compress_conf *conf, void *in, size_t inlen,
			  unsigned char **out, size_t *outlen)
{
  unsigned char *dest;
  size_t rc, nalloc, need;
  ZSTD_CCtx *zctx;
  ZSTD_customMem customMem;
  ZSTD_outBuffer output;
  ZSTD_inBuffer input;

  if (in == NULL)
	return(-1);

  customMem.customAlloc = zstd_alloc;
  customMem.customFree = zstd_dealloc;
  customMem.opaque = conf;
  if ((zctx = ZSTD_createCCtx_advanced(customMem)) == NULL) {
	return(-1);
  }

  rc = ZSTD_CCtx_setParameter(zctx, ZSTD_c_strategy, ZSTD_btultra2);
  if (ZSTD_isError(rc)) {
  err:
	conf->result.error_code = (long) rc;
	conf->result.error_mesg = zstd_error_string(rc);
	return(-1);
  }

  input.src = in;
  input.size = inlen;
  input.pos = 0;

  if ((need = ZSTD_compressBound(inlen)) == ZSTD_CONTENTSIZE_ERROR)
	return(-1);

  if (conf->header_status == COMPRESS_NEED_HEADER) {
	if ((output.size = adjust_for_output_header(conf, NULL, need, &nalloc)) == 0)
	  return(-1);
	dest = conf->result.header_start;
	output.dst = conf->result.data_start;
	output.pos = 0;
  }
  else {
	dest = zstd_alloc(conf, need);
	nalloc = need;
	output.dst = dest;
	output.size = nalloc;
	output.pos = 0;
  }

  while (1) {
	size_t last_alloc;

	rc = ZSTD_compressStream2(zctx, &output, &input, ZSTD_e_continue);
	if (ZSTD_isError(rc))
	  goto err;

	if (rc == 0)
	  break;

	if (input.pos != input.size) {
	  /* XXX allocate more */
	  last_alloc = nalloc;
	  need = nalloc * 2;
	  dest = zstd_realloc(conf, dest, last_alloc, need);
	  output.size = need - nalloc;
	  nalloc = need;
	}
  }

  if (input.pos == input.size) {
	do {
	  rc = ZSTD_compressStream2(zctx, &output, &input, ZSTD_e_end);
	  if (ZSTD_isError(rc))
		goto err;
	} while (rc > 0);
  }

  if (out != NULL)
	*out = dest;
  else
	zstd_dealloc(conf, dest);

  if (outlen != NULL)
	*outlen = output.pos + conf->header_alloc;

  return(0);
}

/*
 * Initialize for Zstandard decompression with configuration CONF.
 * Return a new context or NULL if an error occurs.
 */
static void *
zstd_decompress_open(Compress_conf *conf)
{
  ZSTD_DCtx *zctx;
  ZSTD_customMem customMem;

  customMem.customAlloc = zstd_alloc;
  customMem.customFree = zstd_dealloc;
  customMem.opaque = conf;
  if ((zctx = ZSTD_createDCtx_advanced(customMem)) == NULL)
	return(NULL);

  return(zctx);
}

/*
 * Add INLEN bytes from IN to the Zstandard decompression stream,
 * according to context CONF, putting up to OUTLEN bytes into OUT.
 *
 * Return the number of bytes read from IN, or -1 if an error occurred.
 * Set BUFLEN to the number of bytes copied into OUT.
 */
static int
zstd_decompress_update(Compress_conf *conf, void *in, size_t inlen,
					   unsigned char *out, size_t outlen, size_t *nwritten)
{
  size_t rc;
  ZSTD_outBuffer output;
  ZSTD_inBuffer input;

  if (conf->header_status == COMPRESS_HAVE_HEADER) {
	input.size = adjust_for_input_header(conf, in, inlen);
	input.src = conf->result.data_start;
	input.pos = 0;
  }
  else {
	input.src = in;
	input.size = inlen;
	input.pos = 0;
  }

  output.dst = out;
  output.size = outlen;
  output.pos = 0;

  if (nwritten != NULL)
	*nwritten = 0;

  while (1) {
	rc = ZSTD_decompressStream(conf->subctx, &output, &input);
	if (ZSTD_isError(rc)) {
	  conf->result.error_code = (long) rc;
	  conf->result.error_mesg = zstd_error_string(rc);
	  return(-1);
	}

	if (rc == 0) {
	  conf->is_finished = 1;
	  rc = input.pos;
	  break;
	}

	if (input.pos == input.size) {
	  rc = input.pos;
	  break;
	}

	if (output.pos == output.size) {
	  rc = input.pos;
	  break;
	}
  }

  if (nwritten != NULL)
	*nwritten = output.pos;

  return((int) rc);
}

static int
zstd_decompress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					   size_t *nwritten)
{
  size_t rc;
  ZSTD_outBuffer output;
  ZSTD_inBuffer input;

  output.dst = out;
  output.size = outlen;
  output.pos = 0;

  input.src = NULL;
  input.size = 0;
  input.pos = 0;

  if (nwritten != NULL)
	*nwritten = 0;

  if (conf->is_finished)
	return(0);

  rc = ZSTD_decompressStream(conf->subctx, &output, &input);
  if (ZSTD_isError(rc)) {
	conf->result.error_code = (long) rc;
	conf->result.error_mesg = zstd_error_string(rc);
	return(-1);
  }

  if (nwritten != NULL)
	*nwritten = output.pos;

  if (rc == 0) {
	conf->is_finished = 1;
	return(0);
  }

  return(1);
}

static void
zstd_decompress_close(Compress_conf *conf)
{

  if (conf->subctx != NULL)
	ZSTD_freeDCtx(conf->subctx);
}

/*
 * Decompress INLEN bytes from IN, setting OUT to the allocated buffer
 * containing OUTLEN decompressed bytes.
 */
static int
zstd_decompress(Compress_conf *conf, void *in, size_t inlen,
				unsigned char **out, size_t *outlen)
{
  unsigned char *dest;
  unsigned long long alloc_upper_bound;
  size_t grow_len, hint, nalloc, need;
  ZSTD_DCtx *zctx;
  ZSTD_customMem customMem;
  ZSTD_outBuffer output;
  ZSTD_inBuffer input;

  if (in == NULL)
	return(-1);

  customMem.customAlloc = zstd_alloc;
  customMem.customFree = zstd_dealloc;
  customMem.opaque = conf;
  if ((zctx = ZSTD_createDCtx_advanced(customMem)) == NULL)
	return(-1);

  if (conf->expected_output_size)
	grow_len = max_align_alloc(conf->expected_output_size);
  else
	grow_len = max_align_alloc(inlen);

  input.src = in;
  input.size = inlen;
  input.pos = 0;

  alloc_upper_bound = ZSTD_decompressBound(in, inlen);
  if (alloc_upper_bound == ZSTD_CONTENTSIZE_ERROR) {
	/* IN probably points to something invalid. */
	return(-1);
  }

  need = (size_t) alloc_upper_bound;
  dest = zstd_alloc(conf, need);
  nalloc = need;

  output.dst = dest;
  output.size = nalloc;
  output.pos = 0;

  while (1) {
	size_t last_alloc;

	hint = ZSTD_decompressStream(zctx, &output, &input);
	if (ZSTD_isError(hint)) {
	err:
	  conf->result.error_code = (long) hint;
	  conf->result.error_mesg = zstd_error_string(hint);
	  return(-1);
	}
	if (hint == 0)
	  break;

	/* XXX allocate more */
	last_alloc = need;
	need += grow_len;
	dest = zstd_realloc(conf, dest, last_alloc, need);
	output.size = need - nalloc;
	nalloc = need;
  }

  if (output.pos == output.size) {
	hint = ZSTD_decompressStream(zctx, &output, &input);
	if (ZSTD_isError(hint))
	  goto err;
	if (hint != 0)
	  return(-1);
  }

  if (out != NULL)
	*out = dest;
  else
	zstd_dealloc(conf, dest);

  if (outlen != NULL)
	*outlen = output.pos;

  return(0);
}
#endif

/************************************************************************/

#ifdef HAVE_LIBLZMA

/*
 * LZMA/LZMA2 is used by, and distributed with, the xz(1) utilities.
 * https://tukaani.org/xz/
 * https://www.7-zip.org/sdk.html
 *
 * There does not appear to be any formal documentation for the API,
 * so consult header files in the xz distribution (e.g., under its
 * src/liblzma/api/lzma) for details:
 * https://fossies.org/linux/xz/src/liblzma/api/lzma/base.h
 * https://fossies.org/linux/xz/src/liblzma/api/lzma/container.h
 */

#include <stddef.h>
#include <stdint.h>
#include <lzma.h>

/*
 * The library does not appear to provide a string equivalent for ERROR_CODE...
 */
static MAYBE_UNUSED const char *
lzma_error_string(lzma_ret error_code)
{
  const char *mesg;

  /*
   * XXX
   * this is necessarily very dependent on api/lzma/base.h, which defines
   * not only true error codes, but all return values.
   */
  switch (error_code) {
  case LZMA_OK:
	mesg = "OK"; break;
  case LZMA_STREAM_END:
	mesg = "Stream end"; break;
  case LZMA_NO_CHECK:
	mesg = "No check"; break;
  case LZMA_UNSUPPORTED_CHECK:
	mesg = "Unsupported check"; break;
  case LZMA_GET_CHECK:
	mesg = "Get check"; break;
  case LZMA_MEM_ERROR:
	mesg = "Mem error"; break;
  case LZMA_MEMLIMIT_ERROR:
	mesg = "Memlimit error"; break;
  case LZMA_FORMAT_ERROR:
	mesg = "Format error"; break;
  case LZMA_OPTIONS_ERROR:
	mesg = "Options error"; break;
  case LZMA_DATA_ERROR:
	mesg = "Data error"; break;
  case LZMA_BUF_ERROR:
	mesg = "Buf error"; break;
  case LZMA_PROG_ERROR:
	mesg = "Prog error"; break;
  default:
	mesg = "Unknown error"; break;
  }

  return(mesg);
}


static void *
lzma_alloc(void *opaque, size_t need)
{
  void *ptr;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  ptr = custom_alloc(conf, need);

  return(ptr);
}

/*
 * Memory deallocation callback.
 */
static void
lzma_dealloc(void *opaque, void *ptr)
{
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  custom_dealloc(conf, ptr);
}

static void *
lzma_realloc(void *opaque, void *ptr, size_t last_alloc, size_t need)
{
  void *new_ptr;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  new_ptr = custom_realloc(conf, ptr, last_alloc, need);

  return(new_ptr);
}

/*
 * Initialization for LZMA2.
 * See doc/examples/03_compress_custom.c in the xz distribution.
 */
static void *
lzma_compress_open(Compress_conf *conf)
{
  lzma_options_lzma opt_lzma2;
  lzma_stream *strm;
  lzma_filter filters[] = {
    { .id = LZMA_FILTER_X86, .options = NULL },
	{ .id = LZMA_FILTER_LZMA2, .options = &opt_lzma2 },
	{ .id = LZMA_VLI_UNKNOWN, .options = NULL },
  };
  lzma_ret ret;
  static lzma_stream strm_init = LZMA_STREAM_INIT;

  strm = (lzma_stream *) ALLOC(lzma_stream);
  *strm = strm_init;
  if (lzma_lzma_preset(&opt_lzma2, 9))
	return(NULL);

  ret = lzma_stream_encoder(strm, filters, LZMA_CHECK_CRC64);
  if (ret != LZMA_OK)
	return(NULL);

  return(strm);
}

static int
lzma_compress_update(Compress_conf *conf, void *in, size_t inlen,
					 unsigned char *out, size_t outlen, size_t *nwritten)
{
  int nread;
  lzma_ret rc;
  lzma_stream *strm;

  strm = (lzma_stream *) conf->subctx;
  strm->avail_in = inlen;
  strm->next_in = in;

  if (conf->header_status == COMPRESS_NEED_HEADER) {
	if ((strm->avail_out = adjust_for_output_header(conf, out, outlen, NULL))
		== 0)
	  return(-1);

	strm->next_out = conf->result.data_start;
	/* Below, BUFLEN will be set to a size that includes this header. */
  }
  else {
	strm->next_out = out;
	strm->avail_out = outlen;
  }

  if (nwritten != NULL)
	*nwritten = 0;

  rc = lzma_code(strm, LZMA_RUN);
  if (rc == LZMA_STREAM_END) {
	if (strm->avail_in == 0)
	  conf->is_finished = 1;
  }
  else if (rc != LZMA_OK)
	return(-1);

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  nread = inlen - strm->avail_in;

  return(nread);
}

static int
lzma_compress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					 size_t *nwritten)
{
  lzma_ret rc;
  lzma_stream *strm;

  strm = (lzma_stream *) conf->subctx;
  strm->avail_in = 0;
  strm->next_in = NULL;
  strm->next_out = out;
  strm->avail_out = outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  if (conf->is_finished)
	return(0);

  rc = lzma_code(strm, LZMA_FINISH);
  if (rc == LZMA_STREAM_END) {
	if (strm->avail_out == outlen)
	  conf->is_finished = 1;
  }
  else if (rc != LZMA_OK)
	return(-1);

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  return(conf->is_finished ? 0 : 1);
}

static void
lzma_compress_close(Compress_conf *conf)
{

  if (conf->subctx != NULL)
	lzma_end(conf->subctx);
}

/*
 * Compress INLEN bytes from IN, setting OUT to the allocated buffer containing
 * OUTLEN compressed bytes (which includes any application header).
 */
int
lzma_compress(Compress_conf *conf, void *in, size_t inlen,
			  unsigned char **out, size_t *outlen)
{
  unsigned char *dest;
  size_t need;
  lzma_ret rc;
  lzma_stream s = LZMA_STREAM_INIT;

  dest = NULL;

  rc = lzma_easy_encoder(&s, 9, LZMA_CHECK_CRC32);
  if (rc != LZMA_OK) {
  err:
	if (dest != NULL)
	  lzma_dealloc(conf, dest);
	return(-1);
  }

  s.next_in = in;
  s.avail_in = inlen;

  if (conf->header_status == COMPRESS_NEED_HEADER) {
	need = inlen * 2;
	if ((s.avail_out = adjust_for_output_header(conf, dest, need, NULL)) == 0)
	  return(-1);
	dest = conf->result.header_start;
	s.next_out = conf->result.data_start;
  }
  else {
	need = inlen * 2;
	dest = s.next_out = lzma_alloc(conf, need);
	s.avail_out = need;
  }

  rc = lzma_code(&s, LZMA_FINISH);
  if (rc != LZMA_STREAM_END)
	goto err;

  if (s.avail_in != 0)
	goto err;

  if (out != NULL)
	*out = dest;
  else
	lzma_dealloc(conf, dest);

  if (outlen != NULL)
	*outlen = s.total_out + conf->header_alloc;

  return(0);
}

/*
 * Initialization for LZMA2.
 * See doc/examples/02_decompress.c in the xz distribution.
 */
static void *
lzma_decompress_open(Compress_conf *conf)
{
  lzma_stream *strm;
  lzma_ret ret;
  static lzma_stream strm_init = LZMA_STREAM_INIT;

  strm = (lzma_stream *) ALLOC(lzma_stream);
  *strm = strm_init;
  ret = lzma_stream_decoder(strm, UINT64_MAX, LZMA_CONCATENATED);
  if (ret != LZMA_OK)
	return(NULL);

  return(strm);
}

static int
lzma_decompress_update(Compress_conf *conf, void *in, size_t inlen,
					   unsigned char *out, size_t outlen, size_t *nwritten)
{
  int nread;
  lzma_ret rc;
  lzma_stream *strm;

  strm = (lzma_stream *) conf->subctx;
  if (conf->header_status == COMPRESS_HAVE_HEADER) {
	strm->avail_in = adjust_for_input_header(conf, in, inlen);
	strm->next_in = conf->result.data_start;
  }
  else {
	strm->avail_in = inlen;
	strm->next_in = in;
  }

  strm->next_out = out;
  strm->avail_out = outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  rc = lzma_code(strm, LZMA_RUN);
  if (rc == LZMA_STREAM_END) {
	if (strm->avail_in == 0)
	  conf->is_finished = 1;
  }
  else if (rc != LZMA_OK)
	return(-1);

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  nread = inlen - strm->avail_in;

  return(nread);
}

static int
lzma_decompress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					   size_t *nwritten)
{
  lzma_ret rc;
  lzma_stream *strm;

  strm = (lzma_stream *) conf->subctx;
  strm->avail_in = 0;
  strm->next_in = NULL;
  strm->next_out = out;
  strm->avail_out = outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  if (conf->is_finished)
	return(0);

  rc = lzma_code(strm, LZMA_FINISH);
  if (rc == LZMA_STREAM_END) {
	if (strm->avail_out == outlen)
	  conf->is_finished = 1;
  }
  else if (rc != LZMA_OK)
	return(-1);

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  return(conf->is_finished ? 0 : 1);
}

static void
lzma_decompress_close(Compress_conf *conf)
{

  if (conf->subctx != NULL)
	lzma_end(conf->subctx);
}

/*
 * Decompress INLEN bytes from IN, setting OUT to the allocated buffer
 * containing OUTLEN decompressed bytes.
 * Both IN and INLEN have been modified to account for any application
 * header.
 */
static int
lzma_decompress(Compress_conf *conf, void *in, size_t inlen,
				unsigned char **out, size_t *outlen)
{
  unsigned char *dest;
  size_t grow_len, need, nused;
  lzma_ret rc;
  lzma_stream s = LZMA_STREAM_INIT;

  dest = NULL;

  /*
  rc = lzma_auto_decoder(&s, UINT64_MAX, 0);
  */
  rc = lzma_stream_decoder(&s, UINT64_MAX, LZMA_CONCATENATED);
  if (rc != LZMA_OK) {
  err:
	if (dest != NULL)
	  lzma_dealloc(conf, dest);
	return(-1);
  }

  if (conf->expected_output_size)
	need = grow_len = conf->expected_output_size;
  else {
	grow_len = inlen;
	need = inlen * 2;
  }

  /*
   * Set s.next_in, s.avail_in, s.next_out, s.avail_out to filter data
   * through the LZMA decompression function.
   */
  s.next_in = in;
  s.avail_in = inlen;

  dest = lzma_alloc(conf, need);
  s.next_out = dest;
  s.avail_out = need;
  nused = 0;

  while (1) {
	size_t last_alloc;

	rc = lzma_code(&s, LZMA_FINISH);
	if (rc == LZMA_STREAM_END) {
	  if (s.avail_in == 0) {
		conf->is_finished = 1;
		break;
	  }
	}
	else if (rc != LZMA_OK)
	  goto err;

	/* Allocate more memory. */
	last_alloc = need;
	need += grow_len;
	nused = last_alloc - s.avail_out;
	dest = lzma_realloc(conf, dest, last_alloc, need);
	s.next_out = dest + nused;
	s.avail_out = need - nused;
  }

  if (s.avail_in != 0)
	goto err;

  if (out != NULL)
	*out = dest;
  else
	lzma_dealloc(conf, dest);

  if (outlen != NULL)
	*outlen = s.total_out;
  
  return(0);
}
#endif

/************************************************************************/
#if defined(HAVE_ZLIB) || defined(HAVE_LIBZ)

/*
 * http://zlib.net
 * http://zlib.net/manual.html
 */

#include <zlib.h>

static MAYBE_UNUSED const char *
zlib_error_string(z_stream *strm)
{
  const char *mesg;
  
  if (strm->msg != NULL)
	mesg = strdup(strm->msg);
  else
	mesg = NULL;

  return(mesg);
}

/*
 * Memory allocation callback.
 */
static void *
zlib_alloc(void *opaque, unsigned int n, unsigned int m)
{
  void *ptr;
  size_t need;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  need = n * m;
  ptr = custom_alloc(conf, need);

  return(ptr);
}

/*
 * Memory deallocation callback.
 */
static void
zlib_dealloc(void *opaque, void *ptr)
{
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  custom_dealloc(conf, ptr);
}

static void *
zlib_realloc(void *opaque, void *ptr, size_t last_alloc, size_t need)
{
  void *new_ptr;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  new_ptr = custom_realloc(conf, ptr, last_alloc, need);

  return(new_ptr);
}

static void *
zlib_compress_open(Compress_conf *conf)
{
  int st;
  z_stream *strm;

  strm = ALLOC(z_stream);
  strm->zalloc = zlib_alloc;
  strm->zfree = zlib_dealloc;
  strm->opaque = conf;

  if ((st = deflateInit(strm, Z_BEST_COMPRESSION)) != Z_OK) {
	conf->result.error_code = st;
	free(strm);
	return(NULL);
  }

  return(strm);
}

static int
zlib_compress_update(Compress_conf *conf, void *in, size_t inlen,
					 unsigned char *out, size_t outlen, size_t *nwritten)
{
  int st;
  z_stream *strm;

  strm = (z_stream *) conf->subctx;
  strm->avail_in = (uInt) inlen;
  strm->next_in = (Bytef *) in;
  if (conf->header_status == COMPRESS_NEED_HEADER) {
	if ((strm->avail_out = adjust_for_output_header(conf, out, outlen, NULL))
		== 0)
	  return(-1);

	strm->next_out = conf->result.data_start;
	/* Below, BUFLEN will be set to a size that includes this header. */
  }
  else {
	strm->next_out = (Bytef  *) out;
	strm->avail_out = (uInt) outlen;
  }

  if (nwritten != NULL)
	*nwritten = 0;

  if ((st = deflate(strm, Z_NO_FLUSH)) == Z_STREAM_END)
	conf->is_finished = 1;
  else if (st != Z_OK) {
	conf->result.error_code = st;
	return(-1);
  }

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  st = inlen - strm->avail_in;

  return(st);
}

static int
zlib_compress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					   size_t *nwritten)
{
  int st;
  z_stream *strm;

  strm = (z_stream *) conf->subctx;
  strm->next_out = out;
  strm->avail_out = outlen;

  if (conf->is_finished)
	return(0);

  st = deflate(strm, Z_FINISH);
  if (st == Z_OK || st == Z_BUF_ERROR) {
	/* There is more to output. */
	if (nwritten != NULL)
	  *nwritten = outlen - strm->avail_out;
	return(1);
  }

  if (st != Z_STREAM_END) {
	deflateEnd(strm);
	conf->subctx = NULL;
	conf->result.error_code = st;
	return(-1);
  }

  if ((st = deflateEnd(strm)) != Z_OK) {
	conf->result.error_code = st;
	conf->subctx = NULL;
	return(-1);
  }

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  conf->is_finished = 1;

  return(0);
}

static void
zlib_compress_close(Compress_conf *conf)
{

}

/*
 * Compress INLEN bytes from IN, setting OUT to the allocated buffer containing
 * OUTLEN compressed bytes.
 */
static int
zlib_compress(Compress_conf *conf, void *in, size_t inlen,
			  unsigned char **out, size_t *outlen) {
  int st;
  void *dest;
  uLong destLen;
  z_stream strm;

  if (in == NULL)
	return(-1);

  strm.zalloc = zlib_alloc;
  strm.zfree = zlib_dealloc;
  strm.opaque = conf;

  if ((st = deflateInit(&strm, Z_BEST_COMPRESSION)) != Z_OK) {
	conf->result.error_code = st;
	return(-1);
  }

  /*
   * Determine the upper bound on the compressed size, then allocate
   * the output buffer.
   */
  destLen = deflateBound(&strm, inlen);

  strm.avail_in = (uInt) inlen;
  strm.next_in = (Bytef *) in;

  if (conf->header_status == COMPRESS_NEED_HEADER) {
	size_t need;

	need = destLen;
	if ((strm.avail_out = adjust_for_output_header(conf, NULL, need, NULL))
		== 0)
	  return(-1);
	dest = conf->result.header_start;
	strm.next_out = conf->result.data_start;
  }
  else {
	dest = zlib_alloc(conf, destLen, 1);
	strm.next_out = (Bytef  *) dest;
	strm.avail_out = (uInt) destLen;
  }

  if ((st = deflate(&strm, Z_FINISH)) != Z_STREAM_END) {
	deflateEnd(&strm);
	conf->subctx = NULL;
	conf->result.error_code = st;
	return(-1);
  }

  if ((st = deflateEnd(&strm)) != Z_OK) {
	conf->subctx = NULL;
	conf->result.error_code = st;
	return(-1);
  }

  /*
   * If OUT is NULL, the caller was not interested in the compressed data
   * (probably only its size), so the output buffer should be deallocated.
   */
  if (out != NULL)
	*out = dest;
  else
	zlib_dealloc(conf, dest);

  /* Pass back the size in bytes, if requested. */
  if (outlen != NULL)
	*outlen = strm.total_out + conf->header_alloc;

  return(0);
}

static void *
zlib_decompress_open(Compress_conf *conf)
{
  int st;
  z_stream *strm;

  strm = ALLOC(z_stream);

  strm->zalloc = zlib_alloc;
  strm->zfree = zlib_dealloc;
  strm->opaque = conf;

  /*
   * In case it is not obvious, "inflate" means decompress and
   * "deflate" means compress.
   */
  if ((st = inflateInit(strm)) != Z_OK) {
	conf->result.error_code = st;
	free(strm);
	return(NULL);
  }

  return(strm);
}

static int
zlib_decompress_update(Compress_conf *conf, void *in, size_t inlen,
					   unsigned char *out, size_t outlen, size_t *nwritten)
{
  int st;
  z_stream *strm;

  strm = (z_stream *) conf->subctx;
  if (conf->header_status == COMPRESS_HAVE_HEADER) {
	strm->avail_in = adjust_for_input_header(conf, in, inlen);
	strm->next_in = conf->result.data_start;
  }
  else {
	strm->avail_in = (uInt) inlen;
	strm->next_in = (Bytef *) in;
  }

  strm->next_out = (Bytef  *) out;
  strm->avail_out = (uInt) outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  st = inflate(strm, Z_NO_FLUSH);
  if (st == Z_STREAM_END)
	conf->is_finished = 1;
  else if (st != Z_OK) {
	conf->result.error_code = st;
	return(-1);
  }

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  st = inlen - strm->avail_in;

  return(st);
}

static int
zlib_decompress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					   size_t *nwritten)
{
  int st;
  z_stream *strm;

  strm = (z_stream *) conf->subctx;
  strm->next_out = out;
  strm->avail_out = outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  if (conf->is_finished)
	return(0);

  st = inflate(strm, Z_FINISH);
  if (st == Z_OK || st == Z_BUF_ERROR) {
	/* There is more to output. */
	if (nwritten != NULL)
	  *nwritten = outlen - strm->avail_out;
	return(1);
  }

  if (st != Z_STREAM_END) {
	inflateEnd(strm);
	conf->subctx = NULL;
	conf->result.error_code = st;
	return(-1);
  }

  if ((st = inflateEnd(strm)) != Z_OK) {
	conf->subctx = NULL;
	conf->result.error_code = st;
	return(-1);
  }

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  conf->is_finished = 1;

  return(0);
}

static void
zlib_decompress_close(Compress_conf *conf)
{

}

/*
 * Decompress INLEN bytes from IN, setting OUT to the allocated buffer
 * containing OUTLEN decompressed bytes.
 * Both IN and INLEN have been modified to account for any application
 * header.
 */
static int
zlib_decompress(Compress_conf *conf, void *in, size_t inlen,
				unsigned char **out, size_t *outlen)
{
  int st;
  size_t grow_len, nalloc, need;
  unsigned char *dest;
  z_stream strm;

  if (in == NULL || inlen == 0)
	return(-1);

  if (conf->expected_output_size)
	grow_len = conf->expected_output_size;
  else
	grow_len = inlen;

  strm.next_in = in;
  strm.avail_in = inlen;

  strm.zalloc = zlib_alloc;
  strm.zfree = zlib_dealloc;
  strm.opaque = conf;

  if ((st = inflateInit(&strm)) != Z_OK) {
	conf->result.error_code = st;
	return(-1);
  }

  nalloc = need = inlen;
  dest = (Bytef *) zlib_alloc(conf, need, 1);
  strm.next_out = dest;
  strm.avail_out = need;

  while (1) {
	size_t last_alloc;

	if ((st = inflate(&strm,  Z_SYNC_FLUSH)) == Z_STREAM_END)
	  break;
	if (st != Z_OK)
	  break;

	/* XXX allocate more memory... */
	last_alloc = need;
	need += grow_len;
	dest = zlib_realloc(conf, dest, last_alloc, need);
	strm.next_out = dest + nalloc;
	strm.avail_out = need - nalloc;
	nalloc = need;
  }

  if (st != Z_STREAM_END || strm.avail_in != 0) {
	inflateEnd(&strm);
	conf->subctx = NULL;
	conf->result.error_code = st;
	return(-1);
  }

  if ((st = inflateEnd(&strm)) != Z_OK) {
	conf->subctx = NULL;
	conf->result.error_code = st;
	return(-1);
  }

  if (out != NULL)
	*out = dest;
  else
	zlib_dealloc(conf, dest);

  if (outlen != NULL)
	*outlen = strm.total_out;

  return(0);
}

#endif

/************************************************************************/
#ifdef HAVE_LIBBZ2
/*
 * https://www.sourceware.org/bzip2/
 * https://www.sourceware.org/bzip2/manual/manual.html#libprog
 */

#include <bzlib.h>

/*
 * XXX bzlib.h defines pointers to be (char *) instead of (unsigned char *).
 */

/*
 * The library "unofficially" provides BZ2_bzerror(), which seems unnecessarily
 * restrictive because it requires a (BZFILE *) argument.  And we can't get at
 * bzerrorstrings[] in bzlib.c from here.
 * So as we had to do for lzma_error_string(), we write something that is
 * unpleasantly fragile.
 */
static MAYBE_UNUSED const char *
bz2_error_string(int error_code)
{
  const char *mesg;

  /*
   * XXX
   * this is necessarily very dependent on bzlib.h, which defines
   * not only true error codes, but all return values, and not as enums.
   */
  switch (error_code) {
  case BZ_OK:
	mesg = "OK"; break;
  case BZ_RUN_OK:
	mesg = "Run OK"; break;
  case BZ_FLUSH_OK:
	mesg = "Flush OK"; break;
  case BZ_FINISH_OK:
	mesg = "Finish OK"; break;
  case BZ_STREAM_END:
	mesg = "Stream end"; break;
  case BZ_SEQUENCE_ERROR:
	mesg = "Sequence error"; break;
  case BZ_PARAM_ERROR:
	mesg = "Param error"; break;
  case BZ_MEM_ERROR:
	mesg = "Mem error"; break;
  case BZ_DATA_ERROR:
	mesg = "Data error"; break;
  case BZ_DATA_ERROR_MAGIC:
	mesg = "Data error magic"; break;
  case BZ_IO_ERROR:
	mesg = "IO error"; break;
  case BZ_UNEXPECTED_EOF:
	mesg = "Unexpected EOF"; break;
  case BZ_OUTBUFF_FULL:
	mesg = "Outbuf full"; break;
  case BZ_CONFIG_ERROR:
	mesg = "Config error"; break;
  default:
	mesg = "Unknown error"; break;
  }

  return(mesg);
}

/*
 * Memory allocation callback.
 */
static void *
bz2_alloc(void *opaque, int n, int m)
{
  void *ptr;
  size_t need;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  need = n * m;
  ptr = custom_alloc(conf, need);

  return(ptr);
}

/*
 * Memory deallocation callback.
 */
static void
bz2_dealloc(void *opaque, void *ptr)
{
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  custom_dealloc(conf, ptr);
}

static void *
bz2_realloc(void *opaque, void *ptr, size_t last_alloc, size_t need)
{
  void *new_ptr;
  Compress_conf *conf;

  conf = (Compress_conf *) opaque;
  new_ptr = custom_realloc(conf, ptr, last_alloc, need);

  return(new_ptr);
}

static void *
bz2_compress_open(Compress_conf *conf)
{
  bz_stream *strm;

  strm = ALLOC(bz_stream);
  strm->bzalloc = bz2_alloc;
  strm->bzfree = bz2_dealloc;
  strm->opaque = conf;

  /* Configure: best compression; silent; default workFactor. */
  if (BZ2_bzCompressInit(strm, 9, 0, 0) != BZ_OK)
	return(NULL);

  return(strm);
}

static int
bz2_compress_update(Compress_conf *conf, void *in, size_t inlen,
					unsigned char *out, size_t outlen, size_t *nwritten)
{
  int st;
  bz_stream *strm;

  strm = (bz_stream *) conf->subctx;
  strm->next_in = in;
  strm->avail_in = inlen;
  if (conf->header_status == COMPRESS_NEED_HEADER) {
	if ((strm->avail_out = adjust_for_output_header(conf, out, outlen, NULL))
		== 0)
	  return(-1);

	strm->next_out = conf->result.data_start;
	/* Below, NWRITTEN will be set to a size that includes this header. */
  }
  else {
	strm->next_out = (char *) out;
	strm->avail_out = outlen;
  }

  if (nwritten != NULL)
	*nwritten = 0;

  if ((st = BZ2_bzCompress(strm, BZ_RUN)) != BZ_RUN_OK)
	return(-1);

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  st = inlen - strm->avail_in;

  return(st);
}

static int
bz2_compress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					   size_t *nwritten)
{
  int st;
  bz_stream *strm;

  strm = (bz_stream *) conf->subctx;
  strm->next_out = (char *) out;
  strm->avail_out = outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  if (conf->is_finished)
	return(0);

  if ((st = BZ2_bzCompress(strm, BZ_FINISH)) == BZ_STREAM_END) {
	conf->is_finished = 1;
	if (nwritten != NULL)
	  *nwritten = ((size_t) strm->total_out_hi32 << 32) + strm->total_out_lo32;

	BZ2_bzCompressEnd(strm);
	return(0);
  }

  if (st != BZ_FINISH_OK)
	return(-1);

  return(1);
}

static void
bz2_compress_close(Compress_conf *conf)
{

}

/*
 * Compress INLEN bytes from IN, setting OUT to the allocated buffer containing
 * OUTLEN compressed bytes.
 *
 * For typical and small input (< 200 bytes anyway), bzip2 sucks because
 * of the relatively large overhead.
 */
static int
bz2_compress(Compress_conf *conf, void *in, size_t inlen,
			 unsigned char **out, size_t *outlen)
{
  int st;
  unsigned char *dest;
  size_t need;
  bz_stream strm;

  strm.bzalloc = bz2_alloc;
  strm.bzfree = bz2_dealloc;
  strm.opaque = conf;

  /* Configure: best compression; silent; default workFactor. */
  if (BZ2_bzCompressInit(&strm, 9, 0, 0) != BZ_OK)
	return(-1);

  strm.next_in = in;
  strm.avail_in = inlen;
  if (conf->header_status == COMPRESS_NEED_HEADER) {
	need = inlen * 1.01 + 600;
	if ((strm.avail_out = adjust_for_output_header(conf, NULL, need, NULL))
		== 0)
	  return(-1);
	dest = conf->result.header_start;
	strm.next_out = conf->result.data_start;
  }
  else {
	strm.avail_out = need = inlen * 1.01 + 600;
	dest = bz2_alloc(conf, need, 1);
	strm.next_out = (char *) dest;
  }

  if ((st = BZ2_bzCompress(&strm, BZ_RUN)) != BZ_RUN_OK)
	return(-1);

  while ((st = BZ2_bzCompress(&strm, BZ_FINISH)) == BZ_FINISH_OK)
	;
  if (st != BZ_STREAM_END)
	return(-1);

  BZ2_bzCompressEnd(&strm);

  if (outlen != NULL) {
	*outlen = (((size_t) strm.total_out_hi32 << 32) + strm.total_out_lo32)
	  + conf->header_alloc;
  }

  if (out != NULL)
	*out = dest;
  else
	bz2_dealloc(conf, dest);

  return(0);
}

static void *
bz2_decompress_open(Compress_conf *conf)
{
  bz_stream *strm;

  strm = (bz_stream *) ALLOC(z_stream);
  strm->bzalloc = bz2_alloc;
  strm->bzfree = bz2_dealloc;
  strm->opaque = conf;

  if (BZ2_bzDecompressInit(strm, 0, 0) != BZ_OK)
	return(NULL);

  return(strm);
}

static int
bz2_decompress_update(Compress_conf *conf, void *in, size_t inlen,
					  unsigned char *out, size_t outlen, size_t *nwritten)
{
  int st;
  bz_stream *strm;

  strm = (bz_stream *) conf->subctx;
  if (conf->header_status == COMPRESS_HAVE_HEADER) {
	strm->avail_in = adjust_for_input_header(conf, in, inlen);
	strm->next_in = conf->result.data_start;
  }
  else {
	strm->next_in = in;
	strm->avail_in = inlen;
  }
  strm->next_out = (char *) out;
  strm->avail_out = outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  if ((st = BZ2_bzDecompress(strm)) == BZ_STREAM_END)
	conf->is_finished = 1;
  else if (st != BZ_OK)
	return(-1);

  if (nwritten != NULL)
	*nwritten = outlen - strm->avail_out;

  st = inlen - strm->avail_in;	

  return(st);
}

static int
bz2_decompress_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					  size_t *nwritten)
{
  int st;
  bz_stream *strm;

  strm = (bz_stream *) conf->subctx;
  strm->next_out = (char *) out;
  strm->avail_out = outlen;

  if (nwritten != NULL)
	*nwritten = 0;

  if (conf->is_finished)
	return(0);

  if (!conf->is_finished) {
	if ((st = BZ2_bzDecompress(strm)) == BZ_STREAM_END)
	  conf->is_finished = 1;
	else if (st != BZ_OK)
	  return(-1);

	if (nwritten != NULL)
	  *nwritten = outlen - strm->avail_out;
	return(conf->is_finished);
  }

  BZ2_bzDecompressEnd(strm);
  conf->subctx = NULL;

  return(0);
}

static void
bz2_decompress_close(Compress_conf *conf)
{

}

/*
 * Decompress INLEN bytes from IN, setting OUT to the allocated buffer
 * containing OUTLEN decompressed bytes.
 */
static int
bz2_decompress(Compress_conf *conf, void *in, size_t inlen,
			   unsigned char **out, size_t *outlen)
{
  unsigned char *dest;
  size_t grow_len, need, nalloc;
  bz_stream strm;

  strm.bzalloc = bz2_alloc;
  strm.bzfree = bz2_dealloc;
  strm.opaque = conf;

  if (BZ2_bzDecompressInit(&strm, 0, 0) != BZ_OK)
	return(-1);

  if (conf->expected_output_size)
	grow_len = conf->expected_output_size;
  else
	grow_len = inlen;
  strm.next_in = in;
  strm.avail_in = grow_len = inlen;

  nalloc = need = grow_len;
  dest = bz2_alloc(conf, need, 1);
  strm.next_out = (char *) dest;
  strm.avail_out = need;

  while (1) {
	int rc;
	size_t last_alloc;

	if ((rc = BZ2_bzDecompress(&strm)) == BZ_STREAM_END)
	  break;

	if (rc != BZ_OK)
	  return(-1);

	/* XXX allocate more memory... */
	last_alloc = need;
	need += grow_len;
	dest = bz2_realloc(conf, dest, last_alloc, need);
	strm.next_out = (char *) (dest + nalloc);
	strm.avail_out = need - nalloc;
	nalloc = need;
  }

  if (strm.avail_in != 0)
	return(-1);

  BZ2_bzDecompressEnd(&strm);
  conf->subctx = NULL;

  if (out != NULL)
	*out = dest;
  else
	bz2_dealloc(conf, dest);

  if (outlen)
	*outlen = ((size_t) strm.total_out_hi32 << 32) + strm.total_out_lo32;

  return(0);
} 
#endif

/************************************************************************/

static Compress_desc compress_tab[] = {
#ifdef HAVE_LIBBROTLI
  { "br", "br,.br",
	COMPRESS_BROTLI, COMPRESS_ENABLED, brotli_version_string,
	brotli_compress,
	brotli_compress_open, brotli_compress_update, brotli_compress_finish,
	brotli_compress_close,
	brotli_decompress,
	brotli_decompress_open, brotli_decompress_update, brotli_decompress_finish,
	brotli_decompress_close,
  },
#endif
#ifdef HAVE_LIBLZMA
  { "lzma", "lzma,xz,.xz",
	COMPRESS_LZMA, COMPRESS_ENABLED, lzma_version_string,
	lzma_compress,
	lzma_compress_open, lzma_compress_update, lzma_compress_finish,
	lzma_compress_close,
	lzma_decompress,
	lzma_decompress_open, lzma_decompress_update, lzma_decompress_finish,
	lzma_decompress_close,
  },
#endif
#ifdef HAVE_LIBZSTD
  { "zstd", "zstd,.zst",
	COMPRESS_ZSTD, COMPRESS_ENABLED, ZSTD_versionString,
	zstd_compress,
	zstd_compress_open, zstd_compress_update, zstd_compress_finish,
	zstd_compress_close,
	zstd_decompress,
	zstd_decompress_open, zstd_decompress_update, zstd_decompress_finish,
	zstd_decompress_close
  },
#endif
#ifdef HAVE_LIBZ
  { "zlib", "zlib,gzip,.gz",
	COMPRESS_ZLIB, COMPRESS_ENABLED, zlibVersion,
	zlib_compress,
	zlib_compress_open, zlib_compress_update, zlib_compress_finish,
	zlib_compress_close,
	zlib_decompress,
	zlib_decompress_open, zlib_decompress_update, zlib_decompress_finish,
	zlib_decompress_close,
  },
#endif
#ifdef HAVE_LIBBZ2
  { "bz2", "bz2,bzip2,.bz2",
	COMPRESS_BZ2, COMPRESS_ENABLED, BZ2_bzlibVersion,
	bz2_compress,
	bz2_compress_open, bz2_compress_update, bz2_compress_finish,
	bz2_compress_close,
	bz2_decompress,
	bz2_decompress_open, bz2_decompress_update, bz2_decompress_finish,
	bz2_decompress_close,
  },
#endif
  { NULL, NULL,
	COMPRESS_NONE, COMPRESS_DISABLED, NULL,
	NULL,
	NULL, NULL, NULL, NULL,
	NULL,
	NULL, NULL, NULL, NULL
  }
};

#ifdef HAVE_LIBMAGIC
#include <magic.h>

/*
 * Examine BUF, which contains the first BUFLEN bytes of compressed data,
 * to try to identify the compression algorithm used.
 * Return the algorithm identifier, or COMPRESS_ERROR.
 * If TYPE_STR is not NULL, also return the string descriptor returned by
 * libmagic(3).
 *
 * We assume that BUFLEN is sufficient for the purpose and that inspecting
 * the first few bytes of the string descriptor tells us all we need.
 *
 * XXX at present, libmagic(3) cannot identify Brotli compression,
 * so an application-specific header, a MIME header, or an explicit
 * argument IS REQUIRED, or using a trustworthy file suffix (".br", ".tbr")
 * when available (and, presumably, by process of elimination if the input
 * is of an unknown type and decompresses without error
 */
Compress_alg
compress_type_from_buffer(unsigned char *buf, size_t buflen, char **type_str)
{
  const char *str;
  magic_t m;
  Compress_alg alg;

  if ((m = magic_open(MAGIC_NO_CHECK_COMPRESS)) == NULL)
	return(COMPRESS_ERROR);

#ifndef MAGIC_PATTERN_FILE
#define MAGIC_PATTERN_FILE		"/usr/share/misc/magic"
#endif

  if (magic_load(m, MAGIC_PATTERN_FILE) == -1) {
  err:
	magic_close(m);
	return(COMPRESS_ERROR);
  }

  if ((str = magic_buffer(m, buf, buflen)) == NULL)
	goto err;
  if (debug_flag)
	fprintf(stderr, "compress_type_from_buffer: \"%s\"\n", str);

  if (type_str != NULL)
	*type_str = strdup(str);

  if (strcaseprefix(str, "zstandard "))
	alg = COMPRESS_ZSTD;
  else if (strcaseprefix(str, "zlib ") || strcaseprefix(str, "gzip "))
	alg = COMPRESS_ZLIB;
  else if (strcaseprefix(str, "xz "))
	alg = COMPRESS_LZMA;
  else if (strcaseprefix(str, "bzip2 "))
	alg = COMPRESS_BZ2;
  else
	alg = COMPRESS_ERROR;

  magic_close(m);

  return(alg);
}
#else
Compress_alg
compress_type_from_buffer(unsigned char *buf, size_t buflen, char **type_str)
{

  return(COMPRESS_ERROR);
}
#endif

/*
 * Inspect the start of compressed stream FP to determine the compression
 * algorithm that was used, then reset the file position to the start.
 *
 * Return the algorithm identifier, or COMPRESS_ERROR if it failed.
 */
Compress_alg
compress_type_from_stream(FILE *fp, char **type_str)
{
  unsigned char buf[BUFSIZ];
  size_t nread;
  Compress_alg alg;

  if ((nread = fread(buf, 1, sizeof(buf), fp)) == 0)
	return(COMPRESS_ERROR);

  alg = compress_type_from_buffer(buf, nread, type_str);

  /*
   * XXX Provided we read no more than BUFSIZ bytes from the stream and
   * nothing is done on the stream between the read and the rewind, this should
   * most likely be fine.  But if we ever find that can't reset the file
   * position to the start, we will need to retain buf and nread and pass them
   * to compress_funpack_x() for interpolation.
   */
  if (fseek(fp, 0L, SEEK_SET) == -1) {
	fprintf(stderr, "*** BUG: fseek failed\n");
	return(COMPRESS_ERROR);
  }

  return(alg);
}

Compress_desc *
compress_lookup_by_alg(Compress_alg alg)
{
  Compress_desc *cd;

  for (cd = &compress_tab[0]; cd->name != NULL; cd++) {
	if (cd->alg == alg)
	  return(cd);
  }

  return(NULL);
}

/*
 * Look up NAME, which can be any recognized alias for a compression
 * algorithm, and return its description, or NULL if NAME is not known.
 */
Compress_desc *
compress_lookup_by_name(const char *name)
{
  Compress_desc *cd;

  for (cd = &compress_tab[0]; cd->name != NULL; cd++) {
	int i;
	Dsvec *dsv;

	if ((dsv = strsplit(cd->aliases, ",", 0)) == NULL)
	  continue;
	for (i = 0; i < dsvec_len(dsv); i++) {
	  char *alias;

	  alias = (char *) dsvec_ptr_index(dsv, i);
	  if (strcaseeq(name, alias)) {
		dsvec_free(dsv);
		return(cd);
	  }
	}
	dsvec_free(dsv);
  }

  return(NULL);
}

/*
 * Map compression algorithm ALG to its canonical string name, or NULL.
 * Return NULL if ALG is not recognized or unavailable.
 */
const char *
compress_name(Compress_alg alg)
{
  Compress_desc *cd;

  if ((cd = compress_lookup_by_alg(alg)) != NULL)
	return(cd->name);

  return(NULL);
}

/*
 * Return 1 if ALG can be used, 0 otherwise.
 */
int
compress_alg_is_available(Compress_alg alg)
{
  Compress_desc *cd;

  if ((cd = compress_lookup_by_alg(alg)) != NULL)
	return(1);

  return(0);
}

/*
 * Lookup compress algorithm NAME.
 * Return COMPRESS_NONE if NAME is not recognized or unavailable.
 */
Compress_alg
compress_alg(const char *name)
{
  Compress_desc *cd;

  if ((cd = compress_lookup_by_name(name)) != NULL)
	return(cd->alg);

  return(COMPRESS_NONE);
}

/*
 * Return a pointer to the descriptor for the "first" available
 * compression algorithm, or NULL if none are available.
 */
Compress_desc *
compress_alg_begin(void)
{
  Compress_desc *cd;

  cd = &compress_tab[0];
  if (cd->name == NULL)
	return(NULL);

  return(cd);
}

/*
 * Given a pointer to a descriptor returned by compress_alg_begin() or
 * compress_alg_next(), return a pointer to the descriptor for the
 * next available algorithm, or NULL if PCD was the last one.
 */
Compress_desc *
compress_alg_next(Compress_desc *pcd)
{
  Compress_desc *cd;

  cd = pcd + 1;
  if (cd->name == NULL)
	return(NULL);

  return(cd);
}

/*
 * Return version information for ALG, or NULL.
 */
const char *
compress_version(Compress_alg alg)
{
  const char *str;
  Compress_desc *cd;

  if ((cd = compress_lookup_by_alg(alg)) == NULL)
	return(NULL);
  str = (cd->get_version_str)();

  return(str);
}

void
compress_info(FILE *fp)
{
  Compress_desc *cd;

  for (cd = &compress_tab[0]; cd->name != NULL; cd++)
	fprintf(fp, "\"%s\" (%s)\n", cd->name, (cd->get_version_str)());
#ifdef HAVE_LIBMAGIC
  fprintf(fp, "libmagic: %d\n", magic_version());
#endif
}

/*
 * Produce a compact list of all available algorithms.
 */
char *
compress_alg_list(void)
{
  Compress_desc *cd;
  Ds ds;

  ds_init(&ds);
  for (cd = &compress_tab[0]; cd->name != NULL; cd++) {
	ds_asprintf(&ds, "%s%s (%s)",
				(ds_len(&ds) == 0) ? "" : ", ",
				cd->name, (cd->get_version_str)());
  }
#ifdef HAVE_LIBMAGIC
  ds_asprintf(&ds, "%slibmagic(%d)",
			  (ds_len(&ds) == 0) ? "" : ", ", magic_version());
#endif

  return(ds_buf(&ds));
}

/*
 * These are the defaults used to allocate/deallocate buffers used by
 * compress_pack(), compress_unpack(), compress_fpack(), compress_funpack().
 */
static void *(*default_alloc)(size_t size) = malloc;
static void (*default_dealloc)(void *ptr) = free;
static void *(*default_realloc)(void *ptr, size_t newsize) = realloc;

void
compress_set_default_alloc(void *(*alloc)(size_t size),
						   void (*dealloc)(void *ptr),
						   void *(*realloc)(void *ptr, size_t newsize))
{

  default_alloc = alloc;
  default_dealloc = dealloc;
  default_realloc = realloc;
}

void
compress_get_default_alloc(void *(**alloc)(size_t size),
						   void (**dealloc)(void *ptr),
						   void *(**realloc)(void *ptr, size_t newsize))
{

  *alloc = default_alloc;
  *dealloc = default_dealloc;
  *realloc = default_realloc;
}

void
compress_get_conf_alloc(Compress_conf *conf,
						void *(**alloc)(size_t size),
						void (**dealloc)(void *ptr),
						void *(**realloc)(void *ptr, size_t newsize))
{

  if (conf == NULL) {
	*alloc = default_alloc;
	*dealloc = default_dealloc;
	*realloc = default_realloc;
  }
  else {
	*alloc = conf->alloc;
	*dealloc = conf->dealloc;
	*realloc = conf->realloc;
  }
}

/*
 * Change configuration CONF to note that a header comprised of NEED
 * opaque bytes is to be allocated before the start of the compressed
 * data (when compressing) or is expected to appear before the start of the
 * compressed data (when decompressing).
 * The allocated amount is adjusted so that the byte following the header
 * is maximally aligned, since that address will be passed to the compression
 * library which expects it to be suitably aligned for storage of any type
 * of object.
 *
 * When there is a header the caller must either hide it completely from the
 * decompression functions or this function must be called immediately after
 * compress_configure() so that the functions are aware of it.
 *
 * The contents and format of the header are application dependent.
 * The application might use a header to identify the compression algorithm,
 * specify the uncompressed length (for better output buffer allocation),
 * or for a checksum/hash.
 * If compress_configure() is given COMPRESS_DEFAULT, compress_set_header()
 * must be called before either compress_unpack() or compress_unpack_x() when
 * decompressing.
 */
int
compress_set_header(Compress_conf *conf, size_t need)
{

  if (conf->op == COMPRESS_PACK)
	conf->header_status = COMPRESS_NEED_HEADER;
  else
	conf->header_status = COMPRESS_HAVE_HEADER;

  conf->header_size = need;
  conf->header_alloc = max_align_alloc(need);

  return(1);
}

int
compress_get_header(Compress_conf *conf, void **headerp, size_t *sizep,
					void **datap)
{

  if (conf->header_status == COMPRESS_HAVE_HEADER) {
	*sizep = conf->header_size;
	*headerp = conf->result.header_start;
	*datap = conf->result.data_start;

	return(1);
  }

  return(0);
}

/*
 * Initialize CONF, setting OP and ALG but everything else to its default.
 * Return 1 if ok, -1 if an error occurred.
 */
static int
init_config(Compress_conf *conf, Compress_op op, Compress_alg alg)
{

  conf->op = op;
  conf->alg = alg;
  conf->cd = NULL;
  conf->subctx = NULL;
  conf->is_finished = 0;

  conf->header_status = COMPRESS_NO_HEADER;
  conf->header_size = 0;
  conf->header_alloc = 0;
  conf->result.header_start = NULL;
  conf->result.data_start = NULL;

  conf->result.buffer = NULL;
  conf->result.buflen = 0;

  conf->alloc = default_alloc;
  conf->dealloc = default_dealloc;
  conf->realloc = default_realloc;

  conf->expected_input_size = 0;
  conf->expected_output_size = 0;

  conf->result.total_in = 0;
  conf->result.total_out = 0;

  conf->result.total_alloc = 0;
  conf->result.alloc_count = 0;
  conf->result.dealloc_count = 0;
  conf->result.realloc_count = 0;
  conf->result.dealloc = default_dealloc;

  conf->result.error_code = 0;
  conf->result.error_mesg = NULL;

  return(1);
}

/*
 * Set CONF to the default configuration and then override or refine some
 * of them.  Resolve COMPRESS_DEFAULT or COMPRESS_BEST, if necessary.
 * Return the configuration if ok or NULL if an error occurred.
 * If CONF==NULL, the configuration structure is allocated and the caller
 * is responsible for deallocating it by calling compress_pack_close()
 * or compress_unpack_close().
 */
Compress_conf *
compress_configure(Compress_conf *conf, Compress_op op, Compress_alg req_alg,
				   void *(*alloc)(size_t size), void (*dealloc)(void *ptr),
				   void *(*realloc)(void *ptr, size_t newsize),
				   size_t approx_inlen, size_t approx_outlen)
{
  Compress_alg alg;
  Compress_conf *c;
  Compress_desc *cd;

  c = NULL;

  /*
   * Check that at least one algorithm is available.
   * If the preferred default is unavailable, this one may be used.
   */
  if ((cd = compress_alg_begin()) == NULL) {
	/* No algorithms are available. */
  err:
	if (c != NULL && c->did_alloc)
	  free(c);
	return(NULL);
  }

  if (conf == NULL) {
	c = ALLOC(Compress_conf);
	c->did_alloc = 1;
  }
  else {
	c = conf;
	c->did_alloc = 0;
  }

  /*
   * XXX this needs work.
   * If either of the pseudo selections COMPRESS_DEFAULT or COMPRESS_BEST are
   * requested, use the default (COMPRESS_DEFAULT_ALG), or if that is
   * unavailable, use the one that happens to be the first (or perhaps only)
   * entry in the table of available algorithms.
   *
   * If we wanted to get fancy, we'd introduce a flag to allow a
   * user to describe the range of input characters, say, to be used with
   * input/output sizes to make a more informed decision.
   * For "smallish" inputs, there's not a whole lot of difference between
   * the modern compression algorithms.
   * If you have a favourite, just use it.
   */
  if (op == COMPRESS_PACK) {
	if (req_alg == COMPRESS_DEFAULT || req_alg == COMPRESS_BEST) {
	  if (compress_alg_is_available(COMPRESS_DEFAULT_ALG))
		alg = COMPRESS_DEFAULT_ALG;
	  else
		alg = cd->alg;
	}
	else
	  alg = req_alg;

	if (!compress_alg_is_available(alg))
	  goto err;
  }
  else {
	/*
	 * For decompression, the algorithm must either be explicit or
	 * COMPRESS_DEFAULT, which means it must be deduced.
	 */
	if (req_alg != COMPRESS_DEFAULT && !compress_alg_is_available(req_alg))
	  goto err;
	alg = req_alg;
  }

  if (init_config(c, op, alg) == -1)
	goto err;
  /* Use c->alg from now on. */

  /* Save this to avoid future algorithm lookups. */
  if ((op == COMPRESS_PACK)
	  && (c->cd = compress_lookup_by_alg(c->alg)) == NULL)
	goto err;

  /* Substitute a user-specified allocator and/or deallocator function. */
  if (alloc != NULL)
	c->alloc = alloc;
  if (dealloc != NULL)
	c->dealloc = c->result.dealloc;
  if (realloc != NULL)
	c->realloc = realloc;

  c->expected_input_size = approx_inlen;
  c->expected_output_size = approx_outlen;

  return(c);
}

/*
 * Extended "all-at-once" API to compress exactly one buffer, with configuration
 * and buffer allocation.
 * Compress IN, a contiguous buffer of INLEN bytes, allocating sufficient
 * memory and setting OUTLEN (if non-NULL) to the actual bytes of compressed
 * output (including any header) and OUT (if non-NULL) to the start of the
 * memory allocated for it (which, if requested, will be the application header).
 */
unsigned char *
compress_pack_x(Compress_conf *conf, void *in, size_t inlen,
				unsigned char **out, size_t *outlen)
{
  int st;
  unsigned char *c_text;
  size_t c_text_len;

  if (conf->cd != NULL) {
	/*
	 * Call the configured compression library: zstd_compress(),
	 * zlib_compress(), etc.
	 */
	st = (conf->cd->compress)(conf, in, inlen, &c_text, &c_text_len);
	conf->result.error_code = st;
  }
  else
	st = -1;

  if (st == -1)
	return(NULL);

  if (out != NULL)
	*out = c_text;

  if (outlen != NULL)
	*outlen = c_text_len;

  conf->result.total_in = inlen;
  conf->result.total_out = c_text_len;

  return(c_text);
}

/*
 * The "all-at-once" compression function with minimal configuration.
 * If a little more control is needed, use compress_pack_x().
 * If a lot more control is needed, use a compress_pack_open/_buffer/_close
 * sequence.
 *
 * Compress INLEN bytes from IN using ALG, allocating sufficient memory
 * with the default allocator, and setting OUT (if non-NULL) and OUTLEN
 * (if non-NULL).  Use configuration defaults.
 * OUTLEN will include the size of a header, if enabled.
 * Refer to compress_pack_x() for additional detail.
 *
 * Return a pointer to the compressed output, or NULL if an error occurs.
 * The caller would be well advised to retain OUTLEN so that decompressing
 * can be more efficient.
 */
unsigned char *
compress_pack(Compress_alg alg, void *in, size_t inlen,
			  unsigned char **out, size_t *outlen)
{
  unsigned char *v;
  Compress_conf conf;

  if (compress_configure(&conf, COMPRESS_PACK, alg, NULL, NULL, NULL,
						 0, 0) == NULL)
	return(NULL);

  v = compress_pack_x(&conf, in, inlen, out, outlen);

  compress_pack_close(&conf, NULL);

  return(v);
}

/*
 * Initialize for compressing a sequence of zero or more buffers using
 * algorithm ALG in CONF, passing it algorithm-specific configuration from
 * CONF (if CONF is NULL, use defaults).  After this, call either:
 * 1) compress_pack_update() (as many times as necessary) followed by
 *    compress_pack_finish(), or
 * 2) compress_pack_x() (exactly once).
 *    Calling compress_pack_open() in this case is unnecessary
 *
 * Lastly, call compress_pack_close().
 *
 * Return a context for subsequent compression operations, or NULL if an
 * error occurs.
 * If CONF==NULL, the configuration structure is allocated and the caller
 * is responsible for deallocating it by calling compress_pack_close()
 * or compress_unpack_close().
 */
Compress_conf *
compress_pack_open(Compress_conf *conf)
{
  Compress_conf *c;

  if (conf == NULL) {
	c = ALLOC(Compress_conf);
	init_config(c, COMPRESS_PACK, COMPRESS_DEFAULT);
	c->did_alloc = 1;
  }
  else {
	c = conf;
	c->did_alloc = 0;
  }

  if (conf->cd == NULL) {
  err:
	if (c->did_alloc)
	  free(c);
	return(NULL);
  }

  if ((c->subctx = (conf->cd->compress_open)(c)) == NULL)
	goto err;

  return(c);
}

/*
 * Using configuration CONF, feed INLEN additional bytes from IN to the
 * compression function, putting any output from the compressor into OUT,
 * which can hold at most OUTLEN bytes.
 * If not NULL, set NWRITTEN to the actual number of bytes written to OUT.
 *
 * This intended to be used if the caller wants to allocate and empty the
 * output buffer; e.g., if it is writing the output to a file, stream, list of
 * buffers, etc.
 *
 * Return the number of bytes read from IN, or -1 if an error occurred.
 */
int
compress_pack_update(Compress_conf *conf, void *in, size_t inlen,
					 unsigned char *out, size_t outlen, size_t *nwritten)
{
  int rc;

  if (conf->cd != NULL && conf->subctx != NULL) {
	rc = (conf->cd->compress_update)(conf, in, inlen, out, outlen, nwritten);
	if (rc != -1) {
	  conf->result.total_in += rc;
	  conf->result.total_out += *nwritten;
	}
  }
  else
	rc = -1;

  return(rc);
}

/*
 * There is no more input to compress.
 * Ask the compressor to flush any pending output to OUT, which can hold
 * at most OUTLEN bytes.
 *
 * Return the status code returned by the compressor, which must be 0 if there
 * is no more output, 1 if it must be called again, or -1 if an error occurred.
 * The IS_FINISHED flag is set if compression has completed.
 * The actual number of bytes put in OUT is returned in NWRITTEN.
 *
 * The caller is responsible for eventually deallocating OUT.
 */
int
compress_pack_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					 size_t *nwritten)
{
  int rc;

  if (conf->cd != NULL && conf->subctx != NULL) {
	rc = (conf->cd->compress_finish)(conf, out, outlen, nwritten);
	if (rc != -1) {
	  conf->result.total_in += rc;
	  conf->result.total_out += *nwritten;
	}
  }
  else
	rc = -1;

  return(rc);
}

/*
 * Do final cleanup, including releasing any compression library
 * resources that were created through compress_pack_open().
 *
 * If RESULT is non-NULL, a copy of the final result information is returned,
 * which includes a pointer to the output buffer and a pointer to the
 * deallocator function for the output buffer.  The caller becomes responsible
 * for buffer deallocation (by calling result->dealloc(result->buffer)).
 * If RESULT is NULL, the output buffer is deallocated here and the final result
 * is lost.
 * If the configuration was allocated by compress_pack_open(), it will be
 * deallocated (Compress_result may be lost unless the user has requested
 * a copy).
 * Return RESULT.
 *
 * XXX a difficulty for this API is that the library's context might end up
 * being large - and the user might want to free it asap or reuse it (to avoid
 * deallocating and soon after reallocating).  We also want the caller
 * to be able to hang on to the compressed/decompressed output buffer,
 * and when it is time to deallocate it the correct custom deallocation function
 * needs to be called.  We also need to be able
 * to free the Compress_conf, but it contains pointers required for
 * deallocating the various pieces.
 * We also do not want a proliferation of functions in the API.
 */
Compress_result *
compress_pack_close(Compress_conf *conf, Compress_result *result)
{

  if (conf->op != COMPRESS_PACK)
	return(NULL);

  /* This is expected to deallocate any library resources. */
  (conf->cd->compress_close)(conf);
  conf->subctx = NULL;

  if (result != NULL)
	*result = conf->result;
  else {
	if (conf->result.buffer != NULL) {
	  conf->result.dealloc(conf->result.buffer);
	  /* Just in case... */
	  conf->result.buffer = NULL;
	  conf->result.buflen = 0;
	}
  }

  if (conf->did_alloc)
	free(conf);

  return(result);
}

/*
 * Do final cleanup after decompressing.
 * Currently this is the same as compress_pack_close(), other than the call to
 * decompress_close() instead of compress_close().
 */
Compress_result *
compress_unpack_close(Compress_conf *conf, Compress_result *result)
{

  if (conf->op != COMPRESS_UNPACK)
	return(NULL);

  /* This is expected to deallocate any library resources. */
  (conf->cd->decompress_close)(conf);
  conf->subctx = NULL;

  if (result != NULL)
	*result = conf->result;
  else {
	if (conf->result.buffer != NULL) {
	  custom_dealloc(conf, conf->result.buffer);
	  /* Just in case... */
	  conf->result.buffer = NULL;
	  conf->result.buflen = 0;
	}
  }

  if (conf->did_alloc)
	free(conf);

  return(result);
}

/*
 * Extended "all-at-once" API to decompress exactly one buffer,
 * with buffer allocation and configuration.
 * Use this to unpack one contiguous buffer into a decompressed
 * contiguous buffer.
 *
 * Decompress INLEN bytes from IN, allocating sufficient memory and setting
 * OUTLEN (if non-NULL) to the actual bytes of decompressed output and OUT
 * (if non-NULL) to the allocated memory.
 * If a header is present, compress_set_header() must have been already
 * called and INLEN must include the header size.
 * If the algorithm is COMPRESS_DEFAULT, the input is inspected to
 * determine the compression algorithm that was used.
 */
unsigned char *
compress_unpack_x(Compress_conf *conf, void *in, size_t inlen,
				  unsigned char **out, size_t *outlen)
{
  int st;
  unsigned char *inp, *o_text;
  size_t inp_len, o_text_len;
  Compress_alg alg;

  /*
   * Determine the compression algorithm and make adjustments for an
   * application header, if present.
   */
  if (conf->header_status == COMPRESS_HAVE_HEADER) {
    inp_len = adjust_for_input_header(conf, in, inlen);
    inp = conf->result.data_start;
  }
  else {
    inp_len = inlen;
    inp = (unsigned char *) in;
  }

  if (conf->alg == COMPRESS_DEFAULT) {
	/*
	 * An initial portion of the compressed data must be examined to
	 * determine the algorithm to use.  If there is an application header,
	 * we have already adjusted for that above.
	 */
	if ((alg = compress_type_from_buffer(inp, inp_len, NULL)) == COMPRESS_ERROR)
	  return(NULL);
	conf->alg = alg;
  }

  if ((conf->cd = compress_lookup_by_alg(conf->alg)) != NULL)
	st = (conf->cd->decompress)(conf, inp, inp_len, &o_text, &o_text_len);
  else
	st = -1;

  if (st == -1)
	return(NULL);
  
  if (out != NULL)
	*out = o_text;

  if (outlen != NULL)
	*outlen = o_text_len;

  conf->result.total_in = inlen;
  conf->result.total_out = o_text_len;

  return(o_text);
}

/*
 * The "all-at-once" decompression function with minimal configuration.
 * If a little more control is needed, use compress_unpack_x().
 * If a lot more control is needed, use a compress_unpack_open/_buffer/_close
 * sequence.
 *
 * Uncompress INLEN bytes from IN using ALG, allocating sufficient memory
 * with the default allocator, and setting OUT (if non-NULL) and OUTLEN
 * (if non-NULL).  Use configuration defaults.
 * If the algorithm is COMPRESS_DEFAULT, the input is inspected to
 * determine the compression algorithm that was used.
 * Refer to compress_unpack_x() for additional detail.
 *
 * Return a pointer to the decompressed output, or NULL if an error occurs.
 */
unsigned char *
compress_unpack(Compress_alg alg, void *in, size_t inlen,
				unsigned char **out, size_t *outlen)
{
  unsigned char *v;
  Compress_conf conf;

  if (compress_configure(&conf, COMPRESS_UNPACK, alg, NULL, NULL, NULL,
						 0, 0) == NULL)
	return(NULL);

  v = compress_unpack_x(&conf, in, inlen, out, outlen);

  compress_unpack_close(&conf, NULL);

  return(v);
}

/*
 * Initialize for decompressing zero or more buffers with input coming from
 * stream FP_IN.  This is analogous to compress_pack_open(), other than having
 * the FP_IN parameter.
 *
 * If CONF is NULL, it is allocated and initialized to defaults.
 * If it is not NULL, use its algorithm-specific configuration, requiring
 * ALG to decompress. If ALG is COMPRESS_DEFAULT, the input is inspected to
 * determine the compression algorithm that was used.
 *
 * After this is called, use compress_unpack_update() as many times as
 * required, then compress_unpack_finish() and lastly compress_unpack_close().
 * Calling compress_unpack_open() is unnecessary if compress_unpack_x() is
 * going to be used.
 *
 * Return a new handle for subsequent decompression operations, or NULL if an
 * error occurs.
 * If CONF==NULL, the configuration structure is allocated and the caller
 * is responsible for deallocating it by calling compress_pack_close()
 * or compress_unpack_close().
 */
Compress_conf *
compress_unpack_open(Compress_conf *conf, FILE *fp_in)
{
  Compress_conf *c;

  if (conf == NULL) {
	c = ALLOC(Compress_conf);
	init_config(c, COMPRESS_UNPACK, COMPRESS_DEFAULT);
	c->did_alloc = 1;
  }
  else {
	c = conf;
	c->did_alloc = 0;
  }

  if (c->alg == COMPRESS_DEFAULT) {
	Compress_alg alg;

	/* Determine compression algorithm. */
	if (fp_in == NULL) {
	err:
	  if (c->did_alloc)
		free(c);
	  return(NULL);
	}

	if ((alg = compress_type_from_stream(fp_in, NULL)) == COMPRESS_ERROR)
	  goto err;
	c->alg = alg;
  }

  if ((c->cd = compress_lookup_by_alg(c->alg)) == NULL)
	goto err;

  c->subctx = (conf->cd->decompress_open)(c);

  return(c);
}

/*
 * Uncompress up to INLEN bytes from IN, putting up to OUTLEN bytes into OUT.
 *
 * This intended to be used if the caller wants to allocate and empty the
 * output buffer; e.g., if it is writing the output to a file, stream, list of
 * buffers, etc.
 *
 * Return the number of bytes read from IN, or -1 if an error occurred.
 * If not NULL, NWRITTEN is set to the actual number of bytes put in OUT.
 * The IS_FINISHED flag is set if decompression has completed, in which
 * case compress_unpack_finish() need not be called.
 */
int
compress_unpack_update(Compress_conf *conf, void *in, size_t inlen,
					   unsigned char *out, size_t outlen, size_t *nwritten)
{
  int rc;

  if (conf->cd != NULL && conf->subctx != NULL) {
	rc = (conf->cd->decompress_update)(conf, in, inlen, out, outlen, nwritten);
	if (rc != -1) {
	  conf->result.total_in += rc;
	  conf->result.total_out += *nwritten;
	}
  }
  else
	rc = -1;

  return(rc);
}

/*
 * There is no more input to decompress.
 * Ask the decompressor to flush any pending output to OUT, which can hold
 * at most OUTLEN bytes.  The actual amount of output is returned in NWRITTEN.
 * The IS_FINISHED flag is set if decompression has completed.
 *
 * The decompressor must return 0 if there is no more output, 1 if it must
 * be called again, or -1 if an error occurred.
 * If not NULL, NWRITTEN is set to the actual number of bytes put in OUT.
 * The caller is responsible for eventually deallocating OUT.
 */
int
compress_unpack_finish(Compress_conf *conf, unsigned char *out, size_t outlen,
					   size_t *nwritten)
{
  int rc;

  if (conf->cd != NULL && conf->subctx != NULL) {
	rc = (conf->cd->decompress_finish)(conf, out, outlen, nwritten);
	if (rc != -1) {
	  conf->result.total_in += rc;
	  conf->result.total_out += *nwritten;
	}
  }
  else
	rc = -1;

  return(rc);
}

/*
 * Compress input from FP_IN using configuration CONF, writing to FP_OUT.
 * The amount of input and output does not need to be known in advance.
 * Up to INLEN bytes of input buffer INBUF are available, and up to OUTLEN
 * bytes of output buffer OUTBUF are available.
 * The caller is responsible for closing the streams.
 * Use compress_fpack() if configurability is not needed.
 *
 * Return the number of bytes written to FP_OUT, or -1 if an error occurs.
 * In the event of an error, the caller may need to call feof(3) and/or
 * ferror(3) to determine if an I/O error occurred.
 */
ssize_t
compress_fpack_x(Compress_conf *conf,
				 FILE *fp_in, unsigned char *inbuf, size_t inlen,
				 FILE *fp_out, unsigned char *outbuf, size_t outlen)
{
  int rc;
  size_t inbytes, offset, outbytes, rem;

  /*
   * Recommended buffers sizes for compression:
   * ZSTD_CStreamInSize()
   * ZSTD_CStreamOutSize()
   */
  if (compress_pack_open(conf) == NULL)
	return(-1);

  offset = 0;
  rem = 0;
  while (1) {
	int nread;

	if (rem == 0) {
	  /* XXX probably insufficient for non-file sources... */
	  if ((inbytes = fread(inbuf, 1, inlen, fp_in)) == 0)
		break;
	  rem = inbytes;
	  offset = 0;
	  conf->result.total_in += inbytes;
	}	  

	nread = compress_pack_update(conf, inbuf + offset, inbytes - offset,
								 outbuf, outlen, &outbytes);
	if (nread == -1)
	  return(-1);

	if (outbytes) {
	  if (fwrite(outbuf, 1, outbytes, fp_out) != outbytes)
		return(-1);
	  conf->result.total_out += outbytes;
	}

	rem -= nread;
	if (rem)
	  offset += nread;
  }

  if (ferror(fp_in) || !feof(fp_in))
	return(-1);

  while (1) {
	rc = compress_pack_finish(conf, outbuf, outlen, &outbytes);
	if (rc == -1)
	  return(-1);

	if (outbytes) {
	  /* XXX probably insufficient for non-file sinks... */
	  if (fwrite(outbuf, 1, outbytes, fp_out) != outbytes)
		return(-1);
	  conf->result.total_out += outbytes;
	}

	if (rc == 0) {
	  if (fflush(fp_out) == EOF)
		return(-1);
	  break;
	}
  }

  return(conf->result.total_out);
}

/*
 * Compress input from FP_IN using ALG, writing to FP_OUT.
 * The amount of input and output does not need to be known in advance.
 * The caller is responsible for closing the streams.
 * Use compress_fpack_x() if more configurability is needed.
 *
 * Return the number of bytes written to FP_OUT, or -1 if an error occurs.
 * In the event of an error, the caller may need to call feof(3) and/or
 * ferror(3) to determine if an I/O error occurred.
 */
ssize_t
compress_fpack(FILE *fp_in, FILE *fp_out, Compress_alg alg)
{
  unsigned char inbuf[COMPRESS_DEFAULT_STREAM_INPUT_BUFFER_SIZE];
  unsigned char outbuf[COMPRESS_DEFAULT_STREAM_OUTPUT_BUFFER_SIZE];
  ssize_t rc;
  Compress_conf conf;

  if (compress_configure(&conf, COMPRESS_PACK, alg, NULL, NULL, NULL,
						 0, 0) == NULL)
	return(-1);

  rc = compress_fpack_x(&conf, fp_in, inbuf, sizeof(inbuf),
						fp_out, outbuf, sizeof(outbuf));

  compress_pack_close(&conf, NULL);

  return(rc);
}

/*
 * Decompress input from FP_IN using input buffer INBUF of INLEN bytes
 * and writing output to FP_OUT through output buffer OUTBUF of OUTLEN bytes.
 * If the algorithm is COMPRESS_DEFAULT, the input is inspected to
 * determine the compression algorithm that was used.
 *
 * Return the number of bytes written to FP_OUT, or -1 if an error occurred.
 */
ssize_t
compress_funpack_x(Compress_conf *conf,
				   FILE *fp_in, unsigned char *inbuf, size_t inlen,
				   FILE *fp_out, unsigned char *outbuf, size_t outlen)
{
  int rc;
  size_t inbytes, offset, nwritten, rem;

  /*
   * Recommended buffers sizes for decompression:
   * ZSTD_DStreamInSize()
   * ZSTD_DStreamOutSize()
   */
  if (compress_unpack_open(conf, fp_in) == NULL)
	return(0);

  offset = 0;
  rem = 0;
  while (1) {
	int nread;

	if (rem == 0) {
	  /* XXX probably insufficient for non-file sources... */
	  if ((inbytes = fread(inbuf, 1, inlen, fp_in)) == 0)
		break;
	  rem = inbytes;
	  offset = 0;
	  conf->result.total_in += inbytes;
	}

	nread = compress_unpack_update(conf, inbuf + offset , inbytes - offset,
								   outbuf, outlen, &nwritten);
	if (nread == -1)
	  return(-1);

	if (nwritten) {
	  if (fwrite(outbuf, 1, nwritten, fp_out) != nwritten)
		return(-1);
	  conf->result.total_out += nwritten;
	}

	rem -= nread;
	if (rem)
	  offset += nread;
  }

  if (ferror(fp_in) || !feof(fp_in))
	return(-1);

  while (!conf->is_finished) {
	rc = compress_unpack_finish(conf, outbuf, outlen, &nwritten);
	if (rc == -1)
	  return(-1);

	if (nwritten) {
	  /* XXX probably insufficient for non-file sinks... */
	  if (fwrite(outbuf, 1, nwritten, fp_out) != nwritten)
		return(-1);
	  conf->result.total_out += nwritten;
	}
  }

  if (fflush(fp_out) == EOF)
	return(-1);

  return(conf->result.total_out);
}

/*
 * Decompress input from FP_IN using ALG, writing to FP_OUT.
 *
 * If ALG is COMPRESS_DEFAULT, the input is inspected to
 * determine the compression algorithm that was used.
 */
ssize_t
compress_funpack(FILE *fp_in, FILE *fp_out, Compress_alg alg)
{
  unsigned char inbuf[DECOMPRESS_DEFAULT_STREAM_INPUT_BUFFER_SIZE];
  unsigned char outbuf[DECOMPRESS_DEFAULT_STREAM_OUTPUT_BUFFER_SIZE];
  Compress_conf conf;
  ssize_t rc;

  if (compress_configure(&conf, COMPRESS_UNPACK, alg, NULL, NULL, NULL,
						 0, 0) == NULL)
	return(-1);

  rc = compress_funpack_x(&conf,
						  fp_in, inbuf, sizeof(inbuf),
						  fp_out, outbuf, sizeof(outbuf));

  compress_unpack_close(&conf, NULL);

  return(rc);
}

static MAYBE_UNUSED void
set_all_alg_flags(int flags)
{
  Compress_desc *cd;

  for (cd = &compress_tab[0]; cd->name != NULL; cd++)
    cd->flags = flags;
}

#ifdef PROG

#include "dacs_malloc.h"
#include "dacs.h"

#define MAX_TESTS		20

typedef struct Test {
  int datanum;
  const char *name;
  size_t len;
  unsigned char *text;
} Test;

typedef struct Test2 {
  char *pattern;
  size_t nreps;
} Test2;

#define BINARY_01_TO_FF		  \
  "\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"		\
  "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"	\
  "\x20\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f"	\
  "\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\x3e\x3f"	\
  "\x40\x44\x42\x43\x44\x45\x46\x47\x48\x49\x4a\x4b\x4c\x4d\x4e\x4f"	\
  "\x50\x55\x52\x53\x54\x55\x56\x57\x58\x59\x5a\x5b\x5c\x5d\x5e\x5f"	\
  "\x60\x66\x62\x63\x64\x65\x66\x67\x68\x69\x6a\x6b\x6c\x6d\x6e\x6f"	\
  "\x70\x77\x72\x73\x74\x75\x76\x77\x78\x79\x7a\x7b\x7c\x7d\x7e\x7f"	\
  "\x80\x88\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f"	\
  "\x90\x99\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f"	\
  "\xa0\xaa\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf"	\
  "\xb0\xbb\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf"	\
  "\xc0\xcc\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf"	\
  "\xd0\xdd\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf"	\
  "\xe0\xee\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef"	\
  "\xf0\xff\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff"

static Test2 test2[] = {
  { "a",   1000 },
  { "abc", 1000 },
  { "abcdefghijklmnopqrstuvwxyzzyxwvutsrqponmlkjihgfedcba", 1000 },
  { "Now is the time for all good men to come to the aid of their party", 3141 },
  { "The quick brown fox jumps over the lazy dog", 4821 },
  { BINARY_01_TO_FF, 1 },
  { BINARY_01_TO_FF, 100 },
  { NULL, 0 }
};

static int do_test1, do_test2, do_test3, do_test4;

/*
 * Test the open/update/finish APIs.
 */
static int
run_test4(Compress_alg pack_alg)
{
  int count, i, rc;
  unsigned char *inp, c_buf[10240], d_buf[10240], *outp;
  size_t inlen, nwritten, outlen, rem;
  Compress_conf c_conf, d_conf;

  fprintf(stderr, "  Test4: Streaming API tests:\n");

  /*
   * Configure and compress the test data.
   */
  if (compress_configure(&c_conf, COMPRESS_PACK, pack_alg,
                         NULL, NULL, NULL, 0, 0) == NULL) {
	fprintf(stderr, "compress_configure() failed\n");
	return(-1);
  }

  if (compress_pack_open(&c_conf) == NULL) {
	fprintf(stderr, "compress_pack_open() failed\n");
    return(-1);
  }

  outp = c_buf;
  outlen = sizeof(c_buf);
  count = 3;
  for (i = 0; i < 3; i++) {
	rc = compress_pack_update(&c_conf, "abc", count, outp, outlen, &nwritten);
	if (rc == -1) {
	  fprintf(stderr, "compress_pack_update() failed\n");
	  return(-1);
	}
	if (rc == 3)
	  count = 3;
	else
	  count = 3 - rc;
	  
	outp += nwritten;
	outlen -= nwritten;
  }
#define TEST4_EXPECT	"abcabcabc"

  while (1) {
	rc = compress_pack_finish(&c_conf, outp, outlen, &nwritten);
	if (rc == -1) {
	  fprintf(stderr, "compress_pack_finish() failed\n");
	  break;
	}
	outp += nwritten;
	outlen -= nwritten;

	if (rc == 0)
	  break;
  }

  compress_pack_close(&c_conf, NULL);

  if (rc != 0)
	return(-1);

  /*
   * Configure and decompress the test data.
   */
  if (compress_configure(&d_conf, COMPRESS_UNPACK, pack_alg,
                         NULL, NULL, NULL, 0, 0) == NULL) {
	fprintf(stderr, "decompress: compress_configure() failed\n");
	return(-1);
  }

  if (compress_unpack_open(&d_conf, NULL) == NULL) {
	fprintf(stderr, "compress_unpack_open() failed\n");
    return(-1);
  }

  inp = c_buf;
  inlen = sizeof(c_buf) - outlen;
  outp = d_buf;
  outlen = sizeof(d_buf);

  rem = inlen;
  while (1) {
	rc = compress_unpack_update(&d_conf, inp, inlen, outp, outlen, &nwritten);
	if (rc == -1) {
	  fprintf(stderr, "decompress: compress_unpack_update() failed\n");
	  return(-1);
	}

	outp += nwritten;
	outlen -= nwritten;

	rem -= rc;
	if (rem == 0)
	  break;
  }

  while (1) {
	rc = compress_unpack_finish(&d_conf, outp, outlen, &nwritten);
	if (rc == -1) {
	  fprintf(stderr, "compress_unpack_finish() failed\n");
	  return(-1);
	}

	outp += nwritten;
	outlen -= nwritten;

	if (rc == 0)
	  break;
  }

  compress_unpack_close(&d_conf, NULL);

  nwritten = sizeof(d_buf) - outlen;
  if (nwritten != strlen(TEST4_EXPECT)
	  || memcmp(TEST4_EXPECT, d_buf, nwritten)) {
	fprintf(stderr, "decompress: comparison failed\n");
	return(-1);
  }

  return(0);
}

/*
 * Run a set of tests on printable strings with repeating patterns.
 */
static int
run_test2(Compress_alg pack_alg)
{
  int i, st, tnum;
  unsigned char *c_text, *o_text, *p, *d_text;
  size_t c_text_len, o_text_len, c_nalloc, d_nalloc, patlen, d_text_len;
  Compress_conf c_conf, d_conf;
  Compress_alg unpack_alg;
  Test2 *t2;

  fprintf(stderr, "  Test2: Repeating string tests:\n");
  for (tnum = 0, t2 = &test2[0]; t2->pattern != NULL; t2++, tnum++) {
	patlen = strlen(t2->pattern);
	o_text_len = patlen * t2->nreps;
	o_text = malloc(o_text_len + 1);
	p = o_text;
	for (i = 0; i < t2->nreps; i++) {
	  strncpy((char * restrict) p, t2->pattern, patlen);
	  p += patlen;
	}
	*p = '\0';

	/* Configure and compress the test data. */
	if (compress_configure(&c_conf, COMPRESS_PACK, pack_alg,
						   NULL, NULL, NULL, 0, 0) == NULL)
	  return(-1);

	/* We won't compress the null byte. */
	if (compress_pack_x(&c_conf, o_text, p - o_text,
						&c_text, &c_text_len) == NULL)
	  return(-1);


	/* Configure and decompress, then compare against the original test data. */
	if (pack_alg == COMPRESS_BROTLI)
	  unpack_alg = COMPRESS_BROTLI;
	else
	  unpack_alg = COMPRESS_DEFAULT;
	if (compress_configure(&d_conf, COMPRESS_UNPACK, unpack_alg,
						   NULL, NULL, NULL, 0, 0) == NULL)
	  return(-1);

	if (compress_unpack_x(&d_conf, c_text, c_text_len, &d_text, &d_text_len)
	  == NULL)
	return(-1);

	/* Compare original text with compressed-then-decompressed text. */
	if (o_text_len != d_text_len || memcmp(o_text, d_text, d_text_len)) {
	  fprintf(stderr, "FAILED!\n");
	  return(-1);
	}

	free(o_text);
#ifdef NOTDEF
	free(d_text);
#endif

	compress_pack_close(&c_conf, NULL);
	compress_unpack_close(&d_conf, NULL);

	if (!strprintable(t2->pattern, 0, 1))
	  fprintf(stderr, "%2d. <unprintable> x %s... OK\n",
			  tnum, strcomma(STRNUM_SIZE_T, &t2->nreps, STRCOMMA_DEFAULT));
	else if (patlen > 30)
	  fprintf(stderr, "%2d. \"%.30s\"... x %s... OK\n",
			  tnum, t2->pattern,
			  strcomma(STRNUM_SIZE_T, &t2->nreps, STRCOMMA_DEFAULT));
	else
	  fprintf(stderr, "%2d. \"%s\" x %s... OK\n",
			  tnum, t2->pattern,
			  strcomma(STRNUM_SIZE_T, &t2->nreps, STRCOMMA_DEFAULT));
	fprintf(stderr, "   [%lu -> %lu (%.1f%%)]\n",
			o_text_len, c_text_len,
			(100.0 * ((long) o_text_len - (long) c_text_len)) / o_text_len);
  }

  return(0);
}

static int
test3_compress(Compress_alg alg, unsigned char *in, size_t inlen,
			   size_t header_len,
			   unsigned char **out, size_t *outlen, Compress_result *result)
{
  Compress_conf c_conf;

  if (compress_configure(&c_conf, COMPRESS_PACK, alg,
						 NULL, NULL, NULL, 0, 0) == NULL)
	return(-1);

  if (header_len) {
	if (compress_set_header(&c_conf, header_len) == -1)
	  return(-1);
  }

  if (compress_pack_x(&c_conf, in, inlen, out, outlen) == NULL)
	return(-1);
                             
  compress_pack_close(&c_conf, result);

  return(0);
}

static int
test3_decompress(Compress_alg unpack_alg, size_t header_len, size_t output_len,
				 unsigned char *in, size_t inlen,
				 unsigned char **out, size_t *outlen, Compress_result *result)
{
  Compress_conf d_conf;

  if (compress_configure(&d_conf, COMPRESS_UNPACK, unpack_alg,
						 NULL, NULL, NULL, 0, 0) == NULL)
	return(-1);

  d_conf.expected_output_size = output_len;
  if (header_len) {
	if (compress_set_header(&d_conf, header_len) == -1)
	  return(-1);
  }

  if (compress_unpack_x(&d_conf, in, inlen, out, outlen) == NULL)
	return(-1);

  if (compress_unpack_close(&d_conf, result) == NULL)
	return(-1);

  return(0);
}

typedef struct Test_header {
  uint32_t len;
  uint32_t alg;
  uint32_t cksum;
} Test_header;

/*
 * Run tests on application header handling using PACK_ALG.
 */
static int
run_test3(Compress_alg pack_alg)
{
  int i, st;
  unsigned char *c_text, *o_text, *p, *d_text;
  size_t c_text_len, o_text_len, d_text_len;
  size_t header_len, output_len;
  uint32_t *ui_ptr, ui_val1, ui_val2;
  Compress_alg unpack_alg;
  Compress_conf d_conf;
  Compress_result c_result, d_result;
  Ds ds;
  Test_header th, *thp;

  if (pack_alg == COMPRESS_BROTLI)
	unpack_alg = COMPRESS_BROTLI;
  else
	unpack_alg = COMPRESS_DEFAULT;

  if (pack_alg == COMPRESS_BROTLI) {
	int do_ok;

#ifdef NOTDEF
	return(0);
#endif

	do_ok = 0;
	header_len = 0;
	/* XXX brotli_compress() and/or brotli_decompress() are not working... */
	if (do_ok) {
	  /* With a zero length header, this works... */
	  ds_init(&ds);
	  ds_asprintf(&ds, "https://example.com?abcde");
	}
	else {
	  /* but this does not work. */
	  ds_init(&ds);
	  ds_asprintf(&ds, "https://example.com?abcdeabcde");
	}

	o_text = (unsigned char *) ds_buf(&ds);
	o_text_len = ds_len(&ds) - 1;
	fprintf(stderr, "o_text=\"%s\"\n", o_text);
	fprintf(stderr, "o_text_len=%lu\n", o_text_len);

	if (test3_compress(pack_alg, o_text, o_text_len,
					   header_len, &c_text, &c_text_len, &c_result) == -1) {
	brotli_err:
	  fprintf(stderr, "FAILED!\n");
	  return(-1);
	}

	output_len = o_text_len;
	if (test3_decompress(unpack_alg, header_len, output_len, c_text, c_text_len,
						 &d_text, &d_text_len, &d_result) == -1)
	  goto err;

	/* Verify our decompressed text is what we started with. */
	if (d_text_len != o_text_len || memcmp(d_text, o_text, o_text_len) != 0)
	  goto err;

	fprintf(stderr, "OK\n");
	return(0);
  }

  fprintf(stderr, "  Test3: Application header tests:\n");

  /* Use a simple URL for testing. */
  ds_init(&ds);
  ds_asprintf(&ds, "https://example.com?");
  for (i = 0; i < 26; i++) {
	ds_asprintf(&ds, "%c=%09o&%c=%09o%s",
				'a' + i, i, 'A' + i, i, (i == 25) ? "" : "&");
  }
  o_text = (unsigned char *) ds_buf(&ds);
  o_text_len = ds_len(&ds) + 1;

  /* A simple 32-bit header that will be set to the uncompressed size. */
  header_len = sizeof(uint32_t);
  ui_val1 = o_text_len;

  fprintf(stderr, "   Compressing %lu bytes with %lu byte header...",
		  o_text_len, header_len);
  if (test3_compress(pack_alg, o_text, o_text_len,
					 header_len, &c_text, &c_text_len, &c_result) == -1) {
  err:
	fprintf(stderr, "FAILED!\n");
	return(-1);
  }

  if (header_len) {
	/* Set the header value. */
	ui_ptr = (uint32_t *) c_text;
	*ui_ptr = (uint32_t) htonl(ui_val1);

	/*
	 * Now imagine c_text/c_text_len was read from a file or network...
	 * Decompress and validate.
	 * Extract the header that was prepended.
	 */
	ui_ptr = (uint32_t *) c_text;
	ui_val2 = ntohl(*(uint32_t *) ui_ptr);
	if (ui_val1 != ui_val2)
	  goto err;
	output_len = (size_t) ui_val2;
  }
  else
	output_len = ui_val2 = ui_val1;

  if (test3_decompress(unpack_alg, header_len, output_len, c_text, c_text_len,
					   &d_text, &d_text_len, &d_result) == -1)
	goto err;

  /* Verify our decompressed text is what we started with. */
  if (d_text_len != o_text_len || memcmp(d_text, o_text, o_text_len) != 0)
	goto err;

  fprintf(stderr, "OK\n");
  return(0);

  /* Begin another test. */
  th.len = ds_len(&ds) + 1;
  th.alg = pack_alg;
  th.cksum = 0xbadbadba;

  header_len = sizeof(th);
  fprintf(stderr, "   Compressing %lu bytes with %lu byte header...",
		  o_text_len, header_len);
  if (test3_compress(pack_alg, o_text, o_text_len,
					 header_len, &c_text, &c_text_len, &c_result) == -1)
	goto err;

  /* Set the header values. */
  memcpy(c_text, &th, sizeof(th));

  /*
   * Now imagine c_text/c_text_len was read from a file or network...
   * Decompress and validate.
   * Extract the header that was prepended.
   */
  thp = (Test_header *) c_text;
  if (thp->len != th.len || thp->alg != th.alg || thp->cksum != th.cksum)
	goto err;
  if (test3_decompress(unpack_alg, header_len, thp->len, c_text, c_text_len,
					   &d_text, &d_text_len, &d_result) == -1)
	goto err;

  /* Verify our decompressed text is what we started with. */
  if (d_text_len != o_text_len || memcmp(d_text, o_text, o_text_len) != 0)
	goto err;

  fprintf(stderr, "OK\n");

  return(0);
}

/*
 * Run one test, TEST, compressing using PACK_ALG, then decompressing
 * using UNPACK_ALG and comparing the result against the original
 * input to the test.
 * Return 0 if ok, -1 otherwise.
 */
static int
test1(Test *test, Compress_alg pack_alg, Compress_alg unpack_alg)
{
  int st;
  unsigned char *c_text, *o_text;
  size_t c_text_len, o_text_len, c_nalloc, d_nalloc;
  Compress_conf c_conf, d_conf;

  /* Configure and compress the test data. */
  if (compress_configure(&c_conf, COMPRESS_PACK, pack_alg,
						 NULL, NULL, NULL, 0, 0) == NULL) {
	fprintf(stderr, "test1: compress_configure(), COMPRESS_PACK failed\n");
	return(-1);
  }

  if (compress_pack_x(&c_conf, test->text, test->len + 1,
					  &c_text, &c_text_len) == NULL) {
	fprintf(stderr, "test1: compress_pack_x() failed\n");
	return(-1);
  }
  c_nalloc = c_conf.result.total_alloc;

  fprintf(stderr, "%2d. %7ld --> %7ld... ",
		  test->datanum, test->len + 1, c_text_len);

  /* Configure and decompress, then compare against the original test data. */
  if (compress_configure(&d_conf, COMPRESS_UNPACK, unpack_alg,
						 NULL, NULL, NULL, 0, 0) == NULL) {
	fprintf(stderr, "test1: compress_configure(), COMPRESS_UNPACK failed\n");
	return(-1);
  }

  if (compress_unpack_x(&d_conf, c_text, c_text_len, &o_text, &o_text_len)
	  == NULL) {
	fprintf(stderr, "test1: compress_unpack_x() failed\n");
	return(-1);
  }
  d_nalloc = d_conf.result.total_alloc;

  if (o_text_len != (test->len + 1 ) || memcmp(o_text, test->text, test->len)) {
	fprintf(stderr, "FAILED!\n");
	fprintf(stderr, "  decompressed: %ld\n", o_text_len);
	st = -1;
  }
  else {
	fprintf(stderr, "OK (%s)\n", test->name);
	fprintf(stderr, "%4s[%.1f%%",
			" ",
			100.0 * ((int) o_text_len - (int) c_text_len ) / o_text_len);
	fprintf(stderr, ", c=[%ldK:%ld, %ld/%ld/%ld]",
			c_nalloc  / 1024,
			c_conf.result.buflen,
			c_conf.result.alloc_count,
			c_conf.result.dealloc_count,
			c_conf.result.realloc_count);
	fprintf(stderr, ", d=[%ldK:%ld, %ld/%ld/%ld])]\n",
			d_nalloc / 1024,
			d_conf.result.buflen,
			d_conf.result.alloc_count,
			d_conf.result.dealloc_count,
			d_conf.result.realloc_count);
	st = 0;

	compress_pack_close(&c_conf, NULL);
	compress_unpack_close(&d_conf, NULL);
  }

  return(st);
}

static int
run_test1(Test *tests, unsigned int r, Compress_alg pack_alg,
		  Compress_alg unpack_alg)
{
  int i, st;

  fprintf(stderr, "  Test1: Compress/Decompress:\n");
  for (i = 0; tests[i].len != 0; i++) {
	if ((st = test1(&tests[i], pack_alg, unpack_alg)) == -1)
	  return(-1);
  }

  /* Test when the decompression algorithm must be automatically identified. */
  fprintf(stderr, "  Auto-unpack randomly selected dataset:\n");
  if ((st = test1(&tests[r], pack_alg, COMPRESS_DEFAULT)) == -1)
	return(-1);
  fprintf(stderr, "  OK\n");

  return(0);
}

/*
 * Prepare self test number C with name and specification NAME of LEN bytes
 * in TEST.
 */
static Test *
add_test(Test *test, int c, const char *name, size_t len)
{

  if (c == MAX_TESTS)
	return(NULL);

  test->datanum = c + 1;
  test->name = name;
  test->len = len;
  /*
   * A special case, not handled by crypto_make_random_string_from_spec(),
   * is any byte value between 0x00 and 0xff, inclusive.
   */
  if (streq(name, "[:any:]"))
	test->text = crypto_make_random_buffer(len);
  else
	test->text
	  = (unsigned char *) crypto_make_random_string_from_spec(name, len, 0);

  return(test);
}

/*
 * Run the test suite against each enabled compression library.
 */
static int
run_tests(Test *tests, int ntests)
{
  unsigned int r;
  Compress_desc *cd;

  if (ntests == 0)
	return(-1);

  /* Select a random test to repeat with different parameters. */
  crypto_random_uint(0, ntests - 1, &r);

  fprintf(stderr, "Using %d randomized buffers.\n", ntests);

  for (cd = compress_alg_begin(); cd != NULL; cd = compress_alg_next(cd)) {
	if (cd->flags != COMPRESS_ENABLED)
	  continue;
	
	fprintf(stderr, "\nTesting %s(%s) [%s]:\n",
			compress_name(cd->alg), compress_version(cd->alg), cd->aliases);

	if (do_test1 && run_test1(tests, r, cd->alg, cd->alg) == -1)
	  return(-1);

	if (do_test2 && run_test2(cd->alg) == -1)
	  return(-1);

	if (do_test3 && run_test3(cd->alg) == -1)
	  return(-1);

	if (do_test4 && run_test4(cd->alg) == -1)
	  return(-1);

	fprintf(stderr, "%s: OK\n", compress_name(cd->alg));
  }

  return(0);
}

static void NORETURN
usage(void)
{

  fprintf(stderr, "Usage: compress [flag] [file]\n");
  fprintf(stderr, "-a ALG: use algorithm ALG for compress/decompress\n");
  fprintf(stderr, "-c:     write to stdout (default)\n");
  fprintf(stderr, "-d:     decompress input\n");
  fprintf(stderr, "-debug: enable debugging output\n");
  fprintf(stderr, "-h:     show usage\n");
  fprintf(stderr, "-i:     show info for available algorithms\n");
  fprintf(stderr, "-l:     show one line info\n");
  fprintf(stderr, "-t:     run all tests (default)\n");
  fprintf(stderr, "-ta ALG: enable tests only for ALG (may repeat)\n");
  fprintf(stderr, "-t[1234]: run selected test suite (may repeat)\n");
  fprintf(stderr, "-type:  identify algorithm used by compressed input\n");
  fprintf(stderr, "-z:     compress input\n");

  exit(1);
}

int
main(int argc, char **argv)
{
  int c, do_compress, do_decompress, do_test, do_type, i, use_stdout;
  int include_long_tests, saw_alg;
  char *errmsg;
  Compress_alg alg;
  Test test[MAX_TESTS], *tp;

  if (dacs_init(DACS_STANDALONE, &argc, &argv, NULL, &errmsg) == -1) {
  fail:
    if (errmsg != NULL)
      fprintf(stderr, "dacscompress: %s\n", errmsg);

    exit(1);
  }

  set_dacs_app_name(argv[0]);

  /*
  DACS_ABORT_UTIL(("argc=%d %d xxx", argc, argc));
  */

  /* For testing. */
  dacs_set_malloc(DACS_MALLOC_PHK);
#ifdef NOTDEF
  dacs_debug_malloc(1);
#endif

  alg = COMPRESS_NONE;
  do_compress = do_decompress = 0;
  do_test = 1;
  do_test1 = do_test2 = do_test3 = do_test4 = 0;
  do_type = 0;
  use_stdout = 1;
  saw_alg = 0;

  include_long_tests = 0;

  /* Output is currently always to stdout. */
  for (i = 1; i < argc; i++) {
	if (streq(argv[i], "-a")) {
	  /* Use this algorithm to compress/decompress (not relevant for testing). */
	  if (argv[i + 1] == NULL) {
		usage();
		/*NOTREACHED*/
	  }

	  if (strcaseeq(argv[i + 1], "default"))
		alg = COMPRESS_DEFAULT;
	  else if (strcaseeq(argv[i + 1], "best"))
		alg = COMPRESS_BEST;
	  else {
		if ((alg = compress_alg(argv[i + 1])) == COMPRESS_NONE)  {
		  fprintf(stderr, "\"%s\" is not a supported algorithm.\n", argv[i + 1]);
		  usage();
		  /*NOTREACHED*/
		}
	  }
	  i++;
	}
	else if (streq(argv[i], "-c"))
	  use_stdout = 1;
	else if (streq(argv[i], "-d")) {
	  do_compress = do_test = 0;
	  do_decompress = 1;
	}
	else if (streq(argv[i], "-debug"))
	  debug_flag++;
	else if (streq(argv[i], "-h")) {
	  usage();
	  /*NOTREACHED*/
	}
	else if (streq(argv[i], "-i")) {
	  compress_info(stdout);
	  printf("Also: \"default\" and \"best\"\n");
	  exit(0);
	}
	else if (streq(argv[i], "-l")) {
	  fprintf(stderr, "%s\n", compress_alg_list());
	  exit(0);
	}
	else if (streq(argv[i], "-t")) {
	  do_compress = do_decompress = 0;
	  do_test = 1;
	}
	else if (streq(argv[i], "-t1"))
	  do_test = do_test1 = 1;
	else if (streq(argv[i], "-t2"))
	  do_test = do_test2 = 1;
	else if (streq(argv[i], "-t3"))
	  do_test = do_test3 = 1;
	else if (streq(argv[i], "-t4"))
	  do_test = do_test4 = 1;
	else if (streq(argv[i], "-ta")) {
	  Compress_desc *cd;

	  /* Enable this algorithm for testing. */
	  do_compress = do_decompress = 0;
	  do_test = 1;
	  if (argv[i + 1] == NULL) {
		usage();
		/*NOTREACHED*/
	  }

	  if (!saw_alg) {
		set_all_alg_flags(COMPRESS_DISABLED);
		saw_alg = 1;
	  }

	  if ((cd = compress_lookup_by_name(argv[i + 1])) == NULL)  {
		fprintf(stderr, "\"%s\" is not a supported algorithm.\n", argv[i + 1]);
		usage();
		/*NOTREACHED*/
	  }
	  cd->flags |= COMPRESS_ENABLED;
	  i++;
	}
	else if (streq(argv[i], "-type")) {
	  do_compress = do_decompress = do_test = 0;
	  do_type = 1;
	}
	else if (streq(argv[i], "-z")) {
	  do_test = do_decompress = 0;
	  do_compress = 1;
	}
	else
	  break;
  }

  if (argv[i] != NULL && argv[i + 1] != NULL)
	usage();

  if (!use_stdout) {
	fprintf(stderr, "Output is always to stdout.\n");
	exit(1);
  }

  if ((do_test1 + do_test2 + do_test3 + do_test4) == 0)
	do_test1 = do_test2 = do_test3 = do_test4 = 1;

  if (do_type) {
	FILE *fp;
	Compress_alg alg;

	if (argv[i] == NULL)
	  fp = stdin;
	else {
	  if ((fp = fopen(argv[i], "r")) == NULL)
		exit(1);
	}

	if ((alg = compress_type_from_stream(fp, NULL)) == COMPRESS_ERROR) {
	  fprintf(stderr, "Unknown compression type\n");
	  exit(1);
	}

	printf("%s\n", compress_name(alg));
	exit(0);
  }

  if (do_decompress) {
	size_t n;
	FILE *fp;

	if (alg == COMPRESS_NONE)
	  alg = COMPRESS_DEFAULT;

	if (argv[i] == NULL)
	  fp = stdin;
	else {
	  if ((fp = fopen(argv[i], "r")) == NULL)
		exit(1);
	}
	n = compress_funpack(fp, stdout, alg);
	if (debug_flag)
	  fprintf(stderr, "n=%ld\n", n);

	exit(0);
  }

  if (do_compress) {
	size_t n;
	FILE *fp;

	if (alg == COMPRESS_NONE)
	  alg = COMPRESS_DEFAULT;

	if (argv[i] == NULL)
	  fp = stdin;
	else {
	  if ((fp = fopen(argv[i], "r")) == NULL)
		exit(1);
	}
	n = compress_fpack(fp, stdout, alg);
	if (debug_flag)
	  fprintf(stderr, "n=%ld\n", n);

	exit(0);
  }

  fprintf(stderr, "Supported libraries:\n");
  compress_info(stderr);
  fprintf(stderr, "Generating random input, please wait...\n");
  for (i = 0; i < MAX_TESTS; i++) {
	test[i].name = NULL;
	test[i].len = 0;
	test[i].text = NULL;
  }

  /*
   * Do some tests using arbitrary but representative buffer sizes containing
   * characters randomly selected from various character sets.
   */
  tp = &test[0];
  c = 0;
  add_test(tp++, c++, "[:alpha:]", 250);
  add_test(tp++, c++, "[a-f][0-9]", 333);
  add_test(tp++, c++, "[:any:]", 651);
  add_test(tp++, c++, "[:alpha:]", 877);
  add_test(tp++, c++, "[:alnum:][:punct:]", 1025);
  add_test(tp++, c++, "[:any:]", 4096);
  add_test(tp++, c++, "[:any:]", 8193);
  add_test(tp++, c++, "[a]", 10000);
  add_test(tp++, c++, "[!-~]", 10000);

  if (include_long_tests) {
	add_test(tp++, c++, "[:alnum:][:punct:]", 403562);
	add_test(tp++, c++, "[:print:]", 403562);
	add_test(tp++, c++, "[:any:]", 403562);
	add_test(tp++, c++, "[!-~]", 1000000);
  }

  if (run_tests(test, c) == -1)
	exit(1);

  exit(0);
}
#endif
