/*
 * SHA-2 512-bit Secure Hash Functions (NIST-FIPS 180-4)
 * (excludes SHA-1, SHA-224, and SHA-256)
 */

/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * This is a revised version of the SHA-512 implementation found in
 * distributions of Samba4, Heimdal, and other packages, with the following
 * copyright and licensing terms:
 *
 * Copyright (c) 1995 - 2001 Kungliga Tekniska Högskolan
 * Copyright (c) 1999 Kungliga Tekniska Högskolan
 * Copyright (c) 2006, 2010 Kungliga Tekniska Högskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "dacs_config.h"
#include "str.h"

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "crypto_sha512.h"

#ifndef MAYBE_UNUSED
#define MAYBE_UNUSED   __attribute__ ((__unused__))
#endif

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2003-2024\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: crypto_sha512.c 3321 2024-06-19 23:39:55Z brachman $";
#endif

#undef MINIMUM
#define MINIMUM(a, b) (((a) > (b)) ? (b) : (a))

#define Ch(x, y, z)		(((x) & (y)) ^ ((~(x)) & (z)))
#define Maj(x, y, z)	(((x) & (y)) ^ ((x) & (z)) ^ ((y) & (z)))

#define ROTR(x,n)   (((x) >> (n)) | ((x) << (64 - (n))))

#define Sigma0(x)	(ROTR(x, 28) ^ ROTR(x, 34) ^ ROTR(x, 39))
#define Sigma1(x)	(ROTR(x, 14) ^ ROTR(x, 18) ^ ROTR(x, 41))
#define sigma0(x)	(ROTR(x, 1)  ^ ROTR(x, 8)  ^ ((x) >> 7))
#define sigma1(x)	(ROTR(x, 19) ^ ROTR(x, 61) ^ ((x) >> 6))

/* For debugging only... */
static MAYBE_UNUSED char *
to_hex(unsigned char *value, size_t nbytes)
{
  int i;
  char *p;
  unsigned char *v;
  static char buf[SHA512_DIGEST_BYTE_LENGTH * 2 + 1];

  if (((nbytes * 2) + 1) > sizeof(buf))
	return(NULL);

  v = value;
  p = buf;
  for (i = 0; i < nbytes; i++)
	p += sprintf(p, "%.2x", *v++ & 0xff);

  return(buf);
}

#ifdef NOTDEF
/*
 * Make sure the function isn't optimized away because the string is not
 * read after it is written.
 * See, e.g.:
 *    https://www.mail-archive.com/cryptography%40metzdowd.com/msg08369.html
 */
static void
memzap(void *s, size_t len)
{
  volatile unsigned char *ptr = (volatile unsigned char *) s;
  size_t i;

  for (i = 0; i < len; i++)
    *ptr++ = 0;
}
#endif

#ifndef U64
#if SIZEOF_UNSIGNED_LONG == 8
#define U64(C)		C##UL
#elif SIZEOF_UNSIGNED_LONG_LONG == 8
#define U64(C)		C#UUL
#else
#error "Need 64-bit constant data type"
#endif
#endif

/*
 * Return the big-endian 64-bit decoding of byte vector PP.
 */
static inline uint64_t
bendian_dec64(const void *pp)
{
  const uint8_t *p = (uint8_t const *) pp;

  return((uint64_t) (p[7]) + ((uint64_t) (p[6]) << 8)
         + ((uint64_t) (p[5]) << 16) + ((uint64_t) (p[4]) << 24)
         + ((uint64_t) (p[3]) << 32) + ((uint64_t) (p[2]) << 40)
         + ((uint64_t) (p[1]) << 48) + ((uint64_t) (p[0]) << 56));
}

int
sha512_ctx_digest_length(SHA512_ctx *m)
{

  if (m == NULL)
	return(-1);

  return((m->digest_bits + 7) / 8);
}

int
sha512_digest_length(SHA512_alg alg, unsigned int t)
{
  SHA512_ctx ctx;

  switch (alg) {
  case SHA512_224_alg:
	return(SHA512_224_DIGEST_BYTE_LENGTH);

  case SHA512_256_alg:
	return(SHA512_256_DIGEST_BYTE_LENGTH);

  case SHA512_384_alg:
	return(SHA512_384_DIGEST_BYTE_LENGTH);

  case SHA512_512t_alg:
	if (sha512t_init(&ctx, t) == -1)
	  return(-1);
	return((t + 7) / 8);

  case SHA512_512_alg:
	return(SHA512_DIGEST_BYTE_LENGTH);

  default:
	abort();
	/*NOTREACHED*/
	return(-1);
  }
}

static const uint64_t constant_512[80] = {
  0x428a2f98d728ae22ULL, 0x7137449123ef65cdULL,
  0xb5c0fbcfec4d3b2fULL, 0xe9b5dba58189dbbcULL,
  0x3956c25bf348b538ULL, 0x59f111f1b605d019ULL,
  0x923f82a4af194f9bULL, 0xab1c5ed5da6d8118ULL,
  0xd807aa98a3030242ULL, 0x12835b0145706fbeULL,
  0x243185be4ee4b28cULL, 0x550c7dc3d5ffb4e2ULL,
  0x72be5d74f27b896fULL, 0x80deb1fe3b1696b1ULL,
  0x9bdc06a725c71235ULL, 0xc19bf174cf692694ULL,
  0xe49b69c19ef14ad2ULL, 0xefbe4786384f25e3ULL,
  0x0fc19dc68b8cd5b5ULL, 0x240ca1cc77ac9c65ULL,
  0x2de92c6f592b0275ULL, 0x4a7484aa6ea6e483ULL,
  0x5cb0a9dcbd41fbd4ULL, 0x76f988da831153b5ULL,
  0x983e5152ee66dfabULL, 0xa831c66d2db43210ULL,
  0xb00327c898fb213fULL, 0xbf597fc7beef0ee4ULL,
  0xc6e00bf33da88fc2ULL, 0xd5a79147930aa725ULL,
  0x06ca6351e003826fULL, 0x142929670a0e6e70ULL,
  0x27b70a8546d22ffcULL, 0x2e1b21385c26c926ULL,
  0x4d2c6dfc5ac42aedULL, 0x53380d139d95b3dfULL,
  0x650a73548baf63deULL, 0x766a0abb3c77b2a8ULL,
  0x81c2c92e47edaee6ULL, 0x92722c851482353bULL,
  0xa2bfe8a14cf10364ULL, 0xa81a664bbc423001ULL,
  0xc24b8b70d0f89791ULL, 0xc76c51a30654be30ULL,
  0xd192e819d6ef5218ULL, 0xd69906245565a910ULL,
  0xf40e35855771202aULL, 0x106aa07032bbd1b8ULL,
  0x19a4c116b8d2d0c8ULL, 0x1e376c085141ab53ULL,
  0x2748774cdf8eeb99ULL, 0x34b0bcb5e19b48a8ULL,
  0x391c0cb3c5c95a63ULL, 0x4ed8aa4ae3418acbULL,
  0x5b9cca4f7763e373ULL, 0x682e6ff3d6b2b8a3ULL,
  0x748f82ee5defb2fcULL, 0x78a5636f43172f60ULL,
  0x84c87814a1f0ab72ULL, 0x8cc702081a6439ecULL,
  0x90befffa23631e28ULL, 0xa4506cebde82bde9ULL,
  0xbef9a3f7b2c67915ULL, 0xc67178f2e372532bULL,
  0xca273eceea26619cULL, 0xd186b8c721c0c207ULL,
  0xeada7dd6cde0eb1eULL, 0xf57d4f7fee6ed178ULL,
  0x06f067aa72176fbaULL, 0x0a637dc5a2c898a6ULL,
  0x113f9804bef90daeULL, 0x1b710b35131c471bULL,
  0x28db77f523047d84ULL, 0x32caab7b40c72493ULL,
  0x3c9ebe0a15c9bebcULL, 0x431d67c49c100d4cULL,
  0x4cc5d4becb3e42b6ULL, 0x597f299cfc657e2aULL,
  0x5fcb6fab3ad6faecULL, 0x6c44198c4a475817ULL
};

#define H0(M)	M->counter[0]
#define H1(M)	M->counter[1]
#define H2(M)	M->counter[2]
#define H3(M)	M->counter[3]
#define H4(M)	M->counter[4]
#define H5(M)	M->counter[5]
#define H6(M)	M->counter[6]
#define H7(M)	M->counter[7]

static void
calc(SHA512_ctx *m, uint64_t *in)
{
  int i;
  uint64_t AA, BB, CC, DD, EE, FF, GG, HH;
  uint64_t data[80];

  AA = H0(m);
  BB = H1(m);
  CC = H2(m);
  DD = H3(m);
  EE = H4(m);
  FF = H5(m);
  GG = H6(m);
  HH = H7(m);

  for (i = 0; i < 16; ++i)
	data[i] = in[i];

  for (i = 16; i < 80; ++i)
	data[i] = sigma1(data[i - 2]) + data[i - 7]
	  + sigma0(data[i - 15]) + data[i - 16];

  for (i = 0; i < 80; i++) {
	uint64_t T1, T2;

	T1 = HH + Sigma1(EE) + Ch(EE, FF, GG) + constant_512[i] + data[i];
	T2 = Sigma0(AA) + Maj(AA,BB,CC);

	HH = GG;
	GG = FF;
	FF = EE;
	EE = DD + T1;
	DD = CC;
	CC = BB;
	BB = AA;
	AA = T1 + T2;
  }

  H0(m) += AA;
  H1(m) += BB;
  H2(m) += CC;
  H3(m) += DD;
  H4(m) += EE;
  H5(m) += FF;
  H6(m) += GG;
  H7(m) += HH;
}

/*
 * From `Performance analysis of MD5' by Joseph D. Touch <touch@isi.edu>.
 */

#if !defined(DACS_OS_IS_BIG_ENDIAN)
static inline uint32_t
cshift(uint32_t x, unsigned int n)
{

    return((x << n) | (x >> (32 - n)));
}

static inline uint64_t
cshift64(uint64_t x, unsigned int n)
{
  return((uint64_t) x << (uint64_t) n)
	| ((uint64_t) x >> ((uint64_t) 64 - (uint64_t) n));
}

static inline uint64_t
swap_uint64_t(uint64_t t)
{
  uint64_t temp;

  temp = cshift64(t, 32);
  temp = ((temp & 0xff00ff00ff00ff00ULL) >> 8)
	| ((temp & 0x00ff00ff00ff00ffULL) << 8);

  return((temp & 0xffff0000ffff0000ULL) >> 16)
	| ((temp & 0x0000ffff0000ffffULL) << 16);
}

struct x64 {
  uint64_t a;
  uint64_t b;
};
#endif

/*
 * Start or continue hashing using LEN bytes of V.
 */
void
sha512_update(SHA512_ctx *m, const void *v, size_t len)
{
  const unsigned char *p = v;
  size_t old_sz = m->sz[0];
  size_t offset;

  m->sz[0] += len * 8;
  if (m->sz[0] < old_sz)
	++m->sz[1];
  offset = (old_sz / 8) % 128;
  while (len > 0){
	size_t l = MINIMUM(len, 128 - offset);

	memcpy(m->save + offset, p, l);
	offset += l;
	p += l;
	len -= l;
	if (offset == 128){
#if !defined(DACS_OS_IS_BIG_ENDIAN)
	  int i;
	  uint64_t current[16];
	  struct x64 *us = (struct x64 *) m->save;

	  for (i = 0; i < 8; i++){
		current[2 * i + 0] = swap_uint64_t(us[i].a);
		current[2 * i + 1] = swap_uint64_t(us[i].b);
	  }
	  calc(m, current);
#else
	  calc(m, (uint64_t *) m->save);
#endif
	  offset = 0;
	}
  }
}

/*
 * Finish the current hash and put the result in RES, which must be
 * at least SHA512_512_DIGEST_BYTE_LENGTH bytes.
 * N.B. The output of this function is ALWAYS 64 bytes, regardless of which
 * variant of SHA512 has been requested.
 *
 * If M->alg is SHA512t_hash, then the caller should use the t leftmost
 * (most significant) bits of RES.  The hash will be (M->digest_bits + 7) / 8
 * bytes long, with (M->digest_bits % 8) significant bits in the last byte,
 * unless that remainder is zero, in which case all 8 bits of the last byte are
 * significant.
 */
unsigned char  *
sha512_final(SHA512_ctx *m, void *res)
{
  unsigned char zeros[128 + 16];
  unsigned int offset = (m->sz[0] / 8) % 128;
  unsigned int dstart = (240 - offset - 1) % 128 + 1;

  *zeros = 0x80;
  memset(zeros + 1, 0, sizeof(zeros) - 1);
  zeros[dstart + 15] = (m->sz[0] >>  0) & 0xff;
  zeros[dstart + 14] = (m->sz[0] >>  8) & 0xff;
  zeros[dstart + 13] = (m->sz[0] >> 16) & 0xff;
  zeros[dstart + 12] = (m->sz[0] >> 24) & 0xff;
  zeros[dstart + 11] = (m->sz[0] >> 32) & 0xff;
  zeros[dstart + 10] = (m->sz[0] >> 40) & 0xff;
  zeros[dstart +  9] = (m->sz[0] >> 48) & 0xff;
  zeros[dstart +  8] = (m->sz[0] >> 56) & 0xff;

  zeros[dstart +  7] = (m->sz[1] >>  0) & 0xff;
  zeros[dstart +  6] = (m->sz[1] >>  8) & 0xff;
  zeros[dstart +  5] = (m->sz[1] >> 16) & 0xff;
  zeros[dstart +  4] = (m->sz[1] >> 24) & 0xff;
  zeros[dstart +  3] = (m->sz[1] >> 32) & 0xff;
  zeros[dstart +  2] = (m->sz[1] >> 40) & 0xff;
  zeros[dstart +  1] = (m->sz[1] >> 48) & 0xff;
  zeros[dstart +  0] = (m->sz[1] >> 56) & 0xff;

  sha512_update(m, zeros, dstart + 16);

  {
	int i;
	unsigned char *r = (unsigned char *) res;

	/*
	 * This copies all 64 bytes even though the caller may not care
	 * about all of them.
	 */
	for (i = 0; i < 8; i++) {
	  r[8 * i + 7] = m->counter[i] & 0xff;
	  r[8 * i + 6] = (m->counter[i] >> 8) & 0xff;
	  r[8 * i + 5] = (m->counter[i] >> 16) & 0xff;
	  r[8 * i + 4] = (m->counter[i] >> 24) & 0xff;
	  r[8 * i + 3] = (m->counter[i] >> 32) & 0xff;
	  r[8 * i + 2] = (m->counter[i] >> 40) & 0xff;
	  r[8 * i + 1] = (m->counter[i] >> 48) & 0xff;
	  r[8 * i + 0] = (m->counter[i] >> 56) & 0xff;
	}
  }

  return(res);
}

/*
 * This is identical to sha512t_init(m, 224), but more efficient.
 * N.B. This is not the same algorithm as SHA-224.
 */
void
sha512_224_init(SHA512_ctx *m)
{

  m->alg = SHA512_224_alg;
  m->digest_bits = SHA512_224_DIGEST_BIT_LENGTH;
  m->sz[0] = 0;
  m->sz[1] = 0;

  H0(m) = 0x8c3d37c819544da2ULL;
  H1(m)	= 0x73e1996689dcd4d6ULL;
  H2(m) = 0x1dfab7ae32ff9c82ULL;
  H3(m) = 0x679dd514582f9fcfULL;
  H4(m) = 0x0f6d2b697bd44da8ULL;
  H5(m) = 0x77e36f7304c48942ULL;
  H6(m) = 0x3f9d85a86a1d36c8ULL;
  H7(m) = 0x1112e6ad91d692a1ULL;
}

/*
 * Destroy all potentially sensitive data in M, leaving only the algorithm
 * identifier and its hash size.
 */
void
sha512_ctx_zap(SHA512_ctx *m)
{

  memzap(m->sz, sizeof(m->sz));
  memzap(m->counter, sizeof(m->counter));
  memzap(m->save, sizeof(m->save));
}

unsigned char *
sha512_224_final(SHA512_ctx *m, void *res)
{
  unsigned char buf[SHA512_DIGEST_BYTE_LENGTH];

  sha512_final(m, buf);
  memcpy(res, buf, sha512_ctx_digest_length(m));

  sha512_ctx_zap(m);
  return(res);
}

/*
 * This is identical to sha512t_init(m, 256), but more efficient.
 * N.B. This is not the same algorithm as SHA-256.
 */
void
sha512_256_init(SHA512_ctx *m)
{

  m->alg = SHA512_256_alg;
  m->digest_bits = SHA512_256_DIGEST_BIT_LENGTH;
  m->sz[0] = 0;
  m->sz[1] = 0;

  H0(m) = 0x22312194fc2bf72cULL;
  H1(m) = 0x9f555fa3c84c64c2ULL;
  H2(m) = 0x2393b86b6f53b151ULL;
  H3(m) = 0x963877195940eabdULL;
  H4(m) = 0x96283ee2a88effe3ULL;
  H5(m) = 0xbe5e1e2553863992ULL;
  H6(m) = 0x2b0199fc2c85b8aaULL;
  H7(m) = 0x0eb72ddc81c52ca2ULL;
}

unsigned char *
sha512_256_final(SHA512_ctx *m, void *res)
{
  unsigned char buf[SHA512_DIGEST_BYTE_LENGTH];

  sha512_final(m, buf);
  memcpy(res, buf, sha512_ctx_digest_length(m));

  sha512_ctx_zap(m);
  return((unsigned char *) res);
}

void
sha512_384_init(SHA512_ctx *m)
{

  m->alg = SHA512_384_alg;
  m->digest_bits = SHA512_384_DIGEST_BIT_LENGTH;
  m->sz[0] = 0;
  m->sz[1] = 0;

  H0(m) = 0xcbbb9d5dc1059ed8ULL;
  H1(m) = 0x629a292a367cd507ULL;
  H2(m) = 0x9159015a3070dd17ULL;
  H3(m) = 0x152fecd8f70e5939ULL;
  H4(m) = 0x67332667ffc00b31ULL;
  H5(m) = 0x8eb44a8768581511ULL;
  H6(m) = 0xdb0c2e0d64f98fa7ULL;
  H7(m) = 0x47b5481dbefa4fa4ULL;
}

unsigned char *
sha512_384_final(SHA512_ctx *m, void *res)
{
  unsigned char buf[SHA512_DIGEST_BYTE_LENGTH];

  sha512_final(m, buf);
  memcpy(res, buf, sha512_ctx_digest_length(m));

  sha512_ctx_zap(m);
  return((unsigned char *) res);
}

unsigned char *
sha512_512_final(SHA512_ctx *m, void *res)
{

  sha512_final(m, res);

  return((unsigned char *) res);
}

void
sha512_512_init(SHA512_ctx *m)
{

  m->alg = SHA512_512_alg;
  m->digest_bits = SHA512_DIGEST_BIT_LENGTH;
  m->sz[0] = 0;
  m->sz[1] = 0;

  H0(m) = 0x6a09e667f3bcc908ULL;
  H1(m) = 0xbb67ae8584caa73bULL;
  H2(m) = 0x3c6ef372fe94f82bULL;
  H3(m) = 0xa54ff53a5f1d36f1ULL;
  H4(m) = 0x510e527fade682d1ULL;
  H5(m) = 0x9b05688c2b3e6c1fULL;
  H6(m) = 0x1f83d9abfb41bd6bULL;
  H7(m) = 0x5be0cd19137e2179ULL;
}

void
sha512_init(SHA512_ctx *m)
{

  sha512_512_init(m);
}

/*
 * This is SHA-512 with truncation of the hash.
 * T is the digest length in bits.
 * NIST-FIPS.180-4 5.3.6
 */
int
sha512t_init(SHA512_ctx *m, unsigned int t)
{
  int i;
  uint64_t *ui64p;
  size_t iv_len;
  char iv_name_hash[12];
  unsigned char h0[64];

  /* Reject invalid truncation lengths (NIST-FIPS.180-4 5.3.6). */
  if (t == 0 || t >= 512 || t == 384)
	return(-1);

  iv_len = snprintf(iv_name_hash, sizeof(iv_name_hash), "SHA-512/%u", t);
  if (iv_len > sizeof(iv_name_hash))
	return(-1);

  m->alg = SHA512_512t_alg;
  m->digest_bits = t;
  m->sz[0] = 0;
  m->sz[1] = 0;

  H0(m) = U64(0x6a09e667f3bcc908) ^ U64(0xa5a5a5a5a5a5a5a5);
  H1(m) = U64(0xbb67ae8584caa73b) ^ U64(0xa5a5a5a5a5a5a5a5);
  H2(m) = U64(0x3c6ef372fe94f82b) ^ U64(0xa5a5a5a5a5a5a5a5);
  H3(m) = U64(0xa54ff53a5f1d36f1) ^ U64(0xa5a5a5a5a5a5a5a5);
  H4(m) = U64(0x510e527fade682d1) ^ U64(0xa5a5a5a5a5a5a5a5);
  H5(m) = U64(0x9b05688c2b3e6c1f) ^ U64(0xa5a5a5a5a5a5a5a5);
  H6(m) = U64(0x1f83d9abfb41bd6b) ^ U64(0xa5a5a5a5a5a5a5a5);
  H7(m) = U64(0x5be0cd19137e2179) ^ U64(0xa5a5a5a5a5a5a5a5);

  sha512_update(m, iv_name_hash, iv_len);
  sha512_final(m, h0);

  ui64p = (uint64_t *) h0;
  for (i = 0; i < 8; i++)
	m->counter[i] = bendian_dec64(&ui64p[i]);

  m->sz[0] = 0;
  m->sz[1] = 0;

  return(0);
}

/*
 * Initialize M for digest algorithm ALG.
 *
 * No optional SHA512_ctx allocation (if M == NULL) needed so far.
 */
int
sha512_init_alg(SHA512_ctx *m, SHA512_alg alg)
{
  int rc;

  rc = 0;
  switch (alg) {
  case SHA512_224_alg:
	sha512_224_init(m);
	break;

  case SHA512_256_alg:
	sha512_256_init(m);
	break;

  case SHA512_384_alg:
	sha512_384_init(m);
	break;

  case SHA512_512_alg:
	sha512_512_init(m);
	break;

  default:
	rc = -1;
  }

  return(rc);
}

/*
 * One-shot application of SHA512/224.
 */
unsigned char *
sha512_224(const void *input, size_t input_len, void *result)
{
  SHA512_ctx ctx;

  sha512_224_init(&ctx);

  sha512_update(&ctx, input, input_len);

  sha512_224_final(&ctx, result);
  sha512_ctx_zap(&ctx);

  return(result);
}

/*
 * One-shot application of SHA512/256.
 */
unsigned char *
sha512_256(const void *input, size_t input_len, void *result)
{
  SHA512_ctx ctx;

  sha512_256_init(&ctx);

  sha512_update(&ctx, input, input_len);

  sha512_256_final(&ctx, result);

  sha512_ctx_zap(&ctx);
  return(result);
}

/*
 * One-shot application of SHA512/384.
 */
unsigned char *
sha512_384(const void *input, size_t input_len, void *result)
{
  SHA512_ctx ctx;

  sha512_384_init(&ctx);

  sha512_update(&ctx, input, input_len);

  sha512_384_final(&ctx, result);

  sha512_ctx_zap(&ctx);
  return(result);
}

/*
 * One-shot application of SHA512/t.
 * Caller should use the t leftmost bits.
 */
unsigned char *
sha512_t(unsigned int t, const void *input, size_t input_len, void *result)
{
  SHA512_ctx ctx;

  if (sha512t_init(&ctx, t) == -1)
	return(NULL);

  sha512_update(&ctx, input, input_len);

  sha512_final(&ctx, result);
  sha512_ctx_zap(&ctx);

  return(result);
}

/*
 * One-shot application of SHA512.
 */
unsigned char *
sha512_512(const void *input, size_t input_len, void *result)
{
  SHA512_ctx ctx;

  sha512_512_init(&ctx);
  sha512_update(&ctx, input, input_len);

  sha512_final(&ctx, result);
  sha512_ctx_zap(&ctx);

  return(result);
}

unsigned char *
sha512(const void *input, size_t input_len, void *result)
{

  return(sha512_512(input, input_len, result));
}

/*
 * Given SHA-512t hash value RES and its T, return the number of
 * complete significant bytes (NBYTES), the number of significant "leftover" bits
 * (NBITS), and the leftover bits extracted (masked and shifted).
 *
 * XXX have not been able to find any other implementation or published
 * test vectors for a T non-divisible by 8 case.
 */
int
sha512_t_result(unsigned char *res, unsigned int t, int *nbytes, int *nbits,
				int *bitval)
{
  unsigned char *p;

  if (res == NULL || t == 0 || t >=512 || t == 384)
	return(-1);

  /* The number of complete hash bytes in RES. */
  *nbytes = t / 8;

  /* The number of signficant hash bits in the last partial byte of RES. */
  *nbits = t % 8;

  if (*nbits) {
	/* Point to the last partial byte of RES. */
	p = res + *nbytes;

	/* Extract the leftmost significant bits. */
	*bitval = (*p & 0xff) >> (8 - *nbits);
  }
  else
	*bitval = 0;

  return(0);
}

static MAYBE_UNUSED void
dump_regs(uint64_t *ptr, int count)
{
  int i;

  for (i = 0; i < count; i++)
	printf("reg[%d]: 0x%016llx\n", i, (long long unsigned int) ptr[i]);
}

/* Test vectors. */

typedef struct SHA512t_test {
  SHA512_alg alg;
  unsigned int t;
  char *msg;
  char *digest_value;
} SHA512t_test;

/*
 * Test vectors:
 * https://csrc.nist.gov/groups/ST/toolkit/documents/Examples/SHA512_224.pdf
 * https://csrc.nist.gov/groups/ST/toolkit/documents/Examples/SHA512_256.pdf
 * https://github.com/ottodietz/sha512t/blob/master/sha512t.py
 * Also by using "Digest::SHA" for t==224 and t==256.
 *
 * XXX there are not many published test vectors for SHA512/t (where t is not
 * 224 or 256) and there are few implementations, and some restrict t to a
 * multiple of 8 bits.
 *
 * Also see:
 * https://github.com/haskell-hvr/cryptohash-sha512/blob/master/src-tests/test-sha512t.hs
 * https://csrc.nist.gov/Projects/Cryptographic-Algorithm-Validation-Program/Secure-Hashing#shavs
 */
static SHA512t_test sha512t_test[] = {
  { SHA512_512t_alg, 4, "1234", "d9" },
  { SHA512_512t_alg, 1, "abcdefghijk", "84" },

  { SHA512_512t_alg, 224, "",
    "6ed0dd02806fa89e25de060c19d3ac86cabb87d6a0ddd05c333b84f4" },
  { SHA512_512t_alg, 224, "abc",
	"4634270f707b6a54daae7530460842e20e37ed265ceee9a43e8924aa" },
  { SHA512_512t_alg, 224, "Hi there",
	"9aab5d13e3177bbb44e5aa0b20ec551e3a9633baae1f38e2df55ae4d" },
  { SHA512_512t_alg, 224, "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
	"23fec5bb94d60b23308192640b0c453335d664734fe40e7268674af9" },

  { SHA512_512t_alg, 256, "",
    "c672b8d1ef56ed28ab87c3622c5114069bdd3ad7b8f9737498d0c01ecef0967a" },
  { SHA512_512t_alg, 256, "abc",
	"53048e2681941ef99b2e29b76b4c7dabe4c2d0c634fc6d46e0e2f13107e7af23" },
  { SHA512_512t_alg, 256, "Hi there",
	"407682efd23550957fc62f31087b1c4938bcbc1de36b88d063ad645a323343d2" },
  { SHA512_512t_alg, 256, "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
	"3928e184fb8690f840da3988121d31be65cb9d3ef83ee6146feac861e19b563a" },

  { SHA512_224_alg, 0, "",
    "6ed0dd02806fa89e25de060c19d3ac86cabb87d6a0ddd05c333b84f4" },
  { SHA512_224_alg, 0, "abc",
	"4634270f707b6a54daae7530460842e20e37ed265ceee9a43e8924aa" },
  { SHA512_224_alg, 0, "Hi there",
	"9aab5d13e3177bbb44e5aa0b20ec551e3a9633baae1f38e2df55ae4d" },
  { SHA512_224_alg, 0, "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
	"23fec5bb94d60b23308192640b0c453335d664734fe40e7268674af9" },

  { SHA512_256_alg, 0, "",
    "c672b8d1ef56ed28ab87c3622c5114069bdd3ad7b8f9737498d0c01ecef0967a" },
  { SHA512_256_alg, 0, "abc",
	"53048e2681941ef99b2e29b76b4c7dabe4c2d0c634fc6d46e0e2f13107e7af23" },
  { SHA512_256_alg, 0, "Hi there",
	"407682efd23550957fc62f31087b1c4938bcbc1de36b88d063ad645a323343d2" },
  { SHA512_256_alg, 0, "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
	"3928e184fb8690f840da3988121d31be65cb9d3ef83ee6146feac861e19b563a" },

  { SHA512_384_alg, 0, "abc",
    "cb00753f45a35e8bb5a03d699ac65007272c32ab0eded1631a8b605a43ff5bed8086072ba1e7cc2358baeca134c825a7" },
  { SHA512_384_alg, 0, "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
	"09330c33f71147e83d192fc782cd1b4753111b173b3b05d22fa08086e3b0f712fcc7c71a557e2db966c3e9fa91746039" },

  { SHA512_512t_alg, 8, "000000", "40" },
  { SHA512_512t_alg, 8, "123456", "e6" },
  { SHA512_512t_alg, 8, "uvwxyz", "1c" },

  { SHA512_512t_alg, 16, "000000", "1893" },
  { SHA512_512t_alg, 16, "123456", "ed51" },
  { SHA512_512t_alg, 16, "uvwxyz", "d4f3" },

  { SHA512_512t_alg, 24, "000000", "c36145" },
  { SHA512_512t_alg, 24, "123456", "322e02" },
  { SHA512_512t_alg, 24, "uvwxyz", "b82f96" },

  { SHA512_512t_alg, 32, "000000", "0fb87900" },
  { SHA512_512t_alg, 32, "123456", "b01b3f58" },
  { SHA512_512t_alg, 32, "uvwxyz", "a7802f1c" },

  { SHA512_512t_alg, 40, "000000", "41561e6db1" },
  { SHA512_512t_alg, 40, "123456", "6d92e23336" },
  { SHA512_512t_alg, 40, "uvwxyz", "12807bfe4b" },

  { SHA512_512t_alg, 48, "000000", "e71453add8ee" },
  { SHA512_512t_alg, 48, "123456", "38ebd7973918" },
  { SHA512_512t_alg, 48, "uvwxyz", "d444c346a738" },

  { SHA512_512t_alg, 56, "000000", "34acd645778a15" },
  { SHA512_512t_alg, 56, "123456", "b49acfda08675b" },
  { SHA512_512t_alg, 56, "uvwxyz", "350690fa30f57b" },

  { SHA512_512t_alg, 64, "000000", "0c8beffa2e95addd" },
  { SHA512_512t_alg, 64, "123456", "52261eacb45fe456" },
  { SHA512_512t_alg, 64, "uvwxyz", "ecfacd8d22e401c6" },

  { SHA512_512t_alg, 72, "000000", "857a80e7286803cf57" },
  { SHA512_512t_alg, 72, "123456", "2dc46cba92f277e7ee" },
  { SHA512_512t_alg, 72, "uvwxyz", "804886d65bb10ba6b9" },

  { SHA512_512t_alg, 80, "000000", "21fb1ba1eb71f834d109" },
  { SHA512_512t_alg, 80, "123456", "eeb1e73ed321fb324774" },
  { SHA512_512t_alg, 80, "uvwxyz", "008184950f9c830e7fdc" },

  { SHA512_512t_alg, 88, "000000", "e382856f8019363544df4e" },
  { SHA512_512t_alg, 88, "123456", "0689adfdb5dc251ebee870" },
  { SHA512_512t_alg, 88, "uvwxyz", "298eaa31c4ecf5f5963d09" },

  { SHA512_512t_alg, 96, "000000", "189135861b522f62756400b2" },
  { SHA512_512t_alg, 96, "123456", "7faeba0512326c125ddc89f7" },
  { SHA512_512t_alg, 96, "uvwxyz", "555e97a5903df06d2f818a19" },

  { SHA512_512t_alg, 104, "000000", "b7a5c6dbfad0d3c9d77560b9ea" },
  { SHA512_512t_alg, 104, "123456", "7fa9d643798cb0605eb8b07311" },
  { SHA512_512t_alg, 104, "uvwxyz", "a3d0a1dce670fe231c80a2fbfb" },

  { SHA512_512t_alg, 112, "000000", "b61089a3983050f4fbf8687d6fbf" },
  { SHA512_512t_alg, 112, "123456", "c415857e3a4d15cac1439003f386" },
  { SHA512_512t_alg, 112, "uvwxyz", "36f690af14514051741107661876" },

  { SHA512_512t_alg, 120, "000000", "44e707b93fd9187f8d9a80feb80e23" },

  { SHA512_INVALID_alg, 0, NULL, NULL }
};

int
do_crypto_sha512_tests(void)
{
  int i;
  char *val;
  SHA512t_test *t2;

  /*
   * FIPS-180-4 5.3.6, 6.6, 6.7
   * https://doi.org/10.6028/NIST.FIPS.180-4
   */
  for (i = 0, t2 = &sha512t_test[0]; t2->alg != SHA512_INVALID_alg; t2++, i++) {
	fprintf(stderr, "%2d. ", i + 1);
	if (t2->alg == SHA512_224_alg) {
	  unsigned char result[SHA512_224_DIGEST_BYTE_LENGTH];

	  fprintf(stderr, "SHA512/224(\"%s\"):", t2->msg);
	  sha512_224(t2->msg, strlen(t2->msg), result);
	  val = to_hex(result, SHA512_224_DIGEST_BYTE_LENGTH);
	}
	else if (t2->alg == SHA512_256_alg) {
	  unsigned char result[SHA512_256_DIGEST_BYTE_LENGTH];

	  fprintf(stderr, "SHA512/256(\"%s\"):", t2->msg);
	  sha512_256(t2->msg, strlen(t2->msg), result);
	  val = to_hex(result, SHA512_256_DIGEST_BYTE_LENGTH);
	}
	else if (t2->alg == SHA512_384_alg) {
	  unsigned char result[SHA512_384_DIGEST_BYTE_LENGTH];

	  fprintf(stderr, "SHA512/384(\"%s\"):", t2->msg);
	  sha512_384(t2->msg, strlen(t2->msg), result);
	  val = to_hex(result, SHA512_384_DIGEST_BYTE_LENGTH);
	}
	else if (t2->alg == SHA512_512t_alg) {
	  int nbytes, nbits, bitval;
	  unsigned char result[SHA512_512_DIGEST_BYTE_LENGTH];
	  unsigned char *p;

	  fprintf(stderr, "SHA-512/%u(\"%s\"):", t2->t, t2->msg);
	  if (sha512_t(t2->t, t2->msg, strlen(t2->msg), result) == NULL)
		goto failed;

	  if (t2->t % 8) {
		char bitbuf1[1 * 8 + 1], bitbuf2[1 * 8 + 1], *p1, *p2;

		if (sha512_t_result(result, t2->t, &nbytes, &nbits, &bitval) == -1)
		  goto failed;
		p = result + nbytes;

#ifdef NOTDEF
		fprintf(stderr, "\n  last byte=0x%02x (%08b), %d hashbit%s=%b\n",
				*p & 0xff, *p & 0xff, nbits, (nbits == 1) ? "" : "s", bitval);
#endif
		/*
		 * Turns out... some printf(3) implementations do not grok %b.
		 */
		strui2b((unsigned char *) p, 1, 0, bitbuf1, &p1);
		strui2b((unsigned char *) &bitval, 1, 0, bitbuf2, &p2);
		fprintf(stderr, "\n  last byte=0x%02x (%s), %d hashbit%s=%s ",
				*p & 0xff, p1, nbits, (nbits == 1) ? "" : "s", p2);
	  }
	  val = to_hex(result, sha512_digest_length(SHA512_512t_alg, t2->t));
	}
	else {
	  fprintf(stderr, "Unrecognized SHA512t test\n");
	  return(-1);
	}

	if (val == NULL) {
	failed:
	  fprintf(stderr, " FAILED\n");
	  return(-1);
	}

	fprintf(stderr, "\n%s ", val);
	if (strcmp(val, t2->digest_value)) {
	  fprintf(stderr, "FAILED\n");
	  return(-1);
	}
	fprintf(stderr, "ok\n");
  }

  return(0);
}

#ifdef PROG

int
main(int argc, char **argv)
{
  unsigned char result[SHA512_DIGEST_BYTE_LENGTH];
  SHA512_ctx ctx;

  if (argc == 2 && !strcmp(argv[1], "quick"))
	goto do_quick;

  fprintf(stderr, "Doing FIPS.180-4 SHA-384/SHA-512 tests\n");
  if (do_crypto_sha512_tests() != 0) {
	fprintf(stderr, "FAILED!\n");
	exit(1);
  }

  fprintf(stderr, "All tests succeeded.\n");

  exit(0);

 do_quick:
  sha512_224_init(&ctx);
  sha512_update(&ctx, "", 0);
  sha512_final(&ctx, result);
  printf("%s\n", to_hex(result, sha512_ctx_digest_length(&ctx)));

  sha512_224_init(&ctx);
  sha512_update(&ctx, "abc", 3);
  sha512_final(&ctx, result);
  printf("%s\n", to_hex(result, sha512_ctx_digest_length(&ctx)));

  sha512_init(&ctx);
  sha512_256_init(&ctx);
  sha512_224_init(&ctx);
  sha512t_init(&ctx, 8);
  sha512t_init(&ctx, 224);

  sha512_update(&ctx, "abc", 3);
  sha512_update(&ctx, "000000", 6);
  sha512_update(&ctx, "Hi there", 8);
  sha512_final(&ctx, result);
  printf("%s\n", to_hex(result, ctx.digest_bits  / 8));

  sha512t_init(&ctx, 224);
  sha512_update(&ctx, "abc", 3);
  sha512_final(&ctx, result);
  printf("%s\n", to_hex(result, ctx.digest_bits  / 8));

  exit(0);
}
#endif
