
/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * This is a malloc(3) implementation selector layer.
 * Its purpose is to allow the normal system library malloc(3) implementation
 * to be replaced by another implementation, or to make one or more
 * different implementations available in addition to the system's.
 *
 * Why might this be needed?
 * To use an alternate implementation (usually from source code):
 * o for source-level debugging of the application or the new malloc
 * o to enable malloc performance profiling
 * o to provide an application with a custom implementation that is more
 *   time or space efficient than the system's for its particular
 *   allocation/deallocation behaviour.
 * o that is thread-safe
 * o to help design and test a new malloc library implemention
 * o to add features or make API changes to an implementation
 *
 * There are two main approaches:
 * A) Total or nearly-total transparent replacement
 *    Most calls to malloc(), free(), etc. invoke a specified corresponding
 *    application-supplied function that can then provide the new
 *    equivalent implementation or invoke the system library implementation.
 *    Certain calls at startup time cannot be "wrapped" in this way.
 *   
 *   1) Symbol wrapping through the linker's --wrap option
 *     [FreeBSD/GNU-Linux systems]
 *
 *   2) Function interposing through the DYLD_INTERPOSE macro
 *     [MacOS]
 *
 * B) Application visible extension
 *    The new library's memory management functions have different
 *    names than the system library's; e.g., my_malloc(), my_free(), ...
 *    Calls to malloc() etc. from library functions (e.g., gethostbyname(3)
 *    or getpwent(3)) will invoke the
 *    system malloc normally, but calls at the application level (including
 *    libraries suitably altered and compiled to call the new functions)
 *    will use the new library functions.
 *    Some library functions (endhostent(3), endpwent(3)) will
 *    call free(3).
 *    This will not use the new library for all of the process's memory
 *    management, but may be sufficient for many applications.
 *
 * Regarding A:
 * Although replacing the system malloc implementation is perhaps the
 * main use case for this wrapping feature, it is difficult (or impossible,
 * in the general case) to do safely - without introducing the potential
 * for runtime memory corruption and bugs.
 * This is because to avoid introducing a bug, all relevant system malloc
 * APIs must be known, reimplemented, and wrapped, otherwise a system malloc
 * API might be passed memory allocated by a different malloc implementation.
 * It is not necessarily the case that all system malloc APIs
 * are known and/or documented and/or standard, making it difficult for the
 * writer of the new implementation to perform the necessary wrappings.
 * It is possible, for instance, for some system malloc function to call
 * an undocumented system malloc function; if it happens to pass a memory
 * pointer obtained from the new implementation, the process is likely to crash.
 * The system malloc may change at any time from bug fixes or system upgrades,
 * causing the new implementation to mysteriously fail.
 * For a multi-platform application, this must be done and maintained
 * for each system.
 * This approach is tenable and not too arduous in some contexts, such as
 * if only standardized APIs are used by an application and its
 * libraries, or if an application uses few library APIs that call malloc
 * or if the entire library can be recompiled.
 * For example, the MacOS system malloc library has a (non-standard)
 * malloc_size() (and malloc_good_size()) API. If an implementor is
 * unaware that malloc_size() must be wrapped, if a process calls a
 * library function that calls malloc_size() with a pointer argument obtained
 * from a new malloc implementation, the process is likely to dereference
 * an invalid pointer and raise a segmentation fault.
 *
 * Regarding B:
 * The application must be careful not to use a pointer obtained from
 * a replacement malloc library with a system malloc API, and vice versa.
 * Many system library APIs return information in static buffers or
 * caller-supplied buffers, or provide an API for deallocating previously
 * returned memory, so they do not present a problem.  But the application
 * may be aware of any exceptions, especially from non-standard libraries
 * (e.g., fetchListURL() from libfetch(3)) that allocate memory
 * (using malloc(3)) and require the
 * caller to free it (using free(3) only).
 *
 * Note that it may be possible for a replacement malloc library to provide
 * an API for identifying memory that it has allocated. An application's
 * call to free() or realloc(), for example, might be handled by the
 * replacement library and handled or passed on to another implementation
 * depending on the ownership of the memory.
 *
 * N.B.
 * If DACS_malloc_selector is changed at runtime, the wrong realloc() or free()
 * implementation could be called for an allocation made with a different
 * malloc() etc. implementation.
 * At present, a malloc() etc. implementation should only be selected at
 * build time. It can be changed with care at runtime, at your own risk,
 * perhaps as one of the first things a program does and if it exits
 * without cleaning up (e.g., via _exit(2) or_Exit(3) so that no memory
 * allocations are explicitly freed).
 * On some platforms, malloc() and/or free() are called before main()
 * even begins to execute (e.g., for dynamic linking).
 *
 * XXX could prepend a header to each wrapped allocation that identifies
 * the dynamic allocation package responsible for it. There would be added
 * complexity and memory cost but it would eliminate confusion.
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2024\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: dacs_malloc.c 3334 2024-09-23 20:07:01Z brachman $";
#endif

#include "dacs_config.h"

#include "dacs_malloc.h"
#include "str.h"
#include "log.h"

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef NOTDEF
/* DACS_MALLOC_FRAME */
extern void *frame_malloc(size_t size);
extern void *frame_realloc(void *ptr, size_t size);
extern void *frame_calloc(size_t nmemb, size_t memb_size);
extern void frame_cfree(void *ptr);
extern void frame_free(void *ptr);
#endif

typedef void *(*malloc_func_t)(size_t);
typedef void *(*realloc_func_t)(void *, size_t);
typedef void *(*calloc_func_t)(size_t, size_t);
typedef void (*free_func_t)(void *);

/* Top-level functions for malloc(). */
static DACS_malloc_switch *dacs_malloc_sw = NULL;

/* Lower-level functions for malloc() implementations only. */
static DACS_malloc_switch *dacs_malloc_base_sw = NULL;

static DACS_malloc_switch *dacs_malloc_list = NULL;

#ifdef USE_DACS_MALLOC
static int dacs_malloc_disable = 0;
#else
static int dacs_malloc_disable = 1;
#endif

static int dacs_malloc_did_init = 0;
static int dacs_malloc_debug = 0;

/*
 * If USE_DACS_MALLOC is NOT defined, then the normal system library
 * implementations will be used - so if problems arise, building without
 * defining USE_DACS_MALLOC should bypass almost all of this stuff.
 */

#ifdef USE_DACS_MALLOC
/*
 * Call the system library implementation of malloc(3).
 *
 * Code in other files can call safe_inline_malloc(), safe_malloc(),
 * or MALLOC() if they want to use a wrapper that checks, logs, and exits
 * if an error occurs.
 *
 * Note that redefining malloc(3) and family in this way may not
 * result in ALL of the process's calls to those functions to invoke
 * the routines defined here - for instance, on FreeBSD, none of the malloc()
 * calls performed at process startup time will call here, but subsequent
 * calls to gethostbyname(3) will. There could, of course, be other
 * calls to the system malloc or malloc functionality within libraries
 * that are linked to this code - we must assume that all code within
 * the application is careful that pointers to memory created by one
 * implementation do not find their way to another implementation's
 * realloc() or free().
 */
void *
malloc(size_t size)
{
  void *ptr;
  static void *(*sys_malloc_func)(size_t size) = NULL;

  if (sys_malloc_func == NULL) {
	/* XXX GNU/Linux does not provide dlfunc(3). */
    sys_malloc_func = (malloc_func_t) dacs_dlfunc(RTLD_NEXT, "malloc");
    if (sys_malloc_func == NULL)
      return(NULL);
  }

  ptr = sys_malloc_func(size);

  return(ptr);
}

/*
 * Call the system library implementation of realloc(3).
 */
void *
realloc(void *ptr, size_t size)
{
  void *new_ptr;
  static void *(*sys_realloc_func)(void *ptr, size_t size) = NULL;

  if (sys_realloc_func == NULL) {
	/* XXX GNU/Linux does not provide dlfunc(3). */
    sys_realloc_func = (realloc_func_t) dacs_dlfunc(RTLD_NEXT, "realloc");
    if (sys_realloc_func == NULL)
      return(NULL);
  }

  new_ptr = sys_realloc_func(ptr, size);

  return(new_ptr);
}

/*
 * Call the system library implementation of calloc(3).
 */
void *
calloc(size_t nmemb, size_t nmemb_size)
{
  void *ptr;
  static void *(*sys_calloc_func)(size_t nmemb, size_t memb_size) = NULL;

  if (sys_calloc_func == NULL) {
	/* XXX GNU/Linux does not provide dlfunc(3). */
    sys_calloc_func = (calloc_func_t) dacs_dlfunc(RTLD_NEXT, "calloc");
    if (sys_calloc_func == NULL)
      return(NULL);
  }

  ptr = sys_calloc_func(nmemb, nmemb_size);

  return(ptr);
}

/*
 * Call the system library implementation of free(3).
 */
void
free(void *ptr)
{
  static void (*sys_free_func)(void * size) = NULL;

  if (sys_free_func == NULL) {
	/* XXX GNU/Linux does not provide dlfunc(3). */
    sys_free_func = (free_func_t) dacs_dlfunc(RTLD_NEXT, "free");
    if (sys_free_func == NULL)
      return;
  }

  sys_free_func(ptr);
}
#endif

/*
 * This is an alternate API for malloc() etc. Calls
 * from outside this file that can use these switchable function names.
 * The callers presumably define their own malloc(),
 * realloc(), free(), etc.  That is, they wrap whatever malloc()
 * is actually going to be used.
 *
 * To prevent a loop, this API should not call a malloc() that be wrapping
 * it, so we use dacs_malloc_base_selector whose functions implement
 * malloc() without wrapping.
 * XXX this might be addressed in a more automatic way if a library's
 * API was assigned a "level number", available to this API, and we required
 * the selector in effect for call to use an API with a strictly lower level
 * number.  That way an API could not call itself directly or indirectly.
 */

/*
 * Enable (STATE != 0) or disable (STATE == 0) debugging.
 */
int
dacs_debug_malloc(int state)
{
  int st;

  st = dacs_malloc_debug;
  dacs_malloc_debug = state;

  return(st);
}

/*
 *
 */
int
dacs_get_malloc(DACS_malloc_switch **which_base, DACS_malloc_switch **which)
{

  if (!dacs_malloc_did_init)
	return(-1);

  if (which_base != NULL)
	*which_base = dacs_malloc_base_sw;

  if (which != NULL)
	*which = dacs_malloc_sw;

  return(0);
}

DACS_malloc_switch *
dacs_malloc_get_callback(DACS_malloc_selector which)
{
  DACS_malloc_switch *ms;

  if (which < 1)
	return(NULL);

  for (ms = dacs_malloc_list; ms != NULL; ms = ms->next) {
	if (which == ms->selector)
	  return(ms);
  }

  return(NULL);
}

DACS_malloc_switch *
dacs_malloc_get_callback_byname(const char *name)
{
  DACS_malloc_switch *ms;

  if (name == NULL)
	return(NULL);

  for (ms = dacs_malloc_list; ms != NULL; ms = ms->next) {
	if (name != NULL && streq(name, ms->name))
	  return(ms);
  }

  return(NULL);
}

static int
check_ptr_cb_stub(const void *ptr)
{

  return(0);
}

/*
 * Add a dynamic memory allocation library (API) to the list of
 * available libraries.
 * The API should be uniquely identified by WHICH and also by NAME (non-NULL),
 * so that either parameter can be used for future lookups.
 * After an API has been registered it can be used through this layer.
 * There should not be duplicate identifiers.
 * The lookup will find the most recently added matching entry.
 *
 * If DACS_MALLOC_DISABLE is non-zero and WHICH is not the system library
 * API, however, the library callbacks will not be added - this means that
 * only the system library calls added before disabling are available through
 * this layer.
 * DACS_MALLOC_DISABLE must be non-zero before this function is called
 * to prevent registration of any APIs other than the system's.
 *
 * The CHECK_PTR_CB callback, if non-NULL, is passed a pointer (to allocated
 * memory).
 * If that pointer was obtained from the MALLOC_CB, CALLOC_CB,
 * or REALLOC_CB function for the corresponding library, then CHECK_PTR_CB
 * returns 1, otherwise 0.
 * If CHECK_PTR_CB is NULL, because a suitable function is unavailable, a stub
 * will be invoked to return zero.
 */
DACS_malloc_switch *
dacs_malloc_add_library(DACS_malloc_selector which,
						const char *name,
						void *(*malloc_cb)(size_t),
						void *(*calloc_cb)(size_t, size_t),
						void *(*realloc_cb)(void *, size_t),
						void (*free_cb)(void *),
						int (*check_ptr_cb)(const void *))
{
  DACS_malloc_switch *ms;

  if (which < 1 || name == NULL)
	return(NULL);

  if (dacs_malloc_disable && which != DACS_MALLOC_SYSTEM)
	return(NULL);

  if (malloc_cb == NULL || calloc_cb == NULL || realloc_cb == NULL
	  || free_cb == NULL)
	return(NULL);

  ms = (DACS_malloc_switch *) malloc(sizeof(DACS_malloc_switch));
  ms->selector = which;
  ms->name = strdup(name);
  ms->malloc_func = malloc_cb;
  ms->calloc_func = calloc_cb;
  ms->realloc_func = realloc_cb;
  ms->free_func = free_cb;
  if (check_ptr_cb == NULL)
	ms->check_ptr_func = check_ptr_cb_stub;
  else
	ms->check_ptr_func = check_ptr_cb;

  if (dacs_malloc_list == NULL)
	ms->next = NULL;
  else
	ms->next = dacs_malloc_list;
  dacs_malloc_list = ms;

  return(ms);
}

/*
 * Initialize this layer, ensuring that the system library API is always
 * available.
 *
 * Return 1 if initialization is completed successfully, 0 if initialization
 * has already been done or is unnecessary, or -1 if it fails.
 */
int
dacs_malloc_init(void)
{

  if (dacs_malloc_did_init)
	return(0);

  dacs_malloc_add_library(DACS_MALLOC_SYSTEM, "system",
						  malloc, calloc, realloc, free, NULL);
  dacs_set_malloc(DACS_MALLOC_SYSTEM);
  dacs_set_base_malloc(DACS_MALLOC_SYSTEM);

#ifdef USE_DACS_MALLOC
  /*
   * Discover at runtime which additional malloc implementations have
   * been linked in and make them available.
   *
   * XXX this implementation is tentative...
   */
  {
	typedef int (*dl_adl_t)(void);
	int (*dl_malloc_adl)(void);

	dl_malloc_adl = (dl_adl_t) dacs_dlfunc(NULL, "dl_malloc_add_dacs_library");
	if (dl_malloc_adl)
	  dl_malloc_adl();
  }

  {
	typedef int (*phk_adl_t)(void);
	int (*phk_malloc_adl)(void);

	phk_malloc_adl
	  = (phk_adl_t) dacs_dlfunc(NULL, "phk_malloc_add_dacs_library");
	if (phk_malloc_adl)
	  phk_malloc_adl();
  }
#endif

  dacs_malloc_did_init = 1;

  return(1);
}

/*
 * N.B. This should be used with care.
 * In particular, make sure that allocations from the same memory provider
 * package are all freed (including via realloc() using that provider.
 * Extra caution is required if functions are called via atexit(3) that
 * then call free().
 */
int
dacs_set_malloc(DACS_malloc_selector which)
{
  int n;
  DACS_malloc_switch *ms;

  if ((ms = dacs_malloc_get_callback(which)) == NULL)
	return(-1);

  dacs_malloc_sw = ms;

  return(0);
}

int
dacs_set_base_malloc(DACS_malloc_selector which)
{
  int n;
  DACS_malloc_switch *ms;

  if ((ms = dacs_malloc_get_callback(which)) == NULL)
	return(-1);

  dacs_malloc_base_sw = ms;

  return(0);
}

/*
 * Must not use any malloc(3) memory allocation functions, directly or
 * indirectly.
 */
static void
trace_message(char *label, int selector, void *ptr, size_t size)
{

  if (dacs_malloc_debug) {
	int len;
	char buf[128];

	if (ptr != NULL)
	  len = sprintf(buf, "%s/%d 0x%lx, %ld\n", label, selector,
					(unsigned long) ptr, size);
	else
	  len = sprintf(buf, "%s/%d %ld\n", label, selector, size);
	write(2, buf, len);
  }
}

/*
 * Call the selected base malloc() implementation.
 */
void *
dacs_malloc(size_t size)
{
  void *ptr;

  if (!dacs_malloc_did_init)
	return(NULL);

  trace_message(" dacs_malloc:in", dacs_malloc_base_sw->selector, 0, size);

  if (dacs_malloc_base_sw->malloc_func == NULL)
	ptr = malloc(size);
  else
	ptr = (dacs_malloc_base_sw->malloc_func)(size);

  trace_message("dacs_malloc:out", dacs_malloc_base_sw->selector, ptr, size);

  return(ptr);
}

/*
 * Call the selected base realloc() implementation.
 */
void *
dacs_realloc(void *ptr, size_t size)
{
  void *new_ptr;

  if (!dacs_malloc_did_init)
	return(NULL);

  trace_message(" dacs_realloc:in", dacs_malloc_base_sw->selector, ptr, size);

  new_ptr = (dacs_malloc_base_sw->realloc_func)(ptr, size);

  trace_message("dacs_realloc:out", dacs_malloc_base_sw->selector, new_ptr, size);

  return(new_ptr);
}

/*
 * Call the selected base calloc() implementation.
 */
void *
dacs_calloc(size_t nmemb, size_t memb_size)
{
  size_t need;
  void *p;

  if (!dacs_malloc_did_init)
	return(NULL);

  need = nmemb * memb_size;
  if ((p = dacs_malloc(need)) != NULL)
	memset(p, 0, need);

  return(p);
}

/*
 * An obsolete counterpart to calloc(), equivalent to free(3).
 */
void
dacs_cfree(void *ptr)
{

  if (!dacs_malloc_did_init)
	return;

  dacs_free(ptr);
}

/*
 * Call the selected base free() implementation.
 */
void
dacs_free(void *ptr)
{

  if (!dacs_malloc_did_init)
	return;

  trace_message("dacs_free", dacs_malloc_base_sw->selector, ptr, 0);

  if (ptr != NULL)
	dacs_malloc_base_sw->free_func(ptr);
}

void *
dacs_valloc(size_t size)
{

  if (!dacs_malloc_did_init)
	return(NULL);

  return(malloc(size));
}

/*
 * Given PTR, a pointer to allocated memory, return a descriptor for the
 * library that manages that memory.
 * Return NULL if the library responsible cannot be determined (or PTR
 * is not managed by any library known to this layer).
 */
DACS_malloc_switch *
dacs_malloc_check_ptr(const void *ptr)
{
  DACS_malloc_switch *ms;

  if (!dacs_malloc_did_init)
	return(NULL);

  if (ptr == NULL)
	return(NULL);

  for (ms = dacs_malloc_list; ms != NULL; ms = ms->next) {
	if (ms->check_ptr_func != NULL && (ms->check_ptr_func)(ptr) == 1)
	  return(ms);
  }

  return(NULL);
}

#ifdef XPROG
int
main(int argc, char **argv)
{
  void *ptr;

  ptr = malloc(17);
}
#endif
