
/*
 * JSPR - JSon ParseR
 */

/*
 * Copyright (c) 2017-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: jspr.h 3360 2025-03-31 23:28:47Z brachman $
 */

#ifndef __JSPR_H_
#define __JSPR_H_

#include <sys/types.h>
#include <stdint.h>
#include <inttypes.h>
#include <math.h>
#include <limits.h>
#include <stdarg.h>

#define JSPR_VERSION	"jspr-0.1"

#ifndef INLINE
#define INLINE	inline
#endif

/* Some convenient shorthands. */
#define JSPR_CONFIG(JP)		(&JP->config)
#define JSPR_ERR(JP)		(&JP->err_state)
#define JSPR_RES(JP)		(&JP->res_info)

#define JSPR_WAS_TERMINATED(JP)		(JP->err_state.user_terminated)
#define JSPR_RC(JP)			(JP->err_state.err_rc)

/*
 * Value types.
 * JSON has six structural characters, strings, numbers,
 * and three case-sensitive literal names. A document is parsed into a
 * sequence of values, each represented by a value node.
 * A JSON text is a Value: an Object, Array, Number, String, or a literal name.
 *
 *   o Object: delimited by begin-object '{' and end-object '}'
 *   o Array:  delimited by begin-array '[' and end-array ']'
 *   o String: delimited by double quote characters
 *   o Number: a common signed base 10 integer or floating point representation
 *             (leading plus sign and leading zeroes are disallowed)
 *   o Literal names: true, false, and null
 *
 * See jspr_type_name[].
 *
 * There are no explicit end-of-array and end-of-object value types.
 * The last element of an array or object will have a zero next sibling
 * pointer and, if enabled, an array/object value will have a last child
 * value pointer.
 *
 * Arrays and Objects consist of zero or more values called elements.
 * An Array is a list of zero or more comma-separated elements.
 * An Object is a list of zero or more comma-separated named elements,
 * each element is called a member and consists of a string (its name)
 * and its corresponding value.
 * This terminology deviates slightly from RFC 8259 in that an element
 * refers to a value in an Array or Object. A member only refers to a value
 * in an Object.
 * 
 * There are essentially no a priori limitations (other than available memory
 * and pointer and index sizes) on:
 *   o nesting depth
 *   o string lengths
 *   o number lengths
 *   o number of objects, arrays, or elements
 * The maximum number of values per parse is limited by Jspr_index,
 * which is defined to be a 32-bit unsigned integer.
 */
enum {
  JSPR_INVALID_TYPE    =  0,	/* Reserved and not used. */
  JSPR_OBJECT_TYPE     =  1,	/* Start of object meta value. */
  JSPR_ARRAY_TYPE      =  2,	/* Start of array meta value. */
  JSPR_STRING_TYPE     =  3,
  JSPR_NAME_TYPE       =  4,	/* The string name of a member. */
  JSPR_NUMBER_TYPE     =  5,
  JSPR_BOOLEAN_TYPE    =  6,	/* The literal "true" or "false". */
  JSPR_NULL_TYPE       =  7,	/* The literal "null". */

  JSPR_END_OBJECT_TYPE =  8,	/* XXX End of object marker. */
  JSPR_END_ARRAY_TYPE  =  9,	/* XXX End of array marker. */

  JSPR_COMMENT_TYPE    = 10,	/* For retaining comments. */
  JSPR_ANY_TYPE        = 11,	/* For searching only. */
  JSPR_UNKNOWN_TYPE    = 12		/* An invalid type. */
};

/*
 * Parser result codes.
 * JSPR_OK must be zero.
 * All true error code must be negative.
 *
 * The parser tries to return the most appropriate error code when
 * it encounters invalid input but it may not always be the most informative
 * error code, and/or the best error code may be debatable or dependent
 * on the particular context or user expectation.
 */
typedef enum {
  JSPR_OK                   =   0, /* No error, everything ok. */
  JSPR_ERROR_NOMEM          =  -1, /* Cannot allocate memory. */
  JSPR_ERROR_INVALID_CHAR   =  -2, /* Invalid character within a string. */
  JSPR_ERROR_PART           =  -3, /* Incomplete structure (e.g., no ']'). */
  JSPR_ERROR_SYNTAX         =  -4, /* Unexpected symbol type. */
  JSPR_ERROR_NUMSYNTAX      =  -5, /* Invalid number syntax. */
  JSPR_ERROR_STRSYNTAX      =  -6, /* Invalid string syntax. */
  JSPR_ERROR_ARRSYNTAX      =  -7, /* Invalid Array syntax. */
  JSPR_ERROR_OBJSYNTAX      =  -8, /* Invalid Object syntax. */
  JSPR_ERROR_NAMESYNTAX     =  -9, /* Object name is missing or invalid. */
  JSPR_ERROR_CONFIG         = -10, /* Invalid configuration. */
  JSPR_ERROR_VALUE_LIMIT    = -11, /* Max values limit reached. */
  JSPR_ERROR_DUP_OBJNAME    = -12, /* Duplicated member name with an Object. */
  JSPR_ERROR_USER_TERM      = -13, /* User initiated parsing termination. */
  JSPR_ERROR_VALUE_EXPECTED = -14, /* A required value is missing. */
  JSPR_ERROR_INPUT_OVERFLOW = -15, /* JSON input is too big. */

  JSPR_ERROR_UNKNOWN        = -99  /* Implementation error.. */
} Jspr_rc;

/*
 * A parser's state after the most recent error.
 * Line and position numbers start at one (not zero).
 * The error message is in a static buffer.
 *
 * If an error has occurred and the IS_FATAL element is zero, the end of input
 * was reached; it is possible that calling jspr_reset() with additional
 * input may result in a successful parse; if IS_FATAL is
 * non-zero, the error is unrecoverable.
 */
typedef struct Jspr_err_state {
  Jspr_rc err_rc;			/* Last reported error code. */
  unsigned int err_line;	/* Line number of error. */
  unsigned int err_pos;		/* Approx. byte offset of error. */
  unsigned int err_linepos;	/* Byte position within this line of error. */
  char *err_mesg;			/* Optional additional error text. */
  uint16_t is_fatal;		/* Non-zero if the error is not recoverable. */
  uint16_t user_terminated;	/* Non-zero if a callback wants parsing to stop. */
} Jspr_err_state;

typedef struct Jspr_parser Jspr_parser;
typedef struct Jspr_value Jspr_value;

/*
 * These are bit flags that can be ORed.
 * Comments can be disabled or combinations of styles can be selectively
 * recognized.
 */
typedef enum {
  JSPR_COMMENTS_DISABLED     = 000,		/* Comments are not allowed. */
  JSPR_COMMENTS_CPP_TYPE     = 001,		/* C++ single-line comments. */
  JSPR_COMMENTS_SHELL_TYPE   = 002,		/* Shell # single-line comments. */
  JSPR_COMMENTS_C_TYPE       = 004,		/* C multi-line comments. */
  JSPR_COMMENTS_TYPE_MASK    = 007,		/* For masking type bits. */

  JSPR_COMMENTS_RETAIN_FLAG  = 010		/* Keep comments in parse tree. */
} Jspr_comments;

#define JSPR_COMMENTS_ENABLED(JP, TYPE)						\
  ((JSPR_CONFIG(JP)->enable_comments & JSPR_COMMENTS_TYPE_MASK) & TYPE)

/*
 * Configurable flags, callbacks, etc.
 * Any non-zero value enables a feature/constraint.
 */
typedef struct Jspr_config {
  int enable_expand_values;		/* More values can be allocated. */
  int enable_no_values;			/* Do not emit or allocate any values. */
  int enable_colourize;			/* Colour pretty printed JSON. */
  int enable_debug;
  int enable_verbose;
  int enable_comments;			/* See Jspr_comments. */

  int require_obj_or_arr;		/* Input must be an Object/Array. */
  int require_unique_obj_names;	/* Names within an Object must be unique. */

  int values_limit;				/* Maximum number of values allowed. */

  unsigned int num_values;		/* Current number of values allocated. */
  void  *user_new_member_cb_arg;

  /*
   * For allocating new Jspr_value structures:
   * USER_MALLOC_FUNC: malloc(3) or a functionally similar substitute
   * USER_FREE_FUNC:   free(3) or equivalent, to match USER_MALLOC_FUNC
   * USER_VALUES_PTR:  an unextendable, preallocated buffer to use
   * USER_EXPAND_VALUES_CB: returns the new number of structures to allocate
   * USER_NEW_MEMBER_CB: called when a new Object member is added
   */
  void *(*user_malloc_func)(size_t size);
  void (*user_free_func)(void *ptr);
  void *user_values_ptr;
  int (*user_expand_values_cb)(Jspr_parser *, unsigned int num_values);
  int (*user_new_member_cb)(Jspr_parser *, Jspr_value *member_name,
							Jspr_value *member_value, void *cb_arg);
} Jspr_config;

/* Default configuration. */
enum {
  JSPR_REQUIRE_OBJ_OR_ARR_DEFAULT       = 0,
  JSPR_REQUIRE_UNIQUE_OBJ_NAMES_DEFAULT = 0,
  JSPR_ENABLE_EXPAND_VALUES_DEFAULT     = 1,
  JSPR_ENABLE_COLOURIZE_DEFAULT         = 0,
  JSPR_ENABLE_COMMENTS_DEFAULT
  = (JSPR_COMMENTS_CPP_TYPE | JSPR_COMMENTS_SHELL_TYPE | JSPR_COMMENTS_C_TYPE),
  JSPR_NUM_VALUES_DEFAULT               = 10,
  JSPR_VALUES_LIMIT_DEFAULT             = 1024
};

/* Compile-time buffer sizes. */
enum {
  /*
   * Size in bytes of the static buffer used to temporarily hold values.
   * If this size is insufficient, a buffer will be dynamically allocated.
   */
  JSPR_VALUE_BUFSIZ    = 512,
  JSPR_STRERROR_BUFSIZ = 128
};

/*
 * For tracking resource usage during a parse.
 * See jspr_get_config().
 */
typedef struct Jspr_res_info {
  unsigned int num_values_used;
  unsigned int num_values;	/* Number of values allocated. */
  unsigned int nexpands;	/* Number of times value pool was expanded. */
  unsigned int nstrings;	/* Number of strings allocated. */
  size_t xbuf_count;		/* Number of extra temp buffer allocations. */
  size_t xbuf_maxlen;		/* Maximum extra temp buffer size. */
  size_t string_alloc;		/* Total space allocated for strings. */
} Jspr_res_info;

/*
 * Conversion types for jspr_string_value().
 */
typedef enum {
#define JSPR_I_TYPE  int
  JSPR_I            = 1, JSPR_INT   = 1,
  JSPR_IGZ          = 2, JSPR_INTGZ = 2,       /* Greater than Zero */
  JSPR_INN          = 3, JSPR_INTNN = 3,       /* Not Negative */
#define JSPR_UI_TYPE          unsigned int
  JSPR_UI           = 4, JSPR_UINT  = 4,
  JSPR_UINZ         = 5, JSPR_UINTZ = 5,       /* Not Zero */
#define JSPR_L_TYPE           long
  JSPR_L            = 6, JSPR_LONG  = 6,
#define JSPR_UL_TYPE          unsigned long
  JSPR_UL           = 7, JSPR_ULONG = 7,
#define JSPR_TIME_T_TYPE      time_t
  JSPR_TIME_T       = 8,
#define JSPR_MODE_T_TYPE      mode_t
  JSPR_MODE_T       = 9,
#define JSPR_PID_T_TYPE       pid_t
  JSPR_PID_T        = 10,
#define JSPR_IN_PORT_T_TYPE   in_port_t
  JSPR_IN_PORT_T    = 11,
#define JSPR_STR_TYPE         char *
  JSPR_STR          = 12,
#define JSPR_SIZE_T_TYPE      size_t
  JSPR_SIZE_T       = 13,
#define JSPR_SSIZE_T_TYPE     ssize_t
  JSPR_SSIZE_T      = 14,
#define JSPR_LL_TYPE          long long
  JSPR_LL           = 15, JSPR_LONGLONG = 15,
#define JSPR_ULL_TYPE unsigned long long
  JSPR_ULL          = 16, JSPR_ULONGLONG = 16,
#define JSPR_IMAX_T_TYPE      intmax_t
  JSPR_IMAX_T       = 17, JSPR_INTMAX_T  = 17,
#define JSPR_UIMAX_T_TYPE     uintmax_t
  JSPR_UIMAX_T      = 18, JSPR_UINTMAX_T = 18,
#define JSPR_I32_T_TYPE       int32_t
  JSPR_I32_T        = 19, JSPR_INT32_T   = 19,
#define JSPR_I64_T_TYPE       int64_t
  JSPR_I64_T        = 20, JSPR_INT64_T   = 20,
#define JSPR_UI32_T_TYPE      uint32_t
  JSPR_UI32_T       = 21, JSPR_UINT32_T  = 21,
#define JSPR_UI64_T_TYPE      uint64_t
  JSPR_UI64_T       = 22, JSPR_UINT64_T  = 22,
#define JSPR_FLOAT_TYPE       float
  JSPR_F            = 23, JSPR_FLOAT = 23,
#define JSPR_DOUBLE_TYPE      double
  JSPR_D            = 24, JSPR_DOUBLE = 24,
  JSPR_VALUE_STR    = 25,
  JSPR_VALUE_COPY   = 26,
  JSPR_VALUE_STRLEN = 27,
  JSPR_IBASE        = 256,	/* A bit flag to indicate the optional BASE arg. */
  JSPR_ERR          = -1
} Jspr_string_cvt;

/*
 * Assorted value characteristics.
 * XXX may need improvement
 */
typedef enum {
  /* For JSPR_STRING_TYPE. */
  JSPR_VALUE_STR_CLEAN	   = 0,		/* Printable ASCII only. */
  JSPR_VALUE_STR_UTF8_1	   = 1,		/* UTF8 (1 byte) only. */
  JSPR_VALUE_STR_UTF8_2	   = 2,		/* UTF8 (1 or 2 bytes) only. */
  JSPR_VALUE_STR_UTF8_3	   = 3,		/* UTF8 (1 .. 3 bytes) only. */
  JSPR_VALUE_STR_UTF8_4	   = 4,		/* UTF8 (1 .. 4 bytes). */
  JSPR_VALUE_STR_INVALID   = 5,		/* An invalid character is present. */

  /* For JSPR_NUMBER_TYPE. */
  JSPR_VALUE_NUMBER_POS    = 0,		/* A non-negative integer. */
  JSPR_VALUE_NUMBER_NEG    = 1,		/* A negative integer. */
  JSPR_VALUE_NUMBER_REAL   = 2,		/* A real without an exponent. */
  JSPR_VALUE_NUMBER_EXP    = 3,		/* A real with an exponent. */

  /* For JSPR_BOOLEAN_TYPE. */
  JSPR_VALUE_BOOLEAN_FALSE = 0,		/* Value is 'false'. */
  JSPR_VALUE_BOOLEAN_TRUE  = 1		/* Value is 'true'. */
} Jspr_value_flag_t;

typedef uint8_t Jspr_flags;
typedef uint8_t Jspr_type;
typedef uint32_t Jspr_index;
typedef uint32_t Jspr_js;

typedef enum {
  JSPR_LOOKUP_BREADTH_FIRST = 1,
  JSPR_LOOKUP_PREORDER      = 2
} Jspr_lookup_alg;

/* Various constants/selectors. */
enum {
  /* Index of first valid value, must be greater than zero - zero is not used. */
  JSPR_VALUE_ROOT_INDEX	    = 1
};

/*
 * For both Objects and Arrays, the Object/Array Jspr_value contains
 * the index of its first member (child_first) (and optionally to its
 * last member, child_last).
 * Each element of an Object/Array points to its next member (sib_next)
 * (and optionally to its preceding member, sib_prev).
 * Every Jspr_value contains the index of its parent Object/Array.
 * An index of zero always represents a null pointer (i.e., "no index",
 * no parent, or end of list) - a value with index zero exists but is never used.
 *
 * The "name" field of an Object's member value points to a value that
 * describes its name string; the name value's parent is its
 * corresponding member value, not the containing Object (a member value
 * points to its parent, so the "grandparent" of a name value is its
 * containing Object).
 *
 * Object.child_first --> Member Value --> name --> Member Name string
 *            <-- parent --/               <-- parent --/
 *
 * Or:
 * Object --child_first-->
 *   Member Value 1 --name--> Member Name 1 --sib_next-->
 *   Member Value 2 --name--> Member Name 2 --sib_next-->
 *   ...
 *   Member Value N --name--> Member Name N --zero
 * In the value vector:
 *   Object
 *   Name 1
 *   Value 1
 *   Name 2
 *   Value 2
 *   ...
 *
 * A value references other values via Jspr_index (a base+index) rather
 * than a pointer.
 * Using indexes rather than pointers saves memory and makes dynamic
 * expansion of the value vector easy, but is a little less computationally
 * efficient and complicates the code a little.
 * Using index also simplifies comparing two value vectors for equality.
 *
 * Defining JSPR_COMPACT causes a base+index and length pair to be
 * used to reference the input string instead of begin/end pointers.
 * One advantage of this is that after a value's JSON input text has been
 * scanned, the input text could potentially be freed - a callback, for
 * example, might be used later to resupply the input text in the given
 * byte range for conversion functions.
 *
 * If CHILD_LAST and/or SIB_PREV is undefined, they are computed at
 * runtime - this will not matter during parsing but may make navigating
 * the parse tree less efficient.
 * When both are undefined this data structure may be 8 bytes smaller.
 * XXX this compact representation may be changed or removed; minimizing
 * the size of this structure is not a high priority and not worth the
 * effort of maintaining multiple representations. More lightweight JSON parser
 * implementations are available.
 *
 * The JSPR_END_ARRAY_TYPE and JSPR_END_OBJECT_TYPE are special values that
 * represent end-array ("]") and end-object ("}") structural characters,
 * respectively.
 * These values are not technically required, but they make working with
 * parse trees easier to implement correctly.
 * No parent, sibling, or child index points to one of these values.
 * The parent of a JSPR_END_ARRAY_TYPE is its matching JSPR_ARRAY_TYPE value,
 * and the parent of a JSPR_END_OBJECT_TYPE is its matching JSPR_OBJECT_TYPE
 * value.
 * The Array/Object nesting depths in a value simplify pretty printing
 * a parse tree and may also do so for future parse tree navigation operations.
 * Each new Array/Object (begin-array, begin-object) is assigned a unique
 * numerical identifier and its matching end value (end-array, end-object)
 * is assigned the same identifier - this simplifies identifying and locating
 * structures in the parse tree.
 *
 * XXX the addition of a backpointer to Jspr_parser would cost 8 bytes/value
 * but would eliminate the parameter in many API functions
 */
struct Jspr_value {
  uint8_t type;				/* Jspr_type: the type of parse token. */
  uint8_t flags;			/* Jspr_flags: type-dependent flags. */
#ifdef JSPR_COMPACT
  Jspr_index nchildren;		/* Number of immediate children. */
#else
  uint32_t nchildren;		/* Number of immediate children. */
#endif

#ifdef JSPR_COMPACT
  Jspr_js js_start;			/* JSON text start offset. */
  Jspr_js js_len;			/* JSON text length in bytes. */
#else
  const char *start;		/* First byte of lexical element. */
  const char *end;			/* Last byte of lexical element. */
#endif

  Jspr_index name;			/* Object member name (a string), if applicable. */
  Jspr_index parent;		/* Parent Array or Object, if any. */
  Jspr_index child_first;	/* First child of Array/Object, if any. */
#ifdef CHILD_LAST
  Jspr_index child_last;	/* Last child of Array/Object, if any. */
#endif
#ifdef SIB_PREV
  Jspr_index sib_prev;		/* Previous sibling for Array/Object element. */
#endif
  Jspr_index sib_next;		/* Next sibling for Array/Object element. */

  Jspr_index struct_count;	/* For a structure value type, its identifier. */
  Jspr_index array_level;	/* This value's Array nesting depth. */
  Jspr_index object_level;	/* This value's Object nesting depth. */
};

/*
 * JSON parser context and state for the current parse.
 * The string being parsed (JS) is not copied, so it may not be modified.
 *
 * The JSON string must be passed when the parser is initialized.
 * This simplifies the implementation and is probably most efficient,
 * since the input is fully buffered and its length is known - no additional
 * copying is necessary.
 * Supporting streamed input is possible but would complicate the design,
 * so implementing it is TBD.
 */
struct Jspr_parser {
  const char *js;			/* The JSON string being parsed. */
  size_t js_len;			/* Length of the input buffer in bytes. */

  Jspr_value *values;		/* A vector of used and available values. */
  Jspr_index next_value;	/* Index of the next value to be used. */
  Jspr_config config;		/* Current configuration parameters. */

  Jspr_index array_count;	/* Next identifier for a new Array value type. */
  Jspr_index object_count;	/* Next identifier for a new Object value type. */
  Jspr_index array_level;	/* Current nesting depth of Array values. */
  Jspr_index object_level;	/* Current nesting depth of Object values. */

  Jspr_err_state err_state;	/* Parse error description. */
  Jspr_res_info res_info;	/* Resource usage instrumentation. */
};

extern int jspr_is_compact;

#ifdef __cplusplus
extern "C" {
#endif

/* jspr_parse.c */
extern Jspr_parser *jspr_init(const char *js, size_t js_len);
extern Jspr_parser *jspr_init_x(Jspr_parser *jp, const char *js, size_t jslen,
								Jspr_config *config);
extern Jspr_parser *jspr_reset(Jspr_parser *parser, const char *js,
							   size_t js_len);
extern Jspr_config *jspr_get_config(Jspr_parser *jp, Jspr_config *config);
extern Jspr_value *jspr_get_values(Jspr_parser *jp, Jspr_value **values,
								   Jspr_index *used, Jspr_index *avail);
extern void jspr_end(Jspr_parser *parser);

extern Jspr_rc jspr_parse(Jspr_parser *parser);
extern Jspr_parser *jspr_parse_copy(Jspr_parser *jp);
extern Jspr_rc jspr_parse_value(Jspr_parser *parser, Jspr_value **valuep,
								const char **endp);
extern Jspr_value *jspr_expand_values(Jspr_parser *jp, unsigned int need_values);

extern char *jspr_name_of_value(Jspr_parser *jp, Jspr_value *value);

extern Jspr_res_info *jspr_res_info(Jspr_parser *jp, Jspr_res_info *ri);
extern Jspr_err_state *jspr_err_state(Jspr_parser *parser,
									  Jspr_err_state *err_state);
extern char *jspr_strerror(Jspr_rc rc);

/* jspr_char.c */
extern int jspr_string_cmp(Jspr_parser *jp, Jspr_value *v1, Jspr_value *v2);
extern char *jspr_string_copy(Jspr_parser *parser, Jspr_value *value);
extern int jspr_string_validate(const char *p);
extern int jspr_string_value(Jspr_parser *jp, Jspr_value *value,
							 Jspr_string_cvt type, ...);
extern int jspr_name_eq(Jspr_parser *jp, Jspr_value *v1, char *str, size_t len);
extern int jspr_value_name_eq(Jspr_parser *jp, Jspr_value *value, char *name,
							  size_t name_len);

/* jspr_io.c */
extern void jspr_dump(FILE *ofp, Jspr_parser *jp);
extern void jspr_dump_value(FILE *ofp, Jspr_parser *jp, Jspr_index ind);
extern void jspr_print_value(FILE *fp, Jspr_parser *jp, Jspr_value *v);
extern void jspr_print_values(FILE *fp, Jspr_parser *jp, int pp_level);

/* jspr_nav.c */

/* For traversing the entire parse tree or any subtree. */
extern int jspr_preorder(Jspr_parser *jp, Jspr_value *root,
						 int (*cb)(Jspr_parser *jp, Jspr_value *vm, void *),
						 void *arg);
extern int jspr_breadth_first(Jspr_parser *jp, Jspr_value *root,
							  int (*cb)(Jspr_parser *jp, Jspr_value *v, void *),
							  void *arg);
extern int jspr_element_count(Jspr_parser *jp, Jspr_value *root);
extern int jspr_child_count(Jspr_parser *jp, Jspr_value *value);

/* For searching for Object members. */
extern Jspr_value *jspr_member_lookup_x(Jspr_parser *jp, Jspr_lookup_alg alg,
										Jspr_value *v,
										char *name, Jspr_type type);
extern Jspr_value *jspr_member_lookup(Jspr_parser *jp, Jspr_value *v,
									  char *name, Jspr_type type);
extern Jspr_value *jspr_member_lookup_fast(Jspr_parser *jp, Jspr_value *v,
										   char *name, Jspr_type type);

/* For traversing the elements of an Array or Object. */
extern Jspr_value *jspr_element_first(Jspr_parser *jp, Jspr_value *v);
extern Jspr_value *jspr_element_last(Jspr_parser *jp, Jspr_value *v);
extern Jspr_value *jspr_element_next(Jspr_parser *jp, Jspr_value *v);
extern Jspr_value *jspr_element_previous(Jspr_parser *jp, Jspr_value *v);
extern Jspr_value *jspr_element_nth(Jspr_parser *jp, Jspr_value *root,
								   unsigned int nth);

/* For traversing values in an arbitrary order. */
extern Jspr_value *jspr_value_root(Jspr_parser *jp);
extern Jspr_value *jspr_value_parent(Jspr_parser *jp, Jspr_value *v);
extern Jspr_value *jspr_value_next(Jspr_parser *jp, Jspr_value *v);
extern Jspr_value *jspr_value_previous(Jspr_parser *jp, Jspr_value *v);

#ifdef __cplusplus
}
#endif

#endif /* __JSPR_H_ */
