/*
 * JSPR - JSon ParseR
 *
 * Characters and strings.
 */

/*
 * Copyright (c) 2017-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: jspr_char.c 3360 2025-03-31 23:28:47Z brachman $
 */

#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>

#include "jspr_priv.h"

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2017-2024\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: jspr_char.c 3360 2025-03-31 23:28:47Z brachman $";
#endif

/*
 * UTF-8 syntax check
 *
 * RFC 3629
 * https://datatracker.ietf.org/doc/html/rfc3629
 * https://www.unicode.org/standard/versions/enumeratedversions.html
 */

#define IS_UTF8_TAIL(CH)	((CH) >= 0x80 && (CH) <= 0xBF)
#define IS_UTF8_1(CH)		((CH) >= 0x00 && (CH) <= 0x7f)
#define IS_UTF8_2(CH1, CH2)									\
  (((CH1) >= 0xc2 && (CH1) <= 0xdf) && IS_UTF8_TAIL(CH2))

#define IS_UTF8_3_1(CH1, CH2, CH3)				\
  ((CH1) == 0xe0 && ((CH2) >= 0xa0 && (CH2) <= 0xbf) && IS_UTF8_TAIL(CH3))
#define IS_UTF8_3_2(CH1, CH2, CH3)				\
  (((CH1) >= 0xe1 && (CH1) <= 0xec) && IS_UTF8_TAIL(CH2) && IS_UTF8_TAIL(CH3))
#define IS_UTF8_3_3(CH1, CH2, CH3)				\
  ((CH1) == 0xed && ((CH2) >= 0x80 && (CH2) <= 0x9f) && IS_UTF8_TAIL(CH3))
#define IS_UTF8_3_4(CH1, CH2, CH3)				\
  (((CH1) >= 0xee && (CH1) <= 0xef) && IS_UTF8_TAIL(CH2) && IS_UTF8_TAIL(CH3))
#define IS_UTF8_3(CH1, CH2, CH3)				\
  (IS_UTF8_3_1(CH1, CH2, CH3)					\
   || IS_UTF8_3_2(CH1, CH2, CH3)				\
   || IS_UTF8_3_3(CH1, CH2, CH3)				\
   || IS_UTF8_3_4(CH1, CH2, CH3))

#define IS_UTF8_4_1(CH1, CH2, CH3, CH4)									\
  ((CH1) == 0xf0 && ((CH2) >= 0x90 && (CH2) <= 0xbf)					\
   && IS_UTF8_TAIL(CH3) && IS_UTF8_TAIL(CH4))
#define IS_UTF8_4_2(CH1, CH2, CH3, CH4)									\
  (((CH1) >= 0xf1 && (CH1) <= 0xf3)										\
   && IS_UTF8_TAIL(CH2) && IS_UTF8_TAIL(CH3) && IS_UTF8_TAIL(CH4))
#define IS_UTF8_4_3(CH1, CH2, CH3, CH4)									\
  ((CH1) == 0xf4 && ((CH2) >= 0x80 && (CH2) <= 0x8f)					\
   && IS_UTF8_TAIL(CH3) && IS_UTF8_TAIL(CH4))
#define IS_UTF8_4(CH1, CH2, CH3, CH4)									\
  (IS_UTF8_4_1(CH1, CH2, CH3, CH4)										\
   || IS_UTF8_4_2(CH1, CH2, CH3, CH4)									\
   || IS_UTF8_4_3(CH1, CH2, CH3, CH4))

/*
 * Check if the one to four bytes that begin at P represent a valid UTF-8 byte
 * sequence.
 * Return the character length in bytes, or zero if it is not a recognized
 * sequence.
 */
static int
is_utf8_sequence(const char *p)
{
  unsigned int ch1, ch2, ch3, ch4;

  /* UTF8-1 */
  ch1 = (unsigned int) *p & 0xff;
  if (IS_UTF8_1(ch1))
	return(1);

  /* UTF8-2 */
  ch2 = (unsigned int) *(p + 1) & 0xff;
  if (IS_UTF8_2(ch1, ch2))
	return(2);

  /* UTF8-3 */
  ch3 = (unsigned int) *(p + 2) & 0xff;
  if (IS_UTF8_3(ch1, ch2, ch3))
	return(3);

  /* UTF8-4 */
  ch4 = (unsigned int) *(p + 3) & 0xff;
  if (IS_UTF8_4(ch1, ch2, ch3, ch4))
	return(4);

  return(0);
}

/*
 * Validate the byte sequence starting at P as a character within a
 * JSON string, setting VAL to its value and returning the length in bytes of
 * the character.
 * If the character is invalid, return -1.
 *
 * <char>       = <unesc-char> | <esc-char>
 * ERRATA CHANGE: the '/' (0x2f) is in the unescaped range
 *   <unesc-char> = 0x20..0x21 | 0x23..0x2e | 0x30-0x5b | 0x5d..0x10ffff
 *
 * ORIGINAL:
 *   <unesc-char> = 0x20..0x21 | 0x23..0x5b | 0x5d..0x10ffff
 * <esc-char>   = '\"' | '\\' | '\/' | '\b' | '\f' | '\n' | '\r' | '\t'
 * <xxxx>       = <hexdigit> <hexdigit> <hexdigit> <hexdigit>
 */
int
char_value(Jspr_parser *jp, const char *p, int *val)
{
  int esc_ch;

  if (*p != '\\') {
	int ch, n_utf;

	/* This character was not escaped. */
	ch = (int) *p & 0xff;

	/* Reject characters that must be escaped. */
	if (ch == 0x22 || ch == 0x5c || (ch >= 0x00 && ch <= 0x1f)) {
	  SET_ERROR_MESG(jp, JSPR_ERROR_INVALID_CHAR, p, 1, "Unescaped character");
	  return(-1);
	}

	if ((n_utf = is_utf8_sequence(p)) > 0) {
	  if (n_utf == 1)
		*val = ch;
	  /* No value is set for n_utf > 1 */
	  return(n_utf);
	}

	SET_ERROR_MESG(jp, JSPR_ERROR_INVALID_CHAR, p, 1, NULL);
	return(-1);
  }

  esc_ch = (int) *++p;
  switch (esc_ch) {
  case '"':
	*val = '"';
	return(2);

  case '\\':
	*val = '\\';
	return(2);

  case '/':
	*val = '/';
	return(2);

  case 'b':
	*val = '\b';
	return(2);

  case 'f':
	*val = '\f';
	return(2);

  case 'n':
	*val = '\n';
	return(2);

  case 'r':
	*val = '\r';
	return(2);

  case 't':
	*val = '\t';
	return(2);

  case 'u':
	{
	  int ch, i, v;

	  /*
	   * Four hex digits are expected to follow...
	   */
	  v = 0;
	  for (i = 0; i < 4; i++) {
		ch = tolower((int) *++p);
		if (ch >= '0' && ch <= '9')
		  v = (v * 16) + (ch - '0');
		else if (ch >= 'a' && ch <= 'f')
		  v = (v * 16) + (ch - 'a') + 10;
		else {
		  SET_ERROR_MESG(jp, JSPR_ERROR_INVALID_CHAR, p, 1, NULL);
		  return(-1);
		}
	  }
	  *val = v;
	  return(6);
	}

  default:
	SET_ERROR_MESG(jp, JSPR_ERROR_INVALID_CHAR, p, 1,
				   "Undefined character escape");
	return(-1);
  }

  /*NOTREACHED*/
}

/*
 * Compare two (possibly encoded) characters, possibly of different lenghts,
 * returning 0 if they are equal, 1 if PP is greater, -1 if QQ is greater,
 * and -2 if an error occurs.
 * If equal, advance both pointers to the next character
 * ('b' having length 1, '\b' having length 2, and '\u0062' having length 6,
 * but all being equivalent characters).
 */
int
char_cmp(Jspr_parser *jp, const char **pp, const char **qq)
{
  int pval, pinc, qval, qinc, st;
  const char *p, *q;

  p = *pp;
  q = *qq;
  pval = qval = 0;

  /* An error return may not be expected by the caller... */
  if ((pinc = char_value(jp, p, &pval)) < 0)
	return(-2);

  if ((qinc = char_value(jp, q, &qval)) < 0)
	return(-2);

  /*
   * XXX this doesn't conform to the Unicode collating sequence so it
   * is incorrect.
   * But it might be good enough for our purposes.
   */
  if (pinc > qinc)
	st = 1;
  else if (pinc < qinc)
	st = -1;
  else if (pinc > 1)
	st = memcmp(p, q, pinc);
  else if (pval > qval)
	st = 1;
  else if (pval < qval)
	st = -1;
  else
	st = 0;

  if (st == 0) {
	(*pp) += pinc;
	(*qq) += qinc;
  }

  return(st);
}

/*
 * Compare two JSON text strings, returning -1, 0, or 1 if the first
 * string compares less than, equal to, or greater than the second.
 * If either string contains an invalid encoding, -2 is returned.
 */
int
jspr_string_cmp(Jspr_parser *jp, Jspr_value *v1, Jspr_value *v2)
{
  const char *pptr, *pend, *qptr, *qend;

  if (v1 == NULL || v2 == NULL)
	return(-2);

  pptr = js_start(jp, v1);
  pend = js_end(jp, v1);

  qptr = js_start(jp, v2);
  qend = js_end(jp, v2);

  while (pptr < pend && qptr < qend) {
	int rc;

	if ((rc = char_cmp(NULL, &pptr, &qptr)) != 0)
	  return(rc);
  }

  if (pptr != pend)
	return(1);
  if (qptr != qend)
	return(-1);

  return(0);
}

/*
 * Return 1 if the string associated with value V1 is equivalent
 * to STR (LEN bytes long), otherwise return 0.
 */
int
jspr_name_eq(Jspr_parser *jp, Jspr_value *v1, char *str, size_t len)
{
  const char *pptr, *pend, *qptr, *qend;

  if (v1 == NULL || (v1->type != JSPR_STRING_TYPE && v1->type != JSPR_NAME_TYPE))
	return(-1);

  pptr = js_start(jp, v1);
  pend = js_start(jp, v1);

  qptr = str;
  qend = str + len;

  while (pptr < pend && qptr < qend) {
	int rc;

	if ((rc = char_cmp(NULL, &pptr, &qptr)) != 0)
	  return(rc);
  }

  if (pptr != pend)
	return(0);
  if (qptr != qend)
	return(0);

  return(1);
}

/*
 * Validate a JSON text string, returning 1 if valid, 0 otherwise.
 */
int
jspr_string_validate(const char *p)
{
  const char *pptr, *pend;

  pptr = p;
  pend = p + strlen(p);

  while (pptr < pend) {
	int inc, pval;

	if ((inc = char_value(NULL, pptr, &pval)) < 0)
	  return(0);
	pptr += inc;
  }

  return(1);
}

/*
 * Extract, copy, or convert the JSON text for VALUE according to TYPE.
 * DST_PTR, a pointer to memory of the appropriate type for TYPE, is set.
 * If non-NULL, ENDP is set to the first invalid character in the original
 * JSON text after conversion, or a null byte if conversion is successful.
 *   jspr_string_value(JP, VALUE, TYPE, [BASE], DST_PTR, ENDP)
 *
 * For integer type conversions, base 10 is assumed.
 * If TYPE has the JSPR_IBASE flag set, however, the base (as an int) must
 * immediately follow the TYPE argument.
 * If the BASE argument is provided, the JSPR_IBASE flag must be given even
 * if the conversion does not use it.
 *
 * Return 0 if successful, -1 if an error occurs.
 */
int
jspr_string_value(Jspr_parser *jp, Jspr_value *value, Jspr_string_cvt type,
				  ...)
{
  int base, do_free;
  char buf[JSPR_VALUE_BUFSIZ], *bufp, **endp, *ptr, **str_ptr;
  void *dst_ptr;
  size_t len, *len_ptr, need;
  Jspr_string_cvt cvt_type;
  va_list ap;

  va_start(ap, type);

  if (type & JSPR_IBASE)
	base = va_arg(ap, int);
  else
	base = 10;
  cvt_type = (type & ~JSPR_IBASE);

  if (cvt_type == JSPR_VALUE_STRLEN) {
	str_ptr = va_arg(ap, char **);
	len_ptr = va_arg(ap, size_t *);
  }
  else {
	dst_ptr = va_arg(ap, void *);
	endp = va_arg(ap, char **);
  }
  va_end(ap);

  do_free = 0;
  bufp = NULL;

  len = js_len(jp, value);
  need = len + 1;

  /*
   * There are three different conversion methods for getting a copy of
   * character string values:
   * JSPR_VALUE_STR: Set DST_PTR to the start of the (read only) input
   *   text and set ENDP to point to the byte following its end; this string
   *   may not be null-terminated.
   * JSPR_VALUE_COPY: Given DST_PTR, a sufficiently large buffer, copy the
   *   input text into it, append a null byte, and set ENDP to point to the
   *   null byte.
   * JSPR_STR: allocate a new buffer, set DST_PTR to point to it,
   *   copy the input text into it, append a null byte, and set ENDP to
   *   point to the null byte.
   */
  if (cvt_type == JSPR_VALUE_STR || cvt_type == JSPR_VALUE_COPY) {
	char *dptr;
	char *sptr;

	/*
	 * This is a special case.
	 */
	if (value == NULL || (cvt_type == JSPR_VALUE_STR && endp == NULL))
	  goto err;

	sptr = (char *) js_start(jp, value);

	if (cvt_type == JSPR_VALUE_COPY) {
	  dptr = (char *) dst_ptr;
	  strncpy(dptr, sptr, len);
	  dptr[len] = '\0';
	  if (endp != NULL)
		*endp = dptr + len;
	}
	else {
	  /*
	   * Return a pointer to the start and end of the text - no copying and
	   * unlikely to be null-terminated.
	   * The caller should not alter the pointed-to raw data.
	   * Use JSPR_STR if a (null-terminated) copy is needed.
	   */
	  *(char **) dst_ptr = sptr;
	  if (endp != NULL)
		*endp = sptr + len;
	}

	return(0);
  }

  /*
   * Set DST_PTR to the length of the input text and set ENDP to point
   * to the byte following its end; this string may not be null-terminated.
   */
  if (cvt_type == JSPR_VALUE_STRLEN) {
	char *sptr;
	size_t len;

	if (value == NULL)
	  goto err;

	len = js_len(jp, value);
	*(size_t *) len_ptr = len;

	sptr = (char *) js_start(jp, value);
	*str_ptr = sptr;

	return(0);
  }

  /*
   * Map an explicitly-sized (unsigned) integer type to the appropriate
   * native type.
   */
  if (cvt_type == JSPR_UI32_T) {
    /* 32-bit unsigned int. */
    if (sizeof(JSPR_UI32_T_TYPE) == sizeof(JSPR_UI_TYPE))
      cvt_type = JSPR_UI;
    else {
	err:
	  if (do_free && bufp != NULL)
		free(bufp);

	  return(-1);
	}
  }
  else if (cvt_type == JSPR_UI64_T) { 
    /* 64-bit unsigned int. */
    if (sizeof(JSPR_UI64_T_TYPE) == sizeof(JSPR_UI_TYPE))
      cvt_type = JSPR_UI;
    else if (sizeof(JSPR_UI64_T_TYPE) == sizeof(JSPR_UL_TYPE))
      cvt_type = JSPR_UL;
    else if (sizeof(JSPR_UI64_T_TYPE) == sizeof(JSPR_ULL_TYPE))
      cvt_type = JSPR_ULL;
    else
	  goto err;
  }
  else if (cvt_type == JSPR_I32_T) {
    /* 32-bit int. */
    if (sizeof(JSPR_I32_T_TYPE) == sizeof(JSPR_I_TYPE))
      cvt_type = JSPR_I;
    else
	  goto err;
  }
  else if (cvt_type == JSPR_I64_T) { 
    /* 64-bit int. */
    if (sizeof(JSPR_I64_T_TYPE) == sizeof(JSPR_I_TYPE))
      cvt_type = JSPR_I;
    else if (sizeof(JSPR_I64_T_TYPE) == sizeof(JSPR_L_TYPE))
      cvt_type = JSPR_L;
    else if (sizeof(JSPR_I64_T_TYPE) == sizeof(JSPR_LL_TYPE))
      cvt_type = JSPR_LL;
    else
	  goto err;
  }

  /*
   * Copy the JSON text into a null-terminated buffer.
   * For expected (common) token sizes, use a preallocated buffer to
   * avoid repeatedly allocating and freeing.
   *
   * XXX this could (should!?) be improved by simply writing a NUL byte into
   * value->end and then restoring the original value after conversion;
   * originally wanted to guarantee jp->js would be read-only, but maybe
   * that's not so important.
   */
  if (need >= sizeof(buf)) {
	if ((bufp = malloc(need)) == NULL)
	  goto err;
	/* This will be freed before returning. */
	do_free = 1;
	JSPR_RES(jp)->xbuf_count++;
	if (need > JSPR_RES(jp)->xbuf_maxlen)
	  JSPR_RES(jp)->xbuf_maxlen = need;
  }
  else
	bufp = buf;
  strncpy(bufp, js_start(jp, value), len);
  bufp[len] = '\0';

  errno = 0;
  ptr = NULL;
  switch (cvt_type) {
  case JSPR_STR:
    {
	  char *sptr;

	  /*
	   * Return a null-terminated copy.
	   * This makes little sense if not given a VALUE pointer.
	   */
	  if (dst_ptr == NULL)
		goto err;

	  if (buf == bufp) {
		/* Copy the static buffer. */
		if ((sptr = malloc(need)) == NULL)
		  goto err;
		strcpy(sptr, bufp);
		JSPR_RES(jp)->nstrings++;
		JSPR_RES(jp)->string_alloc += need;
	  }
	  else {
		/* Use the copy made earlier. */
		sptr = bufp;
		do_free = 0;
	  }

	  *(char **) dst_ptr = sptr;

	  if (endp != NULL)
		*endp = sptr + len;
      break;
    }

  case JSPR_I:
    {
      int val_int;
      long val_long;

      val_long = strtol(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_long == 0 && errno == EINVAL))
        goto err;
      val_int = (int) val_long;
      if ((long) val_int != val_long)
        goto err;

      if (dst_ptr != NULL)
        *(int *) dst_ptr = val_int;
      break;
    }

  case JSPR_F:
    {
      float val_float;

      val_float = strtod(bufp, &ptr);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || (errno == ERANGE && val_float == 0.0)
          || (errno == ERANGE
              && (val_float == HUGE_VAL || val_float == -HUGE_VAL)))
        goto err;

      if (dst_ptr != NULL)
        *(float *) dst_ptr = val_float;
      break;
    }

  case JSPR_D:
    {
      double val_double;

      val_double = strtod(bufp, &ptr);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || (errno == ERANGE && val_double == 0.0)
          || (errno == ERANGE
              && (val_double == HUGE_VAL || val_double == -HUGE_VAL)))
        goto err;

      if (dst_ptr != NULL)
        *(double *) dst_ptr = val_double;
      break;
    }
  case JSPR_L:
    {
      long val_long;

      val_long = strtol(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_long == 0 && errno == EINVAL))
        goto err;

      if (dst_ptr != NULL)
        *(long *) dst_ptr = val_long;
      break;
    }

  case JSPR_LL:
    {
      long long val_long_long;

      val_long_long = strtoll(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_long_long == 0 && errno == EINVAL))
        goto err;

      if (dst_ptr != NULL)
        *(long long *) dst_ptr = val_long_long;
      break;
    }

  case JSPR_IMAX_T:
    {
      intmax_t val_intmax;

      val_intmax = strtoimax(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_intmax == 0 && errno == EINVAL))
        goto err;

      if (dst_ptr != NULL)
        *(intmax_t *) dst_ptr = val_intmax;
      break;
    }

  case JSPR_UI:
  case JSPR_UINZ:
    {
      unsigned int val_uint;
      unsigned long val_ulong;

      val_ulong = strtoul(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
        goto err;

      val_uint = (unsigned int) val_ulong;
      if ((unsigned long) val_uint != val_ulong)
        goto err;

      /* Unsigned integer, but not zero. */
      if (cvt_type == JSPR_UINZ && val_uint == 0)
        goto err;

      if (dst_ptr != NULL)
        *(unsigned int *) dst_ptr = val_uint;
      break;
    }

  case JSPR_IN_PORT_T:
    {
      in_port_t val_in_port_t;
      unsigned long val_ulong;

      val_ulong = strtoul(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
        goto err;

      val_in_port_t = (in_port_t) val_ulong;
      if ((unsigned long) val_in_port_t != val_ulong)
        goto err;

      if (dst_ptr != NULL)
        *(in_port_t *) dst_ptr = val_in_port_t;
      break;
    }
  case JSPR_UL:
    {
      unsigned long val_ulong;

      val_ulong = strtoul(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
        goto err;

      if (dst_ptr != NULL)
        *(unsigned long *) dst_ptr = val_ulong;
      break;
    }

  case JSPR_ULL:
    {
      unsigned long long val_ulong_long;

      val_ulong_long = strtoull(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_ulong_long == 0 && errno == EINVAL))
        goto err;

      if (dst_ptr != NULL)
        *(unsigned long long *) dst_ptr = val_ulong_long;
      break;
    }

  case JSPR_UIMAX_T:
    {
      uintmax_t val_uintmax;

      val_uintmax = strtoumax(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_uintmax == 0 && errno == EINVAL))
        goto err;

      if (dst_ptr != NULL)
        *(uintmax_t *) dst_ptr = val_uintmax;
      break;
    }

  case JSPR_TIME_T:
    {
      unsigned long val_ulong;
      time_t val_time_t;

      val_ulong = strtoul(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
        goto err;

      val_time_t = (time_t) val_ulong;
      if ((unsigned long) val_time_t != val_ulong)
        goto err;

      if (dst_ptr != NULL)
        *(time_t *) dst_ptr = val_time_t;
      break;
    }

  case JSPR_SIZE_T:
    {
      unsigned long val_ulong;
      size_t val_size_t;

      val_ulong = strtoul(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
        goto err;

      val_size_t = (size_t) val_ulong;
      if ((unsigned long) val_size_t != val_ulong)
        goto err;

      if (dst_ptr != NULL)
        *(size_t *) dst_ptr = val_size_t;
      break;
    }

  case JSPR_SSIZE_T:
    {
      long val_long;
      ssize_t val_ssize_t;

      val_long = strtol(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_long == 0 && errno == EINVAL))
        goto err;

      val_ssize_t = (ssize_t) val_long;
      if ((long) val_ssize_t != val_long)
        goto err;

      if (dst_ptr != NULL)
        *(ssize_t *) dst_ptr = val_ssize_t;
      break;
    }

  case JSPR_PID_T:
    {
      unsigned long val_ulong;
      pid_t val_pid_t;

      val_ulong = strtoul(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
        goto err;

      val_pid_t = (pid_t) val_ulong;
      if ((unsigned long) val_pid_t != val_ulong)
        goto err;

      if (dst_ptr != NULL)
        *(pid_t *) dst_ptr = val_pid_t;
      break;
    }

  case JSPR_MODE_T:
    {
      unsigned long val_ulong;
      mode_t val_mode_t;

      val_ulong = strtoul(bufp, &ptr, base);
      if (endp != NULL)
        *endp = ptr;
      if ((endp == NULL && *ptr != '\0')
          || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
        goto err;

      val_mode_t = (mode_t) val_ulong;
      if ((unsigned long) val_mode_t != val_ulong)
        goto err;

      if (dst_ptr != NULL)
        *(mode_t *) dst_ptr = val_mode_t;
      break;
    }

  default:
    if (endp != NULL)
      *(const char **) endp = bufp;
	goto err;
  }

  /* Normal return. */

  if (do_free && bufp != NULL)
	free(bufp);

  return(0);
}

/*
 * Return a null-terminated copy of V's text in the original input string.
 * The caller should free the returned string.
 */
char *
jspr_string_copy(Jspr_parser *jp, Jspr_value *value)
{
  char *s;
  size_t name_len;

  name_len = js_end(jp, value) - js_start(jp, value);
  s = (char *) malloc(name_len + 1);
  strncpy(s, js_start(jp, value), name_len);
  s[name_len] = '\0';

  return(s);
}

/*
 * Compare an Object member's name (a JSON string) to NAME of length NAME_LEN
 * for equality.
 * NAME is not surrounded by quotation marks.
 * An encoding error in either string returns "not equal".
 * Note that the start byte of VALUE in the JSON text is the character
 * following the initial quotation mark and its end byte is the closing
 * quotation mark.
 *
 * XXX
 * RFC 8259:
 * Software implementations are typically required to test names of
 * Object members for equality.  Implementations that transform the
 * textual representation into sequences of Unicode code units and then
 * perform the comparison numerically, code unit by code unit, are
 * interoperable in the sense that implementations will agree in all
 * cases on equality or inequality of two strings.  For example,
 * implementations that compare strings with escaped characters
 * unconverted may incorrectly find that "a\\b" and "a\u005Cb" are not equal.
 */
int
jspr_value_name_eq(Jspr_parser *jp, Jspr_value *value, char *name,
				   size_t name_len)
{
  const char *nptr, *nend, *tptr, *tend;

  /*
   * If the strings are both marked as UTF8_1
   * we can use a fast string comparison for this common unencoded case.
   */
  if (value->type != JSPR_STRING_TYPE && value->type != JSPR_NAME_TYPE)
	return(0);

  tptr = js_start(jp, value);
  tend = js_end(jp, value);

  if (*tend != '"')
	return(0);

  nptr = name;
  nend = name + name_len;

  while (nptr < nend && tptr < tend) {
	if (value->flags == JSPR_VALUE_STR_CLEAN
		|| value->flags == JSPR_VALUE_STR_UTF8_1) {
	  if (*nptr != *tptr)
		return(0);
	  nptr++;
	  tptr++;
	}
	else {
	  /* An error in comparison returns "not equal". */
	  if (char_cmp(jp, &tptr, &nptr) != 0)
		return(0);
	}
  }

  if (nptr != nend || tptr != tend)
	return(0);

  return(1);
}

