/*
 * JSPR - JSon ParseR
 *
 * I/O and debugging.
 */

/*
 * Copyright (c) 2017-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: jspr_io.c 3360 2025-03-31 23:28:47Z brachman $
 */

#include <stdlib.h>
#include <stdio.h>

#include "jspr_priv.h"

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2017-2024\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: jspr_io.c 3360 2025-03-31 23:28:47Z brachman $";
#endif

/* Indexed by Jspr_type values, so the order is important. */
static char *jspr_type_name[] = {
  "Invalid", "Object", "Array", "String", "Name",
  "Number", "Boolean", "Null", "EndObject", "EndArray", "Any", "Unknown"
};

/*
 * Print STR to FP and repeat INDENT times for pretty printing.
 */
static void
print_indent(FILE *fp, char *str, int indent)
{
  int i;

  for (i = 0; i < indent; i++)
	fprintf(fp, "%s", str);
}

/*
 * Print information about value IND to OFP (default: stderr) in a
 * simple tabular form, without highlighting, etc.
 *
 * Use jspr_print_value() to pretty print a value with less detail
 * (also see jspr_print_values()).
 */
void
jspr_dump_value(FILE *ofp, Jspr_parser *jp, Jspr_index ind)
{
  Jspr_index child_last, sib_prev;
  Jspr_value *pv, *v;
  FILE *fp;

  if ((fp = ofp) == NULL)
	fp = stderr;

  v = &jp->values[ind];

#ifdef CHILD_LAST
  child_last = v->child_last;
#else
  child_last = find_child_last(jp, v);
#endif

#ifdef SIB_PREV
  sib_prev = v->sib_prev;
#else
  sib_prev = find_sib_prev(jp, v);
#endif

  if (1) {
	Jspr_index ind;

	ind = value_index(jp, v);
	fprintf(fp, "%3u %-9s %4lu/%4lu %3u  #%u: %3u/%-3u %3u/%-3u  %3u   ",
			ind, jspr_type_name[v->type],
			js_start_offset(jp, v), js_end_offset(jp, v),
			v->parent,
			v->nchildren, v->child_first, child_last,
			sib_prev, v->sib_next,
			v->name);
  }

  switch (v->type) {
  case JSPR_OBJECT_TYPE:
	if (v->nchildren == 0)
	  fprintf(fp, "Object { }");
	else {
	  fprintf(fp, "Object {");
	}
	break;

  case JSPR_ARRAY_TYPE:
	if (v->nchildren == 0)
	  fprintf(fp, "Array [ ]");
	else {
	  fprintf(fp, "Array [");
	}
	break;

  case JSPR_NAME_TYPE:
	if (v->flags == JSPR_VALUE_STR_CLEAN) {
	  char *str;
	  size_t len;

	  jspr_string_value(jp, v, JSPR_VALUE_STRLEN, &str, &len);
	  fprintf(fp, "\"%.*s\" :", (int) len, str);
	}
	else
	  fprintf(fp, "(name 0%o)", v->flags);
	break;

  case JSPR_STRING_TYPE:
	if (v->flags == JSPR_VALUE_STR_CLEAN) {
	  char *str;
	  size_t len;

	  jspr_string_value(jp, v, JSPR_VALUE_STRLEN, &str, &len);
	  fprintf(fp, "\"%.*s\"", (int) len, str);
	}
	else
	  fprintf(fp, "(string 0%o)", v->flags);
	break;

  case JSPR_NUMBER_TYPE:
	if (v->flags == JSPR_VALUE_NUMBER_EXP) {
	  double dval;

	  jspr_string_value(jp, v, JSPR_D, &dval, NULL);
	  fprintf(fp, "%E", dval);
	}
	else if (v->flags == JSPR_VALUE_NUMBER_REAL) {
	  double dval;

	  jspr_string_value(jp, v, JSPR_D, &dval, NULL);
	  fprintf(fp, "%f", dval);
	}
	else {
	  int ival;

	  jspr_string_value(jp, v, JSPR_INT, &ival, NULL);
	  fprintf(fp, "%d", ival);
	}
	break;

  case JSPR_BOOLEAN_TYPE:
	fprintf(fp, "%s",
			(v->flags == JSPR_VALUE_BOOLEAN_FALSE) ? "false" : "true");
	break;

  case JSPR_NULL_TYPE:
	fprintf(fp, "null");
	break;

  case JSPR_END_ARRAY_TYPE:
	fprintf(fp, "]");
	break;

  case JSPR_END_OBJECT_TYPE:
	fprintf(fp, "}");
	break;

  case JSPR_COMMENT_TYPE:
	fprintf(fp, "<comment>");

  default:
	fprintf(fp, "[unknown]");
	break;
  }

  /* As in print_value(). */
  pv = value_ptr(jp, v->parent);

  if (value_is_struct(v) && v->sib_next != 0 && v->nchildren == 0) {
	/* The has-sibling, empty Array or Object case. */
	fprintf(fp, ",");
  }
  else if (v->sib_next != 0 && !value_is_struct(v) && !value_is_end_marker(v)) {
	/* The has-sibling, non-Array/Object case. */
	fprintf(fp, ",");
  }
  else if (value_is_end_marker(v) && pv != NULL && pv->sib_next != 0) {
	/* The parent of this end of structure has a sibling. */
	fprintf(fp, ",");
  }

  fprintf(fp, "\n");
  fflush(fp);
}

void
jspr_dump(FILE *ofp, Jspr_parser *jp)
{
  Jspr_index ind;
  FILE *fp;

  if ((fp = ofp) == NULL)
	fp = stderr;

  fprintf(fp,
		  "Ind Type     Start/End   Pt  #C   CF/CL   SP/SN    Name  Src\n");

  ind = JSPR_VALUE_ROOT_INDEX;
  while (ind < jp->next_value) {
	jspr_dump_value(fp, jp, ind);
	ind++;
  }
}

/*
 * Print value V to OFP (default is stderr).
 * The amount of information increases with PP_LEVEL
 * (pretty print level: 0, 1, 2).
 * The current level of indentation for pretty printing is maintained
 * through INDENTP.
 *
 * If the configuration has set enable_colourize, the pretty printed
 * output is coloured to make it easier to identify and match token types.
 *
 * XXX not a priority, but colours should be customizable.
 * (Can copy a complete implementation from lf.c)
 *
 * If enable_verbose is set, pretty printed Array/Object beginnings and ends
 * are marked with a numeric Array/Object identifier, followed by a period,
 * followed by the structure's numeric nesting depth.
 */
static void
print_value(FILE *ofp, Jspr_parser *jp, Jspr_value *v, int pp_level)
{
  int indent;
  char *c_literal, *c_name, *c_inum, *c_reset, *c_rnum, *c_string, *c_struct;
  char *c_error;
  Jspr_index child_last, ind, sib_prev;
  Jspr_value *pv;
  FILE *fp;

  if ((fp = ofp) == NULL)
	fp = stderr;

  if (JSPR_CONFIG(jp)->enable_colourize) {
#ifdef NOTDEF
	c_black = "\033[38;2;0;0;0;1m";
#endif
	c_error = "\033[38;2;255;138;0";
	c_literal = "\033[38;5;28;1m";
	c_name = "\033[31;1m";
	c_inum = "\033[38;5;106;1m";
	c_reset = "\033[0m";
	c_rnum = "\033[91;1m";
	c_string = "\033[38;2;0;0;0;1m";
	c_struct = "\033[34;1m";
  }
  else
	c_error = c_literal = c_name = c_inum = c_reset = c_rnum
	  = c_string = c_struct = "";

#ifdef CHILD_LAST
  child_last = v->child_last;
#else
  child_last = find_child_last(jp, v);
#endif

#ifdef SIB_PREV
  sib_prev = v->sib_prev;
#else
  sib_prev = find_sib_prev(jp, v);
#endif

  if (pp_level > 0) {
	Jspr_index ind;

	ind = value_index(jp, v);
	fprintf(fp, "%3u %-9s %4lu/%4lu %3u  #%u: %3u/%-3u %3u/%-3u  %3u   ",
			ind, jspr_type_name[v->type],
			js_start_offset(jp, v), js_end_offset(jp, v),
			v->parent,
			v->nchildren, v->child_first, child_last,
			sib_prev, v->sib_next,
			v->name);
  }

  /*
   * Pretty printing one value node, V
   *
   * 0. Initialize
   *    0a. Get the current base indent level.
   * 1. Print V:
   *    1a. If V is a member value, temporarily increase the base indent.
   *    1b. Emit the base indent and any temporary indent.
   *    1c. Emit V
   *    1d. If V is the start of a non-empty struct:
   *        2a. Increment the base indent, emit a NL, done.
   *    1e. If V is the last element of a struct:
   *        2a. Emit a NL.
   *        2b. Decrement the base indent and emit a closing brace/bracket.
   *        2c. Make V the parent of V (or zero).
   *        2d. Emit the base indent
   *        2e. Goto 1e.
   *    1f. If V is an element but not the last element:
   *        3a. Emit a comma and NL, done.
   *    1g. If V is not an element, emit a NL, done.
   */

  indent = v->array_level + v->object_level * 2;
  if (indent > 32)
	indent = 36;

  /*
   * If this is a member value, indent it a little more for visibility.
   */
  ind = value_index(jp, v);
  pv = value_parent(jp, ind);
  if ((v->array_level + v->object_level) != 0) {
	if ((v->type != JSPR_NAME_TYPE && v->name != 0)
		|| (v->type != JSPR_NAME_TYPE && pv != NULL && value_is_struct(pv))
		|| (v->type == JSPR_END_OBJECT_TYPE && v->parent != 0))
	  print_indent(fp, "  ", indent + 1);
	else
	  print_indent(fp, "  ", indent);	
  }

  switch (v->type) {
	int rc;

  case JSPR_OBJECT_TYPE:
	if (v->nchildren == 0) {
	  if (JSPR_CONFIG(jp)->enable_verbose)
		fprintf(fp, "%sObject { }%s", c_struct, c_reset);
	  else
		fprintf(fp, "%s{ }%s", c_struct, c_reset);
	}
	else {
	  if (JSPR_CONFIG(jp)->enable_verbose)
		fprintf(fp, "%sObject %u.%u{%s",
				c_struct, v->struct_count, v->object_level, c_reset);
	  else
		fprintf(fp, "%s{%s", c_struct, c_reset);
	}
	break;

  case JSPR_ARRAY_TYPE:
	if (v->nchildren == 0) {
	  if (JSPR_CONFIG(jp)->enable_verbose)
		fprintf(fp, "%sArray [ ]%s", c_struct, c_reset);
	  else
		fprintf(fp, "%s[ ]%s", c_struct, c_reset);
	}
	else {
	  if (JSPR_CONFIG(jp)->enable_verbose)
		fprintf(fp, "%sArray %u.%u[%s",
				c_struct, v->struct_count, v->array_level, c_reset);
	  else
		fprintf(fp, "%s[%s", c_struct, c_reset);
	}
	break;

  case JSPR_NAME_TYPE:
	if (v->flags == JSPR_VALUE_STR_CLEAN) {
	  char *str;
	  size_t len;

	  jspr_string_value(jp, v, JSPR_VALUE_STRLEN, &str, &len);
	  fprintf(fp, "\"%s%.*s%s\" :", c_name, (int) len, str, c_reset);
	}
	else
	  fprintf(fp, "(name 0%o)", v->flags);
	break;

  case JSPR_STRING_TYPE:
	if (v->flags == JSPR_VALUE_STR_CLEAN) {
	  char *str;
	  size_t len;

	  jspr_string_value(jp, v, JSPR_VALUE_STRLEN, &str, &len);
	  fprintf(fp, "\"%s%.*s%s\"", c_string, (int) len, str, c_reset);
	}
	else
	  fprintf(fp, "(string 0%o)", v->flags);
	break;

  case JSPR_NUMBER_TYPE:
	if (v->flags == JSPR_VALUE_NUMBER_EXP) {
	  double dval;

	  rc = jspr_string_value(jp, v, JSPR_D, &dval, NULL);
	  if (rc != -1)
		fprintf(fp, "%s%E%s", c_rnum, dval, c_reset);
	}
	else if (v->flags == JSPR_VALUE_NUMBER_REAL) {
	  double dval;

	  rc = jspr_string_value(jp, v, JSPR_D, &dval, NULL);
	  if (rc != -1)
		fprintf(fp, "%s%f%s", c_rnum, dval, c_reset);
	}
	else {
	  char *s;

	  /*
	   * This is an integer, but in this context there is no integer
	   * subtype (e.g., signed vs. unsigned, radix) or size specification
	   * so we can't check it by doing a conversion.
	   */
	  s = jspr_string_copy(jp, v);
	  fprintf(fp, "%s%s%s", c_inum, s, c_reset);
	  free(s);
	  rc = 0;
	}

	if (rc == -1) {
	  char *s;

	  /* A conversion error occurred, so hilight that. */
	  s = jspr_string_copy(jp, v);
	  fprintf(stderr, "%s(%s)%s", c_error, s, c_reset);
	  free(s);
	}

	break;

  case JSPR_BOOLEAN_TYPE:
	fprintf(fp, "%s%s%s",
			c_literal,
			(v->flags == JSPR_VALUE_BOOLEAN_FALSE) ? "false" : "true",
			c_reset);
	break;

  case JSPR_NULL_TYPE:
	fprintf(fp, "%snull%s", c_literal, c_reset);
	break;

  case JSPR_END_ARRAY_TYPE:
	if (JSPR_CONFIG(jp)->enable_verbose)
	  fprintf(fp, "%s]%u.%u%s",
			  c_struct, v->struct_count, v->array_level, c_reset);
	else
	  fprintf(fp, "%s]%s", c_struct, c_reset);
	break;

  case JSPR_END_OBJECT_TYPE:
	if (JSPR_CONFIG(jp)->enable_verbose)
	  fprintf(fp, "%s}%u.%u%s",
			  c_struct, v->struct_count, v->object_level, c_reset);
	else
	  fprintf(fp, "%s}%s", c_struct, c_reset);
	break;

  case JSPR_COMMENT_TYPE:
	fprintf(fp, "<comment>");

  default:
	fprintf(fp, "[unknown]");
	break;
  }

  pv = value_ptr(jp, v->parent);

  if (value_is_struct(v) && v->sib_next != 0 && v->nchildren == 0) {
	/* The has-sibling, empty Array or Object case. */
	fprintf(fp, ",");
  }
  else if (v->sib_next != 0 && !value_is_struct(v) && !value_is_end_marker(v)) {
	/* The has-sibling, non-Array/Object case. */
	fprintf(fp, ",");
  }
  else if (value_is_end_marker(v) && pv != NULL && pv->sib_next != 0) {
	/* The parent of this end of structure has a sibling. */
	fprintf(fp, ",");
  }

  fprintf(fp, "\n");
  fflush(fp);

}

/*
 *  Print value V to OFP (default stderr).
 */
void
jspr_print_value(FILE *ofp, Jspr_parser *jp, Jspr_value *v)
{
  FILE *fp;

  if ((fp = ofp) == NULL)
	fp = stderr;

  print_value(fp, jp, v, 0);
}

/*
 * Print to OFP (default stderr) all values from the parse tree in JP.
 */
void
jspr_print_values(FILE *ofp, Jspr_parser *jp, int pp_level)
{
  int i;
  FILE *fp;
  Jspr_value *v;

  if ((fp = ofp) == NULL)
	fp = stderr;

  if (pp_level > 1) {
	fprintf(fp, "Input:\n%s\n", jp->js);
	fprintf(fp,
			"Ind Type     Start/End   Pt  #C   CF/CL   SP/SN    Name  Src\n");
  }

  /*
   * The parsing algorithm produces a vector values that represents a
   * left-to-right scan of the input string.
   */
  for (i = JSPR_VALUE_ROOT_INDEX; i < jp->next_value; i++) {
	v = &jp->values[i];
	print_value(fp, jp, v, pp_level);
  }

  if (pp_level > 1) {
	fprintf(fp, "-----\n");
	fprintf(fp, "values used:       %u\n", JSPR_RES(jp)->num_values_used);
	fprintf(fp, "values allocated:  %u (%lu bytes)\n",
			JSPR_RES(jp)->num_values,
			JSPR_RES(jp)->num_values * sizeof(Jspr_value));
	fprintf(fp, "value expands:     %u\n", JSPR_RES(jp)->nexpands);
	fprintf(fp, "temp string count: %u\n", JSPR_RES(jp)->nstrings);
	fprintf(fp, "temp string alloc: %lu bytes\n", JSPR_RES(jp)->string_alloc);
	fprintf(fp, "xbuf_count:        %lu\n", JSPR_RES(jp)->xbuf_count);
	fprintf(fp, "xbuf_maxlen:       %lu bytes\n", JSPR_RES(jp)->xbuf_maxlen);

	fprintf(fp, "\n");
  }
}
