/*
 * JSPR - JSon ParseR
 *
 * Generate a parse tree from a JSON input string using
 * a time and memory efficient RFC 8259 conformant JSON parser.
 * It requires only a few standard C library functions.
 *
 * The design was inspired by the jsmn parser written by Serge A. Zaitsev
 * (approx. 6-Jun-2013) but has been nearly completely redesigned and
 * reimplemented for use with DACS.
 * jsmn was released under the MIT License and is
 * Copyright (c) 2010 Serge A. Zaitsev
 * See:
 *  https://zserge.com/jsmn/
 *  https://github.com/zserge/jsmn
 *
 * The parse is left-to-right recursive and generates a vector of "values"
 * (i.e., tokens or nodes) that refer to the original input text.
 * The validity of values, beyond their syntax, is mostly left up to the
 * caller to decide (e.g., a number might be too large or too small, there might
 * be duplicate Object member names, etc.).
 *
 * Following a successful parse, the caller can search/traverse the resulting
 * parse tree and convert (and validate) raw text values into an internal
 * form. For instance, following a successful parse,
 * numbers have been validated syntactically but not converted to a
 * binary representation. The character string representation in the input
 * text can be examined, converted, and checked for size issues
 * such as overflow; see jspr_string_value().
 *
 * During parsing, a user-supplied callback function (with a user-supplied
 * argument) can be called each time an Object member is completed.
 *
 * For numbers and string, flags can be examined to determine certain
 * characteristics (e.g., whether it is real or integer number).
 *
 * Although two different binary characters/strings can be equivalent
 * according to Unicode normalization rules, the parser will consider them
 * to be different in all contexts; see char_value() and char_cmp().
 *
 * The parse tree is structured so that building it incrementally (by repeatedly
 * calling the parser) is fairly efficient and post-parse operations
 * (which are read-only) are efficient enough.
 *
 * The JSON input string is not copied and must be accessible as long
 * as elements of the parse tree may need to reference it (e.g., for
 * conversion to an internal representation).
 *
 * This parser is more complicated than jsmn, but should still be fast,
 * efficient, portable, and fairly small.
 * Those requiring a truly minimal implementation should consider jsmn.
 *
 * Note:
 *   o minimal memory use is not paramount; the memory pool for values can
 *     grow as needed (optionally, and to a specified limit)
 *   o after a successful parse, parse tree navigation functions can be used.
 *   o there are no compile-time options, some default configuration
 *     can be adjusted at run-time.
 *   o there is some run-time configuration
 *   o there is some parser resource usage instrumentation
 *
 * There are many differences from jsmn, such as:
 *   o the "non-strict" mode of operation is not supported; exactly what
 *     language that accepts is undefined/undocumented, so I'm not sure how
 *     it could tested or used in practice.  Perhaps it's for compatibility
 *     with earlier JSON specifications, which may not have insisted on
 *     quotation marks around member names, for example.
 *     Note: "A JSON parser MAY accept non-JSON forms or extensions."
 *   o "Parent links", used to navigate a parse tree, are optional in the
 *     jspr but always maintained here.
 *   o Escaped hex characters in strings are supported ("/u123A").
 *
 * This parser is expected to conform to RFC 8259.
 *    https://tools.ietf.org/html/rfc8259
 *    https://www.rfc-editor.org/rfc/rfc4627.txt
 * RFC 4627 (2006) was replaced by RFC 7158 (2013), which was
 * replaced by RFC 7159 (2014), which was replaced by RFC 8259 (2017).
 * And don't forget RFC 7493, which defines I-JSON, a "restricted profile
 * of JSON designed to maximize interoperability and increase confidence that
 * software can process it successfully with predictable results".
 *
 * The nice thing about standards is that you have so many to choose from.
 *   A. Tanenbaum, Computer Networks, 2nd Edition, 1988, p. 254
 *
 * Notes from RFC 8259:
 * An implementation may set limits on
 *   o the size of texts that it accepts;
 *   o the maximum depth of nesting;
 *   o the range and precision of numbers;
 *   o the length and character contents of strings.
 *
 * Unicode characters:
 * However, the ABNF in this specification allows member names and
 * string values to contain bit sequences that cannot encode Unicode
 * characters; for example, "\uDEAD" (a single unpaired UTF-16
 * surrogate). Instances of this have been observed, for example, when
 * a library truncates a UTF-16 string without checking whether the
 * truncation split a surrogate pair.  The behavior of software that
 * receives JSON texts containing such values is unpredictable; for
 * example, implementations might return different values for the length
 * of a string value or even suffer fatal runtime exceptions.
 *
 * XXX:
 * - There are only basic accessor functions, but more can easily be added as
 *   needed; e.g., step through duplicate names (first, last, nth), narrow
 *   lookups to names within a specified Object.
 */

/*
 * Copyright (c) 2017-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: jspr_parse.c 3360 2025-03-31 23:28:47Z brachman $
 */

#ifdef NOTDEF
 * Does not support:
 * o trailing commas: [1, 2,]
 * o Special numbers NaN and Infinity (and negative forms)
#endif

#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include "jspr_priv.h"

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2017-2024\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: jspr_parse.c 3360 2025-03-31 23:28:47Z brachman $";
#endif

/* Current line number in current parse. */
unsigned int curr_linenum = 0;

/* Current character position within curr_linenum. */
unsigned int curr_linepos = 0;

/* The six structural characters. */
#define BEGIN_ARRAY_CHAR	((int) '[')
#define END_ARRAY_CHAR		((int) ']')
#define BEGIN_OBJECT_CHAR	((int) '{')
#define END_OBJECT_CHAR		((int) '}')
#define NAME_SEP_CHAR		((int) ':')
#define VALUE_SEP_CHAR		((int) ',')

#ifdef JSPR_COMPACT
int jspr_is_compact = 1;
#else
int jspr_is_compact = 0;
#endif

/* Woof Be Floyd! Embrace double recursion. */
static Jspr_value *match_value(Jspr_parser *, Jspr_index, const char **);

/*
 * Expand the size of the value pool in JP to NEED_VALUES, if possible.
 * The new pool is initialized by copying and adjusting addressed within
 * the current pool and JP is set to use it.
 *
 * Return the root (start) of the new pool, or NULL if an error occurs.
 *
 * This can be called at any time but all active Jspr_value pointers
 * will become invalid and may need to be recomputed using value_ptr()
 * before they are dereferenced.
 */
Jspr_value *
jspr_expand_values(Jspr_parser *jp, unsigned int new_length)
{
  unsigned int need;
  Jspr_value *new_values;

  if (new_length < JSPR_CONFIG(jp)->num_values)
	return(jp->values);

  /*
   * If a limit has been enabled, check if the new number of values
   * would exceed the limit.
   */
  need = new_length;
  if (JSPR_CONFIG(jp)->values_limit > 0
	  && need > JSPR_CONFIG(jp)->values_limit) {
	/*
	 * If the number of values requested would exceed the maximum allowed,
	 * reduce the request to the limit.
	 */
	if (jp->next_value != JSPR_CONFIG(jp)->values_limit)
	  need = jp->config.values_limit;
	else {
	  SET_ERROR(jp, JSPR_ERROR_VALUE_LIMIT, NULL, 1);
	  return(NULL);
	}
  }

  if (!JSPR_CONFIG(jp)->enable_expand_values) {
	SET_ERROR(jp, JSPR_ERROR_VALUE_LIMIT, NULL, 1);
	return(NULL);
  }

  new_values = (Jspr_value *) malloc(sizeof(Jspr_value) * need);
  if (new_values == NULL) {
	SET_ERROR(jp, JSPR_ERROR_NOMEM, NULL, 1);
	return(NULL);
  }
  memcpy(new_values, jp->values,
		 sizeof(Jspr_value) * JSPR_CONFIG(jp)->num_values);

  if (JSPR_CONFIG(jp)->enable_debug)
	fprintf(stderr, "Expanded: %u total values\n", need);

  if (JSPR_CONFIG(jp)->user_free_func != NULL)
	(JSPR_CONFIG(jp)->user_free_func)(jp->values);
  else
	free(jp->values);

  jp->values = new_values;
  JSPR_CONFIG(jp)->num_values = need;
  JSPR_RES(jp)->num_values = need;

  JSPR_RES(jp)->nexpands++;

  return(new_values);
}

/*
 * Allocate a new, expanded value vector for JP, recomputing all of
 * the parent pointers for addresses within the new pool, and freeing
 * the old value vector.
 *
 * The new length is determined from a configured callback, if provided,
 * otherwise by doubling the current length.
 */
static int
expand_value_pool(Jspr_parser *jp)
{
  unsigned int new_length;

  if (JSPR_CONFIG(jp)->user_expand_values_cb != NULL)
	new_length
	  = JSPR_CONFIG(jp)->user_expand_values_cb(jp, JSPR_CONFIG(jp)->num_values);
  else
	new_length = JSPR_CONFIG(jp)->num_values * 2;

  if (jspr_expand_values(jp, new_length) == NULL) {
	/* An error will have been noted. */
	return(-1);
  }

  return(0);
}

static Jspr_value *
value_init(Jspr_parser *jp, Jspr_value *v, Jspr_index parent, Jspr_type type,
		   const char *start, const char *end)
{

  v->type = type;
  v->flags = 0;
  v->nchildren = 0;

#ifdef JSPR_COMPACT
  if (start == NULL) {
	v->js_start = 0;
	v->js_len = 0;
  }
  else {
	v->js_start = start - jp->js;
	v->js_len = end - start;
  }
#else
  v->start = start;
  v->end = end;
#endif

  v->name = 0;
  v->parent = parent;

  v->child_first = 0;
#ifdef CHILD_LAST
  v->child_last = 0;
#endif

#ifdef SIB_PREV
  v->sib_prev = 0;
#endif
  v->sib_next = 0;

  /* Assign a new identifier to structure begin/end value types. */
  if (type == JSPR_ARRAY_TYPE)
	v->struct_count = jp->array_count++;
  else if (type == JSPR_OBJECT_TYPE)
	v->struct_count = jp->object_count++;
  else if (type == JSPR_END_ARRAY_TYPE || type == JSPR_END_OBJECT_TYPE) {
	Jspr_value *pv;

	pv = value_ptr(jp, v->parent);
	v->struct_count = pv->struct_count;
  }
  v->array_level = jp->array_level;
  v->object_level = jp->object_level;

  return(v);
}

/*
 * Return an initialized value.
 * NOTE: after this is called the caller should recompute any active value
 * pointers before they are used again since the value pool may have
 * been reallocated.
 */
static Jspr_value *
value_alloc(Jspr_parser *jp, Jspr_index parent, Jspr_type type,
			const char *start, const char *end)
{
  Jspr_value *v;

  if (JSPR_CONFIG(jp)->enable_no_values) {
	static Jspr_value dummy_value;

	v = &dummy_value;
	value_init(jp, v, parent, type, start, end);
	JSPR_RES(jp)->num_values_used++;
	return(v);
  }

  if (jp->next_value >= JSPR_CONFIG(jp)->num_values) {
	if (expand_value_pool(jp) == -1) {
	  /* An error will have been noted. */
	  return(NULL);
	}
  }

  /* Check if the new size would cause an overflow. */
  if ((jp->next_value + 1) == 0) {
	SET_ERROR(jp, JSPR_ERROR_VALUE_LIMIT, NULL, 1);
	return(NULL);
  }

  if (JSPR_CONFIG(jp)->enable_debug)
	fprintf(stderr, "Alloc: value[%u]\n", jp->next_value);

  v = &jp->values[jp->next_value++];
  value_init(jp, v, parent, type, start, end);

  JSPR_RES(jp)->num_values_used++;

  return(v);
}

static int
match_unesc_char(const char **inp, int ch)
{

  if (**inp == ch) {
	(*inp)++;
	return(1);
  }

  return(0);
}

/*
 * Ignore a C++ style comment.
 * A non-standard extension that must be enabled at runtime.
 * If the start-of-comment sequence (//) is found where whitespace is allowed,
 * ignore the remainder of the input line.
 *
 * <cpp-comment>       = '//' [1*non-eol-char] ('\n' / '\r')
 */
static void
skip_cpp_comment(const char **inpp)
{
  const char *p;

  p = *inpp;
  if (*p == '/' && *(p + 1) == '/') {
	p += 2;
	while (*p != '\n' && *p != '\r')
	  p++;
	curr_linenum++;
	*inpp = ++p;
  }
}

/*
 * Ignore a Shell style comment.
 * A non-standard extension that must be enabled at runtime.
 * If the start-of-comment character (#) is found where whitespace is allowed,
 * ignore the remainder of the input line.
 *
 * <shell-comment>       = '#' [1*non-eol-char] ('\n' / '\r')
 */
static void
skip_shell_comment(const char **inpp)
{
  const char *p;

  p = *inpp;
  if (*p == '#') {
	p++;
	while (*p != '\n' && *p != '\r')
	  p++;
	curr_linenum++;
	*inpp = ++p;
  }
}

/*
 * Ignore a C style comment.
 * A non-standard extension that must be enabled at runtime.
 * If the start-of-comment sequence (slash-star) is found where whitespace
 * is allowed, ignore input until the end-of-comment sequence (star-slash)
 * is seen.
 * Nested comments are not allowed.
 *
 * <c-comment>       = slash-star [1*char] star-slash
 */
static Jspr_rc
skip_c_comment(Jspr_parser *jp, const char **inpp)
{
  const char *p;

  p = *inpp;
  if (*p == '/' && *(p + 1) == '*') {
	p += 2;
	while (*p != '\0') {
	  if (*p == '*' && *(p + 1) == '/')
		break;

	  if (*p == '\n')
		curr_linenum++;
	  p++;
	}

	if (*p == '\0') {
	  SET_ERROR_MESG(jp, JSPR_ERROR_SYNTAX, NULL, 0, "Unterminated comment");
	  return(JSPR_ERROR_SYNTAX);
	}
	*inpp = (p + 2);
  }

  return(JSPR_OK);
}

/*
 * S2:
 * Insignificant whitespace is allowed before or after any of the six
 * structural characters.
 *
 * <ws>         = (' ' | '\t' | '\n' | '\r')*
 */
static void
skip_ws(const char **inpp)
{
  const char *p;

  p = *inpp;
  while (*p == ' ' || *p == '\t' || *p == '\n' || *p == '\r') {
	if (*p == '\n')
	  curr_linenum++;
	p++;
  }

  *inpp = p;
}

/*
 * S2, optionally with one or more comment styles recognized.
 * Insignificant whitespace and comments, if enabled, are allowed before or
 * after any of the six structural characters.
 *
 * An error is returned if an unterminated comment is seen.
 * Nested comments are not reported.
 */
static Jspr_rc
skip_ws_and_comments(Jspr_parser *jp, const char **inpp)
{

  while (1) {
	const char *p;

	p = *inpp;
	if (*p == ' ' || *p == '\t' || *p == '\n' || *p == '\r') {
	  skip_ws(inpp);
	  continue;
	}

	if (JSPR_COMMENTS_ENABLED(jp, JSPR_COMMENTS_CPP_TYPE)) {
	  if (*p == '/' && *(p + 1) == '/') {
		skip_cpp_comment(inpp);
		continue;
	  }
	}

	if (JSPR_COMMENTS_ENABLED(jp, JSPR_COMMENTS_C_TYPE)) {
	  Jspr_rc rc;

	  if (*p == '/' && *(p + 1) == '*') {
		if ((rc = skip_c_comment(jp, inpp)) != JSPR_OK)
		  return(rc);
		continue;
	  }
	}

	if (JSPR_COMMENTS_ENABLED(jp, JSPR_COMMENTS_SHELL_TYPE)) {
	  if (*p == '#') {
		skip_shell_comment(inpp);
		continue;
	  }
	}

	break;
  }

  return(JSPR_OK);
}

/*
 * There are three literal types:
 * o two booleans: 'false', 'true'
 * and:
 * o 'null'
 * (they are not quoted).
 *
 * <literal-name> = 'true' | 'false' | 'null'
 */
static Jspr_value *
match_literal(Jspr_parser *jp, Jspr_index parent, const char **inpp)
{
  const char *s;
  Jspr_value *v;

  /*
   * Do not accept trash that initially matches a recognized literal name.
   */
  s = *inpp;
  if (*s == 't' && *(s + 1) == 'r' && *(s + 2) == 'u'
	  && *(s + 3) == 'e' && !isalpha((int) *(s + 4))) {
	if ((v = value_alloc(jp, parent, JSPR_BOOLEAN_TYPE, s, s + 3)) == NULL)
	  return(NULL);
	v->flags = JSPR_VALUE_BOOLEAN_TRUE;
  }
  else if (*s == 'f' && *(s + 1) == 'a' && *(s + 2) == 'l'
		   && *(s + 3) == 's' && *(s + 4) == 'e'
		   && !isalpha((int) *(s + 5))) {
	if ((v = value_alloc(jp, parent, JSPR_BOOLEAN_TYPE, s, s + 4)) == NULL)
	  return(NULL);
	v->flags = JSPR_VALUE_BOOLEAN_FALSE;
  }
  else if (*s == 'n' && *(s + 1) == 'u' && *(s + 2) == 'l'
		   && *(s + 3) == 'l' && !isalpha((int) *(s + 4))) {
	if ((v = value_alloc(jp, parent, JSPR_NULL_TYPE, s, s + 3)) == NULL)
	  return(NULL);
  }
  else
	v = NULL;

  if (v != NULL)
	*inpp = js_end(jp, v) + 1;

  return(v);
}

/*
 * <string>     = '"' <chars> '"'
 * <chars>      = <empty> | <char> <chars>
 */
static Jspr_value *
match_string(Jspr_parser *jp, Jspr_index parent, const char **inpp)
{
  int saw_badchar, max_char_len;
  const char *p, *s;
  Jspr_value *v;

  saw_badchar = 0;
  max_char_len = 0;

  s = p = *inpp;

  if (*p++ != '"') {
	/* Not a string. */
	return(NULL);
  }

  while (1) {
	int char_len, val;

	if (*p == '"')
	  break;

	if ((char_len = char_value(jp, p, &val)) == -1) {
	  SET_ERROR_MESG(jp, JSPR_ERROR_STRSYNTAX, p, 1, NULL);
	  return(NULL);
	}

	if (char_len > max_char_len)
	  max_char_len = char_len;
	if (char_len == 1) {
	  if (!isprint(val))
		saw_badchar++;
	}

	p += char_len;
  }

  /* The string's value is between the quotation marks. */
  if ((v = value_alloc(jp, parent, JSPR_STRING_TYPE, s + 1, p)) == NULL)
	return(NULL);

  if (max_char_len == 1) {
	if (saw_badchar)
	  v->flags = JSPR_VALUE_STR_UTF8_1;
	else
	  v->flags = JSPR_VALUE_STR_CLEAN;
  }
  else if (max_char_len == 2)
	v->flags = JSPR_VALUE_STR_UTF8_2;
  else if (max_char_len == 3)
	v->flags = JSPR_VALUE_STR_UTF8_3;
  else if (max_char_len == 4)
	v->flags = JSPR_VALUE_STR_UTF8_4;
  else
	v->flags = JSPR_VALUE_STR_INVALID;

  *inpp = p + 1;

  return(v);
}

/*
 * <number>     = <opt-msign> <int> <frac-part> <exp-part>
 * <opt-msign>  = '-' | <empty>
 * <int>        = <dec-digit> | <nz-digit> <dec-digits>
 * <nz-digit>   = '1..9'
 * <frac-part>  = <empty> | '.' <dec-digits>
 * <dec-digits> = <dec-digit> | <dec-digit> <dec_digits>
 * <dec-digit>  = '0..9'
 * <exp-part>   = <empty> | ( 'e' | 'E' ) <opt-sign> <dec-digits>
 * <opt-sign>   = '+' | '-' | <empty>
 */
static Jspr_value *
match_number(Jspr_parser *jp, Jspr_index parent, const char **inpp)
{
  int saw_exp, saw_minus, saw_real;
  const char *p, *s;
  Jspr_value *v;

  saw_exp = saw_minus = saw_real = 0;

  s = p = *inpp;

  /*
   * <opt-msign>
   * A '+' is invalid - if it's followed by a digit we will assume
   * nothing else can match and parsing can stop.
   */
  if (*p == '+' && isdigit((int) *(p + 1))) {
	SET_ERROR_MESG(jp, JSPR_ERROR_NUMSYNTAX, p, 1, NULL);
	return(NULL);
  }
  if (*p == '-') {
	saw_minus++;
	p++;
  }

  /* Not a number. */
  if (!isdigit((int) *p)) {
	/* A digit must be first or follow a minus sign. */
	return(NULL);
  }

  /* <int>, with leading zeroes disallowed. */
  if (*p == '0' && isdigit((int) *(p + 1))) {
	SET_ERROR_MESG(jp, JSPR_ERROR_NUMSYNTAX, p, 1, NULL);
	return(NULL);
  }

  while (isdigit((int) *p))
	p++;

  /* <frac-part> */
  if (*p == '.') {
	const char *dot;

	dot = p++;
	while (isdigit((int) *p))
	  p++;
	if (dot == (p - 1)) {
	  int is_fatal;

	  is_fatal = (*p != '\0');
	  SET_ERROR_MESG(jp, JSPR_ERROR_NUMSYNTAX, p, is_fatal,
					 "Digit must follow '.'");
	  return(NULL);
	}
	saw_real++;
  }

  /* <exp-part> */
  if (*p == 'e' || *p == 'E') {
	const char *expect_digit;

	p++;
	if (*p == '+' || *p == '-')
	  p++;

	saw_exp++;
	saw_real++;
	expect_digit = p;
	while (isdigit((int) *p))
	  p++;
	if (p == expect_digit) {
	  int is_fatal;

	  is_fatal = (*p != '\0');
	  /* The 'e', 'E', '+', or '-') was not followed by a digit. */
	  SET_ERROR_MESG(jp, JSPR_ERROR_NUMSYNTAX, p, is_fatal,
					 "Exponent needs digit(s)");
	  return(NULL);
	}
  }

  if ((v = value_alloc(jp, parent, JSPR_NUMBER_TYPE, s, p)) == NULL)
	return(NULL);

  if (saw_exp)
	v->flags = JSPR_VALUE_NUMBER_EXP;
  else if (saw_real)
	v->flags = JSPR_VALUE_NUMBER_REAL;
  else if (saw_minus)
	v->flags = JSPR_VALUE_NUMBER_NEG;

  *inpp = p;

  return(v);
}

/*
 * An unordered set of zero or more name/value pairs.
 *
 * <object>     = <begin-obj> <end-obj>
 *                | <begin-obj> <members> <end-obj>
 * <members>    = <string> <name-sep> <value>
 *                | <string> <name-sep> <value> <value-sep> <members>
 */
static Jspr_value *
match_object(Jspr_parser *jp, Jspr_index parent, const char **inpp)
{
  unsigned int nsibs;
  const char *s;
  Jspr_index name_ind, v_curr, v_first, v_ind, v_obj_ind;
  Jspr_value *v, *v_end, *v_obj;

  if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	return(NULL);

  s = *inpp;
  if (!match_unesc_char(inpp, BEGIN_OBJECT_CHAR))
	return(NULL);

  if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	return(NULL);

  /* Match the empty Object. */
  if (match_unesc_char(inpp, END_OBJECT_CHAR)) {
	if ((v = value_alloc(jp, parent, JSPR_OBJECT_TYPE, s, (*inpp) - 1)) == NULL)
	  return(NULL);
	return(v);
  }

  if ((v_obj = value_alloc(jp, parent, JSPR_OBJECT_TYPE,
						   s, (*inpp) - 1)) == NULL)
	return(NULL);
  jp->object_level++;
  v_obj_ind = value_index(jp, v_obj);

  nsibs = 0;
  v_curr = v_first = 0;
  while (1) {
	Jspr_value *name, *v;

	if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	  return(NULL);

	/*
	 * A <member> is expected:
	 *   <string> <name-separator> <value>
	 * where <string> is the name of the member.
	 *
	 * First, expect the name part of the Object member.
	 * The member name is a string like any other but it is convenient
	 * to assign it a special type.
	 */
	if ((name = match_string(jp, v_obj_ind, inpp)) == NULL) {
	  SET_ERROR(jp, JSPR_ERROR_NAMESYNTAX, *inpp, -1);
	  jp->object_level--;
	  return(NULL);
	}

	/*
	 * This is a member name type rather than the more general string type.
	 */
	name->type = JSPR_NAME_TYPE;
	name_ind = value_index(jp, name);

	/*
	 * RFC 8259 says member names should be unique within an Object.
	 * Enforcements depends on configuration.
	 */
	if (JSPR_CONFIG(jp)->require_unique_obj_names && v_first != 0) {
	  char *dupname;
	  Jspr_index ind;
	  Jspr_value *vn;

	  for (ind = v_first; ind != 0; ind = vn->sib_next) {
		vn = value_ptr(jp, ind);
		if (vn->name != 0
			&& jspr_string_cmp(jp, name, value_ptr(jp, vn->name)) == 0) {
		  if ((dupname = jspr_name_of_value(jp, vn)) == NULL)
			dupname = "???";
		  SET_ERROR_MESG(jp, JSPR_ERROR_DUP_OBJNAME, *inpp, 1, dupname);
		  jp->object_level--;
		  return(NULL);
		}
	  }
	}

	if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	  return(NULL);

	if (!match_unesc_char(inpp, NAME_SEP_CHAR)) {
	  int is_fatal;

	  is_fatal = (**inpp != '\0');
	  SET_ERROR_MESG(jp, JSPR_ERROR_OBJSYNTAX, *inpp, is_fatal,
					 "Invalid Object name separator");
	  jp->object_level--;
	  return(NULL);
	}

	if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	  return(NULL);

	/*
	 * Expect the value part of an Object member.
	 */
	if ((v = match_value(jp, v_obj_ind, inpp)) == NULL) {
	  SET_ERROR(jp, JSPR_ERROR_OBJSYNTAX, *inpp, -1);
	  jp->object_level--;
	  return(NULL);
	}
	v_ind = value_index(jp, v);

	if (v_first == 0)
	  v_curr = v_first = v_ind;
	else {
	  jp->values[v_curr].sib_next = v_ind;
#ifdef SIB_PREV
	  v->sib_prev = v_curr;
#endif
	  v_curr = v_ind;
	}
	v->parent = v_obj_ind;
	/*
	 * The member value points to its name value via its name field
	 * and the member name points to its corresponding value via its parent
	 * field.
	 * There is no other link to the name, so no child_first, child_last,
	 * sib_next, or sib_prev point to it.
	 * Only the member values are children of the containing Object.
	 */
	v->name = name_ind;
	name = value_ptr(jp, name_ind);
	name->parent = v_ind;
	nsibs++;

	if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	  return(NULL);

	/*
	 * The purpose of this callback is to give the user an opportunity
	 * to process a new Object member as soon as it is seen, perhaps to
	 * add it to a symbol table, or to stop parsing immediately.
	 * Because this is a pre-order traversal, the member's value is available
	 * at this point but not any siblings that follow, and its parent Object
	 * has not been completed.
	 *
	 * If the callback returns -1, parsing terminates at this point.
	 *
	 * XXX might want to pass additional context and/or add it to Jspr_value
	 */
	if (JSPR_CONFIG(jp)->user_new_member_cb != NULL) {
	  int rc;
	  void *cb_arg;

	  cb_arg = JSPR_CONFIG(jp)->user_new_member_cb_arg;
	  rc = JSPR_CONFIG(jp)->user_new_member_cb(jp, name, v, cb_arg);
	  if (rc == -1) {
		/* Tidy the parent Object and note that parsing is complete. */
		JSPR_WAS_TERMINATED(jp) = 1;
		jp->object_level--;
		break;
	  }
	}

	/* Either the end of the Object or a value separator is expected. */
	if (match_unesc_char(inpp, END_OBJECT_CHAR))
	  break;

	if (!match_unesc_char(inpp, VALUE_SEP_CHAR)) {
	  int is_fatal;

	  is_fatal = (**inpp != '\0');
	  SET_ERROR_MESG(jp, JSPR_ERROR_OBJSYNTAX, *inpp, is_fatal,
					 "Invalid Object member separator");
	  jp->object_level--;
	  return(NULL);
	}
  }

  v_obj = value_ptr(jp, v_obj_ind);
  v_obj->child_first = v_first;
#ifdef CHILD_LAST
  v_obj->child_last = v_curr;
#endif
  /*
   * An element can determine how many siblings it has by inspecting
   * its parent's nchildren field.
   */
  v_obj->nchildren = nsibs;
  /* Patch end pointer */
#ifdef JSPR_COMPACT
  v_obj->js_len = ((*inpp) - 1) - jp->js - v_obj->js_start;
#else
  v_obj->end = (*inpp) - 1;
#endif

  jp->object_level--;
  v_end = value_alloc(jp, v_obj_ind, JSPR_END_OBJECT_TYPE, (*inpp) - 1, *inpp);
  if (v_end == NULL)
	return(NULL);

#ifdef NOTDEF
  if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	return(NULL);
#endif

  if (JSPR_WAS_TERMINATED(jp)) {
	SET_ERROR_MESG(jp, JSPR_ERROR_USER_TERM, *inpp, 1, "User terminated");
	return(NULL);
  }

  v_obj = value_ptr(jp, v_obj_ind);
  return(v_obj);
}

/*
 * An ordered sequence of zero or more values.
 *
 * <array>      = <begin-array> <end-array>
 *                | <begin-array> <values> <end-array>
 * <values>     = <value>
 *                | <value> <value-sep> <values>
 * <begin-array>, <end-array>, and <value-sep> may be optionally preceded
 * or followed by whitespace.
 * <value-sep> is a comma.
 */
static Jspr_value *
match_array(Jspr_parser *jp, Jspr_index parent, const char **inpp)
{
  unsigned int nsibs;
  const char *s;
  Jspr_index v_curr, v_first, v_ind, v_new_ind;
  Jspr_value *v, *v_end, *v_new;

  if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	return(NULL);

  s = *inpp;
  if (!match_unesc_char(inpp, BEGIN_ARRAY_CHAR))
	return(NULL);

  if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	return(NULL);

  /*
   * Three cases:
   *   1. empty []
   *   2. one value [<value>]
   *   3. two or more comma-separated value [<value>,<value>[,...]
   */

  if (match_unesc_char(inpp, END_ARRAY_CHAR)) {
	if ((v = value_alloc(jp, parent, JSPR_ARRAY_TYPE, s, (*inpp) - 1)) == NULL)
	  return(NULL);
	return(v);
  }

  if ((v_new = value_alloc(jp, parent, JSPR_ARRAY_TYPE,
						   s,  (*inpp) - 1)) == NULL)
	return(NULL);
  jp->array_level++;
  v_new_ind = value_index(jp, v_new);

  nsibs = 0;
  v_curr = v_first = 0;
  while (1) {
	if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	  return(NULL);

	if ((v = match_value(jp, v_new_ind, inpp)) == NULL) {
	  SET_ERROR(jp, JSPR_ERROR_ARRSYNTAX, *inpp, -1);
	  jp->array_level--;
	  return(NULL);
	}
	v_ind = value_index(jp, v);

	if (v_first == 0)
	  v_curr = v_first = v_ind;
	else {
	  jp->values[v_curr].sib_next = v_ind;
#ifdef SIB_PREV
	  v->sib_prev = v_curr;
#endif
	  v_curr = v_ind;
	}
	v->parent = v_new_ind;
	nsibs++;

	if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	  return(NULL);

	if (match_unesc_char(inpp, END_ARRAY_CHAR))
	  break;

	if (!match_unesc_char(inpp, VALUE_SEP_CHAR)) {
	  int is_fatal;

	  is_fatal = (**inpp != '\0');
	  SET_ERROR_MESG(jp, JSPR_ERROR_ARRSYNTAX, *inpp, is_fatal,
					 "Invalid Array value separator");
	  jp->array_level--;
	  return(NULL);
	}
  }

  v_new = value_ptr(jp, v_new_ind);
  v_new->child_first = v_first;
#ifdef CHILD_LAST
  v_new->child_last = v_curr;
#endif
  v_new->nchildren = nsibs;

  /* Patch end pointer */
#ifdef JSPR_COMPACT
  v_new->js_len = ((*inpp) - 1) - jp->js - v_new->js_start;
#else
  v_new->end = (*inpp) - 1;
#endif

  jp->array_level--;
  v_end = value_alloc(jp, v_new_ind, JSPR_END_ARRAY_TYPE, (*inpp) - 1, *inpp);
  if (v_end == NULL)
	return(NULL);

#ifdef NOTDEF
  if (skip_ws_and_comments(jp, inpp) != JSPR_OK)
	return(NULL);
#endif

  v_new = value_ptr(jp, v_new_ind);
  return(v_new);
}

/*
 * <value> = <literal-name> | <object> | <array> | <number> | <string>
 */
static Jspr_value *
match_value(Jspr_parser *jp, Jspr_index parent, const char **inpp)
{
  int is_fatal;
  Jspr_value *v;

  if ((v = match_literal(jp, parent, inpp)) != NULL || JSPR_WAS_TERMINATED(jp))
	return(v);

  if ((v = match_number(jp, parent, inpp)) != NULL || JSPR_WAS_TERMINATED(jp))
	return(v);

  if ((v = match_string(jp, parent, inpp)) != NULL || JSPR_WAS_TERMINATED(jp))
	return(v);

  if ((v = match_object(jp, parent, inpp)) != NULL || JSPR_WAS_TERMINATED(jp))
	return(v);

  if ((v = match_array(jp, parent, inpp)) != NULL || JSPR_WAS_TERMINATED(jp))
	return(v);

  is_fatal = (JSPR_ERR(jp)->is_fatal ? 1 : (**inpp != '\0'));
  SET_ERROR_MESG(jp, JSPR_ERROR_VALUE_EXPECTED, *inpp, is_fatal,
				 "A value is expected");
  return(NULL);
}

/*
 * Return the error message string corresponding to RC.
 */
char *
jspr_strerror(Jspr_rc rc)
{
  static char errbuf[JSPR_STRERROR_BUFSIZ + 1];

  switch (rc) {
  case JSPR_ERROR_NOMEM:
	snprintf(errbuf, sizeof(errbuf),
			 "Cannot allocate memory [JSPR_ERROR_NOMEM=%d]",
			 JSPR_ERROR_NOMEM);
	break;

  case JSPR_ERROR_INVALID_CHAR:
	snprintf(errbuf, sizeof(errbuf),
			 "Invalid character in JSON string [JSPR_ERROR_INVALID_CHAR=%d]",
			 JSPR_ERROR_INVALID_CHAR);
	break;

  case JSPR_ERROR_PART:
	snprintf(errbuf, sizeof(errbuf),
			 "Incomplete JSON structure [JSPR_ERROR_PART=%d]",
			 JSPR_ERROR_PART);
	break;

  case JSPR_ERROR_SYNTAX:
	snprintf(errbuf, sizeof(errbuf),
			 "Unexpected character or value type [JSPR_ERROR_SYNTAX=%d]",
			 JSPR_ERROR_SYNTAX);
	break;

  case JSPR_ERROR_NUMSYNTAX:
	snprintf(errbuf, sizeof(errbuf),
			 "Invalid number syntax [JSPR_ERROR_NUMSYNTAX=%d]",
			 JSPR_ERROR_NUMSYNTAX);
	break;

  case JSPR_ERROR_STRSYNTAX:
	snprintf(errbuf, sizeof(errbuf),
			 "Invalid string syntax [JSPR_ERROR_STRSYNTAX=%d]",
			 JSPR_ERROR_STRSYNTAX);
	break;

  case JSPR_ERROR_OBJSYNTAX:
	snprintf(errbuf, sizeof(errbuf),
			 "Invalid Object syntax [JSPR_ERROR_OBJSYNTAX=%d]",
			 JSPR_ERROR_OBJSYNTAX);
	break;

  case JSPR_ERROR_ARRSYNTAX:
	snprintf(errbuf, sizeof(errbuf),
			 "Invalid Array syntax [JSPR_ERROR_ARRSYNTAX=%d]",
			 JSPR_ERROR_ARRSYNTAX);
	break;

  case JSPR_ERROR_NAMESYNTAX:
	snprintf(errbuf, sizeof(errbuf),
			 "Object member name is missing or invalid [JSPR_ERROR_NAMESYNTAX=%d]",
			 JSPR_ERROR_NAMESYNTAX);
	break;

  case JSPR_ERROR_CONFIG:
	snprintf(errbuf, sizeof(errbuf),
			 "Invalid configuration [JSPR_ERROR_CONFIG=%d]",
			 JSPR_ERROR_CONFIG);
	break;

  case JSPR_ERROR_VALUE_LIMIT:
	snprintf(errbuf, sizeof(errbuf),
			 "Maximum number of values reached [JSPR_ERROR_VALUE_LIMIT=%d]",
			 JSPR_ERROR_VALUE_LIMIT);
	break;

  case JSPR_ERROR_DUP_OBJNAME:
	snprintf(errbuf, sizeof(errbuf),
			 "Object has a duplicate member name [JSPR_ERROR_DUP_OBJNAME=%d]",
			 JSPR_ERROR_DUP_OBJNAME);
	break;

  case JSPR_ERROR_USER_TERM:
	snprintf(errbuf, sizeof(errbuf),
			 "User initiated termination [JSPR_ERROR_USER_TERM=%d]",
			 JSPR_ERROR_USER_TERM);
	break;

  case JSPR_ERROR_VALUE_EXPECTED:
	snprintf(errbuf, sizeof(errbuf),
			 "A value was expected [JSPR_ERROR_VALUE_EXPECTED=%d]",
			 JSPR_ERROR_VALUE_EXPECTED);
	break;

  case JSPR_ERROR_INPUT_OVERFLOW:
	snprintf(errbuf, sizeof(errbuf),
			 "JSON input overflow - reconfigure with bigger buffers [JSPR_ERROR_INPUT_OVERFLOW=%d]",
			 JSPR_ERROR_INPUT_OVERFLOW);
	break;

  case JSPR_ERROR_UNKNOWN:
	snprintf(errbuf, sizeof(errbuf),
			 "Unknown error - bug! [JSPR_ERROR_UNKNOWN=%d]",
			 JSPR_ERROR_UNKNOWN);
	break;

  case JSPR_OK:
	snprintf(errbuf, sizeof(errbuf), "No error [JSPR_OK=%d]",
			 JSPR_OK);
	break;

  default:
	snprintf(errbuf, sizeof(errbuf), "Unknown error [%d]", rc);
	break;
  }

  return(errbuf);
}

/*
 * XXX
 * See: RFC 6901 and RFC 9535
 *
 * Search for a value by navigating through the parse tree in JP
 * according to PATH.
 * Examples:
 * "/foo/bar/baz"
 * "foo/bar/baz"
 * "foo/bar[1]"
 * "a/b/[]"
 *
 * Path grammar:
 * <path>            = '/' <path-list>
 *                     | <path-list>
 * <path-list>       = <path-component>
 *                     | <path-component> '/' <path-list>
 * <path-component>  = <object-name> <index> | <object-name> | <index> | '*'
 * <object-name>     = <string>
 * <index>           = '[' <decimal-digits> ']' | '[' ']'
 * <unquoted-string> = <char> | <char> <unquoted-string>
 */

/*
 * Return a copy of the Object member name associated with VALUE from JP,
 * or NULL. VALUE must be an Object member value.
 * The caller should free the returned string.
 */
char *
jspr_name_of_value(Jspr_parser *jp, Jspr_value *value)
{
  Jspr_value *v;

  if (value == NULL || value->name == 0)
    return(NULL);

  v = value_ptr(jp, value->name);
  if (v->type != JSPR_NAME_TYPE)
    return(NULL);

  return(strndup(js_start(jp, v), js_len(jp, v)));
}

/*
 * Copy the default configuration into CONFIG and return it.
 * Since a copy is provided, the caller may customize it.
 */
static Jspr_config *
get_default_config(Jspr_config *config)
{
  static Jspr_config *conf = NULL;

  if (config == NULL)
	return(NULL);

  if (conf == NULL) {
	static Jspr_config default_config;

	default_config.enable_expand_values = JSPR_ENABLE_EXPAND_VALUES_DEFAULT;
	default_config.enable_colourize = JSPR_ENABLE_COLOURIZE_DEFAULT;
	default_config.enable_comments = JSPR_ENABLE_COMMENTS_DEFAULT;
	default_config.require_obj_or_arr = JSPR_REQUIRE_OBJ_OR_ARR_DEFAULT;
	default_config.require_unique_obj_names
	  = JSPR_REQUIRE_UNIQUE_OBJ_NAMES_DEFAULT;
	default_config.enable_no_values = 0;
	default_config.values_limit = JSPR_VALUES_LIMIT_DEFAULT;
	default_config.num_values = JSPR_NUM_VALUES_DEFAULT;

	default_config.enable_debug = 0;
	default_config.enable_verbose = 0;

	default_config.user_malloc_func = NULL;
	default_config.user_free_func = NULL;
	default_config.user_expand_values_cb = NULL;
	default_config.user_new_member_cb = NULL;
	default_config.user_new_member_cb_arg = NULL;

	conf = &default_config;
  }

  *config = *conf;

  return(conf);
}

/*
 * Return parser error information for JP.
 */
Jspr_err_state *
jspr_err_state(Jspr_parser *jp, Jspr_err_state *err_state)
{

  *err_state = jp->err_state;

  return(err_state);
}

/*
 * Copy the resource usage data for parser JP into RI and return its address.
 * The resource data is reset by jspr_init(), jspr_init_x(), and jspr_reset().
 */
Jspr_res_info *
jspr_res_info(Jspr_parser *jp, Jspr_res_info *ri)
{

  if (ri == NULL)
	ri = (Jspr_res_info *) malloc(sizeof(Jspr_res_info));

  *ri = jp->res_info;

  return(ri);
}

/*
 * If JP is non-NULL, return its current configuration and copy it into
 * CONFIG.
 * If JP is NULL, return the default configuration and copy it into CONFIG.
 */
Jspr_config *
jspr_get_config(Jspr_parser *jp, Jspr_config *config)
{

  if (config == NULL)
	return(NULL);

  if (jp == NULL)
	get_default_config(config);
  else
	*config = jp->config;

  return(config);
}

/*
 * Return a pointer to the first value in JP's vector of values.
 * If non-NULL, also set any of VALUES, USED (the number of values in
 * the vector currently being used), and AVAIL (the total number of values
 * allocated for the vector - which may increase if expansion is allowed).
 */
Jspr_value *
jspr_get_values(Jspr_parser *jp, Jspr_value **values, Jspr_index *used,
			Jspr_index *avail)
{

  if (values != NULL)
	*values = jp->values;

  if (used != NULL)
	*used = jp->next_value;

  if (avail != NULL)
	*avail = jp->config.num_values;

  return(jp->values);
}

/*
 * Create a new parser initialized to parse JS, a null-terminated JSON input
 * string of length JS_LEN bytes, using configuration CONFIG.
 * If OJP is NULL, allocate a new data structure, otherwise reuse OJP.
 * If CONFIG is NULL, use defaults.
 * JS is not copied and must continue to exist after parsing if anything
 * beyond validation is needed.
 *
 * Return the parser, which will be deallocated by jspr_end().
 *
 * Note:
 * A NUL (zero) byte is not valid within JS (it must be escaped)
 * The length (JS_LEN) of the input string is passed to detect this error and
 * to help discover application and parsing bugs and/or malicious input,
 * since a valid parse must terminate at the true end of the input buffer.
 *
 * CONFIG specifies how to obtain the memory required for values produced
 * when JS is parsed:
 *  o NUM_VALUES is the number of values that will be (or are) allocated
 *    and the maximum number of values available to the parser; if zero,
 *    a compile time default is used (except this is an error if
 *    USER_VALUES_PTR is non-NULL)
 *
 * o VALUES_LIMIT specifies the maximum number of values in the pool;
 *   if negative, no limit is enforced;
 *   if the pool cannot grow, the limit is set to NUM_VALUES.
 *
 * o ENABLE_EXPAND_VALUES specifies the behaviour for growing the value
 *   pool; if it is negative, growing is disabled; if it is zero, use the
 *   compile time default; if it is positive, growing is enabled.
 *   Only a dynamically allocated pool can grow, however.
 *
 * o ENABLE_NO_VALUES: if non-zero, the parser will validate its input but
 *   will not allocate or emit any value tokens (ignoring VALUES_LIMIT) but
 *   num_values_used will be set.
 *   This mode can be used to quickly validate the syntax of an input string.
 *   It determines how many value tokens would be needed for the
 *   input string's parse tree.
 *   No parse tree is produced so obviously the caller may not inspect
 *   or print values (e.g., via callbacks or searching) after this parse
 *   completes.
 *
 * o REQUIRE_OBJ_OR_ARR: if non-zero, the top-level JSON string
 *   must be an Object or Array.  RFC 8259 does not require this constraint
 *   but some applications might.
 *
 * o REQUIRE_UNIQUE_OBJ_NAMES: if non-zero, enforces the constraint that
 *   member names within an Object must be unique. This is not applied
 *   recursively.  RFC 8259 does not have this requirement.
 *
 * o USER_VALUES_PTR, if non-NULL, points to memory that has already been
 *   allocated to hold NUM_VALUES; the caller is responsible for deallocating
 *   this memory, if that is necessary; this might point to a static buffer,
 *   for instance;
 *   if USER_VALUES_PTR is NULL, USER_MALLOC_FUNC must be provided and will
 *   be called to allocate NUM_VALUES; if USER_FREE_FUNC is provided, it will
 *   be called by jspr_end() to deallocate the memory.
 *
 * Refer to the definition of Jspr_config in jspr.h
 */
Jspr_parser *
jspr_init_x(Jspr_parser *ojp, const char *js, size_t js_len, Jspr_config *config)
{
  int enable_expand_values, values_limit;
  unsigned int num_values;
  Jspr_config *conf, default_config;
  Jspr_parser *jp;

  if ((conf = config) == NULL)
	conf = get_default_config(&default_config);

  if ((jp = ojp) == NULL) {
	if ((jp = (Jspr_parser *) malloc(sizeof(Jspr_parser))) == NULL) {
	  SET_ERROR(jp, JSPR_ERROR_NOMEM, NULL, 1);
	  return(NULL);
	}
  }

  jp->js = js;
  jp->js_len = js_len;

  jp->array_level = jp->object_level = 0;
  jp->array_count = jp->object_count = 0;

  num_values = conf->num_values;
  values_limit = conf->values_limit;
  if (conf->enable_no_values) {
	static Jspr_value dummy_values[2];

	jp->values = dummy_values;
	enable_expand_values = 0;
	values_limit = -1;
	num_values = 1;
  }
  else if (conf->user_values_ptr == NULL) {
	size_t need;

	/*
	 * Values are being allocated dynamically.
	 * If no initial number was specified, use the default.
	 * If enable_expand_values is negative, never expand the pool;
	 * if it is 0, use the default; and if it is anything else, enable
	 * expansion.
	 */
	if (num_values == 0)
	  num_values = JSPR_NUM_VALUES_DEFAULT;

	if (conf->enable_expand_values < 0)
	  enable_expand_values = 0;
	else if (conf->enable_expand_values == 0)
	  enable_expand_values = JSPR_ENABLE_EXPAND_VALUES_DEFAULT;
	else
	  enable_expand_values = 1;

	/*
	 * If VALUES_LIMIT is negative, there is no predetermined maximum number
	 * of allocated values.
	 * If the pool cannot be grown, the values limit is effectively NUM_VALUES.
	 * The values limit can be less than NUM_VALUES.
	 */
	if (values_limit < 0)
	  values_limit = -1;
	else if (!enable_expand_values && values_limit > num_values)
	  values_limit = num_values;

	need = num_values * sizeof(Jspr_value);
	if (conf->user_malloc_func != NULL)
	  jp->values = (Jspr_value *) (conf->user_malloc_func)(need);
	else
	  jp->values = (Jspr_value *) malloc(need);
	if (jp->values == NULL) {
	  SET_ERROR(jp, JSPR_ERROR_NOMEM, NULL, 1);
	  free(jp);
	  return(NULL);
	}
	if (JSPR_CONFIG(jp)->enable_debug)
	  fprintf(stderr, "Init: %u values\n", num_values);
  }
  else {
	/*
	 * Values are being allocated from a fixed size, user-supplied buffer.
	 * The number of values must be non-zero and pool expansion is not
	 * possible.
	 */
	if (num_values == 0) {
	  SET_ERROR_MESG(jp, JSPR_ERROR_SYNTAX, NULL, 1, "No values provided");
	  free(jp);
	  return(NULL);
	}
	values_limit = num_values;
	enable_expand_values = 0;
	jp->values = (Jspr_value *) conf->user_values_ptr;
  }

  JSPR_CONFIG(jp)->num_values = num_values;
  JSPR_CONFIG(jp)->enable_expand_values = enable_expand_values;
  JSPR_CONFIG(jp)->values_limit = values_limit;

  JSPR_CONFIG(jp)->enable_colourize = conf->enable_colourize;
  JSPR_CONFIG(jp)->enable_comments = conf->enable_comments;
  JSPR_CONFIG(jp)->enable_no_values = conf->enable_no_values;
  JSPR_CONFIG(jp)->enable_debug = conf->enable_debug;
  JSPR_CONFIG(jp)->enable_verbose = conf->enable_verbose;

  JSPR_CONFIG(jp)->require_obj_or_arr = conf->require_obj_or_arr;
  JSPR_CONFIG(jp)->require_unique_obj_names = conf->require_unique_obj_names;
  JSPR_CONFIG(jp)->user_malloc_func = conf->user_malloc_func;
  JSPR_CONFIG(jp)->user_free_func = conf->user_free_func;
  JSPR_CONFIG(jp)->user_values_ptr = conf->user_values_ptr;
  JSPR_CONFIG(jp)->user_expand_values_cb = conf->user_expand_values_cb;
  JSPR_CONFIG(jp)->user_new_member_cb = conf->user_new_member_cb;
  JSPR_CONFIG(jp)->user_new_member_cb_arg = conf->user_new_member_cb_arg;

  jp->next_value = JSPR_VALUE_ROOT_INDEX;	/* Don't use index 0. */

  JSPR_ERR(jp)->err_pos = 0;
  JSPR_ERR(jp)->err_line = 0;
  JSPR_ERR(jp)->err_linepos = 0;
  JSPR_ERR(jp)->err_mesg = NULL;
  JSPR_ERR(jp)->err_rc = JSPR_OK;
  JSPR_ERR(jp)->is_fatal = 0;
  JSPR_ERR(jp)->user_terminated = 0;

  JSPR_RES(jp)->num_values_used = 0;
  JSPR_RES(jp)->num_values = 0;
  JSPR_RES(jp)->nexpands = 0;
  JSPR_RES(jp)->nstrings = 0;
  JSPR_RES(jp)->string_alloc = 0;
  JSPR_RES(jp)->xbuf_maxlen = 0;
  JSPR_RES(jp)->xbuf_count = 0;

  curr_linenum = 1;
  curr_linepos = 0;

  return(jp);
}

/*
 * Create a new parser initialized to parse JS, a null-terminated JSON input
 * string of length JS_LEN, using the default configuration.
 * Use the compile-time default configuration.
 * The default is to call malloc() and free(3) to allocate and
 * deallocate a value.
 *
 * JS is not copied and must continue to exist after parsing if anything
 * beyond validation is needed.
 * NUM_VALUES will be the maximum number of values available to the parser;
 * if zero, a typical default will be used, which may not be adequate.
 *
 * See jspr_init_x().
 */
Jspr_parser *
jspr_init(const char *js, size_t js_len)
{
  Jspr_parser *jp;

  jp = jspr_init_x(NULL, js, js_len, NULL);
  if (jp == NULL)
	return(NULL);

  return(jp);
}

/*
 * Reset the context of JP so that it can be used to parse JSON
 * string JS of length JS_LEN bytes.
 * If JS is NULL, do not reset the string being parsed and ignore JS_LEN.
 * JS is not copied and must continue to exist after parsing if
 * anything beyond validation is needed.
 *
 * The parser's configuration is not changed, including the space
 * used for values and maximum number of available values, but its state
 * is reset, including instrumentation counters and error information.
 * Any previous parse information is lost and should not be referenced.
 * Return the parser, or NULL if an error occurred.
 */
Jspr_parser *
jspr_reset(Jspr_parser *jp, const char *js, size_t js_len)
{

  if (jp == NULL)
	return(NULL);

  if (js != NULL) {
	jp->js = js;
	jp->js_len = js_len;
  }
  else {
	if (jp->js == NULL || jp->js_len == 0) {
	  SET_ERROR_MESG(jp, JSPR_ERROR_CONFIG, NULL, 1, "No JSON input");
	  return(NULL);
	}
  }

  jp->array_count = jp->object_count = 0;
  jp->array_level = jp->object_level = 0;

  memset(jp->values, 0, sizeof(Jspr_value) * jp->next_value);
  jp->next_value = JSPR_VALUE_ROOT_INDEX;	/* Don't use index 0. */

  JSPR_ERR(jp)->err_pos = 0;
  JSPR_ERR(jp)->err_line = 0;
  JSPR_ERR(jp)->err_linepos = 0;
  JSPR_ERR(jp)->err_mesg = NULL;
  JSPR_ERR(jp)->err_rc = JSPR_OK;
  JSPR_ERR(jp)->is_fatal = 0;
  JSPR_ERR(jp)->user_terminated = 0;

  JSPR_RES(jp)->num_values_used = 0;
  JSPR_RES(jp)->num_values = 0;
  JSPR_RES(jp)->nexpands = 0;
  JSPR_RES(jp)->nstrings = 0;
  JSPR_RES(jp)->string_alloc = 0;
  JSPR_RES(jp)->xbuf_maxlen = 0;
  JSPR_RES(jp)->xbuf_count = 0;

  curr_linenum = 1;
  curr_linepos = 0;

  return(jp);
}

/*
 * Parse one (top-level) JSON value (a "JSON-text"), defined by the
 * JSON grammar as a value optionally surrounded by whitespace:
 *   <json-text>  = <ws> <value> <ws>
 *
 * If ENDP is non-NULL, then it is set to the first non-whitespace
 * character that follows the value, or the input's terminating null byte.
 * Any non-conformant characters may follow the value (including any
 * trailing whitespace).
 *
 * If ENDP is NULL, then the entire input string must be exactly
 * one (top-level) value.
 *
 * The parser state should have been initialized and may be altered.
 *
 * The REQUIRE_OBJ_OR_ARR configuration option is honoured.
 * RFC 8259 does not require the JSON text to be an Object or an Array,
 * but some earlier specifications and applications may have that constraint.
 *
 * Return 0 if a valid value was parsed, -1 otherwise.
 */
Jspr_rc
jspr_parse_value(Jspr_parser *jp, Jspr_value **v_result, const char **endp)
{
  const char *inp;
  Jspr_rc rc;
  Jspr_value *v;

  inp = jp->js;

  /* White space is allowed before the value. */
  if ((rc = skip_ws_and_comments(jp, &inp)) != JSPR_OK)
	return(rc);

  /* The top element has no parent. */
  if ((v = match_value(jp, 0, &inp)) == NULL) {
	if (JSPR_RC(jp) != JSPR_OK) {
	  SET_ERROR(jp, JSPR_RC(jp), inp, -1);
	  return(JSPR_RC(jp));
	}
	SET_ERROR(jp, JSPR_ERROR_UNKNOWN, inp, 1);
	return(JSPR_ERROR_UNKNOWN);
  }

  if (JSPR_CONFIG(jp)->require_obj_or_arr) {
    if (!value_is_struct(v)) {
      SET_ERROR_MESG(jp, JSPR_ERROR_PART, NULL, 1, "Invalid top-level value");
      return(JSPR_ERROR_PART);
    }
  }

  /* White space is allowed after the value. */
  if ((rc = skip_ws_and_comments(jp, &inp)) != JSPR_OK)
	return(rc);

  if (endp == NULL) {
	/* Check that parsing has reached the end of the input. */
	if (*inp != '\0' || (inp - jp->js) != jp->js_len) {
	  SET_ERROR(jp, JSPR_ERROR_SYNTAX, inp, 1);
	  return(JSPR_ERROR_SYNTAX);
	}
  }
  else
	*endp = inp;

  if (v_result != NULL)
	*v_result = v;

  return(JSPR_OK);
}

/*
 * Run the JSON parser on the configured JSON text input, generating an
 * array of values.  Returns a result code.
 *
 * Note that the parse is a left-to-right scan.
 * The values are produced in that order.
 *
 * The JSON grammar (RFC 8259)
 *
 * <json-text>  = <ws> <value> <ws>
 * <ws>         = (' ' | '\t' | '\n' | '\r')*
 * <value>      = <literal-name> | <object> | <array> | <number> | <string>
 * <literal-name> = 'true' | 'false' | 'null'
 * <values>     = <value>
 *                | <value> <value-sep> <values>
 * <object>     = <begin-obj> <end-obj>
 *                | <begin-obj> <members> <end-obj>
 * <members>    = <string> <name-sep> <value>
 *                | <string> <name-sep> <value> <value-sep> <members>
 * <array>      = <begin-array> <end-array>
 *                | <begin-array> <values> <end-array>
 *
 * <number>     = <opt-msign> <int> <frac-part> <exp-part>
 * <opt-msign>  = '-' | <empty>
 * <int>        = <dec-digit> | <nz-digit> <dec-digits>
 * <nz-digit>   = '1..9'
 * <frac-part>  = <empty> | '.' <dec-digits>
 * <dec-digits> = <dec-digit> | <dec-digit> <dec_digits>
 * <dec-digit>  = '0..9'
 * <exp-part>   = <empty> | ( 'e' | 'E' ) <opt-sign> <dec-digits>
 * <opt-sign>   = '+' | '-' | <empty>
 *
 * <string>     = '"' <chars> '"'
 * <chars>      = <empty> | <char> <chars>
 * <char>       = <unesc-char> | <esc-char>
 * ERRATA CHANGE: the '/' (0x2f) is in the unescaped range
 *   <unesc-char> = 0x20..0x21 | 0x23..0x2e | 0x30-0x5b | 0x5d..0x10ffff
 * ORIGINAL:
 *   <unesc-char> = 0x20..0x21 | 0x23..0x5b | 0x5d..0x10ffff
 * <esc-char>   = '\"' | '\\' | '\/' | '\b' | '\f' | '\n' | '\r' | '\t'
 *                | \u<xxxx>
 * <xxxx>       = <hexdigit> <hexdigit> <hexdigit> <hexdigit>
 *
 * <begin-obj>  = <ws> '{' <ws>
 * <end-obj>    = <ws> '}' <ws>
 * <begin-array>= <ws> '[' <ws>
 * <end-array>  = <ws> ']' <ws>
 * <name-sep>   = <ws> ':' <ws>
 * <value-sep>  = <ws> ',' <ws>
 */

/*
 * Simple top-level parser invocation, most useful when the caller is
 * only needs to check for valid syntax.
 * Post-parse details can still be accessed through JP.
 */
Jspr_rc
jspr_parse(Jspr_parser *jp)
{
  Jspr_rc rc;
  Jspr_value *v;

  rc = jspr_parse_value(jp, &v, NULL);
  if (rc == JSPR_OK && JSPR_CONFIG(jp)->require_obj_or_arr) {
	if (!value_is_struct(v)) {
	  SET_ERROR_MESG(jp, JSPR_ERROR_PART, NULL, 1, "Invalid top-level value");
	  return(JSPR_ERROR_PART);
	}
  }

  return(rc);
}

/*
 * Deallocate JP, including any memory for values allocated by jspr_init()
 * or jspr_init_x().
 * If memory for the values was provided by the user, the user is
 * responsible for any deallocation.
 */
void
jspr_end(Jspr_parser *jp)
{

  if (jp == NULL)
	return;

  if (jp->values != NULL) {
	memset(jp->values, 0, sizeof(Jspr_value) * JSPR_CONFIG(jp)->num_values);
	if (JSPR_CONFIG(jp)->user_values_ptr == NULL) {
	  if (JSPR_CONFIG(jp)->user_free_func != NULL)
		(JSPR_CONFIG(jp)->user_free_func)(jp->values);
	  else
		free(jp->values);
	}
  }

  memset(jp, 0, sizeof(Jspr_parser));
  free(jp);
}

/*
 * Return a snapshot of parser JP, which includes a copy of its input
 * string, all of its values, configuration, and other state information.
 * Parser JP continues to exist, is unmodified, can be navigated,
 * reinitialized, and so on.
 *
 * Also see jspr_get_values().
 */
Jspr_parser *
jspr_parse_copy(Jspr_parser *jp)
{
  size_t need;
  Jspr_parser *jp_copy;

  if (jp == NULL || jp->js == NULL)
	return(NULL);

  jp_copy = (Jspr_parser *) malloc(sizeof(Jspr_parser));
  jp_copy->js = strdup(jp->js);
  jp_copy->js_len = jp->js_len;

  need = sizeof(Jspr_value) * jp->config.num_values;
  jp_copy->values = (Jspr_value *) malloc(need);
  memcpy(jp_copy->values, jp->values, need);

  jp_copy->array_count = jp->array_count;
  jp_copy->array_level = jp->array_level;
  jp_copy->object_count = jp->object_count;
  jp_copy->object_level = jp->object_level;

  jp_copy->next_value = jp->next_value;
  jp_copy->config = jp->config;
  jp_copy->err_state = jp->err_state;
  jp_copy->res_info = jp->res_info;

  return(jp_copy);
}

#ifdef NOTDEF
/*
 * Create a serialized string representation of the state of
 * the parse tree in JP. If non-NULL, return it through SNAP and SNAP_LEN
 * (otherwise it will be discarded).
 * Return a hash of the string representation.
 *
 * This snapshot can be saved and/or compared to another snapshot
 * of the same parse tree for the same input string;
 * e.g., if parses of the same input string result in different snapshot
 * hash values, then something has changed and one of the parses may be
 * erroneous - can be useful for testing).
 */
char *
jspr_parse_snap(Jspr_parser *jp, char **snap, size_t *snap_len)
{

  /*
   * Parser state: js_len, next_value
   * Values: type, flags, nchildren, js_start, js_len, name, parent, child_first,
   * sib_next
   */

}
#endif
