
/*
 * JSPR - JSon ParseR
 */

/*
 * Copyright (c) 2017-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: jspr_priv.h 3360 2025-03-31 23:28:47Z brachman $
 */

#ifndef __JSPR_PRIV_H_
#define __JSPR_PRIV_H_

#include <sys/types.h>
#include <stdint.h>
#include <inttypes.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <stdarg.h>

/*
 * The default is not to define JSPR_COMPACT, which creates larger Jspr_value
 * structures.
 * If JSPR_COMPACT is defined use a more compact representation that may be
 * a little less CPU efficient.
 */
#ifdef JSPR_COMPACT
#undef CHILD_LAST
#undef SIB_PREV
#endif

#include "jspr.h"

#if defined(__linux__)
#include <netinet/in.h>
#endif

#define INLINE          inline
#define MAYBE_UNUSED    __attribute__ ((__unused__))
#define PACKED          __attribute__ ((__packed__))

/*
 * If an error has not already been set, note this innermost error.
 * IS_FATAL should be 0 if the error may be recoverable, 1 if it is
 * unrecoverable, or -1 if the current flag value should not be changed.
 */
#define SET_ERROR(JP, RC, POS, IS_FATAL)								\
  do {																	\
	if (JP && RC != JSPR_OK && JSPR_RC(JP) == JSPR_OK) {				\
	  if ((char *) POS != NULL)											\
		JSPR_ERR(JP)->err_pos = (char *) POS - JP->js;					\
	  else																\
		JSPR_ERR(JP)->err_pos = 0;										\
	  JSPR_ERR(JP)->err_linepos = curr_linepos;							\
	  JSPR_ERR(JP)->err_line = curr_linenum;							\
	  JSPR_ERR(JP)->err_rc = RC;										\
	  if (IS_FATAL != -1)												\
		JSPR_ERR(JP)->is_fatal = IS_FATAL;								\
	}																	\
  } while (0);

/* Like SET_ERROR() but with a custom message. */
#define SET_ERROR_MESG(JP, RC, POS, IS_FATAL, MESG)						\
  do {																	\
	if (JP && RC != JSPR_OK && JSPR_RC(JP) == JSPR_OK) {				\
	  if ((char *) POS != NULL)											\
		JSPR_ERR(JP)->err_pos = (char *) POS - JP->js;					\
	  else																\
		JSPR_ERR(JP)->err_pos = 0;										\
	  JSPR_ERR(JP)->err_linepos = curr_linepos;							\
	  JSPR_ERR(JP)->err_line = curr_linenum;							\
	  JSPR_ERR(JP)->err_mesg = MESG;									\
	  JSPR_ERR(JP)->err_rc = RC;										\
	  if (IS_FATAL != -1)												\
		JSPR_ERR(JP)->is_fatal = IS_FATAL;								\
	}																	\
  } while (0);


static INLINE MAYBE_UNUSED int
streq(const char *p, const char *q)
{

  return(strcmp(p, q) == 0);
}

static INLINE MAYBE_UNUSED int
strneq(const char *p, const char *q, size_t n)
{

  return(strncmp(p, q, n) == 0);
}

/*******/

/* Functions for accessing the JSON text being parsed. */

/*
 * Return a pointer to the start of the text of V in the input string.
 */
static MAYBE_UNUSED INLINE const char *
js_start(Jspr_parser *jp, Jspr_value *v)
{
  const char *ptr;

#ifdef JSPR_COMPACT
  ptr = jp->js + v->js_start;
#else
  ptr = v->start;
#endif

  return(ptr);
}

/*
 * Return the length in bytes of the text of V in the input string.
 */
static MAYBE_UNUSED INLINE size_t
js_len(Jspr_parser *jp, Jspr_value *v)
{
  size_t len;

#ifdef JSPR_COMPACT
  len = v->js_len;
#else
  len = v->end - v->start;
#endif

  return(len);
}

/*
 * Return a pointer to the byte after the end of the text of V in the
 * input string.
 */
static MAYBE_UNUSED INLINE const char *
js_end(Jspr_parser *jp, Jspr_value *v)
{
  const char *ptr;

#ifdef JSPR_COMPACT
  ptr = jp->js + v->js_start + v->js_len;
#else
  ptr = v->end;
#endif

  return(ptr);
}

/*
 * Return the offset in the input text to the byte after the end of V.
 */
static MAYBE_UNUSED INLINE unsigned long
js_start_offset(Jspr_parser *jp, Jspr_value *v)
{
  unsigned long off;

#ifdef JSPR_COMPACT
  off = v->js_start;
#else
  off = v->start - jp->js;
#endif

  return(off);
}

/*
 * Return the offset in the input text to the byte after the end of V.
 */
static MAYBE_UNUSED INLINE unsigned long
js_end_offset(Jspr_parser *jp, Jspr_value *v)
{
  unsigned long off;

#ifdef JSPR_COMPACT
  off = v->js_start + v->js_len;
#else
  off = v->end - jp->js;
#endif

  return(off);
}

/*
 * The end of an Array or Object is noted by a special marker value type.
 */
static INLINE int
value_is_end_marker(Jspr_value *v)
{

  return(v->type == JSPR_END_OBJECT_TYPE || v->type == JSPR_END_ARRAY_TYPE);
}

/*******/

/* Functions to convert between value offsets and value pointers. */

/* Return 1 if V is an Object or Array, 0 otherwise. */
static MAYBE_UNUSED INLINE int
value_is_struct(Jspr_value *v)
{

  return(v->type == JSPR_OBJECT_TYPE || v->type == JSPR_ARRAY_TYPE);
}

static MAYBE_UNUSED INLINE Jspr_index
value_index(Jspr_parser *jp, Jspr_value *v)
{

  if (JSPR_CONFIG(jp)->enable_no_values)
	return(1);

  if (v < &jp->values[0])
	return(0);

  return((Jspr_index) (v - jp->values));
}

static MAYBE_UNUSED INLINE Jspr_value *
value_ptr(Jspr_parser *jp, Jspr_index ind)
{

  if (JSPR_CONFIG(jp)->enable_no_values)
	return(&jp->values[1]);

  if (ind == 0 || ind >= jp->next_value)
	return(NULL);

  return(&jp->values[ind]);
}

static MAYBE_UNUSED INLINE Jspr_value *
value_parent(Jspr_parser *jp, Jspr_index ind)
{
  Jspr_value *v;

  if (JSPR_CONFIG(jp)->enable_no_values)
	return(&jp->values[1]);

  if (ind == 0)
	return(NULL);

  v = value_ptr(jp, ind);
  if (v->parent == 0)
	return(NULL);

  return(value_ptr(jp, v->parent));
}

extern unsigned int curr_linenum;
extern unsigned int curr_linepos;

#ifdef __cplusplus
extern "C" {
#endif

extern Jspr_index find_child_last(Jspr_parser *jp, Jspr_value *pv);
extern Jspr_index find_sib_prev(Jspr_parser *jp, Jspr_value *v);

extern int char_value(Jspr_parser *jp, const char *p, int *val);
extern int char_cmp(Jspr_parser *jp, const char **pp, const char **qq);

#ifdef __cplusplus
}
#endif

#endif /* __JSPR_H_ */
