
/*
 * This runs sets of parser tests, some written specifically for jspr
 * (mostly built in to this program) but most were collected from a variety
 * of free sources, including the 300+ tests in the JSON Parsing Test Suite
 * (as individual files).
 * https://github.com/nst/JSONTestSuite
 *
 * Also may include tests based on jsmn code written by Serge A. Zaitsev and
 * released by him under the MIT License.
 * Copyright (c) 2010 Serge A. Zaitsev
 * See:
 *  https://zserge.com/jsmn/
 *  https://github.com/zserge/jsmn
 *
 * Jspr should pass all tests.
 * Some tests could be essentially duplicates.
 */

/*
 * Copyright (c) 2017-2025
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 *
 * $Id: jspr_test.c 3398 2025-04-09 20:30:08Z brachman $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "jspr_priv.h"

#ifndef INLINE
#define INLINE          inline
#endif

#ifndef MAYBE_UNUSED
#define MAYBE_UNUSED    __attribute__ ((__unused__))
#endif


#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2017-2025\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: jspr_test.c 3398 2025-04-09 20:30:08Z brachman $";
#endif

static int tests_passed = 0;
static int tests_failed = 0;

/* Terminate current test with error. */
#define FAIL()  { tests_failed++; return(__LINE__); }

#define TEST(TNUM, CASE, LINE, WHICH, COND)								\
  do {																	\
	if (!(COND)) {														\
	  if (CASE >= 0)													\
		fprintf(stderr, "FAIL: Test %d[line %u], Case %d (%s test): ",	\
				TNUM, LINE, CASE, WHICH);								\
	  else																\
		fprintf(stderr, "FAIL: Test %d[line %u], %s (%s test): ",		\
				TNUM, LINE, jspr_strerror(CASE), WHICH);				\
	  return(__LINE__);													\
	}																	\
	else																\
	  tests_passed++;													\
  } while (0)

/* Check single condition */
#define CHECKRC(JP, RC)													\
  do {																	\
	if (RC != JSPR_OK) {												\
	  fprintf(stderr, "At line %d: %s", __LINE__, jspr_strerror(RC));	\
	  if (JP->err_state.err_mesg)										\
		fprintf(stderr, ":  \"%s\"\n", JP->err_state.err_mesg);			\
	  return(-1);														\
	}																	\
	else																\
	  tests_passed++;													\
  } while (0)

#define CHECK(COND)								\
  do {											\
	if (!(COND)) FAIL();						\
	tests_passed++;								\
  } while (0)

/* Test runner. */
static int
do_tests(const char *name, int tnum, int (*func)(void))
{
  int r;
  static int exit_on_error = 1;
  
  fprintf(stderr, "Test \"%s\" (Set %d): ", name, tnum);

  r = func();

  if (r == 0) {
	tests_passed++;
	fprintf(stderr, "OK\n");
	return(0);
  } else {
	tests_failed++;
	if (r > 0)
	  fprintf(stderr, "(at line %d)\n", r);
	if (exit_on_error)
	  exit(1);
	return(-1);
  }
}

#ifdef NOTDEF
static inline int
streq(const char *p, const char *q)
{

    return(strcmp(p, q) == 0);
}
#endif

static char *
strsuffix(char *p, size_t len, const char *suffix)
{
  char *q;
  size_t suffix_len;

  suffix_len = strlen(suffix);
  if (suffix_len > len)
    return(0);
  q = p + len - suffix_len;
  return(streq(q, suffix) ? q : NULL);
}

/******/

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <dirent.h>

/*
 * Read one test from FILENAME into BUF of size BUFSIZE bytes and
 * null terminate it.
 *
 * If HEADER_ENDP is non-NULL, the input may begin with a header of
 * zero or more lines. Each line must begin with the two characters "//".
 * HEADER_ENDP will be set to the start of the first line that follows the
 * last header line. This allows a version stamp or control information
 * to be embedded at the start of a test file.
 *
 * Return the total number of bytes read (which excludes the null byte),
 * or -1 if an error occurred.
 */
static ssize_t
load_suite_test(char *filename, char *buf, size_t bufsize, char **header_endp)
{
  int err, fd;
  ssize_t n;

  if ((fd = open(filename, O_RDONLY)) == -1) {
	fprintf(stderr, "Cannot read \"%s\"\n", filename);
	return(-1);
  }

  n = read(fd, buf, bufsize);
  err = errno;
  close(fd);

  /* A zero-length test is allowed. */
  if (n < 0) {
	fprintf(stderr, "Cannot read test: \"%s\" (%s)\n", filename, strerror(err));
	return(-1);
  }

  buf[n] = '\0';

  if (n == bufsize) {
	fprintf(stderr, "Test is too big: \"%s\"\n", filename);
	return(-1);
  }

  if (header_endp != NULL) {
	char *p;

	for (p = buf; *p != '\0'; p++) {
	  if (*p != '/' || *(p + 1) != '/')
		break;
	  p += 2;
	  while (*p != '\n' && *p != '\0')
		p++;
	}
	*header_endp = p;
  }

  return(n);
}

/*
 * Run the single test in FILENAME, for which the parser should
 * pass/fail/dontcare according to EXPECTED_RESULT:
 *  1 -> expect JSPR_OK parser result
 * -1 -> do not expect JSPR_OK parser result
 *  0 -> test must be run but do not care about the parser result
 *
 * Return 1 if the test produced the expected result, 0 if it did not,
 * or -1 if the test could not be run.
 */
static int
run_suite_test(char *filename, int expected_result)
{
  char *header_endp;
  ssize_t len, test_len;
  Jspr_rc r;
  static Jspr_parser *jp = NULL;
  /* Big enough to slurp in the biggest test case. */
  static char bigbuf[262144];

  len = load_suite_test(filename, bigbuf, sizeof(bigbuf), &header_endp);
  if (len == -1) {
	fprintf(stderr, "Cannot load test file: \"%s\"\n", filename);
	return(-1);
  }

  /* For now, just ignore any header. */
  test_len = len - (header_endp - bigbuf);
  if (jp == NULL)
	jp = jspr_init_x(NULL, header_endp, test_len, NULL);
  else
	jspr_reset(jp, header_endp, test_len);

  r = jspr_parse_value(jp, NULL, NULL);

  if ((r == JSPR_OK && expected_result == 1)
	  || (r != JSPR_OK && expected_result == -1)
	  || (expected_result == 0))
	return(1);

  fprintf(stdout, "ERROR [unexpected result: %s]\n", jspr_strerror(r));
  if (r != JSPR_OK) {
	Jspr_err_state err;

	jspr_err_state(jp, &err);
	fprintf(stdout, "At line %u: %s\n",
			err.err_line,
			err.err_mesg != NULL ? err.err_mesg : jspr_strerror(r));
  }

  return(0);
}

static MAYBE_UNUSED size_t
dirent_len(struct dirent *dp)
{
#if defined(__linux__)
    extern size_t strlen(const char *s);

    return(strlen(dp->d_name));
#else
    return(dp->d_namlen);
#endif
}

/*
 * Run every test case found in directory DIR (non-recursively).
 * Each test case is a regular file with a name ending with ".json".
 * Each filename must begin with a 'y_' if the test is expected to pass,
 * 'n_' if the test is expected to fail, or 'i_' if the result may
 * be either because of implementation/configuration choices.
 *
 * If a test produces the expected result, tests_passed is incremented,
 * otherwise tests_failed is incremented.
 * A misconfiguration or other error causes an exit with a non-zero status code.
 */
static int
do_test_suite(char *dirpath)
{
  int count, rc;
  char *test_file_path;
  size_t path_prefix_len, test_file_pathlen;
  struct dirent *de;
  DIR *dirp;
  static int exit_on_error = 1;

  fprintf(stderr, "Test Suite \"%s\" begin:\n", dirpath);

  if ((dirp = opendir((const char *) dirpath)) == NULL) {
	fprintf(stderr, "Cannot read directory \"%s\"\n", dirpath);
	exit(1);
  }

  test_file_path = NULL;
  test_file_pathlen = 0;
  path_prefix_len = strlen(dirpath) + 1;

  count = 0;
  while ((de = readdir(dirp)) != NULL) {
	int expected_result;
	char *expect_label;
	size_t need;

	if (de->d_type != DT_REG)
	  continue;
	if (!strsuffix(de->d_name, dirent_len(de), ".json"))
	  continue;

	if (de->d_name[0] == 'y' && de->d_name[1] == '_') {
	  expected_result = 1;
	  expect_label = "Pass";
	}
	else if (de->d_name[0] == 'n' && de->d_name[1] == '_') {
	  expected_result = -1;
	  expect_label = "Fail";
	}
	else if (de->d_name[0] == 'i' && de->d_name[1] == '_') {
	  expected_result = 0;
	  expect_label = "Implementation dependent";
	}
	else {
	  fprintf(stderr, "Invalid filename syntax: \"%s\"\n", de->d_name);
	  exit(1);
	}

	printf("%3d. %s (expect %s)... ", count, de->d_name, expect_label);
	fflush(stdout);

	need = path_prefix_len + dirent_len(de) + 1;
	if (need > test_file_pathlen) {
	  test_file_path = (char *) realloc(test_file_path, need);
	  test_file_pathlen = need;
	}
	sprintf(test_file_path, "%s/%s", dirpath, de->d_name);

	rc = run_suite_test(test_file_path, expected_result);
	if (rc == 1) {
	  tests_passed++;
	  fprintf(stdout, "OK\n");
	}
	else if (rc == 0) {
	  if (exit_on_error)
		exit(1);
	  tests_failed++;
	}
	else {
	  fprintf(stderr, "Invalid test case - exiting.\n");
	  exit(1);
	}

	count++;
  }

  if (errno != 0) {
	fprintf(stderr, "readdir() error: %s\n", strerror(errno));
	exit(1);
  }

  closedir(dirp);

  if (test_file_path)
	free(test_file_path);

  fprintf(stderr, "Test suite \"%s\" end: %d tests\n", dirpath, count);
  return(0);
}

/******/

#define VALUE_EQ(JP, V, VAL_START, VAL_END, VAL_TYPE)					\
  (js_start(JP, V) == VAL_START											\
   && js_end(JP) == VAL_END												\
   && (V)->type == (VAL_TYPE))

#define VALUE_STRING(JP, V, S)						\
  (strncmp(js_start(JP, V), S, js_len(JP, V)) == 0	\
   && strlen(S) == js_len(JP, V))

#define VALUE_PRINT(JP, V)												\
  fprintf(stderr, "start: %d, end: %d, type: %d, nchildren: %d\n",		\
		  js_start_offset(JP,V), js_end_offset(JP, V), (V).type, (V).nchildren)

static int
test_empty(void)
{
  const char *js;
  char *str;
  Jspr_rc r;
  Jspr_parser *jp;
  Jspr_value *v;

  js = "{}";
  jp = jspr_init(js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);
  v = jspr_value_root(jp);
  CHECK(v->type == JSPR_OBJECT_TYPE);
  CHECK(js_start(jp, v) == js && js_end(jp, v) == js + 1);

  js = "[]";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);
  v = jspr_value_root(jp);
  CHECK(v->type == JSPR_ARRAY_TYPE);
  CHECK(js_start(jp, v) == js && js_end(jp, v) == js + 1);

  js = "{\"a\":[]}";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);
  v = jspr_element_nth(jp, NULL, 0);
  CHECK(v && (v->type == JSPR_ARRAY_TYPE));
  str = jspr_name_of_value(jp, v);
  CHECK(str && (streq(str, "a")));
  CHECK(jspr_element_count(jp, v) == 0);

  js = "[{},{}]";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);
  CHECK(jspr_element_count(jp, NULL) == 2);

  jspr_end(jp);
  return(0);
}

static int
test_simple(void)
{
  const char *js;
  Jspr_parser *jp;
  Jspr_rc r;

  js = "{\"a\": 0}";

  jp = jspr_init(js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);
  jspr_print_values(NULL, jp, 2);

  js = "{\"a\":{},\"b\":{}}";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);

  js = "{\n\"Day\": 26,\n\"Month\": 9,\n\"Year\": 12\n}";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);

  jspr_end(jp);
  return(0);
}

typedef struct Test_value {
  int ind;			/* Index of value to test. */
  Jspr_type type;	/* Expected type of selected value. */
  char *str;		/* Expected string value of selected value. */
} Test_value;

typedef struct Test {
  char *text;
  Jspr_rc rc;				/* Expected parsing return code. */
  unsigned int linenum;		/* Line number of test data. */
  Test_value expected[4];	/* Vector of expected values. */
} Test;

static Test bjb_tests[] = {
  { "{ \"c\": { \"a\": [0, 1, 2, { \"b\": [ 3, 4]}, 5, 6] }}",
	JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  /* The most random test. */
  { "17",
	JSPR_OK, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "+17",
	JSPR_ERROR_NUMSYNTAX, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "-17",
	JSPR_OK, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "17a17", JSPR_ERROR_SYNTAX, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "0017", JSPR_ERROR_NUMSYNTAX, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "-017.17E+17", JSPR_ERROR_NUMSYNTAX, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "-17.17E+17", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "-17.17", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  /* There must be at least one value. */
  { "", JSPR_ERROR_VALUE_EXPECTED, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "\"\"", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "\"hello, world.\"", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "\"\\u0030\\u0031\\u0032\\u0033\"", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  /* There must be at least one value. */
  { "	", JSPR_ERROR_VALUE_EXPECTED, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "true", 0, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  /* Literal names must be lowercase. */
  { "TRUE", JSPR_ERROR_VALUE_EXPECTED, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "truez", JSPR_ERROR_VALUE_EXPECTED, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "true}", JSPR_ERROR_SYNTAX, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "t\\u0072rue", JSPR_ERROR_VALUE_EXPECTED, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "null", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "false", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{\"boolVar\" : true}", JSPR_OK, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },
										
  { "{\"boolVar\" : false}", JSPR_OK, __LINE__,
	{
#ifdef NOTDEF
	  { 1, JSPR_NAME_TYPE, "boolVar" },
	  { 2, JSPR_BOOLEAN_TYPE, "false" },
#endif
	  { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{\"nullVar\" : null}", JSPR_OK, __LINE__,
	{
#ifdef NOTDEF
	  { 1, JSPR_NAME_TYPE, "nullVar" },
	  { 2, JSPR_NULL_TYPE, "null" },
#endif
	  { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{\"intVar\" : 12345}", JSPR_OK, __LINE__,
	{
#ifdef NOTDEF
	  { 1, JSPR_NAME_TYPE, "intVar" },
	  { 2, JSPR_NUMBER_TYPE, "12345" },
#endif
	  { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{\"floatVar\" : 12.345}", JSPR_OK, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } }},

  { "{\"floatVar\" : -12.345e+58}", JSPR_OK, __LINE__,
	{ { -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{\"floatVar\" : +12.345e+58}", JSPR_ERROR_NUMSYNTAX, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "\"a\", \"b\"", JSPR_ERROR_SYNTAX, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{ \"foo\" : 17, \"baz\":  18 } ", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{ \"foo\" : 17 } ", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{  }", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "[1, 2, 3]", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "[1,2]", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "[17]", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "[]", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "[[[0]]]", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "[[[0]]", JSPR_ERROR_ARRSYNTAX, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "[ 0, [ 1, [ 2 ]]]", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "[ 0, [ 1, [ 2 ]], [3], [4]]", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { "{ \"a\": [0, 1, 2, { \"b\": [ 3, 4]}, 5, 6] }", JSPR_OK, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  /* Ends with a comma. */
  { "{ \"a\": [0, 1, 2, { \"b\": [ 3, 4]}, 5, 6] },",
	JSPR_ERROR_SYNTAX, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  /* Bogus comma. */
  { "{ \"a\": [0, 1, 2, { \"b\": [ 3,, 4]}, 5, 6] },",
	JSPR_ERROR_VALUE_EXPECTED, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  /* Missing right brace. */
  { "{ \"c\": { \"a\": [0, 1, 2, { \"b\": [ 3, 4]}, 5, 6] },",
	JSPR_ERROR_NAMESYNTAX, __LINE__,
	{{ -1, JSPR_UNKNOWN_TYPE, NULL } } },

  { NULL, JSPR_OK, __LINE__, { { -1, JSPR_UNKNOWN_TYPE, NULL } } }
};

static int
test_bjb(void)
{
  int tnum;
  Jspr_parser *jp;
  Jspr_rc rc;
  Test *test;

  jp = NULL;
  for (tnum = 1, test = &bjb_tests[0]; test->text != NULL; test++, tnum++) {
	if (jp == NULL)
	  jp = jspr_init(test->text, strlen(test->text));
	else
	  jspr_reset(jp, test->text, strlen(test->text));

	rc = jspr_parse_value(jp, NULL, NULL);
	if (rc != test->rc) {
	  fprintf(stderr, "Test %d (line %d) failed, rc=%d\n",
			  tnum, test->linenum, rc);
	  tests_failed++;
	  return(-1);
	}
	tests_passed++;
  }

  if (jp != NULL)
	jspr_end(jp);

  fprintf(stderr, "(Ran %d tests)\n", tnum);
  return(0);
}

static int
test_string(void)
{
  const char *js;
  Jspr_parser *jp;
  Jspr_rc r;
  Jspr_value *v;

  js = "[\" \t \"]";
  jp = jspr_init(js, strlen(js));
  r = jspr_parse_value(jp, &v, NULL);
  CHECK(r == JSPR_ERROR_INVALID_CHAR);

  js = "[\" \\t \"]";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse_value(jp, NULL, NULL);
  CHECKRC(jp, r);

  js = "{\"strVar\" : \"hello world\"}";
  jp = jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);
#ifdef NOTDEF
  v = jp->values;
  CHECK(v[0].type == JSPR_NAME_TYPE && v[1].type == JSPR_STRING_TYPE);
  CHECK(VALUE_STRING(js, v[0], "strVar"));
  CHECK(VALUE_STRING(js, v[1], "hello world"));
#endif

  js = "{\"strVar\" : \"escapes: \\/\\r\\n\\t\\b\\f\\\"\\\\\"}";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  v = jp->values;
  CHECKRC(jp, r);
#ifdef NOTDEF
  CHECK(v[0].type == JSPR_NAME_TYPE && v[1].type == JSPR_STRING_TYPE);
  CHECK(VALUE_STRING(js, v[0], "strVar"));
  CHECK(VALUE_STRING(js, v[1], "escapes: \\/\\r\\n\\t\\b\\f\\\"\\\\"));
#endif
 
  js = "{\"strVar\" : \"\"}";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);
#ifdef NOTDEF
  v = jp->values;
  CHECK(v[0].type == JSPR_NAME_TYPE && v[1].type == JSPR_STRING_TYPE);
  CHECK(VALUE_STRING(js, v[0], "strVar"));
  CHECK(VALUE_STRING(js, v[1], ""));
#endif

  jspr_end(jp);
  return(0);
}

static int
test_partial_string(void)
{
  const char *js;
  Jspr_parser *p;
  Jspr_rc r;

  js = "\"x\": \"va";
  p = jspr_init(js, strlen(js));
  r = jspr_parse(p);
  CHECK(r == JSPR_ERROR_SYNTAX);

#ifdef NOTDEF
  v = p->values;
  CHECK(r == JSPR_ERROR_PART && v[0].type == JSPR_NAME_TYPE);
  CHECK(VALUE_STRING(js, v[0], "x"));
  CHECK(p->next_value == 1);
#endif

  js = "\"x\": \"valu";
  jspr_reset(p, js, strlen(js));
  r = jspr_parse(p);
  CHECK(r == JSPR_ERROR_SYNTAX);
#ifdef NOTDEF
  v = p->values;
  CHECK(r == JSPR_ERROR_PART && v[0].type == JSPR_NAME_TYPE);
  CHECK(VALUE_STRING(js, v[0], "x"));
  CHECK(p->next_value == 1);
#endif

  js = "\"x\": \"value\"";
  jspr_reset(p, js, strlen(js));
  r = jspr_parse(p);
  CHECK(r == JSPR_ERROR_SYNTAX);
#ifdef NOTDEF
  v = p->values;
  CHECK(r == JSPR_OK && v[0].type == JSPR_NAME_TYPE
		&& v[1].type == JSPR_STRING_TYPE);
  CHECK(VALUE_STRING(js, v[0], "x"));
  CHECK(VALUE_STRING(js, v[1], "value"));
#endif

  js = "\"x\": \"value\", \"y\": \"value y\"";
  jspr_reset(p, js, strlen(js));
  r = jspr_parse(p);
  CHECK(r == JSPR_ERROR_SYNTAX);
#ifdef NOTDEF
  v = p->values;
  CHECK(r == JSPR_OK && v[0].type == JSPR_NAME_TYPE
		&& v[1].type == JSPR_STRING_TYPE && v[2].type == JSPR_NAME_TYPE
		&& v[3].type == JSPR_STRING_TYPE);
  CHECK(VALUE_STRING(js, v[0], "x"));
  CHECK(VALUE_STRING(js, v[1], "value"));
  CHECK(VALUE_STRING(js, v[2], "y"));
  CHECK(VALUE_STRING(js, v[3], "value y"));
#endif

  jspr_end(p);
  return(0);
}

static int
test_partial_array(void)
{
  const char *js;
  Jspr_parser *jp;
  Jspr_rc r;

  js = "  [ 1, true, ";
  jp = jspr_init(js, strlen(js));
  r = jspr_parse(jp);
  CHECK(r == JSPR_ERROR_VALUE_EXPECTED);
#ifdef NOTDEF
  v = jp->values;
  CHECK(r == JSPR_ERROR_PART && v[0].type == JSPR_ARRAY_TYPE 
		&& v[1].type == JSPR_NUMBER_TYPE && v[2].type == JSPR_BOOLEAN_TYPE);
#endif

  js = "  [ 1, true, [123, \"hello";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
#ifdef NOTDEF
  v = p->values;
  CHECK(r == JSPR_ERROR_PART && v[0].type == JSPR_ARRAY_TYPE 
		&& v[1].type == JSPR_NUMBER_TYPE && v[2].type == JSPR_BOOLEAN_TYPE
		&& v[3].type == JSPR_ARRAY_TYPE && v[4].type == JSPR_NUMBER_TYPE);
#endif

  js = "  [ 1, true, [123, \"hello\"]";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
#ifdef NOTDEF
  v = p->values;
  CHECK(r == JSPR_ERROR_PART && v[0].type == JSPR_ARRAY_TYPE 
		&& v[1].type == JSPR_NUMBER_TYPE && v[2].type == JSPR_BOOLEAN_TYPE
		&& v[3].type == JSPR_ARRAY_TYPE && v[4].type == JSPR_NUMBER_TYPE
		&& v[5].type == JSPR_STRING_TYPE);
  /* check child nodes of the 2nd array */
  CHECK(v[3].nchildren == 2);
#endif

  js = "  [ 1, true, [123, \"hello\"]]";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
#ifdef NOTDEF
  v = p->values;
  CHECK(r == JSPR_OK && v[0].type == JSPR_ARRAY_TYPE 
		&& v[1].type == JSPR_NUMBER_TYPE && v[2].type == JSPR_BOOLEAN_TYPE
		&& v[3].type == JSPR_ARRAY_TYPE && v[4].type == JSPR_NUMBER_TYPE
		&& v[5].type == JSPR_STRING_TYPE);
  CHECK(v[3].nchildren == 2);
  CHECK(v[0].nchildren == 3);
#endif

  jspr_end(jp);
  return(0);
}

static int
user_expand_values_cb(Jspr_parser *jp, unsigned int num_values)
{

  return(num_values + 1);
}

static int
test_array_nomem(void)
{
  int i;
  char buf[10240];
  const char *js;
  Jspr_config config_small;
  Jspr_parser *jp_small;
  Jspr_rc r;

  js = "[ 1, true, [123, \"hello\"], false, [17, 17, 17, 17]]";

  jspr_get_config(NULL, &config_small);
  config_small.enable_expand_values = 1;
  config_small.num_values = 1;
  config_small.values_limit = -1;
  config_small.user_expand_values_cb = user_expand_values_cb;

  jp_small = NULL;
  for (i = 0; i < 100; i++) {
	if (jp_small == NULL) {
	  strcpy(buf, js);
	  jp_small = jspr_init_x(NULL, buf, strlen(buf), &config_small);
	}
	else {
	  char *p;

	  p = buf + strlen(buf) - 1;
	  *p++ = ',';
	  strcpy(p, js + 1);
	  jspr_reset(jp_small, buf, strlen(buf));
	}

	r = jspr_parse(jp_small);
	CHECKRC(jp_small, r);

#ifdef NOTDEF
	vsmall = jp_small->values;
	CHECK(vsmall[0].type == JSPR_ARRAY_TYPE && vsmall[0].nchildren == 5);
	CHECK(vsmall[3].type == JSPR_ARRAY_TYPE && vsmall[3].nchildren == 2);
#endif
  }

  if (jp_small != NULL)
	jspr_end(jp_small);

  return(0);
}

static int
test_objects_arrays(void)
{
  const char *js;
  Jspr_parser *jp;
  Jspr_rc r;

  js = "[10}";
  jp = jspr_init(js, strlen(js));
  r = jspr_parse(jp);
  CHECK(r == JSPR_ERROR_ARRSYNTAX);

  js = "[10]";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECK(r == JSPR_OK);

  js = "{\"a\": 1]";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECK(r == JSPR_ERROR_OBJSYNTAX);

  js = "{\"a\": 1}";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse(jp);
  CHECK(r == JSPR_OK);

  jspr_end(jp);
  return(0);
}

static int
test_bjb1(void)
{
  const char *js;
  Jspr_rc r;
  Jspr_parser *jp;
  Jspr_value *cities0, *cities3;
  Jspr_value *city0, *city1;
  Jspr_value *index0, *index1, *index3;
  Jspr_value *v;

  js = " { \"foo\": { \"bar\": [ 80, 443 ], \"baz\": 1.2e-21 } } ";
  jp = jspr_init(js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);

  js = " { \"foo\": 1, \"bar\": 2, \"baz\": 3, \"foo\": 4 } ";
  jp = jspr_init(js, strlen(js));
  JSPR_CONFIG(jp)->require_unique_obj_names = 1;
  r = jspr_parse(jp);
  if (r != JSPR_ERROR_DUP_OBJNAME)
	CHECKRC(jp, r);
  JSPR_CONFIG(jp)->require_unique_obj_names = 0;

  js = "{\"accuracy\":654,\"real\":2.54e-3}";
/* XXX */
  jp = jspr_init(js, strlen(js));
  r = jspr_parse(jp);
  CHECKRC(jp, r);

#define EXAMPLE_JSON_DOC0 "[ \
  { \"Index\": 0, \"Country\": \"CA\",			   \
    \"Cities\": [ \"Victoria\", \"Vancouver\" ] }, \
  { \"Index\": 1, \"Country\": \"JP\",			   \
    \"Cities\": [ \"Tokyo\", \"Kyoto\" ] },		   \
  { \"Index\": 2, \"Country\": \"US\",			   \
    \"Cities\": [ \"New York\", \"Seattle\" ] },   \
  { \"Index\": 3, \"Country\": \"GB\",			   \
    \"Cities\": [ \"London\", \"Reading\" ] }	   \
  ]"

  js = EXAMPLE_JSON_DOC0;
  jp = jspr_init(js, strlen(js));
  r = jspr_parse_value(jp, &v, NULL);
  CHECKRC(jp, r);
  printf("Element count=%d\n", jspr_element_count(jp, NULL));

  index0 = jspr_element_nth(jp, v, 0);
  CHECK(index0 != NULL);
  printf("Index 0:\n");
  jspr_print_value(stdout, jp, index0);

  index1 = jspr_element_nth(jp, v, 1);
  CHECK(index1 != NULL);
  printf("Index 1:\n");
  jspr_print_value(stdout, jp, index1);

  cities0 = jspr_member_lookup(jp, index0, "Cities", JSPR_ANY_TYPE);
  CHECK(cities0 != NULL);
  city0 = jspr_element_nth(jp, cities0, 0);
  CHECK(city0 != NULL);
  printf("City 0, first member of Cities 0:\n");
  jspr_print_value(stdout, jp, city0);

  city1 = jspr_element_nth(jp, cities0, 1);
  CHECK(city1 != NULL);
  printf("City 1, second member of Cities 0:\n");
  jspr_print_value(stdout, jp, city1);

  index3 = jspr_element_nth(jp, v, 2);
  CHECK(index3 != NULL);
  printf("Index 3:\n");
  jspr_print_value(stdout, jp, index3);

  cities3 = jspr_member_lookup(jp, index3, "Cities", JSPR_ANY_TYPE);
  CHECK(cities3 != NULL);
  printf("City2, two Cities:\n");
  jspr_print_value(stdout, jp, cities3);

  printf("\n");

  jspr_print_values(stdout, jp, 2);
  printf("\n");

  {
	Jspr_res_info res_info;

	jspr_res_info(jp, &res_info);
	printf("-----\n");
	printf("values used:      %u\n", res_info.num_values_used);
	printf("values allocated: %u\n", res_info.num_values);
	printf("nexpands:         %u\n", res_info.nexpands);
	printf("nstrings:         %u\n", res_info.nstrings);
	printf("string_alloc:     %lu bytes\n", res_info.string_alloc);
	printf("xbuf_count:       %lu\n", res_info.xbuf_count);
	printf("xbuf_maxlen:      %lu bytes\n", res_info.xbuf_maxlen);
	printf("-----\n");
  }
  jspr_end(jp);

#define EXAMPLE_JSON_DOC1 "[ { \"id\": \"18241\" ,\"t\" : \"IBM\" ,\"e\" : \"NYSE\" ,\"l\" : \"158.23\" ,\"l_fix\" : \"158.23\" ,\"l_cur\" : \"158.23\" ,\"s\": \"0\" ,\"ltt\":\"12:10PM EST\" ,\"lt\" : \"Jan 8, 12:10PM EST\" ,\"lt_dts\" : \"2015-01-08T12:10:13Z\" ,\"c\" : \"+3.18\" ,\"c_fix\" : \"3.18\" ,\"cp\" : \"2.05\" ,\"cp_fix\" : \"2.05\" ,\"ccol\" : \"chg\" ,\"pcls_fix\" : \"155.05\" } ]"

  js = EXAMPLE_JSON_DOC1;
  jp = jspr_init(js, strlen(js));
  r = jspr_parse_value(jp, &v, NULL);
  CHECKRC(jp, r);

  printf("JSON DOC1:\n");
  jspr_print_values(stdout, jp, 2);

  {
	char *str;
	Jspr_value *l_cur;

	l_cur = jspr_member_lookup_fast(jp, NULL, "l_cur", JSPR_ANY_TYPE);
	CHECK(l_cur != NULL);
	str = jspr_string_copy(jp, l_cur);
	CHECK(str != NULL);
	CHECK(streq(str, "158.23"));
  }
  printf("\n");

  printf("JSON foo1:\n");
  js = "17";
  jspr_reset(jp, js, strlen(js));
  r = jspr_parse_value(jp, NULL, NULL);
  CHECKRC(jp, r);
  printf("OK\n");

  {
#define EXAMPLE_JSON_DOC2 "{\"response\": { 					\
	\"statusCode\": 200,										\
   	\"loginId\": \"outer@example.com\",							\
	\"data\": {													\
    	\"loginId\": \"middle@example.com\",					\
	    \"sessionSecret\": \"LwB3o09clyHoZpj1\",				\
	    \"token\": {											\
			\"expiresIn\": 1209600,								\
    		\"a\": \"%2FwQAAAAAAebSpiSuLfZH19dYJR3PZge5cYPAaj\" \
	      	\"loginId\": \"inner@example.com\",					\
		} 	\
	},		\
    \"statusText\": \"OK\"} \
}"

	js = EXAMPLE_JSON_DOC2;
	jspr_reset(jp, js, strlen(js));
	r = jspr_parse(jp);
	CHECK(r == JSPR_ERROR_OBJSYNTAX);
  }

  {
	unsigned int val;
	char *ptr;
	
#define EXAMPLE_JSON_DOC3 "{										\
\"response\": {														\
	\"statusCode\": 200,											\
   	\"loginId\": \"outer@example.com\",								\
	\"data\": {														\
    	\"loginId\": \"middle@example.com\",						\
	    \"sessionSecret\": \"LwB3o09clyHoZpj1\",					\
	    \"exponent\": 0.3141592e+1,									\
	    \"token\": {												\
			\"expiresIn\": 1209600,									\
    		\"a\": \"%2FwQAAAAAAebSpiSuLfZH19dYJR3PZge5cYPAaj\",	\
	      	\"loginId\": \"inner@example.com\"						\
		}															\
	},																\
    \"statusText\": \"OK\"											\
}																	\
}"

	printf("\n");
	printf("JSON DOC3:\n");

	js = EXAMPLE_JSON_DOC3;
	jspr_reset(jp, js, strlen(js));
	r = jspr_parse(jp);
	CHECKRC(jp, r);
	jspr_print_values(stdout, jp, 2);

	v = jspr_member_lookup_x(jp, JSPR_LOOKUP_PREORDER, NULL,
							  "statusCode", JSPR_ANY_TYPE);
	CHECK(v != NULL);
	if (jspr_string_value(jp, v, JSPR_UI, &val, &ptr) == 0) {
	  printf("Expect 200: %u\n", val);
	}
  }

  jspr_end(jp);
  printf("\n");

  return(r);
}

static int
test_parser1(void)
{
  char *js;
  Jspr_config config;
  Jspr_parser *jp;
  Jspr_rc rc;
  Jspr_value values[100];

  js = "{\"f\":\n 17}";
  jspr_get_config(NULL, &config);
  config.num_values = 10;
  config.user_values_ptr = (Jspr_value *) values;
  if ((jp = jspr_init_x(NULL, js, strlen(js), &config)) == NULL) {
	printf("Cannot create parser\n");
	return(-1);
  }

  rc = jspr_parse(jp);
  CHECKRC(jp, rc);

  jspr_end(jp);
  printf("OK\n");

  return(0);
}

typedef struct Cb_arg {
  FILE *fp;
} Cb_arg;

/*
 * Return 1 if ok, anything else causes traversal to stop with that value
 * returned.
 */
static MAYBE_UNUSED int
traverse_cb(Jspr_parser *jp, Jspr_value *v, void *arg)
{
  Jspr_index ind;
  Cb_arg *cb_arg;

  ind = value_index(jp, v);
  cb_arg = (Cb_arg *) arg;
  jspr_dump_value(cb_arg->fp, jp, ind);

  return(1);
}

/*
 * Callback for test_member_name_cb().
 */
static int
nm_cb(Jspr_parser *jp, Jspr_value *name, Jspr_value *v, void *cb_arg)
{

  if (v != NULL) {
	char *name_str;

	if ((name_str = jspr_name_of_value(jp, v)) != NULL) {
	  if (streq(name_str, "x")) {
		Jspr_value **cb_v;

		/* Terminate parsing. */
		fprintf(stderr, "Found name: \"%s\"\n", name_str);
		jspr_print_values(stderr, jp, 0);
		cb_v = (Jspr_value **) cb_arg;
		*cb_v = v;
		free(name_str);
		return(-1);
	  }
	  free(name_str);
	}
  }

  /* Continue parsing. */
  return(0);
}

/*
 * Test the new member callback mechanism.
 */
static int
test_member_name_cb(void)
{
  char *js;
  Jspr_config config;
  Jspr_parser *jp;
  Jspr_rc r;
  Jspr_value *cb_v;

  /* nm_cb() will look for the member named "x". */
  js = "{ \"one\": 1, \"two\": true, \"3\": \"foo\", \"obj\" : { \"x\": 17  }, \"minus\": -17  }";

  jspr_get_config(NULL, &config);
  config.user_new_member_cb = nm_cb;
  cb_v = NULL;
  config.user_new_member_cb_arg = &cb_v;
  jp = jspr_init_x(NULL, js, strlen(js), &config);
  r = jspr_parse(jp);

  if (r != JSPR_ERROR_USER_TERM || !jp->err_state.user_terminated) {
	jspr_end(jp);
	return(-1);
  }

  jspr_print_value(stderr, jp, cb_v);
  jspr_end(jp);
  return(0);
}

static void
usage(void)
{

  fprintf(stderr, "Usage: jspr_test ");
  fprintf(stderr, "[-a|-b|-f|-h|-t] [-c] [-d] [-v] [DIR ...]\n");
  fprintf(stderr, "-a | --all :    run all builtin tests, then exit.\n");
  fprintf(stderr, "-b | --builtin : run all builtin tests, then exit.\n");
  fprintf(stderr, "-f | --fast    : run fast builtin tests, then exit.\n");
  fprintf(stderr, "-c | --colour  : colourize JSON output.\n");
  fprintf(stderr, "-d | --debug   : enable debugging output.\n");
  fprintf(stderr, "-h | --help    : print this, then exit.\n");
  fprintf(stderr, "-s | --suite   : run test case suites (in DIR...)\n");
  fprintf(stderr, "-v | --verbose : enable verbose output.\n");
  fprintf(stderr, "DIR : directory containing test cases.\n");

  exit(0);
}

int
main(int argc, char **argv)
{
  int builtin_only_flag, fast_only_flag, suites_only_flag;
  int all_flag, colour_flag, debug_flag, i, verbose_flag;

  all_flag = 1;
  colour_flag = debug_flag = verbose_flag = 0;
  builtin_only_flag = fast_only_flag = suites_only_flag = 0;

  for (i = 1; argv[i] != NULL; i++) {
	if (streq(argv[i], "-a") || streq(argv[i], "--all"))
	  all_flag = 1;
	else if (streq(argv[i], "-b") || streq(argv[i], "--builtin")) {
	  builtin_only_flag = 1;
	  all_flag = 0;
	}
	else if (streq(argv[i], "-c") || streq(argv[i], "--colour"))
	  colour_flag = 1;
	else if (streq(argv[i], "-d") || streq(argv[i], "--debug"))
	  debug_flag = 1;
	else if (streq(argv[i], "-f") || streq(argv[i], "--fast")) {
	  fast_only_flag = 1;
	  all_flag = 0;
	}
	else if (streq(argv[i], "-h") || streq(argv[i], "--help")) {
	  usage();
	  /*NOTREACHED*/
	  exit(0);
	}
	else if (streq(argv[i], "-s") || streq(argv[i], "--suite")) {
	  suites_only_flag = 1;
	  all_flag = 0;
	}
	else if (streq(argv[i], "-v") || streq(argv[i], "--verbose"))
	  verbose_flag = 1;
	else if (argv[i][0] != '-')
	  break;
	else {
	  usage();
	  /*NOTREACHED*/
	  exit(1);
	}
  }

  fprintf(stdout, "(Using %s values)\n",
		  jspr_is_compact ? "Compact" : "Standard");

  /* Some tests for convenient development and debugging. */
  if (all_flag || fast_only_flag) {
	int i;
	const char *endp;
	Jspr_config config;
	Jspr_parser *jp;
	Jspr_rc r;
	Jspr_value *v;
	static char *jsv[] = {
	  "[ 1 ]",
	  "[ [ ] ]",
	  "[ 1, [ 2 ] ]",
	  "{ \"one\":1, \"two\":2}",
	  "[ 1, 2, [ 3, 4], 5]",
	  "[ 1, [ 2, 3], [ 4, 5], [6, 7]]",
	  "[ 1, [ 2, [ 3, [ 4, [ 5, [ 6, [ 7, [ 8, [ 9 ] ] ] ] ] ] ] ] ]",
	  "{ \"one\": 1, \"two\": true, \"3\": \"foo\", \"obj\" : { \"x\": 17  }, \"minus\": -17  }",
	  "{ \"one\": 1, \"two\" : [ 2 ], \"three\" : 3.0 }",
	  NULL
	};

	fprintf(stdout, "Builtin fast tests:\n");
	jspr_get_config(NULL, &config);
	config.enable_colourize = colour_flag;
	config.enable_debug = debug_flag;
	config.enable_verbose = verbose_flag;

	if (0)
	  jspr_expand_values(jp, 50);

	for (i = 0; jsv[i] != NULL; i++) {
	  if (i == 0)
		jp = jspr_init_x(NULL, jsv[i], strlen(jsv[i]), &config);
	  else
		jspr_reset(jp, jsv[i], strlen(jsv[i]));
	  fprintf(stdout, "\nInput:\n%s\n\n", jsv[i]);

	  if ((r = jspr_parse_value(jp, &v, &endp)) == JSPR_OK) {
		jspr_print_values(stdout, jp, 0);
		jspr_dump(stdout, jp);
	  }
	  else {
		fprintf(stdout, "Failed: %s\n", jspr_strerror(r));
		exit(1);
	  }
	}
	fprintf(stdout, "OK.\n\n");

	/* Try some callback tests. */
	{
	  Cb_arg cb_arg;
	  jspr_dump(stdout, jp);

	  fprintf(stdout, "\nBuiltin callback tests:\n");
	  cb_arg.fp = stdout;
	  /*
		fprintf(stdout,
		"Ind Type     Start/End   Pt  #C   CF/CL   SP/SN    Name  Src\n");
		jspr_preorder(jp, v, traverse_cb, &cb_arg);
	  */

	  v = jspr_member_lookup_fast(jp, NULL, "obj", JSPR_OBJECT_TYPE);
	  v = jspr_member_lookup_fast(jp, v, "x", JSPR_ANY_TYPE);
	  if (v != NULL) {
		fprintf(stdout, "Match: ");
		jspr_print_value(stdout, jp, v);
	  }

	  v = jspr_member_lookup(jp, NULL, "3", JSPR_ANY_TYPE);
	  if (v != NULL) {
		fprintf(stdout, "Match: ");
		jspr_print_value(stdout, jp, v);
	  }

	  cb_arg.fp = stdout;
	  jspr_breadth_first(jp, v, traverse_cb, &cb_arg);
	  fprintf(stdout, "\n\n");
	  jspr_preorder(jp, v, traverse_cb, &cb_arg);

	  fprintf(stdout, "OK.\n");
	}

	if (jp != NULL)
	  jspr_end(jp);

	if (fast_only_flag)
	  exit(0);
  }

  if (all_flag || builtin_only_flag) {
	int n;

	/* Some of these tests could be moved to an (external) test suite now. */
	n = 1;
	do_tests("test_member_name_cb", n++, test_member_name_cb);

	if (do_tests("test_bjb", n++, test_bjb) != 0)
	  exit(1);

	do_tests("test_parser1", n++, test_parser1);
	do_tests("test_bjb1", n++, test_bjb1);

	do_tests("Empty JSON objects/arrays", n++, test_empty);
	do_tests("Simple JSON string", n++, test_simple);
	do_tests("JSON data types", n++, test_string);
	do_tests("Partial JSON string parsing", n++, test_partial_string);
	do_tests("Partial array reading", n++, test_partial_array);
	do_tests("Array reading with a smaller number of tokens", n++,
			 test_array_nomem);
	do_tests("Objects and arrays", n++, test_objects_arrays);

	printf("\n");
  }

  if (all_flag || suites_only_flag) {
	/* Process each test suite (a directory of JSON files). */
	for (; argv[i] != NULL; i++) {
	  /* Argument is pathname of the directory containing the test files. */
	  if (do_test_suite(argv[i]) != 0)
		exit(1);
	}
  }

  printf("Tests passed: %2d\n", tests_passed);
  printf("Tests failed: %2d\n", tests_failed);
  if (tests_failed) {
	printf("FAILED\n");
	exit(1);
  }
  
  printf("OK\n");
  exit(0);
}
