/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*
 * In development... usable but incomplete and a bit buggy.
 *
 * XXX as of Id stamp 3320, there are still some issues with cases that
 * recurse/list a subdirectory when it probably should not
 */

/*
 * lf - list files
 *
 * An alternative, not fully compatible ls(1) command, originally written
 * to test fts(3) library functions and various DACS library functions.
 *
 * Implemented as three consecutive phases:
 * 1) Generate sorted lists of files from file/directory arguments
 * 2) Modify lists to select/deselect files of interest from arguments
 * 3) Format and print lists according to arguments
 *
 * Several new/convenient features that have proved to be difficult/inconvient
 * to approximate using scripts or aliases have been included, such as the
 * ability to:
 * 1) specify the formatting of each listed item (modelled after stat(1);
 * 2) list only directories or only non-directories;
 * 3) list only files modified within a certain number of days;
 * 4) configure coloured printing of file names or listing lines
 *    based on a characteristic of that file or its name.
 *
 * Some rarely-used features and behaviours have been changed or omitted.
 * If you need these, use one of the many other implementations.
 *
 * Most ls(1) commonly used command line flags work as expected, but
 * some have been renamed and new ones have been added.
 * Like ls(1), many flags are idiosyncratic and/or implementation dependent
 * and/or platform dependent and/or strangely context sensitive, so output
 * formatting can be unexpectedly dependent on flags and file types, or work
 * a little differently than ls(1) and/or gnuls(1) on other systems.
 *
 * Flags are roughly divided into three kinds:
 * 1) File/Directory selection flags: determines the files to be listed
 * 2) Information flags: for selected files, indicates the file characteristics
 *    of interest
 * 3) Formatting flags: configures the appearance of the listings
 *
 * At present there is no manual page.
 * For documentation, do: lf --help
 *
 * Alternate command line entry points:
 * 1) lsd   - list only directories (also "lfd" and "lf -Dd")
 * 2) lsf   - list only non-directory files (also "ls -Dx")
 * 2) lss   - list only symbolic links (also "ls -Ds")
 * 3) today - list regular files modified/changed/accessed today or within the
 *            previous # days (also "lf -t[#]")
 *
 * Environment (also see -Coff):
 * LF_COLOURS: default colour-related command line arguments, e.g.:
 *   setenv LF_COLOURS "-C- red -C/ blue"
 *   setenv LF_COLOURS "-C/ blue"
 *
 * Info about missing/incompatible flags. Individual and combined flag
 * semantics can differ in different implementations, sometimes subtly so:
 * o POSIX:
 *   -C: force multi-column, sorted output (default)
 *       [FreeBSD]
 *   -H: for a symlink-to-directory argument, print that name but with info for
 *       the dereferenced directory (overrides -F/-P)
 *       [FreeBSD: Flag applies only to arguments; implied if no -F, -d, or -l]
 *   -L: For ALL symbolic links, dereference to get all file info, including
 *       type, but print name of link (overrides -F/-P)
 *       [FreeBSD: Same, but applies only to arguments]
 *   -f: force each argument to be interpreted as a directory and list names
 *       [FreeBSD: unsorted output, various flag interactions]
 *   -g: like -l, except do not write owner
 *       [FreeBSD: no effect]
 *   -m: stream output format, separate files by a comma
 *       [FreeBSD]
 *   -n: like -l, except use numeric UID and GID
 *       [FreeBSD: effective with -l]
 *   -o: like -l, except do not write group
 *       [FreeBSD: include file flags with -l]
 *   -q: non-printable and tab chars are written as '?'
 *       [FreeBSD]
 *   -t: sort by most recent time modified/filename
 *   -x: like -C except sort across columns
 *    Environment variables: COLUMNS, LANG, LC_ALL, LC_COLLAGE, LC_CTYPE,
 *    LC_MESSAGES, LC_TIME, NLSPATH, TZ
 *    Output formats are specified for some flags and flag combos
 *    See:
 *    https://www.unix.com/man-page/posix/1posix/ls/
 * FreeBSD:
 *   -P: List symbolic link arguments; cancels -H and -L
 * o GNU:
 *    https://linux.die.net/man/1/ls
 *
 * Implementation:
 * This could (should?) be implemented as three independent, pipeline-able
 * programs by:
 * 1) Generating an ordered list of filenames, possibly involving directory
 *    traversals, following symbolic links, etc. with needed meta information
 * 2) Selecting and rejecting files from that list according to options,
 *    calculating blocks used by directories
 * 3) Formatting and emitting output
 *
 * Though this component architecture would be somewhat less efficient
 * (runtime and implementation-wise), involving more
 * configuration/build/maintenance/documentation effort,
 * it's not clear whether the individual programs would be of much use on
 * their own since much of their commonly-used functionality is available in
 * existing commands (such as find(1)).
 * On the other hand, this implementation is perhaps a little too monolithic.
 * After some of its useful and interesting functionality is explored
 * a better implementation can be weighed.
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2024\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: lf.c 3329 2024-09-23 19:59:02Z brachman $";
#endif

#include "dacs_config.h"

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "dacs_malloc.h"
#ifdef ENABLE_FTS
#include "dacs_fts.h"
#else
#include <fts.h>
#endif

#include "ds.h"
#include "dacs_qsort.h"
#include "str.h"

static char *log_module_name = "lf";

/* Undefine to remove testing code. */
#define LF_TESTS

/* Define for "ls -l" date logic; see format_line(). */
#undef USE_LS_COMMAND_DATES

#define  DEFAULT_SHORT_FORMAT	"%n"
#define  DEFAULT_LONG_FORMAT	"%m %l %u %g %s %dd %n"

/* A multiple of 512. */
#define DEFAULT_LONG_BLOCKSIZE	512

#define DEFAULT_RGB_SEARCH_PATH											\
  "/usr/lib/X11:/usr/local/lib/X11:/usr/share/X11:/opt/X11/share/X11"

/*
 * Writing new settings overrides current settings, so all necessary settings
 * must be written at the same time rather than separately; e.g., you must
 * set bold and underline in the same operation, otherwise only
 * one of the two effects will be used.
 *
 * \033[38;2;<r>;<g>;<b>m : Set foreground colour using RGB
 * \033[31;<c>            : Font effects
 *   0 = reset, 1 = bold, 4 = underline (and others)
 *
 * Examples:
 * Black & Bold: "\033[38;2;0;0;0;1m"
 * Red & Bold: "\033[31;1m"
 * Bold: "\033[1m"
 * Underline: "\033[4m"
 * Reset: "\033[0m"
 * See, e.g.,
 * https://stackoverflow.com/questions/4842424/list-of-ansi-color-escape-sequences
 */

#define ANSI_TTY_RESET_STR		"\033[0m"
#define COLOUR_RGB_FILENAME		"rgb.txt"

#define DEFAULT_TTY_BOLD_OP		1

/* A parsed rgb.txt entry. */
typedef struct {
  char *desc;
  char *name;
  unsigned char red;
  unsigned char green;
  unsigned char blue;
} Colour_rgb;

/* Maximum number of entries read from rgb.txt. */
#define MAX_COLOURS	1000
static Colour_rgb *colour_rgb[MAX_COLOURS];

/* Bits. */
typedef enum {
  TTY_RESET_OP     = 0,
  TTY_PLAIN_OP     = 1,
  TTY_BOLD_OP      = 2,
  TTY_UNDERLINE_OP = 4
} Tty_op;

typedef enum {
  SORT_UNSORTED      = 0,
  SORT_NEW_TO_OLD    = 1,
  SORT_OLD_TO_NEW    = 2,
  SORT_BIG_TO_SMALL  = 3,
  SORT_SMALL_TO_BIG  = 4,
  SORT_FILENAME_AZ   = 5,	/* Case sensitive, ascending. */
  SORT_FILENAME_ZA   = 6,	/* Case sensisitve, descending. */
  SORT_FILENAME_Az   = 7,	/* Case insensitive, ascending. */
  SORT_FILENAME_zA   = 8,	/* Case insensitive, descending. */
  SORT_DECREASING_LINK_COUNT = 9
} Sort_type;

typedef enum {
  TIME_ATIME    = 1,	/* Access time (st_atime). */
  TIME_BTIME    = 2,	/* Birth time (st_birthtime, FreeBSD). */
  TIME_CTIME    = 3,	/* Change time (st_ctime). */
  TIME_MTIME    = 4		/* Modification time (st_mtime). */
} Time_type;

/* Various flags that determine the computation for displaying blocksizes. */
typedef enum {
  BS_DEFAULT_UNITS = 0,	/* Default. */
  BS_HUMAN_UNITS   = 1,	/* -h flag. */
  BS_K_UNITS       = 2,	/* -k flag. */
  BS_BS_UNITS      = 3	/* --blocksize flag. */
} Blocksize_flag;

/*
 * Information extracted from fts(3) traversals, stored as a vector.
 * Level FTS_ROOTPARENTLEVEL is the parent of the root (starting point) of
 * a traversal and level FTS_ROOTLEVEL is the root of a traversal.
 */
typedef struct File_d {
  char *fts_path;
  size_t fts_pathlen;
  size_t fts_namelen;
  long fts_level;
  struct stat *fts_statp;
} File_d;

typedef struct Hier {
  char *filename;			/* As it appears on the command line. */
  struct stat stat;
  int st;  /* -1 -> stat failed; 0 -> ENOENT; 1 -> ok; 2 -> broken symlink */
} Hier;

typedef struct Colour_effect {
  char *arg;
  char *effect_str;
} Colour_effect;

/*
 * Text highlighting configuration.
 * Indexes into colour_effect[].
 */
typedef enum {
  COLOUR_EFFECT_DEFAULT_ALL     = 0,	/* None of the others apply. */
  COLOUR_EFFECT_DEFAULT_SPECIAL = 1,	/* For non-regulars if nothing else. */
  COLOUR_EFFECT_REGULAR         = 2,	/* For regular files  "-" */
  COLOUR_EFFECT_DIR             = 3,	/* For directories    "/" */
  COLOUR_EFFECT_LINK            = 4,	/* For links          "@" */
  COLOUR_EFFECT_EXEC            = 5,	/* For executables    "*" */
  COLOUR_EFFECT_SOCK            = 6,	/* For sockets        "=" */
  COLOUR_EFFECT_FIFO            = 7,	/* For FIFOs          "|" */
  COLOUR_EFFECT_WHITEOUT        = 8,	/* For FreeBSD whiteouts "%" */
  COLOUR_EFFECT_UNPRINTABLE     = 9,	/* For unprintables   "x" */
  COLOUR_EFFECT_DOT             = 10,	/* A regular dot file "." */
  COLOUR_EFFECT_GLOB            = 11,	/* A glob expression  "g" */
  COLOUR_EFFECT_NTYPES          = 12	/* The number of type slots needed. */
} Colour_effect_type;

/* Selectors for today_count_flag, if it is not zero. */
#define TODAY_COUNT_WITHIN_DAYS	1
#define TODAY_COUNT_EXACT_DAYS	2

#define TODAY_COUNT_DEFAULT_DAYS	1

#include <time.h>
#if defined(DACS_OS_LINUX)
#include <sys/vfs.h>
#endif

#if defined(DACS_OS_MACOSX)
#include <sys/param.h>
#include <sys/mount.h>
/* Not sure why/if this is required for only the MacOS/arm64 build... */
int current_log_level;
#endif

static Colour_effect colour_effect[COLOUR_EFFECT_NTYPES];

typedef struct Colour_effect_glob {
  char *glob;
  Colour_effect effect_arg;
} Colour_effect_glob;

static Colour_effect_glob *colour_effect_glob = NULL;
static int colour_effect_glob_nalloc = 0;
static int colour_effect_glob_count = 0;

static File_d **file_set = NULL;
static int nfiles = 0;
static size_t nfiles_alloc = 0;

/*
 * -A
 * On FreeBSD, this flag is a default for the super-user unless -I is used.
 * For GNU ls, --almost-all is a synonym for -A, which is not a default,
 * and -I has different semantics.
 * See: -aa/--almost-all and -aar/--almost-all-root
 */
static int absolute_flag = 0;

/* Post command line flag processing, honour this blocksize config. */
static Blocksize_flag which_bs_flag = BS_DEFAULT_UNITS;

/*
 * When all_flag/almost_all_flag is NOT set, only dot files/directories that
 * appear on the command line are listed.
 * The all_flag will list all files/directories that are encountered whereas
 * the almost_all_flag will list all files/directories except that
 * "." and ".." will be excluded unless they appear on the command line.
 *
 * XXX consider GNU ls: --block-size=SIZE,  --si
 * GNU ls has quite a few esoteric options...
 */
static int all_flag = 0;
static int almost_all_flag = 0;
static int almost_all_root_flag = 0;
static int blocks_flag = 0;
static char *blocksize_units_flag = NULL;
static int bs_si_flag = 0;
static int colour_long_only_flag = 0;
static int colour_line_flag = 0;
static int dirsfirst_flag = 0;
static int filetype_flag = 0;
static int groupname_flag = 0;
static int human_units_flag = 0;
static int human_units_base_flag = 0;
static int inum_flag = 0;
static int is_today_prog = 0;
static unsigned int limit_count_flag = 0;
static int long_flag = 0;
static unsigned int user_spec_blocksize = DEFAULT_LONG_BLOCKSIZE;
static int lsd_flag = 0;
static int lsf_flag = 0;
static int lss_flag = 0;
/* Merge one or more filenames on to a maximum-length output line. */
static int merge_flag = 0;
/* Was given an argument from which to read filenames to list, one per line. */
static int namefile_flag = 0;
static int number_flag = 0;

/* -d: do not descend into any listed directories. */
static int onelevel_flag = 0;

/*
 * Surround filenames with single quotes; useful when filenames will be used
 * by a shell. Not so useful if file type suffixes are enabled.
 */
static int quoted_flag = 0;

/*
 * FreeBSD -R does a depth-first recursive listing.
 * If set to -1 (negative), recurse indefinitely.
 * If set to 0, do not recurse (enables onelevel_flag).
 * If set to a positive number, recurse up to that many levels deep.
 */
static int recurse_flag = 1;

#define S_IPMT			07777

typedef enum {
  PERMS_MATCH_EXACT     = 1,		/* Permission bits must match exactly. */
  PERMS_MATCH_SET       = 2,		/* Permission bits must be set. */
  PERMS_MATCH_NOTSET    = 3,		/* Permission bits must not be set. */
  PERMS_MATCH_IGNORE    = 0		/* Permission bits not examined. */
} Perms_match_flag;

static Perms_match_flag perms_match_flag = PERMS_MATCH_IGNORE;
static mode_t perms_match = 0;

static int reverse_sort_flag = 0;
static int shell_escape_flag = 0;	/* XXX */
static int single_column_flag = 0;
static int slash_flag = 0;
static int sortgid_flag = 0;
static int sortuid_flag = 0;
static int symlink_Hflag = 0;
static int symlink_Lflag = 0;
static int symlink_Pflag = 0;
static int today_count_flag = 0;
static unsigned int today_count = 0;
static int unprintable_filenames_flag = 0;
static int username_flag = 0;
static int verbose_flag = 0;

static char *type_match_flag = NULL;

/* A strftime(3) specification for printing a date/time. */
static char *dateformat_flag = NULL;
/* A user-specified format to apply to each output line. */
static char *format_flag = NULL;
/* A user-specified string to prepend to each filename. */
static char *prefix_path_flag = NULL;
/* A user-specified path for the RGB colour specification file to use. */
static char *rgbpath_flag = NULL;

/* Initialize to the defaults. */
static Sort_type sort_type_flag = SORT_FILENAME_Az;
static Time_type time_type_flag = TIME_MTIME;

/*
 * If 1, colour initialization is required and an RGB colour definition file
 * must be loaded.
 * If -1, colour rendering is not required or has been disabled, so just ignore
 * any flags associated with it.
 * Otherwise (if 0), no colour initialization has been done yet.
 */
static int colours_enabled = 0;

/*
 * If non-zero, a flag specifically requested colourization so do not
 * disable it when that is the default behaviour.
 */
static int colours_forced_flag = 0;

/* The format to apply to each output line. */
static char *line_format = NULL;

/* Number of command line file arguments. */
static int nfile_args = 0;

static int fts_read_count = 0;
static int saw_dot_file_arg = 0;

static char *progname = NULL;

#ifdef NOTDEF
static inline MAYBE_UNUSED int
streq(const char *p, const char *q)
{

    return(strcmp(p, q) == 0);
}

static inline MAYBE_UNUSED int
strcaseeq(const char *p, const char *q)
{

    return(strcasecmp(p, q) == 0);
}

static inline MAYBE_UNUSED int
strneq(const char *p, const char *q, size_t n)
{

    return(strncmp(p, q, n) == 0);
}

static MAYBE_UNUSED char *
strprefix(const char *p, const char *prefix)
{
  size_t len;

  len = strlen(prefix);
  if (strneq(p, prefix, len))
    return((char *) p + len);

  return(NULL);
}

/*
 * Map STR to lower case in supplied buffer BUF, assumed to be large enough.
 * They can be the same buffer.
 */
char *
strlowerbuf(char *buf, char *str)
{
  char *p, *q;

  for (p = str, q = buf; *p != '\0'; p++, q++)
    *q = tolower((int) *p);

  *q = '\0';

  return(buf);
}
#endif

#ifndef HAVE_STRNDUP
/*
 * Return a copy of P, up to a maximum of MAXLEN bytes.
 * That is, strlen() of the returned string will be the lessor of
 * strlen(p) and MAXLEN.
 * The returned string is always null-terminated.
 */
static inline MAYBE_UNUSED char *
strndup(const char *p, size_t maxlen)
{
  char *str;
  size_t len;

  len = strnlen(p, maxlen);

  str = (char *) malloc(len + 1);
  memcpy(str, p, len);
  str[len] = '\0';

  return(str);
}

/*
 * Create an inode mode descriptor string from MODE.
 * Alternate implementation of FreeBSD's strmode(3).
 * The string is always 11 bytes long (plus a null).
 * The last character is always a space.
 *
 * If OUTP is non-NULL, it is assumed to be at least 12 bytes; if OUTP is
 * NULL, allocate the 12 bytes.
 * Return a pointer to the formatted string.
 */
static char *
strfilemode(mode_t mode, char *outp)
{
  char *p, *s;

  if (outp == NULL)
	s = p = (char *) malloc(12);
  else
	s = p = outp;

  /* Inode type (one character). */
  switch (mode & S_IFMT) {
  case S_IFREG:           /* Regular */
	*p++ = '-';
	break;
  case S_IFBLK:           /* Block special */
	*p++ = 'b';
	break;
  case S_IFCHR:           /* Character special */
	*p++ = 'c';
	break;
  case S_IFDIR:           /* Directory */
	*p++ = 'd';
	break;
  case S_IFLNK:           /* Symbolic link */
	*p++ = 'l';
	break;
#ifdef S_IFIFO
  case S_IFIFO:           /* FIFO */
	*p++ = 'p';
	break;
#endif
  case S_IFSOCK:          /* Socket */
	*p++ = 's';
	break;
#ifdef S_IFWHT
  case S_IFWHT:           /* Whiteout (BSD) */
	*p++ = 'w';
	break;
#endif
  default:            /* Unknown inode type */
	*p++ = '?';
	break;
  }

  /* Owner's permissions (3 characters). */
  if (mode & S_IRUSR)
	*p++ = 'r';
  else
	*p++ = '-';

  if (mode & S_IWUSR)
	*p++ = 'w';
  else
	*p++ = '-';

  switch (mode & (S_IXUSR | S_ISUID)) {
  case 0:
	*p++ = '-';
	break;
  case S_IXUSR:
	*p++ = 'x';
	break;
  case S_ISUID:
	*p++ = 'S';
	break;
  case S_IXUSR | S_ISUID:
	*p++ = 's';
	break;
  default:            /* Unknown? */
	*p++ = '?';
	break;
  }

  /* Group's permissions (3 characters). */
  if (mode & S_IRGRP)
	*p++ = 'r';
  else
	*p++ = '-';

  if (mode & S_IWGRP)
	*p++ = 'w';
  else
	*p++ = '-';

  switch (mode & (S_IXGRP | S_ISGID)) {
  case 0:
	*p++ = '-';
	break;
  case S_IXGRP:
	*p++ = 'x';
	break;
  case S_ISGID:
	*p++ = 'S';
	break;
  case S_IXGRP | S_ISGID:
	*p++ = 's';
	break;
  default:            /* Unknown? */
	*p++ = '?';
	break;
  }

  /* Other's (default) permissions (3 characters). */
  if (mode & S_IROTH)
	*p++ = 'r';
  else
	*p++ = '-';

  if (mode & S_IWOTH)
	*p++ = 'w';
  else
	*p++ = '-';

  switch (mode & (S_IXOTH | S_ISVTX)) {
  case 0:
	*p++ = '-';
	break;
  case S_IXOTH:
	*p++ = 'x';
	break;
  case S_ISVTX:
	*p++ = 'T';
	break;
  case S_IXOTH | S_ISVTX:
	*p++ = 't';
	break;
  default:            /* Unknown? */
	*p++ = '?';
	break;
  }

  *p++ = ' ';
  *p = '\0';

  return(s);
}

/*
 * Given string P of length LEN, if P ends with the string SUFFIX,
 * return a pointer to the start of SUFFIX in P, NULL otherwise.
 * SUFFIX is a suffix of P if they are identical.
 */
static char *
strsuffix(char *p, size_t len, const char *suffix)
{
  char *q;
  size_t suffix_len;

  suffix_len = strlen(suffix);
  if (suffix_len > len)
    return(0);
  q = p + len - suffix_len;
  return(streq(q, suffix) ? q : NULL);
}

/*
 * Return a copy of PATHNAME after deleting 1) any prefix ending with the last
 * slash character (after first stripping trailing slashes), and 2) a SUFFIX,
 * if given.  The SUFFIX string is not stripped if it is identical to the
 * remaining characters in PATHNAME.  The resulting string is returned.
 * A non-existent suffix is ignored.
 * This is intended to be like the basename(1) command.
 */
static char *
strbasename(const char *pathname, const char *suffix)
{
  char *last, *p, *q, *s;
  size_t len;

  len = strlen(pathname);
  if (len == 0)
    return(strdup(""));

  s = strdup(pathname);
  p = s + len - 1;

  /* Delete trailing slashes first. */
  while (p > s && *p == '/')
    *p-- = '\0';

  /* Is the pathname equivalent to "/"? */
  if (*p == '/')
    return(strdup("/"));

  if ((last = rindex(s, (int) '/')) != NULL)
    *last++ = '\0';
  else
    last = s;

  /*
   * We've found the end of the last component.
   * If SUFFIX was given, check if it needs to be deleted from the end of
   * the string.
   */
  if (suffix != NULL && !streq(last, suffix)
      && (q = strsuffix(last, strlen(last), suffix)) != NULL)
    *q = '\0';

  return(last);
}

#endif

/*
 * Return an approximation of VAL in terms of four or fewer characters by
 * appending a unit suffix of Bytes (B), Kilobytes (K), Megabytes (M),
 * Gigabytes (G), Terabytes (T), or Petabytes (P).
 * KBASE is either 1000 (if a kilobytes is 1000 bytes) or 1024.
 * E.g.,
 * 58 -> "58B", 4096 -> "4K" (KBASE=1024) or "4.1K" (KBASE=1000),
 * 389711 -> "381K", 8611240 -> "8.2M", 13494456508 -> "13G"
 *
 * Return NULL if VAL cannot be approximated.
 */
typedef struct Unit_tab {
  char *suffix;
  unsigned long max;
} Unit_tab;

static Unit_tab unit_tab1000[] = {
  /* 10^3 */  { "B",                1000 },
  /* 10^6 */  { "K",             1000000 },
  /* 10^6 */  { "M",          1000000000 },
  /* 10^9 */  { "G",       1000000000000 },
  /* 10*12 */ { "T",    1000000000000000 },
  /* 10^15 */ { "P", 1000000000000000000 },
  { NULL, 0 }
};

static Unit_tab unit_tab1024[] = {
  /* 2^10 */ { "B",              0x400 },
  /* 2^20 */ { "K",           0x100000 },
  /* 2^30 */ { "M",         0x40000000 },
  /* 2^40 */ { "G",      0x10000000000 },
  /* 2^50 */ { "T",    0x4000000000000 },
  /* 2^60 */ { "P", 0x1000000000000000 },
  { NULL, 0 }
};

static char *
str_unit_suffix(unsigned long val, unsigned int units_base)
{
  int i;
  char *str;
  Unit_tab *ut;

  if (units_base == 10)
	ut = unit_tab1000;
  else if (units_base == 2)
	ut = unit_tab1024;
  else
	return(NULL);

  for (i = 0; ut->suffix != NULL && val >= ut->max; i++)
	ut++;

  if (ut->suffix == NULL)
	return(strdup("????"));

  if (i == 0)
	asprintf(&str, "%ld%s", val, ut->suffix);
  else {
	long xx;
	double fracpart, intpart, x;

	x = (double) val / (double) (ut - 1)->max;
	fracpart = modf(x, &intpart);
	xx = lrint(fracpart * 10.0);
	if (val < 1000000)
	  asprintf(&str, "%lu%s", (unsigned long) lround(x), ut->suffix);
	else if (xx <  1)
	  asprintf(&str, "%lu%s", (unsigned long) intpart, ut->suffix);
	else
	  asprintf(&str, "%.1f%s", x, ut->suffix);
  }

  return(str);
}

/* Default field widths. */
#define MIN_BLOCKS_FIELD_WIDTH 		3
#define MIN_DEV_FIELD_WIDTH 		3
#define MIN_RDEV_FIELD_WIDTH		8
#define MIN_NLINK_FIELD_WIDTH		1
#define MIN_INO_FIELD_WIDTH			1
#define MIN_SIZE_FIELD_WIDTH		4
#define MIN_PW_NAME_FIELD_WIDTH		6
#define MIN_GR_NAME_FIELD_WIDTH		6
#define MIN_UID_FIELD_WIDTH			4
#define MIN_GID_FIELD_WIDTH			4

/* The mode string width excludes the space appended by strfilemode(). */
#define MODE_FIELD_WIDTH			10
#define FLAGS_FIELD_WIDTH			8

/* Maximum field values to calculate field widths. */
static unsigned long st_dev_max = 0;
static unsigned long st_rdev_max = 0;
static unsigned long st_nlink_max = 0;
static unsigned long st_ino_max = 0;
static long st_size_max = 0;
static size_t st_uid_max = 0;
static size_t st_gid_max = 0;
static size_t pw_name_max = 0;
static size_t gr_name_max = 0;
static size_t path_name_max = 0;

/* Computed field widths to be used. */
static int field_blocks_width = MIN_BLOCKS_FIELD_WIDTH;
static int field_dev_width = MIN_DEV_FIELD_WIDTH;
static int field_rdev_width = MIN_RDEV_FIELD_WIDTH;
static int field_nlink_width = MIN_NLINK_FIELD_WIDTH;
static int field_ino_width = MIN_INO_FIELD_WIDTH;
static int field_size_width = MIN_SIZE_FIELD_WIDTH;
static int field_pw_name_width = MIN_PW_NAME_FIELD_WIDTH;
static int field_gr_name_width = MIN_GR_NAME_FIELD_WIDTH;
static int field_uid_width = MIN_UID_FIELD_WIDTH;
static int field_gid_width = MIN_GID_FIELD_WIDTH;
static int field_linenum_width = 1;

/*
 * Determine the blocksize to use for reporting the space usage displayed for
 * a directory with the -l (ell) flag or for each file with the -s flag.
 *
 * In order of increasing priority, this will be:
 * 1) the program default (DEFAULT_LONG_BLOCKSIZE)
 * 2) a value selected by environment variable LF_BLOCKSIZE
 * 3) a value selected by environment variable BLOCKSIZE
 * 4) a value implicitly selected by the last appearance on the command line
 *    of: -h, -k, --blocksize=VALUE (any of which can be modified by --bssi)
 * 5) --bssi always selects SI units; i.e., powers of 1000 instead of 1024
 *
 * Note: though not documented on FreeBSD within ls(1), the command apparently
 * uses getbsize(3) to determine the "preferred block size" from the
 * environment variable BLOCKSIZE set to a string of the form:
 *   [unsigned-int][K,k,M,m,G,g].
 * This specification will be recognized.
 *
 * GNU ls(1) uses --block-size=SIZE, where SIZE is an integer and optional
 * unit (example: "10M" is 10*1024*1024). Units are: K, M, G, T, P, E, Z, Y
 * (powers of 1024) or KB, MB, ... (powers of 1000).  No environment
 * variable akin to BLOCKSIZE is documented.
 * The --block-size syntax will also be recognized.
 */
static unsigned int
get_preferred_block_size(void)
{
  unsigned int bs;
  unsigned long block_unit, ipart, spart;
  char *msg, *ptr, *spec;

#ifdef DACS_OS_FREEBSD
  if (0) {
	int i;
	long l;
	msg = getbsize(&i, &l);
	fprintf(stderr, "msg=\"%s\" (%d, %ld)\n", msg, i, l);
  }
#endif

  /* Use this if nothing else is valid. */
  bs = DEFAULT_LONG_BLOCKSIZE;

  switch (which_bs_flag) {
  case BS_K_UNITS:
	if (bs_si_flag)
      bs = 1000;
    else
      bs = 1024;;
	break;

  case BS_BS_UNITS:
  case BS_DEFAULT_UNITS:
	if (which_bs_flag == BS_DEFAULT_UNITS) {
	  /* If there is a specification in the environment, use it. */
      if ((spec = getenv("LF_BLOCKSIZE")) == NULL
	  && (spec = getenv("BLOCKSIZE")) == NULL)
	  break;
    }
    else
	  spec = blocksize_units_flag;

	block_unit = 512;
	msg = NULL;

	/* ipart = integer part, spart = suffix part */
	ipart = spart = 1;

	/* Extract the integer part. */
	ptr = spec;
	if (isdigit((int) *spec)) {
	  ipart = strtoul(spec, &ptr, 10);
	  if (errno == ERANGE || errno == EINVAL) {
		msg = "Invalid blocksize integer";
		goto err;
	  }
	  else
		spec = ptr;
	}

	/* Extract the suffix, if any. */
	if (tolower((int) *spec) == 'k') {
	  spec++;
	  if (tolower((int) *spec) == 'b' && *(spec + 1) == '\0') {
		spart = 1000;			/* 10^3 */
		block_unit = 1000;
	  }
	  else if (*spec == '\0') {
		if (bs_si_flag) {
		  spart = 1000;			/* 10^3 */
		  block_unit = 1000;
	    }
	    else
		  spart = 1024;			/* 2^10 */
	  }
	  else {
		msg = "Invalid blocksize spec";
		goto err;
	  }
	}
	else if (tolower((int) *spec) == 'm') {
	  spec++;
	  if (tolower((int) *spec) == 'b' && *(spec + 1) == '\0') {
		spart = 1000000;			/* 10^6 */
		block_unit = 1000;
	  }
	  else if (*spec == '\0') {
		if (bs_si_flag) {
		  spart = 1000000;			/* 10^6 */
		  block_unit = 1000;
		}
		else
		  spart = 1048576;			/* 2^20 */
	  }
	  else {
		msg = "Invalid blocksize spec";
		goto err;
	  }
	}
	else if (tolower((int) *spec) == 'g') {
	  spec++;
	  if (tolower((int) *spec) == 'b' && *(spec + 1) == '\0') {
		spart = 1000000000;		/* 10^9 */
		block_unit = 1000;
	  }
	  else if (*spec == '\0') {
		if (bs_si_flag) {
		  spart = 1000000000;		/* 10^9 */
		  block_unit = 1000;
		}
		else
		  spart = 1073741824;		/* 2^30 */
	  }
	  else {
		msg = "Invalid blocksize spec";
		goto err;
	  }
	}
	else if (*spec != '\0') {
	  msg = "Invalid blocksize spec";

	err:
	  if (msg != NULL)
		fprintf(stderr, "%s\n", msg);
	  ipart = DEFAULT_LONG_BLOCKSIZE;
	  spart = 1;
	}

    if ((ipart * spart) <= block_unit)
	  bs = DEFAULT_LONG_BLOCKSIZE;
    else
	  bs = ((ipart * spart) / block_unit) * block_unit;
	break;

  case BS_HUMAN_UNITS:
    /* The units will be variable and indicated by a suffix. */
    user_spec_blocksize = DEFAULT_LONG_BLOCKSIZE;
	break;
		
  default:
    /* Something is wrong so revert to the default. */
	fprintf(stderr, "Internal error\n");
    user_spec_blocksize = DEFAULT_LONG_BLOCKSIZE;
	break;
  }

  return(bs);
}

#ifdef NOTDEF
/*
 * Return the effective blocksize that is used with the -l (ell) and
 * -s flags when computing the space usage of a directory or file.
 * For FreeBSD, this number will always be a multiple of 512 (BLOCK_UNIT) and
 * sizes less than 512 bytes are rounded to 512 bytes and sizes greater than
 * 1G bytes are rounded down to 1G bytes.
 * For GNU ls(1), the blocksize can be certain powers of 10, so we will
 * round to multiples of 1000 instead of 512.
 *
 * On FreeBSD, see the desscription of BLOCKSIZE in the ENVIRONMENT section
 * of ls(1) and getbsize(3).
 */
static unsigned int
get_user_spec_blocksize(unsigned int value, unsigned int block_unit)
{
  unsigned int bs;

  if (value <= block_unit)
	bs = DEFAULT_LONG_BLOCKSIZE;
  else
	bs = (value / block_unit) * block_unit;

  return(bs);
}
#endif

#include <sys/param.h>
#include <sys/mount.h>

/*
 * Return the blocksize (allocation size) for the filesystem on which
 * PATH resides, or 0 if that is unknown.
 */
static unsigned int
filesystem_blocksize(const char *path)
{
	struct statfs sfs;

	if (statfs(path, &sfs) != -1) {
		fprintf(stderr, "f_bsize=%u\n", (unsigned int) sfs.f_bsize);
		return((unsigned int) sfs.f_bsize);
	}

	return(0);
}

/*
 * Return the number of blocks (as defined by user_spec_blocksize) used by
 * the file associated with STATP.  This number may be zero, otherwise
 * it may be rounded up to count a partial block.
 * This is used for blocks_flag (-s) output.
 */
static unsigned long
filesystem_blocks_used(struct stat *statp)
{
  unsigned long nb;

  if (statp->st_size == 0)
	return(0);

  nb = (statp->st_blocks * 512) / user_spec_blocksize;

  return(nb);
}

#include <curses.h>
#include <term.h>

static size_t
get_maxlen(void *base, int num, size_t (*length)(void *, int), int *which)
{
  int i;
  size_t max;
  File_d **dp;

  max = 0;
  if (which != NULL)
	*which = -1;

  dp = (File_d **) base;
  for (i = 0; i < num; i++) {
	File_d *d;

	d = dp[i];
	if (d->fts_pathlen > max) {
	  max = d->fts_pathlen;
	  if (which != NULL)
		*which = i;
	}
  }

  return(max);
}

static int
get_tty_columns(void)
{
  int co;
  char *term, tbuf[2024];

  if ((term = getenv("TERM")) == NULL)
	return(0);

  if (tgetent(tbuf, term) == ERR)
	return(0);

  co = tgetnum("co");
  if (co == ERR)
	return(-1);

  return(co);
}

static int
cfun(char ch, void *farg)
{

  return(1);
}

/*
 * Routine to print a table
 * Modified from 'ls.c' mods (BJB/1983)
 * Arguments:
 *  base    - address of first entry
 *  num     - number of entries
 *  d_cols  - number of columns to use if > 0, "best" size if == 0
 *  width   - max line width if not zero
 *  prentry - function to print the string
 *  length  - function  to determine the length of string to be printed 
 *
 * prtable and length are called with the the address of the base and
 * an index
 */
static void
prtable(void *base, int num, int d_cols, int width,
		int (*prentry)(void *, int, void *), void *prentry_arg,
		size_t (*length)(void *, int))
{
  int c, j;
  int a, b, cols, loc, nrows, z;
  size_t maxlen;

  if (num == 0)
	return;

  maxlen = get_maxlen(base, num, length, NULL) + 1;
  if (number_flag)
	maxlen += field_linenum_width + 1;
  if (blocks_flag)
	maxlen += field_blocks_width + 1;
  if (inum_flag)
	maxlen += field_ino_width + 1;

  if (filetype_flag)
	maxlen++;

  if (d_cols > 0)
	cols = d_cols;
  else if (width == 0)
	cols = get_tty_columns() / maxlen;
  else
	cols = width / maxlen;

  if (cols == 0)
	cols = 1;

  nrows = (num - 1) / cols + 1;

  if (verbose_flag) {
	fprintf(stderr,
			"prtable: maxlen=%ld, cols=%d, nrows=%d, num=%d, d_cols=%d, width=%d\n",
			maxlen, cols, nrows, num, d_cols, width);
  }

  for (a = 1; a <= nrows; a++) {
	b = c = z = loc = 0;
	for (j = 0; j < num; j++) {
	  c++;
	  if (c >= a + b)
		break;
	}

	while (j < num) {
	  int len;

	  len = (*prentry)(base, j, prentry_arg);
	  if (len > 0)
		loc += len;
	  z++;
	  b += nrows;
	  for (j++; j < num; j++) {
		c++;
		if (c >= a + b)
		  break;
	  }

	  if (j < num) {
		while (loc < (z * maxlen)) {
		  printf(" ");
		  loc++;
		}
	  }
	}
	printf("\n");
  }
}

static size_t
name_len(void *base, int num)
{
  size_t len;
  File_d *d, **dp;

  dp = (File_d **) base;
  d = dp[num];
  len = strlen(d->fts_path);

  return(len);
}

/*
 * Search for NAME in colon-separated directory search path DIR_LIST,
 * terminating when it finds the first occurrence that satisfies MODE, which
 * is a bitwise-inclusive OR of the access permissions to be checked, as with
 * access(2): R_OK, W_OK, X_OK, or F_OK.
 * Two consecutive colons are ignored as a search path prefix.
 *
 * If found, return an allocated buffer containing its path, or NULL if
 * nothing was found.  The caller is responsible for freeing the buffer.
 * The returned path does not reveal anything non-essential about DIR_LIST
 * (i.e., the caller's caller is not given any information about DIR_LIST).
 *
 * The arguments are not modified.
 * Note that a colon in DIR_LIST cannot be escaped.
 *
 * Example: file_path_search("foo", "a::b:c", R_OK) will return the first
 * of "a/foo", "b/foo", "c/foo" that exists and is readable, or NULL.
 */
char *
file_path_search(char *name, char *dir_list, int mode)
{
  char *buf, *s;
  size_t buflen, dir_listlen, namelen;

  if (name == NULL || *name == '\0' || dir_list == NULL || *dir_list == '\0')
	return(NULL);

  /* Allocate a buffer for the longest possible matching path. */
  namelen = strlen(name);
  dir_listlen = strlen(dir_list);

  /* Include the slash and null bytes. */
  buflen = namelen + 1 + dir_listlen + 1;
  buf = (char *) malloc(buflen);

  if (verbose_flag > 1)
	fprintf(stderr, "file_path_search: searching \"%s\" for \"%s\"\n",
			dir_list, name);

  s = dir_list;
  while (s != NULL && *s != '\0') {
	size_t dirlen;
	char *e;

	if ((e = strchr(s, (int) ':')) != NULL) {
	  if ((dirlen = e - s) != 0) {
		strncpy(buf, s, dirlen);
		buf[dirlen] = '/';
		strncpy(buf + dirlen + 1, name, namelen);
		buf[dirlen + 1 + namelen] = '\0';
	  }
	  e++;
	}
	else {
	  /* The last directory path component does not end with a separator. */
	  dirlen = dir_listlen - (s - dir_list);
	  strncpy(buf, s, dirlen);
	  buf[dirlen] = '/';
	  strncpy(buf + dirlen + 1, name, namelen);
	  buf[dirlen + 1 + namelen] = '\0';
	}

	/* If the search component was the empty string, ignore it. */
	if (dirlen) {
	  if (verbose_flag > 1)
		fprintf(stderr, "file_path_search: checking \"%s\"\n", buf);
	  if (access(buf, mode) == 0)
		return(buf);
	}

	s = e;
  }

  /* Nothing was found. */
  free(buf);

  return(NULL);
}

/* For debugging. */
static int
is_directory(mode_t m)
{

  if (S_ISDIR(m))
	return(1); 

  return(0);
}

static int
is_regular(mode_t m)
{

  if (S_ISREG(m))
	return(1); 

  return(0);
}

static int
is_slink(mode_t m)
{

  if (S_ISLNK(m))
	return(1); 

  return(0);
}

/*
 * If non-NULL, use PATHNAME as the pathname for the RGB colour specification
 * file (in the X11 format, called rgb.txt by convention).
 * If PATHNAME is NULL, look for the first occurrence by:
 * 1) trying the pathname given by the RGBPATH environment variable
 * 2) searching for rgb.txt in the colon-separated list of dictories given
 *    by the RGBSEARCHPATH environment variable
 * 3) looking for rgb.txt in the directory given by the HOME environment variable
 * 4) searching a default colon-separated list (DEFAULT_RGB_SEARCH_PATH) for the
 *    first occurrence of a directory that contains rgb.txt.
 *
 * Any directory may be absolute or relative to the current working directory.
 *
 * Return an allocated copy of the path for the colour spec file, or NULL.
 */ 
static char *
colour_config_rgb_path(char *pathname)
{
  char *path, *pn;

  if (pathname != NULL) {
	if (access(pathname, R_OK) == 0)
	  return(strdup(pathname));

	return(NULL);
  }
 
  if ((pn = getenv("RGBPATH")) != NULL) {
	if (access(pn, R_OK) != -1)
	  return(strdup(pn));
  }

  if ((pn = getenv("RGBSEARCHPATH")) != NULL) {
	if ((path = file_path_search("rgb.txt", pn, R_OK)) != NULL)
	  return(path);
  }

  if ((pn = getenv("HOME")) != NULL) {
	asprintf(&path, "%s/%s", pn, COLOUR_RGB_FILENAME);
	if (access(path, R_OK) == 0)
	  return(path);
  }

  path = file_path_search("rgb.txt", DEFAULT_RGB_SEARCH_PATH, R_OK);

  return(path);
}

static int
colour_load_rgb_file(char *path)
{
  int i, n;
  FILE *fp;
  static char buf[BUFSIZ];

  if (path == NULL || (fp = fopen(path, "r")) == NULL)
	return(-1);

  for (i = 0; i < MAX_COLOURS; i++)
	colour_rgb[i] = NULL;

  n = 0;
  while (fgets(buf, BUFSIZ, fp) != NULL) {
	int nfields;
	char *p;
	Colour_rgb *desc;

	p = buf;
	/* Skip leading whitespace. */
	while (isspace((int) *p))
	  p++;

	/* Ignore comments. */
	if (*p == '#' || *p == '\0')
	  continue;

	if (n == MAX_COLOURS)
	  goto err;

	desc = ALLOC(Colour_rgb);
	desc->name = (char *) malloc(strlen(p) + 1);
	nfields = sscanf(p, "%hhu %hhu %hhu %s",
					 &desc->red, &desc->green, &desc->blue, desc->name);
	if (nfields != 4)
	  goto err;

	/* XXX maybe also elide spaces here. */
	strlowerbuf(desc->name, desc->name);
	desc->desc = strdup(p);

	colour_rgb[n++] = desc;
  }
  
  if (feof(fp)) {
	fclose(fp);
	return(n);
  }

 err:
  fclose(fp);
  return(-1);	
}

static Colour_rgb *
colour_lookup(char *name)
{
  int i;
  Colour_rgb *rgb;

  if (name == NULL)
	return(NULL);

  /*
   * If NAME starts with a '#', it is expected to be an rgb tupple, with
   * the numbers separated by a semicolon.
   * The static buffer is reused.
   */
  if (*name == '#') {
	static Colour_rgb desc;

	if (sscanf(name + 1, "%hhu;%hhu;%hhu",
			   &desc.red, &desc.green, &desc.blue) != 3)
	  return(NULL);
	desc.desc = strdup(name);
	desc.name = strdup(name + 1);

	return(&desc);
  }

  for (i = 0; colour_rgb[i] != NULL; i++) {
	rgb = colour_rgb[i];
	if (strcaseeq(rgb->name, name))
	  return(rgb);
  }

  return(NULL);
}

#include "glob.h"

/*
 * If filename FN matches a user-configured pattern, return the
 * associated colour rendering string.
 */
static char *
colour_effect_lookup_filename(char *fn)
{
  int flags, i, st;
  extern int globmatch(const char * __restrict str,
					   const char * __restrict pattern, int flags);

  for (i = 0; i < colour_effect_glob_count; i++) {
	flags = GLOB_NOSORT | GLOB_BRACE;
	st = globmatch(fn, colour_effect_glob[i].glob, flags);
	if (st == 1)
	  return(colour_effect_glob[i].effect_arg.effect_str);
  }

  return(NULL);
}

/*
 * Return the colour rendering string to use for the file named FN with mode M,
 * or NULL if there is nothing apropos.
 */
static char *
colour_effect_lookup(char *fn, mode_t m)
{
  char *e;

  if ((e = colour_effect_lookup_filename(fn)) != NULL)
	return(e);

  if (S_ISDIR(m))
	return(colour_effect[COLOUR_EFFECT_DIR].effect_str);

  if (m & (S_IXUSR | S_IXGRP | S_IXOTH))
	return(colour_effect[COLOUR_EFFECT_EXEC].effect_str);

  if (S_ISLNK(m))
	return(colour_effect[COLOUR_EFFECT_LINK].effect_str);

  if (S_ISREG(m)) {
	if (fn != NULL && fn[0] == '.')
	  return(colour_effect[COLOUR_EFFECT_DOT].effect_str);
	else
	  return(colour_effect[COLOUR_EFFECT_REGULAR].effect_str);
  }

  if (S_ISSOCK(m))
	return(colour_effect[COLOUR_EFFECT_SOCK].effect_str);

  if (S_ISFIFO(m))
	return(colour_effect[COLOUR_EFFECT_FIFO].effect_str);

#ifdef S_ISWHT
  if (S_ISWHT(m))
	return(colour_effect[COLOUR_EFFECT_WHITEOUT].effect_str);
#endif

  return(NULL);
}

/*
 * Copy into BUF the string to be emitted to render text in COLOUR
 * with FLAGS modifiers.
 *
 * Return that buffer.
 */
static char *
colour_effect_str(char *buf, Tty_op flags, char *colour)
{
  int len;
  char *p;

  if (flags == TTY_RESET_OP) {
	strcpy(buf, ANSI_TTY_RESET_STR);
	return(buf);
  }

  len = 0;
  p = buf;
  if (colour != NULL) {
	Colour_rgb *d;

	/*
	 * If COLOUR is not recognized, make a few basic built-ins available.
	 */
	if ((d = colour_lookup(colour)) != NULL)
	  len = sprintf(p, "\033[38;2;%d;%d;%d", d->red, d->green, d->blue);
	else if (streq(colour, "black"))
	  len = sprintf(p, "\033[38;2;0;0;0");
	else if (streq(colour, "red"))
	  len = sprintf(p, "\033[31");
	else if (streq(colour, "green"))
	  len = sprintf(p, "\033[32");
	else if (streq(colour, "yellow"))
	  len = sprintf(p, "\033[33");
	else if (streq(colour, "blue"))
	  len = sprintf(p, "\033[34");
	else if (streq(colour, "magenta"))
	  len = sprintf(p, "\033[35");
	else if (streq(colour, "cyan"))
	  len = sprintf(p, "\033[36");
	else if (streq(colour, "white"))
	  len = sprintf(p, "\033[37");
	else if (streq(colour, "orange"))
	  len = sprintf(p, "\033[38;2;255;138;0");
	else {
	  len = sprintf(p, "\033[38;2;0;0;0");
	  fprintf(stderr, "Colour \"%s\" is undefined\n", colour);
	}
	p += len;
  }

  if ((flags & TTY_BOLD_OP) != 0) {
	if (len)
	  len = sprintf(p, ";1");
	else
	  len = sprintf(p, "\033[1");
	p += len;
  }

  if ((flags & TTY_UNDERLINE_OP) != 0) {
	if (len)
	  len = sprintf(p, ";4");
	else
	  len = sprintf(p, "\033[4");
	p += len;
  }

  if (len)
	sprintf(p, "m");

  return(buf);
}

/*
 * Given the colour specification in E->ARG, initialize the corresponding
 * rendering string in E->EFFECT_STR.
 */
static void
colour_compile_effect(Colour_effect *e)
{
  int cflags;
  char *cstr;

  if (e->arg == NULL)
	return;

  if (streq(e->arg, "bold")) {
	cflags = TTY_BOLD_OP;
	cstr = NULL;
  }
  else if (streq(e->arg, "underline")) {
	cflags = TTY_UNDERLINE_OP;
	cstr = NULL;
  }
  else {
#ifdef DEFAULT_TTY_BOLD_OP
	cflags = TTY_BOLD_OP;
#else
	cflags = TTY_PLAIN_OP;
#endif
	cstr = e->arg;
  }

  /* XXX */
  e->effect_str = (char *) malloc(32);
  colour_effect_str(e->effect_str, cflags, cstr);
}

static void
colour_compile_effects(void)
{
  int i;

  for (i = 0; i < COLOUR_EFFECT_NTYPES; i++) {
	colour_compile_effect(&colour_effect[i]);
  }
}

/*
 * Text colourization configuration.
 * Attempt to recognize and process one colour specification option FLAG
 * with non-NULL argument ARG.
 * Any existing specification is overridden by a subsequent one,
 * so default specs, then environment specs, then command line specs should
 * be processed in that order.
 *
 * Return values:
 *   0 if not passed a recognized colour config argument.
 *  -1 if passed an invalid colour config argument. *
 *  1 if successful and a single flag was seen.
 *  2 if successful and a flag with one parameter was seen.
 *  3 if successful and a flag with two parameters was seen.
 */
static int
colour_parse_flag_arg(char *args[])
{
  int rc;
  char *a, *flag, *p;

  /* This is not a colour-related argument. */
  if (args == NULL
	  || (flag = args[0]) == NULL || flag[0] != '-' || flag[1] != 'C')
	return(0);

  if (colours_enabled == -1)
	return(0);

  /* First, flags that take no arguments. */
  if (streq(flag, "-Ca")) {
	colour_line_flag = 1;
	return(1);
  }

  if (streq(flag, "-Cl")) {
	colour_long_only_flag = 1;
	return(1);
  }

  if (streq(flag, "-Coff")) {
	/*
	 * Force text colourization off, regardless of any other config.
	 * Useful for disabling specifications passed in the environment.
	 */
	colours_enabled = -1;
	return(1);
  }

  if (streq(flag, "-Con")) {
	/* Force text colourization on, regardless of any other config. */
	colours_enabled = 0;
	colours_forced_flag = 1;
	return(1);
  }
 
  /* Next, flags that take one or two arguments. */
  if (args[1] == NULL)
	return(-1);

  a = strdup(args[1]);
  if (streq(flag, "-C")) {
	colour_effect[COLOUR_EFFECT_DEFAULT_ALL].arg = a;
	return(2);
  }

  if (streq(flag, "-Cp")) {
	rgbpath_flag = a;
	return(2);
  }

  /*
   * Colour selection by glob matching the filename.
   * Can be repeated.
   */
  if (streq(flag, "-Cg")) {
	Colour_effect_glob *ceg;

	if (args[2] == NULL)
	  return(-1);

	if (colour_effect_glob_count == colour_effect_glob_nalloc) {
	  colour_effect_glob_nalloc += 10;
	  colour_effect_glob
		= (Colour_effect_glob *) realloc(colour_effect_glob,
										 sizeof(Colour_effect_glob)
										 * colour_effect_glob_nalloc);
	}
	ceg = &colour_effect_glob[colour_effect_glob_count++];
	ceg->glob = a;
	ceg->effect_arg.arg = strdup(args[2]);
	ceg->effect_arg.effect_str = NULL;
	colour_compile_effect(&ceg->effect_arg);

	return(3);
  }

  /*
   * One or more of the file type selectors are expected.
   * Example: "-C/. red"
   * specifies red for both the directory and dot file types.
   */
  rc = 2;
  for (p = flag + 2; rc == 2 && *p != '\0'; p++) {
	switch ((int) *p) {
	case '+':
	  colour_effect[COLOUR_EFFECT_DEFAULT_SPECIAL].arg = a;
	  break;

	case '-':
	  colour_effect[COLOUR_EFFECT_REGULAR].arg = a;
	  break;

	case '/':
	  colour_effect[COLOUR_EFFECT_DIR].arg = a;
	  break;

	case '*':
	  colour_effect[COLOUR_EFFECT_EXEC].arg = a;
	  break;

	case '=':
	  colour_effect[COLOUR_EFFECT_SOCK].arg = a;
	  break;

	case '|':
	  colour_effect[COLOUR_EFFECT_FIFO].arg = a;
	  break;

	case '.':
	  colour_effect[COLOUR_EFFECT_DOT].arg = a;
	  break;

	case '%':
	  colour_effect[COLOUR_EFFECT_WHITEOUT].arg = a;
	  break;

	case 'x':
	  colour_effect[COLOUR_EFFECT_UNPRINTABLE].arg = a;
	  break;

	default:
	  rc = -1;
	  break;
	}
  }

  if (rc == 2) {
	/*
	 * XXX This signals that an RGB colour definition file is to be loaded,
	 * but assuming this here is pessimistic.
	 */
	colours_enabled = 1;
  }

  return(rc);
}

/*
 * Process all colour specifications given by STR, which are zero or more
 * space-separated flag/parameters recognized on the command line.
 * Return the number of flags successfully processed, or -1 if an error occurs.
 *
 * XXX quotes are not recognized
 */
static int
colour_parse_flag_str(char *str)
{
  int n;
  char *p, *sflag, *sarg1;

  if (str == NULL)
	return(0);

  /* Break STR into space-separated tokens, then process each pair. */
  p = strdup(str);
  n = 0;
  while (*p != '\0' && *p != '\n') {
	/* Skip leading space. */
	while (*p == ' ')
	  p++;

	/* Flag starts... */
	sflag = p;
	while (*p != ' ' && *p != '\n' && *p != '\0')
	  p++;
	if (*p != ' ')
	  break;

	/* End of the flag. */
	if (*p != '\0')
	  *p++ = '\0';

	/* Start of the parameter... */
	sarg1 = p;
	while (*p != ' ' && *p != '\n' && *p != '\0')
	  p++;
	/* End of the parameter. */
	if (*p != '\0')
	  *p++ = '\0';
	
	/* XXX kludge... this is the only flag that has two parameters. */
	if (streq(sflag, "-Cg")) {
	  char *args[4], *sarg2;

	  /* Skip leading space. */
	  while (*p == ' ')
		p++;

	  sarg2 = p;
	  while (*p != ' ' && *p != '\n' && *p != '\0')
		p++;
	  /* End of the parameter. */
	  if (*p != '\0')
		*p++ = '\0';

	  args[0] = sflag;
	  args[1] = sarg1;
	  args[2] = sarg2;
	  args[3] = NULL;
	  if (colour_parse_flag_arg(args) == -1)
		return(-1);
	}
	else {
	  char *args[3];

	  args[0] = sflag;
	  args[1] = sarg1;
	  args[2] = NULL;
	  if (colour_parse_flag_arg(args) == -1)
		return(-1);
	}

	n++;
  }

  return(n);
}

/*
 * Add the interesting bits from F to a linked list.
 */
static File_d *
add_file_d(FTSENT *f)
{
  struct stat *s;
  File_d *d;

  if (verbose_flag > 1) {
	printf("   L%d, ", f->fts_level);

	if (f->fts_info == FTS_D)
	  printf("FTS_D");
	else if (f->fts_info == FTS_DC)
	  printf("FTS_DC");
	else if (f->fts_info == FTS_DEFAULT)
	  printf("FTS_DEFAULT");
	else if (f->fts_info == FTS_DNR)
	  printf("FTS_DNR (%s)", strerror(f->fts_errno));
	else if (f->fts_info == FTS_DOT)
	  printf("FTS_DOT");
	else if (f->fts_info == FTS_DP)
	  printf("FTS_DP");
	else if (f->fts_info == FTS_F)
	  printf("FTS_F");
	else if (f->fts_info == FTS_NS)
	  printf("FTS_NS (%s)", strerror(f->fts_errno));
	else if (f->fts_info == FTS_NSOK)
	  printf("FTS_NSOK");
	else if (f->fts_info == FTS_SL)
	  printf("FTS_SL");
	else if (f->fts_info == FTS_SLNONE)
	  printf("FTS_SLNONE");
	else if (f->fts_info == FTS_ERR)
	  printf("FTS_ERR (%s)", strerror(f->fts_errno));
	else
	  printf("FTS_???");

	if (f->fts_info == FTS_NS || f->fts_info == FTS_NSOK)
	  printf(" (nostat)");

	printf("\n");
  }

  d = ALLOC(File_d);
  d->fts_path = strdup(f->fts_path);
  d->fts_pathlen = f->fts_pathlen;
  d->fts_namelen = f->fts_namelen;
  d->fts_level = f->fts_level;
  if (d->fts_pathlen > path_name_max)
	path_name_max = d->fts_pathlen;

  s = d->fts_statp = f->fts_statp;

  if (s != NULL) {
	struct group *g;
	struct passwd *u;

	d->fts_statp = ALLOC(struct stat);
	*d->fts_statp = *s;

	/* Maintain maximum values for column width calculations. */
	if (s->st_dev > st_dev_max)
	  st_dev_max = s->st_dev;
	if (s->st_rdev > st_rdev_max)
	  st_rdev_max = s->st_rdev;
	if (s->st_nlink > st_nlink_max)
	  st_nlink_max = s->st_nlink;
	if (s->st_ino > st_ino_max)
	  st_ino_max = s->st_ino;
	if (s->st_size > st_size_max)
	  st_size_max = s->st_size;
	if (s->st_uid > st_uid_max)
	  st_uid_max = s->st_uid;
	if (s->st_gid > st_gid_max)
	  st_gid_max = s->st_gid;

	/* XXX only need to bother with this if usernames are printed. */
	errno = 0;
	if ((u = getpwuid(s->st_uid)) != NULL) {
	  size_t len;

	  len = strlen(u->pw_name);
	  if (len > pw_name_max)
		pw_name_max = len;
	}

	/* XXX only need to bother with this if groupnames are printed. */
	if ((g = getgrgid(s->st_gid)) != NULL) {
	  size_t len;

	  len = strlen(g->gr_name);
	  if (len > gr_name_max)
		gr_name_max = len;
	}
  }

  if (file_set == NULL) {
	size_t need;

	need = 512;
	file_set = (File_d **) malloc(sizeof(File_d *) * need);
	nfiles_alloc = need;
  }
  else if (nfiles == nfiles_alloc) {
	size_t need;

	need = nfiles_alloc * 2;
	file_set = (File_d **) realloc(file_set, sizeof(File_d *) * need);
	if (file_set == NULL)
	  exit(1);
	nfiles_alloc = need;
  }

  file_set[nfiles++] = d;

  return(d);
}

static void
config_field_widths(void)
{
  int w, wb;

  w = (int) log10((double) st_dev_max) + 1;
  if (w > field_dev_width)
	field_dev_width = w;

  w = (int) log10((double) st_rdev_max) + 1;
  if (w > field_rdev_width)
	field_rdev_width = w;

  w = (int) log10((double) st_nlink_max) + 1;
  if (w > field_nlink_width)
	field_nlink_width = w;

  w = (int) log10((double) st_ino_max) + 1;
  if (w > field_ino_width)
	field_ino_width = w;

  w = (int) log10((double) st_size_max) + 1;
  if (w > field_size_width) {
	field_size_width = w;
	wb = (int) log10((double) st_size_max / user_spec_blocksize) + 1;
	if (wb > field_blocks_width)
	  field_blocks_width = wb;
  }

  w = (int) log10((double) st_uid_max) + 1;
  if (w > field_uid_width)
	field_uid_width = w;

  w = (int) log10((double) st_gid_max) + 1;
  if (w > field_gid_width)
	field_gid_width = w;

  field_pw_name_width = pw_name_max;
  field_gr_name_width = gr_name_max;
}

/*
 * Compare for sorting, using only stat elements.
 */
static int
compar_stat(struct stat *sa, struct stat *sb)
{
  int st;

  switch (sort_type_flag) {
  case SORT_NEW_TO_OLD:
	if (sa->st_mtime < sb->st_mtime)
	  return(1);
	if (sa->st_mtime > sb->st_mtime)
	  return(-1);
	return(0);
	/*NOTREACHED*/

  case SORT_OLD_TO_NEW:
	if (sa->st_mtime > sb->st_mtime)
	  return(1);
	if (sa->st_mtime < sb->st_mtime)
	  return(-1);
	return(0);
	/*NOTREACHED*/
  
  case SORT_SMALL_TO_BIG:
	if (sa->st_size > sb->st_size)
	  return(1);
	if (sa->st_size < sb->st_size)
	  return(-1);
	return(0);
	/*NOTREACHED*/
  
  case SORT_BIG_TO_SMALL:
	if (sa->st_size < sb->st_size)
	  return(1);
	if (sa->st_size > sb->st_size)
	  return(-1);
	return(0);
	/*NOTREACHED*/

  case SORT_DECREASING_LINK_COUNT:
	if (sa->st_nlink < sb->st_nlink)
	  st = 1;
	else if (sa->st_nlink > sb->st_nlink)
	  st = -1;
	else
	  st = 0;
	/* XXX there is no SORT_INCREASING_LINK_COUNT flag... */
	if (reverse_sort_flag)
	  st = -st;

	return(st);
	/*NOTREACHED*/

  default:
	return(0);
	/*NOTREACHED*/
  }

  return(0);
}

/*
 * Compare for sorting, using filenames or stat elements.
 * May never use fts_accpath, fts_path, or fts_pathlen.
 */
static int
compar_fts(const void *aa, const void *bb)
{
  int st;
  const FTSENT *a, *b;
  struct stat *sa, *sb;

  a = *(FTSENT **) aa;
  b = *(FTSENT **) bb;

  sa = a->fts_statp;
  sb = b->fts_statp;

  if (sort_type_flag != SORT_FILENAME_AZ
	  && sort_type_flag != SORT_FILENAME_ZA
	  && sort_type_flag != SORT_FILENAME_Az
	  && sort_type_flag != SORT_FILENAME_zA) {
	/* If fts_info is FTS_NS or FTS_NSOK, fts_statp may not be used. */
	if (a->fts_info == FTS_NS || a->fts_info == FTS_NSOK
		|| b->fts_info == FTS_NS || b->fts_info == FTS_NSOK)
	  return(0);

	if (sa == NULL || sb == NULL)
	  return(0);
  }

  switch (sort_type_flag) {
  case SORT_FILENAME_AZ:
  case SORT_FILENAME_ZA:
	/* Put "." files ahead of (or behind) the others. */
	if (a->fts_path[0] == '.' && b->fts_path[0] != '.')
	  st = -1;
	else if (a->fts_path[0] != '.' && b->fts_path[0] == '.')
	  st = 1;
	else
	  st = strcmp(a->fts_name, b->fts_name);
	return((sort_type_flag == SORT_FILENAME_AZ) ? st : -st);
	/*NOTREACHED*/

  case SORT_FILENAME_Az:
  case SORT_FILENAME_zA:
	/* Put "." files ahead of (or behind) the others. */
	if (a->fts_path[0] == '.' && b->fts_path[0] != '.')
	  st = -1;
	else if (a->fts_path[0] != '.' && b->fts_path[0] == '.')
	  st = 1;
	else
	  st = strcasecmp(a->fts_name, b->fts_name);
	return((sort_type_flag == SORT_FILENAME_Az) ? st : -st);
	/*NOTREACHED*/

  default:
	break;
  }

  return(compar_stat(sa, sb));
}

#ifdef NOTDEF
  if (sortuid_flag) {
	if (sa->st_uid < sb->st_uid)
	  return(1);
	if (sa->st_uid > sb->st_uid)
	  return(-1);
	return(0);
  }
  else if (sortgid_flag) {
	if (sa->st_gid < sb->st_gid)
	  return(1);
	if (sa->st_gid > sb->st_gid)
	  return(-1);
	return(0);
  }
#endif

/*
 * For (re)sorting after all pathnames have been collected in file_set.
 */
static int
compar_merge(const void *fa, const void *fb)
{
  const File_d *a, *b;
  struct stat *sa, *sb;

  a = *(File_d **) fa;
  b = *(File_d **) fb;

  sa = a->fts_statp;
  sb = b->fts_statp;

  if (sa == NULL || sb == NULL)
	return(0);

  if (dirsfirst_flag) {
	if (S_ISDIR(sa->st_mode) && !S_ISDIR(sb->st_mode))
	  return(-1);
	if (!S_ISDIR(sa->st_mode) && S_ISDIR(sb->st_mode))
	  return(1);
  }


  else {
	return(compar_stat(sa, sb));
  }
#ifdef NOTDEF
  else if (sortuid_flag) {
	if (sa->st_uid < sb->st_uid)
	  return(1);
	if (sa->st_uid > sb->st_uid)
	  return(-1);
	return(0);
  }
  else if (sortgid_flag) {
	if (sa->st_gid < sb->st_gid)
	  return(1);
	if (sa->st_gid > sb->st_gid)
	  return(-1);
	return(0);
  }
#endif

  return(0);
}

/*
 * Return the number of "unprintable" characters in FN.
 * These are characters that would likely be problematic when printed.
 */
static size_t
count_unprintable_chars(char *fn)
{
  char *p;
  size_t count;

  count = 0;
  for (p = fn; *p != '\0'; p++) {
	if (!isprint((int) *p))
	  count++;
  }

  return(count);
}

/*
 * Depending on the particular shell being used, some/all of these characters
 * need to be escaped if a filename output by this program is going to be
 * stuffed onto a command line.
 * Also see: -Q flag (quoted_flag).
 */
static char *shell_escaped_chars = "`~!$&*(){}[]|\\;'\"<>? \t";

static int
is_shell_escaped_char(int ch)
{

  if (!shell_escape_flag)
	return(0);

  if (shell_escape_flag > 1 && !isalnum(ch))
	return(1);

  if (strchr(shell_escaped_chars, ch) != NULL)
	return(1);

  return(0);
}

/*
 *
 */
static size_t
count_shell_escaped_chars(char *fn)
{
  char *p;
  size_t count;

  if (!shell_escape_flag)
	return(0);

  count = 0;
  for (p = fn; *p != '\0'; p++) {
	if (is_shell_escaped_char((int) *p))
	  count++;
  }

  return(count);
}

/*
 * Return the number of whitespace characters in FN.
 */
static size_t
count_whitespace_chars(char *fn)
{
  char *p;
  size_t count;

  count = 0;
  for (p = fn; *p != '\0'; p++) {
	if (!isprint((int) *p))
	  count++;
  }

  return(count);
}

/*
 * Reformat every non-printable character in FN as \xxx, where xxx is the
 * character's octal value, otherwise copy it verbatim.
 * The caller should free the returned string, which may be a copy of the
 * original.
 *
 * Note that if FN actually contains a substring like "\223", that will
 * indistinguishable from the case where FN contains the byte with value
 * 0223 (octal).
 */
static char *
format_safe_filename(char *fn)
{
  size_t esc_count, unprintable_count;

  /* If unprintable characters are allowed, return an exact copy. */
  if (unprintable_filenames_flag)
	return(strdup(fn));

  /* If there are unprintables, escape them. */
  if ((unprintable_count = count_unprintable_chars(fn)) != 0) {
	char *p, *q, *sfn;
	size_t need;

	/* The single byte 0234 becomes the four bytes "\234" */
	need = strlen(fn) + (unprintable_count * 3);
	sfn = q = (char *) malloc(need + 1);
	for (p = fn; *p != '\0'; p++) {
	  if (isprint((int) *p))
		*q++ = *p;
	  else {
		*q++ = '\\';
		*q++ = ((*p & 0200) >> 6) + '0';
		*q++ = ((*p & 0070) >> 3) + '0';
		*q++ = (*p & 0007) + '0';
	  }
	}

	*q = '\0';

	return(sfn);
  }

  /*
   * If there are printable but potentially problematic characters
   * escape each one.
   */
  if (shell_escape_flag
	  && (esc_count = count_shell_escaped_chars(fn)) != 0) {
	char *p, *q, *sfn;
	size_t need;

	/* Replace each problematic character with an escaped string. */
	need = strlen(fn) + esc_count;
	sfn = q = (char *) malloc(need + 1);
	for (p = fn; *p != '\0'; p++) {
	  if (is_shell_escaped_char((int) *p))
		*q++ = '\\';
	  *q++ = *p;
	}

	*q = '\0';

	return(sfn);
  }

  /* The argument was ok. */

  return(strdup(fn));
}

/*
 * These symbols follow ls(1) (-F flag) and IEEE Std 1003.1-2017.
 * Note that there is no indicator character for a regular file.
 * The standard permits the use of other symbols for other file types.
 */
static char *
file_type_ind(struct stat *statp)
{
  char *s;

  if (statp) {
	/* Check for a whiteout first. */
#ifdef S_ISWHT
	if (S_ISWHT(statp->st_mode))
	  s = "%";
#else
	if (0)
	  ;
#endif
	/* Then check for a link. */
	else if (S_ISLNK(statp->st_mode))
	  s = "@";
	/* Then check for a directory before executable. */
	else if (S_ISDIR(statp->st_mode))
	  s = "/";
	else if (statp->st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))
	  s = "*";
	else if (S_ISSOCK(statp->st_mode))
	  s = "=";
	else if (S_ISFIFO(statp->st_mode))
	  s = "|";
	else
	  s = NULL;
  }
  else
	s = NULL;

  return(s);
}

/*
 * Return 1 if STATP meets any one of the type specifications
 * expressed by SPEC (each specifier type is a single character, with
 * some synonyms), 0 otherwise.
 * Ignore unrecognized specifier types.
 *
 * Specifier characters (conflated into SPEC):
 * block device:       b
 * character device:   c
 * directory:          d, / (NOT directory:      D)
 * regular file:       f, r (NOT regular file:   F, R)
 * symbolic link:      l, @
 * hard linked:        n
 * FIFO/pipe:          p, |
 * socket:             s, =
 * regular executable: *, e (NOT regular executable: E)
 * whiteout:           % (when available)
 */
static int
file_type_match(char *spec, struct stat *statp)
{
  char *p;
  mode_t m;

  m = statp->st_mode;

  for (p = spec; *p != '\0'; p++) {
	switch (*p) {
	case 'b':
	  if (S_ISBLK(m))
		return(1);
	  break;

	case 'c':
	  if (S_ISCHR(m))
		return(1);
	  break;

	case 'd':
	case '/':
	  if (S_ISDIR(m))
		return(1);
	  break;

	case 'D':
	  if (!S_ISDIR(m))
		return(1);
	  break;

	case 'f':
	case 'r':
	  if (S_ISREG(m))
		return(1);
	  break;

	case 'F':
	case 'R':
	  if (!S_ISREG(m))
		return(1);
	  break;

	case 'l':
	case '@':
	  if (S_ISLNK(m))
		return(1);
	  break;

	case 'n':
	  /* Regular and more than one hard link, so not a directory. */
	  if (S_ISREG(m) && statp->st_nlink > 1)
		return(1);
	  break;

	case 'p':
	case '|':
	  if (S_ISFIFO(m))
		return(1);
	  break;

	case 's':
	case '=':
	  if (S_ISSOCK(m))
		return(1);
	  break;

	case '*':
	case 'e':
	  /* Regular and executable, so not a directory, link, etc. */
	  if (S_ISREG(m) && (m & (S_IXUSR | S_IXGRP | S_IXOTH)))
		return(1);
	  break;

	case 'E':
	  /* NOT (regular and executable). */
	  if (!(S_ISREG(m) && (m & (S_IXUSR | S_IXGRP | S_IXOTH))))
		return(1);
	  break;

#ifdef S_ISWHT
	case '%':
	  if (S_ISWHT(m))
		return(1);
	  break;
#endif

	default:
	  break;
	}
  }

  return(0);
}

static char *cwd = NULL;

static char *
get_hilite_colour(char *fn, struct stat *statp)
{
  char *effect_str;

  if ((effect_str = colour_effect_lookup(fn, statp->st_mode)) == NULL) {
	if (!S_ISREG(statp->st_mode))
	  effect_str = colour_effect[COLOUR_EFFECT_DEFAULT_SPECIAL].effect_str;
	if (effect_str == NULL)
	  effect_str = colour_effect[COLOUR_EFFECT_DEFAULT_ALL].effect_str;
  }

  return(effect_str);
}

/*
 * Format output for D, according to FMT_SPEC, as output line number LINENUM.
 * If DSP is NULL print the line to stdout, otherwise set DSP to a
 * buffer containing the formatted line - the caller is responsible for
 * calling ds_free() on the buffer.
 *
 * Return the printing length of the formatted line (the "visible line"),
 * which does not include the length of any escape sequences that might be
 * embedded in the line.
 * N.B. Be extra careful when generating the line because returning an
 * inaccurate line length will mess up formatting alignment.
 * For example, being able to highlight certain fields might be a nice
 * feature (e.g., the file type indicator or a file name extension), but
 * this would require rewriting the field composition code to compute the
 * visible line length correctly.
 *
 * When printing a command line argument that is a directory
 * the -1 flag (single_column_flag) is enabled and
 * output is formated as
 *   <directory-path>:
 * followed by the files within that directory, with
 * a blank line preceding <directory-path>: (except if the first output line),
 * but ONLY when these flags are unset: -d (onelevel_flag), -A (absolute_flag),
 * -S flag (prefix_path_flag).
 */
static int
format_line(char *fmt_spec, File_d *d, int linenum, Ds **dsp)
{
  int hilited, olen, suppress_filetype, unprintable_count;
  char bp[12], *effect_str, *fmt, *fts_path, *ind, *ofn;
  char *fn;
  struct group *g;
  struct passwd *u;
  struct stat *statp;
  Ds *ds;

  fmt = fmt_spec;
  suppress_filetype = 0;
  effect_str = NULL;

  fts_path = d->fts_path;
  statp = d->fts_statp;

  ds = ds_init(NULL);

  /* Get the name of the file to be listed. */
  ofn = NULL;
  if (fts_path[0] == '/') {
	ofn = strndup(fts_path + d->fts_pathlen - d->fts_namelen,
				  d->fts_namelen);
	fn = ofn;
  }
  else {
	/* Trim an initial "./", which is implicit. */
	if (fts_path[0] == '.' && fts_path[1] == '/' && fts_path[2] != '\0')
	  fn = fts_path + 2;
	else
	  fn = fts_path;

	if (prefix_path_flag != NULL)
	  asprintf(&ofn, "%s%s", prefix_path_flag, fn);
	else if (absolute_flag) {
	  if (cwd == NULL) {
		cwd = (char *) malloc(MAXPATHLEN + 1);
		getcwd(cwd, MAXPATHLEN + 1);
	  }
	  asprintf(&ofn, "%s/%s", cwd, fn);
	}
	else
	  asprintf(&ofn, "%s", fn);
  }

  if (number_flag)
	olen = ds_asprintf(ds, "%*d ", field_linenum_width, linenum);
  else
	olen = 0;

  /*
   * IEEE Std 1003.1 says that the -s flag is used to:
   * Indicate the total number of file system blocks consumed by each file
   * displayed. The block size is implementation-defined.
   * Also, this value should appear first on the output line followed by
   * a space.
   */
  if (blocks_flag) {
	if (statp)
	  olen += ds_asprintf(ds, "%*lu ",
					   field_blocks_width, filesystem_blocks_used(statp));
	else
	  olen += ds_asprintf(ds, "%*s ", field_blocks_width, " ");
  }

  if (inum_flag) {
	if (statp)
	  olen += ds_asprintf(ds,
					   "%-*lu ", field_ino_width, (unsigned long) statp->st_ino);
	else
	  olen += ds_asprintf(ds, "%*s ", field_ino_width, " ");
  }

  while (*fmt != '\0') {
	hilited = 0;
	if (colour_line_flag) {
	  hilited = 1;
	  effect_str = get_hilite_colour(fn, statp);
	}

	if (*fmt == '%') {
	  fmt++;

	  /* XXX should recognize an explicit integer field width here. */
		 
	  if (*fmt == '+') {
		fmt++;
		hilited = 1;
		effect_str = get_hilite_colour(fn, statp);
	  }

	  if (hilited && effect_str != NULL)
		ds_asprintf(ds, "%s", effect_str);

	  switch (*fmt) {
		int show_secs;
		char buf[32];
		size_t st;
		time_t now, t;
		struct tm *tm, *tm_now, *tm_t;

	  case 'b':
		if (statp)
		  olen += ds_asprintf(ds, "%-*lu ",
						 field_blocks_width, filesystem_blocks_used(statp));
		else
		  olen += ds_asprintf(ds, "%*s ", field_blocks_width, " ");
		break;

	  case 'c':
		olen += ds_asprintf(ds, "%*d", field_linenum_width, linenum);
		break;

	  case 'd':
		/*
		 * Date/Time.
		 * The next character is a selector that must be present.
		 * 'a': st_atime:     last access time
		 * 'b': st_birthtime: creation time (non-POSIX)
		 * 'c': st_ctime:     last inode change time
		 * 'd': respect the time_type_flag
		 * 'm': st_mtime:     last modification time
		 */
		show_secs = 0;
		if (statp) {
		  char nextch;

		  /* Modifier 's' requests seconds-since-the-epoch. */
		  if ((nextch = *(fmt + 1)) == 's') {
			show_secs = 1;
			fmt++;
			nextch = *(fmt + 1);
		  }

		  if (nextch == 'a') {
			t = statp->st_atime;
			fmt++;
		  }
#ifdef DACS_OS_FREEBSD
		  else if (nextch == 'b') {
			t = statp->st_birthtime;
			fmt++;
		  }
#endif
		  else if (nextch == 'c') {
			t = statp->st_ctime;
			fmt++;
		  }
		  else if (nextch == 'd') {
			if (time_type_flag == TIME_ATIME)
			  t = statp->st_atime;
			else if (time_type_flag == TIME_MTIME)
			  t = statp->st_mtime;
			else if (time_type_flag == TIME_CTIME)
			  t = statp->st_ctime;
#ifdef DACS_OS_FREEBSD
			else if (time_type_flag == TIME_BTIME)
			  t = statp->st_birthtime;
#endif
			else
			  t = statp->st_mtime;
			fmt++;
		  }
		  else if (nextch == 'm') {
			t = statp->st_mtime;
			fmt++;
		  }
		  else {
			/* ??? */
			/* Copy verbatim. */
			olen += ds_asprintf(ds, "?%c?", (char) *fmt);
			t = 0;
		  }
		}
		else
		  t = 0;

		if (show_secs)
		  olen += ds_asprintf(ds, "%ld", t);
		else {
		  time(&now);
		  tm_now = tm_t = NULL;
		  if ((tm = localtime(&now)) != NULL) {
			tm_now = ALLOC(struct tm);
			memcpy(tm_now, tm, sizeof(struct tm));
		  }
		  if ((tm = localtime(&t)) != NULL) {
			tm_t = ALLOC(struct tm);
			memcpy(tm_t, tm, sizeof(struct tm));
		  }
		  if (tm_now == NULL || tm_t == NULL) {
			break;
		  }

		  /*
		   * To approximate "ls -l", list only the year for dates greater than
		   * 6 months ago or in the future, otherwise only the time.
		   * [see: Single UNIX Spec, Version 7, 2018]
		   * Is it not better to list the time for dates within the current
		   * year, otherwise the year (except for dates in the future)?
		   * At least this provides a quick visual indication of dates that
		   * fall within the current calendar year.
		   */
		  if (dateformat_flag != NULL)
			st = strftime(buf, sizeof(buf), dateformat_flag, tm_t);
#ifdef USE_LS_COMMAND_DATES
#define APPROX_SECS_PER_6MONTHS		((365.5 * (24 * 60 * 60)) / 2)
		  else if ((now >= t) && (now - t) <= APPROX_SECS_PER_6MONTHS)
			st = strftime(buf, sizeof(buf), "%b %e %H:%M", tm_t);
#else
		  else if ((now >= t) && (tm_now->tm_year == tm_t->tm_year))
			st = strftime(buf, sizeof(buf), "%b %e %H:%M", tm_t);
#endif
		  else
			st = strftime(buf, sizeof(buf), "%b %e  %Y", tm_t);
		  if (st != 0)
			olen += ds_asprintf(ds, "%s", buf);
		  else
			olen += ds_asprintf(ds, "???");
		}

		break;

	  case 'D':
		/* Device number. */
		if (statp)
		  olen += ds_asprintf(ds, "%*lu", field_dev_width,
						   (unsigned long) statp->st_dev);
		else
		  olen += ds_asprintf(ds, "%*s", field_dev_width, " ");
		break;

	  case 'l':
		/* Link count. */
		if (statp)
		  olen += ds_asprintf(ds, "%*lu", field_nlink_width,
						   (unsigned long) statp->st_nlink);
		else
		  olen += ds_asprintf(ds, "%*s", field_nlink_width, " ");
		break;

	  case 'm':
	  case 'M':
		/* Mode bits. */
		if (statp) {
		  /*
		   * strmode(3) is non-standard and may be unavailable on CentOS etc.
		   * Use the portable DACS implementation.
		   */
		  if (*fmt == 'm') {
			/*
			 * The string is always 11 chars long and ends with a space.
			 * We need to zap the space to avoid an extra one.
			 */
			strfilemode(statp->st_mode, bp);
			bp[10] = '\0';
			olen += ds_asprintf(ds, "%*s", MODE_FIELD_WIDTH, bp);
		  }
		  else
			olen += ds_asprintf(ds, "%*o", MODE_FIELD_WIDTH,
							 (unsigned int) statp->st_mode);
		}
		else
		  olen += ds_asprintf(ds, "%*s", MODE_FIELD_WIDTH, " ");
		break;

	  case 'R':
		/* Device number for device special files. */
		if (statp)
		  olen += ds_asprintf(ds, "%-*lu", field_rdev_width,
						 (unsigned long) statp->st_rdev);
		else
		  olen += ds_asprintf(ds, "%*s", field_rdev_width, " ");
		break;

	  case 'T':
		/*
		 * File type indicator character (as per ls(1)).
		 *   Directory: '/'
		 *   Executable: '*'
		 *   Symbolic link: '@'
		 *   Socket: '='
		 *   FIFO: '|'
		 *   Whiteout: '%'
		 */
		if (!suppress_filetype) {
		  ind = file_type_ind(statp);
		  olen += ds_asprintf(ds, "%s", ind ? ind : "");
		}
		else
		  suppress_filetype = 0;
		break;

	  case 'u':
		/* Username, if possible. */
		errno = 0;
		if (statp && (u = getpwuid(statp->st_uid)) != NULL)
		  olen += ds_asprintf(ds, "%-*s", field_pw_name_width, u->pw_name);
		else
		  olen += ds_asprintf(ds, "%*s", field_pw_name_width, " ");
		break;

	  case 'U':
		/* UID. */
		if (statp)
		  olen += ds_asprintf(ds, "%-*u", field_uid_width, statp->st_uid);
		else
		  olen += ds_asprintf(ds, "%*s", field_uid_width, " ");
		break;

	  case 'g':
		/* Groupname, if possible. */
		if (statp && (g = getgrgid(statp->st_gid)) != NULL)
		  olen += ds_asprintf(ds, "%-*s", field_gr_name_width, g->gr_name);
		else
		  olen += ds_asprintf(ds, "%*s", field_gr_name_width, " ");
		break;

	  case 'G':
		/* GID. */
		if (statp)
		  olen += ds_asprintf(ds, "%-*u", field_gid_width, statp->st_gid);
		else
		  olen += ds_asprintf(ds, "%*s", field_gid_width, " ");
		break;

	  case 'f':
#ifdef DACS_OS_FREEBSD
		if (statp)
		  olen += ds_asprintf(ds, "%-*x", FLAGS_FIELD_WIDTH, statp->st_flags);
		else
		  olen += ds_asprintf(ds, "%*s", FLAGS_FIELD_WIDTH, " ");
#else
		olen += ds_asprintf(ds, "%*s", FLAGS_FIELD_WIDTH, " ");
#endif
		break;

	  case 'i':
		if (statp)
		  olen += ds_asprintf(ds, "%-*lu", field_ino_width,
						   (unsigned long) statp->st_ino);
		else
		  olen += ds_asprintf(ds, "%*s", field_ino_width, " ");
		break;

	  case 's':
	  case 'h':
		/* Size, in bytes. */
		/* 'h' selects "human readable" format. */
		if (statp) {
		  if (human_units_flag || *fmt == 'h')
			olen += ds_asprintf(ds, "%*s",
						   field_size_width,
						   str_unit_suffix((unsigned long) statp->st_size,
										   human_units_base_flag));
		  else
			olen += ds_asprintf(ds, "%*lu", field_size_width,
						   (unsigned long) statp->st_size);
		}
		else
		  olen += ds_asprintf(ds, "%*s", field_size_width, " ");
		break;

		if (statp)
		  olen += ds_asprintf(ds, "%*lu", field_size_width,
						 (unsigned long) statp->st_size);
		else
		  olen += ds_asprintf(ds, "%*s", field_size_width, " ");
		break;

	  case 'n':
		/*
		 * Filename.
		 */
		if (filetype_flag || (slash_flag && S_ISDIR(statp->st_mode))) {
		  /* If the path already ends with a "/", don't add another one. */
		  if (strsuffix(fts_path, strlen(fts_path), "/") == NULL)
			ind = file_type_ind(statp);
		  else
			ind = NULL;

		  /*
		   * Peek ahead to see if an explicit filetype indicator follows
		   * immediately in the format.  If so, suppress the duplicate
		   * in this context.
		   */
		  if (*(fmt + 1) == '%' && *(fmt + 2) == 'T')
			suppress_filetype = 1 ;
		}
		else
		  ind = NULL;

		/*
		 * Determine if/how the filename should be highlighted.
		 * It will be highlighted if:
		 * 1) the filename contains an unprintable character and
		 *    COLOUR_EFFECT_UNPRINTABLE is configured, or
		 * Note: an unprintable string may not actually appear highlighted
		 * 2) highlighting has been explicitly configured for the file type, or
		 * 3) it's not a regular file and COLOUR_EFFECT_DEFAULT_SPECIAL
		 *    is configured, or
		 * 4) it's a regular file and COLOUR_EFFECT_DEFAULT_ALL is configured
		 *
		 * XXX might be cute to add a "rainbow" effect option to colourize
		 * adjacent characters using alternating colours...
		 * "\033[31;1mA", "\033[33;1mB", "\033[31;1mC", ...
		 */
		effect_str = NULL;
		unprintable_count = count_unprintable_chars(ofn);
		if (format_flag == NULL
			&& (long_flag || !colour_long_only_flag)
			&& (unprintable_count == 0 || unprintable_filenames_flag == 0)) {
		  if (colour_effect[COLOUR_EFFECT_UNPRINTABLE].effect_str != NULL
			  && unprintable_count)
			effect_str = colour_effect[COLOUR_EFFECT_UNPRINTABLE].effect_str;
		  else if ((effect_str = colour_effect_lookup(fn, statp->st_mode))
				   == NULL) {
			if (!S_ISREG(statp->st_mode))
			  effect_str
				= colour_effect[COLOUR_EFFECT_DEFAULT_SPECIAL].effect_str;
			if (effect_str == NULL)
			  effect_str = colour_effect[COLOUR_EFFECT_DEFAULT_ALL].effect_str;
		  }
		}

		if (effect_str != NULL) {
		  ds_asprintf(ds, "%s", effect_str);
		  /* Make sure highlighting is turned off later. */
		  hilited = 1;
		}

		if (verbose_flag > 1) {
		  if (ind == NULL)
			asprintf(&ind, "@%ld", d->fts_level);
		  else
			asprintf(&ind, "%s@%ld", ind, d->fts_level);
		}

		if (ofn != NULL) {
		  if (quoted_flag)
			olen += ds_asprintf(ds, "'%s'%s",
								format_safe_filename(ofn),
								(ind != NULL) ? ind : "");
		  else
			olen += ds_asprintf(ds, "%s%s",
								format_safe_filename(ofn),
								(ind != NULL) ? ind : "");

		  if (long_flag && S_ISLNK(statp->st_mode)) {
			ssize_t count;
			char target_buf[MAXPATHLEN + 1];
			
			/*
			 * Careful... readlink(2) does not append a null...
			 * XXX the max path length? PATH_MAX, MAX_PATH_LENGTH
			 */
			count = readlink(fts_path, target_buf, sizeof(target_buf) - 1);
			if (count == -1)
			  olen += ds_asprintf(ds, " -> ???");
			else {
			  /*
			   * Indicate whether symlink is broken or ok.
			   */
			  target_buf[count] = '\0';
			  olen += ds_asprintf(ds, " %s %s",
								  (access(target_buf, F_OK) == -1) ? "-?" : "->",
								  target_buf);
			}
		  }
		  free(ofn);
		  ofn = NULL;
		}
		break;

	  case '\0':
		/* Format ends with a '%'. */
		olen += ds_asprintf(ds, "%%");
		break;

	  case '%':
		/*  */
		olen += ds_asprintf(ds, "%%");
		break;

	  default:
		/* Unknown escape. */
		break;
	  }

	  if (hilited) {
		ds_asprintf(ds, "%s", ANSI_TTY_RESET_STR);
		hilited = 0;
	  }
	}
	else {
	  /* Copy verbatim. */
	  olen += ds_asprintf(ds, "%c", (char) *fmt);
	}

	fmt++;
  }

  if (dsp != NULL)
	*dsp = ds;
  else {
	printf("%s", ds_buf(ds));
	ds_free(ds);
  }

  return(olen);
}

typedef struct Printit_arg {
  char *line_format;
  int c;
} Printit_arg;

static int
printit(void *base, int num, void *arg)
{
  int len;
  File_d *d, **dp;
  Printit_arg *ff;

  dp = (File_d **) base;
  d = dp[num];

  ff = (Printit_arg *) arg;
  len = format_line(ff->line_format, d, ff->c, NULL);
  if (len)
	ff->c++;

  return(len);
}

enum {
  SECS_PER_MINUTE = 60,
  SECS_PER_HOUR   = (SECS_PER_MINUTE * 60),
  SECS_PER_DAY    = (SECS_PER_HOUR * 24),
  SECS_PER_WEEK   = (SECS_PER_DAY * 7),
  SECS_PER_MONTH  = (SECS_PER_DAY * 30),
  SECS_PER_YEAR   = (SECS_PER_DAY * 365)
};

static time_t now = 0;

/*
 * Return the number of time units (UNITS) between now and ST_TIME,
 * both assumed to be within the same time zone.
 *
 * This is an approximation.
 * For example, when the units are "days", the time of day does not matter,
 * only the difference in calendar days.  So if it is currently
 * 1-Aug-2022 at 12:01AM and ST_TIME represents 31-Jul-2022 at 11:59PM,
 * a value of one will be returned because the times occur on different days.
 * If it is currently 1-Aug-2022 at 11:59PM and ST_TIME represents
 * 1-Aug-2022 at 12:01AM, a value of zero will be returned because both times
 * are on the same calendar day (zero days separate them).
 *
 * Note that ST_TIME cannot represent a time before the start of the epoch
 * (1-Jan-1970 at hour zero).  If ST_TIME is in the future it is taken to
 * be equivalent to "now".
 */
typedef enum {
  AGE_IN_SECONDS = 0,
  AGE_IN_MINUTES = 1,
  AGE_IN_HOURS   = 2,
  AGE_IN_DAYS    = 3,
  AGE_IN_WEEKS   = 4,
  AGE_IN_MONTHS  = 5,
  AGE_IN_YEARS   = 6
} Age_units;

static Age_units age_units = AGE_IN_DAYS;

static unsigned int
age_since(Age_units units, time_t st_time)
{
  unsigned int age;
  time_t rsecs;

  /* Initialize the current time, just once. */
  if (now == 0)
	time(&now);

  /* If ST_TIME is not in the past, consider it today/now. */
  if (now <= st_time)
	return(0);

  switch (units) {
  case AGE_IN_SECONDS:
	age = now - st_time;
	break;

  case AGE_IN_MINUTES:
	age = (now / SECS_PER_MINUTE) - (st_time / SECS_PER_MINUTE);
	break;

  case AGE_IN_HOURS:
	age = (now / SECS_PER_HOUR) - (st_time / SECS_PER_HOUR);
	break;

  case AGE_IN_DAYS:
	age = (now / SECS_PER_DAY) - (st_time / SECS_PER_DAY);
	break;

  case AGE_IN_WEEKS:
	age = (now / SECS_PER_WEEK) - (st_time / SECS_PER_WEEK);
	break;

  case AGE_IN_MONTHS:
	/* A rough approximation. */
	age = (now / SECS_PER_MONTH) - (st_time / SECS_PER_MONTH);
	break;

  case AGE_IN_YEARS:
	/* A rough approximation. */
	age = (now / SECS_PER_YEAR) - (st_time / SECS_PER_YEAR);
	break;

  default:
	age = 0;
	break;
  }
  
  return(age);
}

/*
 * Allocate and initialize a new Hier structure for listing PATH and set
 * HIERP to point to it.  Its ST element is set to 1 if all is ok, 2 if
 * PATH is a broken link, 0 if no information about PATH was found, or -1 for
 * any other error.
 *
 * If stat() fails and: if it is a broken symbolic link,
 * note that it is and return 1; if PATH does not exist, return 1; and for any
 * other error (e.g., it is an unreadable directory), return -1.
 * If OK, return 1.
 *
 * XXX this is probably overly complicated
 */
static int
is_listable_arg(char *path, Hier **hierp)
{
  Hier *hier;

  hier = ALLOC(Hier);
  hier->filename = strdup(path);
  hier->st = 1;
  *hierp = hier;

  if (stat(path, &hier->stat) == -1) {
	int err;

	err = errno;
	if (lstat(path, &hier->stat) == 0) {
	  /* PATH exists but is a broken link? */
	  hier->st = 2;
	  return(1);
	}

	memset(&hier->stat, 0, sizeof(struct stat));
	if (err == ENOENT) {
	  hier->st = 0;
	  return(0);
	}
	hier->st = -1;
	return(-1);
  }

  return(1);
}

/*
 * Return 1 if F was specified on the command line (i.e., it appears in HIER),
 * otherwise 0.
 */
static int
is_cmdline_file(Hier **hier, FTSENT *f)
{
  int i;

  if (nfile_args == 0)
	return(0);

  for (i = 0; hier[i]->filename != NULL; i++) {
	if ((hier[i]->stat.st_dev == f->fts_statp->st_dev)
		&& (hier[i]->stat.st_ino == f->fts_statp->st_ino))
	  return(1);

	if (streq(hier[i]->filename, f->fts_path))
	  return(1);

	if (streq(hier[i]->filename, f->fts_name))
	  return(1);
  }

  return(0);
}

/*
 * Order command line file arguments.
 * Defaults:
 * 1) non-directory arguments are displayed first;
 * 2) Directory and non-directory arguments are sorted separately; and
 * 3) lexicographical ordering is applied.
 */
static int
compar_file_args(const void *aa, const void *bb)
{
  int dirs_first;
  Hier *a, *b;
  struct stat *sa, *sb;

  a = *(Hier **) aa;
  b = *(Hier **) bb;

  sa = &a->stat;
  sb = &b->stat;

  dirs_first = dirsfirst_flag ? -1 : 1;


  if (a->st == 1) {
	if (b->st == 1) {
	  if (S_ISDIR(sa->st_mode) && !S_ISDIR(sb->st_mode))
		return(dirs_first);

	  if (!S_ISDIR(sa->st_mode) && S_ISDIR(sb->st_mode))
		return(-dirs_first);
	}
	else if (S_ISDIR(sa->st_mode))
	  return(dirs_first);
  }
  else if (b->st == 1) {
	if (S_ISDIR(sb->st_mode))
	  return(-dirs_first);
  }

  return(strcmp(a->filename, b->filename));
}

static void
sort_file_args(int nargs, Hier **file_args)
{

  if (nargs <= 1)
	return;

  qsort(file_args, nargs, sizeof(Hier *), compar_file_args);
}

static int
do_du(Hier **hier_arg, char *path)
{
  int count, fts_options;
  long du;
  char *hier[2];
  File_d *d;
  FTS *ftsp;

  du = 0;

  fts_options = FTS_NOCHDIR | FTS_XDEV | FTS_SEEDOT;
  if (symlink_Pflag)
	fts_options |= FTS_PHYSICAL;
  else
	fts_options |= FTS_LOGICAL;

  hier[0] = path;
  hier[1] = NULL;
  ftsp = fts_open(hier, fts_options, NULL);
  if (ftsp == NULL) {
	perror("fts_open");
	return(-1);
  }

  count = 0;
  while (1) {
	int is_dot_file, is_dot_dir;
	char *b;
	FTSENT *f;

	if ((f = fts_read(ftsp)) == NULL)
	  break;

	/*
	 * If this is the second visit to a directory, ignore it to avoid
	 * adding it twice.
	 */
	if (f->fts_info == FTS_DP)
	  continue;

	fts_read_count++;
	if (verbose_flag > 1)
	  fprintf(stderr, ">>>%5d %s\n", fts_read_count, f->fts_accpath);

	/*
	 * If only directories are to be listed, ignore anything else.
	 */
	if (lsd_flag && !S_ISDIR(f->fts_statp->st_mode))
	  continue;

	b = strbasename(f->fts_name, NULL);
	is_dot_file = (*b == '.');
	is_dot_dir = (S_ISDIR(f->fts_statp->st_mode)
		&& ((*b == '.' && *(b + 1) == '\0')
			|| ((*b == '.') && *(b + 1) == '.' && *(b + 2) == '\0')));

	if (f->fts_level != FTS_ROOTLEVEL && !is_cmdline_file(hier_arg, f)) {
	  if ((is_dot_dir && !almost_all_flag) || (is_dot_file && !all_flag)) {
		if (fts_set(ftsp, f, FTS_SKIP) == -1) {
		  perror("fts_set");
		  exit(1);
		}
		continue;
	  }
	}

	/*
	 * If this is a directory, it will only be listed if it is a command line
	 * argument or it is a dot file, and onelevel_flag is enabled.
	 */
	if (S_ISDIR(f->fts_statp->st_mode)
		&& ((nfile_args && is_cmdline_file(hier_arg, f))
			|| (nfile_args == 0 && is_dot_file)))
	  {
	  if (onelevel_flag) {
		if ((f->fts_statp->st_mode & S_IFMT) == S_IFREG)
		  /* du += f->fts_statp->st_size; */
		  /* du += filesystem_blocks_used(f->fts_statp); */
		  du += f->fts_statp->st_blocks;
		count++;

		if (limit_count_flag && count == limit_count_flag)
		  break;

		if (fts_set(ftsp, f, FTS_SKIP) == -1) {
		  perror("fts_set");
		  exit(1);
		}
	  }

	  continue;
	}

	if (today_count_flag) {
	  unsigned int n;

	  if (!S_ISREG(f->fts_statp->st_mode))
		continue;

	  if (time_type_flag == TIME_ATIME)
		n = age_since(age_units, f->fts_statp->st_atime);
#ifdef DACS_OS_FREEBSD
	  else if (time_type_flag == TIME_BTIME)
		n = age_since(age_units, f->fts_statp->st_birthtime);
#endif
	  else if (time_type_flag == TIME_CTIME)
		n = age_since(age_units, f->fts_statp->st_ctime);
	  else
		n = age_since(age_units, f->fts_statp->st_mtime);

	  if (n > today_count)
		continue;
	  if (today_count_flag == TODAY_COUNT_EXACT_DAYS && n != today_count)
		continue;
	}

	if (perms_match_flag != PERMS_MATCH_IGNORE) {
	  mode_t m;

	  m = f->fts_statp->st_mode & S_IPMT;
	  if (perms_match_flag == PERMS_MATCH_EXACT && m != perms_match)
		continue;

	  if (perms_match_flag == PERMS_MATCH_SET
		  && ((m & perms_match) != perms_match))
		continue;

	  if (perms_match_flag == PERMS_MATCH_NOTSET && ((m & ~perms_match) == 0))
		continue;
	}

	if (type_match_flag != NULL) {
	  if (!file_type_match(type_match_flag, f->fts_statp))
		continue;
	}

	/*
	 * If only non-directories are to be listed, do not print anything else.
	 */
	if (!lsf_flag || !S_ISDIR(f->fts_statp->st_mode)) {
	  if (!all_flag) {
		/* Ignore a dot file unless it's a command line argument. */
		if (!is_dot_file || is_cmdline_file(hier_arg, f)) {
		    du += f->fts_statp->st_blocks;
		  count++;
		}
	  }
	  else {
		/* It looks like "." is seen twice: once as "." and again as "./." */
		if (f->fts_info != FTS_DOT || !streq(f->fts_name, ".")) {
		    /* du += f->fts_statp->st_size; */
		    /* du += filesystem_blocks_used(f->fts_statp); */
		    du += f->fts_statp->st_blocks;
		  count++;
		}
	  }
	}

	if (limit_count_flag && count == limit_count_flag)
	  break;

	/* Check for pruning. */
	if (f->fts_info == FTS_D && recurse_flag >= 0) {
	  if (onelevel_flag
		  || (lsd_flag && f->fts_level > 1)
		  || f->fts_level >= recurse_flag) {
		if (fts_set(ftsp, f, FTS_SKIP) == -1) {
		  perror("fts_set");
		  exit(1);
		}
	  }
	}
  }

  if (errno != 0) {
	perror("fts_read");
	if (fts_close(ftsp) == -1)
	  perror("fts_close");

	return(-1);
  }

  if (fts_close(ftsp) == -1) {
	perror("fts_close");
	return(-1);
  }

  return(du);
}

/*
 * List command line file or directory PATH (with additional detail in
 * HIER_ARG).
 * If PATH is a dot file, it appeared on the command line and so will be
 * listed regardless of any suppression implied by all_flag or almost_all_flag.
 *
 * The contents of a directory may be listed, but this function is not called
 * recursively.
 *
 * Return the number of entries listed, or -1 if an error occurred.
 */
static int
do_ls(Hier **hier_arg, char *path)
{
  int count, do_list, fts_options;
  char *hier[2];
  File_d *d;
  FTS *ftsp;

  fts_options = FTS_NOCHDIR | FTS_XDEV;
  if (symlink_Pflag)
	fts_options |= FTS_PHYSICAL;
  else
	fts_options |= FTS_LOGICAL;

  /*
   * Without FTS_SEEDOT dot and dotdot are not returned by fts_read()
   * unless they are PATH.
   */
  if (all_flag)
	fts_options |= FTS_SEEDOT;

  hier[0] = path;
  hier[1] = NULL;
  if (sort_type_flag  == SORT_UNSORTED)
	ftsp = fts_open(hier, fts_options, NULL);
  else
	ftsp = fts_open(hier, fts_options, compar_fts);
  if (ftsp == NULL) {
	perror("fts_open");
	return(-1);
  }

  count = 0;
  while (1) {
	int is_dot_file, is_dot_dir;
	char *bn;
	FTSENT *f;

	/*
	 * Read entries until there are no more or we've reached a limit.
	 *
	 * For each entry, decide:
	 * 1) whether it needs to be listed, AND IF SO AND IT IS A DIRECTORY,
	 * 2) whether its contents need to be listed.
	 */
	if ((f = fts_read(ftsp)) == NULL)
	  break;

	/*
	 * If this is the second visit to a directory, ignore it to avoid
	 * adding it twice.
	 */
	if (f->fts_info == FTS_DP)
	  continue;

	/*
	 * If fts_open() is given ".", fts_read() returns both "." and "./."
	 */
	if (streq(f->fts_path, "./."))
	  continue;

	fts_read_count++;
	if (verbose_flag > 1)
	  fprintf(stderr, ">>>%5d %s\n", fts_read_count, f->fts_accpath);

	/* If only directories are to be listed, ignore anything else. */
	if (lsd_flag && !S_ISDIR(f->fts_statp->st_mode))
	  continue;

	/* If only non-directories are to be listed, ignore anything else. */
	if (lsf_flag && S_ISDIR(f->fts_statp->st_mode))
	  continue;

	/* If only symbolic links are to be listed, ignore anything else. */
	if (lss_flag && !S_ISLNK(f->fts_statp->st_mode)) {
	  if (!is_cmdline_file(hier_arg, f)) {
		if (fts_set(ftsp, f, FTS_SKIP) == -1) {
		  perror("fts_set");
		  exit(1);
		}
	  }

	  continue;
	}

#ifdef NOTDEF
	bn = strbasename(f->fts_name, NULL);
#endif
	bn = f->fts_name;

	/* A file/directory name beginning with a dot, including dot and dot-dot. */
	is_dot_file = (*bn == '.');

	/* The dot or dot-dot directories. */
	is_dot_dir = (S_ISDIR(f->fts_statp->st_mode)
				  && (*bn == '.'
					  && (*(bn + 1) == '\0'
						  || (*(bn + 1) == '.' && *(bn + 2) == '\0'))));

	/*
	 * Decide if F should be listed, traversed, neither, or both.
	 *
	 * 1. A file/directory that appears on the command line should always be
	 * listed (unless its file type is inapplicable or some limit is reached).
	 *
	 * 2. A file/directory is listed subject to: all_flag, almost_all_flag,
	 *    and other criteria
	 * 2a. A listed directory is traversed unless onelevel_flag is set
	 *     or a maximum depth limit has been reached
	 *
	 * The only case of a directory that is "traversed-but-not-listed" is
	 * for "." when it is the default case
	 * (nfile_args is zero because no files were given on the command line)
	 * and it may not appear in the listing (all_flag is zero) but
	 * its contents are eligible for being listed.
	 *
	 * Note: f->fts_level is the depth of the traversal where f is.
	 * A level of FTS_ROOTPARENT_LEVEL (-1) means that f is the parent/root
	 * directory of PATH, and a level of FTS_ROOTLEVEL (0) means that f
	 * corresponds to PATH.
	 *
	 * If F is a directory that should not be listed, call fts_set() with
	 * FTS_SKIP.
	 */

	/*
	 * Decide if F should be listed.
	 * Assume that it will be listed, then check if a flag asks for listing
	 * to be suppressed.
	 */
	do_list = 1;
	if (!is_cmdline_file(hier_arg, f)) {
	  if (f->fts_info == FTS_D) {
		if (!all_flag) {
		  if (is_dot_dir && almost_all_flag)
			do_list = 0;
		  else if (is_dot_file)
			do_list = 0;
		}

		if (!do_list && f->fts_level != FTS_ROOTLEVEL) {
		  if (fts_set(ftsp, f, FTS_SKIP) == -1) {
			perror("fts_set");
			exit(1);
		  }

		  continue;
		}
	  }
	  else {
		if (is_dot_file && !all_flag && !almost_all_flag)
		  do_list = 0;
		else if (!recurse_flag && f->fts_level > 1)
		  do_list = 0;
	  }
	}
	else {
	  /*
	   * If invoked as: lf nas/
	   * then "nas" does not appear in the listing
	   * If invoked as: lf nas/ bear/
	   * then "nas" and "bear" appear as separate listing sections.
	   */
	  if (S_ISDIR(f->fts_statp->st_mode))
		do_list = 0;
	}

	/*
	 * If this is a directory, it will be listed if:
	 * 1) it is a command line argument or
	 * 2) it is a dot file and onelevel_flag (-d) is enabled.
	 */
	if (S_ISDIR(f->fts_statp->st_mode)
		&& (is_cmdline_file(hier_arg, f) || (nfile_args == 0 && is_dot_file))) {
	  if (do_list || onelevel_flag) {
		if ((d = add_file_d(f)) == NULL)
		  return(-1);
		count++;

		if (limit_count_flag && count == limit_count_flag)
		  break;

		if (onelevel_flag && fts_set(ftsp, f, FTS_SKIP) == -1) {
		  perror("fts_set");
		  exit(1);
		}
	  }

	  continue;
	}

	/* Reject if F does not satisfy a date attribute requirement. */
	if (today_count_flag) {
	  unsigned int n;

	  if (!S_ISREG(f->fts_statp->st_mode))
		continue;

	  if (time_type_flag == TIME_ATIME)
		n = age_since(age_units, f->fts_statp->st_atime);
#ifdef DACS_OS_FREEBSD
	  else if (time_type_flag == TIME_BTIME)
		n = age_since(age_units, f->fts_statp->st_birthtime);
#endif
	  else if (time_type_flag == TIME_CTIME)
		n = age_since(age_units, f->fts_statp->st_ctime);
	  else
		n = age_since(age_units, f->fts_statp->st_mtime);

	  if (n > today_count)
		continue;
	  if (today_count_flag == TODAY_COUNT_EXACT_DAYS && n != today_count)
		continue;
	}

	if (perms_match_flag != PERMS_MATCH_IGNORE) {
	  mode_t m;

	  m = f->fts_statp->st_mode & S_IPMT;
	  /* Deselect if permission bits are not exactly PERMS_MATCH. */
	  if (perms_match_flag == PERMS_MATCH_EXACT && m != perms_match)
		continue;

	  /* Deselect if permission bits do not include all of PERMS_MATCH. */
	  if (perms_match_flag == PERMS_MATCH_SET
		  && ((m & perms_match) != perms_match))
		continue;

	  /* Deselect if permission bits include any of PERMS_MATCH. */
	  if (perms_match_flag == PERMS_MATCH_NOTSET && ((m & perms_match) != 0))
		continue;
	}

	if (type_match_flag != NULL) {
	  if (!file_type_match(type_match_flag, f->fts_statp)) {
		/* XXX do this for all no-list cases? */
		if (fts_set(ftsp, f, FTS_SKIP) == -1) {
		  perror("fts_set");
		  exit(1);
		}

		continue;
	  }
	}

	/*
	 * If only non-directories are to be listed, do not print anything else.
	 */
	if (lsf_flag && S_ISDIR(f->fts_statp->st_mode))
	  continue;
	if (lsd_flag && !S_ISDIR(f->fts_statp->st_mode))
	  continue;

	if (do_list) {
	  if ((d = add_file_d(f)) == NULL)
		return(-1);
	  count++;
	}

	/* We are done if the maximum listing count has been reached. */
	if (limit_count_flag && count == limit_count_flag)
	  break;

	/* Prune F if it is a directory at the maximum listing depth. */
	if (f->fts_info == FTS_D && recurse_flag >= 0) {
	  if (onelevel_flag
		  || (lsd_flag && f->fts_level > 1)
		  || f->fts_level >= recurse_flag) {
		if (fts_set(ftsp, f, FTS_SKIP) == -1) {
		  perror("fts_set");
		  exit(1);
		}
	  }
	}
  }

  if (errno != 0) {
	perror("fts_read");
	if (fts_close(ftsp) == -1)
	  perror("fts_close");

	return(-1);
  }

  if (fts_close(ftsp) == -1) {
	perror("fts_close");
	return(-1);
  }

  return(count);
}

/*
 * Read FILENAME for a sequence of filenames, one per line, and append each
 * to the list of filenames to be listed.  If FILENAME is "-", read stdin
 * (use "./-" if the input file is actually "-").
 *
 * Return the number of filenames read if ok, -1 if an error occurs.
 */
static int
add_namefile_files(Hier ***hierp, char *filename)
{
  int count, i, n;
  Dsvec *dsv;
  Hier **hier;

  if (filename == NULL)
	return(-1);

  if (streq(filename, "-"))
	dsv = dsvec_load_lines(NULL, NULL);
  else
	dsv = dsvec_load_lines(NULL, filename);

  if (dsv == NULL) {
	fprintf(stderr, "Cannot open namelist \"%s\"\n", filename);
	return(-1);
  }

  if (hierp == NULL)
	hier = NULL;
  else
	hier = *hierp;

  count = 0;
  n = nfile_args;
  /* Allocate an extra for the sentinel. */
  hier = REALLOC_N(Hier *, hier, nfile_args + dsvec_len(dsv) + 1);

  for (i = 0; i < dsvec_len(dsv); i++) {
	int st;
	char *fn;

	fn = (char *) dsvec_ptr_index(dsv, i);

	if ((st = is_listable_arg(fn, &hier[n])) == 1) {
	  if (fn[0] == '.')
		saw_dot_file_arg = 1;
	  n++;
	  count++;
	}
	else {
	  if (st == 0)
		fprintf(stderr, "%s: \"%s\": No such file or directory\n", progname, fn);
	  else
		fprintf(stderr, "%s: \"%s\": Permission denied\n", progname, fn);
	}
  }

  *hierp = hier;
  dsvec_free(dsv);

  return(count);
}

#include "dacs_version.h"

static char *
version_string(void)
{
  char *vs;

  asprintf(&vs, "lf %s %s (%s)",
		   DACS_VERSION_RELEASE, DACS_VERSION_DATE, DACS_VERSION_REV);

  return(vs);
}

typedef enum {
  HELP_NONE   = 0x00,
  HELP_JFLAGS = 0x01,
  HELP_CFLAGS = 0x02,
  HELP_IFLAGS = 0x04,
  HELP_SFLAGS = 0x08,
  HELP_FFLAGS = 0x10,
  HELP_HFLAGS = 0x20,
  HELP_PFLAGS = 0x40,
  HELP_ALL    = (HELP_JFLAGS | HELP_CFLAGS | HELP_IFLAGS
				 | HELP_SFLAGS | HELP_FFLAGS | HELP_HFLAGS | HELP_PFLAGS)
} Help_type;

static void
usage(Help_type which)
{
  int need_nl;
  char *bold_off, *bold_on;
  FILE *ofile;

  /* Use simple highlighting. */
  bold_on = "\033[1m";
  bold_off = "\033[0m";

  ofile = stdout;

  fprintf(ofile, "%sUsage: lf ", bold_on);
  fprintf(ofile, "[fflag] [jflags] [cflags] [hflags] [iflags] [sort-flag] [file ...]%s\n", bold_off);

  need_nl = 1;
  fprintf(ofile, "%spflags (program, must be first argument, pick one):%s\n",
		  bold_on, bold_off);
  fprintf(ofile, " --lsd:   List only directories (also: --lfd)\n");
  fprintf(ofile, " --lsf:   List only non-directories\n");
  fprintf(ofile, " --lss:   List only symbolic links\n");
  fprintf(ofile, " --today: Equivalent to invoking as 'today'\n");

  if ((which == HELP_ALL) || (which & HELP_JFLAGS)) {
	/* General: joinable flags. */
	if (need_nl)
	  fprintf(ofile, "\n");
	fprintf(ofile, "%sjflags (single, joinable):%s\n", bold_on, bold_off);
	fprintf(ofile, " -1: single column output (also: --single-column)\n");
	fprintf(ofile, " -a: include 'dot' files when listing directories\n");
	fprintf(ofile, " -A: print absolute paths\n");
	fprintf(ofile, " -c: use the last change time instead of last modification time\n");
	fprintf(ofile, " -d: include the root level directory (also -r0)\n");
	fprintf(ofile, " -F: append file type indicator to file names\n");
	fprintf(ofile, " -h: with -l, use human-friendly unit suffixes for sizes\n");
	fprintf(ofile, " -i: print the inode number (unless -fmt)\n");
	fprintf(ofile, " -k: use 1024 for the blocksize (overrides prior -h)\n");
	fprintf(ofile, " -l: use the long format (unless -fmt)\n");
	fprintf(ofile, " -m: merge filenames into lines\n");
	fprintf(ofile, " -n: prepend a line number to the format, starting at 1\n");
	fprintf(ofile, " -p: append a slash to each filename that is a directory\n");
	fprintf(ofile, " -Q: surround filenames with single quotes\n");
	fprintf(ofile, " -r: reverse (default or selected) sort order\n");
	fprintf(ofile, " -R: recurse\n");
	fprintf(ofile, " -s: display number of file system blocks used by file\n");
	fprintf(ofile, " -u: use the last access time instead of last modification time\n");
	fprintf(ofile, " -U: use the creation time instead of last modification time\n");
	fprintf(ofile, " -v: verbose output\n");
	fprintf(ofile, " -w: emit unprintable filenames (do not escape them)\n");

    need_nl = 1;
  }

  if ((which == HELP_ALL) || (which & HELP_IFLAGS)) {
	/* Separate, Individual: general flags. */
	if (need_nl)
	  fprintf(ofile, "\n");
	fprintf(ofile, "%siflags (individual):%s\n", bold_on, bold_off);
	fprintf(ofile, " -Df: order directories separately and list them first\n");
	fprintf(ofile, " -Dd: list directories only (also if command name is 'lsd')\n");
	fprintf(ofile, " -Ds: list symbolic links only\n");
	fprintf(ofile, " -Dx: list non-directories only\n");
	fprintf(ofile, " -dfmt datefmt: strftime format spec for each date\n");
	fprintf(ofile, " -f namefile: read the names of files to be listed from namefile, one per line\n");
	fprintf(ofile, " -fmt linefmt: custom format spec for output lines (overrides -l)\n");
	fprintf(ofile, " -o#: max # items (0 means unlimited)\n");
	fprintf(ofile, " -S STR: prepend STR to file names (overrides -A)\n");
	fprintf(ofile, " -t[#]: only regular files modified/changed/accessed within past # time units\n");
	fprintf(ofile, "        (default: days; # > 0)\n");
	fprintf(ofile, "        (a single letter time unit may follow the #: dwmyHMS:)\n");
	fprintf(ofile, "        (days, weeks, months, years, Hours, Mins, Secs)\n");
	fprintf(ofile, "        (used as 'today', can shorten to '-#' where # > 0)\n");
	fprintf(ofile, " -te[#]: only regular files modified/changed/accessed exactly # time units\n");
	fprintf(ofile, "        (default: days; # > 0)\n");
	fprintf(ofile, "        (optional time unit suffix as above)\n");
	fprintf(ofile, " --aa: include dot directories other than '.' and '..' (also --almost-all)\n");
	fprintf(ofile, " --aar: for super-user, include dot directories other than '.' and '..'\n");
	fprintf(ofile, "        (also --almost-all-root)\n");
	fprintf(ofile, " --almost-all: include directory names that begin with a dot except for . and ..\n");
	fprintf(ofile, " --bssi: use SI units (powers of 10) for blocksize reporting \n");
	fprintf(ofile, " --esc: escape individual problematic filename characters (repeat for effect)\n");
	fprintf(ofile, " --h#: set units for -h flag to 1024 or 1000\n");
	fprintf(ofile, " --perms [+|-]PERMS: select if octal permission PERMS match exactly/all/none\n");
	fprintf(ofile, " --recurse #: recurse at most # levels deep, # >= zero\n");
	fprintf(ofile, " --type TYPES: select files of any of TYPES (b,c,d/,fr,l@,n,p|,s=,*e,D,E,FR");
#ifdef S_ISWHT
	fprintf(ofile, ",%%");
#endif
	fprintf(ofile, ")\n");
	fprintf(ofile, " --version: print version stamp, then exit\n");
	fprintf(ofile, " --: end of flags\n");
    need_nl = 1;
  }

  if ((which == HELP_ALL) || (which & HELP_CFLAGS)) {
	/* Separate: text colour specifications. */
	if (need_nl)
	  fprintf(ofile, "\n");
	fprintf(ofile, "%scflags (colours, individual, LF_COLOURS envar):%s\n",
			bold_on, bold_off);
	fprintf(ofile, " -Ca: default to colour entire line not only filename\n");
	fprintf(ofile, " -Cl: use colour with long format or custom format only\n");
	fprintf(ofile, " -Coff: force text colourization off\n");
	fprintf(ofile, " -Con: force text colourization on\n");
	fprintf(ofile, " -Cp FILENAME: use RGB definitions in FILENAME\n");
	fprintf(ofile, " -C COLOUR :default for all\n");

	fprintf(ofile, "  COLOUR is one of:\n");
 	fprintf(ofile, "    {black|red|green|yellow|blue|magenta|cyan|orange|white|black}\n");
	fprintf(ofile, "    or {#r;g;b|rgb-colour-name|bold|underline}\n");

	fprintf(ofile,
			" -C<SPEC> COLOUR: highlight in COLOUR files selected by <SPEC>\n");
	fprintf(ofile, "  <SPEC> is one or more of:\n");
	fprintf(ofile, "  + :default for non-regular\n");
	fprintf(ofile, "  - :for regular, -C/ :for directory\n");
	fprintf(ofile, "  . :for regular dot, ");
	fprintf(ofile, "  * :for executable, -C@ :for link\n");
	fprintf(ofile, "  = :for socket\n");
	fprintf(ofile, "  | :for FIFO\n");
	fprintf(ofile, "  %% :for whiteout\n");

	fprintf(ofile, " -Cg GLOB COLOUR: use COLOUR for filenames selected by GLOB, repeatable:\n");
	fprintf(ofile, "   lf -Cg '*.pdf' red -Cg '*.core' yellow\n");
    need_nl = 1;
  }

  if ((which == HELP_ALL) || (which & HELP_SFLAGS)) {
	/* Separate: sorting flags. */
	if (need_nl)
	  fprintf(ofile, "\n");
	fprintf(ofile, "%ssflags (sort flags, pick one):%s\n", bold_on, bold_off);
	fprintf(ofile, " --sa:   sort alphabetically, a-z, case sensitively\n");
	fprintf(ofile, " --sai:  sort alphabetically, a-z, case insensitively\n");
	fprintf(ofile, " --sb:   sort size bigger to smaller\n");
	fprintf(ofile, " --sl:   sort by decreasing hard link count\n");
	fprintf(ofile, " --sn:   sort newest modified to oldest\n");
	fprintf(ofile, " --so:   sort oldest modified to newest\n");
	fprintf(ofile, " --soff: sorting is turned off\n");
	fprintf(ofile, " --ss:   sort size smaller to bigger \n");
	fprintf(ofile, " --sz:   sort alphabetically, z-a, case sensitively\n");
	fprintf(ofile, " --szi:  sort alphabetically, z-a, case insensitively\n");
	fprintf(ofile, "sort-extra:\n");
	fprintf(ofile, " --sg:   separate by gid\n");
	fprintf(ofile, " --su:   separate by uid\n");
	fprintf(ofile, "(also see -r flag)\n");
    need_nl = 1;
  }

  if ((which == HELP_ALL) || (which & HELP_FFLAGS)) {
	/* Output format selection/conversion specifiers. */
	if (need_nl)
	  fprintf(ofile, "\n");
	fprintf(ofile, "%sfflags (formatting, interpolation) (-fmt str):%s\n",
			bold_on, bold_off);
	fputs("  %b : blocks used\n", ofile);
	fputs("  %c : output line number, starting at one\n", ofile);
	fputs("  %d{abcdm}  : use access/birth/changed-inode/default/mod date\n",
		  ofile);
	fputs("  %ds{abcdm} : use seconds timestamp for access/birth/changed-inode/default/mod\n",
		  ofile);
	fputs("  %D : device number (st_dev)\n", ofile);
	fputs("  %m : mode bits as text\n", ofile);
	fputs("  %M : mode bits in octal\n", ofile);
	fputs("  %f : flag bits\n", ofile);
	fputs("  %i : inode number\n", ofile);
	fputs("  %l : link count\n", ofile);
	fputs("  %n : filename\n", ofile);
	fputs("  %R : device number for device special files (st_rdev)\n", ofile);
	fputs("  %s : size in bytes\n", ofile);
	fputs("  %T : file type indicator character\n", ofile);
	fputs("  %{uU} : Username or UID\n", ofile);
	fputs("  %{gG} : Groupname or GID\n", ofile);
	fputs("  %% : literal %\n", ofile);
	fputs("  (Use %+ form to highlight the field: %+s\n", ofile);
	fputs("  (If no -fmt flag, always highlight filename)\n", ofile);
	fputs("  anything else: literal\n", ofile);
    need_nl = 1;
  }

  /* Print this last so it isn't scrolled away. */
  if ((which == HELP_ALL) || (which & HELP_HFLAGS)) {
	if (need_nl)
	  fprintf(ofile, "\n");
	fprintf(ofile, "%shflags (help) one or more, print help to stdout then exit):%s\n",
			bold_on, bold_off);
	fprintf(ofile, " --help:   print all help\n");
	fprintf(ofile, " --help-a: print all help\n");
	fprintf(ofile, " --help-c: print cflags help\n");
	fprintf(ofile, " --help-f: print fflags help\n");
	fprintf(ofile, " --help-h: print help help\n");
	fprintf(ofile, " --help-i: print iflags help\n");
	fprintf(ofile, " --help-j: print jflags help\n");
	fprintf(ofile, " --help-s: print sflags help\n");
	fprintf(ofile, " Also: combined, e.g., --help-cij or long, e.g., --help-sort\n");
    need_nl = 1;
  }

  if (which == HELP_ALL) {
	fprintf(ofile, "\n");

	fprintf(ofile, "%sSynonyms:%s\n", bold_on, bold_off);
	fprintf(ofile, "lsd [...] is equivalent to: lf -Dd [...]\n");
	fprintf(ofile, "lsf [...] is equivalent to: lf -Dx [...]\n");
	fprintf(ofile, "lss [...] is equivalent to: lf -Ds [...]\n");
	fprintf(ofile, "today [-#] [...] is equivalent to lf -t[#] [...] (default: 1)\n");
	fprintf(ofile, "today -e [-#] [...] is equivalent to lf -te[#] [...]\n");
  }

  exit(1);
}

int
main(int argc, char **argv)
{
  int count, i, ncolours;
  int help_flag, saw_file_args;
  char *lf_colours_env, *p;
  File_d *d;
  Help_type help_type;
  Hier **hier;
  extern int disable_frame_memory;

  progname = argv[0];

  /*
   * XXX some operations are super slow when using frame memory support
   * (e.g., "lf -Dx" on a very large/deep directory)
   * Disable now, investigate later.
   */
  disable_frame_memory = 1;

#ifdef LF_TESTS
  int xxx_flag;

  /*
   * If invoked as lf-test, look for an integer argument, run the
   * corresponding test, then exit.
   */
  xxx_flag = 0;
  if (streq(progname, "lf-test")) {
	if (argv[1] == NULL)
	  exit(1);
	if ((xxx_flag = (int) strtol(argv[1], NULL, 10)) == 0)
	  exit(1);
  }

  if (xxx_flag == 1) {
	char *fn, *path;

	fn = "rgb.txt";
	path = "/usr/lib/X11:::/usr/local/lib/X11:/usr/share/X11:/opt/X11/share/X11";
	verbose_flag = 2;
	p = file_path_search(fn, path, R_OK);
	if (p == NULL)
	  fprintf(stderr, "Not found: \"%s\" in \"%s\"\n", fn, path);
	else
	  fprintf(stderr, "Found: \"%s\"\n", p);
	exit(0);
  }

  if (xxx_flag == 2) {
	colour_parse_flag_str("-C red -C/ blue");
	exit(0);
  }

  if (xxx_flag == 3) {
	char *f, fname[128];

	sprintf(fname, "abc%cdef%c%cghi%c",
			0x01, 0x02, 0x80, 0xff);
	f = format_safe_filename(fname);
	fprintf(stderr, "%s\n", f);
	exit(0);
  }

  if (xxx_flag == 4) {
	static char xxx[10] = "\221\222\223";
	FILE *fp;

	fp = fopen(xxx, "w");
	fclose(fp);
	exit(0);
  }

  if (xxx_flag == 5) {
	static char xxx[128], reset[128];

	colour_effect_str(xxx, TTY_PLAIN_OP, "blue");
	colour_effect_str(reset, TTY_RESET_OP, NULL);
	printf("%s  MooshyBluePlain  %s", xxx, reset);

	colour_effect_str(xxx, TTY_BOLD_OP, "blue");
	printf("%s  MooshyBlueBold  %s", xxx, reset);

	colour_effect_str(xxx, TTY_BOLD_OP | TTY_UNDERLINE_OP, "yellow");
	printf("%s  MooshyYellowUnderBold  %s", xxx, reset);

	colour_effect_str(xxx, TTY_BOLD_OP, "red");
	printf("%s  MooshyRedBold  %s", xxx, reset);

	colour_effect_str(xxx, TTY_BOLD_OP, "orange");
	printf("%s  MooshyOrangeBold  %s", xxx, reset);

	colour_effect_str(xxx, TTY_BOLD_OP, "orange");
	printf("%s  MooshyOrangeBold  %s", xxx, reset);

	printf("\n");
	exit(0);
  }

  if (xxx_flag)
	exit(1);
#endif

  absolute_flag = 0;
  bs_si_flag = 0;
  colour_line_flag = 0;
  colour_long_only_flag = 1;
  dirsfirst_flag = 0;
  filetype_flag = 0;
  help_flag = 0;
  help_type = HELP_NONE;
  human_units_flag = 0;
  human_units_base_flag = 2;
  inum_flag = 0;
  long_flag = 0;
  lsd_flag = 0;
  lsf_flag = 0;
  lss_flag = 0;
  all_flag = almost_all_flag = almost_all_root_flag = 0;
  number_flag = 0;
  onelevel_flag = 0;
  recurse_flag = 1;
  single_column_flag = 0;
  symlink_Hflag = 1;

  /*
  groupname_flag = 1;
  username_flag = 1;
  */

  verbose_flag = 0;

  prefix_path_flag = NULL;
  dateformat_flag = NULL;
  format_flag = NULL;
  rgbpath_flag = NULL;
  limit_count_flag = 0;

  nfile_args = 0;
  hier = NULL;

  for (i = 0; i < COLOUR_EFFECT_NTYPES; i++)
	colour_effect[i].arg = colour_effect[i].effect_str = NULL;

  lf_colours_env = getenv("LF_COLOURS");
  colour_parse_flag_str(lf_colours_env);

  saw_file_args = saw_dot_file_arg = 0;

  if (streq(progname, "lsd")
	  || streq(progname, "lfd")
	  || strsuffix(progname, strlen(progname), "/lsd") != NULL
	  || strsuffix(progname, strlen(progname), "/lfd") != NULL) {
	  lsd_flag = 1;
  }
  else if (streq(progname, "lsf")
		   || strsuffix(progname, strlen(progname), "/lsf") != NULL) {
	lsf_flag = 1;
  }
  else if (streq(progname, "lss")
		   || strsuffix(progname, strlen(progname), "/lss") != NULL) {
	lss_flag = 1;
  }
  else if (streq(progname, "today")
		   || strsuffix(progname, strlen(progname), "/today") != NULL) {
	is_today_prog = 1;
	today_count_flag = TODAY_COUNT_WITHIN_DAYS;
  }

  if (argv[1] != NULL) {
	int found;
	char *first_flag;

	first_flag = argv[1];
	found = 0;
	if (streq(first_flag, "--lsd") || streq(first_flag, "--lfd")) {
	  lsd_flag = 1;
	  found = 1;
	}
	else if (streq(first_flag, "--lsf")) {
	  lsf_flag = 1;
	  found = 1;
	}
	else if (streq(first_flag, "--lss")) {
	  lss_flag = 1;
	  found = 1;
	}
	else if (streq(first_flag, "--today")) {
	  is_today_prog = 1;
	  today_count_flag = TODAY_COUNT_WITHIN_DAYS;
	  age_units = AGE_IN_DAYS;
	  found = 1;
	}

	if (found) {
	  argv++;
	  argc--;
	}
  }

  /* It's Command Line Argument time! */
  for (i = 1; i < argc; i++) {
	int c;
	char *flagp;

	if (argv[i][0] != '-')
	  break;

	flagp = &argv[i][1];
	if (*flagp == '\0') {
	  usage(HELP_ALL);
	  /*NOTREACHED*/
	}

  deconflate:
	if (*flagp == '\0')
	  continue;

	/* -h */
	if (*flagp == 'h') {
	  human_units_flag = 1;
	  which_bs_flag = BS_HUMAN_UNITS;
	  flagp++;
	  goto deconflate;
	}

	/* -H */
	if (*flagp == 'H') {
	  /* Applies only to command line symlinks. */
	  symlink_Hflag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -1 */
	if (!is_today_prog && *flagp == '1') {
	  /*
	   * If invoked as "today", "-1" flag means "-t1" so use "--single-column".
	   */
	  single_column_flag = 1;
	  merge_flag = 0;
	  long_flag = 0;
	  flagp++;
	  goto deconflate;
	}

	/* -a */
	if (*flagp == 'a') {
	  all_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -A */
	if (*flagp == 'A') {
	  /* XXX change to --abs because this is different; also see -I */
	  absolute_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -c */
	if (*flagp == 'c') {
	  time_type_flag = TIME_CTIME;
	  flagp++;
	  goto deconflate;
	}

	/* -i */
	if (*flagp == 'i') {
	  inum_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -F */
	if (*flagp == 'F') {
	  filetype_flag = 1;
	  /*
	   * POSIX says this flag implies not to follow symlink operands (on command
	   * line) unless -H or -L are also specified.
	   */
	  flagp++;
	  goto deconflate;
	}

	/* -m */
	if (*flagp == 'm') {
	  merge_flag = 1;
	  long_flag = 0;
	  single_column_flag = 0;
	  flagp++;
	  goto deconflate;
	}

	/* -n */
	if (*flagp == 'n') {
	  number_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -d */
	if (*flagp == 'd') {
	  onelevel_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -l */
	if (*flagp == 'l') {
	  long_flag = 1;
	  merge_flag = 0;
	  single_column_flag = 0;
	  flagp++;
	  goto deconflate;
	}

	/* -L */
	if (*flagp == 'L') {
	  /*
	   * This should apply to all listed files, not just those
	   * on the command line.
	   */
	  symlink_Lflag = 1;
	  symlink_Pflag = 0;
	  flagp++;
	  goto deconflate;
	}

	/* -p */
	if (*flagp == 'p') {
	  slash_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -P */
	if (*flagp == 'P') {
	  /* Applies only to command line symlinks. */
	  symlink_Pflag = 1;
	  symlink_Hflag = 0;
	  symlink_Lflag = 0;
	  flagp++;
	  goto deconflate;
	}

	/* -Q */
	if (*flagp == 'Q') {
	  quoted_flag++;
	  flagp++;
	  goto deconflate;
	}

	/* -r */
	if (*flagp == 'r') {
	  reverse_sort_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	if (*flagp == 'R') {
	  recurse_flag = -1;
	  onelevel_flag = 0;
	  flagp++;
	  goto deconflate;
	}

	/* -s */
	if (*flagp == 's') {
	  blocks_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -u */
	if (*flagp == 'u') {
	  time_type_flag = TIME_ATIME;
	  flagp++;
	  goto deconflate;
	}

	/* -v */
	if (*flagp == 'v') {
	  verbose_flag++;
	  flagp++;
	  goto deconflate;
	}

	/* -w */
	if (*flagp == 'w') {
	  unprintable_filenames_flag = 1;
	  flagp++;
	  goto deconflate;
	}

	/* -U */
	if (*flagp == 'U') {
	  time_type_flag = TIME_BTIME;
	  flagp++;
	  goto deconflate;
	}
	/* End of conflatable flags. */

	/* Start of individual/separate flags. */

	/*
	 * Various kinds of help.
	 * For full help, display it immediately and then exit.
	 * For special help, collect all requested types and then display them.
	 * Recognize some likely flag prefixes as appropriate.
	 */
	if (streq(argv[i], "--help")
		|| streq(argv[i], "--help-")
		|| streq(argv[i], "--help-all")) {
	  help_type = HELP_ALL;
	  help_flag = 1;
	  usage(help_type);
	  /*NOTREACHED*/
	}
    else if (streq(argv[i], "--help-colour")
			 || streq(argv[i], "--help-color")) {
      help_type |= HELP_CFLAGS;
	  help_flag = 1;
    }
    else if (streq(argv[i], "--help-format")
			 || streq(argv[i], "--help-form")) {
      help_type |= HELP_FFLAGS;
	  help_flag = 1;
	}
	else if (streq(argv[i], "--help-help")) {
      help_type |= HELP_HFLAGS;
	  help_flag = 1;
	}
	else if (streq(argv[i], "--help-individual")
			 || streq(argv[i], "--help-ind")) {
      help_type |= HELP_IFLAGS;
	  help_flag = 1;
	}
	else if (streq(argv[i], "--help-join")) {
      help_type |= HELP_JFLAGS;
	  help_flag = 1;
	}
	else if (streq(argv[i], "--help-sort")) {
      help_type |= HELP_SFLAGS;
	  help_flag = 1;
	}
	else if ((p = strprefix(argv[i], "--help")) != NULL) {
	  /*
	   * Handle conflated help types or an unrecognized type.
	   * Examples:
	   * --help // --helpi // --help-i // --help-cf
	   */
	  help_flag = 1;
	  /* Optional hyphen separator allowed. */
	  if (*p == '-')
		p++;
	  while (*p != '\0') {
		if (*p == 'a')
		  help_type |= HELP_ALL;
		else if (*p == 'c')
		  help_type |= HELP_CFLAGS;
		else if (*p == 'f')
		  help_type |= HELP_FFLAGS;
		else if (*p == 'h')
		  help_type |= HELP_HFLAGS;
		else if (*p == 'i')
		  help_type |= HELP_IFLAGS;
		else if (*p == 'j')
		  help_type |= HELP_JFLAGS;
		else if (*p == 's')
		  help_type |= HELP_SFLAGS;
		else {
		  /* An unrecognized help type. */
		  help_type |= HELP_HFLAGS;
		  break;
		}
		p++;
	  }
	}
	else if (streq(argv[i], "--version")) {
	  fprintf(stderr, "%s\n", version_string());
	  exit(0);
	  /*NOTREACHED*/
	}
	else if (streq(argv[i], "-Df")) {
	  /*
	   * Only when not recursing, list all directories first (using the
	   * sort method in effect), followed by all non-directories (reapplying
	   * the same sort method to them).
	   */
	  if (recurse_flag == 1)
		dirsfirst_flag = 1;
	}
	else if (streq(argv[i], "-Dd")) {
	  /* List only directories. */
	  lsd_flag = 1;
	}
	else if (streq(argv[i], "-Ds")) {
	  /* List only symbolic links. */
	  lss_flag = 1;
	  onelevel_flag = 1;
	}
	else if (streq(argv[i], "-Dx")) {
	  /* List only non-directories. */
	  lsf_flag = 1;
	  onelevel_flag = 1;
	}
    else if (streq(argv[i], "--aa"))
		almost_all_flag = 1;
	else if (streq(argv[i], "--aar")) {
	  if (getuid() == 0)
	    almost_all_flag = 1;
	  almost_all_root_flag = 1;
	}
	else if ((p = strprefix(argv[i], "--blocksize=")) != NULL) {
	  blocksize_units_flag = p + 1;
	  human_units_flag = 0;
	  which_bs_flag = BS_BS_UNITS;
    }
	else if ((c = colour_parse_flag_arg(argv + i)) != 0) {
	  if (c < 0) {
		fprintf(stderr, "Invalid parameter: \"%s\"\n", argv[i]);
		c++;
	  }
	  else
		i = i + (c - 1);
	}
	else if (streq(argv[i], "-f")) {
	  int count;

	  /*
	   * A namefile follows. It contains a sequence of filenames to list.
	   * Note that pathnames are literal - there is no globbing.
	   * This flag may be repeated.
	   */
	  if (++i == argc) {
		usage(HELP_IFLAGS);
		/*NOTREACHED*/
	  }
	  if ((count = add_namefile_files(&hier, argv[i])) == -1)
		exit(1);
	  if (count > 0) {
		nfile_args += count;
		namefile_flag++;
		saw_file_args = 1;
	  }
	}
	else if (streq(argv[i], "-fmt")) {
	  if (++i == argc) {
		usage(HELP_FFLAGS);
		/*NOTREACHED*/
	  }
	  format_flag = argv[i];
	}
	else if (streq(argv[i], "-dfmt")) {
	  if (++i == argc) {
		usage(HELP_ALL);
		/*NOTREACHED*/
	  }
	  dateformat_flag = argv[i];
	}
	else if (is_today_prog && argv[i][0] == '-' && isdigit(argv[i][1])) {
	  today_count = (unsigned int) strtoul(&argv[i][1], NULL, 10);
	  if ((today_count == 0 && errno == EINVAL) || errno == ERANGE) {
		fprintf(stderr, "Invalid limit count: \"%s\"\n", argv[i]);
		usage(HELP_ALL);
		/*NOTREACHED*/
	  }
	  today_count_flag = TODAY_COUNT_WITHIN_DAYS;
	}
	else if (is_today_prog && streq(argv[i], "-e")) {
	  today_count_flag = TODAY_COUNT_EXACT_DAYS;
	}
	else if (argv[i][0] == '-' && argv[i][1] == '-' && argv[i][2] == 'h') {
	  if (argv[i][3] == '\0')
		human_units_base_flag = 2;
	  else if (isdigit((int) argv[i][3])) {
		if ((human_units_base_flag = (int) strtol(&argv[i][3], NULL, 10)) < 0
			|| human_units_base_flag != 2
			|| human_units_base_flag != 10) {
		  usage(HELP_ALL);
		  /*NOTREACHED*/
		}
	  }
	  else {
		usage(HELP_ALL);
		/*NOTREACHED*/
	  }
	}
	else if (streq(argv[i], "-S")) {
	  if (++i == argc) {
		usage(HELP_ALL);
		/*NOTREACHED*/
	  }
	  prefix_path_flag = argv[i];
	}
	else if (streq(argv[i], "-k")) {
	  user_spec_blocksize = 1024;
	  which_bs_flag = BS_K_UNITS;
	  human_units_flag = 0;
	}
	else if (streq(argv[i], "--almost-all"))
	  almost_all_flag = 1;
	else if (streq(argv[i], "--almost-all-root")) {
	  if (getuid() == 0)
	    almost_all_flag = 1;
	  almost_all_root_flag = 1;
	}
	else if (streq(argv[i], "--bssi"))
	  bs_si_flag = 1;
	else if (streq(argv[i], "--esc"))
	  shell_escape_flag++;
	else if (streq(argv[i], "--recurse")) {
		char *ptr;

		/* --recurse # */
	    if (++i == argc) {
		usage(HELP_ALL);
		/*NOTREACHED*/
	  }
		recurse_flag = (int) strtol(argv[i], &ptr, 10);
		if ((recurse_flag == 0 && errno == EINVAL) || errno == ERANGE) {
		  usage(HELP_ALL);
		  /*NOTREACHED*/
		}

		if (recurse_flag == 0)
		  onelevel_flag = 1;
		else
		  onelevel_flag = 0;
	}
	else if (streq(argv[i], "--sa"))
	  sort_type_flag = SORT_FILENAME_AZ;
	else if (streq(argv[i], "--sai"))
	  sort_type_flag = SORT_FILENAME_Az;
	else if (streq(argv[i], "--sb"))
	  sort_type_flag = SORT_BIG_TO_SMALL;
 	else if (streq(argv[i], "--sg"))
	  sortgid_flag = 1;
 	else if (streq(argv[i], "--sl"))
	  sort_type_flag = SORT_DECREASING_LINK_COUNT;
	else if (streq(argv[i], "--sn"))
	  sort_type_flag = SORT_NEW_TO_OLD;
	else if (streq(argv[i], "--so"))
	  sort_type_flag = SORT_OLD_TO_NEW;
	else if (streq(argv[i], "--soff"))
	  sort_type_flag = SORT_UNSORTED;
	else if (streq(argv[i], "--ss"))
	  sort_type_flag = SORT_SMALL_TO_BIG;
	else if (streq(argv[i], "--su"))
	  sortuid_flag = 1;
	else if (streq(argv[i], "--sz"))
	  sort_type_flag = SORT_FILENAME_ZA;
	else if (streq(argv[i], "--szi"))
	  sort_type_flag = SORT_FILENAME_zA;
	else if (strneq(&argv[i][0], "-o", 2) && isdigit(argv[i][2])) {
	  limit_count_flag = (unsigned int) strtoul(&argv[i][2], NULL, 10);
	  if ((limit_count_flag == 0 || errno == EINVAL) || errno == ERANGE) {
		fprintf(stderr, "Invalid limit count: \"%s\"\n", argv[i]);
		usage(HELP_ALL);
		/*NOTREACHED*/
	  }
	}
	else if (argv[i][0] == '-' && argv[i][1] == 't') {
	  int opt;

	  /*
	   * -t[#] and -te[#]
	   *
	   * Select only files less than or equal to (or exactly) a certain age,
	   * according to (some) file time attribute, and relative to "now".
	   * The default is "today", which is equivalent to "-t0".
	   * The files' time attribute that is used in the comparison defaults
	   * to its modification time, but can be selected using
	   * '-c' (file status change time), '-u' (file data accessed time), or
	   * '-U' (file birth time, for FreeBSD only).
	   * The -te variant matches exactly # days ago (default is 0).
	   * A single letter time unit suffix may follow the number.
	   */
	  opt = 2;
	  today_count_flag = TODAY_COUNT_WITHIN_DAYS;
	  age_units = AGE_IN_DAYS;
	  if (argv[i][2] == 'e') {
		opt++;
		today_count_flag = TODAY_COUNT_EXACT_DAYS;
	  }

	  if (argv[i][opt] == '\0')
		today_count = 0;
	  else {
		char *endptr;

		endptr = NULL;
		today_count = (unsigned int) strtoul(&argv[i][opt], &endptr, 10);
		if ((today_count == 0 && errno == EINVAL) || errno == ERANGE) {
		  fprintf(stderr, "Invalid limit count: \"%s\"\n", argv[i]);
		  usage(HELP_ALL);
		  /*NOTREACHED*/
		}
		if (*endptr != '\0') {
		  if (*(endptr + 1) != '\0') {
		  bad_count_suffix:
			fprintf(stderr, "Invalid limit count suffix: \"%s\"\n", argv[i]);
			usage(HELP_ALL);
			/*NOTREACHED*/
		  }

		  if (*endptr == 'S')
			age_units = AGE_IN_SECONDS;
		  else if (*endptr == 'M')
			age_units = AGE_IN_MINUTES;
		  else if (*endptr == 'H')
			age_units = AGE_IN_HOURS;
		  else if (*endptr == 'd')
			age_units = AGE_IN_DAYS;
		  else if (*endptr == 'w')
			age_units = AGE_IN_WEEKS;
		  else if (*endptr == 'm')
			age_units = AGE_IN_MONTHS;
		  else if (*endptr == 'y')
			age_units = AGE_IN_YEARS;
		  else
			goto bad_count_suffix;
		}
	  }
	}
	else if (streq(argv[i], "--perms")) {
	  char *endptr, *octal_mode_str;
	  long m;

	  /* --perms PERMS */
	  if ((perms_match_flag != PERMS_MATCH_IGNORE) || ++i == argc) {
		usage(HELP_ALL);
		/*NOTREACHED*/
	  }

	  /*
	   * Select only files with mode permissions that match:
	   * --perms PERMS:  true if file permissions match exactly
	   * --perms +PERMS: true if all of PERMS are set, but others may also be
	   * --perms -PERMS: true if none of PERMS are set
	   */
	  octal_mode_str = argv[i];
	  if (*octal_mode_str == '+') {
		perms_match_flag = PERMS_MATCH_SET;
		octal_mode_str++;
	  }
	  else if (*octal_mode_str == '-') {
		perms_match_flag = PERMS_MATCH_NOTSET;
		octal_mode_str++;
	  }
	  else
		perms_match_flag = PERMS_MATCH_EXACT;

	  m = strtol(octal_mode_str, &endptr, 8);
	  if ((m == 0 && errno == EINVAL) || errno == ERANGE) {
		fprintf(stderr, "Invalid --perms argument.\n");
		exit(1);
	  }
	  perms_match = m & S_IPMT;
	}
	else if (streq(argv[i], "--type")) {
	  /*
	   * Select only files with any specified file type (the types are
	   * conflated into a string).  See file_type_match().
	   *
	   * Includes syntax from the -type flag of find(1):
	   * And indicators used by the -F flag of ls(1) - see file_type_ind():
	   * Plus some convenient additions and synonyms.
	   */
	  if (type_match_flag != NULL || ++i == argc) {
		usage(HELP_ALL);
		/*NOTREACHED*/
	  }
	  type_match_flag = argv[i];
	}
	else if (streq(argv[i], "--single-column")) {
	  single_column_flag = 1;
	}
	else if (streq(argv[i], "--")) {
	  i++;
	  break;
	}
	else {
	  usage(HELP_ALL);
	  /*NOTREACHED*/
	}
  }

  /*
   * The -H flag (i.e., follow symlinks for command line arguments) "is assumed
   * if none of the -F, -D, or -l options are specified" (FreeBSD).
   */
  if (!symlink_Lflag && !symlink_Pflag
	  && !onelevel_flag && !filetype_flag && !long_flag)
	symlink_Hflag = 1;

  if (symlink_Lflag || symlink_Pflag) {
	fprintf(stderr, "Unimplemented flag.\n");
	usage(HELP_JFLAGS);
	/*NOTREACHED*/
  }

  if (help_flag) {
	usage(help_type);
	/*NOTREACHED*/
  }

  /* If no count was provided, use the default. */
  if (is_today_prog && today_count == 0) {
	today_count = TODAY_COUNT_DEFAULT_DAYS;
	age_units = AGE_IN_DAYS;
  }

  if (format_flag == NULL) {
	if (long_flag)
	  line_format = DEFAULT_LONG_FORMAT;
	else
	  line_format = DEFAULT_SHORT_FORMAT;
  }
  else
	line_format = format_flag;

  /*
   * If output is not to a tty, single_column_flag is enabled unless
   * merge_flag has already been enabled
   * (e.g., when it's to a pipe). Colourization is turned off unless
   * it has been specifically requested.
   */
  if (isatty(1) != 1) {
	long_flag = 0;
	if (!merge_flag)
	  single_column_flag = 1;

	if (!colours_forced_flag) {
	  colours_enabled = -1;
	  colour_line_flag = 0;
	}
  }

  if (lsf_flag)
	recurse_flag = 0;

  /*
   * Long format implies single column in that the listing is one file
   * per line; i.e., there is an implicit newline.
   */
  if (long_flag)
	single_column_flag = 1;

  user_spec_blocksize = get_preferred_block_size();

  /*
   * Any other arguments left on the command line are files or directories.
   * If there are none, this is the same as having "." as the argument,
   * which lists the current directory.
   *
   * If one or more namefiles were given but there were no listable files,
   * list the current directory.
   */
  if (saw_file_args == 0 && argv[i] == NULL) {
	int st;

	/*
	 * There were no file arguments, so pretend "." was given.
	 * Allocate an extra element for the sentinel.
	 */
	hier = ALLOC_N(Hier *, 2);
	if ((st = is_listable_arg(".", &hier[0])) != 1) {
	  if (st == 0)
		fprintf(stderr, "%s: '.': No such file or directory\n", progname);
	  else
		fprintf(stderr, "%s: '.': Permission denied\n", progname);
	  exit(1);
	}

	nfile_args = 1;
  }
  else if (argv[i] != NULL) {
	int count, j, n, st;

	/*
	 * In addition to any namelist files, there are command line files.
	 * These are appended.
	 */
	count = argc - i;
	/* Allocate an extra for the sentinel. */
	hier = REALLOC_N(Hier *, hier, nfile_args + count + 1);
	n = nfile_args;
	for (j = 0; j < count; j++) {
	  if ((st = is_listable_arg(argv[i], &hier[n])) == 1) {
		if (argv[i][0] == '.')
		  saw_dot_file_arg = 1;
		i++;
		n++;
	  }
	  else {
		if (st == 0)
		  fprintf(stderr,
				  "%s: %s: No such file or directory\n", progname, argv[i]);
		else
		  fprintf(stderr, "%s: %s: Permission denied\n", progname, argv[i]);
	  }
	}

	if (n == nfile_args) {
	  /* No valid file arguments? */
	  exit(1);
	}

	nfile_args = n;
	saw_file_args = 1;
  }

  /* Append an EOL sentinel. */
  hier[nfile_args] = ALLOC(Hier);
  hier[nfile_args]->filename = NULL;
  hier[nfile_args]->st = -1;
  
  if (reverse_sort_flag) {
	switch (sort_type_flag) {
	case SORT_FILENAME_AZ:
	  sort_type_flag = SORT_FILENAME_ZA;
	  break;
	case SORT_FILENAME_ZA:
	  sort_type_flag = SORT_FILENAME_AZ;
	  break;
	case SORT_FILENAME_Az:	  
	  sort_type_flag = SORT_FILENAME_zA;
	  break;
	case SORT_FILENAME_zA:
	  sort_type_flag = SORT_FILENAME_AZ;
	  break;
	case SORT_BIG_TO_SMALL:
	  sort_type_flag = SORT_SMALL_TO_BIG;
	  break;
	case SORT_SMALL_TO_BIG:
	  sort_type_flag = SORT_BIG_TO_SMALL;
	  break;
	case SORT_NEW_TO_OLD:
	  sort_type_flag = SORT_OLD_TO_NEW;
	  break;
	case SORT_OLD_TO_NEW:
	  sort_type_flag = SORT_NEW_TO_OLD;
	  break;
	default:
	  break;
	}
  }

  if (nfile_args > 1)
	sort_file_args(nfile_args, hier);

  if (verbose_flag) {
	fprintf(stderr, "absolute_flag=%d\n", absolute_flag);
	fprintf(stderr, "all_flag=%d\n", all_flag);
	fprintf(stderr, "almost_all_flag=%d\n", almost_all_flag);
	fprintf(stderr, "almost_all_root_flag=%d\n", almost_all_root_flag);
	fprintf(stderr, "blocks_flag=%d\n", blocks_flag);
	fprintf(stderr, "colours_enabled=%d\n", colours_enabled);
	fprintf(stderr, "colours_forced_flag=%d\n", colours_forced_flag);
	fprintf(stderr, "colour_line_flag=%d\n", colour_line_flag);
	fprintf(stderr, "dirsfirst_flag=%d\n", dirsfirst_flag);
	fprintf(stderr, "filetype_flag=%d\n", filetype_flag);
	fprintf(stderr, "format_flag=\"%s\"\n", format_flag ? format_flag : "");
	fprintf(stderr, "groupname_flag=%d\n", groupname_flag);
	fprintf(stderr, "human_units_flag=%d\n", human_units_flag);
	fprintf(stderr, "inum_flag=%d\n", inum_flag);
	fprintf(stderr, "user_spec_blocksize=%u\n", user_spec_blocksize);
	fprintf(stderr, "limit_count_flag=%d\n", limit_count_flag);
	fprintf(stderr, "long_flag=%d\n", long_flag);
	fprintf(stderr, "lsd_flag=%d\n", lsd_flag);
	fprintf(stderr, "lsf_flag=%d\n", lsf_flag);
	fprintf(stderr, "lss_flag=%d\n", lss_flag);
	fprintf(stderr, "merge_flag=%d\n", merge_flag);
	fprintf(stderr, "onelevel_flag=%d\n", onelevel_flag);
	fprintf(stderr, "prefix_path_flag=\"%s\"\n",
			prefix_path_flag ? prefix_path_flag : "");
	fprintf(stderr, "recurse_flag=%d\n", recurse_flag);
	fprintf(stderr, "reverse_sort_flag=%d\n", reverse_sort_flag);
	fprintf(stderr, "rgbpath_flag=\"%s\"\n", rgbpath_flag ? rgbpath_flag : "");
	fprintf(stderr, "saw_file_args=%d\n", saw_file_args);
	fprintf(stderr, "single_column_flag=%d\n", single_column_flag);
	fprintf(stderr, "slash_flag=%d\n", slash_flag);
	fprintf(stderr, "sortgid_flag=%d\n", sortgid_flag);
	fprintf(stderr, "sort_type_flag=%d\n", sort_type_flag);
	fprintf(stderr, "sortuid_flag=%d\n", sortuid_flag);
	fprintf(stderr, "time_type_flag=%d\n", time_type_flag);
	fprintf(stderr, "username_flag=%d\n", username_flag);
	fprintf(stderr, "verbose_flag=%d\n", verbose_flag);
  }

  if (hier[0]->filename == NULL) {
	/* All done - there were no valid arguments to list. */
	exit(0);
  }

  /*
   * Read and process each entry.
   * For file arguments to ls(1):
   * o For each argument that names a file of a type other than directory,
   *   ls displays its name as well as any requested, associated information.
   * o For each argument that names a file of type directory, ls displays the
   *   names of files contained within that directory, as well as any requested,
   *   associated information.
   * o If no arguments are given, the contents of the current directory are
   *   displayed.
   * If more than one argument is given, non-directory arguments are
   * displayed first; directory and non-directory arguments are sorted
   * separately and in lexicographical order.
   */
  count = 0;
  nfiles = 0;
  file_set = NULL;

  for (i = 0; hier[i]->filename != NULL; i++) {
	/*
	 * Reporting the number of "blocks" used/allocated (as when requested by
	 * the -s and/or -l flags) is so underspecified and non-portable as to
	 * almost be useless.  For the -l flag, the number printed seems to
	 * be the same as the sum of the number of blocks for all files
	 * that are (or would be) reported by the -s flag.
	 *
	 * There are two units involved in the calculation of the blocks used:
	 * a) the number of bytes per block that the user wants applied (e.g.,
	 * 512B, 1KB, 4KB, ...) and b) the number of bytes per block used by
	 * the file system containing a file (e.g., 4KB).
	 *
	 * In the case of sparse files, the displayed file size
	 * (in bytes) may seem much larger than would appear to be the case from
	 * the number of blocks.  It is also possible for the file size (in bytes)
	 * to seem much smaller than would appear to be the case from the number
	 * of blocks, since a) there might be a minimum number of blocks used per
	 * file and b) the number of blocks used is integral and therefore
	 * always rounded up (e.g., a file containing 1 byte may use one
	 * 4KB block).
	 *
	 * Furthermore, when doing a recursive listing, it is possible to cross a
	 * mount point when listing a directory and different file systems may
	 * use a different bytes-per-block quantity.  This make computation of
	 * the number of blocks used by files somewhat more complicated since
	 * different files may reside on different file systems with different
	 * block sizes.
	 *
	 * The listing of a (complete named/recursed/implicit) directory's contents
	 * is preceded by a labeled total number of blocks (see below) used by the
	 * listed files (which may or may not include . and .. and other files that
	 * start with a dot, depending on other options).
	 * The default block size for the calculation is 512 bytes.
	 * The block size may be set with option -k or environment variable
	 * BLOCKSIZE.  Numbers of blocks in the output will have been rounded up
	 * so the numbers of bytes is at least as many as used by the corresponding
	 * file system blocks (which might have a different size).
	 *
	 * IEEE Std 1003.1 says that the block size is implementation-defined,
	 * as is the number of disk blocks occupied by a listed file, and that
	 * "-s uses implementation-defined units".
	 * Also that in any "long" listing, each list of files within the
	 * directory is preceded by a status line indicating the number of file
	 * system blocks occupied by files in the directory in 512-byte units,
	 * rounded up to the next integral number of units. The format is:
	 * "total %u\n", <number of units in the directory>
	 * If more than one directory, or a combination of non-directory files and
	 * directories are written, each list of files within a directory is to be
	 * preceded by:
	 * "\n%s:\n", <directory name>
	 * Except that if this string is the first output line, the first
	 * <newline> is omitted.
	 */
	if (!onelevel_flag && !lsd_flag
		&& long_flag && S_ISDIR(hier[i]->stat.st_mode)) {
		long du;

		if ((du = do_du(hier, hier[i]->filename)) != -1)
	      printf("total %ld (blocksize=%u)\n",
				 du * 512 / user_spec_blocksize, user_spec_blocksize);
		else
	      printf("total xxxx (blocksize=%u)\n", user_spec_blocksize);
	 }

	count += do_ls(hier, hier[i]->filename);
  }
  if (verbose_flag)
	fprintf(stderr, "fts_read_count=%d\n", fts_read_count);

  /* Output. */
  field_linenum_width = (int) log10((double) count) + 1;
  config_field_widths();

  if (verbose_flag) {
	fprintf(stderr, "field_dev_width=%d (st_dev_max=%lu)\n",
			field_dev_width, st_dev_max);
	fprintf(stderr, "field_rdev_width=%d (st_rdev_max=%lu)\n",
			field_rdev_width, st_rdev_max);
	fprintf(stderr, "field_nlink_width=%d (st_nlink_max=%lu)\n",
			field_nlink_width, st_nlink_max);
	fprintf(stderr, "field_ino_width=%d (st_ino_max=%lu)\n",
			field_ino_width, st_ino_max);
	fprintf(stderr, "field_blocks_width=%d (st_size_max=%ld)\n",
			field_blocks_width, st_size_max);
	fprintf(stderr, "field_size_width=%d (st_size_max=%ld)\n",
			field_size_width, st_size_max);
	fprintf(stderr, "field_uid_width=%d (st_uid_max=%ld)\n",
			field_uid_width, st_uid_max);
	fprintf(stderr, "field_gid_width=%d (st_gid_max=%ld)\n",
			field_gid_width, st_gid_max);
	fprintf(stderr, "field_pw_name_width=%d (pw_name_max=%ld)\n",
			field_pw_name_width, pw_name_max);
	fprintf(stderr, "field_gr_name_width=%d (gr_name_max=%ld)\n",
			field_gr_name_width, gr_name_max);
	fprintf(stderr, "path_name_max=%ld\n", path_name_max);
  }

  /*
   * In some cases, sorting over the entire list must be done here,
   * after all the items of interest have been collected.
   * Use a stable sort to not move items that compare equal here.
   *
   * mergesort(3) is non-standard so use the portable implementation
   * included with DACS.  Believe it or not, some GNU/Linux distros do not
   * include mergesort(3) or a substitute.
   */
#ifndef DACS_OS_LINUX
#endif
	bsd_mergesort(file_set, nfiles, sizeof(File_d *), compar_merge);

  /* If colours may be needed, try to load the RGB definition database. */
  if (colours_enabled == 1) {
	char *colour_rgb_path;

	if ((colour_rgb_path = colour_config_rgb_path(rgbpath_flag)) == NULL)
	  fprintf(stderr, "Could not find RGB file \"%s\"\n", rgbpath_flag);

	if ((ncolours = colour_load_rgb_file(colour_rgb_path)) == -1) {
	  fprintf(stderr, "Could not load rgb.txt\n");
	  colours_enabled = -1;
	}
	else {
	  if (verbose_flag)
		fprintf(stderr, "Got %d colours from \"%s\"\n",
				ncolours, colour_rgb_path);

	  /* This must happen after loading the RGB file. */
	  colour_compile_effects();
	}
  }

  /*
   * XXX
   * A long listing of a directory argument, and when recursing, should be
   * preceded by:
   * 1) a blank line (except when the first output line)
   * 2) the directory path prefix and a colon, and
   * 3) a count of the number of blocks used by the contents
   */

  if (single_column_flag) {
	for (i = 0; i < nfiles; i++) {
	  d = file_set[i];
	  if (format_line(line_format, d, i + 1, NULL))
		printf("\n");
	}
  }
  else if (merge_flag) {
	int c, nc;

	/* The last column number will be nc - 1. */
	if ((nc = get_tty_columns()) == -1)
	  nc = 80;

	c = 0;
	for (i = 0; i < nfiles; i++) {
	  int len;
	  Ds *ds;

	  d = file_set[i];
	  len = format_line(line_format, d, i + 1, &ds);
	  if (len > 0) {
		/*
		 * If this won't fit on the current line, and this isn't the
		 * beginning of a line, end the current line and start a new one.
		 */
		if (c && (c + len + 2) >= nc) {
		  printf(",\n%s", ds_buf(ds));
		  c = len;
		}
		else {
		  /* The first item on a line does not need a comma. */
		  if (c == 0) {
			printf("%s", ds_buf(ds));
			c += len;
		  }
		  else {
			/* End the current line with a comma, start a new line. */
			printf(", %s", ds_buf(ds));
			c += (len + 2);
		  }
		}
	  }
	  if (ds != NULL)
		ds_free(ds);
	}
	printf("\n");
  }
  else {
	int nc, w;
	size_t xxx;
	Printit_arg arg;

	if (verbose_flag) {
	  nc = get_tty_columns();
	  fprintf(stderr, "columns: %d\n", nc);
	  xxx = get_maxlen(file_set, nfiles, name_len, &w);
	  if (number_flag)
		xxx += field_linenum_width;
	  if (blocks_flag)
		xxx += field_blocks_width;
	  fprintf(stderr, "maxlen: %ld (%d)\n", xxx, w);
	}

	arg.line_format = line_format;
	arg.c = 1;
	prtable(file_set, nfiles, 0, 0, printit, &arg, name_len);

	exit(0);
  }

  if (verbose_flag) {
	fprintf(stderr, "%d files\n", nfiles);
	fprintf(stderr, "%lu pointers allocated\n", nfiles_alloc);
  }

  exit(0);
}
