
/*
 * Auxiliary cryptographic functions
 * These are mainly older functions that are needed for interoperability
 * or completeness reasons (such as for NTLM authentication) and for which we'd
 * rather not require any additional dependencies (e.g., OpenSSL).
 * Efficiency is not much of a consideration.
 *
 * Note: this should be standalone - do not bring in any DACS dependencies.
 * XXX This may lead to some redundant code, so fix
 *
 * RFC 2104 Appendix
 * https://tools.ietf.org/html/rfc2104
 */

/*
 * Copyright (c) 2003-2024
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "crypto_aux.h"

#ifndef MAYBE_UNUSED
#define MAYBE_UNUSED   __attribute__ ((__unused__))
#endif

#ifndef lint
static MAYBE_UNUSED const char copyright[] =
"Copyright (c) 2003-2024\n\
Distributed Systems Software.  All rights reserved.";
static MAYBE_UNUSED const char revid[] =
  "$Id: crypto_aux.c 3321 2024-06-19 23:39:55Z brachman $";
#endif

/*
 * rc4.c
 *
 * Copyright (c) 1996-2000 Whistle Communications, Inc.
 * All rights reserved.
 * 
 * Subject to the following obligations and disclaimer of warranty, use and
 * redistribution of this software, in source or object code forms, with or
 * without modifications are expressly permitted by Whistle Communications;
 * provided, however, that:
 * 1. Any and all reproductions of the source or object code must include the
 *    copyright notice above and the following disclaimer of warranties; and
 * 2. No rights are granted, in any manner or form, to use Whistle
 *    Communications, Inc. trademarks, including the mark "WHISTLE
 *    COMMUNICATIONS" on advertising, endorsements, or otherwise except as
 *    such appears in the above copyright notice or in the software.
 * 
 * THIS SOFTWARE IS BEING PROVIDED BY WHISTLE COMMUNICATIONS "AS IS", AND
 * TO THE MAXIMUM EXTENT PERMITTED BY LAW, WHISTLE COMMUNICATIONS MAKES NO
 * REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, REGARDING THIS SOFTWARE,
 * INCLUDING WITHOUT LIMITATION, ANY AND ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
 * WHISTLE COMMUNICATIONS DOES NOT WARRANT, GUARANTEE, OR MAKE ANY
 * REPRESENTATIONS REGARDING THE USE OF, OR THE RESULTS OF THE USE OF THIS
 * SOFTWARE IN TERMS OF ITS CORRECTNESS, ACCURACY, RELIABILITY OR OTHERWISE.
 * IN NO EVENT SHALL WHISTLE COMMUNICATIONS BE LIABLE FOR ANY DAMAGES
 * RESULTING FROM OR ARISING OUT OF ANY USE OF THIS SOFTWARE, INCLUDING
 * WITHOUT LIMITATION, ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * PUNITIVE, OR CONSEQUENTIAL DAMAGES, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES, LOSS OF USE, DATA OR PROFITS, HOWEVER CAUSED AND UNDER ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF WHISTLE COMMUNICATIONS IS ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */

/*
 * This implementation of the Alleged RC4 (ARC4) stream cipher is from FreeBSD:
 * /usr/src/sys/crypto/rc4/rc4.[ch]
 * $FreeBSD: releng/10.3/sys/crypto/rc4/rc4.c 241394 2012-10-10 08:36:38Z kevlo $
 * $FreeBSD: src/sys/crypto/rc4/rc4.h,v 1.3 2000/07/16 05:53:14 peter Exp $
 *
 * NOTE:
 * RC4 has a variety of severe cryptographic weaknesses and is considered to be
 * insecure, particularly if the beginning of the generated keystream is not
 * discarded or if nonrandom or related keys are used.
 *
 * It is included here only for interoperation with NTLM and should
 * not be used for any (other) applications that require secure encryption.
 */

static __inline void
swap_bytes(uint8_t *a, uint8_t *b)
{
  uint8_t temp;

  temp = *a;
  *a = *b;
  *b = temp;
}

/*
 * Initialize an RC4 state buffer using the supplied key,
 * which can have arbitrary length.
 */
RC4_ctx *
rc4_init(RC4_ctx *const state, const uint8_t *key, size_t keylen)
{
  size_t i, k;
  uint8_t j;

  /* Initialize state with identity permutation. */
  for (i = 0; i < 256; i++)
	state->perm[i] = (uint8_t) i; 

  state->index1 = 0;
  state->index2 = 0;
  
  /* Randomize the permutation using key data. */
  for (j = i = k = 0; i < 256; i++) {
	j += state->perm[i] + key[k]; 
	swap_bytes(&state->perm[i], &state->perm[j]);
	if (++k >= keylen)
	  k = 0;
  }

  return(state);
}

/*
 * Encrypt some data using the supplied RC4 state buffer.
 * The input and output buffers may be the same buffer.
 * Since RC4 is a stream cipher, this function is used
 * for both encryption and decryption.
 */
void
rc4_crypt(RC4_ctx *const state, const uint8_t *inbuf, uint8_t *outbuf,
		  size_t buflen)
{
  size_t i;
  uint8_t j;

  for (i = 0; i < buflen; i++) {
	/* Update modification indices. */
	state->index1++;
	state->index2 += state->perm[state->index1];

	/* Modify permutation. */
	swap_bytes(&state->perm[state->index1], &state->perm[state->index2]);

	/* Encrypt (or decrypt) next byte. */
	j = state->perm[state->index1] + state->perm[state->index2];
	outbuf[i] = inbuf[i] ^ state->perm[j];
  }
}

/*
 * This code implements the MD5 message-digest algorithm.
 * The algorithm is due to Ron Rivest.  This code was
 * written by Colin Plumb in 1993, no copyright is claimed.
 * This code is in the public domain; do with it what you wish.
 *
 * Equivalent code is available from RSA Data Security, Inc.
 * This code has been tested against that, and is equivalent,
 * except that you don't need to include two pages of legalese
 * with every copy.
 *
 * To compute the message digest of a chunk of bytes,
 * pass an MD5 context structure to MD5_init(), pass MD5_update()
 * a buffer of bytes as many times as necessary, and then call MD5_final to
 * fill the supplied 16-byte (MD5_DIGEST_LEN) array with the computed digest.
 *
 * Warning: this digest algorithm is no longer considered cryptographically
 * secure and is provided only for compatibility purposes and
 * non-cryptographic applications.  Use a better cryptographic digest
 * algorithm for applications requiring security.
 *
 * RFC 1321 https://tools.ietf.org/html/rfc1321
 */

/*
 * Includes minor modifications for DACS.
 */

static void MD5_transform(uint32_t buf[4], uint32_t const in[16]);

/*
 * Note: this code is harmless on little-endian machines.
 */
static void
byteReverse(uint8_t *buf, unsigned int longs)
{
  uint32_t t;

  do {
    t = (uint32_t) ((unsigned int) buf[3] << 8 | buf[2]) << 16
	  | ((unsigned int) buf[1] << 8 | buf[0]);
    *(uint32_t *) buf = t;
    buf += 4;
  } while (--longs);
}

/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
void
MD5_init(MD5_ctx *ctx)
{

  ctx->buf[0] = 0x67452301;
  ctx->buf[1] = 0xefcdab89;
  ctx->buf[2] = 0x98badcfe;
  ctx->buf[3] = 0x10325476;

  ctx->bits[0] = 0;
  ctx->bits[1] = 0;
}

/*
 * Update context to reflect the concatenation of another buffer full of bytes.
 */
void
MD5_update(MD5_ctx *ctx, const uint8_t *buf, size_t len)
{
  uint32_t t;

  /* Update bitcount */

  t = ctx->bits[0];
  if ((ctx->bits[0] = t + ((uint32_t) len << 3)) < t)
    ctx->bits[1]++;     /* Carry from low to high */
  ctx->bits[1] += len >> 29;

  t = (t >> 3) & 0x3f;    /* Bytes already in shsInfo->data. */

  /* Handle any leading odd-sized chunks. */
  if (t) {
    uint8_t *p = (uint8_t *) ctx->in + t;

    t = 64 - t;
    if (len < t) {
	  memmove(p, buf, len);
	  return;
    }

    memmove(p, buf, t);
    byteReverse(ctx->in, 16);
    MD5_transform(ctx->buf, (uint32_t *) ctx->in);
    buf += t;
    len -= t;
  }

  /* Process data in 64-byte chunks. */
  while (len >= 64) {
    memmove(ctx->in, buf, 64);
    byteReverse(ctx->in, 16);
    MD5_transform(ctx->buf, (uint32_t *) ctx->in);
    buf += 64;
    len -= 64;
  }

  /* Handle any remaining bytes of data. */
  memmove(ctx->in, buf, len);
}

/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, MSB-first).
 * DIGEST is a 16-byte buffer (MD5_DIGEST_LEN) where the result is copied.
 */
uint8_t *
MD5_final(MD5_ctx *ctx, uint8_t *digest)
{
  unsigned int count;
  uint8_t *p;

  /* Compute number of bytes mod 64. */
  count = (ctx->bits[0] >> 3) & 0x3F;

  /*
   * Set the first char of padding to 0x80.  This is safe since there is
   * always at least one byte free.
   */
  p = ctx->in + count;
  *p++ = 0x80;

  /* Bytes of padding needed to make 64 bytes. */
  count = 64 - 1 - count;

  /* Pad out to 56 mod 64. */
  if (count < 8) {
    /* Two lots of padding:  Pad the first block to 64 bytes. */
    memset(p, 0, count);
    byteReverse(ctx->in, 16);
    MD5_transform(ctx->buf, (uint32_t *) ctx->in);

    /* Now fill the next block with 56 bytes. */
    memset(ctx->in, 0, 56);
  } else {
    /* Pad block to 56 bytes. */
    memset(p, 0, count - 8);
  }
  byteReverse(ctx->in, 14);

  /* Append length in bits and transform. */
  ((uint32_t *) ctx->in)[14] = ctx->bits[0];
  ((uint32_t *) ctx->in)[15] = ctx->bits[1];

  MD5_transform(ctx->buf, (uint32_t *) ctx->in);
  byteReverse((uint8_t *) ctx->buf, 4);
  memmove(digest, ctx->buf, 16);
  memset(ctx, 0, sizeof(*ctx));   /* In case it's sensitive. */

  return(digest);
}

/* The four core functions - F1 is optimized somewhat. */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f, w, x, y, z, data, s)							\
  ( w += f(x, y, z) + data,  w = w<<s | w>>(32-s),  w += x )

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
static void
MD5_transform(uint32_t buf[4], uint32_t const in[16])
{
  uint32_t a, b, c, d;

  a = buf[0];
  b = buf[1];
  c = buf[2];
  d = buf[3];

  MD5STEP(F1, a, b, c, d, in[ 0] + 0xd76aa478, 7);
  MD5STEP(F1, d, a, b, c, in[ 1] + 0xe8c7b756, 12);
  MD5STEP(F1, c, d, a, b, in[ 2] + 0x242070db, 17);
  MD5STEP(F1, b, c, d, a, in[ 3] + 0xc1bdceee, 22);
  MD5STEP(F1, a, b, c, d, in[ 4] + 0xf57c0faf, 7);
  MD5STEP(F1, d, a, b, c, in[ 5] + 0x4787c62a, 12);
  MD5STEP(F1, c, d, a, b, in[ 6] + 0xa8304613, 17);
  MD5STEP(F1, b, c, d, a, in[ 7] + 0xfd469501, 22);
  MD5STEP(F1, a, b, c, d, in[ 8] + 0x698098d8, 7);
  MD5STEP(F1, d, a, b, c, in[ 9] + 0x8b44f7af, 12);
  MD5STEP(F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
  MD5STEP(F1, b, c, d, a, in[11] + 0x895cd7be, 22);
  MD5STEP(F1, a, b, c, d, in[12] + 0x6b901122, 7);
  MD5STEP(F1, d, a, b, c, in[13] + 0xfd987193, 12);
  MD5STEP(F1, c, d, a, b, in[14] + 0xa679438e, 17);
  MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

  MD5STEP(F2, a, b, c, d, in[ 1] + 0xf61e2562, 5);
  MD5STEP(F2, d, a, b, c, in[ 6] + 0xc040b340, 9);
  MD5STEP(F2, c, d, a, b, in[11] + 0x265e5a51, 14);
  MD5STEP(F2, b, c, d, a, in[ 0] + 0xe9b6c7aa, 20);
  MD5STEP(F2, a, b, c, d, in[ 5] + 0xd62f105d, 5);
  MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
  MD5STEP(F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
  MD5STEP(F2, b, c, d, a, in[ 4] + 0xe7d3fbc8, 20);
  MD5STEP(F2, a, b, c, d, in[ 9] + 0x21e1cde6, 5);
  MD5STEP(F2, d, a, b, c, in[14] + 0xc33707d6, 9);
  MD5STEP(F2, c, d, a, b, in[ 3] + 0xf4d50d87, 14);
  MD5STEP(F2, b, c, d, a, in[ 8] + 0x455a14ed, 20);
  MD5STEP(F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
  MD5STEP(F2, d, a, b, c, in[ 2] + 0xfcefa3f8, 9);
  MD5STEP(F2, c, d, a, b, in[ 7] + 0x676f02d9, 14);
  MD5STEP(F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

  MD5STEP(F3, a, b, c, d, in[ 5] + 0xfffa3942, 4);
  MD5STEP(F3, d, a, b, c, in[ 8] + 0x8771f681, 11);
  MD5STEP(F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
  MD5STEP(F3, b, c, d, a, in[14] + 0xfde5380c, 23);
  MD5STEP(F3, a, b, c, d, in[ 1] + 0xa4beea44, 4);
  MD5STEP(F3, d, a, b, c, in[ 4] + 0x4bdecfa9, 11);
  MD5STEP(F3, c, d, a, b, in[ 7] + 0xf6bb4b60, 16);
  MD5STEP(F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
  MD5STEP(F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
  MD5STEP(F3, d, a, b, c, in[ 0] + 0xeaa127fa, 11);
  MD5STEP(F3, c, d, a, b, in[ 3] + 0xd4ef3085, 16);
  MD5STEP(F3, b, c, d, a, in[ 6] + 0x04881d05, 23);
  MD5STEP(F3, a, b, c, d, in[ 9] + 0xd9d4d039, 4);
  MD5STEP(F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
  MD5STEP(F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
  MD5STEP(F3, b, c, d, a, in[ 2] + 0xc4ac5665, 23);

  MD5STEP(F4, a, b, c, d, in[ 0] + 0xf4292244, 6);
  MD5STEP(F4, d, a, b, c, in[ 7] + 0x432aff97, 10);
  MD5STEP(F4, c, d, a, b, in[14] + 0xab9423a7, 15);
  MD5STEP(F4, b, c, d, a, in[ 5] + 0xfc93a039, 21);
  MD5STEP(F4, a, b, c, d, in[12] + 0x655b59c3, 6);
  MD5STEP(F4, d, a, b, c, in[ 3] + 0x8f0ccc92, 10);
  MD5STEP(F4, c, d, a, b, in[10] + 0xffeff47d, 15);
  MD5STEP(F4, b, c, d, a, in[ 1] + 0x85845dd1, 21);
  MD5STEP(F4, a, b, c, d, in[ 8] + 0x6fa87e4f, 6);
  MD5STEP(F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
  MD5STEP(F4, c, d, a, b, in[ 6] + 0xa3014314, 15);
  MD5STEP(F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
  MD5STEP(F4, a, b, c, d, in[ 4] + 0xf7537e82, 6);
  MD5STEP(F4, d, a, b, c, in[11] + 0xbd3af235, 10);
  MD5STEP(F4, c, d, a, b, in[ 2] + 0x2ad7d2bb, 15);
  MD5STEP(F4, b, c, d, a, in[ 9] + 0xeb86d391, 21);

  buf[0] += a;
  buf[1] += b;
  buf[2] += c;
  buf[3] += d;
}

uint8_t *
crypto_hmac_md5(uint8_t *key, size_t key_len, uint8_t *text, size_t text_len,
				uint8_t **inner, size_t *ilen, uint8_t **hmac, size_t *hlen)
{
  int i;
  MD5_ctx context;
  uint8_t *digest;
  uint8_t k_ipad[65];    /* inner padding - key XORd with ipad */
  uint8_t k_opad[65];    /* outer padding - key XORd with opad */
  uint8_t tk[16];

  /* If key is longer than 64 bytes, reset it to key=MD5(key). */
  if (key_len > 64) {
	MD5_ctx tctx;

	MD5_init(&tctx);
	MD5_update(&tctx, key, key_len);
	MD5_final(&tctx, tk);

	key = tk;
	key_len = 16;
  }

  /*
   * The HMAC_MD5 transform looks like:
   *
   * MD5(K XOR opad, MD5(K XOR ipad, text))
   *
   * where K is an n byte key
   * ipad is the byte 0x36 repeated 64 times
   * opad is the byte 0x5c repeated 64 times
   * and text is the data being protected
   */

  /* Start out by storing key in pads. */
  memset(k_ipad, 0, sizeof(k_ipad));
  memset(k_opad, 0, sizeof(k_opad));
  memcpy(k_ipad, key, key_len);
  memcpy(k_opad, key, key_len);

  /* XOR key with ipad and opad values. */
  for (i = 0; i < 64; i++) {
	k_ipad[i] ^= 0x36;
	k_opad[i] ^= 0x5c;
  }

  if ((digest = (uint8_t *) malloc(16)) == NULL) {
	fprintf(stderr, "crypto_hmac_md5: malloc failed.\n");
	exit(1);
  }
  *hmac = digest;
  *hlen = 16;

  /* Perform inner MD5. */
  MD5_init(&context);                   /* init context for 1st pass */
  MD5_update(&context, k_ipad, 64);      /* start with inner pad */
  MD5_update(&context, text, text_len); /* then text of datagram */
  MD5_final(&context, digest);          /* finish up 1st pass */

  if (inner != NULL) {
	if ((*inner = (uint8_t *) malloc(16)) == NULL) {
	  fprintf(stderr, "crypto_hmac_md5: malloc failed.\n");
	  exit(1);
	}
	memcpy(*inner, digest, 16);
	*ilen = 16;
  }

  /* Perform outer MD5. */
  MD5_init(&context);                   /* init context for 2nd pass */
  MD5_update(&context, k_opad, 64);     /* start with outer pad */
  MD5_update(&context, digest, 16);     /* then results of 1st hash */
  MD5_final(&context, digest);          /* finish up 2nd pass */

  return(digest);
}

/*
 * This is an OpenSSL-compatible implementation of the RSA Data Security, Inc.
 * MD4 Message-Digest Algorithm (RFC 1320).
 * https://tools.ietf.org/html/rfc1320
 *
 * Homepage:
 * http://openwall.info/wiki/people/solar/software/public-domain-source-code/md4
 *
 * Author:
 * Alexander Peslyak, better known as Solar Designer <solar at openwall.com>
 *
 * This software was written by Alexander Peslyak in 2001.  No copyright is
 * claimed, and the software is hereby placed in the public domain.
 * In case this attempt to disclaim copyright and place the software in the
 * public domain is deemed null and void, then the software is
 * Copyright (c) 2001 Alexander Peslyak and it is hereby released to the
 * general public under the following terms:
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted.
 *
 * There's ABSOLUTELY NO WARRANTY, express or implied.
 *
 * (This is a heavily cut-down "BSD license".)
 *
 * This differs from Colin Plumb's older public domain implementation in that
 * no exactly 32-bit integer data type is required (any 32-bit or wider
 * unsigned integer data type will do), there's no compile-time endianness
 * configuration, and the function prototypes match OpenSSL's.  No code from
 * Colin Plumb's implementation has been reused; this comment merely compares
 * the properties of the two independent implementations.
 *
 * The primary goals of this implementation are portability and ease of use.
 * It is meant to be fast, but not as fast as possible.  Some known
 * optimizations are not included to reduce source code size and avoid
 * compile-time configuration.
 */

/*
 * Includes minor modifications for DACS.
 */

/*
 * The basic MD4 functions.
 *
 * F and G are optimized compared to their RFC 1320 definitions, with the
 * optimization for F borrowed from Colin Plumb's MD5 implementation.
 */
#define F(x, y, z)          ((z) ^ ((x) & ((y) ^ (z))))
#define G(x, y, z)          (((x) & ((y) | (z))) | ((y) & (z)))
#define H(x, y, z)          ((x) ^ (y) ^ (z))

/*
 * The MD4 transformation for all three rounds.
 */
#define STEP(f, a, b, c, d, x, s)								\
  (a) += f((b), (c), (d)) + (x);								\
  (a) = (((a) << (s)) | (((a) & 0xffffffff) >> (32 - (s))));

/*
 * SET reads 4 input bytes in little-endian byte order and stores them
 * in a properly aligned word in host byte order.
 *
 * The check for little-endian architectures that tolerate unaligned
 * memory accesses is just an optimization.  Nothing will break if it
 * doesn't work.
 */
#if defined(__i386__) || defined(__x86_64__)
#define SET(n)				(*(MD4_u32plus *) &ptr[(n) * 4])
#define GET(n)				SET(n)

#else

#define SET(n)								\
  (ctx->block[(n)] =						\
   (MD4_u32plus)ptr[(n) * 4] |				\
   ((MD4_u32plus)ptr[(n) * 4 + 1] << 8) |	\
   ((MD4_u32plus)ptr[(n) * 4 + 2] << 16) |	\
   ((MD4_u32plus)ptr[(n) * 4 + 3] << 24))
#define GET(n)				(ctx->block[(n)])
#endif

/*
 * This processes one or more 64-byte data blocks, but does NOT update
 * the bit counters.  There are no alignment requirements.
 */
static const void *
body(MD4_ctx *ctx, const void *data, unsigned long size)
{
  const unsigned char *ptr;
  MD4_u32plus a, b, c, d;
  MD4_u32plus saved_a, saved_b, saved_c, saved_d;

  ptr = (const unsigned char *) data;

  a = ctx->a;
  b = ctx->b;
  c = ctx->c;
  d = ctx->d;

  do {
	saved_a = a;
	saved_b = b;
	saved_c = c;
	saved_d = d;

	/* Round 1 */
	STEP(F, a, b, c, d, SET(0), 3);
	STEP(F, d, a, b, c, SET(1), 7);
	STEP(F, c, d, a, b, SET(2), 11);
	STEP(F, b, c, d, a, SET(3), 19);
	STEP(F, a, b, c, d, SET(4), 3);
	STEP(F, d, a, b, c, SET(5), 7);
	STEP(F, c, d, a, b, SET(6), 11);
	STEP(F, b, c, d, a, SET(7), 19);
	STEP(F, a, b, c, d, SET(8), 3);
	STEP(F, d, a, b, c, SET(9), 7);
	STEP(F, c, d, a, b, SET(10), 11);
	STEP(F, b, c, d, a, SET(11), 19);
	STEP(F, a, b, c, d, SET(12), 3);
	STEP(F, d, a, b, c, SET(13), 7);
	STEP(F, c, d, a, b, SET(14), 11);
	STEP(F, b, c, d, a, SET(15), 19);

	/* Round 2 */
	STEP(G, a, b, c, d, GET(0) + 0x5a827999, 3);
	STEP(G, d, a, b, c, GET(4) + 0x5a827999, 5);
	STEP(G, c, d, a, b, GET(8) + 0x5a827999, 9);
	STEP(G, b, c, d, a, GET(12) + 0x5a827999, 13);
	STEP(G, a, b, c, d, GET(1) + 0x5a827999, 3);
	STEP(G, d, a, b, c, GET(5) + 0x5a827999, 5);
	STEP(G, c, d, a, b, GET(9) + 0x5a827999, 9);
	STEP(G, b, c, d, a, GET(13) + 0x5a827999, 13);
	STEP(G, a, b, c, d, GET(2) + 0x5a827999, 3);
	STEP(G, d, a, b, c, GET(6) + 0x5a827999, 5);
	STEP(G, c, d, a, b, GET(10) + 0x5a827999, 9);
	STEP(G, b, c, d, a, GET(14) + 0x5a827999, 13);
	STEP(G, a, b, c, d, GET(3) + 0x5a827999, 3);
	STEP(G, d, a, b, c, GET(7) + 0x5a827999, 5);
	STEP(G, c, d, a, b, GET(11) + 0x5a827999, 9);
	STEP(G, b, c, d, a, GET(15) + 0x5a827999, 13);

	/* Round 3 */
	STEP(H, a, b, c, d, GET(0) + 0x6ed9eba1, 3);
	STEP(H, d, a, b, c, GET(8) + 0x6ed9eba1, 9);
	STEP(H, c, d, a, b, GET(4) + 0x6ed9eba1, 11);
	STEP(H, b, c, d, a, GET(12) + 0x6ed9eba1, 15);
	STEP(H, a, b, c, d, GET(2) + 0x6ed9eba1, 3);
	STEP(H, d, a, b, c, GET(10) + 0x6ed9eba1, 9);
	STEP(H, c, d, a, b, GET(6) + 0x6ed9eba1, 11);
	STEP(H, b, c, d, a, GET(14) + 0x6ed9eba1, 15);
	STEP(H, a, b, c, d, GET(1) + 0x6ed9eba1, 3);
	STEP(H, d, a, b, c, GET(9) + 0x6ed9eba1, 9);
	STEP(H, c, d, a, b, GET(5) + 0x6ed9eba1, 11);
	STEP(H, b, c, d, a, GET(13) + 0x6ed9eba1, 15);
	STEP(H, a, b, c, d, GET(3) + 0x6ed9eba1, 3);
	STEP(H, d, a, b, c, GET(11) + 0x6ed9eba1, 9);
	STEP(H, c, d, a, b, GET(7) + 0x6ed9eba1, 11);
	STEP(H, b, c, d, a, GET(15) + 0x6ed9eba1, 15);

	a += saved_a;
	b += saved_b;
	c += saved_c;
	d += saved_d;

	ptr += 64;
  } while (size -= 64);

  ctx->a = a;
  ctx->b = b;
  ctx->c = c;
  ctx->d = d;

  return(ptr);
}

void
MD4_init(MD4_ctx *ctx)
{

  ctx->a = 0x67452301;
  ctx->b = 0xefcdab89;
  ctx->c = 0x98badcfe;
  ctx->d = 0x10325476;

  ctx->lo = 0;
  ctx->hi = 0;
}

void
MD4_update(MD4_ctx *ctx, const void *data, unsigned long size)
{
  MD4_u32plus saved_lo;
  unsigned long used, available;

  saved_lo = ctx->lo;
  if ((ctx->lo = (saved_lo + size) & 0x1fffffff) < saved_lo)
	ctx->hi++;
  ctx->hi += size >> 29;

  used = saved_lo & 0x3f;

  if (used) {
	available = 64 - used;

	if (size < available) {
	  memcpy(&ctx->buffer[used], data, size);
	  return;
	}

	memcpy(&ctx->buffer[used], data, available);
	data = (const unsigned char *) data + available;
	size -= available;
	body(ctx, ctx->buffer, 64);
  }

  if (size >= 64) {
	data = body(ctx, data, size & ~(unsigned long) 0x3f);
	size &= 0x3f;
  }

  memcpy(ctx->buffer, data, size);
}

void
MD4_final(MD4_ctx *ctx, unsigned char *result)
{
  unsigned long used, available;

  used = ctx->lo & 0x3f;

  ctx->buffer[used++] = 0x80;

  available = 64 - used;

  if (available < 8) {
	memset(&ctx->buffer[used], 0, available);
	body(ctx, ctx->buffer, 64);
	used = 0;
	available = 64;
  }

  memset(&ctx->buffer[used], 0, available - 8);

  ctx->lo <<= 3;
  ctx->buffer[56] = ctx->lo;
  ctx->buffer[57] = ctx->lo >> 8;
  ctx->buffer[58] = ctx->lo >> 16;
  ctx->buffer[59] = ctx->lo >> 24;
  ctx->buffer[60] = ctx->hi;
  ctx->buffer[61] = ctx->hi >> 8;
  ctx->buffer[62] = ctx->hi >> 16;
  ctx->buffer[63] = ctx->hi >> 24;

  body(ctx, ctx->buffer, 64);

  result[ 0] = ctx->a;
  result[ 1] = ctx->a >> 8;
  result[ 2] = ctx->a >> 16;
  result[ 3] = ctx->a >> 24;
  result[ 4] = ctx->b;
  result[ 5] = ctx->b >> 8;
  result[ 6] = ctx->b >> 16;
  result[ 7] = ctx->b >> 24;
  result[ 8] = ctx->c;
  result[ 9] = ctx->c >> 8;
  result[10] = ctx->c >> 16;
  result[11] = ctx->c >> 24;
  result[12] = ctx->d;
  result[13] = ctx->d >> 8;
  result[14] = ctx->d >> 16;
  result[15] = ctx->d >> 24;

  memset(ctx, 0, sizeof(*ctx));
}

/*
 * Slightly modified for use with DACS, but originally:
 *
 * FIPS pub 180-1: Secure Hash Algorithm (SHA-1)
 * based on: http://csrc.nist.gov/fips/fip180-1.txt
 * implemented by Jun-ichiro itojun Itoh <itojun@itojun.org>
 *
 * $KAME: sha1.c,v 1.5 2000/11/08 06:13:08 itojun Exp $
 *
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *    84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1
 * A million repetitions of "a":
 *    34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F
 */

/* Constant table */
static u_int32_t _K[] = { 0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xca62c1d6 };
#define	K(t)	_K[(t) / 20]

#undef F0
#undef F1
#undef F2
#undef F3
#undef H

#define	F0(b, c, d)	(((b) & (c)) | ((~(b)) & (d)))
#define	F1(b, c, d)	(((b) ^ (c)) ^ (d))
#define	F2(b, c, d)	(((b) & (c)) | ((b) & (d)) | ((c) & (d)))
#define	F3(b, c, d)	(((b) ^ (c)) ^ (d))

#define	S(n, x)		(((x) << (n)) | ((x) >> (32 - n)))

#define	H(n)	(ctxt->h.b32[(n)])
#define	COUNT	(ctxt->count)
#define	BCOUNT	(ctxt->c.b64[0] / 8)
#define	W(n)	(ctxt->m.b32[(n)])

#define	PUTBYTE(x)	{					\
	ctxt->m.b8[(COUNT % 64)] = (x);		\
	COUNT++;							\
	COUNT %= 64;						\
	ctxt->c.b64[0] += 8;				\
	if (COUNT % 64 == 0)				\
	  sha1_step(ctxt);					\
  }

#define	PUTPAD(x)	{					\
	ctxt->m.b8[(COUNT % 64)] = (x);		\
	COUNT++;							\
	COUNT %= 64;						\
	if (COUNT % 64 == 0)				\
	  sha1_step(ctxt);					\
  }

static void sha1_step(struct SHA1_ctx *);

static void
sha1_step(SHA1_ctx *ctxt)
{
  u_int32_t	a, b, c, d, e;
  size_t t, s;
  u_int32_t	tmp;
  
#if BYTE_ORDER == LITTLE_ENDIAN
  SHA1_ctx tctxt;

  bcopy(&ctxt->m.b8[0], &tctxt.m.b8[0], 64);
  ctxt->m.b8[0] = tctxt.m.b8[3]; ctxt->m.b8[1] = tctxt.m.b8[2];
  ctxt->m.b8[2] = tctxt.m.b8[1]; ctxt->m.b8[3] = tctxt.m.b8[0];
  ctxt->m.b8[4] = tctxt.m.b8[7]; ctxt->m.b8[5] = tctxt.m.b8[6];
  ctxt->m.b8[6] = tctxt.m.b8[5]; ctxt->m.b8[7] = tctxt.m.b8[4];
  ctxt->m.b8[8] = tctxt.m.b8[11]; ctxt->m.b8[9] = tctxt.m.b8[10];
  ctxt->m.b8[10] = tctxt.m.b8[9]; ctxt->m.b8[11] = tctxt.m.b8[8];
  ctxt->m.b8[12] = tctxt.m.b8[15]; ctxt->m.b8[13] = tctxt.m.b8[14];
  ctxt->m.b8[14] = tctxt.m.b8[13]; ctxt->m.b8[15] = tctxt.m.b8[12];
  ctxt->m.b8[16] = tctxt.m.b8[19]; ctxt->m.b8[17] = tctxt.m.b8[18];
  ctxt->m.b8[18] = tctxt.m.b8[17]; ctxt->m.b8[19] = tctxt.m.b8[16];
  ctxt->m.b8[20] = tctxt.m.b8[23]; ctxt->m.b8[21] = tctxt.m.b8[22];
  ctxt->m.b8[22] = tctxt.m.b8[21]; ctxt->m.b8[23] = tctxt.m.b8[20];
  ctxt->m.b8[24] = tctxt.m.b8[27]; ctxt->m.b8[25] = tctxt.m.b8[26];
  ctxt->m.b8[26] = tctxt.m.b8[25]; ctxt->m.b8[27] = tctxt.m.b8[24];
  ctxt->m.b8[28] = tctxt.m.b8[31]; ctxt->m.b8[29] = tctxt.m.b8[30];
  ctxt->m.b8[30] = tctxt.m.b8[29]; ctxt->m.b8[31] = tctxt.m.b8[28];
  ctxt->m.b8[32] = tctxt.m.b8[35]; ctxt->m.b8[33] = tctxt.m.b8[34];
  ctxt->m.b8[34] = tctxt.m.b8[33]; ctxt->m.b8[35] = tctxt.m.b8[32];
  ctxt->m.b8[36] = tctxt.m.b8[39]; ctxt->m.b8[37] = tctxt.m.b8[38];
  ctxt->m.b8[38] = tctxt.m.b8[37]; ctxt->m.b8[39] = tctxt.m.b8[36];
  ctxt->m.b8[40] = tctxt.m.b8[43]; ctxt->m.b8[41] = tctxt.m.b8[42];
  ctxt->m.b8[42] = tctxt.m.b8[41]; ctxt->m.b8[43] = tctxt.m.b8[40];
  ctxt->m.b8[44] = tctxt.m.b8[47]; ctxt->m.b8[45] = tctxt.m.b8[46];
  ctxt->m.b8[46] = tctxt.m.b8[45]; ctxt->m.b8[47] = tctxt.m.b8[44];
  ctxt->m.b8[48] = tctxt.m.b8[51]; ctxt->m.b8[49] = tctxt.m.b8[50];
  ctxt->m.b8[50] = tctxt.m.b8[49]; ctxt->m.b8[51] = tctxt.m.b8[48];
  ctxt->m.b8[52] = tctxt.m.b8[55]; ctxt->m.b8[53] = tctxt.m.b8[54];
  ctxt->m.b8[54] = tctxt.m.b8[53]; ctxt->m.b8[55] = tctxt.m.b8[52];
  ctxt->m.b8[56] = tctxt.m.b8[59]; ctxt->m.b8[57] = tctxt.m.b8[58];
  ctxt->m.b8[58] = tctxt.m.b8[57]; ctxt->m.b8[59] = tctxt.m.b8[56];
  ctxt->m.b8[60] = tctxt.m.b8[63]; ctxt->m.b8[61] = tctxt.m.b8[62];
  ctxt->m.b8[62] = tctxt.m.b8[61]; ctxt->m.b8[63] = tctxt.m.b8[60];
#endif

  a = H(0); b = H(1); c = H(2); d = H(3); e = H(4);

  for (t = 0; t < 20; t++) {
	s = t & 0x0f;
	if (t >= 16) {
	  W(s) = S(1, W((s+13) & 0x0f) ^ W((s+8) & 0x0f) ^ W((s+2) & 0x0f) ^ W(s));
	}
	tmp = S(5, a) + F0(b, c, d) + e + W(s) + K(t);
	e = d; d = c; c = S(30, b); b = a; a = tmp;
  }
  for (t = 20; t < 40; t++) {
	s = t & 0x0f;
	W(s) = S(1, W((s+13) & 0x0f) ^ W((s+8) & 0x0f) ^ W((s+2) & 0x0f) ^ W(s));
	tmp = S(5, a) + F1(b, c, d) + e + W(s) + K(t);
	e = d; d = c; c = S(30, b); b = a; a = tmp;
  }
  for (t = 40; t < 60; t++) {
	s = t & 0x0f;
	W(s) = S(1, W((s+13) & 0x0f) ^ W((s+8) & 0x0f) ^ W((s+2) & 0x0f) ^ W(s));
	tmp = S(5, a) + F2(b, c, d) + e + W(s) + K(t);
	e = d; d = c; c = S(30, b); b = a; a = tmp;
  }
  for (t = 60; t < 80; t++) {
	s = t & 0x0f;
	W(s) = S(1, W((s+13) & 0x0f) ^ W((s+8) & 0x0f) ^ W((s+2) & 0x0f) ^ W(s));
	tmp = S(5, a) + F3(b, c, d) + e + W(s) + K(t);
	e = d; d = c; c = S(30, b); b = a; a = tmp;
  }

  H(0) = H(0) + a;
  H(1) = H(1) + b;
  H(2) = H(2) + c;
  H(3) = H(3) + d;
  H(4) = H(4) + e;

  bzero(&ctxt->m.b8[0], 64);
}

/*------------------------------------------------------------*/

void
SHA1_init(SHA1_ctx *ctxt)
{
	bzero(ctxt, sizeof(SHA1_ctx));

	H(0) = 0x67452301;
	H(1) = 0xefcdab89;
	H(2) = 0x98badcfe;
	H(3) = 0x10325476;
	H(4) = 0xc3d2e1f0;
}

void
sha1_pad(SHA1_ctx *ctxt)
{
  size_t padlen;		/*pad length in bytes*/
  size_t padstart;

  PUTPAD(0x80);

  padstart = COUNT % 64;
  padlen = 64 - padstart;
  if (padlen < 8) {
	bzero(&ctxt->m.b8[padstart], padlen);
	COUNT += padlen;
	COUNT %= 64;
	sha1_step(ctxt);
	padstart = COUNT % 64;	/* should be 0 */
	padlen = 64 - padstart;	/* should be 64 */
  }
  bzero(&ctxt->m.b8[padstart], padlen - 8);
  COUNT += (padlen - 8);
  COUNT %= 64;
#if BYTE_ORDER == BIG_ENDIAN
  PUTPAD(ctxt->c.b8[0]); PUTPAD(ctxt->c.b8[1]);
  PUTPAD(ctxt->c.b8[2]); PUTPAD(ctxt->c.b8[3]);
  PUTPAD(ctxt->c.b8[4]); PUTPAD(ctxt->c.b8[5]);
  PUTPAD(ctxt->c.b8[6]); PUTPAD(ctxt->c.b8[7]);
#else
  PUTPAD(ctxt->c.b8[7]); PUTPAD(ctxt->c.b8[6]);
  PUTPAD(ctxt->c.b8[5]); PUTPAD(ctxt->c.b8[4]);
  PUTPAD(ctxt->c.b8[3]); PUTPAD(ctxt->c.b8[2]);
  PUTPAD(ctxt->c.b8[1]); PUTPAD(ctxt->c.b8[0]);
#endif
}

void
SHA1_update(SHA1_ctx *ctxt, const unsigned char *input, size_t len)
{
  size_t gaplen;
  size_t gapstart;
  size_t off;
  size_t copysiz;

  off = 0;

  while (off < len) {
	gapstart = COUNT % 64;
	gaplen = 64 - gapstart;

	copysiz = (gaplen < len - off) ? gaplen : len - off;
	bcopy(&input[off], &ctxt->m.b8[gapstart], copysiz);
	COUNT += copysiz;
	COUNT %= 64;
	ctxt->c.b64[0] += copysiz * 8;
	if (COUNT % 64 == 0)
	  sha1_step(ctxt);
	off += copysiz;
  }
}

void
SHA1_final(SHA1_ctx *ctxt, unsigned char digest0[SHA1_DIGEST_LEN])
{
  uint8_t *digest;

  digest = (uint8_t *)digest0;
  sha1_pad(ctxt);
#if BYTE_ORDER == BIG_ENDIAN
  bcopy(&ctxt->h.b8[0], digest, SHA1_RESULTLEN);
#else
  digest[0] = ctxt->h.b8[3]; digest[1] = ctxt->h.b8[2];
  digest[2] = ctxt->h.b8[1]; digest[3] = ctxt->h.b8[0];
  digest[4] = ctxt->h.b8[7]; digest[5] = ctxt->h.b8[6];
  digest[6] = ctxt->h.b8[5]; digest[7] = ctxt->h.b8[4];
  digest[8] = ctxt->h.b8[11]; digest[9] = ctxt->h.b8[10];
  digest[10] = ctxt->h.b8[9]; digest[11] = ctxt->h.b8[8];
  digest[12] = ctxt->h.b8[15]; digest[13] = ctxt->h.b8[14];
  digest[14] = ctxt->h.b8[13]; digest[15] = ctxt->h.b8[12];
  digest[16] = ctxt->h.b8[19]; digest[17] = ctxt->h.b8[18];
  digest[18] = ctxt->h.b8[17]; digest[19] = ctxt->h.b8[16];
#endif
}

/*
 * Without any arguments, a series of self-tests are run.
 * If the single argument "md5" (or "md4") is given, the program instead
 * reads stdin and emits the hex encoded MD5 (MD4) digest to stdout.
 * Exit status is zero iff no error occurs.
 */

static char *
to_hex(uint8_t *value, size_t nbytes)
{
  size_t i;
  char *p;
  uint8_t *v;
  static char buf[512];

  if (((nbytes * 2) + 1) > sizeof(buf))
	return(NULL);

  v = value;
  p = buf;
  for (i = 0; i < nbytes; i++)
    p += sprintf(p, "%.2x", *v++ & 0xff);

  return(buf);
}

/* Also defined in str.c, but we do not want that entire file. */
static int
ca_hexpair2int(char *pair)
{
  int byte, hi, lo;

  lo = pair[1];
  hi = pair[0];

  if (!isxdigit(lo) || !isxdigit(hi))
    return(-1);

  hi = tolower(hi);
  lo = tolower(lo);
  byte = (hi >= 'a') ? (hi - 'a' + 10) : (hi - '0');
  byte <<= 4;
  byte += ((lo >= 'a') ? (lo - 'a' + 10) : (lo - '0'));

  return(byte);
}

/*
 * Also defined in str.c, but we do not want that entire file.
 * Convert STR, a null-terminated string consisting of hex character pairs,
 * into binary.  Set LENP to the size of the binary vector, in bytes,
 * which will be zero if an error occurs.
 * Return the allocated buffer, or NULL if the string contains an
 * invalid character.
 */
static unsigned char *
ca_strhextob(char *str, unsigned int *lenp)
{
  int blen, bval, i, n;
  char *p;
  size_t slen;
  unsigned char *uc;

  slen = strlen(str);
  if (slen == 0 || (slen & 01)) {
    if (lenp != NULL)
      *lenp = 0;
    return(NULL);
  }

  blen = slen / 2;
  if (lenp != NULL)
    *lenp = blen;
  if ((uc = (unsigned char *) malloc(blen)) == NULL) {
	fprintf(stderr, "ca_strhextob: malloc failed.\n");
	exit(1);
  }

  p = str;
  n = 0;
  for (i = 0; i < blen; i++) {
    if ((bval = ca_hexpair2int(p)) == -1) {
      if (lenp != NULL)
        *lenp = 0;
      return(NULL);
    }
    uc[n++] = bval;
    p += 2;
  }

  return(uc);
}

/*
 * Put the SHA1 hash of LEN bytes of MEM into DIGESTP, which is assumed to
 * be at least SHA1_DIGEST_LEN bytes long.
 */
void
sha1_mem(void *mem, size_t len, uint8_t *digestp)
{
  SHA1_ctx ctx;

  SHA1_init(&ctx);
  SHA1_update(&ctx, (uint8_t *) mem, len);
  SHA1_final(&ctx, digestp);
}

/*
 * Put the SHA1 hash of string STR into DIGESTP, which is assumed to
 * be at least SHA1_DIGEST_LEN bytes long.
 */
void
sha1_str(char *str, uint8_t *digestp)
{

  sha1_mem(str, strlen(str), digestp);
}


#include <sys/types.h>
#include <unistd.h>

/* The biggest digest output length we are testing here. */
#define MAX_DIGEST_LEN		SHA1_DIGEST_LEN

/* https://en.wikipedia.org/wiki/MD5 */
typedef struct Digest_test {
  int is_hex;		/* A text string? (or a hex-encoded binary string) */
  char *msg;
  uint8_t digest[MAX_DIGEST_LEN];
} Digest_test;

static Digest_test sha1_self_test[] = {
  { 1, "",
	{ 0xda, 0x39, 0xa3, 0xee, 0x5e, 0x6b, 0x4b, 0x0d, 0x32, 0x55,
	  0xbf, 0xef, 0x95, 0x60, 0x18, 0x90, 0xaf, 0xd8, 0x07, 0x09}
  },

  { 0, "The quick brown fox jumps over the lazy dog",
	{ 0x2f, 0xd4, 0xe1, 0xc6, 0x7a, 0x2d, 0x28, 0xfc, 0xed, 0x84,
	  0x9e, 0xe1, 0xbb, 0x76, 0xe7, 0x39, 0x1b, 0x93, 0xeb, 0x12 }
  },

  { 0, "The quick brown fox jumps over the lazy cog",
	{ 0xde, 0x9f, 0x2c, 0x7f, 0xd2, 0x5e, 0x1b, 0x3a, 0xfa, 0xd3,
	  0xe8, 0x5a, 0x0b, 0xd1, 0x7d, 0x9b, 0x10, 0x0d, 0xb4, 0xb3 }
  },

  { 0, "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
	{ 0x84, 0x98, 0x3e, 0x44, 0x1c, 0x3b, 0xd2, 0x6e, 0xba, 0xae,
	  0x4a, 0xa1, 0xf9, 0x51, 0x29, 0xe5, 0xe5, 0x46, 0x70, 0xf1 }
  },

  { 0, "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
	{ 
	 0xa4, 0x9b, 0x24, 0x46, 0xa0, 0x2c, 0x64, 0x5b, 0xf4, 0x19,
	 0xf9, 0x95, 0xb6, 0x70, 0x91, 0x25, 0x3a, 0x04, 0xa2, 0x59 }
  },

  { 0, "abc",
	{ 0xa9, 0x99, 0x3e, 0x36, 0x47, 0x06, 0x81, 0x6a, 0xba, 0x3e,
	  0x25, 0x71, 0x78, 0x50, 0xc2, 0x6c, 0x9c, 0xd0, 0xd8, 0x9d }
  },

  { 0, NULL, { 0x00 } }
};

static int
do_sha1_self_test(void)
{
  uint8_t digest[SHA1_DIGEST_LEN];
  Digest_test *test;

  fprintf(stderr, "Doing SHA1 tests:\n");

  for (test = &sha1_self_test[0]; test->msg != NULL; test++) {
    if (test->is_hex) {
      unsigned int len;
      unsigned char *m;

      m = ca_strhextob(test->msg, &len);
      sha1_mem(m, len, digest);
    }
    else
      sha1_str(test->msg, digest);

    fprintf(stderr, "\"%s\": ", test->msg);
    if (memcmp(digest, test->digest, SHA1_DIGEST_LEN) != 0) {
      fprintf(stderr, "failed!\n");
      fprintf(stderr, "expected: %s\n",
              to_hex(test->digest, SHA1_DIGEST_LEN));
      fprintf(stderr, "got:      %s\n", to_hex(digest, SHA1_DIGEST_LEN));
      return(-1);
    }
    fprintf(stderr, "ok\n");
  }

  /* A special case test... */
  {
	unsigned char buf[1000000];
    static uint8_t expected_digest[SHA1_DIGEST_LEN] =
	  { 0x34, 0xaa, 0x97, 0x3c, 0xd4, 0xc4, 0xda, 0xa4, 0xf6, 0x1e,
		0xeb, 0x2b, 0xdb, 0xad, 0x27, 0x31, 0x65, 0x34, 0x01, 0x6f };

	fprintf(stderr, "One million bytes of 'a': ");

	memset(buf, (int) 'a', sizeof(buf));
	sha1_mem(buf, sizeof(buf), digest);

    if (memcmp(digest, expected_digest, SHA1_DIGEST_LEN) != 0) {
      fprintf(stderr, "failed!\n");
      fprintf(stderr, "expected: %s\n",
              to_hex(test->digest, SHA1_DIGEST_LEN));
      fprintf(stderr, "got:      %s\n", to_hex(digest, SHA1_DIGEST_LEN));
      return(-1);
    }
	fprintf(stderr, "ok\n");
  }

  fprintf(stderr, "All ok\n");
  return(0);
}

static Digest_test md5_self_test[] = {  
  { 0, "The quick brown fox jumps over the lazy dog",
	{ 0x9e, 0x10, 0x7d, 0x9d, 0x37, 0x2b, 0xb6, 0x82,
	  0x6b, 0xd8, 0x1d, 0x35, 0x42, 0xa4, 0x19, 0xd6 },
  },

  { 0, "The quick brown fox jumps over the lazy cog",
	{ 0x10, 0x55, 0xd3, 0xe6, 0x98, 0xd2, 0x89, 0xf2,
	  0xaf, 0x86, 0x63, 0x72, 0x51, 0x27, 0xbd, 0x4b },
  },

  { 0, "The quick brown fox jumps over the lazy dog.",
	{
	  0xe4, 0xd9, 0x09, 0xc2, 0x90, 0xd0, 0xfb, 0x1c,
	  0xa0, 0x68, 0xff, 0xad, 0xdf, 0x22, 0xcb, 0xd0
	},
  },

  { 0, "",
	{
	  0xd4, 0x1d, 0x8c, 0xd9, 0x8f, 0x00, 0xb2, 0x04,
	  0xe9, 0x80, 0x09, 0x98, 0xec, 0xf8, 0x42, 0x7e
	},
  },

  /* RFC 1321 */
  { 0, "a",
	{
	 0x0c, 0xc1, 0x75, 0xb9, 0xc0, 0xf1, 0xb6, 0xa8,
	 0x31, 0xc3, 0x99, 0xe2, 0x69, 0x77, 0x26, 0x61
	},
  },

  /* RFC 1321 */
  { 0, "abc",
	{
	 0x90, 0x01, 0x50, 0x98, 0x3c, 0xd2, 0x4f, 0xb0,
	 0xd6, 0x96, 0x3f, 0x7d, 0x28, 0xe1, 0x7f, 0x72
	},
  },

  /* RFC 1321 */
  { 0, "message digest",
	{
	 0xf9, 0x6b, 0x69, 0x7d, 0x7c, 0xb7, 0x93, 0x8d,
	 0x52, 0x5a, 0x2f, 0x31, 0xaa, 0xf1, 0x61, 0xd0
	},
  },

  /* RFC 1321 */
  { 0, "abcdefghijklmnopqrstuvwxyz",
	{
	 0xc3, 0xfc, 0xd3, 0xd7, 0x61, 0x92, 0xe4, 0x00,
	 0x7d, 0xfb, 0x49, 0x6c, 0xca, 0x67, 0xe1, 0x3b
	},
  },

  /* RFC 1321 */
  { 0, "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
	{
	 0xd1, 0x74, 0xab, 0x98, 0xd2, 0x77, 0xd9, 0xf5,
	 0xa5, 0x61, 0x1c, 0x2c, 0x9f, 0x41, 0x9d, 0x9f
	},
  },

  /* RFC 1321 */
  { 0, "12345678901234567890123456789012345678901234567890123456789012345678901234567890",
	{
	 0x57, 0xed, 0xf4, 0xa2, 0x2b, 0xe3, 0xc9, 0x55,
	 0xac, 0x49, 0xda, 0x2e, 0x21, 0x07, 0xb6, 0x7a
	},
  },

  /*
   * From: "MD5 To Be Considered Harmful Someday", Dan Kaminsky, 6-Dec-2004.
   * This test vector and the next are examples of different strings (bytes
   * at +19, +45, +59, +83, +109, and +123 differ) that have the same MD5 hash.
   * https://eprint.iacr.org/2004/357.pdf
   */
  { 1, "d131dd02c5e6eec4693d9a0698aff95c2fcab58712467eab4004583eb8fb7f8955ad340609f4b30283e488832571415a085125e8f7cdc99fd91dbdf280373c5bd8823e3156348f5bae6dacd436c919c6dd53e2b487da03fd02396306d248cda0e99f33420f577ee8ce54b67080a80d1ec69821bcb6a8839396f9652b6ff72a70",
	{
	 0x79, 0x05, 0x40, 0x25, 0x25, 0x5f, 0xb1, 0xa2,
	 0x6e, 0x4b, 0xc4, 0x22, 0xae, 0xf5, 0x4e, 0xb4
	},
  },

  { 1, "d131dd02c5e6eec4693d9a0698aff95c2fcab50712467eab4004583eb8fb7f8955ad340609f4b30283e4888325f1415a085125e8f7cdc99fd91dbd7280373c5bd8823e3156348f5bae6dacd436c919c6dd53e23487da03fd02396306d248cda0e99f33420f577ee8ce54b67080280d1ec69821bcb6a8839396f965ab6ff72a70",
	{
	 0x79, 0x05, 0x40, 0x25, 0x25, 0x5f, 0xb1, 0xa2,
	 0x6e, 0x4b, 0xc4, 0x22, 0xae, 0xf5, 0x4e, 0xb4
	},
  },

  { 0, NULL, { 0x00 } }
};

/*
 * This is a convenience function to calculate the MD5 hash of the first
 * LEN bytes of MEM into DIGESTP, which is assumed to be at least
 * MD5_DIGEST_LEN bytes long.
 */
void
md5(void *mem, size_t len, uint8_t *digestp)
{
  MD5_ctx ctx;

  MD5_init(&ctx);
  MD5_update(&ctx, (uint8_t *) mem, len);
  MD5_final(&ctx, digestp);
}

/*
 * This convenience function puts the MD5 hash of the null-terminated string
 * STR into DIGESTP, which is assumed to be at least MD5_DIGEST_LEN bytes long.
 */
void
md5_str(char *str, uint8_t *digestp)
{

  md5(str, strlen(str), digestp);
}

static int
do_md5_self_test(void)
{
  uint8_t digest[MD5_DIGEST_LEN];
  Digest_test *test;

  fprintf(stderr, "Doing MD5 tests:\n");

  for (test = &md5_self_test[0]; test->msg != NULL; test++) {
	if (test->is_hex) {
	  unsigned int len;
	  unsigned char *m;

	  m = ca_strhextob(test->msg, &len);
	  md5(m, len, digest);
	}
	else
	  md5_str(test->msg, digest);

	fprintf(stderr, "\"%s\": ", test->msg);
	if (memcmp(digest, test->digest, MD5_DIGEST_LEN) != 0) {
	  fprintf(stderr, "failed!\n");
	  fprintf(stderr, "expected: %s\n",
			  to_hex(test->digest, MD5_DIGEST_LEN));
	  fprintf(stderr, "got:      %s\n", to_hex(digest, MD5_DIGEST_LEN));
	  return(-1);
	}
	fprintf(stderr, "ok\n");
  }
  fprintf(stderr, "All ok\n");
  return(0);
}

/* Test vectors are from RFC 1320. */
static Digest_test md4_self_test[] = {  
  { 0, "",
	{ 0x31, 0xd6, 0xcf, 0xe0, 0xd1, 0x6a, 0xe9, 0x31,
	  0xb7, 0x3c, 0x59, 0xd7, 0xe0, 0xc0, 0x89, 0xc0 }
  },

  { 0, "a",
	{ 0xbd, 0xe5, 0x2c, 0xb3, 0x1d, 0xe3, 0x3e, 0x46,
	  0x24, 0x5e, 0x05, 0xfb, 0xdb, 0xd6, 0xfb, 0x24 }
  },

  { 0, "abc",
	{ 0xa4, 0x48, 0x01, 0x7a, 0xaf, 0x21, 0xd8, 0x52,
	  0x5f, 0xc1, 0x0a, 0xe8, 0x7a, 0xa6, 0x72, 0x9d }
  },

  { 0, "message digest",
	{ 0xd9, 0x13, 0x0a, 0x81, 0x64, 0x54, 0x9f, 0xe8,
	  0x18, 0x87, 0x48, 0x06, 0xe1, 0xc7, 0x01, 0x4b }
  },

  { 0, "abcdefghijklmnopqrstuvwxyz",
	{ 0xd7, 0x9e, 0x1c, 0x30, 0x8a, 0xa5, 0xbb, 0xcd,
	  0xee, 0xa8, 0xed, 0x63, 0xdf, 0x41, 0x2d, 0xa9 }
  },

  { 0, "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
	{ 0x04, 0x3f, 0x85, 0x82, 0xf2, 0x41, 0xdb, 0x35,
	  0x1c, 0xe6, 0x27, 0xe1, 0x53, 0xe7, 0xf0, 0xe4 }
  },

  { 0, "12345678901234567890123456789012345678901234567890123456789012345678901234567890",
	{ 0xe3, 0x3b, 0x4d, 0xdc, 0x9c, 0x38, 0xf2, 0x19,
	  0x9c, 0x3e, 0x7b, 0x16, 0x4f, 0xcc, 0x05, 0x36 }
  },

  { 0, "The quick brown fox jumps over the lazy dog",
	{ 0x1b, 0xee, 0x69, 0xa4, 0x6b, 0xa8, 0x11, 0x18,
	  0x5c, 0x19, 0x47, 0x62, 0xab, 0xae, 0xae, 0x90 }
  },

  { 0, "The quick brown fox jumps over the lazy cog",
	{ 0xb8, 0x6e, 0x13, 0x0c, 0xe7, 0x02, 0x8d, 0xa5,
	  0x9e, 0x67, 0x2d, 0x56, 0xad, 0x01, 0x13, 0xdf }
  },

  /*
   * This test vector and the next are examples of different strings that
   * have the same MD4 hash.
   */
  { 1, "839c7a4d7a92cb5678a5d5b9eea5a7573c8a74deb366c3dc20a083b69f5d2a3bb3719dc69891e9f95e809fd7e8b23ba6318edd45e51fe39708bf9427e9c3e8b9",
	{ 0x4d, 0x7e, 0x6a, 0x1d, 0xef, 0xa9, 0x3d, 0x2d,
	  0xde, 0x05, 0xb4, 0x5d, 0x86, 0x4c, 0x42, 0x9b }
  },

  { 1, "839c7a4d7a92cbd678a5d529eea5a7573c8a74deb366c3dc20a083b69f5d2a3bb3719dc69891e9f95e809fd7e8b23ba6318edc45e51fe39708bf9427e9c3e8b9",
	{ 0x4d, 0x7e, 0x6a, 0x1d, 0xef, 0xa9, 0x3d, 0x2d,
	  0xde, 0x05, 0xb4, 0x5d, 0x86, 0x4c, 0x42, 0x9b }
  },

  { 0, NULL, { 0x00 } }
};

static void
md4_str(char *str, uint8_t *digestp)
{
  MD4_ctx ctx;

  MD4_init(&ctx);
  MD4_update(&ctx, str, strlen(str));
  MD4_final(&ctx, digestp);
}

/*
 * Put the MD4 hash of LEN bytes of MEM into DIGESTP, which is assumed to
 * be at least MD4_DIGEST_LEN bytes long.
 */
void
md4_mem(void *mem, size_t len, uint8_t *digestp)
{
  MD4_ctx ctx;

  MD4_init(&ctx);
  MD4_update(&ctx, (uint8_t *) mem, len);
  MD4_final(&ctx, digestp);
}

static int
do_md4_self_test(void)
{
  uint8_t digest[MD4_DIGEST_LEN];
  Digest_test *test;

  fprintf(stderr, "Doing MD4 tests:\n");

  for (test = &md4_self_test[0]; test->msg != NULL; test++) {
	if (test->is_hex) {
	  unsigned int len;
      unsigned char *m;

      m = ca_strhextob(test->msg, &len);
      md4_mem(m, len, digest);
    }
    else
	  md4_str(test->msg, digest);

	fprintf(stderr, "\"%s\": ", test->msg);
	if (memcmp(digest, test->digest, MD4_DIGEST_LEN) != 0) {
	  fprintf(stderr, "failed!\n");
	  fprintf(stderr, "expected: %s\n",
			  to_hex(test->digest, MD4_DIGEST_LEN));
	  fprintf(stderr, "got:      %s\n", to_hex(digest, MD4_DIGEST_LEN));
	  return(-1);
	}
	fprintf(stderr, "ok\n");
  }
  fprintf(stderr, "All ok\n");
  return(0);
}

/*
 * Test vectors are from various sources.
 * See RFC 6229 https://tools.ietf.org/html/rfc6229
 */
typedef struct RC4_test {
  size_t key_len;
  size_t offset;		/* Starting offset for key stream test. */
  size_t slen;			/* Length in bytes of key stream test. */
  uint8_t key[128];		/* Key. */
  size_t msg_len;		/* If zero, this is a key stream test. */
  uint8_t msg[128];		/* Message to encipher/decipher. */
  uint8_t exp[128];		/* Expected cipher output or key stream. */
} RC4_test;

static RC4_test rc4_self_test[] = {
  { 8, 0, 0,
	{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef },
	8, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x74, 0x94, 0xc2, 0xe7, 0x10, 0x4b, 0x08, 0x79 }
  },

  { 8, 0, 0,
	{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef },
	8, { 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef },
	{ 0x75, 0xb7, 0x87, 0x80, 0x99, 0xe0, 0xc5, 0x96 }
  },

  { 8, 0, 0,
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	8, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xde, 0x18, 0x89, 0x41, 0xa3, 0x37, 0x5d, 0x3a }
  },

  { 5, 0, 0,
	{ 0x61, 0x8a, 0x63, 0xd2, 0xfb },
	5, { 0xdc, 0xee, 0x4c, 0xf9, 0x2c },
	{ 0xf1, 0x38, 0x29, 0xc9, 0xde }
  },

  { 16, 0, 16,
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	0, { 0x00 },
	{ 0xde, 0x18, 0x89, 0x41, 0xa3, 0x37, 0x5d, 0x3a,
	  0x8a, 0x06, 0x1e, 0x67, 0x57, 0x6e, 0x92, 0x6d }
  },

  { 5, 0, 16,
	{ 0x01, 0x02, 0x03, 0x04, 0x05 },
	0, { 0x00 },
	{ 0xb2, 0x39, 0x63, 0x05, 0xf0, 0x3d, 0xc0, 0x27,
	  0xcc, 0xc3, 0x52, 0x4a, 0x0a, 0x11, 0x18, 0xa8 }
  },

  { 5, 0, 16,
	{ 0x01, 0x02, 0x03, 0x04, 0x05 },
	0, { 0x00 },
	{ 0xb2, 0x39, 0x63, 0x05, 0xf0, 0x3d, 0xc0, 0x27,
	  0xcc, 0xc3, 0x52, 0x4a, 0x0a, 0x11, 0x18, 0xa8 }
  },

  { 5, 0, 16,
	{ 0x01, 0x02, 0x03, 0x04, 0x05 },
	0, { 0x00 },
	{ 0xb2, 0x39, 0x63, 0x05, 0xf0, 0x3d, 0xc0, 0x27,
	  0xcc, 0xc3, 0x52, 0x4a, 0x0a, 0x11, 0x18, 0xa8 }
  },

  { 5, 1536, 16,
	{ 0x01, 0x02, 0x03, 0x04, 0x05 },
	0, { 0x00 },
	{ 0xd8, 0x72, 0x9d, 0xb4, 0x18, 0x82, 0x25, 0x9b,
	  0xee, 0x4f, 0x82, 0x53, 0x25, 0xf5, 0xa1, 0x30 }
  },

  { 10, 4096, 16,
	{ 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a },
	0, { 0x00 },
	{ 0x08, 0xb6, 0xbe, 0x45, 0x12, 0x4a, 0x43, 0xe2,
	  0xeb, 0x77, 0x95, 0x3f, 0x84, 0xdc, 0x85, 0x53 }
  },

  { 5, 0, 16,
	{ 0x83, 0x32, 0x22, 0x77, 0x2a },
	0, { 0x00 },
	{ 0x80, 0xad, 0x97, 0xbd, 0xc9, 0x73, 0xdf, 0x8a,
	  0x2e, 0x87, 0x9e, 0x92, 0xa4, 0x97, 0xef, 0xda }
  },

  { 5, 3056, 16,
	{ 0x83, 0x32, 0x22, 0x77, 0x2a },
	0, { 0x00 },
	{ 0xaf, 0xec, 0xc0, 0x37, 0xfd, 0xb7, 0xb0, 0x83,
	  0x8e, 0xb3, 0xd7, 0x0b, 0xcd, 0x26, 0x83, 0x82 }
  },

  { 32, 0, 16,
	{ 0x1a, 0xda, 0x31, 0xd5, 0xcf, 0x68, 0x82, 0x21,
	  0xc1, 0x09, 0x16, 0x39, 0x08, 0xeb, 0xe5, 0x1d,
	  0xeb, 0xb4, 0x62, 0x27, 0xc6, 0xcc, 0x8b, 0x37,
	  0x64, 0x19, 0x10, 0x83, 0x32, 0x22, 0x77, 0x2a },
	0, { 0x00 },
	{ 0xdd, 0x5b, 0xcb, 0x00, 0x18, 0xe9, 0x22, 0xd4,
	  0x94, 0x75, 0x9d, 0x7c, 0x39, 0x5d, 0x02, 0xd3 }
  },

  { 32, 768, 16,
	{ 0x1a, 0xda, 0x31, 0xd5, 0xcf, 0x68, 0x82, 0x21,
	  0xc1, 0x09, 0x16, 0x39, 0x08, 0xeb, 0xe5, 0x1d,
	  0xeb, 0xb4, 0x62, 0x27, 0xc6, 0xcc, 0x8b, 0x37,
	  0x64, 0x19, 0x10, 0x83, 0x32, 0x22, 0x77, 0x2a },
	0, { 0x00 },
	{ 0x85, 0x14, 0xa5, 0x49, 0x58, 0x58, 0x09, 0x6f,
	  0x59, 0x6e, 0x4b, 0xcd, 0x66, 0xb1, 0x06, 0x65 }
  },

  { 0, 0, 0,
	{ 0x00 },
	0, { 0x00 },
	{ 0x00 }
  }
};

/*
 * For testing purposes, copy LEN bytes of the key stream for the given STATE
 * into SBUF, starting OFFSET bytes from the current stream position.
 */
static void
rc4_key_stream(RC4_ctx *const state, uint8_t *sbuf, size_t offset, size_t len)
{
  size_t i, s;
  uint8_t j;

  for (i = s = 0; i < (offset + len); i++) {
	/* Update modification indices. */
	state->index1++;
	state->index2 += state->perm[state->index1];

	/* Modify permutation. */
	swap_bytes(&state->perm[state->index1], &state->perm[state->index2]);

	/* Encrypt (or decrypt) next byte. */
	j = state->perm[state->index1] + state->perm[state->index2];
	if (i >= offset)
	  sbuf[s++] = state->perm[j];
  }
}

static int
do_rc4_self_test(void)
{
  uint8_t buf[512];
  RC4_ctx ctx;
  RC4_test *test;

  fprintf(stderr, "Doing RC4 tests:\n");

  for (test = &rc4_self_test[0]; test->key_len != 0; test++) {
	rc4_init(&ctx, test->key, test->key_len);
	if (test->msg_len) {
	  rc4_crypt(&ctx, test->msg, buf, test->msg_len);
	  if (memcmp(buf, test->exp, test->msg_len) != 0) {
		fprintf(stderr, "failed!\n");
		fprintf(stderr, "expected: %s\n",
				to_hex(test->exp, test->msg_len));
		fprintf(stderr, "got:      %s\n", to_hex(buf, test->msg_len));
		return(-1);
	  }
	}
	else {
	  rc4_init(&ctx, test->key, test->key_len);
	  rc4_key_stream(&ctx, buf, test->offset, test->slen);
	  if (memcmp(buf, test->exp, test->slen) != 0) {
		fprintf(stderr, "failed!\n");
		fprintf(stderr, "expected: %s\n", to_hex(test->exp, test->slen));
		fprintf(stderr, "got:      %s\n", to_hex(buf, test->slen));
		return(-1);
	  }
	}
	fprintf(stderr, ".");
  }
  fprintf(stderr, "\nAll ok\n");
  return(0);
}

static int
do_hmac_md5_self_test(void)
{
  size_t hlen;
  uint8_t *hmac;
  static uint8_t key1[] = {
	0x0b, 0x0b, 0x0b, 0x0b, 0x0b, 0x0b, 0x0b, 0x0b,
	0x0b, 0x0b, 0x0b, 0x0b, 0x0b, 0x0b, 0x0b, 0x0b
  };
  static uint8_t exp1[] = {
	0x92, 0x94, 0x72, 0x7a, 0x36, 0x38, 0xbb, 0x1c,
	0x13, 0xf4, 0x8e, 0xf8, 0x15, 0x8b, 0xfc, 0x9d
  };
  static uint8_t key2[] = {
	0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
	0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa
  };
  static uint8_t data2[] = {
	0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd,
	0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd,
	0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd,
	0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd,
	0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd, 0xdd
  };
  static uint8_t exp2[] = {
	0x56, 0xbe, 0x34, 0x52, 0x1d, 0x14, 0x4c, 0x88,
	0xdb, 0xb8, 0xc7, 0x33, 0xf0, 0xe8, 0xb3, 0xf6
  };

  fprintf(stderr, "Doing HMAC-MD5 tests:\n");

  crypto_hmac_md5(key1, sizeof(key1), (uint8_t *) "Hi There", 8,
				  NULL, NULL, &hmac, &hlen);
  if (memcmp(hmac, exp1, hlen) != 0) {
	fprintf(stderr, "failed!\n");
	fprintf(stderr, "expected: %s\n", to_hex(exp1, sizeof(exp1)));
	fprintf(stderr, "got:      %s\n", to_hex(hmac, hlen));
	return(-1);
  }
  fprintf(stderr, "%s\n", to_hex(hmac, hlen));

  {
	static uint8_t exp[] = {
	  0x75, 0x0c, 0x78, 0x3e, 0x6a, 0xb0, 0xb5, 0x03,
	  0xea, 0xa8, 0x6e, 0x31, 0x0a, 0x5d, 0xb7, 0x38
	};

	crypto_hmac_md5((uint8_t *) "Jefe", 4,
					(uint8_t *) "what do ya want for nothing?", 28,
					NULL, NULL, &hmac, &hlen);
	if (memcmp(hmac, exp, hlen) != 0) {
	  fprintf(stderr, "failed!\n");
	  fprintf(stderr, "expected: %s\n", to_hex(exp, sizeof(exp)));
	  fprintf(stderr, "got:      %s\n", to_hex(hmac, hlen));
	  return(-1);
	}
	fprintf(stderr, "%s\n", to_hex(hmac, hlen));
  }

  crypto_hmac_md5(key2, sizeof(key2), data2, sizeof(data2), NULL, NULL,
				  &hmac, &hlen);
  if (memcmp(hmac, exp2, hlen) != 0) {
	fprintf(stderr, "failed!\n");
	fprintf(stderr, "expected: %s\n", to_hex(exp2, sizeof(exp2)));
	fprintf(stderr, "got:      %s\n", to_hex(hmac, hlen));
	return(-1);
  }
  fprintf(stderr, "%s\n", to_hex(hmac, hlen));

  fprintf(stderr, "All ok\n");
  return(0);
}

int
do_crypto_aux_tests(void)
{

  if (do_sha1_self_test() == -1)
	return(-1);

  if (do_md4_self_test() == -1)
	return(-1);

  if (do_md5_self_test() == -1)
	return(-1);

  if (do_rc4_self_test() == -1)
	return(-1);

  if (do_hmac_md5_self_test() == -1)
	return(-1);

  return(0);
}

#ifdef PROG

int
main(int argc, char **argv)
{
  int i;
  uint8_t buf[BUFSIZ];
  ssize_t n;

  if (argc == 2) {
	if (strcasecmp(argv[1], "md5") == 0) {
	  uint8_t digest[MD5_DIGEST_LEN];
	  MD5_ctx ctx;

	  MD5_init(&ctx);
	  while ((n = read(0, (void *) buf, sizeof(buf))) > 0)
		MD5_update(&ctx, buf, n);
	  MD5_final(&ctx, digest);

	  fprintf(stdout, "%s\n", to_hex(digest, MD5_DIGEST_LEN));

	  exit(0);
	}

	if (strcasecmp(argv[1], "md4") == 0) {
	  uint8_t digest[MD4_DIGEST_LEN];
	  MD4_ctx ctx;

	  MD4_init(&ctx);
	  while ((n = read(0, (void *) buf, sizeof(buf))) > 0)
		MD4_update(&ctx, buf, n);
	  MD4_final(&ctx, digest);

	  fprintf(stdout, "%s\n", to_hex(digest, MD4_DIGEST_LEN));

	  exit(0);
	}

	fprintf(stderr, "?\n");
	exit(1);
  }

  fprintf(stderr, "Doing auxiliary cryptographic function tests\n");
  do_md4_self_test();
  do_md5_self_test();
  do_rc4_self_test();
  do_hmac_md5_self_test();
  fprintf(stderr, "All auxiliary cryptographic function tests succeeded\n");

  exit(0);
}
#endif
