/*
 * Copyright (c) 2003-2025
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*****************************************************************************
 * COPYRIGHT AND PERMISSION NOTICE
 * 
 * Copyright (c) 2001-2003 The Queen in Right of Canada
 * 
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, and/or sell
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, provided that the above copyright notice(s) and this
 * permission notice appear in all copies of the Software and that both the
 * above copyright notice(s) and this permission notice appear in supporting
 * documentation.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE 
 * BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 * 
 * Except as contained in this notice, the name of a copyright holder shall not
 * be used in advertising or otherwise to promote the sale, use or other
 * dealings in this Software without prior written authorization of the
 * copyright holder.
 ***************************************************************************/

/*
 * MIT License
 *
 * Copyright (c) His Majesty the King in Right of Canada, as represented by
 * the Minister of Natural Resources, 2023.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * Signout (logoff/signout) service
 *
 * By specifying DACS_USERNAME, only logins under a username that exactly
 * matches DACS_USERNAME will be signed out.
 * By specifying DACS_JURISDICTION, only logins under a home jurisdiction
 * that exactly matches DACS_JURISDICTION will be signed out.
 * Both parameters are optional and both may be given.
 * Expired and otherwise invalid credentials are always deleted.
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2025\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: signout.c 3385 2025-04-01 00:26:15Z brachman $";
#endif

#include "dacs.h"

static const char *log_module_name = "dacs_signout";

typedef enum {
  SIGNOUT_HANDLER_URL         = 1,
  SIGNOUT_HANDLER_DEFAULT_URL = 2,
  SIGNOUT_HANDLER_FILE        = 3,
  SIGNOUT_HANDLER_MESSAGE     = 4,
  SIGNOUT_HANDLER_CREDENTIALS = 5,
  SIGNOUT_HANDLER_NONE        = 6
} Signout_handler_type;

typedef struct Signout_handler {
  Signout_handler_type type;
  char *signout_string;
} Signout_handler;

static int emit_xml = 0;
static int emit_json = 0;

/*
 * Return the configured SIGNOUT_HANDLER, a default, or NULL if something
 * is broken.
 */
static Signout_handler *
lookup_signout_handler(void)
{
  int n;
  char **argv;
  Kwv_pair *v;
  Mkargv conf = { 1, 0, NULL, NULL, NULL };
  static Signout_handler current;

  current.type = SIGNOUT_HANDLER_CREDENTIALS;
  current.signout_string = NULL;

  if ((v = conf_var(CONF_SIGNOUT_HANDLER)) == NULL)
	return(&current);

  if ((n = mkargv(v->val, &conf, &argv)) != 1 && n != 2) {
	log_msg((LOG_ERROR_LEVEL, "Invalid directive: %s", v->val));
	return(NULL);
  }

  if (n == 2) {
	current.signout_string = argv[1];
	if (strcaseeq(argv[0], "url"))
	  current.type = SIGNOUT_HANDLER_URL;
	else if (strcaseeq(argv[0], "defaulturl"))
	  current.type = SIGNOUT_HANDLER_DEFAULT_URL;
	else if (strcaseeq(argv[0], "file"))
	  current.type = SIGNOUT_HANDLER_FILE;
	else if (strcaseeq(argv[0], "message"))
	  current.type = SIGNOUT_HANDLER_MESSAGE;
	else {
	  log_msg((LOG_ERROR_LEVEL,
			   "Error parsing SIGNOUT_HANDLER directive: %s", v->val));
	  return(NULL);
	}

	log_msg((LOG_TRACE_LEVEL, "Using signout handler: %s %s",
			 argv[0], argv[1]));
  }
  else {
	/* The default URL case is not possible in the implicit syntax. */
	current.signout_string = argv[0];
	if (strcaseprefix(argv[0], "http"))
	  current.type = SIGNOUT_HANDLER_URL;
	else if (argv[0][0] == '/')
	  current.type = SIGNOUT_HANDLER_FILE;
	else if (strcaseeq(argv[0], "credentials"))
	  current.type = SIGNOUT_HANDLER_CREDENTIALS;
	else if (argv[0][0] == '"')
	  current.type = SIGNOUT_HANDLER_MESSAGE;
	else {
	  log_msg((LOG_ERROR_LEVEL,
			   "Error parsing SIGNOUT_HANDLER directive: %s", v->val));
	  return(NULL);
	}

	log_msg((LOG_TRACE_LEVEL, "Using signout handler: %s", argv[0]));
  }

  return(&current);
}

static int
void_cookie(char **set_cookie_buf, Credentials *credentials)
{
  char *name;

  log_msg((LOG_INFO_LEVEL, "%s:%s",
		   credentials->home_jurisdiction, credentials->username));

  name = credentials->cookie_name;

  return(make_set_void_cookie_header(name, 0, set_cookie_buf));
}

int
main(int argc, char **argv)
{
  int handler_error;
  int is_oidc_frontchannel_logout, is_oidc_rp_initiated_logout;
  int is_oidc_dacs_only_logout, is_oidc_op_only_logout, need_oidc_logout;
  char *bp, *jurisdiction_arg, *username_arg;
  char *dacs_signout_handler, *cookie_syntax;
  Kwv *kwv;
  char *errmsg;
  Credentials *cr, *credentials;
  Cookie *c, *cookies;
  unsigned int ncookies, ndeleted, nignored, ninvalid, nselected;
  Signout_handler *handler;
#ifdef ENABLE_OIDC_AUTH
  char *iss, *oidc_logout, *sid;
  char *client_id, *end_session_endpoint, *post_logout_redirect_uri;
#endif
  errmsg = "Internal error";
  handler_error = 0;
  emit_json = 0;
  emit_xml = 0;
  dacs_signout_handler = NULL;

  if (dacs_init(DACS_WEB_SERVICE, &argc, &argv, &kwv, &errmsg) == -1)
	goto fail;

  emit_xml = test_emit_xml_format();
  emit_json = test_emit_format(EMIT_FORMAT_JSON);

  if (should_use_argv) {
	if (argc > 1) {
	  errmsg = "Usage: unrecognized parameter";
	  goto fail;
	}
  }

  jurisdiction_arg = kwv_lookup_value(kwv, "DACS_JURISDICTION");
  username_arg = kwv_lookup_value(kwv, "DACS_USERNAME");

  dacs_signout_handler = kwv_lookup_value(kwv, "DACS_SIGNOUT_HANDLER");

  need_oidc_logout = 0;
  is_oidc_frontchannel_logout = 0;
  is_oidc_rp_initiated_logout = 0;
  is_oidc_dacs_only_logout = is_oidc_op_only_logout = 0;

#ifdef ENABLE_OIDC_AUTH
  /*
   * The trick here is to decide:
   * 1) if this is (or might be) an OIDC logout and
   * 2) if it is an OIDC logout, whether it is
   *    a) RP-only, b) OP-only, or c) both RP and OP
   * There will be default choices that depend on the exact context,
   * but it is assumed that the user has seen a signout page from which
   * the desired behaviour can be selected.
   *
   * N.B. the jurisdiction that did the OIDC authentication must be
   * the one to receive the OIDC logout because other jurisdiction may not
   * be configured; so unless an admin ensures that all jurisdictions
   * in the federation that can be asked to do an OIDC logout have the
   * (same) necessary configuration, the login/logout services and pages
   * must be carefully organized.
   *
   * There are two main OIDC logout cases.
   * Case 1: RP-Initiated Logout
   * A user who invokes dacs_signout directly has credentials derived
   * from an OIDC authentication (or may have).
   * 1) Perform the usual DACS signout, including cookie deletion, and then
   * 2) Redirect the user to the OIDC end_session_endpoint URI, attaching
   *    the post_logout_redirect_uri parameter that has been registered
   *    with the OIDC Provider - that is where the user will be redirected
   *    by the OIDC Provider when it finishes.
   * 3) The OP will trigger a Front-Channel Logout
   * 4) The OP will redirect the user to the post_logout_redirect_uri
   *
   * Case 2: Front-Channel Logout
   * A user has invoked signout at the OIDC Provider, triggering
   * Front-Channel Logout, which calls the frontchannel_logout_uri that has
   * been registered with that OIDC Provider. This case can potentially
   * be identified by the presence of 'iss' and 'sid' query parameters,
   * but they 'iss' and 'sid' are optional; if they are not used,
   * it will be necessary to invoke dacs_signout in such a way that it
   * can tell that it was invoked via the frontchannel_logout_uri
   * e.g., by using a new, explicit web service argument
   * (e.g., query parameter OIDC_LOGOUT=FRONTCHANNEL) or environment variable,
   * or if the context implies this must be a signout of DACS credentials
   * obtained from OIDC.
   *
   * Operation is the same (or nearly) as the general case:
   * 0) Ideally, the sid can be correlated with the credentials
   *    (OIDC Front-Channel Logout 5.); this would require a separate
   *    OIDC context cookie or adding OIDC context to DACS credentials
   * 1) Prompt the user about logging out from DACS; if approved, perform
   *    the usual DACS signout, including cookie deletion
   * 2) Exit with a status code of 200 and a Cache-Control "no-store".
   * Any handler configuration is ignored.
   *
   * Another use case is where the user wants to logout from the OP
   * but not from DACS (is_oidc_op_only_logout).
   * We do not inspect or delete any DACS cookies (or we could if we need to
   * identify the OP, but if only one OIDC Provider can be configured for
   * a jurisdiction, we can assume it is ok to use that one).
   * Otherwise it looks like the usual RP-Initiated Logout.
   * When (and if) the Front-Channel Logout follows up, 
   * the user should be prompted about a DACS signout.
   */

  /*
   * The OIDC_LOGOUT parameter can be used to explicitly select functionality.
   * frontchannel - delete DACS/OIDC cookies, then exit
   * rpinitiated  - delete DACS/OIDC cookies, then redirect to the endsession
   *                endpoint with parameters and exit
   * oponly       - do not delete DACS/OIDC cookies, then redirect to the
   *                endsession endpoint with parameters and exit
   * dacsonly     - normal (non-OIDC) behaviour; delete DACS/OIDC cookies,
   *                invoke any configured handlers and exit
   */
  client_id = end_session_endpoint = post_logout_redirect_uri = NULL;
  iss = sid = NULL;

  if ((oidc_logout = kwv_lookup_value(kwv, "OIDC_LOGOUT")) != NULL) {
	log_msg((LOG_TRACE_LEVEL, "oidc_logout=\"%s\"", oidc_logout));
	if (strcaseeq(oidc_logout, "frontchannel"))
	  is_oidc_frontchannel_logout = 1;
	else if (strcaseeq(oidc_logout, "rpinitiated"))
	  is_oidc_rp_initiated_logout = 1;
	else if (strcaseeq(oidc_logout, "oponly"))
	  is_oidc_op_only_logout = 1;
	else if (strcaseeq(oidc_logout, "dacsonly"))
	  is_oidc_dacs_only_logout = 1;
	else {
	  log_msg((LOG_ERROR_LEVEL,
			  "Unrecognized OIDC_LOGOUT parameter value: \"%s\"", oidc_logout));
	  goto find_cookies;
	}
  }
  else {
	/*
	 * If there is no OIDC_LOGOUT parameter but iss/sid are present,
	 * assume this is a Front-Channel Logout.
	 */
	iss = kwv_lookup_value(kwv, "iss");
	sid = kwv_lookup_value(kwv, "sid");
	if (iss != NULL && sid != NULL) {
	  is_oidc_frontchannel_logout = 1;
	  log_msg((LOG_DEBUG_LEVEL, "iss=\"%s\", sid=\"%s\"", iss, sid));
	}
  }

  if (is_oidc_frontchannel_logout)
	log_msg((LOG_DEBUG_LEVEL, "This is an OIDC Front-Channel logout"));
  else if (is_oidc_rp_initiated_logout)
	log_msg((LOG_DEBUG_LEVEL, "This is an OIDC RP-Initiated logout"));
  else if (is_oidc_dacs_only_logout)
	log_msg((LOG_DEBUG_LEVEL, "This is an OIDC DACS-Only logout"));
  else if (is_oidc_op_only_logout) {
	log_msg((LOG_DEBUG_LEVEL, "This is an OIDC RP-Initiated OP Only logout"));
	goto do_rp_logout;
  }

  /*
   * XXX see comments below
   * If signout is allowed at all jurisdictions, then all jurisdictions
   * should have these directives configured - they don't necessarily
   * have to configure directives for authentication though.
   */
  if ((client_id = conf_auth_option_lookup("oidc", "OIDC_CLIENT_ID")) == NULL) {
	log_msg((LOG_ERROR_LEVEL, "Cannot find OIDC_CLIENT_ID"));
  }
  else
	log_msg((LOG_TRACE_LEVEL, "OIDC_CLIENT_ID=\"%s\"", client_id));

  if ((end_session_endpoint
	   = conf_auth_option_lookup("oidc", "OIDC_END_SESSION_ENDPOINT")) == NULL) {
	log_msg((LOG_ERROR_LEVEL, "Cannot find OIDC_END_SESSION_ENDPOINT"));
  }
  else
	log_msg((LOG_TRACE_LEVEL, "OIDC_END_SESSION_ENDPOINT=\"%s\"",
			 end_session_endpoint));

  if ((post_logout_redirect_uri
	   = conf_auth_option_lookup("oidc", "OIDC_POST_LOGOUT_REDIRECT_URI"))
	  == NULL) {
	log_msg((LOG_ERROR_LEVEL, "Cannot find OIDC_POST_LOGOUT_REDIRECT_URI"));
  }
  else
	log_msg((LOG_TRACE_LEVEL, "OIDC_POST_LOGOUT_REDIRECT_URI=\"%s\"",
			 post_logout_redirect_uri));
#endif

 find_cookies:

  /*
   * If there are no cookies, a handler should still be called if one is
   * configured and when appropriate.
   * Also, the case where there are no cookies and the case where there are
   * cookies but there is no actual signout should result in the same behaviour.
   */

  if (get_cookies(NULL, &cookies, &ncookies) == -1) {
	errmsg = "Cookie parse error";
	goto fail;
  }

  if ((cookie_syntax = kwv_lookup_value(kwv, "COOKIE_SYNTAX")) != NULL) {
	if (configure_cookie_syntax(cookie_syntax, &errmsg) == -1)
	  goto fail;
  }

  credentials = NULL;
  ndeleted = ninvalid = nignored = nselected = 0;

  /*
   * In case OIDC authentication is enabled and configured, watch for
   * DACS cookies obtained from OIDC. If they are found, RP-Initiated Logout or
   * Front-Channel Logout specifications may need to be followed.
   */
  for (c = cookies; c != NULL; c = c->next) {
	if (is_selected_cookie_name(c->name)) {
	  nselected++;
	  continue;
	}

	if (!is_auth_cookie_name(c->name)) {
	  /* What is this cookie? */
	  nignored++;
	  continue;
	}

	/* Try to delete any invalid cookies. */
	if (cookie_to_credentials(c, NULL, &cr) == -1) {
	  if (cr != NULL) {
		if (cr->auth_style == AUTH_STYLE_OIDC) {
		  /*
		   * XXX the current jurisdiction may not be configured correctly
		   * to handle this - the federation must have administrative rules
		   * to ensure all jurisdictions are configured for this, or we
		   * need to redirect this user to the correct jurisdiction
		   * so that we have access to the configuration.
		   * Also: in the initial implementation at most one OIDC Provider
		   * is allowed (for the entire federation).
		   */
		  if (!name_eq(cr->home_jurisdiction, conf_val(CONF_JURISDICTION_NAME),
													   DACS_NAME_CMP_CONFIG))
			log_msg((LOG_WARN_LEVEL,
					 "OIDC credentials are from another jurisdiction"));
		  need_oidc_logout++;
		}

		void_cookie(&bp, cr);
		printf("%s", bp);
	  }
	  ninvalid++;
	  continue;
	}

	if ((username_arg == NULL
		 || name_eq(username_arg, cr->username, DACS_NAME_CMP_CONFIG))
		&& (jurisdiction_arg == NULL
			|| name_eq(jurisdiction_arg, cr->home_jurisdiction,
					   DACS_NAME_CMP_CONFIG))) {
	  if (cr->auth_style == AUTH_STYLE_OIDC) {
		/*
		 * XXX as above
		 */
		if (!name_eq(cr->home_jurisdiction, conf_val(CONF_JURISDICTION_NAME),
													 DACS_NAME_CMP_CONFIG))
		  log_msg((LOG_WARN_LEVEL,
				   "OIDC credentials are from another jurisdiction"));
		need_oidc_logout++;
	  }

	  void_cookie(&bp, cr);
	  printf("%s", bp);

	  cr->expires_secs = 0;
	  ndeleted++;
	  log_msg((LOG_DEBUG_LEVEL | LOG_AUDIT_FLAG,
			   "Signout succeeded for %s",
			   auth_identity_from_credentials_track(cr)));
#ifdef ENABLE_USER_INFO
	  user_info_signout(cr);
#endif
	}

	/* Link together the remaining credentials. */
	if (credentials == NULL)
	  cr->next = NULL;
	else
	  cr->next = credentials;
	credentials = cr;
  }

  log_msg((LOG_TRACE_LEVEL,
		   "ncookies=%u, nignored=%u, nselected=%u, ninvalid=%u",
		   ncookies, nignored, nselected, ninvalid));
  log_msg((LOG_TRACE_LEVEL,
		   "ndeleted=%u, need_oidc_logout=%d", ndeleted, need_oidc_logout));

  /*
   * If we're deleting the last set of credentials, only then remove the
   * selected credentials cookie.
   */
  if (nselected
	  && ((username_arg == NULL && jurisdiction_arg == NULL)
		  || credentials == NULL)) {
	make_set_void_selected_cookie_header(&bp);
	printf("%s", bp);
  }

#ifdef ENABLE_OIDC_AUTH
  if (is_oidc_dacs_only_logout)
	goto normal_logout;

  /*
   * If any DACS cookies from OIDC authentication were found,
   * check if the RP-Initiated Logout flow or the Front-Channel Logout flow
   * need to be followed.
   * If so, the usual DACS signout handlers are disabled.
   * The OIDC Provider should call the post_logout_redirect_uri which should
   * essentially provide that functionality.
   */
  if (is_oidc_frontchannel_logout) {
	printf("Cache-Control: no-cache\n");
	printf("\n");

	log_msg((LOG_TRACE_LEVEL, "OIDC Front-Channel Logout completed"));
	exit(0);
  }

  if (need_oidc_logout || is_oidc_rp_initiated_logout) {
	Ds ds;

  do_rp_logout:

	log_msg((LOG_DEBUG_LEVEL, "This is an RP-Initiated Logout"));
	if (need_oidc_logout > 1)
	  log_msg((LOG_WARN_LEVEL, "Found %d OIDC style credentials",
			   is_oidc_rp_initiated_logout));

	/*
	 * Suppress any handlers.
	 * Redirect to the end_session_endpoint, which will eventually
	 * redirect back to the post_logout_redirect_uri.
	 */
	if (client_id == NULL || end_session_endpoint == NULL
		|| post_logout_redirect_uri == NULL) {
	  /*
	   * This jurisdiction is not correctly configured for OIDC.
	   * XXX see comments above.
	   */
	  log_msg((LOG_ERROR_LEVEL,
			   "This jurisdiction is not configured for OIDC signout"));
	  goto normal_logout;
	}

	ds_init(&ds);
	ds_asprintf(&ds, "%s", end_session_endpoint);
	ds_asprintf(&ds, "%sclient_id=%s",
				(strchr(end_session_endpoint, (int) '?') != NULL) ? "&" : "?",
				url_encode(client_id, 0));
	ds_asprintf(&ds, "&post_logout_redirect_uri=%s",
				url_encode(post_logout_redirect_uri, 0));

	emit_http_header_redirect(stdout, ds_buf(&ds));

	log_msg((LOG_TRACE_LEVEL, "OIDC RP-Initiated Logout completed"));
	exit(0);
  }
#endif

 normal_logout:

  /*
   * Check if an XML or JSON reply was requested instead of HTML.
   * XXX it might be useful to include the redirect URL in the returned
   * document if it has been configured
   */
  if (emit_xml) {
	emit_xml_header(stdout, "dacs_current_credentials");
	printf("%s", make_xml_dacs_current_credentials(credentials, NULL, 0));
	emit_xml_trailer(stdout);

	exit(0);
  }

  if (emit_json) {
	emit_json_header(stdout, "dacs_current_credentials");
	printf("%s", make_json_dacs_current_credentials(credentials, NULL, 0));
	emit_json_trailer(stdout);

	exit(0);
  }

  /*
   * What follows is only for HTML output.
   */

  /*
   * Get the configured handler or a default.
   * If NULL is returned, something bad happened.
   */
  if ((handler = lookup_signout_handler()) == NULL) {
	log_msg((LOG_ALERT_LEVEL, "lookup_signout_handler() failed?!"));
	errmsg = "Configuration error";
	goto fail;
  }

  /*
   * If the SIGNOUT_HANDLER is configured for SIGNOUT_HANDLER_DEFAULT_URL,
   * use the URL given by the DACS_SIGNOUT_HANDLER parameter when present,
   * otherwise use the configured URL.
   * In all other cases a DACS_SIGNOUT_HANDLER parameter is silently ignored.
   * No validity check is performed on DACS_SIGNOUT_HANDLER.
   */
  if (handler->type == SIGNOUT_HANDLER_DEFAULT_URL
	  || handler->type == SIGNOUT_HANDLER_URL) {
	Ds ds;
	char *p, *url;

	if (handler->type == SIGNOUT_HANDLER_DEFAULT_URL
		&& dacs_signout_handler != NULL)
	  url = dacs_signout_handler;
	else
	  url = handler->signout_string;

	ds_init(&ds);
	ds_asprintf(&ds, "%s%s", url,
				(strchr(url, '?') != NULL) ? "&" : "?");
	ds_asprintf(&ds, "DACS_VERSION=%s", DACS_VERSION_NUMBER);

	if ((p = conf_val(CONF_FEDERATION_NAME)) != NULL)
	  ds_asprintf(&ds, "&DACS_FEDERATION=%s", p);

	if ((p = conf_val(CONF_JURISDICTION_NAME)) != NULL)
	  ds_asprintf(&ds, "&DACS_JURISDICTION=%s", p);

	ds_asprintf(&ds, "&DACS_SIGNOUT_RESULT=%d", ndeleted);

	emit_http_header_redirect(stdout, ds_buf(&ds));
  }
  else if (handler->type == SIGNOUT_HANDLER_FILE) {
	char *ptr;
	size_t file_size;

	if (load_file(handler->signout_string, &ptr, &file_size) == -1) {
	  log_msg((LOG_ALERT_LEVEL,
			   "Could not load file \"%s\" for SIGNOUT_HANDLER",
			   handler->signout_string));
	  handler_error = 1;
	}
	else {
	  fflush(stdout);
	  /* XXX Errors ignored... */
	  write_buffer(fileno(stdout), ptr, file_size);
	}
  }
  else if (handler->type == SIGNOUT_HANDLER_MESSAGE) {
	emit_html_header(stdout, NULL);
	printf("%s\n", handler->signout_string);
	emit_html_trailer(stdout);
  }

  if (handler_error || handler->type == SIGNOUT_HANDLER_CREDENTIALS) {
	emit_html_header(stdout, NULL);
	if (ndeleted) {
	  for (cr = credentials; cr != NULL; cr = cr->next) {
		if (cr->expires_secs == 0) {
		  if (!streq(cr->federation, conf_val(CONF_FEDERATION_NAME)))
			printf("<p>You have signed off as <tt><b>%s</b></tt><br>",
				   auth_identity(cr->federation, cr->home_jurisdiction,
								 cr->username, NULL));
		  else
			printf("<p>You have signed off as <tt><b>%s</b></tt><br>",
				   auth_identity(NULL, cr->home_jurisdiction, cr->username,
								 NULL));
		}
	  }
	}

	if (ncookies - ndeleted - ninvalid) {
	  printf("<p>");
	  for (cr = credentials; cr != NULL; cr = cr->next) {
		if (cr->expires_secs != 0)
		  printf("<b>You are still authenticated as <tt><b>%s</b></tt><br>",
				 auth_identity(cr->federation, cr->home_jurisdiction,
							   cr->username, NULL));
	  }
	  printf("</p>");
	}
	else
	  printf("<p>You are not authenticated within federation <b>%s</b>.</p>\n",
			 conf_val(CONF_FEDERATION_NAME));

	emit_html_trailer(stdout);
  }

  exit(0);

 fail:
  if (errmsg != NULL)
	log_msg((LOG_ERROR_LEVEL, "Failed: %s", errmsg));

  if (emit_xml) {
	Common_status status;

	emit_xml_header(stdout, "dacs_current_credentials");
	fprintf(stdout, "<%s", make_xml_root_element("dacs_current_credentials"));
	init_common_status(&status, NULL, NULL, errmsg);
	fprintf(stdout, "%s", make_xml_common_status(&status));
	fprintf(stdout, "</dacs_current_credentials>\n");
	emit_xml_trailer(stdout);
  }
  else if (emit_json) {
	Common_status status;

	emit_json_header(stdout, "dacs_current_credentials");
	init_common_status(&status, NULL, NULL, errmsg);
	printf("%s", make_json_common_status(&status));
	emit_json_trailer(stdout);
  }
  else {
	emit_html_header(stdout, NULL);
	printf("<p>An error occurred during the service request.\n");
	if (errmsg != NULL)
	  printf("<p>Reason: %s.\n", errmsg);
	emit_html_trailer(stdout);
  }

  exit(1);
}
