/*
 * Copyright (c) 2003-2025
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*****************************************************************************
 * COPYRIGHT AND PERMISSION NOTICE
 * 
 * Copyright (c) 2001-2003 The Queen in Right of Canada
 * 
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, and/or sell
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, provided that the above copyright notice(s) and this
 * permission notice appear in all copies of the Software and that both the
 * above copyright notice(s) and this permission notice appear in supporting
 * documentation.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE 
 * BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 * 
 * Except as contained in this notice, the name of a copyright holder shall not
 * be used in advertising or otherwise to promote the sale, use or other
 * dealings in this Software without prior written authorization of the
 * copyright holder.
 ***************************************************************************/

/*
 * Various string manipulation functions
 * These aren't necessarily the fastest or most space efficient algorithms,
 * but that's ok for our applications.
 *
 * Simple things should be simple, complex things should be possible.
 * -- Alan Kay
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2025\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: str.c 3365 2025-03-31 23:54:10Z brachman $";
#endif

#ifdef DSSLIB
#include "dsslib.h"
#else
#include "local.h"
#endif

static const char *log_module_name = "str";

/*
 * This is like strncpy() except the copy is always null-terminated.
 */
char *
strcpyn(char *dst, const char *src, size_t n)
{
  unsigned int ui;
  char *a;
  const char *b;

  a = dst;
  b = src;
  for (ui = 0; ui < n; ui++) {
	if ((*a++ = *b++) == '\0')
	  return(dst);
  }
  *a = '\0';
  return(dst);
}

/*
 * STR is one or more lines.
 * Fold continued lines (ones that end in backslash-newline) into one
 * long line.
 */
Ds *
strfold(char *str)
{
  char *p;
  Ds *ds;

  ds = ds_init(NULL);

  p = str;
  while (*p != '\0') {
	if (*p == '\\') {
	  /*
	   * Combine lines by not copying "\\n".
	   * Copy a backslash followed by anything else but delete/ignore
	   * a backslash if it is followed by a null.
	   */
	  if (*(p + 1) == '\n')
		p += 2;
	  else if (*(p + 1) != '\0') {
		ds_appendc(ds, (int) *p++);
		ds_appendc(ds, (int) *p++);
	  }
	  else
		p++;
	}
	else
	  ds_appendc(ds, (int) *p++);
  }

  ds_appendc(ds, (int) '\0');
  return(ds);
}

/*
 * Destructively trim up to LIMIT characters (unlimited if LIMIT == 0) in
 * the set TRIM_CHARS from the end of STR by putting a null character on top
 * of the leftmost such character.  Return the first argument.
 */
char *
strtrim(char *str, char *trim_chars, int limit)
{
  int n;
  char *p;

  p = str + strlen(str) - 1;
  n = 0;
  while ((!limit || n < limit) && p >= str
		 && strtr_char((int) *p, trim_chars, 0) == 1)
	p--, n++;

  *(p + 1) = '\0';

  return(str);
}

#ifndef HAVE_STRSEP
char *
strsep(char **stringp, const char *delim)
{
  char *s0, *s;

  if (stringp == NULL || *stringp == NULL || delim == NULL)
	return(NULL);

  /* Locate a char in *stringp matching any char in DELIM. */
  s0 = *stringp;
  for (s = *stringp;  *s != '\0';  s++) {
	const char *d;

	/* Look for a match of *s to any char in delim. */
	for (d = delim;  *d != '\0';  d++) {
	  if (*d == *s) {
		*s = '\0';
		*stringp = s + 1;
		return(s0);
	  }
    }
  }

  *stringp = NULL;
  return(s0);
}
#endif

#ifdef NOTDEF
/*
 * Return the length of the initial substring of s1 which does not
 * contain any characters in s2.
 * Use another algorithm on long strings.
 */
int
strcspn(char *s1, char *s2)
{
  char *p1, *p2;

  for (p1 = s1; *p1 != '\0'; p1++) {
	for (p2 = s2; *p2 != *p1 && *p2 != '\0'; p2++)
	  ;
	if (*p2 != '\0')
	  break;
  }
  return((int) (p1 - s1));
}
#endif

/*
 * Return the length of the initial substring of S1 that does not
 * contain any characters in S2, ignoring case.
 * Use another algorithm on long strings.
 */
int
strcasecspn(const char *s1, const char *s2)
{
  const char *p1, *p2;

  for (p1 = s1; *p1 != '\0'; p1++) {
	for (p2 = s2; tolower((int) *p2) != tolower((int) *p1) && *p2 != '\0'; p2++)
	  ;
	if (*p2 != '\0')
	  break;
  }
  return((int) (p1 - s1));
}

/*
 * Duplicate S until STOP_CH (or the end of S) is seen and return the new
 * string. If ENDP is not NULL, it will be set to point to STOP_CH if that
 * character was found, otherwise to the terminating null byte if STOP_CH
 * was not found.  This can be useful when parsing, for instance.
 * The function name is supposed to be analogous to 'strndup()'.
 * This is a functional subset of strcspndup().
 *
 * Also see memccpy(3).
 */
char *
strchrdup(const char *s, int stop_ch, char **endp)
{
  const char *p;

  for (p = s; *p != '\0'; p++) {
	if (*p == stop_ch)
	  break;
  }

  if (endp != NULL)
	*endp = (char *) p;

  if (*p == '\0')
	return(strdup(s));

  return(strndup(s, p - s));
}

/*
 * Duplicate S until any character in CHARSET (or the end of S) is seen, then
 * set ENDP (if non-NULL) to point to that character in S (or the NUL in S).
 * The function name is supposed to be analogous to 'strndup()'.
 * It can be useful when parsing, for instance.
 */
char *
strcspndup(const char *s, const char *charset, char **endp)
{
  const char *e, *p;

  for (p = s; *p != '\0'; p++) {
	if ((e = strchr(charset, (int) *p)) != NULL)
	  break;
  }

  if (endp != NULL)
	*endp = (char *) p;

  if (*p == '\0')
	return(strdup(s));

  return(strndup(s, p - s));
}

/*
 * Make a copy of characters from S until the first one that does not
 * satisfy ALLOWABLE (see is_strclass()).
 * If ENDP is not NULL, set it to point to the character in S that caused
 * copying to stop.
 * This is basically strcspndup() using classes of characters instead of
 * a list of specific characters.
 */
char *
strclassdup(const char *s, int allowable, char **endp)
{
  const char *p;

  for (p = s; *p != '\0'; p++) {
	if (!is_strclass_char((int) *p, allowable, NULL))
	  break;
  }

  if (endp != NULL)
	*endp = (char *) p;

  if (*p == '\0')
	return(strdup(s));

  return(strndup(s, p - s));
}

/*
 * Return a copy of STR, mapped to lowercase.
 */
char *
strtolower(const char *str)
{
  char *p, *q;

  p = strdup(str);
  for (q = p; *q != '\0'; q++) {
	if (isupper((int) *q))
	  *q = tolower((int) *q);
  }

  return(p);
}

/*
 * Map STR to lower case in supplied buffer BUF, assumed to be large enough.
 * They can be the same buffer.
 */
char *
strlowerbuf(char *buf, char *str)
{
  char *p, *q;

  for (p = str, q = buf; *p != '\0'; p++, q++)
	*q = tolower((int) *p);

  *q = '\0';

  return(buf);
}

/*
 * Return a copy of STR, mapped to uppercase.
 */
char *
strtoupper(const char *str)
{
  char *p, *q;

  p = strdup(str);
  for (q = p; *q != '\0'; q++) {
	if (islower((int) *q))
	  *q = toupper((int) *q);
  }

  return(p);
}

char *
strcapitalize(Ds *ods, char *str)
{
  char *s;
  Ds *ds;

  if ((ds = ods) == NULL)
	ds = ds_init(NULL);

  s = str;
  while (*s != '\0') {
	if (s == str) {
	  if (islower((int) *s))
		ds_appendc(ds, toupper((int) *s));
	  else
		ds_appendc(ds, (int) *s);
	}
	else if (isupper((int) *s))
	  ds_appendc(ds, tolower((int) *s));
	else
	  ds_appendc(ds, (int) *s);

	s++;
  }
  ds_appendc(ds, (int) '\0');

  return(ds_buf(ds));
}

/*********************************************************************/

/*
 * String to (binary) number conversions.
 */

/*
 * Convert the NBYTES long unsigned integer (or buffer) pointed to by VAL into
 * its base 2 (binary) string representation.
 * If STARTP is non-NULL, it is set to point to the most signficant non-zero
 * bit in the null-terminated output string (for the caller to elide leading
 * zeroes), or to the last zero if VAL is zero). If OUTBUF is non-NULL,
 * it will hold the output string and is assumed to be sufficiently large;
 * if OUTBUF is NULL, the string buffer is allocated.
 * A pointer to the output string is returned, or NULL if an argument is
 * invalid.
 * If BIG_ENDIAN is non-zero, VAL points to the most significant byte,
 * otherwise it points to the least significant byte. 
 */
char *
strui2b(unsigned char *val, size_t nbytes, int big_endian,
		char *outbuf, char **startp)
{
  int i, ind, incr, j, in_leading_zeroes, oind, zind;
  char *result;

  if (nbytes == 0 || val == NULL)
	return(NULL);

  if (outbuf != NULL)
	result = outbuf;
  else
	result = (char *) malloc(nbytes * 8 + 1);

  /*
   * Identify the index of the most significant byte of the unsigned integer
   * value and the increment to use to advance to its least significant byte.
   */
  if (big_endian) {
	ind = 0;
	incr = 1;
  }
  else {
	ind = nbytes - 1;
	incr = -1;
  }

  oind = 0;
  zind = 0;
  in_leading_zeroes = (startp != NULL);
  for (i = 0; i < nbytes; i++) {
	unsigned char mask;

	mask = 0x80;
	for (j = 0; j < 8; j++) {
	  char ch;

	  ch = result[oind++] = (*(val + ind) & mask) ? '1' : '0';
	  /* Count the leading zeroes. */
	  if (in_leading_zeroes && ch == '0')
		zind++;
	  else
		in_leading_zeroes = 0;
	  mask >>= 1;
	}
	ind += incr;
  }

  result[oind] = '\0';

  if (startp != NULL) {
	if (in_leading_zeroes)
	  *startp = result + oind - 1;
	else
	  *startp = result + zind;
  }

  return(result);
}

static void
test_strui2b(void)
{
  unsigned char xxx[16] = {
	0xff, 0x00, 0xff, 0x00, 0xff, 0x00, 0xff, 0x00,
	0xff, 0x00, 0xff, 0x00, 0xff, 0x00, 0xff, 0x01
  };

  fprintf(stderr, "littleEndian=%s\n", strui2b(xxx, 16, 0, NULL, NULL));
  fprintf(stderr, "bigEndian   =%s\n", strui2b(xxx, 16, 1, NULL, NULL));
}

/* Stringification. */
#define STRNUM_TYPENAME(NAME)		STRNUM_TYPENAME_X(NAME)
#define STRNUM_TYPENAME_X(NAME)		#NAME

/* XXX maybe this should be run-time extensible? */
/* XXX it might be handy to generate/define printf conversion specs. */
static Strnum_tab strnum_tab[] = {
  { STRNUM_I,      "STRNUM_I",
	sizeof(STRNUM_I_TYPE),   STRNUM_TYPENAME(STRNUM_I_TYPE) },
  { STRNUM_UI,     "STRNUM_UI",
	sizeof(STRNUM_UI_TYPE),  STRNUM_TYPENAME(STRNUM_UI_TYPE) },
  { STRNUM_L,      "STRNUM_L",
	sizeof(STRNUM_L_TYPE),   STRNUM_TYPENAME(STRNUM_L_TYPE) },
  { STRNUM_UL,     "STRNUM_UL",
	sizeof(STRNUM_UL_TYPE),  STRNUM_TYPENAME(STRNUM_UL_TYPE) },
  { STRNUM_LL,      "STRNUM_LL",
	sizeof(STRNUM_LL_TYPE),  STRNUM_TYPENAME(STRNUM_LL_TYPE) },
  { STRNUM_ULL,     "STRNUM_ULL",
	sizeof(STRNUM_ULL_TYPE), STRNUM_TYPENAME(STRNUM_ULL_TYPE) },
  { STRNUM_STR,     "STRNUM_STR",
    sizeof(STRNUM_STR_TYPE), STRNUM_TYPENAME(STRNUM_STR_TYPE) },
  { STRNUM_TIME_T,  "STRNUM_TIME_T",
	sizeof(STRNUM_TIME_T_TYPE), STRNUM_TYPENAME(STRNUM_TIME_T_TYPE) },
  { STRNUM_MODE_T,  "STRNUM_MODE_T",
	sizeof(STRNUM_MODE_T_TYPE), STRNUM_TYPENAME(STRNUM_MODE_T_TYPE) },
  { STRNUM_PID_T,   "STRNUM_PID_T",
	sizeof(STRNUM_PID_T_TYPE), STRNUM_TYPENAME(STRNUM_PID_T_TYPE) },
  { STRNUM_IN_PORT_T, "STRNUM_IN_PORT_T",
	sizeof(STRNUM_IN_PORT_T_TYPE), STRNUM_TYPENAME(STRNUM_IN_PORT_T_TYPE) },
  { STRNUM_SIZE_T,  "STRNUM_SIZE_T",
	sizeof(STRNUM_SIZE_T_TYPE), STRNUM_TYPENAME(STRNUM_SIZE_T_TYPE) },
  { STRNUM_SSIZE_T, "STRNUM_SSIZE_T",
	sizeof(STRNUM_SSIZE_T_TYPE), STRNUM_TYPENAME(STRNUM_SSIZE_T_TYPE) },
  { STRNUM_IMAX_T,  "STRNUM_IMAX_T",
	sizeof(STRNUM_IMAX_T_TYPE), STRNUM_TYPENAME(STRNUM_IMAX_T_TYPE) },
  { STRNUM_UIMAX_T, "STRNUM_UIMAX_T",
	sizeof(STRNUM_UIMAX_T_TYPE), STRNUM_TYPENAME(STRNUM_UIMAX_T_TYPE) },
  { STRNUM_UI32_T,  "STRNUM_PID_T",
	sizeof(STRNUM_UI32_T_TYPE), STRNUM_TYPENAME(STRNUM_UI32_T_TYPE) },
  { STRNUM_UI64_T,  "STRNUM_UI64_T",
	sizeof(STRNUM_UI64_T_TYPE), STRNUM_TYPENAME(STRNUM_UI64_T_TYPE) },
  { STRNUM_FLOAT,  "STRNUM_FLOAT",
	sizeof(STRNUM_FLOAT_TYPE), STRNUM_TYPENAME(STRNUM_FLOAT_TYPE) },
  { STRNUM_DOUBLE,  "STRNUM_DOUBLE",
	sizeof(STRNUM_DOUBLE_TYPE), STRNUM_TYPENAME(STRNUM_DOUBLE_TYPE) },
  { STRNUM_LDOUBLE, "STRNUM_LDOUBLE",
	sizeof(STRNUM_LDOUBLE_TYPE), STRNUM_TYPENAME(STRNUM_LDOUBLE_TYPE) },
  { STRNUM_ERR,  NULL, 0, NULL }
};

Strnum_tab *
strnum_ind(int ind)
{
  static int max_ind = -1;

  if (max_ind == -1) {
	int i;

	for (i = 0; strnum_tab[i].name != NULL; i++)
	  ;
	max_ind = i;
  }

  if (ind < 0 || ind >= max_ind)
	return(NULL);

  return(&strnum_tab[ind]);
}

Strnum_tab *
strnum_lookup_type(Strnum type)
{
  Strnum_tab *tab;

  tab = &strnum_tab[0];
  while (tab->name != NULL) {
	if (tab->type == type)
	  return(tab);
  }

  return(NULL);
}

Strnum_tab *
strnum_lookup_name(char *name)
{
  Strnum_tab *tab;

  tab = &strnum_tab[0];
  while (tab->name != NULL) {
	if (strcaseeq(name, tab->name))
	  return(tab);
  }

  return(NULL);
}

size_t
strnum_size(Strnum type)
{
  Strnum_tab *t;

  if ((t = strnum_lookup_type(type)) == NULL)
	return(0);

  return(t->size);
}

Strnum
strnum_type(char *name)
{
  Strnum_tab *s;

  if ((s = strnum_lookup_name(name)) == NULL)
	return(STRNUM_ERR);

  return(s->type);
}

/*
 * This function and its variants provide uniform APIs for converting
 * common string representations into their native binary representations,
 * with type and error checking.  For instance, instead of using
 * strtol(3), strtod(3), strtoul(3), and their variants, there is basically one
 * function.  This is similar to BSD's strtonum(3) approach, but portable,
 * generalized, and extensible.
 *
 * Return 0 if ok, -1 otherwise.
 */
int
strnum(char *str, Strnum type, void *value)
{
  Strnum_mode mode;

  if (value == NULL)
	mode = STRNUM_CHECK;
  else {
	if (type == STRNUM_STR)		/* Special case for backward compatibility. */
	  mode = STRNUM_ALLOC;
	else
	  mode = STRNUM_STORE;
  }

  return(strnum_b(str, type, mode, 10, value, NULL));
}

/*
 * How-to demo and some quick & dirty tests.
 */
static MAYBE_UNUSED int
test_strnum(char *str)
{
  int count, st;

#define CHECK(ST, FMT, VAR)												\
  if (ST == -1) {														\
	fprintf(stderr, "\nFailed test %d: \"%s\"\n", count, FMT);			\
	return(-1);															\
  }																		\
  else {																\
	fprintf(stderr, "%s[%d] ", (count == 1) ? "" : ", ", count);		\
	fprintf(stderr, FMT, VAR);											\
	count++;															\
  }

  fprintf(stderr, "Testing strnum(): ");

  count = 1;
  {
	char s[10], *sp, *strp;
	/* Copy into s[]. */
	memset(s, 0, sizeof(s));
	st = strnum_b(str, STRNUM_STR, STRNUM_STORE, 0, s, NULL);
	CHECK(st, "s=\"%s\"", s);

	/* Allocate a copy of str. */
	sp = NULL;
	st = strnum_b(str, STRNUM_STR, STRNUM_ALLOC, 0, &sp, NULL);
	CHECK(st, "sp=\"%s\"", sp);

	/* Set strp to str. */
	strp = NULL;
	st = strnum_b(str, STRNUM_STR, STRNUM_NOCOPY, 0, &strp, NULL);
	CHECK(st, "strp=\"%s\"", strp);
  }

  /* XXX should all be explicitly cast for reasonable conversion specs. */

  {
	int i, *ip;

	i = 0;
	st = strnum_b(str, STRNUM_I, STRNUM_STORE, 10, &i, NULL);
	CHECK(st, "i=%d", i);
	ip = NULL;
	st = strnum_b(str, STRNUM_I, STRNUM_ALLOC, 10, &ip, NULL);
	CHECK(st, "i=%d", *ip);
  }

  {
	unsigned int u, *up;

	u = 0;
	st = strnum_b(str, STRNUM_UI, STRNUM_STORE, 10, &u, NULL);
	CHECK(st, "u=%u", u);
	up = NULL;
	st = strnum_b(str, STRNUM_UI, STRNUM_ALLOC, 10, &up, NULL);
	CHECK(st, "up=%u", *up);
  }

  {
	in_port_t ipt, *iptp;

	ipt = 0;
	st = strnum_b(str, STRNUM_IN_PORT_T, STRNUM_STORE, 0, &ipt, NULL);
	CHECK(st, "ipt=%u", ipt);
	iptp = NULL;
	st = strnum_b(str, STRNUM_IN_PORT_T, STRNUM_ALLOC, 0, &iptp, NULL);
	CHECK(st, "ipt=%u", *iptp);
  }

  {
	long l, *lp;

	l = 0;
	st = strnum_b(str, STRNUM_L, STRNUM_STORE, 0, &l, NULL);
	CHECK(st, "l=%ld", l);
	lp = NULL;
	st = strnum_b(str, STRNUM_L, STRNUM_ALLOC, 0, &lp, NULL);
	CHECK(st, "l=%ld", *lp);
  }

  {
	unsigned long ul, *ulp;

	ul = 0;
	st = strnum_b(str, STRNUM_UL, STRNUM_STORE, 0, &ul, NULL);
	CHECK(st, "ul=%lu", ul);
	ulp = NULL;
	st = strnum_b(str, STRNUM_UL, STRNUM_ALLOC, 0, &ulp, NULL);
	CHECK(st, "ul=%lu", *ulp);
  }

  {
	unsigned long long ull, *ullp;

	ull = 0;
	st = strnum_b(str, STRNUM_ULL, STRNUM_STORE, 0, &ull, NULL);
	CHECK(st, "ull=%llu", ull);
	ullp = NULL;
	st = strnum_b(str, STRNUM_ULL, STRNUM_ALLOC, 0, &ullp, NULL);
	CHECK(st, "ull=%llu", *ullp);
  }

  {
	long long ll, *llp;

	ll = 0;
	st = strnum_b(str, STRNUM_LL, STRNUM_STORE, 0, &ll, NULL);
	CHECK(st, "ll=%lld", ll);
	llp = NULL;
	st = strnum_b(str, STRNUM_LL, STRNUM_ALLOC, 0, &llp, NULL);
	CHECK(st, "llp=%lld", *llp);
  }

  {
	time_t t, *tp;

	t = 0;
	st = strnum_b(str, STRNUM_TIME_T, STRNUM_STORE, 0, &t, NULL);
	CHECK(st, "t=%ld", t);
	tp = NULL;
	st = strnum_b(str, STRNUM_TIME_T, STRNUM_ALLOC, 0, &tp, NULL);
	CHECK(st, "t=%ld", *tp);
  }

  {
	pid_t p, *pp;

	p = 0;
	st = strnum_b(str, STRNUM_PID_T, STRNUM_STORE, 0, &p, NULL);
	CHECK(st, "p=%d", p);
	pp = NULL;
	st = strnum_b(str, STRNUM_PID_T, STRNUM_ALLOC, 0, &pp, NULL);
	CHECK(st, "p=%d", *pp);
  }

  {
	mode_t m, *mp;

	m = 0;
	st = strnum_b(str, STRNUM_MODE_T, STRNUM_STORE, 0, &m, NULL);
	CHECK(st, "m=%d", m);
	mp = NULL;
	st = strnum_b(str, STRNUM_MODE_T, STRNUM_ALLOC, 0, &mp, NULL);
	CHECK(st, "m=%d", *mp);
  }

  {
	size_t st, *stp;

	st = 0;
	st = strnum_b(str, STRNUM_SIZE_T, STRNUM_STORE, 0, &st, NULL);
	CHECK(st, "st=%ld", st);
	stp = NULL;
	st = strnum_b(str, STRNUM_SIZE_T, STRNUM_ALLOC, 0, &stp, NULL);
	CHECK(st, "st=%ld", *stp);
  }

  {
	ssize_t sst, *sstp;

	sst = 0;
	st = strnum_b(str, STRNUM_SSIZE_T, STRNUM_STORE, 0, &sst, NULL);
	CHECK(st, "sst=%ld", sst);
	sstp = NULL;
	st = strnum_b(str, STRNUM_SSIZE_T, STRNUM_ALLOC, 0, &sstp, NULL);
	CHECK(st, "sst=%ld", *sstp);
  }

  {
	intmax_t im, *imp;

	im = 0;
	st = strnum_b(str, STRNUM_IMAX_T, STRNUM_STORE, 0, &im, NULL);
	CHECK(st, "im=%ld", im);
	imp = NULL;
	st = strnum_b(str, STRNUM_IMAX_T, STRNUM_ALLOC, 0, &imp, NULL);
	CHECK(st, "imp=%ld", *imp);
  }

  {
	uintmax_t uim, *uimp;

	uim = 0;
	st = strnum_b(str, STRNUM_UIMAX_T, STRNUM_STORE, 0, &uim, NULL);
	CHECK(st, "uim=%lu", uim);
	uimp = NULL;
	st = strnum_b(str, STRNUM_UIMAX_T, STRNUM_ALLOC, 0, &uimp, NULL);
	CHECK(st, "uimp=%lu", *uimp);
  }

  {
	float f, *fp;

	f = 0.0;
	st = strnum_b(str, STRNUM_F, STRNUM_STORE, 0, &f, NULL);
	CHECK(st, "f=%f", f);
	fp = NULL;
	st = strnum_b(str, STRNUM_F, STRNUM_ALLOC, 0, &fp, NULL);
	CHECK(st, "f=%f", *fp);
  }

  {
	double d, *dp;

	d = 0.0;
	st = strnum_b(str, STRNUM_D, STRNUM_STORE, 0, &d, NULL);
	CHECK(st, "d=%lf", d);
	dp = NULL;
	st = strnum_b(str, STRNUM_D, STRNUM_ALLOC, 0, &dp, NULL);
	CHECK(st, "d=%lf", *dp);
  }

  {
	long double L, *Lp;

	L = 0.0;
	st = strnum_b(str, STRNUM_LD, STRNUM_STORE, 0, &L, NULL);
	CHECK(st, "L=%Lf", L);
	Lp = NULL;
	st = strnum_b(str, STRNUM_LD, STRNUM_ALLOC, 0, &Lp, NULL);
	CHECK(st, "L=%Lf", *Lp);
  }
#undef CHECK

  fprintf(stderr, " OK\n");
  return(0);
}

/*
 * Alternate interface that stops conversion at the first invalid
 * character and returns a pointer to that character.
 */
int
strnumx(const char *str, Strnum type, void *value, char **endp)
{
  Strnum_mode mode;

  if (value == NULL)
	mode = STRNUM_CHECK;
  else {
	if (type == STRNUM_STR)		/* Special case for backward compatibility. */
	  mode = STRNUM_ALLOC;
	else
	  mode = STRNUM_STORE;
  }

  return(strnum_b(str, type, mode, 10, value, endp));
}

/*
 * Convert STR to a number of type TYPE in base BASE, watching for errors.
 * As a special case, STR can be copied unchanged.
 * Upon success, 0 is returned, -1 otherwise.
 * In the event of an error, errno will (usually) be set to indicate
 * whether some sort of range error (ERANGE) or usage error (EINVAL) occurred.
 *
 * If mode is:
 * o STRNUM_CHECK, the function returns 0 but does not set the
 *   converted value.  This is useful for type checking.
 * o STRNUM_STORE and the conversion succeeds, VALUE is expected to
 *   be a pointer to properly aligned and sized memory for TYPE and the
 *   converted value is stored there.
 * o STRNUM_ALLOC and the conversion succeeds, then properly aligned
 *   and sized memory for TYPE is allocated, the converted value is stored
 *   in it. VALUE, expected to be a pointer of the correct type, is set to
 *   point to the allocated memory.
 * o STRNUM_NOCOPY is only valid for type STRNUM_STR.  VALUE, expected to be
 *   a (char **), is set to point to STR (no copying, no allocation).
 *
 * If mode is not STRNUM_CHECK, it is an error if VALUE is NULL.
 *
 * Note:
 * For a STRNUM_STR conversion, if STR is NULL and mode is STRNUM_STORE,
 * the first byte of value is set to a null byte; if mode is STRNUM_ALLOC
 * or STRNUM_NOCOPY, VALUE (a pointer) is set to NULL.
 *
 * If ENDP is non-NULL, it is set to point to the first invalid character
 * (with respect to the conversion type) found.  For STRNUM_STR, ENDP is never
 * set. If ENDP is NULL it is an error if an invalid character is found in STR,
 * otherwise no error will occur and ENDP will be set to point at that character.
 */
int
strnum_b(const char *str, Strnum type, Strnum_mode mode, int base,
		 void *value, char **endp)
{
  char *ptr;
  Strnum num_type;

  /* Reject various illegal usages. */
  if (*str == '\0' && type != STRNUM_STR) {
	errno = EINVAL;
	return(-1);
  }

  if (mode != STRNUM_CHECK && value == NULL) {
	errno = EINVAL;
	return(-1);
  }

  if (type != STRNUM_STR && mode == STRNUM_NOCOPY) {
	errno = EINVAL;
	return(-1);
  }

  /*
   * Map an explicitly-sized (unsigned) integer type to the appropriate
   * native type.
   */
  num_type = type;
  if (type == STRNUM_UI32_T) {
	/* 32-bit unsigned int. */
	if (sizeof(STRNUM_UI32_T_TYPE) == sizeof(STRNUM_UI_TYPE))
	  num_type = STRNUM_UI;
	else {
	  errno = EINVAL;
	  return(-1);
	}
  }
  else if (type == STRNUM_UI64_T) { 
	/* 64-bit unsigned int. */
	if (sizeof(STRNUM_UI64_T_TYPE) == sizeof(STRNUM_UI_TYPE))
	  num_type = STRNUM_UI;
	else if (sizeof(STRNUM_UI64_T_TYPE) == sizeof(STRNUM_UL_TYPE))
	  num_type = STRNUM_UL;
	else if (sizeof(STRNUM_UI64_T_TYPE) == sizeof(STRNUM_ULL_TYPE))
	  num_type = STRNUM_ULL;
	else {
	  errno = EINVAL;
	  return(-1);
	}
  }
  else if (type == STRNUM_I32_T) {
	/* 32-bit int. */
	if (sizeof(STRNUM_I32_T_TYPE) == sizeof(STRNUM_I_TYPE))
	  num_type = STRNUM_I;
	else {
	  errno = EINVAL;
	  return(-1);
	}
  }
  else if (type == STRNUM_I64_T) { 
	/* 64-bit int. */
	if (sizeof(STRNUM_I64_T_TYPE) == sizeof(STRNUM_I_TYPE))
	  num_type = STRNUM_I;
	else if (sizeof(STRNUM_I64_T_TYPE) == sizeof(STRNUM_L_TYPE))
	  num_type = STRNUM_L;
	else if (sizeof(STRNUM_I64_T_TYPE) == sizeof(STRNUM_LL_TYPE))
	  num_type = STRNUM_LL;
	else {
	  errno = EINVAL;
	  return(-1);
	}
  }

  errno = 0;
  switch (num_type) {
  case STRNUM_STR:
	{
	  /*
	   * This is a special but handy case, mostly for strpack().
	   * There is no conversion, just copy the string.
	   */
	  if (value != NULL && mode != STRNUM_CHECK) {
		if (str == NULL) {
		  if (mode == STRNUM_STORE)
			*(char *) value = '\0';
		  else if (mode == STRNUM_NOCOPY)
			*(char **) value = NULL;
		  else
			*(char **) value = NULL;
		}
		else {
		  if (mode == STRNUM_STORE)
			strcpy((char *) value, str);
		  else if (mode == STRNUM_NOCOPY)
			*(char **) value = (char *) str;	/* Do not copy the string. */
		  else
			*(char **) value = strdup(str);
		}
	  }

	  /* ENDP is not set. */
	  break;
	}
	
  case STRNUM_I:
  case STRNUM_IGZ:
  case STRNUM_INN:
	{
	  int val_int;
	  long val_long;

	  val_long = strtol(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_long == 0 && errno == EINVAL))
		return(-1);
	  val_int = (int) val_long;
	  if ((long) val_int != val_long) {
		errno = ERANGE;
		return(-1);
	  }

	  /* Does the integer meet the requested constraint? */
	  if  (num_type == STRNUM_IGZ && val_int <= 0) {
		errno = EINVAL;
		return(-1);
	  }
	  if  (num_type == STRNUM_INN && val_int < 0) {
		errno = EINVAL;
		return(-1);
	  }

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(int *) value = val_int;
		else {
		  *(int **) value = ALLOC(int);
		  **(int **) value = val_int;
		}
	  }
	  break;
	}

  case STRNUM_F:
	{
	  float val_float;

	  val_float = strtod(str, &ptr);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || (errno == ERANGE && val_float == 0.0)
		  || (errno == ERANGE
			  && (val_float == HUGE_VAL || val_float == -HUGE_VAL)))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(float *) value = val_float;
		else {
		  *(float **) value = ALLOC(float);
		  **(float **) value = val_float;
		}
	  }
	  break;
	}

  case STRNUM_D:
	{
	  double val_double;

	  val_double = strtod(str, &ptr);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || (errno == ERANGE && val_double == 0.0)
		  || (errno == ERANGE
			  && (val_double == HUGE_VAL || val_double == -HUGE_VAL)))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(double *) value = val_double;
		else {
		  *(double **) value = ALLOC(double);
		  **(double **) value = val_double;
		}
	  }
	  break;
	}

  case STRNUM_LD:
	{
	  long double val_ldouble;

	  val_ldouble = strtold(str, &ptr);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || (errno == ERANGE && val_ldouble == 0.0)
		  || (errno == ERANGE
			  && (val_ldouble == HUGE_VAL || val_ldouble == -HUGE_VAL)))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(long double *) value = val_ldouble;
		else {
		  *(long double **) value = ALLOC(long double);
		  **(long double **) value = val_ldouble;
		}
	  }
	  break;
	}

  case STRNUM_L:
	{
	  long val_long;

	  val_long = strtol(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_long == 0 && errno == EINVAL))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(long *) value = val_long;
		else {
		  *(long **) value = ALLOC(long);
		  **(long **) value = val_long;
		}
	  }
	  break;
	}

  case STRNUM_LL:
	{
	  long long val_long_long;

	  val_long_long = strtoll(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_long_long == 0 && errno == EINVAL))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(long long *) value = val_long_long;
		else {
		  *(long long **) value = ALLOC(long long);
		  **(long long **) value = val_long_long;
		}
	  }
	  break;
	}

  case STRNUM_IMAX_T:
	{
	  intmax_t val_intmax;

	  val_intmax = strtoimax(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_intmax == 0 && errno == EINVAL))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(intmax_t *) value = val_intmax;
		else {
		  *(intmax_t **) value = ALLOC(intmax_t);
		  **(intmax_t **) value = val_intmax;
		}
	  }
	  break;
	}

  case STRNUM_UI:
  case STRNUM_UINZ:
	{
	  unsigned int val_uint;
	  unsigned long val_ulong;

	  val_ulong = strtoul(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
		return(-1);
	  val_uint = (unsigned int) val_ulong;
	  if ((unsigned long) val_uint != val_ulong) {
		errno = ERANGE;
		return(-1);
	  }

	  /* Unsigned integer, but not zero. */
	  if (num_type == STRNUM_UINZ && val_uint == 0) {
		errno = EINVAL;
		return(-1);
	  }

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(unsigned int *) value = val_uint;
		else {
		  *(unsigned int **) value = ALLOC(unsigned int);
		  **(unsigned int **) value = val_uint;
		}
	  }
	  break;
	}

  case STRNUM_IN_PORT_T:
	{
	  in_port_t val_in_port_t;
	  unsigned long val_ulong;

	  val_ulong = strtoul(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
		return(-1);
	  val_in_port_t = (in_port_t) val_ulong;
	  if ((unsigned long) val_in_port_t != val_ulong) {
		errno = ERANGE;
		return(-1);
	  }

	  if (value != NULL) {
		if (mode == STRNUM_STORE)		  
		  *(in_port_t *) value = val_in_port_t;
		else {
		  *(in_port_t **) value = ALLOC(in_port_t);
		  **(in_port_t **) value = val_in_port_t;
		}
	  }
	  break;
	}

  case STRNUM_UL:
	{
	  unsigned long val_ulong;

	  val_ulong = strtoul(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(unsigned long *) value = val_ulong;
		else {
		  *(unsigned long **) value = ALLOC(unsigned long);
		  **(unsigned long **) value = val_ulong;
		}
	  }
	  break;
	}

  case STRNUM_ULL:
	{
	  unsigned long long val_ulong_long;

	  val_ulong_long = strtoull(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_ulong_long == 0 && errno == EINVAL))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(unsigned long long *) value = val_ulong_long;
		else {
		  *(unsigned long long **) value = ALLOC(unsigned long long);
		  **(unsigned long long **) value = val_ulong_long;
		}
	  }
	  break;
	}

  case STRNUM_UIMAX_T:
	{
	  uintmax_t val_uintmax;

	  val_uintmax = strtoumax(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_uintmax == 0 && errno == EINVAL))
		return(-1);

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(uintmax_t *) value = val_uintmax;
		else {
		  *(uintmax_t **) value = ALLOC(uintmax_t);
		  **(uintmax_t **) value = val_uintmax;
		}
	  }
	  break;
	}

  case STRNUM_TIME_T:
	{
	  unsigned long val_ulong;
	  time_t val_time_t;

	  val_ulong = strtoul(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
		return(-1);
	  val_time_t = (time_t) val_ulong;
	  if ((unsigned long) val_time_t != val_ulong) {
		errno = ERANGE;
		return(-1);
	  }

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		*(time_t *) value = val_time_t;
		else {
		  *(time_t **) value = ALLOC(time_t);
		  **(time_t **) value = val_time_t;
		}
	  }
	  break;
	}

  case STRNUM_SIZE_T:
	{
	  unsigned long val_ulong;
	  size_t val_size_t;

	  val_ulong = strtoul(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
		return(-1);
	  val_size_t = (size_t) val_ulong;
	  if ((unsigned long) val_size_t != val_ulong) {
		errno = ERANGE;
		return(-1);
	  }

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(size_t *) value = val_size_t;
		else {
		  *(size_t **) value = ALLOC(size_t);
		  **(size_t **) value = val_size_t;
		}
	  }
	  break;
	}

  case STRNUM_SSIZE_T:
	{
	  long val_long;
	  ssize_t val_ssize_t;

	  val_long = strtol(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_long == 0 && errno == EINVAL))
		return(-1);
	  val_ssize_t = (ssize_t) val_long;
	  if ((long) val_ssize_t != val_long) {
		errno = ERANGE;
		return(-1);
	  }

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		*(ssize_t *) value = val_ssize_t;
		else {
		  *(ssize_t **) value = ALLOC(ssize_t);
		  **(ssize_t **) value = val_ssize_t;
		}
	  }
	  break;
	}

  case STRNUM_PID_T:
	{
	  unsigned long val_ulong;
	  pid_t val_pid_t;

	  val_ulong = strtoul(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
		return(-1);
	  val_pid_t = (pid_t) val_ulong;
	  if ((unsigned long) val_pid_t != val_ulong) {
		errno = ERANGE;
		return(-1);
	  }

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		*(pid_t *) value = val_pid_t;
		else {
		  *(pid_t **) value = ALLOC(pid_t);
		  **(pid_t **) value = val_pid_t;
		}
	  }
	  break;
	}

  case STRNUM_MODE_T:
	{
	  unsigned long val_ulong;
	  mode_t val_mode_t;

	  val_ulong = strtoul(str, &ptr, base);
	  if (endp != NULL)
		*endp = ptr;
	  if ((endp == NULL && *ptr != '\0')
		  || errno == ERANGE || (val_ulong == 0 && errno == EINVAL))
		return(-1);
	  val_mode_t = (mode_t) val_ulong;
	  if ((unsigned long) val_mode_t != val_ulong) {
		errno = ERANGE;
		return(-1);
	  }

	  if (value != NULL) {
		if (mode == STRNUM_STORE)
		  *(mode_t *) value = val_mode_t;
		else {
		  *(mode_t **) value = ALLOC(mode_t);
		  **(mode_t **) value = val_mode_t;
		}
	  }
	  break;
	}

  default:
	if (endp != NULL)
	  *(const char **) endp = str;
	return(-1);
  }

  return(0);
}

/* Access union element VAR in Strnum_var variable X. */
#define STRNUM_U_VAR(X, VAR)	X->u.STRNUM_VAR_NAME(VAR)

/* To help generate some repetitive code below... */
#define STRNUM_VAR_CASE_CODE(VAR, VAR_ENUM, VAR_TYPE, VOID_VAL, FMT) \
  case VAR_ENUM:													 \
  STRNUM_U_VAR(VAR, VAR_ENUM) = *(VAR_TYPE *) VOID_VAL;				 \
  if (FMT != NULL)													 \
	VAR->str = ds_xprintf(FMT, STRNUM_U_VAR(VAR, VAR_ENUM));		 \
  else																 \
	VAR->str = NULL;												 \
  break;

/*
 * Where VAL points to memory containing a value of TYPE, convert it to
 * an actual variable of TYPE.  Also, if FMT is non-NULL, it is a printf(3)
 * type format that will be used to generate a string representation of VAL.
 * If OVAR is non-NULL, store the new representation there, otherwise allocate
 * a new structure.
 * Return the new representation, or NULL if an error occurs.
 */
Strnum_var *
strnum_var(Strnum_var *ovar, Strnum type, char *fmt, void *val)
{
  Strnum_var *var;

  if (val == NULL || type == STRNUM_ERR)
	return(NULL);

  if (ovar == NULL)
	var = ALLOC(Strnum_var);
  else
	var = ovar;

  var->type = type;
  switch (type) {
	/*
	 * Expand code for this case to illustrate what the code looks like.
	 */
  case STRNUM_I:
	/* The union elements have mixed-case variable names. */
	var->u.STRNUM_I_var = *(STRNUM_I_TYPE *) val;

	if (fmt != NULL)
	  var->str = ds_xprintf(fmt, var->u.STRNUM_I_var);
	else
	  var->str = NULL;
	break;

	/* Insert similar cases for the data types recognized by strnum(). */

	STRNUM_VAR_CASE_CODE(var, STRNUM_UI, STRNUM_UI_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_L, STRNUM_L_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_UL, STRNUM_UL_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_TIME_T, STRNUM_TIME_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_MODE_T, STRNUM_MODE_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_PID_T, STRNUM_PID_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_IN_PORT_T, STRNUM_IN_PORT_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_STR, STRNUM_STR_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_SIZE_T, STRNUM_SIZE_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_SSIZE_T, STRNUM_SSIZE_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_LL, STRNUM_LL_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_ULL, STRNUM_ULL_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_IMAX_T, STRNUM_IMAX_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_UIMAX_T, STRNUM_UIMAX_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_I32_T, STRNUM_I32_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_I64_T, STRNUM_I64_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_UI32_T, STRNUM_UI32_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_UI64_T, STRNUM_UI64_T_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_FLOAT, STRNUM_FLOAT_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_DOUBLE, STRNUM_DOUBLE_TYPE, val, fmt);
	STRNUM_VAR_CASE_CODE(var, STRNUM_LDOUBLE, STRNUM_LDOUBLE_TYPE, val, fmt);

  default:
	var->type = STRNUM_ERR;
	var->str = NULL;
	break;
  }

  return(var);
}

/*********************************************************************/

/*
 * String encoding and decoding.
 *
 * XXX the APIs have grown a little messy and should be refactored for
 * generality, consistency, and to eliminate redundancies.
 */

/*
 * The DACS64 encoding algorithm
 * This encoding scheme is similar to (but not identical to) the one
 * described in Section 6.8 of RFC 2045.
 * The reason for the difference is that we need to avoid the '=' and '+'
 * characters, which are used by RFC 2045 but cause problems in a variety of
 * contexts unless they are specially escaped - and we don't want to require
 * another layer of encoding.  Other characters can be similarly problematic.
 * We want something that can be used in a query string or URI, filename or
 * URI path, or MIME header value.  For instance, the '/' is a problem if the
 * value can appear in a path element or in a URI.
 * A comma is a problem because it can be used to separate multiple cookie
 * values in a Cookie header (see get_cookies()).
 *
 * Prior to 1.4.19, '.' mapped to 0 and '/' mapped to 1.
 * Starting with 1.4.19, we would also like to avoid '/', so we will map
 * '_' (0x5f) to 0 and '-' (0x2d) to 1.
 *
 * XXX consider changing '-' to '@' for even greater application, since
 * if we're generating a filename, a leading '-' can be interpreted as a
 * flag in a command line context.  This might be stretching it though.
 *
 * XXX Also see RFC 4648 (http://www.rfc-editor.org/rfc/rfc4648.txt) for the
 * "base64url" encoding, which appeared after this encoding was developed
 * and uses the same alphabet but a different mapping.
 */

enum {
  /* Lowest ASCII code in the table (055 == 45 == '-') */
  DACS64_TABLE_BASE     = 0x2d,
  /* Number of entries in the reverse lookup table */
  DACS64_DEC_TABLE_SIZE = 78,
  DACS64_ENC_TABLE_SIZE = 64,
  DACS64_TABLE_NOXBITS  = 16,
  DACS64_TABLE_XX       = 0x40
};

static const char dacs64_enc_table[DACS64_ENC_TABLE_SIZE] = {
  '_', '-', '0', '1', '2', '3', '4', '5',	/*  0.. 7 */
  '6', '7', '8', '9', 'A', 'B', 'C', 'D',	/*  8..15 */
  'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',	/* 16..23 */
  'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',	/* 24..31 */
  'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b',	/* 32..39 */
  'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',	/* 40..47 */
  'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r',	/* 48..55 */
  's', 't', 'u', 'v', 'w', 'x', 'y', 'z'	/* 56..63 */
};

/*
 * These are the mappings from a base-64 symbol to its decimal value.
 * DACS64_TABLE_XX represents ASCII values that do not map to anything.
 *
 * Originally, and many versions ago, this table mapped '.' to 0 and '/' to 1.
 */
#define XX DACS64_TABLE_XX
static const char dacs64_dec_table[DACS64_DEC_TABLE_SIZE] = {
  /* 0x2d */                                                       1, XX, XX,
  /* 0x30 */   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, XX, XX, XX, XX, XX, XX,
  /* 0x40 */  XX, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26,
  /* 0x50 */  27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, XX, XX, XX, XX,  0,
  /* 0x60 */  XX, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52,
  /* 0x70 */  53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63
};
#undef XX

/*
 * Base64URL encoder/decoder.
 * RFC 4648 S5
 */
enum {
  /* Lowest ASCII code in the table (055 == 45 == '-') */
  B64URL_TABLE_BASE     = 0x2d,
  /* Number of entries in the reverse lookup table */
  B64URL_DEC_TABLE_SIZE = 78,
  B64URL_ENC_TABLE_SIZE = 64,
  B64URL_TABLE_NOXBITS  = 0,
  B64URL_TABLE_XX       = 0x40
};

static const char b64url_enc_table[B64URL_ENC_TABLE_SIZE] = {
  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',	/*  0.. 7 */
  'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',	/*  8..15 */
  'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',	/* 16..23 */
  'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',	/* 24..31 */
  'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',	/* 32..39 */
  'o', 'p', 'q', 'r', 's', 't', 'u', 'v',	/* 40..47 */
  'w', 'x', 'y', 'z', '0', '1', '2', '3',	/* 48..55 */
  '4', '5', '6', '7', '8', '9', '-', '_'	/* 56..63 */
};

/*
 * These are the mappings from a base-64 symbol to its decimal value.
 * This table will still map '.' to 0 and '/' to 1 for backward
 * compatibility.  With suitable warning, this should be removed in a
 * future release. B64URL_TABLE_XX represents ASCII values that do not
 * map to anything.
 */
#define XX B64URL_TABLE_XX
static const char b64url_dec_table[B64URL_DEC_TABLE_SIZE] = {
  /* 0x2d */                                                      62, XX, XX,
  /* 0x30 */  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, XX, XX, XX, XX, XX, XX,
  /* 0x40 */  XX,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  /* 0x50 */  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, XX, XX, XX, XX, 63,
  /* 0x60 */  XX, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  /* 0x70 */  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51
};
#undef XX

/*
 * Base64 encoder/decoder.
 * RFC 4648 S4
 */
enum {
  /* Lowest ASCII code in the table (053 == 43 == '+') */
  B64_TABLE_BASE     = 0x2b,
  /* Number of entries in the reverse lookup table */
  B64_DEC_TABLE_SIZE = 80,
  B64_ENC_TABLE_SIZE = 64,
  B64_TABLE_NOXBITS  = 0,
  B64_TABLE_XX       = 0x40
};

static const char b64_enc_table[B64_ENC_TABLE_SIZE] = {
  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',	/*  0.. 7 */
  'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',	/*  8..15 */
  'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',	/* 16..23 */
  'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',	/* 24..31 */
  'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',	/* 32..39 */
  'o', 'p', 'q', 'r', 's', 't', 'u', 'v',	/* 40..47 */
  'w', 'x', 'y', 'z', '0', '1', '2', '3',	/* 48..55 */
  '4', '5', '6', '7', '8', '9', '+', '/'	/* 56..63 */
};

/*
 * These are the mappings from a base-64 symbol to its decimal value.
 * This table will still map '.' to 0 and '/' to 1 for backward
 * compatibility.  With suitable warning, this should be removed in a
 * future release. B64URL_TABLE_XX represents ASCII values that do not
 * map to anything.
 */
#define XX B64_TABLE_XX
static const char b64_dec_table[B64_DEC_TABLE_SIZE] = {
  /* 0x2b */                                              62, XX, XX, XX, 63,
  /* 0x30 */  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, XX, XX, XX, XX, XX, XX,
  /* 0x40 */  XX,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  /* 0x50 */  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, XX, XX, XX, XX, XX,
  /* 0x60 */  XX, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  /* 0x70 */  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51
};
#undef XX

/*
 * INP is an arbitrary string, NBYTES long.  Encode it to a null-terminated
 * ASCII text string, as radix-64, and set OUTP to point to an allocated
 * buffer.  This maps each 6 bits of INP into a printable character.
 * Return the length of the encoded (text) string, which excludes the
 * null byte.
 */
static Ds *
ds_strba64_internal(Ds *ods, Strba64_alg alg, const unsigned char *inp,
					unsigned int nbytes, const char *enc_table,
					unsigned int table_size, unsigned char table_noxbits)
{
  unsigned int fq, n, rem, rem_bits, val;
  const unsigned char *p;
  Ds *ds;

  if (ods == NULL) {
	if ((ds = ds_init(NULL)) == NULL)
	  return(NULL);
  }
  else
	ds = ods;

  n = nbytes;
  p = inp;
  rem = 0;

  rem_bits = 0;
  while (n > 0) {
	if (alg != STRBA64_DACS64 && rem_bits == 0 && n <= 2)
	  break;

	switch (rem_bits) {
	case 0:
	  val = (*p & 0xfc) >> 2;
	  ds_appendc(ds, enc_table[val]);
	  rem = *p & 0x03;
	  rem_bits = 2;
	  break;

	case 2:
	  val = (rem << 4) | (*p & 0xf0) >> 4;
	  ds_appendc(ds, enc_table[val]);
	  rem = *p & 0x0f;
	  rem_bits = 4;
	  break;

	case 4:
	  val = (rem << 2) | (*p & 0xc0) >> 6;
	  ds_appendc(ds, enc_table[val]);
	  val = *p & 0x3f;
	  ds_appendc(ds, enc_table[val]);
	  rem_bits = 0;
	  break;
	}

	p++;
	n--;
  }

  if (alg == STRBA64_DACS64) {
	/*
	 * An extra byte, containing any leftover bits, is always appended.
	 * There can be at most 4 leftover bits, so their maximum value is 15.
	 */
	if (rem_bits == 0) {
	  ds_appendc(ds, enc_table[table_noxbits]);
	}
	else if (rem < table_size)
	  ds_appendc(ds, enc_table[rem]);
	else {
	  fprintf(stderr, "ds_strba64_internal(): FIX ENCODING BUG!\n");
	  abort();
	  /*NOTREACHED*/
	}
  }
  else {
	rem_bits += (n * 8);
	if (n == 1) {
	  /*
	   * There is one byte to encode.
	   * Output two encoded bytes and two padding characters.
	   */
	  val = (*p & 0xfc) >> 2;
	  ds_appendc(ds, enc_table[val]);
	  rem = (*p & 0x03) << 4;
	  ds_appendc(ds, enc_table[rem]);
	  if (alg == STRBA64_B64) {
		ds_appendc(ds, '=');
		ds_appendc(ds, '=');
	  }
	}
	else if (n == 2) {
	  /*
	   * There are two bytes to encode.
	   * Output three encoded bytes and one padding character.
	   */
	  val = (*p & 0xfc) >> 2;
	  ds_appendc(ds, enc_table[val]);
	  rem = *p & 0x03;
	  p++;
	  val = (rem << 4) | ((*p & 0xf0) >> 4);
	  ds_appendc(ds, enc_table[val]);
	  rem = (*p & 0x0f) << 2;
	  ds_appendc(ds, enc_table[rem]);
	  if (alg == STRBA64_B64)
		ds_appendc(ds, '=');
	}
	else {
	  /* Should not be any other cases... */
	  if (rem_bits != 0)
		abort();
	}
  }

  ds_appendc(ds, (int) '\0');

  return(ds);
}

/*
 * DACS64 encoder.
 */
Ds *
ds_strba64(Ds *ods, const unsigned char *inp, unsigned int nbytes)
{
  Ds *ds;

  ds = ds_strba64_internal(ods, STRBA64_DACS64, inp, nbytes, dacs64_enc_table,
						   DACS64_ENC_TABLE_SIZE,
						   DACS64_TABLE_NOXBITS);

  return(ds);
}

/*
 * Return the base 64 encoding of INP (NBYTES long) using ALG.
 */
Ds *
ds_strba64_x(Ds *ods, Strba64_alg alg, const unsigned char *inp,
			 unsigned int nbytes)
{
  Ds *ds;

  ds = NULL;
  if (alg == STRBA64_DACS64)
	ds = ds_strba64_internal(ods, STRBA64_DACS64, inp, nbytes, dacs64_enc_table,
							 DACS64_ENC_TABLE_SIZE,
							 DACS64_TABLE_NOXBITS);
  else if (alg == STRBA64_B64)
	ds = ds_strba64_internal(ods, STRBA64_B64, inp, nbytes, b64_enc_table,
							 B64_ENC_TABLE_SIZE, 0);
  else if (alg == STRBA64_B64URL)
	ds = ds_strba64_internal(ods, STRBA64_B64URL, inp, nbytes, b64url_enc_table,
							 B64URL_ENC_TABLE_SIZE, 0);
  return(ds);
}

/*
 * Note that the DACS64 encoding algorithm cannot return a zero-length
 * string, unlike some other algorithms.
 * Therefore zero is returned here to indicate an error, since that is
 * not a valid encoding length.
 */
unsigned int
strba64(const unsigned char *inp, unsigned int nbytes, char **outp)
{
  Ds *ds;

  if ((ds = ds_strba64(NULL, inp, nbytes)) == NULL)
	return(0);

  if (outp != NULL)
	*outp = ds_buf(ds);

  return((int) ds_len(ds) - 1);
}

static Ds *
ds_stra64b_internal(Ds *ods, Stra64b_alg alg, const char *inp,
					int end_char,
					const char *dec_table, unsigned int table_base,
					unsigned int table_size, unsigned int table_noxbits,
					unsigned char table_xx)
{
  unsigned int ind, got_bits, val, x;
  char *p;
  Ds *ds;

  if (ods == NULL) {
    if ((ds = ds_init(NULL)) == NULL)
      return(NULL);
  }
  else
    ds = ods;

  if (*inp == '\0' || *inp == end_char)
	return(ds);

  p = (char *) inp;
  val = 0;

  got_bits = 0;
  while (*p != '\0' && *p != end_char) {
	if (alg == STRA64B_DACS64 && (*(p + 1) == '\0' || *(p +  1) == end_char))
	  break;
	if (alg == STRA64B_B64 && *p == '=')
	  break;

	if ((ind = *p - table_base) >= table_size
			 || (x = dec_table[ind]) == table_xx)
	  return(NULL);

	switch (got_bits) {
	case 0:
	  val = x << 2;
	  got_bits = 6;
	  break;

	case 6:
	  val |= ((x & 0x30) >> 4);
	  ds_appendc(ds, val);
	  val = (x & 0x0f) << 4;
	  got_bits = 4;
	  break;

	case 4:
	  val |= (x & 0x3c) >> 2;
	  ds_appendc(ds, val);
	  val = (x & 0x03) << 6;
	  got_bits = 2;
	  break;

	case 2:
	  val |= x;
	  ds_appendc(ds, val);
	  got_bits = 0;
	  break;
	}

	p++;
  }

  if (alg == STRA64B_DACS64) {
	if ((ind = *p - table_base) >= table_size
		|| (x = dec_table[ind]) == table_xx)
	  return(NULL);

	if (x != table_noxbits && alg != STRA64B_B64URL)
	  ds_appendc(ds, val | x);
  }

  return(ds);
}

/*
 * DACS64 decoder.
 */
Ds *
ds_stra64b(Ds *ods, const char *inp)
{
  Ds *ds;

  ds = ds_stra64b_internal(ods, STRA64B_DACS64, inp, 0, dacs64_dec_table,
						   DACS64_TABLE_BASE, DACS64_DEC_TABLE_SIZE,
						   DACS64_TABLE_NOXBITS, DACS64_TABLE_XX);

  return(ds);
}

/*
 * Return the base 64 decoding of INP using ALG.
 */
Ds *
ds_stra64b_x(Ds *ods, Stra64b_alg alg, const char *inp, int end_char)
{
  Ds *ds;

  if (alg == STRA64B_DACS64)
	ds = ds_stra64b_internal(ods, STRA64B_DACS64, inp, end_char,
							 dacs64_dec_table,
							 DACS64_TABLE_BASE, DACS64_DEC_TABLE_SIZE,
							 DACS64_TABLE_NOXBITS, DACS64_TABLE_XX);
  else if (alg == STRA64B_B64)
	ds = ds_stra64b_internal(ods, STRA64B_B64, inp, end_char,
							 b64_dec_table,
							 B64_TABLE_BASE, B64_DEC_TABLE_SIZE,
							 B64_TABLE_NOXBITS, B64_TABLE_XX);
  else if (alg == STRA64B_B64URL)
	ds = ds_stra64b_internal(ods, STRA64B_B64URL, inp, end_char,
							 b64url_dec_table,
							 B64URL_TABLE_BASE, B64URL_DEC_TABLE_SIZE,
							 B64URL_TABLE_NOXBITS, B64URL_TABLE_XX);
  else
	ds = NULL;

  return(ds);
}

/*
 * INP is a null-terminated ASCII string produced by strba64().
 * Convert INP to binary, storing it in OUTP, and setting NBYTES to its length.
 * OUTP and/or NBYTES can be NULL if its value isn't wanted
 * (when the output is known to be a null-terminated string, for example).
 * OUTP needs to be acceptably aligned.
 * Return a pointer to the output buffer and (when requested) set NBYTES to
 * the length of the output buffer in bytes.
 * If decoding fails, NULL is returned and NBYTES (if non-NULL) is set to zero.
 */
unsigned char *
stra64b(const char *inp, unsigned char **outp, unsigned int *nbytes)
{
  unsigned char *out;
  Ds ds;

  if (*inp == '\0') {
	if (nbytes != NULL)
	  *nbytes = 0;
	out = (unsigned char *) malloc(1);
	*out = '\0';
	if (outp != NULL)
	  *outp = out;
	return(out);
  }

  ds_init(&ds);

  if (ds_stra64b(&ds, inp) == NULL) {
	if (nbytes != NULL)
	  *nbytes = 0;
	return(NULL);
  }

  if (nbytes != NULL)
	*nbytes = ds_len(&ds);
  out = (unsigned char *) ds_buf(&ds);
  if (outp != NULL)
	*outp = out;

  return(out);
}

unsigned char *
strdec64(const char *inp)
{
  unsigned char *out;

  if (inp == NULL)
	return(NULL);

  out = stra64b(inp, NULL, NULL);

  return(out);
}

char *
strbenc64(const unsigned char *str, unsigned int len)
{
  char *out;

  if (str == NULL)
	return(NULL);

  strba64(str, len, &out);

  return(out);
}

char *
strenc64(char *str)
{
  char *out;

  if (str == NULL)
	return(NULL);

  /* Include the null byte. */
  out = strbenc64((unsigned char *) str, (unsigned int) strlen(str) + 1);

  return(out);
}
 
/*
 * Here is an algorithm to convert DACS64 to base64.
 * It may be useful for converting the DACS format into the base64 format that
 * can be decoded by a standard base64 library.
 * It can also be of use for test purposes:
 * encode-to-DACS64 -> convert-to-base64 -> decode-base64 -> compare to original
 *
 * 1. Ignore the DACS64 encoding's final byte (FB) and map everything else
 * from DACS64 to Base64
 *
 * 2. If FB is 'E', you are done.
 *
 * 3. Compute Remainder: 8 - ((total-encoded-length-in-bytes - 1) * 6) % 8
 * Remainder must be either 2 or 4 (otherwise something is wrong).
 *
 * 4. If Remainder is 2, append the Base64 suffix string for FB in the Case 2
 * Mapping Table (something ending in '=='), and you are done.
 *
 * 5. If Remainder is 4, append the Base64 suffix string for FB in the Case 4
 * Mapping Table (something ending in '='), and you are done.
 *
 * 6. Any other Remainder indicates an error
 *
 * Case 2 Mapping Table
 * If (unencoded input length in bits) % 6 == 2 (bits left over)
 * DACS64 final byte            -> Base64 (3 final bytes appended)
 *
 * Element : char in dacs64_enc_table -> Base64 mapping
 * 0 :  '_'   -> 'A=='  [0b00 << 4 == element  0 in mime64_conv_table]
 * 1 :  '-'   -> 'Q=='  [0b01 << 4 == element 16]
 * 2 :  '0'   -> 'g=='  [0b10 << 4 == element 32]
 * 3 :  '1'   -> 'w=='  [0b11 << 4 == element 48]
 *
 * Case 4 Mapping Table
 * If (unencoded input length in bits) % 6 == 4 (bits left over)
 * DACS64 final byte            -> Base64 (2 final bytes appended)
 *
 * Element : char in dacs64_enc_table -> Base64 mapping
 *  0 : '_'   -> 'A='   [0b0000 << 2 ==  element 0 in mime64_conv_table]
 *  1 : '-'   -> 'E='   [0b0001 << 2 ==  element 4]
 *  2 : '0'   -> 'I='   [0b0010 << 2 ==  8]
 *  3 : '1'   -> 'M='   [0b0011 << 2 == 12]
 *  4 : '2'   -> 'Q='   [0b0100 << 2 == 16]
 *  5 : '3'   -> 'U='   [0b0101 << 2 == 20]
 *  6 : '4'   -> 'Y='   [0b0110 << 2 == 24]
 *  7 : '5'   -> 'c='   [0b0111 << 2 == 28]
 *  8 : '6'   -> 'g='   [0b1000 << 2 == 32]
 *  9 : '7'   -> 'k='   [0b1001 << 2 == 36]
 * 10 : '8'   -> 'o='   [0b1010 << 2 == 40]
 * 11 : '9'   -> 's='   [0b1011 << 2 == 44]
 * 12 : 'A'   -> 'w='   [0b1100 << 2 == 48]
 * 13 : 'B'   -> '0='   [0b1101 << 2 == 52]
 * 14 : 'C'   -> '4='   [0b1110 << 2 == 56]
 * 15 : 'D'   -> '8='   [0b1111 << 2 == 60]
 *
 * Examples:
 * encode(dacs64, "A") gives "E-"
 * Discard '-', DACS64 'E' maps to character 16, which in Base64 is 'Q'
 * Remainder is 2   = 8 - ((2 - 1) * 6) % 8
 * Find '-' in Case 2 Mapping Table, get "Q=="
 * Base64: "QQ=="
 * encode(mime, "A") gives "QQ=="
 *
 * encode(dacs64,"12345678") gives "AH6nB1IqBn6"
 * "AH6nB1IqBn" maps to "MTIzNDU2Nz"
 * Remainder is 4   = 8 - ((11 - 1) * 6) % 8
 * Lookup '6' in Case 4 Mapping Table, get "g="
 * Base64: "MTIzNDU2Nzg="
 *
 * encode(dacs64,"1234567") gives "AH6nB1IqB1"
 * "AH6nB1IqB" maps to "MTIzNDU2N"
 * Remainder is 2    = 8 - ((10 - 1) * 6) % 8
 * Lookup '1' in Case 2 table, get "w=="
 * Base64: "MTIzNDU2Nw=="
 *
 * encode(dacs64, "abcde") gives "MK7XN43"
 * "MK7XN4" maps to "YWJjZG"
 * Remainder is 4    = 8 - ((7 - 1) * 6) % 8
 * Lookup '3' in Case 4 table, get "U="
 * Base64: "YWJjZGU="
 *
 * encode(dacs64, "1110") gives "AH2lA_"
 * "AH2lA" maps to "MTExM"
 * Remainder is 2    = 8 - ((6 - 1) * 6) % 8
 * Lookup '_' in Case 2 table, get "A=="
 * Base64: "MTExMA=="
 *
 */

enum {
  XX32                      = 0xff,
  CONV32_INPUT_GROUP        = 8,
  CONV32_MAX_VALUE          = 31,
  CONV32_MAX_PADDED_VALUE   = 32,
  CONV32_TABLE_SIZE			= 128
};
 
/*
 * RFC 4648 Base 32 mappings (symbol to decimal):
 *
 * 0=='A' (0x41/0x61),  9=='J' (0x4a/0x6a), 18=='S' (0x53/0x73), 27=='3' (0x33)
 * 1=='B' (0x42/0x62), 10=='K' (0x4b/0x6b), 19=='T' (0x54/0x74), 28=='4' (0x34)
 * 2=='C' (0x43/0x63), 11=='L' (0x4c/0x6c), 20=='U' (0x55/0x75), 29=='5' (0x35)
 * 3=='D' (0x44/0x64), 12=='M' (0x4d/0x6d), 21=='V' (0x56/0x76), 30=='6' (0x36)
 * 4=='E' (0x45/0x65), 13=='N' (0x4e/0x6e), 22=='W' (0x57/0x77), 31=='7' (0x37)
 * 5=='F' (0x46/0x66), 14=='O' (0x4f/0x6f), 23=='X' (0x58/0x78),
 * 6=='G' (0x47/0x67), 15=='P' (0x50/0x70), 24=='Y' (0x59/0x79),
 * 7=='H' (0x48/0x68), 16=='Q' (0x51/0x71), 25=='Z' (0x5a/0x7a),
 * 8=='I' (0x49/0x69), 17=='R' (0x52/0x72), 26=='2' (0x32)
 * (pad)=='=' (0x3d)
 */
static const unsigned char strba32_conv_table[CONV32_TABLE_SIZE] = {
   /* 00-07 (0x00-0x07) */
   XX32, XX32, XX32, XX32, XX32, XX32, XX32, XX32,
   /* 08-15 (0x08-0x0f) */
   XX32, XX32, XX32, XX32, XX32, XX32, XX32, XX32,
   /* 16-23 (0x10-0x17) */
   XX32, XX32, XX32, XX32, XX32, XX32, XX32, XX32,
   /* 24-31 (0x18-0x1f) */
   XX32, XX32, XX32, XX32, XX32, XX32, XX32, XX32,
   /* 32-39 (0x20-0x27) */
   XX32, XX32, XX32, XX32, XX32, XX32, XX32, XX32,
   /* 40-47 (0x28-0x2f) */
   XX32, 0x00, XX32, XX32, XX32, XX32, XX32, XX32,
   /* 48-55 (0x30-0x37) */
   XX32, XX32, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,	/* '2' .. '7' */
   /* 56-63 (0x38-0x3f) */
   XX32, XX32, XX32, XX32, XX32, 0x20, XX32, XX32,	/* '=' */
   /* 64-71 (0x40-0x47) */
   XX32, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,	/* 'A' .. 'G' */
   /* 72-79 (0x48-0x4f) */
   0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,	/* 'H' .. 'O' */
   /* 80-87 (0x50-0x57) */
   0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,	/* 'P' .. 'W' */
   /* 88-95 (0x58-0x5f) */
   0x17, 0x18, 0x19, XX32, XX32, XX32, XX32, XX32,	/* 'X' .. 'Z' */
   /* 96-103 (0x60-0x67) */
   XX32, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,	/* 'a' .. 'g' */
   /* 104-111 (0x68-0x6f) */
   0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,	/* 'h' .. 'o' */
   /* 112-119 (0x70-0x77) */
   0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,	/* 'p' .. 'w' */
   /* 120-127 (0x78-0x7f) */
   0x17, 0x18, 0x19, XX32, XX32, XX32, XX32, XX32	/* 'x' .. 'z' */
 };

/* Helper for stra32b(). */
static char
nextc(char **ptr)
{
  char ch;

  if ((ch = **ptr) == '\0')
	return('=');

  (*ptr)++;

  return(ch);
}

/*
 * Decode an RFC 4648 base-32 encoded string value to arbitrary bytes.
 * If IGNORE_KEY_LENGTH is non-zero, padding can be absent; if zero,
 * explicit padding is required.
 *
 * Examples (encoded strings exclude a terminating NUL):
 * "NBSWY3DPEB3W64TMMQGQU===" --> "hello world\r\n"
 * "IF2WOZ3JMUQEI33HM5UWK===" --> "Auggie Doggie"
 * "PB4HS6Q="                 --> "xxyz"
 * "MFRGGZDFMZTWQ2LKNNWG23TPOBYXE43UOV3HO6DZPIYDCMRTGQ2TMNZYHE======"
 *     --> "abcdefghijklmnopqrstuvwxyz0123456789"
 * "NI======" --> "j"
 * "MFRGGzdf" --> "abcde"
 * "JBSWY3DPEHPK3PXP" --> "Hello!0xde0xad0xbe0xef"
 * "MZXW6YTBOI======" --> "foobar"
 * For testing, also see online utilities, such as:
 * http://online-calculators.appspot.com/base32/
 *
 * RFC 4648 S3.2 specifies that "in some circumstances" the use of padding
 * in base-encoded data is not required or used "if the specification
 * referring to this document [RFC 4648] explicitly states otherwise".
 * Sure enough, some applications (e.g., TOTP-based authenticators) store
 * their base-32 encoded secrets without any padding, likely because in
 * some contexts (e.g., a URL query string) the '=' character can be
 * problematic.  Examples where padding "is not required and should be omitted":
 *   https://github.com/google/google-authenticator/wiki/Key-Uri-Format
 *   https://datatracker.ietf.org/doc/draft-linuxgemini-otpauth-uri/
 */
unsigned char *
stra32b_x(const char *inp, unsigned char **outp, unsigned int *nbytes,
		  int ignore_key_length)
{
  unsigned char *out;
  char *p;
  size_t remaining;
  Ds ds;

  if (*inp == '\0') {
	if (nbytes != NULL)
	  *nbytes = 0;
	out = (unsigned char *) malloc(1);
	*out = '\0';
	if (outp != NULL)
	  *outp = out;
	return(out);
  }

  remaining = strlen(inp);
  if (!ignore_key_length && (remaining % CONV32_INPUT_GROUP) != 0)
	return(NULL);

  ds_init(&ds);
  
  p = (char *) inp;
  while (remaining) {
	unsigned char in1, in2, in3, in4, in5, in6, in7, in8;
	unsigned char inv1, inv2, inv3, inv4, inv5, inv6, inv7, inv8;

#ifdef ORIG
	in1 = *p++;
	in2 = *p++;
	in3 = *p++;
	in4 = *p++;
	in5 = *p++;
	in6 = *p++;
	in7 = *p++;
	in8 = *p++;
	remaining -= 8;
#else
	in1 = nextc(&p);
	in2 = nextc(&p);
	in3 = nextc(&p);
	in4 = nextc(&p);
	in5 = nextc(&p);
	in6 = nextc(&p);
	in7 = nextc(&p);
	in8 = nextc(&p);

	if (ignore_key_length && remaining < 8)
	  remaining = 0;
	else
	  remaining -= 8;
#endif

	if (in1 >= CONV32_TABLE_SIZE || in2 >= CONV32_TABLE_SIZE
		|| in3 >= CONV32_TABLE_SIZE || in4 >= CONV32_TABLE_SIZE
		|| in5 >= CONV32_TABLE_SIZE || in6 >= CONV32_TABLE_SIZE
		|| in7 >= CONV32_TABLE_SIZE || in8 >= CONV32_TABLE_SIZE)
	  return(NULL);

	inv1 = strba32_conv_table[in1];
	inv2 = strba32_conv_table[in2];
	inv3 = strba32_conv_table[in3];
	inv4 = strba32_conv_table[in4];
	inv5 = strba32_conv_table[in5];
	inv6 = strba32_conv_table[in6];
	inv7 = strba32_conv_table[in7];
	inv8 = strba32_conv_table[in8];

	if (inv1 > CONV32_MAX_VALUE || inv2 > CONV32_MAX_VALUE)
	  return(NULL);

	if (inv3 > CONV32_MAX_PADDED_VALUE || inv4 > CONV32_MAX_PADDED_VALUE
		|| inv5 > CONV32_MAX_PADDED_VALUE || inv6 > CONV32_MAX_PADDED_VALUE
		|| inv7 > CONV32_MAX_PADDED_VALUE || inv8 > CONV32_MAX_PADDED_VALUE)
	  return(NULL);

	if (inv1 != CONV32_MAX_PADDED_VALUE
		&& inv2 != CONV32_MAX_PADDED_VALUE)
	  ds_appendc(&ds, (int) ((inv1 & 0x1f) << 3) | ((inv2 & 0x1c) >> 2));

	if (inv2 != CONV32_MAX_PADDED_VALUE
		&& inv3 != CONV32_MAX_PADDED_VALUE
		&& inv4 != CONV32_MAX_PADDED_VALUE)
	  ds_appendc(&ds, (int) ((inv2 & 0x03) << 6) | ((inv3 & 0x1f) << 1)
				 | ((inv4 & 0x10) >> 4));

	if (inv4 != CONV32_MAX_PADDED_VALUE
		&& inv5 != CONV32_MAX_PADDED_VALUE)
	  ds_appendc(&ds, (int) ((inv4 & 0x0f) << 4) | ((inv5 & 0x1e) >> 1));

	if (inv5 != CONV32_MAX_PADDED_VALUE
		&& inv6 != CONV32_MAX_PADDED_VALUE
		&& inv7 != CONV32_MAX_PADDED_VALUE)
	  ds_appendc(&ds, (int) ((inv5 & 0x01) << 7) | ((inv6 & 0x1f) << 2)
				 | ((inv7 & 0x18) >> 3));

	if (inv7 != CONV32_MAX_PADDED_VALUE
		&& inv8 != CONV32_MAX_PADDED_VALUE)
	  ds_appendc(&ds, (int) ((inv7 & 0x07) << 5) | (inv8 & 0x1f));
  }
#ifdef NOTDEF
  /* Not necessarily a text string, so don't null-terminate. */
 ds_appendc(&ds, (int) '\0');
#endif

 if (nbytes != NULL)
   *nbytes = ds_len(&ds);

 out = (unsigned char *) ds_buf(&ds);
 if (outp != NULL)
   *outp = out;

 return(out);
}

/*
 * By default, INP is expected to have correct padding.
 */
unsigned char *
stra32b(const char *inp, unsigned char **outp, unsigned int *nbytes)
{
  unsigned char *out;

  out = stra32b_x(inp, outp, nbytes, 0);

  return(out);
}

unsigned char *
strdec32(const char *inp)
{
  unsigned char *out;

  if (inp == NULL)
	return(NULL);

  out = stra32b(inp, NULL, NULL);

  return(out);
}

static inline int
b32ind(const unsigned char *base, unsigned int offset, int shift)
{
  int ch;
#ifdef NOTDEF
  static const char b32str[32] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";
#endif

  if (shift >= 0)
	ch = base[offset] << shift;
  else
	ch = base[offset] >> (-shift);

  return(ch);
}

/*
 * Convert binary to a base-32 ASCII string.
 * See strba32().
 */
Ds *
ds_strba32(Ds *ods, const unsigned char *inp, unsigned int nbytes,
		   int suppress_padding)
{
  int ch;
  unsigned int n;
  const unsigned char *p;
  Ds *ds;
  static const char b32str[32] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";

  if (ods == NULL) {
    if ((ds = ds_init(NULL)) == NULL)
      return(NULL);
  }
  else
    ds = ods;

  n = nbytes;
  p = inp;

  while (n > 0) {
	ch = b32str[b32ind(p, 0, -3) & 0x1f];
	ds_appendc(ds, ch);

	n--;
	if (n > 0)
	  ch = b32str[(b32ind(p, 0, 2) + b32ind(p, 1, -6)) & 0x1f];
	else
	  ch = b32str[b32ind(p, 0, 2) & 0x1f];
	ds_appendc(ds, ch);

	if (n > 0)
	  ch = b32str[b32ind(p, 1, -1) & 0x1f];
	else {
	  if (suppress_padding)
		break;
	  ch = '=';
	}
	ds_appendc(ds, ch);

	if (n > 0) {
	  if (--n > 0)
		ch = b32str[(b32ind(p, 1, 4) + b32ind(p, 2, -4)) & 0x1f];
	  else
		ch = b32str[b32ind(p, 1, 4) & 0x1f];
	}
	else {
	  if (suppress_padding)
		break;
	  ch = '=';
	}
	ds_appendc(ds, ch);

	if (n > 0) {
	  if (--n > 0)
		ch = b32str[(b32ind(p, 2, 1) + b32ind(p, 3, -7)) & 0x1f];
	  else
		ch = b32str[b32ind(p, 2, 1) & 0x1f];
	}
	else {
	  if (suppress_padding)
		break;
	  ch = '=';
	}
	ds_appendc(ds, ch);

	if (n > 0)
	  ch = b32str[b32ind(p, 3, -2) & 0x1f];
	else {
	  if (suppress_padding)
		break;
	  ch = '=';
	}
	ds_appendc(ds, ch);

	if (n > 0) {
	  if (--n > 0)
		ch = b32str[(b32ind(p, 3, 3) + b32ind(p, 4, -5)) & 0x1f];
	  else
		ch = b32str[b32ind(p, 3, 3) & 0x1f];
	}
	else {
	  if (suppress_padding)
		break;
	  ch = '=';
	}
	ds_appendc(ds, ch);

	if (n > 0) {
	  ch = b32str[b32ind(p, 4, 0) & 0x1f];
	  n--;
	}
	else {
	  if (suppress_padding)
		break;
	  ch = '=';
	}
	ds_appendc(ds, ch);

	p += 5;
  }

  ds_appendc(ds, '\0');

  return(ds);
}

/*
 * INP is an arbitrary string, NBYTES long.  Encode it into a null-terminated
 * ASCII text string, as RFC 4648 "base32", and set OUTP to point to an
 * allocated buffer.
 * Return the length of the encoded (text) string, which excludes the
 * null byte.
 */
unsigned int
strba32(const unsigned char *inp, unsigned int nbytes, int suppress_padding,
		char **outp)
{
  Ds *ds;

  ds = ds_strba32(NULL, inp, nbytes, suppress_padding);

  if (outp != NULL)
	*outp = ds_buf(ds);

  return(ds_len(ds) - 1);
}

char *
strbenc32(const unsigned char *str, unsigned int len)
{
  char *out;

  if (str == NULL)
	return(NULL);

  strba32(str, len, 0, &out);

  return(out);
}

/*
 * Like strbenc32(), except suppress padding characters.
 */
char *
strbenc32_np(const unsigned char *str, unsigned int len)
{
  char *out;

  if (str == NULL)
	return(NULL);

  strba32(str, len, 1, &out);

  return(out);
}

char *
strenc32(char *str)
{
  char *out;

  if (str == NULL)
	return(NULL);

  /* Include the null byte. */
  out = strbenc32((unsigned char *) str, (unsigned int) strlen(str) + 1);

  return(out);
}

/*
 * Different encoded character sets are used for the radix-85 representation,
 * also called ASCII85.  This table, which uses characters 0x21 through 0x75,
 * is used by PostScript and PDF (ASCII85Decode).  See
 *    http://en.wikipedia.org/wiki/Ascii85
 * and Adobe PDF Reference, Version 1.7, (November, 2006), Section 3.3.2.
 * The relevant RFC is:
 *    http://www.ietf.org/rfc/rfc1924.txt
 * although that uses a different mapping:
 *   '0'..'9', 'A'..'Z', 'a'..'z', '!', '#', '$', '%', '&', '(',
 *   ')', '*', '+', '-', ';', '<', '=', '>', '?', '@', '^', '_',
 *   '`', '{', '|', '}', '~'
 */
static MAYBE_UNUSED const char conv85_std_table[85] = {
  '!', '"', '#', '$', '%', '&', '\'', '(', ')', '*',
  '+', ',', '-', '.', '/',
  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
  ':', ';', '<', '=', '>', '?', '@',
  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
  'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
  'U', 'V', 'W', 'X', 'Y', 'Z',
  '[', '\\', ']', '^', '_', '`',
  'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
  'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
  'u'
};

static MAYBE_UNUSED const unsigned char conv85_std_test1[] = {
  0x12, 0x34, 0x56, 0x78, 0x9a
};
static MAYBE_UNUSED const char *conv85_std_enc1 = "&i<X6RK";

/*
 * Some characters in the "standard" radix-85 encoded character set are not
 * well-suited to appearing in the query component of a URL, so we need to
 * make some adjustments.  This change should not matter much because the
 * encoding should only be used internally (between DACS components).
 * Troublesome characters:
 * '#' (start of fragment), '=' (variable/value internal separator),
 * '&' (variable/value external separator), '%' (hex encoding escape),
 * '+' (blank encoding)
 * '?' (allowed, but may be confusing), '\\' (allowed, but may be confusing)
 * '~' is reserved for the start/end delimiting pair.
 */
static MAYBE_UNUSED const char conv85_dacs_table[85] = {
  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
  'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
  'U', 'V', 'W', 'X', 'Y', 'Z',
  'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
  'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
  'u', 'v', 'w', 'x', 'y', 'z',
  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
  '!', '"', '#', '$', '\'', '(', ')', '*', ',', '-',
  '.', ':', ';', '^', '>', '_', '@', '{', '|', '}',
  '`', '[', ']'
};

int strba85_demark = 0;
const char *conv85_enc = conv85_std_table;

static void
encode85(Ds *ds, const char *enc_tab, unsigned int val, unsigned int count)
{
  int i;
  unsigned char enc[5], *p;

  /* Compute them "forwards" but generate output in reverse. */
  p = enc;
  for (i = 0; i < 5; i++) {
	*p++ = val % 85;
	val /= 85;
  }

  for (i = 0; i <= count; i++)
	ds_appendc(ds, (int) enc_tab[*--p]);
}

Ds *
ds_strba85(Ds *ods, const unsigned char *inp, unsigned int nbytes)
{
  unsigned int count, i, val;
  const unsigned char *p;
  Ds *ds;

  if (ods == NULL) {
	if ((ds = ds_init(NULL)) == NULL)
	  return(NULL);
  }
  else
	ds = ods;

  count = 0;
  val = 0;
  p = inp;

  if (strba85_demark) {
	ds_appendc(ds, (int) '<');
	ds_appendc(ds, (int) '~');
  }

  for (i = 0; i < nbytes; i++) {
	switch (count++) {
	case 0:
	  val |= (*p++ << 24);
	  break;

	case 1:
	  val |= (*p++ << 16);
	  break;

	case 2:
	  val |= (*p++ << 8);
	  break;

	case 3:
	  val |= *p++;
	  if (val == 0)
		ds_appendc(ds, (int) 'z');
	  else
		encode85(ds, conv85_enc, val, count);
	  val = count = 0;
	  break;
	}
  }

  if (count)
	encode85(ds, conv85_enc, val, count);

  if (strba85_demark) {
	ds_appendc(ds, (int) '~');
	ds_appendc(ds, (int) '>');
  }

  ds_appendc(ds, '\0');

  return(ds);
}

/*
 * INP is a binary string, NBYTES long.  Convert it to a null-terminated
 * ASCII text string, as radix-85, and set OUTP to point to an allocated
 * buffer.  This maps each 4 bytes of INP into 5 printable characters.
 * Return the length of the text string.
 * We optionally emit the conventional leading "<~" and trailing "~>",
 * because they are not strictly necessary for our purposes, but always
 * recognize them when decoding.
 *
 * Return the length of the encoded (text) string, which excludes the
 * null byte.
 */
unsigned int
strba85(const unsigned char *inp, unsigned int nbytes, char **outp)
{
  Ds *ds;

  if ((ds = ds_strba85(NULL, inp, nbytes)) == NULL)
	return(0);

  if (outp != NULL)
	*outp = ds_buf(ds);

  return((unsigned int) ds_len(ds) - 1);
}

static MAYBE_UNUSED void
test85(void)
{
  char *out;

  conv85_enc = conv85_std_table;
  strba85(conv85_std_test1, 5, &out);
  printf("%s", out);
  if (!streq(out, conv85_std_enc1))
	printf(" (wrong!)\n");
  else
	printf(" (ok!)\n");
}
  
static void
decode85(Ds *ds, unsigned int val, int count)
{

  switch (count) {
  case 4:
	ds_appendc(ds, (int) (val >> 24));
	ds_appendc(ds, (int) (val >> 16));
	ds_appendc(ds, (int) (val >>  8));
	ds_appendc(ds, (int) val);
	break;

  case 3:
	ds_appendc(ds, (int) (val >> 24));
	ds_appendc(ds, (int) (val >> 16));
	ds_appendc(ds, (int) (val >>  8));
	break;

  case 2:
	ds_appendc(ds, (int) (val >> 24));
	ds_appendc(ds, (int) (val >> 16));
	break;

  case 1:
	ds_appendc(ds, (int) (val >> 24));
	break;
  }
}

static int
lookup85(const char *enc_tab, int ch)
{
  int i;

  for (i = 0; i < 85; i++) {
	if ((int) enc_tab[i] == ch)
	  return(i);
  }

  return(-1);
}

/*
 * INP is a null-terminated ASCII string produced by strba85().
 * OUTP and/or NBYTES can be NULL if their value isn't wanted
 * (when the output is known to be a null-terminated string, for example).
 * Internal whitespace is ignored.
 * Convert it to binary, storing it in OUTP, and setting NBYTES to its length.
 * OUTP needs to be acceptably aligned.
 * Return a pointer to the output buffer and set NBYTES to the number of
 * bytes copied to OUTP.
 * If decoding fails, NULL is returned and NBYTES is set to zero.
 */
unsigned char *
stra85b(const char *inp, unsigned char **outp, unsigned int *nbytes)
{
  unsigned int ch, count, v85, val;
  const char *p;
  Ds ds;
  static unsigned int pow85[5] = {
	(85 * 85 * 85 * 85), (85 * 85 * 85), (85 * 85), 85, 1
  };

  p = inp;

  if (*p == '<' && *(p + 1) == '~')
	p += 2;

  count = 0;
  val = 0;
  ds_init(&ds);

  while ((ch = *p++) != '\0') {
	switch (ch) {
	case 'z':
	  if (count) {
		/* z inside 5-tuple? */
		return(NULL);
	  }

	  ds_appendc(&ds, 0);
	  ds_appendc(&ds, 0);
	  ds_appendc(&ds, 0);
	  ds_appendc(&ds, 0);
	  break;

	case '~':
	  if (*p != '>')
		return(NULL);

	  /* This is the end. */
	  goto done;
	  /*NOTREACHED*/

	case '\n':
	case '\r':
	case '\t':
	case ' ':
	  break;

	default:
	  if ((v85 = lookup85(conv85_enc, ch)) == -1) {
		/* Bad character */
		return(NULL);
	  }
	  val += v85 * pow85[count++];
	  if (count == 5) {
		decode85(&ds, val, 4);
		count = 0;
		val = 0;
	  }
	  break;
	}
  }

 done:

  if (count > 0) {
	count--;
 	val += pow85[count];
	decode85(&ds, val, count);
  }

  if (nbytes != NULL)
	*nbytes = ds_len(&ds);
  if (outp != NULL)
	*outp = (unsigned char *) ds_buf(&ds);

  return((unsigned char *) ds_buf(&ds));
}

typedef struct Str_test_encdec {
  char *name;				/* Name of encode/decode algorithm. */
  char *test_str;			/* String to encode. */
  size_t test_str_len;		/* Optional explicit length of test_str. */
  char *enc_str;			/* Expected encoded string. */
  size_t enc_str_len;		/* Optional explicit length of enc_str. */
} Str_test_encdec;

/*
 * Keep these grouped by name.
 */
static Str_test_encdec str_test_encdec[] = {
  /* RFC 4648 S10. */
  { "base64", "",       0, "",         0 },
  { "base64", "f",      0, "Zg==",     0 },
  { "base64", "fo",     0, "Zm8=",     0 },
  { "base64", "foo",    0, "Zm9v",     0 },
  { "base64", "foob",   0, "Zm9vYg==", 0 },
  { "base64", "fooba",  0, "Zm9vYmE=", 0 },
  { "base64", "foobar", 0, "Zm9vYmFy", 0 },
  
  /* RFC 4648 S9 and RFC 2440 S6.5 */
  { "base64", "\x14\xfb\x9c\x03\xd9\x7e",  6, "FPucA9l+", 0 },
  { "base64", "\x14\xfb\x9c\x03\xd9",      5, "FPucA9k=", 0 },
  { "base64", "\x14\xfb\x9c\x03",          4, "FPucAw==", 0 },

  { "base64",    "<<??\?>>",  0, "PDw/Pz8+Pg==", 0 },
  { "base64",    "abc1230901", 0, "YWJjMTIzMDkwMQ==", 0 },
  { "base64",    "\x62\x6c\xc3\xa5\x62\xc3\xa6\x72\x67\x72\xc3\xb8\x64", 0,
	"YmzDpWLDpnJncsO4ZA==", 0 },
  { "base64",    "This is an encoded string", 0,
	"VGhpcyBpcyBhbiBlbmNvZGVkIHN0cmluZw==", 0 },
  { "base64",    "light w",   0, "bGlnaHQgdw==", 0 },
  { "base64",    "light wo",  0, "bGlnaHQgd28=", 0 },
  { "base64",    "light wor", 0, "bGlnaHQgd29y", 0 },
  { "base64",    "Many hands make light work.", 0,
	"TWFueSBoYW5kcyBtYWtlIGxpZ2h0IHdvcmsu", 0 },

  { "base64url", "f",      0, "Zg",       0 },
  { "base64url", "fo",     0, "Zm8",      0 },
  { "base64url", "foo",    0, "Zm9v",     0 },
  { "base64url", "foob",   0, "Zm9vYg",   0 },
  { "base64url", "fooba",  0, "Zm9vYmE",  0 },
  { "base64url", "foobar", 0, "Zm9vYmFy", 0 },

  { "base64url", "<<??\?>>", 0, "PDw_Pz8-Pg", 0 },
  { "base64url", "abc1230901", 0, "YWJjMTIzMDkwMQ", 0 },
  { "base64url", "\x62\x6c\xc3\xa5\x62\xc3\xa6\x72\x67\x72\xc3\xb8\x64", 0,
	"YmzDpWLDpnJncsO4ZA", 0 },
  { "base64url", "light w",   0, "bGlnaHQgdw", 0 },
  { "base64url", "light wo",  0, "bGlnaHQgd28", 0 },
  { "base64url", "light wor", 0, "bGlnaHQgd29y", 0 },
  { "base64url", "Many hands make light work.", 0,
	"TWFueSBoYW5kcyBtYWtlIGxpZ2h0IHdvcmsu", 0 },

  /* RFC 7515  S3.3 */
  { "base64url", "{\"typ\":\"JWT\",\x0d\x0a \"alg\":\"HS256\"}", 0,
	"eyJ0eXAiOiJKV1QiLA0KICJhbGciOiJIUzI1NiJ9", 0 },
  { "base64url", "{\"iss\":\"joe\",\x0d\x0a \"exp\":1300819380,\x0d\x0a \"http://example.com/is_root\":true}", 0,
	"eyJpc3MiOiJqb2UiLA0KICJleHAiOjEzMDA4MTkzODAsDQogImh0dHA6Ly9leGFtcGxlLmNvbS9pc19yb290Ijp0cnVlfQ", 0 },

  { "dacs64", "\000\001\002\003\004",  5, "__20_k2", 0 },
  { "dacs64",  "\x01\x02\x03\x04\x05", 5, "_E61-_3", 0 },
  { "dacs64",  "Hello, world.", 0,   "G4JgP4wg65RjQalY90", 0 },
  { "dacs64", "f",      0, "N0", 0 },
  { "dacs64", "fo",     0, "NaD",       0 },
  { "dacs64", "foo",    0, "NaxjE",     0 },
  { "dacs64", "foob",   0, "NaxjM0",    0 },
  { "dacs64", "fooba",  0, "NaxjMa-",   0 },
  { "dacs64", "foobar", 0, "NaxjMa3mE", 0 },

  { "dacs64",    "<<??\?>>", 0, "D1kzDnwyD0", 0 },

  { "hex", "ZZZ", 0,         "5a5a5a", 0 },
  { "hex", "\0\0\0\0\0", 5,  "0000000000", 0 },
  { "hex", "hexadecimal", 0, "68657861646563696d616c", 0 },

  { "base32", "j", 0,               "NI======", 0 },
  { "base32", "hello world\r\n", 0, "NBSWY3DPEB3W64TMMQGQU===", 0 },
  { "base32", "Auggie Doggie", 0,   "IF2WOZ3JMUQEI33HM5UWK===", 0 },
  { "base32", "xxyz", 0,            "PB4HS6Q=", 0 },
  { "base32", "abcdefghijklmnopqrstuvwxyz0123456789", 0,
	"MFRGGZDFMZTWQ2LKNNWG23TPOBYXE43UOV3HO6DZPIYDCMRTGQ2TMNZYHE======", 0 },
  { "base32", "abcde", 0,           "MFRGGzdf", 0 },
  { "base32", "Hello!\xde\xad\xbe\xef", 10, "JBSWY3DPEHPK3PXP", 0 },
  { "base32", "foobar", 0,          "MZXW6YTBOI======", 0 },

  /* RFC 4648 S10. */
  { "base32", "", 0,                "",         0 },
  { "base32", "f", 0,               "MY======", 0 },
  { "base32", "fo", 0,              "MZXQ====", 0 },
  { "base32", "foo", 0,             "MZXW6===", 0 },
  { "base32", "foob", 0,            "MZXW6YQ=", 0 },
  { "base32", "fooba", 0,           "MZXW6YTB", 0 },
  { "base32", "foobar", 0,          "MZXW6YTBOI======", 0 },

  { "base32", "12345", 0,           "GEZDGNBV", 0 },

  { "ascii85", "hello", 0,           "BOu!rDZ", 0 },
  { "ascii85", "sample test", 0,     "F(&p)Ch4`2ATMq", 0 },
  { "ascii85", "Hello, world.", 0,   "87cURD_*#TDfTZ)/c", 0 },
  { "ascii85", "The encoding is only 25% larger than the original, arbitrary bytes.", 0,
	"<+ohcASu!rA8,po+DGm>Df0B:+>Pf#+DbJ,B4Z*+FD,*#+EV:.+E):7B5)6lCb-];Eaia)Eaa!6+C]V<ATKI", 0 },

  { "mime64",  "f", 0,               "Zg==", 0 },
  { "mime64",  "fo", 0,              "Zm8=", 0 },
  { "mime64",  "foo", 0,             "Zm9v", 0 },
  { "mime64",  "foob", 0,            "Zm9vYg==", 0 },
  { "mime64",  "fooba", 0,           "Zm9vYmE=", 0 },
  { "mime64",  "foobar", 0,          "Zm9vYmFy", 0 },
  { "mime64",  "Hello, world.", 0,   "SGVsbG8sIHdvcmxkLg==", 0 },
  { "mime64",  "\x01\x02\x03\x04\x05", 5, "AQIDBAU=", 0 },
  { "mime64",  "a",    1,                "YQ==",     4 },
  { "mime64",  "ab",   2,                "YWI=",     4 },
  { "mime64",  "abc",  3,                "YWJj",     4 },
  { "mime64",  "abcd", 4,                "YWJjZA==", 8 },
  { "mime64",  "abcdefghijklmnop", 0,    "YWJjZGVmZ2hpamtsbW5vcA==", 0 },
  { "mime64",  "\x00\x45\xf2", 3,        "AEXy", 0 },
  { "mime64",
	"Man is distinguished, not only by his reason, but by this \
singular passion from other animals, which is a lust of the mind, that by a \
perseverance of delight in the continued and indefatigable generation of \
knowledge, exceeds the short vehemence of any carnal pleasure.", 0,
"TWFuIGlzIGRpc3Rpbmd1aXNoZWQsIG5vdCBvbmx5IGJ5IGhpcyByZWFzb24sIGJ1dCBieSB0aGlz\
IHNpbmd1bGFyIHBhc3Npb24gZnJvbSBvdGhlciBhbmltYWxzLCB3aGljaCBpcyBhIGx1c3Qgb2Yg\
dGhlIG1pbmQsIHRoYXQgYnkgYSBwZXJzZXZlcmFuY2Ugb2YgZGVsaWdodCBpbiB0aGUgY29udGlu\
dWVkIGFuZCBpbmRlZmF0aWdhYmxlIGdlbmVyYXRpb24gb2Yga25vd2xlZGdlLCBleGNlZWRzIHRo\
ZSBzaG9ydCB2ZWhlbWVuY2Ugb2YgYW55IGNhcm5hbCBwbGVhc3VyZS4=", 0 },

  { 0, NULL, 0,               NULL, 0 }
};

/*
 * These tests are by no means comprehensive ...
 * XXX maybe add encode/decode of pseudo-random generated material.
 */
static MAYBE_UNUSED int
test_strencdec(FILE *fp_out)
{
  int n, test_both, test_dec, test_enc;
  Str_test_encdec *prev_test, *test;
  Ds *ds;
  FILE *fp;

  if (fp_out == NULL)
	fp = stderr;
  else
	fp = fp_out;

  fprintf(fp, "Testing string encoding/decoding functions:\n");

  test_enc = 1;
  test_dec = 1;
  prev_test = NULL;
  for (n = 0, test = str_test_encdec; test->test_str != NULL; test++, n++) {
	unsigned int enc_len, test_str_len, val_len;
	char *enc_val;
	unsigned char *dec_val;

	if ((test_str_len = test->test_str_len) == 0)
	  test_str_len = strlen(test->test_str);
	if (test_str_len == 2342430)
	  continue;

	if ((enc_len = test->enc_str_len) == 0)
	  enc_len = strlen(test->enc_str);

	if (prev_test == NULL || !streq(prev_test->name, test->name)) {
	  if (prev_test != NULL)
		fprintf(fp, "\n");
	  fprintf(fp, "%s: ", test->name);
	}

	if (streq(test->name, "base32")) {
	  char *not_padded;

	  if (test_enc) {
		/* Test encoding. */
		val_len = strba32((unsigned char *) test->test_str, test_str_len,
						  0, &enc_val);
		if (val_len != enc_len
			|| !memcaseeq(enc_val, test->enc_str, enc_len)) {
		  fprintf(fp, "strba32 encoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }

	  if (test_dec) {
		/* Test decoding. */
		stra32b(test->enc_str, &dec_val, &val_len);
		if (val_len != test_str_len
			|| memcmp(dec_val, test->test_str, test_str_len)) {
		  fprintf(fp, "stra32b decoding failed, case %d:\n",  n);
		  return(-1);
		}

		/* Test decoding when padding may be omitted/implied. */
		not_padded = strchop(test->enc_str, "=");
		stra32b_x(not_padded, &dec_val, &val_len, 1);
		if (val_len != test_str_len
			|| memcmp(dec_val, test->test_str, test_str_len)) {
		  fprintf(fp, "stra32b_x decoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }
	  fputc((int) '.', fp);
	}
	else if (streq(test->name, "dacs64")
			 || streq(test->name, "base64")
			 || streq(test->name, "base64url")) {
	  Strba64_alg enc_alg;
	  Stra64b_alg dec_alg;

	  if (streq(test->name, "dacs64")) {
		enc_alg = STRBA64_DACS64;
		dec_alg = STRA64B_DACS64;
	  }
	  else if (streq(test->name, "base64")) {
		enc_alg = STRBA64_B64;
		dec_alg = STRA64B_B64;
	  }
	  else {
		enc_alg = STRBA64_B64URL;
		dec_alg = STRA64B_B64URL;
	  }

	  if (test_enc) {
		/* Test encoding. */
		ds = ds_strba64_x(NULL, enc_alg,
						  (unsigned char *) test->test_str, test_str_len);
		if (ds == NULL) {
		  fprintf(fp, "stra64b encoding failed, case %d:\n",  n);
		  return(-1);
		}
		enc_val = ds_buf(ds);
		val_len = ds_len(ds) - 1;
		if (val_len != enc_len
			|| !strneq(enc_val, test->enc_str, enc_len)) {
		  fprintf(fp, "strba64 encoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }

	  if (test_dec) {
		/* Test decoding. */
		if (test_enc) 
		  ds = ds_stra64b_x(NULL, dec_alg, enc_val, 0);
		else
		  ds = ds_stra64b_x(NULL, dec_alg, test->enc_str, 0);

		if (ds == NULL) {
		  fprintf(fp, "stra64b decoding failed, case %d:\n",  n);
		  return(-1);
		}
		dec_val = (unsigned char *) ds_buf(ds);
		val_len = ds_len(ds);
		if (val_len != test_str_len
			|| memcmp(dec_val, test->test_str, test_str_len)) {
		  fprintf(fp, "stra64b decoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }
	  fputc((int) '.', fp);
	}
	else if (streq(test->name, "ascii85")) {
	  if (test_enc) {
		/* Test encoding. */
		val_len = strba85((unsigned char *) test->test_str, test_str_len,
						  &enc_val);
		if (val_len != enc_len
			|| !strneq(enc_val, test->enc_str, enc_len)) {
		  fprintf(fp, "strba85 encoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }

	  if (test_dec) {
		/* Test decoding. */
		stra85b(test->enc_str, &dec_val, &val_len);
		if (val_len != test_str_len
			|| memcmp(dec_val, test->test_str, test_str_len)) {
		  fprintf(fp, "stra85b decoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }
	  fputc((int) '.', fp);
	}
	else if (streq(test->name, "mime64") || streq(test->name, "mime")) {
	  if (test_enc) {
		/* Test encoding. */
		val_len = mime_encode_base64((unsigned char *) test->test_str,
									 test_str_len, &enc_val);
		if (val_len != enc_len
			|| !strneq(enc_val, test->enc_str, enc_len)) {
		  fprintf(fp, "mime encoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }

	  if (test_dec) {
		/* Test decoding. */
		val_len = mime_decode_base64(test->enc_str, &dec_val);
		if (val_len != test_str_len
			|| memcmp(dec_val, test->test_str, test_str_len)) {
		  fprintf(fp, "mime decoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }
	  fputc((int) '.', fp);
	}
	else if (streq(test->name, "hex")) {
	  char *enc;
	  unsigned char *dec;

	  if (test_enc) {
		/* Test encoding. */
		enc = strbtohex((unsigned char *) test->test_str,
						test_str_len, STRBTOHEX_DEFAULT);
		/* The hex digits a-f in the encoded output are case insensitive. */
		if (enc == NULL || strlen(enc) != enc_len
			|| !memcaseeq(enc, test->enc_str, enc_len)) {
		  fprintf(fp, "hex encoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }

	  if (test_dec) {
		/* Test decoding. */
		dec = stra16b(test->enc_str, &dec_val, &val_len);
		if (dec == NULL || val_len != test_str_len
			|| memcmp(dec_val, test->test_str, test_str_len)) {
		  fprintf(fp, "hex decoding failed, case %d:\n",  n);
		  return(-1);
		}
	  }
	  fputc((int) '.', fp);
	}
	else {
	  fprintf(fp, "Whoa! Unrecognized test case in test_strencdec()\n");
	  return(-1);
	}
	prev_test = test;
  }

  fprintf(fp, "\n%d tests ok\n", n);
  return(0);
}

/*
 * Return 1 if PTR points to LEN zero bytes, 0 otherwise.
 */
int
memzeq(const void *ptr, size_t len)
{
  uint8_t *endp, *p;

  p = (uint8_t *) ptr;
  endp = p + len;
  while (p < endp) {
	if (*p++ != '\0')
	  return(0);
  }

  return(1);
}

/*
 * Return 1 if the first LEN bytes of PP are the same as QQ's, 0 otherwise.
 */
int
memeq(void *pp, void *qq, size_t len)
{
  int i;
  uint8_t *p, *q;

  p = pp;
  q = qq;
  for (i = 0; i < len; i++) {
	if (*p++ != *q++)
	  return(0);
  }

  return(1);
}

/*
 * Compare B1 to B2, each assumed to be LEN bytes long.
 * Return 1 if the two strings are equal, case insensitively, 0 otherwise.
 */
int
memcaseeq(const void *b1, const void *b2, size_t len)
{
  unsigned char *a, *b;
  size_t i;

  a = (unsigned char *) b1;
  b = (unsigned char *) b2;
  for (i = 0; i < len; i++) {
	if (*a != *b) {
	  if (tolower((int) *a) != tolower((int) *b))
		return(0);
	}
	a++;
	b++;
  }

  return(1);
}

/*
 * Make sure the function isn't optimized away because the string is not
 * read after it is written.
 * See, e.g.:
 *    http://www.mail-archive.com/cryptography%40metzdowd.com/msg08369.html
 */
void
memzap(void *s, size_t len)
{
  volatile unsigned char *ptr = (volatile unsigned char *) s;
  size_t i;

  for (i = 0; i < len; i++)
	*ptr++ = 0;
}

/*
 * Also see memset_s(3) from C11, where available.
 */
void
memzapb(void *s, int val, size_t len)
{
  volatile unsigned char *ptr = (volatile unsigned char *) s;
  size_t i;

  for (i = 0; i < len; i++)
	*ptr++ = val;
}

void *
memdupn(void *s, size_t len)
{
  void *ptr;

  ptr = malloc(len);
  memcpy(ptr, s, len);

  return(ptr);
}

/*
 * Destroy the value of environment variable NAME if it is found.
 * For greater certainty, we do this directly rather than by invoking
 * getenv(), putenv(), etc.
 * XXX presumably there aren't any cached copies of the value in
 * library functions
 */
void
envzap(char *name)
{
  int i, len;
  char *cp, *np, **p;
  extern char **environ;

  if (name == NULL || environ == NULL)
	return;

  for (np = name; *np != '\0' && *np != '='; np++)
	continue;
  len = np - name;

  for (p = environ; (cp = *p) != NULL; p++) {
	for (np = name, i = len; i != 0 && *cp != '\0'; i--) {
	  if (*cp++ != *np++)
		break;
	}
	if (i == 0 && *cp == '=') {
	  char **dp;

	  strzap(*p);
	  /*
	   * Do the equivalent of unsetenv(), which is not available everywhere.
	   * The memory is not freed, if it should be.
	   */
	  dp = p;
	  do {
		dp[0] = dp[1];
	  } while (*dp++ != NULL);
	  /* Continue in case NAME appears again. */
	}
  }
}

/*
 * For each occurrence of any of QUOTE_CHARS in STR, quote the character by
 * preceding it with a backslash character.
 * Return the new string, which might simply be a copy of the original.
 */
char *
strquote(char *str, char *quote_chars)
{
  char *p, *qchars;
  Ds ds;
  static char *default_quote_chars = "\"\'";

  if (quote_chars == NULL)
	qchars = default_quote_chars;
  else
	qchars = quote_chars;

  ds_init(&ds);
  for (p = str; *p != '\0'; p++) {
	if (index(qchars, (int) *p) != NULL)
	  ds_appendc(&ds, (int) '\\');
	ds_appendc(&ds, (int) *p);
  }

  ds_appendc(&ds, (int) '\0');
  return(ds_buf(&ds));
}

/*
 * Strip the outermost "quotes" surrounding STR.
 * STR is a string with matching "quote" characters, where
 * the character in START_Q at index N is a valid starting quote character
 * and the character in END_Q at index N is the matching ending quote
 * character.  Return the embedded string, or NULL.
 *
 * Example: strunquote("[foo]", "([{", ")]}") gives "foo"
 * Example: strunquote("'foo'", "'\"`", "'\"`") also gives "foo"
 */
char *
strextract(char *str, char *start_q, char *end_q)
{
  char e_ch, *p;
  size_t slen;

  if ((p = strchr(start_q, (int) str[0])) == NULL)
	return(NULL);
  if (strlen(start_q) != strlen(end_q))
	return(NULL);
  e_ch = end_q[p - start_q];

  slen = strlen(str);
  if (slen < 2 || str[slen - 1] != e_ch)
	return(NULL);

  return(strndup(str + 1, slen - 2));
}

/*
 * If the argument is surrounded by matching quote characters
 * (single, double, or backtick quotes), return a new string with those quotes
 * elided; otherwise, return a copy of the original string.
 * This is just a convenient interface to the more general strextract().
 */
char *
strdequote(char *p)
{
  char *embedded;

  if ((embedded = strextract(p, "'\"`", "'\"`")) == NULL)
	return(strdup(p));

  return(embedded);
}

/*
 * Given string P of length LEN, if P ends with the string SUFFIX,
 * return a pointer to the start of SUFFIX in P, NULL otherwise.
 * SUFFIX is a suffix of P if they are identical.
 */
char *
strsuffix(char *p, size_t len, const char *suffix)
{
  char *q;
  size_t suffix_len;

  suffix_len = strlen(suffix);
  if (suffix_len > len)
	return(0);
  q = p + len - suffix_len;
  return(streq(q, suffix) ? q : NULL);
}

/*
 * Like strsuffix(), except case insensitive.
 */
char *
strcasesuffix(const char *p, size_t len, const char *suffix)
{
  const char *q;
  size_t suffix_len;

  suffix_len = strlen(suffix);
  if (suffix_len > len)
	return(0);
  q = p + len - suffix_len;
  return(strcaseeq(q, suffix) ? (char *) q : NULL);
}

/*
 * Test if the initial substring of P is PREFIX (they are if they are
 * identical, too).
 * If it is, return a pointer to the character in P following the prefix,
 * otherwise return NULL.
 */
char *
strprefix(const char *p, const char *prefix)
{
  size_t len;

  len = strlen(prefix);
  if (strneq(p, prefix, len))
	return((char *) p + len);
  return(NULL);
}

/*
 * Like strprefix(), except case insensitive.
 */
char *
strcaseprefix(const char *p, const char *prefix)
{
  size_t len;

  len = strlen(prefix);
  if (strncaseeq(p, prefix, len))
	return((char *) p + len);
  return(NULL);
}

/*
 * Test if PREFIX is a prefix of WORD and PREFIX is at least MATCH_LEN
 * bytes long.  Comparisons are case insensitive.
 * Examples:
 *   strncaseprefix("pass", "password", 4) is 1
 *   strncaseprefix("PASSWord", "password", 8) is 1
 *   strncaseprefix("passz", "password", 4) is 0
 *   strncaseprefix("pass", "password", 5) is 0
 * Return 1 if so, 0 otherwise.
 */
int
strncaseprefix(char *prefix, char *word, size_t match_len)
{
  size_t prefix_len, word_len;

  prefix_len = strlen(prefix);
  word_len = strlen(word);
  if (prefix_len > word_len || prefix_len < match_len)
	return(0);

  if (strncaseeq(prefix, word, prefix_len))
	return(1);

  return(0);
}

/*
 * Count the number of times character C appears in string P.
 */
int
strchrcount(const char *p, int c)
{
  int n;
  const char *q;

  if (p == NULL || *p == '\0')
	return(0);

  q = p;
  n = 0;
  while ((q = strchr(q, c)) != NULL) {
	q++;
	n++;
  }

  return(n);
}

/*
 * Count the number of times any character in CHARSET appears in string P.
 * CHARSET must be comprised of unique characters (no duplicates).
 * Return -1 if CHARSET is invalid.
 */
int
strchrscount(const char *p, const char *charset)
{
  int n;
  const char *q;

  if (p == NULL || *p == '\0')
	return(0);

  if (charset == NULL || *charset == '\0')
	return(-1);

  n = 0;
  for (q = charset; *q != '\0'; q++) {
	if (strchrcount(charset, (int) *q) != 1)
	  return(-1);
	n += strchrcount(p, (int) *q);
  }

  return(n);
}

#ifndef HAVE_ISBLANK
/*
 * Test for a space or tab character.
 * Return zero if the character tests false and
 * non-zero if the character tests true.
 */
int
isblank(int ch)
{

  if (ch == ' ' || ch == '\t')
	return(1);

  return(0);
}
#endif

/*
 * Check if STR consists of alphabetics only.
 */
int
is_alpha_string(const char *str)
{
  const char *p;

  p = str;
  do {
	if (!isalpha((int) *p))
	  return(0);
	p++;
  } while (*p != '\0');

  return(1);
}

/*
 * Check if STR consists of one or more decimal digits.
 */
int
is_digit_string(const char *str)
{
  const char *p;

  p = str;
  do {
	if (!isdigit((int) *p))
	  return(0);
	p++;
  } while (*p != '\0');

  return(1);
}

int
is_signed_digit_string(const char *str)
{
  const char *p;

  p = str;
  if (*p == '+' || *p == '-')
	p++;

  while (isdigit((int) *p))
	p++;
  if (*p == '\0')
	return(1);
  return(0);
}

/*
 * If ENV_STRING looks like "<name>=<value>", then set ENV_NAME to point
 * to the <name> and ENV_VAL to point to the <value>.
 * Either of ENV_NAME and ENV_VAL may be NULL, which means the substring
 * is not to be returned.
 * <name> may not contain a '=' character.
 * If no '=' character is present, it is treated as if <name> is absent
 * and <value> is the entire string.
 * Return a pointer to the <name> part (which may be NULL); if ENV_NAME is
 * not NULL, it will be set to point to the same string.
 */
char *
env_parse(char *env_string, char **env_name, char **env_val)
{
  char *e, *p, *q;

  p = q = NULL;
  if (env_string != NULL && env_string[0] != '\0') {
	if ((e = strchr(env_string, '=')) == NULL) {
	  if (env_val != NULL)
		q = strdup(env_string);
	}
	else {
	  p = (char *) malloc(e - env_string + 1);
	  strncpy(p, env_string, e - env_string);
	  p[e - env_string] = '\0';
	  if (env_val != NULL)
		q = strdup(e + 1);
	}
  }

  if (env_name != NULL)
	*env_name = p;
  if (env_val != NULL)
	*env_val = q;

  return(p);
}

char *
ul_str(unsigned long val)
{
  char *buf;

  /*
   * 32 bits -> 10 digits max
   * 64 bits -> 20 digits max
   */
  buf = (char *) safe_malloc(32);
  sprintf(buf, "%lu", val);

  return(buf);
}

#ifndef HAVE_STRNCASECMP
int
strncasecmp(const char *s1, const char *s2, size_t len)
{
  int uc1, uc2;
  const char *p1, *p2;
  size_t c;

  p1 = (const char *) s1;
  p2 = (const char *) s2;
  c = len;

  while (*p1 != '\0' && *p2 != '\0' && c--) {
	uc1 = tolower((int) *p1++);
	uc2 = tolower((int) *p2++);
	if (uc1 != uc2)
	  return(((unsigned char) uc1 > (unsigned char) uc2) ? 1 : -1);
  }

  if (*p1 != '\0' && *p2 == '\0')
	return(1);
  if (*p1 == '\0' && *p2 != '\0')
	return(-1);

  return(0);
}
#endif

/*
 * Return 1 if CH1 and CH2 are equal, case-independently if NOCASE is non-zero,
 * 0 otherwise.
 */
static int
char_eq(int ch1, int ch2, int nocase)
{

  if (nocase)
	return(tolower(ch1) == tolower(ch2));

  return(ch1 == ch2);
}

static char *
strstr_internal(const char *str, const char *substr, int nocase)
{
  char ch, ch_first, *s, *ss;
  size_t len;
  int (*cmp)(const char *s1, const char *s2, size_t len);

  if (nocase)
	cmp = strncasecmp;
  else
	cmp = strncmp;

  s = (char *) str;
  if ((ch_first = *substr++) != '\0') {
	len = strlen(substr);
	ss = (char *) substr;
	do {
	  do {
		if ((ch = *s++) == '\0')
		  return(NULL);
	  } while (!char_eq(ch, ch_first, nocase));
	} while (cmp(s, ss, len) != 0);
	s--;
  }

  return(s);
}

/*
 * We use libraries that include a strstr() function; life is easier
 * if we can explicitly call this implementation.
 * But the native version might be optimized, so use it when possible.
 *
 * This implementation is poor if STR is long - there are much better
 * methods - but in our typical applications STR is short enough that we
 * won't care.
 */
char *
dacs_strstr(const char *str, const char *substr)
{

  return(strstr_internal(str, substr, 0));
}

/*
 * Like strstr(), but use case insensitive character comparison.
 *
 * We use libraries that include a strcasestr() function; life is easier
 * if we can explicitly call this implementation.
 * But the native version might be optimized, so use it when possible.
 *
 * This implementation is poor if STR is long - there are much better methods
 * (such as strqs()) - but in our typical applications STR is short enough
 * that we won't care.
 */
char *
dacs_strcasestr(const char *str, const char *substr)
{

  return(strstr_internal(str, substr, 1));
}

static char *
strrstr_internal(const char *str, const char *substr, int nocase)
{
  const char *found, *p;
  char *(*cmp)(const char *str, const char *substr);

  if (nocase)
	cmp = strcasestr;
  else
	cmp = strstr;

  found = NULL;

  p = str;
  do {
	if ((p = cmp(p, substr)) == NULL)
	  break;
	found = p++;
  } while (*p != '\0');

  return((char *) found);
}

/*
 * Look for the start of the *last* occurrence of SUBSTR in STR.
 * This may not be as efficient as searching the string backwards, but it
 * may leverage an optimized strstr() implementation and in any case we
 * probably don't care much about its relative efficiency when STR is not
 * very long.
 */
char *
strrstr(const char *str, const char *substr)
{

  return(strrstr_internal(str, substr, 0));
}

char *
strrcasestr(const char *str, const char *substr)
{

  return(strrstr_internal(str, substr, 1));
}

/******/

enum {
  STRQS_ALPHABET_SIZE = 128
};

/*
 * Construct the delta 1 shift table from PATTERN of length PATTERN_LEN
 * and return it.  If TD_ONE is non-NULL, assume it is a buffer of the correct
 * length to use, otherwise allocate such a buffer.
 * Compute the length of PATTERN if PATTERN_LEN is zero.
 */
unsigned int *
strqs_init(int *td_one, const char *pattern, size_t pattern_len, int nocase)
{
  int i;
  unsigned int *td;
  const char *p;
  size_t plen;

  if (td_one == NULL)
	td = (unsigned int *) malloc(STRQS_ALPHABET_SIZE * sizeof(unsigned int));
  else
	td = (unsigned int *) td_one;

  if (pattern_len == 0)
	plen = strlen(pattern);
  else
	plen = pattern_len;

  for (i = 0; i < STRQS_ALPHABET_SIZE; i++)
	td[i] = plen + 1;

  for (p = pattern; *p != '\0'; p++) {
	if (nocase && isalpha((int) *p)) {
	  int cl, cu;

	  cl = tolower((int) *p);
	  cu = toupper((int) *p);
	  td[cl] = td[cu] = plen - (p - pattern);
	}
	else {
	  int c;

	  c = (int) (*p & 0377);
	  td[c] = plen - (p - pattern);
	}
  }

  return(td);
}

/*
 * Extended interface to strqs() where all of the arguments have been
 * provided.  This can avoid recomputing them if the caller needs to look
 * for the same pattern again.
 * TD_ONE is the delta 1 shift table produced by strqs_init(), ENDP points to
 * the character following the end of the text (STR), PLEN is the length of the
 * pattern (SUBSTR), and NOCASE is non-zero if case-insensitive comparisons
 * should be used (in which case TD_ONE must have been produced with the
 * NOCASE flag to strqs_init()).
 */
char *
strqsx(const char *str, const char *endp, const char *substr,
	   unsigned int *td_one, size_t plen, int nocase)
{
  const char *p, *t, *tx;

  tx = str;
  while ((tx + plen) < endp) {
	for (p = substr, t = tx; *p != '\0'; p++, t++) {
	  if (nocase) {
		if (tolower((int) *p) != tolower((int) *t))
		  break;
	  }
	  else if (*p != *t)
		break;
	}
	if (*p == '\0')
	  return((char *) tx);
	tx += td_one[*(tx + plen) & 0377];
  }

  return(NULL);
}

static char *
strqs_internal(const char *str, const char *substr, int nocase,
			   unsigned int **td_one_ret)
{
  unsigned int *td_one;
  char *val;
  size_t str_len, substr_len;

  str_len = strlen(str);
  substr_len = strlen(substr);

  td_one = strqs_init(NULL, substr, substr_len, nocase);

  val = strqsx(str, str + str_len, substr, td_one, substr_len, nocase);

  if (td_one_ret != NULL)
	*td_one_ret = td_one;
  else
	free(td_one);

  return(val);
}

/*
 * Find the start of the first occurrence of SUBSTR in STR, like strstr().
 * This is an implementation of "Quick Search":
 * "A Very Fast Substring Search Algorithm", Daniel M. Sunday, CACM, Vol. 33,
 * No. 8 (August 1990), pp. 132-142.
 * The paper describes some variations of this algorithm that may be a little
 * better.
 * 
 */
char *
strqs(const char *str, const char *substr)
{
  char *val;

  val = strqs_internal(str, substr, 0, NULL);

  return(val);
}

/*
 * A case-insensitive variant of strqs().
 */
char *
strcaseqs(const char *str, const char *substr)
{
  char *val;

  val = strqs_internal(str, substr, 1, NULL);

  return(val);
}

/******/

typedef struct Str_test_parse {
  char *str;
  char *name;
  char *value;
  char *endstr;
  int flags;
  int sep_char;
  int end_char;
  int rc;
} Str_test_parse;

static Str_test_parse str_test_parse_strings[] = {
  { "a: b",        "a",  " b",  "",
	KWV_PARSE_DEFAULT,          ':', 0,   0 },
  { "a: \"bcd\"",         "a",  "bcd",  "",
	KWV_PARSE_DEFAULT,          ':', 0,   0 },
  { "a=\"\"",      "a",  "",    "",
	KWV_PARSE_NEED_VALUE,       0, 0,   0 },
  { "a='b'",       "a",  "'b'",   "",
	KWV_PARSE_KEEP_QUOTES,      0, 0,   0 },
  { "a=\"b\"; c=d","a",  "b",     "; c=d",
	KWV_PARSE_DEFAULT,          0, ';', 0 },
  { "a=b; c=d",    "a",  "b",     "; c=d",
	KWV_PARSE_DEFAULT,          0, ';', 0 },
  { "a=\"b=\"c\"", "a",  "b=\"c", "",
	KWV_PARSE_INTERNAL_QUOTES,  0, 0,   0 },
  { "ab=cd\n",     "ab", "cd",    "\n",
    KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "a = \"bb\" ", "a",  "bb",    " ",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "a = b",       "a",  " b",    "",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "=a",          "",   "",      "",
	KWV_PARSE_DEFAULT,          0, 0,  -1 },
  { "a=",          "a",  "",      "",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "a=  ",        "a",  "  ",    "",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "a=b",         "a",  "b",     "",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "a='b'",       "a",  "b",     "",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "a=\"b\"",     "a",  "b",     "",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "  a=\"b\"",   "a",  "b",     "",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "  a=\"b\"  ", "a",  "b",     "  ",
    KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "a = \"b\"",   "a",  "b",     "",
	KWV_PARSE_DEFAULT,          0, 0,   0 },
  { "a = \"b",     "a",  "b",     "",
	KWV_PARSE_DEFAULT,          0, 0,  -1 },
  { NULL,          NULL, NULL,    NULL,
    KWV_PARSE_DEFAULT,          0, 0,   0 }
};

static MAYBE_UNUSED int
test_parse(FILE *fp)
{
  int i, rc;
  char *endptr, *name, *value;
  Kwv_parse_conf conf;
  Str_test_parse *stp;

  fprintf(fp, "Testing string parsing functions...\n");
  for (i = 0, stp = str_test_parse_strings; stp->str != NULL; stp++, i++) {
	name = NULL;
	value = NULL;
	endptr = NULL;
	conf.sep_char = stp->sep_char;
	conf.flags = stp->flags;
	conf.end_char = stp->end_char;
	conf.errmsg = NULL;
	rc = kwv_parse_qstr(stp->str, &name, &value, &endptr, &conf);
	if (rc == stp->rc && rc == -1) {
	  /* This is supposed to be an error case, so it's ok. */
	  continue;
	}

	if (rc != stp->rc || name == NULL || value == NULL || endptr == NULL
		|| !streq(name, stp->name) || !streq(value, stp->value)
		|| !streq(endptr, stp->endstr)) {
	  fprintf(fp, "test_parse case %d failed:\n", i);
	  fprintf(fp, "  string:%s\n", stp->str);
	  fprintf(fp, "  rc:%d (expected:%d)\n", rc, stp->rc);
	  fprintf(fp, "  name:%s (expected:%s)\n",
			  name == NULL ? "NULL" : name,
			  stp->name == NULL ? "NULL" : stp->name);
	  fprintf(fp, "  value:%s (expected:%s)\n",
			  value == NULL ? "NULL" : value,
			  stp->value == NULL ? "NULL" : stp->value);
	  fprintf(fp, "  endptr:%s (expected:%s)\n",
			  endptr == NULL ? "NULL" : endptr,
			  stp->endstr == NULL ? "NULL" : stp->endstr);
	  return(-1);
	}
  }

  return(0);
}

/*
 * Destructively carve STR into a vector of substrings.  Each element of STR
 * is separated from the next by DELIMIT, which is a string.
 * At most LIMIT substrings will be returned (a LIMIT of zero implies no
 * limit).
 * If DELIMIT is "", then STR is broken up into individual characters.
 *
 * If INC_DELIMIT is 1 it means that the delimiter should be included
 * as the first character of each vector element; if it is -1 it means the
 * delimiter should be included as the last character of each vector element;
 * and if 0 or any undefined value it means that the delimiter should not
 * be included in vector elements.
 *
 * XXX there is no way to escape the delimiter string or to quote it;
 * for example, strsplit("a,'b,c',d", ",", 0) will return "a", "'b", "c'", "d",
 * when the caller would probably like to specify the single quote as being
 * special so that "a", "b,c", and "d" are parsed out.
 *
 * Example cases:
 *                                  INC_DELIMIT
 *                     0                 1                   -1
 *
 *  "/"         -> ["",""]             ["/"]               ["/"]
 *  "//"        -> ["","",""]          ["/","/"]           ["/","/"]
 *  "/xxx"      -> ["","xxx"]          ["/xxx"]            ["/","xxx"]
 *  "/xxx/"     -> ["","xxx", ""]      ["/xxx","/"]        ["/","xxx/"]
 *  "xxx"       -> ["xxx"]             ["xxx"]             ["xxx"]
 *  "/xxx/yyy"  -> ["","xxx","yyy"]    ["/xxx","/yyy"]     ["/","xxx/","yyy"]
 *  "xxx/yyy"   -> ["xxx","yyy"]       ["xxx","/yyy"]      ["xxx/","yyy"]
 *  "xxx/yyy/"  -> ["xxx","yyy",""]    ["xxx","/yyy","/"]  ["xxx/","yyy/"]
 *  "/xxx/yyy/" -> ["","xxx","yyy",""] ["/xxx","/yyy","/"] ["/","xxx/","yyy/"]
 *
 * Two consecutive DELIMIT strings, or a trailing or initial DELIMIT string,
 * may create an empty element, depending on INC_DELIMIT.
 *
 * Example: strsplit("a=1\nb=2\nc=3\n", "\n", 0)
 * will break the string into a set of three pointers to null-terminated
 * strings (without newlines):
 * [0] -> "a=1"
 * [1] -> "b=2"
 * [2] -> "c=3"
 */
static Dsvec *
strsplit_internal(Dsvec *odsv, char *str, const char *delimit, int limit,
				  int inc_delimit)
{
  int ndel;
  char *e, *dstart, *s, *scanp;
  size_t dlen;
  Dsvec *dsv;

  if (*str == '\0')
	return(NULL);

  dlen = strlen(delimit);
  if (odsv != NULL)
	dsv  = odsv;
  else
	dsv = dsvec_init(NULL, sizeof(char **));

  s = scanp = str;
  ndel = 0;
  while (*s != '\0') {
	if ((limit != 0 && dsvec_len(dsv) == (limit - 1))
		|| (dlen > 0 && (dstart = strstr(scanp, delimit)) == NULL)) {
	  dsvec_add_ptr(dsv, s);
	  break;
	}

	ndel++;
	if (dlen > 0) {
	  /*
	   * A non-empty delimiter has been found, starting at S and ending
	   * at DEND.
	   */
	  if (inc_delimit < 0) {
		/* Include delimiter at end of S. */
		e = dstart + dlen;
		dsvec_add_ptr(dsv, strndup(s, e - s));
		s = scanp = e;
	  }
	  else if (inc_delimit > 0) {
		if (dstart != s) {
		  /*
		   * This element does not begin with a delimiter.  Copy up to
		   * but not including the delimiter.
		   */
		  e = dstart;
		  dsvec_add_ptr(dsv, strndup(s, e - s));
		  s = scanp = dstart;
		}
		else {
		  /*
		   * This element begins with a delimiter.  Just move past it and
		   * continue.
		   */
		  scanp = dstart + dlen;
		}
	  }
	  else if (inc_delimit == 0) {
		dsvec_add_ptr(dsv, s);
		*dstart = '\0';
		s = scanp = dstart + dlen;
	  }
	}
	else {
	  dsvec_add_ptr(dsv, strndup(s, 1));
	  s++;
	}
  }

  /* If STR ends with a delimiter, there is an "empty" element on the end. */
  if (dsvec_len(dsv) == ndel && inc_delimit == 0)
	dsvec_add_ptr(dsv, "");

  return(dsv);
}

Dsvec *
strsplit(const char *str, const char *delimit, int limit)
{

  return(strsplit_internal(NULL, strdup(str), delimit, limit, 0));
}

Dsvec *
strsplitx(Dsvec *odsv, const char *str, const char *delimit, int limit)
{

  return(strsplit_internal(odsv, strdup(str), delimit, limit, 0));
}

Dsvec *
strsplitd(const char *str, const char *delimit, int limit, int inc_delimit)
{

  return(strsplit_internal(NULL, strdup(str), delimit, limit, inc_delimit));
}

Dsvec *
strsplit_nocopy(char *str, const char *delimit, int limit)
{

  return(strsplit_internal(NULL, str, delimit, limit, 0));
}

/*
 * DSV is a vector of lines of text.
 * Return a vector of "words" obtained by splitting each line of DSV into
 * substrings that are delimited by string "DELIMIT".
 *
 * Previously known as strsplit_x().
 */
Dsvec *
strsplit_vec(Dsvec *dsv, const char *delimit, int limit)
{
  int i;
  char *line;
  Dsvec *words;

  words = dsvec_init(NULL, sizeof(char *));
  for (i = 0; i < dsvec_len(dsv); i++) {
	line = (char *) dsvec_ptr_index(dsv, i);
	strsplit_internal(words, line, delimit, limit, 0);
  }

  return(words);
}

/*
 * Like strsplit(), except regular expression matching is used.
 * If an error occurs and ERRMSG is non-NULL, it is set to a descriptive
 * error message.
 */
Dsvec *
strsplit_re(const char *str, const char *regex, int limit, int inc_delimit,
			char **errmsg)
{
  int n, st;
  char errbuf[100], *s;
  regex_t preg;
  regmatch_t pmatch[1];
  Dsvec *dsv;

  if (*regex == '\0') {
	if (errmsg != NULL)
	  *errmsg = ds_xprintf("Invalid regular expression");
	return(NULL);
  }

  if ((st = regcomp(&preg, regex, REG_EXTENDED)) != 0) {
	if (errmsg != NULL) {
	  regerror(st, &preg, errbuf, sizeof(errbuf));
	  *errmsg = strdup(errbuf);
	}
	return(NULL);
  }

  s = strdup(str);
  dsv = dsvec_init(NULL, sizeof(char **));

  n = 1;
  while (*s != '\0') {
	if (limit != 0 && n == limit) {
	  dsvec_add_ptr(dsv, s);
	  break;
	}

	/* Advance to the next delimiter, if present. */
	st = regexec(&preg, s, 1, pmatch, 0);
	if (st != 0) {
	  if (st != REG_NOMATCH) {
		if (errmsg != NULL) {
		  regerror(st, &preg, errbuf, sizeof(errbuf));
		  *errmsg = strdup(errbuf);
		}
		return(NULL);
	  }
	  /* No more elements, this is the last one. */
	  dsvec_add_ptr(dsv, s);
	  break;
	}
	else {
	  if (inc_delimit == 1)
		dsvec_add_ptr(dsv, strndup(s, pmatch[0].rm_eo));
	  else {
		s[pmatch[0].rm_so] = '\0';
		dsvec_add_ptr(dsv, s);
	  }
	  s += pmatch[0].rm_eo;
	}

	n++;
  }

  regfree(&preg);

  return(dsv);
}

/*
 * Break pathname PATH into its slash-delimited components, any of which may
 * be empty.  Slashes cannot be escaped.
 * The slashes are *not* kept.  A trailing slash does not create an element.
 * Set IS_ABS to non-zero if PATH is absolute, or zero if it is relative.
 * Set TRAILING to non-zero if PATH is not "/" and ends with a slash.
 * Try strjoin() to reconstruct a path from components.
 *
 * Examples:
 *  PATH      Return                    IS_ABS TRAILING
 *  "/"        NULL                        1       0
 *  "//"       one element (null string)   1       1
 *  "a"        one element ("a")           0       0
 *  "a/"       one element ("a")           0       1
 *  "/a"       one element ("a")           1       0
 *  "/a/"      one element ("a")           1       1
 *  "/a/b"     two elements ("a", "b")     1       0
 *  "///"      two elements (null strings) 1       1
 */
Dsvec *
strpathparse(const char *path, int *is_abs, int *trailing)
{
  int is_root;
  const char *p, *s;
  Dsvec *dsv;

  s = path;
  is_root = 0;
  if (trailing != NULL)
	*trailing = 0;

  if (*s == '/') {
	if (is_abs != NULL)
	  *is_abs = 1;
	s++;
	if (*s == '\0')
	  is_root = 1;
  }
  else {
	if (is_abs != NULL)
	  *is_abs = 0;
  }

  dsv = NULL;
  p = s;
  while (*p != '\0') {
	if (*p == '/') {
	  if (dsv == NULL)
		dsv = dsvec_init(NULL, sizeof(char *));
	  dsvec_add_ptr(dsv, strndup(s, p - s));
	  s = p + 1;
	  if (*(p + 1) == '\0' && !is_root && trailing != NULL)
		*trailing = 1;
	}
	p++;
  }

  if (*s != '\0') {
	if (dsv == NULL)
	  dsv = dsvec_init(NULL, sizeof(char *));
	dsvec_add_ptr(dsv, strndup(s, p - s + 1));
  }

  return(dsv);
}

/*
 * Convert the elementary types described by DESC into a string,
 * with elements separated by DELIMIT.
 * This is largely an experimental kludge that leverages existing functions
 * to externalize and internalize data structures but it will do until there's
 * a need for something better.
 *
 * XXX consider extending this to handle various string encodings; a catch is
 * that a resulting encoding must either not contain DELIMIT or must
 * automagically escape/unescape every occurrence.
 */
char *
strpack(Strpack *desc, const char *delimit)
{
  int n;
  Ds *ds;
  Strpack *dp;

  dp = desc;
  ds = ds_init(NULL);
  for (n = 0, dp = desc; dp->ptr != NULL; dp++, n++) {
	char *el;
	const char *sep;

	sep = (n == 0) ? "" : delimit;
	switch (dp->type) {
	case STRNUM_STR:
	  el = *(char **) dp->ptr;
	  ds_asprintf(ds, "%s%s", sep, non_null(el));
	  break;
	case STRNUM_I:
	  ds_asprintf(ds, "%s%d", sep, *(int *) dp->ptr);
	  break;
	case STRNUM_L:
	  ds_asprintf(ds, "%s%ld", sep, *(long *) dp->ptr);
	  break;
	case STRNUM_UL:
	  ds_asprintf(ds, "%s%lu", sep, *(unsigned long *) dp->ptr);
	  break;
	case STRNUM_TIME_T:
	  ds_asprintf(ds, "%s%lu", sep, *(time_t *) dp->ptr);
	  break;
	case STRNUM_MODE_T:
	  ds_asprintf(ds, "%s%lu", sep, *(mode_t *) dp->ptr);
	  break;
	case STRNUM_PID_T:
	  ds_asprintf(ds, "%s%lu", sep, *(pid_t *) dp->ptr);
	  break;
	case STRNUM_IN_PORT_T:
	  ds_asprintf(ds, "%s%lu", sep, *(in_port_t *) dp->ptr);
	  break;
	default:
	  return(NULL);
	}
  }

  return(ds_buf(ds));
}

/*
 * Split STR into elements that are separated by DELIMIT (up to LIMIT of them)
 * and convert and assign each element according to DESC, a vector of
 * descriptions.
 *
 * The caller is responsible for ensuring that any string elements are
 * properly encoded so that they do not include DELIMIT.
 *
 * Return the number of unpacked elements, or -1 if an error occurs.
 */
int
strunpack(const char *str, const char *delimit, int limit, Strpack *desc)
{
  int i;
  Dsvec *dsv;
  Strpack *dp;

  if ((dsv = strsplit(str, delimit, limit)) == NULL)
	return(-1);

  dp = desc;
  for (i = 0; i < dsvec_len(dsv); i++) {
	char *el;

	if (dp->ptr == NULL)
	  return(-1);

	el = (char *) dsvec_ptr_index(dsv, i);
	if (strnum(el, dp->type, dp->ptr) == -1)
	  return(-1);

	dp++;
  }

  if (dp->ptr != NULL)
	return(-1);

  return(i);
}

static MAYBE_UNUSED void
strpack_unpack_test(void)
{
  int st;
  char *desc;
  struct Strpack_test {
	int iii;
	char *str;
	time_t ttt;
	unsigned long ulul;
  } unpack_test;
  Strpack test_desc[] = {
	{ STRNUM_I,      &unpack_test.iii },
	{ STRNUM_STR,    &unpack_test.str },
	{ STRNUM_TIME_T, &unpack_test.ttt },
	{ STRNUM_UL,     &unpack_test.ulul },
	{ STRNUM_ERR,    NULL }
  };

  st = strunpack("17,hello world,1234567,1010101010", ",", 0, test_desc);
  fprintf(stderr, "%d\n", st);
  desc = strpack(test_desc, ",");
  fprintf(stderr, "%s\n", desc);
}

char *
strjoin_subset(Dsvec *dsv, unsigned int start, int len, char *sepstr)
{
  unsigned int ui;
  char *s, *sep;
  Ds ds;

  if (sepstr == NULL)
	sep = "";
  else
	sep = sepstr;

  if (len == -1)
	len = dsvec_len(dsv);

  ds_init(&ds);
  for (ui = start; ui < dsvec_len(dsv) && ui < len; ui++) {
	s = (char *) dsvec_ptr_index(dsv, ui);
	if (s != NULL)
	  ds_asprintf(&ds, "%s%s", (ui == start) ? "" : sep, s);
  }

  return(ds_buf(&ds));
}

/*
 * Join together the elements of DSV, each separated from the next
 * by SEPSTR (unless it is NULL).  NULL elements are ignored, although
 * empty (nul) strings are not.
 * Return the new string.
 */
char *
strjoin(Dsvec *dsv, char *sepstr)
{

  return(strjoin_subset(dsv, 0, -1, sepstr));
}

static MAYBE_UNUSED Strsubst test_strsubst[] = {
  { "<",  "&lt;"  },
  { "&",  "&amp;" },
  { "hi", "hello" },
  { NULL, NULL    }
};

/*
 * Return a new string formed by replacing (exactly) matching substrings with
 * corresponding replacement text (without rescanning the replacement text).
 */
Ds *
strsubst(char *str, Strsubst *tab)
{
  char *e, *p;
  Ds *ds;
  Strsubst *t;

  ds = ds_init(NULL);
  p = str;
  while (*p != '\0') {
	for (t = tab; t->src_text != NULL; t++) {
	  if ((e = strprefix(p, t->src_text)) != NULL)
		break;
	}
	if (t->src_text != NULL) {
	  ds_append(ds, t->subst_text);
	  p = e;
	}
	else {
	  ds_appendc(ds, (int) *p);
	  p++;
	}
  }
  ds_appendc(ds, (int) '\0');

  return(ds);
}

/*
 * Scan STR, decoding/unescaping C-style character constants, and return
 * a new string or NULL if invalid encoding is found.
 * Octal and hexadecimal byte encodings are recognized (but not hex byte
 * sequences like the three byte long value \xabcdef).
 * Note that the decoded string might be binary (e.g., "\001\002\003").
 */
Ds *
struncescape(char *str)
{
  int cc, i;
  char *p;
  Ds *ds;

  ds = ds_init(NULL);
  p = str;

  while (*p != '\0') {
	if (*p == '\\') {
	  /* Handle C-style character constants. */
	  p++;
	  if (*p == '\0')	{ /* Ends with a backslash? */
		return(NULL);
	  }

	  if (is_octdigit((int) *p)) {
		/* Handle e.g., \0123 */
		cc = 0;
		for (i = 0; i < 3; i++) {
		  if (!is_octdigit((int) *p))
			break;
		  cc = cc * 8 + octdigit_val((int) *p);
		  p++;
		}
		ds_appendc(ds, cc);
	  }
	  else if (*p == 'x') {
		/* Handle e.g., \xa3 */
		p++;
		if (!is_hexdigit((int) *p))
		  return(NULL);
		cc = hexdigit_val((int) *p);
		p++;
		if (is_hexdigit(*p)) {
		  cc = cc * 16 + hexdigit_val((int) *p);
		  p++;
		}
		ds_appendc(ds, cc);
	  }
	  else {
		/* Harbison & Steele 2.7.6 */
		switch ((int) *p) {
		case 'a':
		  cc = '\a'; break;
		case 'b':
		  cc = '\b'; break;
		case 'f':
		  cc = '\f'; break;
		case 'n':
		  cc = '\n'; break;
		case 'r':
		  cc = '\r'; break;
		case 't':
		  cc = '\t'; break;
		case 'v':
		  cc = '\v'; break;
		default:
		  cc = *p; break;
		}

		ds_appendc(ds, cc);
		p++;
	  }
	}
	else
	  ds_appendc(ds, (int) *p++);
  }

  ds_appendc(ds, (int) '\0');
  return(ds);
}

/*
 * Convert the null-terminated string PTR (if SLEN == 0) or the
 * byte sequence PTR of length SLEN (SLEN != 0) into a C-style escaped
 * string.
 */
char *
strcescape(char *ptr, size_t slen)
{
  int ch, esc, i;
  char *s;
  size_t len;
  Ds ds;

  if (slen == 0)
	len = strlen(ptr);
  else
	len = slen;

  s = ptr;
  ds_init(&ds);
  for (i = 0; i < len; i++) {
	ch = *s++;
	switch (ch) {
	case '\a':
	  esc = 'a'; break;
	case '\b':
	  esc = 'b'; break;
	case '\f':
	  esc = 'f'; break;
	case '\n':
	  esc = 'n'; break;
	case '\r':
	  esc = 'r'; break;
	case '\t':
	  esc = 't'; break;
	case '\v':
	  esc = 'v'; break;

	default:
	  esc = 0;
	  if (ch == 0)
		ds_append(&ds, "\\0");		/* A special case */
	  else if (ch == 0177 || ch < 040)
		ds_append(&ds, ds_xprintf("\\%.3o", ch));
	  else
		ds_appendc(&ds, ch);
	  break;
	}

	if (esc) {
	  ds_appendc(&ds, (int) '\\');
	  ds_appendc(&ds, esc);
	}
  }

  ds_appendc(&ds, (int) '\0');
  return(ds_buf(&ds));
}

/*
 * Parse a string into fields.
 *
 * This is a quick & dirty hack to leverage mkargv() to extract fields.
 */
Strfields *
strpfields(const char *str, Strfields *fields)
{
  Mkargv *c;
  Strfields *f;

  if (fields == NULL) {
	f = ALLOC(Strfields);
	f->conf = NULL;
  }
  else
	f = fields;

  f->argc = 0;
  f->argv = NULL;

  if (f->conf == NULL) {
	c = ALLOC(Mkargv);
	c->keepq = c->keepws = 0;
	c->startq = c->endq = NULL;
	c->ifs = " \t";
	f->conf = c;
  }

  if ((f->argc = mkargv((char *) str, f->conf, &f->argv)) == -1)
	  return(NULL);

  return(f);
}

#include <limits.h>

/*
 * Format (join) selected fields into a single string.
 * Return that string, or NULL if an error is encountered.
 *
 * Format:
 *   %j - set field join separator to the string argument
 *   %nf, %:nf, %n:f, %n:mf, %:f - field selector, where n and m are integers
 *     o "%1f" means interpolate spec->argv[1], "%2:3f" means interpolate
 *       spec->argv[2], the join separator (unless NULL), and spec->argv[3])
 *     o a reference to a non-existant field is an error
 *   %s - interpolate the string argument
 *   %% - interpolate a per cent character
 * Anything else is interpolated verbatim.
 */
/*VARARGS*/
char *
strffields(const char *fmt, Strfields *spec, ...)
{
  int first, i;
  const char *endp, *f;
  char *errmsg, *j, *range, *s;
  va_list ap;
  Ds ds;
  static const char *range_char = "0123456789-,:+";

  if (fmt == NULL || spec == NULL || spec->argc == 0 || spec->argv == NULL)
	return(NULL);

  va_start(ap, spec);
  ds_init(&ds);
  f = fmt;
  j = NULL;

  while (*f != '\0') {
	if (*f == '%') {
	  f++;

	  if (*f == '%')
		ds_appendc(&ds, '%');
	  else if (*f == 'j')
		j = va_arg(ap, char *);
	  else if (*f == 's') {
		s = va_arg(ap, char *);
		ds_append(&ds, s);
	  }
	  else if (strchr(range_char, (int) *f) != NULL) {
		endp = f;
		while (*endp != '\0' && strchr(range_char, (int) *endp) != NULL)
		  endp++;
		if (*endp != 'f') {
		  va_end(ap);
		  return(NULL);
		}
		range = strndup(f, endp - f);
		if (range_is_valid(range, NULL, &errmsg) == -1) {
		  va_end(ap);
		  return(NULL);
		}
		first = 1;
		for (i = 0; i < spec->argc; i++) {
		  if (range_test(i, range, NULL, NULL) == 1) {
			if (first)
			  first = 0;
			else if (j != NULL)
			  ds_append(&ds, j);
			ds_append(&ds, spec->argv[i]);
		  }
		}
		f = endp;
	  }
	  else
		ds_appendc(&ds, *f);
	}
	else
	  ds_appendc(&ds, *f);
	f++;
  }
  ds_appendc(&ds, '\0');

  va_end(ap);

  return(ds_buf(&ds));
}

#ifndef REG_BASIC
/*
 * A FreeBSD synonym for 0, provided as a counterpart to REG_EXTENDED to improve
 * readability.
 */
#define REG_BASIC		0
#endif

/*
 * Convert a string representation of a regex(3) flag expression into an integer
 * suitable for use as an argument to strregex(), etc.
 * Refer to regex(3) for a flag definitions; the "REG_" prefix may be omitted,
 * and "0" is recognized as zero.
 * Spaces are ignored in STR, elements are separated by a comma or '|'.
 * Return -1 if STR is invalid, which assumes that -1 is not a legitimate value.
 */
int
strregexflags(char *str)
{
  int cflags, i;
  char *errmsg, *s;
  Dsvec *dsv;

  s = strelide(str, " ");
  errmsg = NULL;
  dsv = strsplit_re(s, "\\||,", 0, 0, &errmsg);
  if (dsv == NULL) {
	if (errmsg != NULL)
	  fprintf(stderr, "%s\n", errmsg);
	return(-1);
  }

  cflags = 0;
  for (i = 0; i < dsvec_len(dsv); i++) {
	char *flag;

	flag = (char *) dsvec_ptr_index(dsv, i);
	if (streq(flag, "0"))
	  ;
	else if (streq(flag, "REG_ICASE") || streq(flag, "ICASE"))
	  cflags |= REG_ICASE;
	else if (streq(flag, "REG_EXTENDED") || streq(flag, "EXTENDED"))
	  cflags |= REG_EXTENDED;
	else if (streq(flag, "REG_BASIC") || streq(flag, "BASIC"))
	  cflags |= REG_BASIC;
#ifdef REG_NOSPEC
	else if (streq(flag, "REG_NOSPEC") || streq(flag, "NOSPEC"))
	  cflags |= REG_NOSPEC;
#endif
	else if (streq(flag, "REG_NOSUB") || streq(flag, "NOSUB"))
	  cflags |= REG_NOSUB;
	else if (streq(flag, "REG_NEWLINE") || streq(flag, "NEWLINE"))
	  cflags |= REG_NEWLINE;
#ifdef REG_PEND
	else if (streq(flag, "REG_PEND") || streq(flag, "PEND"))
	  cflags |= REG_PEND;
#endif
	else
	  return(-1);
  }

  return(cflags);
}

/*
 * Match regular expression REGEX (modified by CFLAGS) against SRC.
 *
 * If successful return 1.
 * If there was no match, return 0.
 * If an error occurred, return -1 and (optionally) set ERRMSG.
 *
 * If non-NULL, MATCH will be set to NULL (if no match) or, if REG_NOSUB
 * was not specified, to an allocated copy of the matching substring.
 *
 * If non-NULL, ERRMSG is set to NULL or a descriptive error message.
 */
int
strregex(char *src, char *regex, int cflags, char **match, char **errmsg)
{
  int rc, st;
  regmatch_t m[2];
  regex_t preg;

  if (match != NULL)
	*match = NULL;

  if (src == NULL || regex == NULL) {
	if (errmsg != NULL)
	  *errmsg = "Invalid argument";
	return(-1);
  }

  if (errmsg != NULL)
	*errmsg = NULL;

  m[0].rm_so = m[0].rm_eo = -1;
  m[1].rm_so = m[1].rm_eo = -1;

  if ((st = regcomp(&preg, regex, cflags)) != 0) {
	char errbuf[100];

  fail:
	if (errmsg != NULL) {
	  errbuf[0] = '\0';
	  regerror(st, &preg, errbuf, sizeof(errbuf));
	  *errmsg = ds_xprintf("Invalid regular expression: \"%s\": %s",
						   regex, errbuf);
	}

	return(-1);
  }

  if ((st = regexec(&preg, src, 1, m, 0)) == 0) {
	/* It's a match... */
	rc = 1;
	if ((cflags & REG_NOSUB) == 0 && match != NULL) {
	  int n;

	  n = m[0].rm_eo - m[0].rm_so;
	  *match = strndup(src + m[0].rm_so, n);
	}
  }
  else if (st != REG_NOMATCH)
	goto fail;
  else
	rc = 0;

  regfree(&preg);

  return(rc);
}

/*
 * Initialize (or optionally allocating) a configuration structure to
 * be used with strregex_x().
 * Any previous memory allocation will be lost.
 * Returns the initialized structure.
 */
Strregex_conf *
strregex_init(Strregex_conf *conf, char *regex_str, int cflags,
			  unsigned int nmatch)
{
  Strregex_conf *c;

  if (conf == NULL)
	c = ALLOC(Strregex_conf);
  else
	c = conf;

  c->regex = regex_str;
  c->cflags = cflags;
  c->preg_in = NULL;
  c->nmatch = nmatch;
  c->errcode = 0;

  return(c);
}

/*
 * Extended regular expression matching.
 * SRC is the string that is matched against.
 * The CONF argument provides regular expression REGEX and
 * modification flags CFLAGS (see regcomp(3) and strregexflags()).
 * Frequently used flags:
 * REG_EXTENDED (extended/modern REs),
 * REG_ICASE (case insensitive matching),
 * REG_NOSUB (do not need matched subexpressions),
 * REG_NOSPEC (literal matching, pattern matching characters not special).
 *
 * PREG_IN will be used if non-NULL - it is assumed to have been obtained
 * from the given REGEX and CFLAGS (which may matter if an error is reported).
 * If PREG_IN is NULL, it will be set to the newly compiled REGEX.
 * The caller is responsible for calling regfree(3).
 * The NMATCH argument is passed to regexec(3).
 *
 * If NMATCH is non-NULL, element 0 will be the substring of SRC that was
 * matched by the entire regular expression; element 1 will be the substring
 * matched by the first parenthesized subexpression (if any), and so on
 * for each subexpression up to NMATCH.
 *
 * The errcode element will be a non-zero regex(3) error code if an error occurs.
 * The caller may use regerror(3) to produce a message.
 *
 * Return 0 if there is no match, -1 if there is an error; otherwise,
 * return the number of subexpression matches plus 1.
 *
 * XXX Not intended to be a complete interface; e.g.,
 * the EFLAGS argument to regexec(3) is not supported.
 */
int
strregex_x(char *src, Strregex_conf *conf, Dsvec *matches)
{
  int i, rc, st;
  Ds ds;
  /* There may need to be at least one element; see man 3 regex. */
  regmatch_t m[1 + conf->nmatch];
  regex_t *preg;

  if (src == NULL || conf == NULL)
	return(-1);

  rc = -1;

  if (conf->preg_in == NULL) {
	if (conf->regex == NULL) {
	  conf->errcode = -1;
	  return(-1);
	}

	conf->preg_in = ALLOC(regex_t);
	if ((st = regcomp(conf->preg_in, conf->regex, conf->cflags)) != 0)
	  goto fail;
  }

  conf->errcode = 0;
  preg = conf->preg_in;

  /*
   * Initialize in case no matching subexpressions are returned by regex().
   * Because the regex may have already been compiled (regex == NULL and
   * CFLAGS is immaterial), we don't know if regexec() will be returning
   * matching subexpressions.
   */
  for (i = 0; i < (1 + conf->nmatch); i++)
	m[i].rm_so = m[i].rm_eo = -1;

  if ((st = regexec(preg, src, conf->nmatch, m, 0)) == 0) {
	/* It's a match... */
	rc = 0;
	for (i = 0; m[i].rm_so != -1 && i < conf->nmatch; i++) {
	  if (matches != NULL) {
		int n;

		n = m[i].rm_eo - m[i].rm_so;
		ds_init(&ds);
		ds_concatn(&ds, src + m[i].rm_so, n);
		dsvec_add_ptr(matches, ds_buf(&ds));
	  }
	  /* Counts the entire match plus the substring matches. */
	  rc++;
	}
	if (i == 0) {
	  /* There were no substring matches. */
	  rc = 1;
	}
  }
  else if (st != REG_NOMATCH) {
  fail:
	conf->errcode = st;
	rc = -1;
  }
  else
	rc = 0;

  if (rc == -1) {
	if (conf->preg_in != NULL) {
	  regfree(conf->preg_in);
	  conf->preg_in = NULL;
	}
  }

  return(rc);
}

/*
 * Simple interface to strregex() to match REGEX against SRC, with default
 * matching behaviour.
 *
 * Use strregex() or strregex_x() for more matching and output options.
 *
 * Return 0 if there is no match, 1 if there is a match, or
 * -1 if there is a problem with REGEX.
 */
int
strmatch(char *src, char *regex)
{
  int st;

  st = strregex(src, regex, 0, NULL, NULL);

  return(st);
}

/*
 * Do the hard part of regex-based substitution, applying REGEX to
 * SRC and replacing the matched part of SRC according to REPL.
 * CFLAGS are regex matching flags, REPEAT is non-zero if
 * matching/replacement should be repeated until unsuccessful (like the
 * 'g' modifier in s/REGEX/REPL/g).
 *
 * Submatches of SRC specified in REGEX (e.g., "foo (.*) baz") are
 * referenced as variables ${1} through ${9}, with '&' and ${0} representing
 * the entire part of SRC that was matched.
 *
 * Return the resulting string, or NULL upon error.
 * If there's no match at all, NULL is returned.
 */
char *
strregexsub(char *src, char *regex, char *repl, int cflags, int repeat,
			char **errmsg)
{
  int ind, st;
  char *r, *text;
  regex_t preg;
  regoff_t start_off, next_off;
  regmatch_t pmatch[10];
  Ds value;

  if ((st = regcomp(&preg, regex, cflags)) != 0)
	goto fail;

  ds_init(&value);

  text = src;
  while (1) {
	if ((st = regexec(&preg, text, 10, pmatch, 0)) != 0) {
	  if (st != REG_NOMATCH)
		goto fail;
	  if (ds_len(&value) == 0)
		ds_set(&value, "");
	  else if (*text != '\0')
		ds_concat(&value, text);
	  break;
	}

	start_off = pmatch[0].rm_so;
	next_off = pmatch[0].rm_eo;
	if (start_off != 0)
	  ds_concatn(&value, text, start_off);

	for (r = repl; *r != '\0'; r++) {
	  if (*r == '&')
		ds_concatn(&value, text + start_off, next_off - start_off);
	  else if (*r == '$' && *(r + 1) == '{'
			   && isdigit((int) *(r + 2))
			   && *(r + 3) == '}') {
		r += 2;
		ind = *r - '0';
		if (pmatch[ind].rm_so == -1 || pmatch[ind].rm_eo == -1)
		  return(NULL);
		ds_concatn(&value, text + pmatch[ind].rm_so,
				   pmatch[ind].rm_eo - pmatch[ind].rm_so);
		r++;
	  }		
	  else {
		if (*r == '\\')
		  r++;
		ds_concatn(&value, r, 1);
	  }
	}

	text += next_off;
	if (!repeat) {
	  if (*text != '\0')
		ds_concat(&value, text);
	  break;
	}
  }

  regfree(&preg);
  return(ds_buf(&value));

 fail:
  if (errmsg != NULL) {
	char errbuf[1024];

	errbuf[0] = '\0';
	regerror(st, &preg, errbuf, sizeof(errbuf));
	*errmsg = ds_xprintf("bad regular expression: \"%s\": %s",
						 regex, errbuf);
  }

  regfree(&preg);

  return(NULL);
}

/*
 * Return a copy of PATHNAME after deleting 1) any prefix ending with the last
 * slash character (after first stripping trailing slashes), and 2) a SUFFIX,
 * if given.  The SUFFIX string is not stripped if it is identical to the
 * remaining characters in PATHNAME.  The resulting string is returned.
 * A non-existent suffix is ignored.
 * This is intended to be like the basename(1) command.
 */
char *
strbasename(const char *pathname, const char *suffix)
{
  char *last, *p, *q, *s;
  size_t len;

  len = strlen(pathname);
  if (len == 0)
	return(strdup(""));

  s = strdup(pathname);
  p = s + len - 1;

  /* Delete trailing slashes first. */
  while (p > s && *p == '/')
	*p-- = '\0';

  /* Is the pathname equivalent to "/"? */
  if (*p == '/')
	return(strdup("/"));

  if ((last = rindex(s, (int) '/')) != NULL)
	*last++ = '\0';
  else
	last = s;

  /*
   * We've found the end of the last component.
   * If SUFFIX was given, check if it needs to be deleted from the end of
   * the string.
   */
  if (suffix != NULL && !streq(last, suffix)
	  && (q = strsuffix(last, strlen(last), suffix)) != NULL)
	*q = '\0';

  return(last);
}

/*
 * Return a copy of PATHNAME with its filename portion deleted; i.e., the
 * string beginning with the last slash character continuing to the end of
 * string (after first stripping trailing slashes).
 * Notes:
 * dirname("/") is "/".
 * dirname("/a/b/c") is "/a/b".
 * dirname("") is dirname("./") is dirname(".") is ".".
 * dirname("foo") is dirname("./foo") is dirname("foo/") is "."
 * dirname("foo/baz") is "foo"
 * dirname("foo/bar/baz") is "foo/bar"
 * This is intended to be like the dirname(1) command.
 */
char *
strdirname(const char *pathname)
{
  char *p, *s;
  size_t len;

  len = strlen(pathname);
  if (len == 0)
	return(strdup("."));

  s = strdup(pathname);
  p = s + len - 1;

  /* Delete consecutive trailing slashes. */
  while (p > s && *p == '/')
	*p-- = '\0';

  /* Does PATHNAME only consist of slashes? */
  if (*p == '/')
	return(strdup("/"));

  /* Scan backwards until the first slash. */
  while (p > s && *p != '/')
	p--;

  /* Found the right-most slash of interest? */
  if (*p == '/') {
	*p = '\0';
	return(s);
  }

  return(strdup("."));
}

/*
 * Return a copy of the extension part (including the dot) of the last
 * component of PATHNAME, or NULL if no extension is found.
 */
char *
strextname(const char *pathname)
{
  char *dot, *slash;

  dot = strrchr(pathname, (int) '.');
  slash = strrchr(pathname, (int) '/');

  if (dot != NULL) {
	if (slash == NULL || slash < dot)
	  return(strdup(dot));
  }

  return(NULL);
}

/*
 * XXX buggy, needs work
 * Create a copy of STR with some simple formatting.
 * If any line of STR has more than MAX_LEN characters, try to split it.
 * Any character in BRKSTR can be replaced with a line break; if BRKSTR
 * is NULL, a space is used as the break character.
 * Return the new string, or the original string if no formatting is possible.
 */
char *
strfmt(const char *str, const char *brkstr, size_t max_len)
{
  char *lastp, *nstr, *p;
  const char *brks;
  size_t len;

  if (brkstr == NULL)
	brks = " ";
  else
	brks = brkstr;

  nstr = strdup(str);

  lastp = NULL;
  len = 0;
  for (p = nstr; *p != '\0'; p++) {
	len++;
	if (strchr(brks, (int) *p) == NULL)
	  continue;
	if (len > max_len) {
	  if (lastp == NULL) {
		/* XXX What to do? */
		return(strdup(str));
	  }
	  *lastp++ = '\n';
	  len = p - lastp;
	}
	lastp = p;
  }

  if (len > max_len) {
	if (lastp == NULL) {
	  /* XXX What to do? */
	  return(strdup(str));
	}
	*lastp = '\n';
  }

  return(nstr);
}

/*
 * Convert a pair of hex digit characters into a byte value
 * (0 <= value <= 255), or return -1 if the pair is invalid.
 */
int
hexpair2int(char *pair)
{
  int byte, hi, lo;

  lo = pair[1];
  hi = pair[0];

  if (!isxdigit(lo) || !isxdigit(hi))
	return(-1);

  hi = tolower(hi);
  lo = tolower(lo);
  byte = (hi >= 'a') ? (hi - 'a' + 10) : (hi - '0');
  byte <<= 4;
  byte += ((lo >= 'a') ? (lo - 'a' + 10) : (lo - '0'));

  return(byte);
}

/*
 * Alternate API for strhextob(), converting string STR consisting of
 * hex character pairs (0-9 and [aA-fF]) into binary.
 * Return a pointer to the (allocated) buffer, set OUTP to point to it
 * (if it is non-NULL), and set LENP to the length of the output buffer in
 * bytes (if LENP is non-NULL), which will be zero if an error occurs.
 */
unsigned char *
stra16b(char *str, unsigned char **outp, unsigned int *lenp)
{
  unsigned char *p;

  p = strhextob(str, lenp);
  if (outp != NULL)
	*outp = p;

  return(p);
}

/*
 * Convert STR, a null-terminated string consisting of hex character pairs,
 * into binary.  Set LENP to the size of the binary vector, in bytes,
 * which will be zero if an error occurs.
 * Return the allocated buffer, or NULL if the string contains an
 * invalid character.
 */
unsigned char *
strhextob(char *str, unsigned int *lenp)
{
  int blen, bval, i, n;
  char *p;
  size_t slen;
  unsigned char *uc;

  slen = strlen(str);
  if (slen == 0 || (slen & 01)) {
	if (lenp != NULL)
	  *lenp = 0;
	return(NULL);
  }

  blen = slen / 2;
  if (lenp != NULL)
	*lenp = blen;
  uc = (unsigned char *) malloc(blen);

  p = str;
  n = 0;
  for (i = 0; i < blen; i++) {
	if ((bval = hexpair2int(p)) == -1) {
	  if (lenp != NULL)
		*lenp = 0;
	  return(NULL);
	}
	uc[n++] = bval;
	p += 2;
  }

  return(uc);
} 

/*
 * See strbtohex().
 */
Ds *
ds_strba16(Ds *ods, unsigned char *buf, size_t len, int flags)
{
  int hi, lo;
  size_t ui;
  Ds *ds;
  static char *hex_upper = "0123456789ABCDEF";
  static char *hex_lower = "0123456789abcdef";

  if (ods == NULL) {
    if ((ds = ds_init(NULL)) == NULL)
      return(NULL);
  }
  else
    ds = ods;

  for (ui = 0; ui < len; ui++) {
	hi = (buf[ui] & 0xf0) >> 4;
	lo = (buf[ui] & 0x0f);
	if (flags & STRBTOHEX_UPPER) {
	  if (! (ui == 0 && hi == 0 && (flags & STRBTOHEX_NOZ)))
		ds_appendc(ds, (int) hex_upper[hi]);
	  ds_appendc(ds, (int) hex_upper[lo]);
	}
	else {
	  if (! (ui == 0 && hi == 0 && (flags & STRBTOHEX_NOZ)))
		ds_appendc(ds, (int) hex_lower[hi]);
	  ds_appendc(ds, (int) hex_lower[lo]);
	}
  }
  ds_appendc(ds, (int) '\0');

  return(ds);
}

/*
 * Convert binary string BUF of length LEN bytes into an ASCII hex character
 * encoding.  Use upper case hex digits if flag STRBTOHEX_UPPER is present,
 * suppress a leading zero if flag STRBTOHEX_NOZ is present.
 * The flags can be ORed; if no flags are needed, FLAGS should be
 * STRBTOHEX_DEFAULT (which is zero).
 */
char *
strbtohex(unsigned char *buf, size_t len, int flags)
{
  Ds *ds;

  ds = ds_strba16(NULL, buf, len, flags);

  return(ds_buf(ds));
}

#ifdef NOTDEF
/*
 * ModHex alphabet (lowercase only), valued from 0x00 to 0x0f:
 *   c,b,d,e,f,g,h,i,j,k,l,n,r,t,u,v   (ModHex)
 *   0,1,2,3,4,5,6,7,8,9,A,B,C,D,E,F   (Hex)
 * These characters use the same scan codes across many USB HID-mode
 * (Human Interface Device) keyboard layouts.
 * E.g.,  0x4711 == fibb
 *
 * /usr/src/share/misc/usb_hid_usages
 * https://forum.yubico.com/viewtopica943.html?p=252
 * https://demo.yubico.com/php-yubico/Modhex_Calculator.php
 */
#endif

/*
 * Return 1 if CH meets classification/constraint ALLOWABLE with
 * optional SPEC, 0 otherwise.
 */
int
is_strclass_char(int ch, int allowable, char *spec)
{

  if (IS_STRCLASS(allowable, STRCLASS_NUL) && ch == '\0')
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_ALNUM) && isalnum(ch))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_ALPHA) && isalpha(ch))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_DECIMAL) && isdigit(ch))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_LC) && islower(ch))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_UC) && isupper(ch))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_SPACE) && ch == ' ')
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_EOL) && (ch == '\n' || ch == '\r'))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_PUNCT) && ispunct(ch))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_OCTAL) && (ch >= '0' && ch <= '7'))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_HEX_UC) && (ch >= 'A' && ch <= 'F'))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_HEX_LC) && (ch >= 'a' && ch <= 'f'))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_GENNUM)
	  && (ch == '+' || ch == '-' || isdigit(ch) || ch == '.' || ch == ','))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_NUMERIC_DATE)
	  && (isdigit(ch) || ch == '/' || ch == '-'))
	return(1);

  if (IS_STRCLASS(allowable, STRCLASS_SPEC)
	  && spec != NULL && strchr(spec, ch) != NULL)
	return(1);

  return(0);
}

/*
 * Return 1 if the characters in STR satisfy ALLOWABLE, which is a bit
 * mask formed from defined character classes, 0 otherwise.
 *
 * Special control flags that flag arguments that follow ALLOWABLE are processed
 * in the following order:
 * 1) If ALLOWABLE includes STRCLASS_SPEC, then an argument is expected
 * that is a string specifying valid characters.
 * 2) If ALLOWABLE includes STRCLASS_STOPCH, then an argument is
 * expected; that argument will be set to point to the first character
 * that does not meet the constraint.
 *
 * Examples:
 * is_strclass("aaa8",STRCLASS_ALNUM|STRCLASS_STOPCH, &pos)
 * Returns 1 with *pos == '\0'.
 * is_strclass("aaa8;",STRCLASS_ALNUM|STRCLASS_STOPCH, &pos)
 * Returns 0 with *pos == ';'.
 *
 * Note that a return of 1 does not necessarily mean that STR is syntactically
 * correct, only that it is comprised of the specified character set.
 *
 * Also see: is_strclass_char() and strnum().
 * This is an attempt at reducing the proliferation of functions that test
 * for various syntaxes.
 */
/*VARARGS*/
int
is_strclass(char *str, int allowable, ...)
{
  char *p, *spec, **stopch;
  va_list ap;

  va_start(ap, allowable);
  if (allowable & STRCLASS_SPEC)
	spec = va_arg(ap, char *);
  else
	spec = NULL;

  if (allowable & STRCLASS_STOPCH) {
	stopch = va_arg(ap, char **);
	**stopch = '\0';
  }
  else
	stopch = NULL;
  va_end(ap);

  for (p = str; *p != '\0'; p++) {
	if (is_strclass_char((int) *p, allowable, spec))
	  continue;

	if (IS_STRCLASS(allowable, STRCLASS_SIGNED_DIGITS)) {
	  if (*p == '+' || *p == '-' || isdigit((int) *p))
		continue;
	}

	if (IS_STRCLASS(allowable, STRCLASS_HOSTNAME)) {
	  if (isalnum((int) *p) || *p == '-' || *p == '.')
		continue;
	}

	if (stopch != NULL)
	  *stopch = p;

	return(0);
  }

  return(1);
}

/*
 * Convert binary byte string BUF of length LEN bytes into a
 * UTF-16LE encoding.
 * http://en.wikipedia.org/wiki/UTF-16
 */
Ds *
strbtoutf16le(unsigned char *buf, size_t nbytes)
{
  int i;
  Ds *ds;

  ds = ds_init(NULL);
  for (i = 0; i < nbytes; i++) {
	ds_appendc(ds, (int) buf[i]);
	ds_appendc(ds, (int) '\0');
  }

  return(ds);
}

/*
 * Return 1 iff STR is (safely) printable.  If ALLOW_SP_WS is non-zero,
 * treat tabs, newlines, and carriage returns as safe, otherwise reject them.
 * If NBYTES is 0, then compute the length of STR, assuming that it is
 * null-terminated and that the first null really is supposed to be the end.
 * If NBYTES is not 0, then it is the number of bytes to look at in STR,
 * assuming that a null character is invalid.
 */
int
strprintable(char *str, size_t nbytes, int allow_sp_ws)
{
  int i;
  unsigned char *p;
  size_t len;

  if (nbytes == 0)
	len = strlen(str);
  else
	len = nbytes;

  p = (unsigned char *) str;
  for (i = 0; i < len; i++) {
	if (*p == '\t' || *p == '\n' || *p == '\r') {
	  if (!allow_sp_ws)
		return(0);
	}
	else if (!isprint((int) *p))
	  return(0);

	p++;
  }

  return(1);
}

/*
 * Return a string consisting of unique characters in STR.
 * The returned string, which is null-terminated, will not contain any duplicate
 * characters. STR may contain any byte values (except a null byte).
 */
char *
strunique(char *str)
{
  char *p, *q, *s;
  size_t len;
  Ds ds;

  ds_init(&ds);
  s = strdup(str);
  len = strlen(s);

  for (p = s; (p - s) < len; p++) {
	if (*p == '\0') {
	  /* A repeated character appeared here. */
	  continue;
	}

	for (q = p + 1; (q - s) < len; q++) {
	  if (*p == *q) {
		/* "Erase" the repeated character. */
		*q = '\0';
	  }
	}
	ds_appendc(&ds, (int) *p);
  }
  ds_appendc(&ds, (int) '\0');

  free(s);
  return(ds_buf(&ds));
}

/*
 * Return a new string with every instance of characters in DEL deleted from
 * STR.
 */
char *
strelide(char *str, char *del)
{
  char *p;
  Ds ds;

  ds_init(&ds);
  for (p = str; *p != '\0'; p++) {
	if (strchr(del, (int) *p) == NULL)
	  ds_appendc(&ds, (int) *p);
  }
  ds_appendc(&ds, (int) '\0');

  return(ds_buf(&ds));
}

/*
 * Collapse (delete) consecutive whitespace characters in ODS, returning
 * a new copy of the string.
 * To delete all whitespace characters, see strelide().
 */
Ds *
strelidews(Ds *nds, Ds *ods)
{
  int prev_ch;
  char *p;
  Ds *ds;

  if (nds == ods)
	return(NULL);

  if (nds == NULL)
	ds = ds_init(NULL);
  else
	ds = nds;

  prev_ch = '\0';
  for (p = ds_buf(ods); *p != '\0'; p++) {
	if (*p != ' ' && *p != '\t')
	  ds_appendc(ds, (int) *p);
	else if (prev_ch != ' ' && prev_ch != '\t')
	  ds_appendc(ds, (int) *p);

	prev_ch = *p;
  }
  ds_appendc(ds, (int) '\0');

  return(ds);
}

/*
 * Return a new string with any contiguous sequence of characters from DEL
 * deleted from the end of STR; e.g.,
 *   strchop("foo4.859", ".56789") -> "foo4"
 *   strchop("foo7.859", ".5679")  -> "foo7.8"
 *   strchop("/a/b/c///", "/")     -> "/a/b/c"
 *   strchop("hello    ", " ")     -> "hello"
 * Particularly useful for deleting whitespace or newlines from the end of STR.
 * XXX should probably be a regex version
 */
char *
strchop(char *str, char *del)
{
  char *e;
  Ds ds;

  if (del == NULL || *del == '\0')
	return(strdup(str));

  e = str + strlen(str) - 1;
  while (e >= str) {
	if (strchr(del, (int) *e) == NULL)
	  break;
	e--;
  }

  ds_init(&ds);
  ds_setn(&ds, (unsigned char *) str, e - str + 1);
  ds_appendc(&ds, (int) '\0');

  return(ds_buf(&ds));
}

/*
 * Reinitialize state in CONV_TAB so that it can be used again.
 */
void
stropt_reset(Stropt_tab *conv_tab)
{
  Stropt_tab *oc;

  for (oc = conv_tab; oc->name != NULL; oc++) {
	oc->count = 0;
	oc->value = NULL;		/* XXX leaky */
  }

}

/*
 * Scan CONV_TAB for key NAME, returning its entry (the first instance found),
 * or NULL if it is not present.
 * If NAME is NULL, return the last entry in CONV_TAB, which can be used
 * to handle any NAME that does not appear in CONV_TAB.
 */
Stropt_tab *
stropt_lookup(Stropt_tab *conv_tab, char *name)
{
  Stropt_tab *oc;

  for (oc = conv_tab; oc->name != NULL; oc++) {
	int st;

	if (name != NULL) {
	  if (oc->flags & STROPT_NOCASE)
		st = strcaseeq(oc->name, name);
	  else
		st = streq(oc->name, name);

	  if (st)
		return(oc);
	}
  }

  if (name == NULL)
	return(oc);

  return(NULL);
}

/*
 * Simple table-based utility for parsing and converting NAME=VALUE strings,
 * where NAME is an option (variable) being assigned VALUE.
 *
 * Search for the entry for NAME in CONV_TAB, which maintains state between
 * calls, and if it is found convert VALUE according to the entry's conversion
 * rule, identify the entry for the caller if WHICH is non-NULL, and return
 * STROPT_OK.
 * If NAME is not found, if conversion fails, or if any other error occurs,
 * return a descriptive result code.
 *
 * Only the first instance of NAME that is found in the search of CONV_TAB
 * is examined.  It is possible, but probably a bad idea for a name to appear
 * more than once in a given Strop_tab.
 */
Stropt_rc
stropt(Stropt_tab *conv_tab, char *name, char *value, Stropt_tab **which)
{
  int st, key_mode;
  char *conv_val, *value_str;
  void *v;
  Stropt_tab *oc;

  /*
   * XXX could add a flag that permits VALUE to be NULL, in which case
   * NAME would be a Boolean that defaults to false and is set to true
   * if VALUE is NULL.
   */
  if (name == NULL || *name == '\0' || value == NULL)
	return(STROPT_PARAM_ERR);

  /* Scan for NAME. */
  if ((oc = stropt_lookup(conv_tab, name)) == NULL) {
	oc = stropt_lookup(conv_tab, NULL);
	/*
	 * XXX if the last entry is configured to catch default cases,
	 * process this one.
	 */
	return(STROPT_NOTFOUND_ERR);
  }

  /* Found... now what? */
  key_mode = oc->flags & STROPT_KEY_MODE_MASK;
  if (key_mode == STROPT_KEY_IGNORE)
	goto done_ignore;

  /* Only the first occurrence is important so terminate. */
  if (oc->count) {
	/* Already saw NAME... */
	if (key_mode == STROPT_KEY_KEEP_FIRST) {
	  if (which != NULL)
		*which = oc;
	  return(STROPT_OK);
	}

	/* Supposed to be unique? */
	if (key_mode != STROPT_KEY_KEEP_LAST && key_mode != STROPT_KEY_MULTI) {
	  /* Duplicates disallowed. */
	  return(STROPT_DUP_ERR);
	}
  }

  /*
   * If pre_callback() has been configured, it is called here,
   * before conversion, in case any massaging of the VALUE is needed.
   * It is passed a pointer to a copy of VALUE so that it can modify
   * VALUE if it wants.
   */
  if (oc->pre_callback != NULL) {
	Stropt_rc rc;

	conv_val = strdup(value);
	rc = oc->pre_callback(oc, name, &conv_val);
	if (rc != STROPT_OK)
	  return(rc);
  }
  else
	conv_val = value;

  /*
   * If there is a callback function call it to set VALUE, otherwise
   * use the default conversion function.
   */
  if (oc->callback_func == NULL) {
	st = strnum_b(conv_val, oc->type, STRNUM_ALLOC, 0, &v, NULL);
	if (st == -1)
	  return(STROPT_TYPE_ERR);
  }
  else {
	/* If the callback returns -1, we are done. */
	st = (oc->callback_func)(oc, name, conv_val, &v);
	if (st == -1)
	  return(STROPT_CONVERSION_ERR);
  }

  /*
   * If post_callback() has been configured, it should be called here,
   * just after conversion, in case error checking etc. is needed.
   */
  if (oc->post_callback != NULL) {
	Stropt_rc rc;

	rc = oc->post_callback(oc, name, conv_val, &v);
	if (rc != STROPT_OK)
	  return(rc);
  }

  /* Only the last occurrence is kept. */
  if (oc->count == 0 || key_mode != STROPT_KEY_KEEP_LAST)
	oc->count++;

  /*
   * If this is the first instance of this option, store it in the usual way.
   * If this is not the first instance, all of the values are kept,
   * so create or append to the vector of values.
   * Keep the original string value(s), if requested.
   */

  if ((oc->flags & STROPT_VALUE_MASK) == STROPT_VALUE_SET_PTR)
	value_str = value;
  else if ((oc->flags & STROPT_VALUE_MASK) == STROPT_VALUE_COPY)
	value_str = strdup(value);
  else
	value_str = NULL;

  if (oc->count == 1) {
	/* XXX leaky if STROPT_KEY_KEEP_LAST. */
	oc->value = v;
	oc->value_str = value_str;
  }
  else {
	Dsvec *dsv;

	if (oc->count == 2) {
	  dsv = dsvec_init(NULL, sizeof(Dsvec *));
	  dsvec_add_ptr(dsv, oc->value);
	  oc->value = dsv;

	  if ((oc->flags & STROPT_VALUE_MASK) != STROPT_VALUE_NOCOPY) {
		dsv = dsvec_init(NULL, sizeof(Dsvec *));
		dsvec_add_ptr(dsv, oc->value_str);
		oc->value_str = dsv;
	  }
	}
	dsvec_add_ptr((Dsvec *) oc->value, v);
	if (oc->value_str != NULL)
	  dsvec_add_ptr((Dsvec *) oc->value_str, value_str);
  }

 done_ignore:

  if (which != NULL)
	*which = oc;

  return(STROPT_OK);
}

/*
 * A little "value added" function to do some generic error checking
 * after all options in CONV_TAB have been fully processed.
 */
Stropt_rc
stropt_check(Stropt_tab *conv_tab, Stropt_tab **which)
{
  Stropt_tab *oc;

  for (oc = conv_tab; oc->name != NULL; oc++) {
	if ((oc->flags & STROPT_REQUIRED) && oc->count == 0) {
	  if (which != NULL)
		*which = oc;
	  return(STROPT_NOTFOUND_ERR);
	}
  }

  return(STROPT_OK);
}

/*
 * This is primarily a demonstration of how stropt() might be used.
 * Like kwv_opt(), except each element of ARGV has a "name=value"
 * format and the last element is NULL.
 * Return STROPT_OK if successful, otherwise a descriptive error code.
 * If IND is non-NULL, it is always set to the index of the first element
 * (starting at zero) that was not converted; ARGV[*IND] will be NULL if all
 * elements were converted.  This allows the caller to handle special
 * cases and then resume option processing by calling here again with ARGV
 * adjusted to point to the new starting element.
 */
int
stropt_argv(char **argv, Stropt_tab *conv_tab, int *ind)
{
  int i;
  Stropt_rc rc;

  for (i = 0; argv[i] != NULL; i++) {
	char *endp, *name, *val;
	Stropt_tab *oc;

	/* Parse out each keyword/option name and its value. */
	if ((name = strchrdup(argv[i], (int) '=', &endp)) == NULL
		|| *endp == '\0') {
	  rc = STROPT_PARSE_ERR;

	err:
	  if (ind != NULL)
		*ind = i;
	  return(rc);
	}

	val = endp + 1;

	if ((rc = stropt(conv_tab, name, val, &oc)) != STROPT_OK)
	  goto err;
  }

  if (ind != NULL)
	*ind = i;

  return(STROPT_OK);
}

/*
 * A custom callback for converting VALUE into its internal equivalent code.
 * Return 0 if ok, -1 otherwise.
 */
static int
stropt_method_callback(Stropt_tab *m, char *name, char *value, void **v)
{
  Http_method method;

  method = http_string_to_method(value);
  if (method == HTTP_UNKNOWN_METHOD)
	return(-1);

  *v = ALLOC(int);
  *(int *) *v = method;

  return(0);
}

static Stropt_tab option_tab[] = {
  { "AUXILIARY", STRNUM_STR, STROPT_NOCASE|STROPT_KEY_KEEP_LAST,
	NULL, NULL, NULL, NULL,
	0, NULL, NULL },
  { "AUTH_METHOD", STRNUM_INT, STROPT_VALUE_COPY|STROPT_KEY_KEEP_FIRST,
	NULL, NULL, stropt_method_callback, NULL,
	0, NULL, NULL },
  { "AUTH_URL", STRNUM_STR, STROPT_KEY_MULTI,
	NULL, NULL, NULL, NULL,
	0, NULL, NULL },
  { "USERNAME",  STRNUM_STR, 0,
	NULL, NULL, NULL, NULL,
	0, NULL, NULL },
  { "PASSWORD",  STRNUM_STR, STROPT_KEY_UNIQUE,
	NULL, NULL, NULL, NULL,
	0, NULL, NULL },
  { NULL,        STRNUM_ERR, 0,
	NULL, NULL, NULL, NULL,
	0, NULL, NULL }
};

/*
 * Option functionality self tests.
 */
static MAYBE_UNUSED int
stropt_test(void)
{
  int ind;
  Kwv *kwv;
  Stropt_tab *oc;

  kwv = kwv_init(16);
  kwv->dup_mode = KWV_ALLOW_DUPS;
#ifdef NOTDEF
  kwv_add(kwv, "USERNAME", "bob");
  kwv_add(kwv, "PASSWORD", "%%SeCrET%%");
#endif
  kwv_add(kwv, "AUTH_METHOD", "post");
  kwv_add(kwv, "AUTH_METHOD", "get");
  kwv_add(kwv, "auxiliary", "myaux1");
  kwv_add(kwv, "auxiliary", "myaux2");
  kwv_add(kwv, "AUTH_URL", "http://example.com/my/url-1");
  kwv_add(kwv, "AUTH_URL", "http://example.com/my/url-2");
  kwv_add(kwv, "BOGUS", "bogus");
  kwv_add(kwv, "auxiliary", "myaux3");

  if (stropt(option_tab, "PASSWORD", "foozle", &oc) != STROPT_OK)
	exit(1);
  fprintf(stderr, "\"%s\" : \"%s\"\n", oc->name,  (char *) oc->value);

  stropt_reset(option_tab);
  ind = 0;
  if (kwv_opt(kwv, option_tab, &ind) != STROPT_OK) {
	fprintf(stderr, "ind=%d\n", ind);
  }

  for (oc = option_tab; oc->name != NULL; oc++) {
	if (oc->count == 0)
	  fprintf(stderr, "\"%s\" : NULL\n", oc->name);
	else {
	  if (oc->count == 1) {
		fprintf(stderr, "\"%s\" : \"%s\"\n", oc->name,
				(oc->value_str == NULL) ?
				(char *) oc->value : (char *) oc->value_str);
	  }
	  else {
		int i;
		Dsvec *dsv;

		if ((dsv = (Dsvec *) oc->value_str) == NULL)
		  dsv = (Dsvec *) oc->value;

		fprintf(stderr, "\"%s\" (%d):\n", oc->name, dsvec_len(dsv));
		for (i = 0; i < dsvec_len(dsv); i++)
		  fprintf(stderr, "  \"%s\"\n", (char *) dsvec_ptr_index(dsv, i));
	  }
	}
  }

  return(0);
}

/*
 * Simple tabular formatting for strings.
 * NCOLS is the number of columns in the table, COLUMN_SEP is the minimum
 * amount of blank space between columns, and if ALIGN is non-zero, the
 * lhs of each column is aligned (left justified).
 * Originally written to pretty print program usage displays (command flags).
 */
Strtable *
strtable_init(int ncols, int column_sep, int align)
{
  int i;
  Strtable *stab;

  stab = ALLOC(Strtable);
  stab->ncols = ncols;
  stab->nrows = 0;
  stab->column_sep = column_sep;
  stab->align = align;
  stab->columns = ALLOC_N(Dsvec *, ncols);
  for (i = 0; i < ncols; i++)
	stab->columns[i] = dsvec_init(NULL, sizeof(char *));

  return(stab);
}

/*
 * Add a row to table ST.
 * The caller must supply the correct number of string arguments.
 */
/*VARARGS*/
void
strtable_add_row(Strtable *stab, ...)
{
  int col;
  char *s;
  Dsvec *dsv;
  va_list ap;
  
  va_start(ap, stab);
  for (col = 0; col < stab->ncols; col++) {
	dsv = stab->columns[col];
	s = va_arg(ap, char *);
	dsvec_add_ptr(dsv, s);
  }
  va_end(ap);
  stab->nrows++;
}

/*
 * Default comparison function for sorting a column (case sensitive).
 */
static int
strtable_column_compar(const void *ap, const void *bp)
{
  Strtable_row_sort *a, *b;

  a = *(Strtable_row_sort **) ap;
  b = *(Strtable_row_sort **) bp;

  return(strcmp(a->colval, b->colval));
}

/*
 * Sort column COL of table ST using COMPAR.
 * Column numbers start at zero.
 * If COMPAR is NULL, use the default collation.
 * Returns a vector that describes the sorted row permutation.
 */
Dsvec *
strtable_sort(Strtable *stab, int col,
			  int (*compar)(const void *, const void *))
{
  int row;
  int (*c)(const void *, const void *);
  Dsvec *dsv;

  if (compar == NULL)
	c = strtable_column_compar;
  else
	c = compar;

  dsv = dsvec_init(NULL, sizeof(Strtable_row_sort *));
  for (row = 0; row < stab->nrows; row++) {
	char *s;
	Strtable_row_sort *cs;

	s = (char *) dsvec_ptr_index(stab->columns[col], row);
	cs = ALLOC(Strtable_row_sort);
	cs->rownum = row;
	cs->colval = s;
	dsvec_add_ptr(dsv, cs);
  }

  dsvec_sort(dsv, c);

  return(dsv);
}

/*
 * Format and output table ST to FP_OUT.
 * If DSV_SORTED is non-NULL, use the row permutation that it describes;
 * otherwise the rows are unsorted.
 */
void
strtable_format(FILE *fp_out, Strtable *stab, Dsvec *dsv_sorted)
{
  int col, len, maxlen, row, *widths;
  char *s;
  Dsvec *dsv;
  FILE *fp;

  if (stab == NULL || stab->nrows == 0)
	return;

  if ((fp = fp_out) == NULL)
	fp = stdout;

  widths = ALLOC_N(int, stab->ncols);
  for (col = 0; col < stab->ncols; col++) {
	dsv = stab->columns[col];
	maxlen = 0;
	for (row = 0; row < stab->nrows; row++) {
	  s = dsvec_ptr_index(dsv, row);
	  if ((len = strlen(s)) > maxlen)
		maxlen = len;
	}
	widths[col] = maxlen;
  }

  for (row = 0; row < stab->nrows; row++) {
	Strtable_row_sort *cs;

	if (dsv_sorted != NULL)
	  cs = (Strtable_row_sort *) dsvec_ptr_index(dsv_sorted, row);
	else
	  cs = NULL;

	for (col = 0; col < stab->ncols; col++) {
	  dsv = stab->columns[col];
	  if (cs != NULL)
		s = dsvec_ptr_index(dsv, cs->rownum);
	  else
		s = dsvec_ptr_index(dsv, row);

	  len = fprintf(fp, "%s", s);
	  if (col == (stab->ncols - 1))
		fprintf(fp, "\n");
	  else
		fprintf(fp, "%*s", stab->column_sep + widths[col] - len, " ");
	}
  }

  free(widths);
}

void
strtable_free(Strtable *stab)
{

}

#define STRTABLE_RULER	\
  "123456789a123456789b123456789c123456789d123456789e123456789f"

static MAYBE_UNUSED void
strtable_test(void)
{
  int align, column_sep, i, ncols, sort_col;
  Dsvec *dsv_sorted;
  Strtable *stab;

  ncols = 3;
  align = 1;
  column_sep = 3;
  sort_col = 2;
  fprintf(stderr, "columns=%d, column_sep=%d, align=%d\n",
		  ncols, column_sep, align);

  stab = strtable_init(ncols, column_sep, align);
  strtable_add_row(stab, "First",    "hello",     "zippy");
  strtable_add_row(stab, "Second",   "alphabeta", "zardoz");
  strtable_add_row(stab, "Third",    "xx",        "zinc");
  strtable_add_row(stab, "4th",      "flibber",   "zero");
  strtable_add_row(stab, "Careless", "frogs",     "zoltan");

  fprintf(stderr, "%s (unsorted)\n", STRTABLE_RULER);
  strtable_format(stderr, stab, NULL);
  fprintf(stderr, "\n");

  dsv_sorted = strtable_sort(stab, sort_col, NULL);
  fprintf(stderr, "Sorted on column %d (numbered from zero):", sort_col);
  for (i = 0; i < dsvec_len(dsv_sorted); i++) {
	Strtable_row_sort *cs;

	cs = (Strtable_row_sort *) dsvec_ptr_index(dsv_sorted, i);
	printf("%s%d", (i == 0) ? " " : ", ", cs->rownum);
  }
  fprintf(stderr, "\n");

  fprintf(stderr, "%s (sorted)\n", STRTABLE_RULER);
  strtable_format(stderr, stab, dsv_sorted);
  fprintf(stderr, "%s\n", STRTABLE_RULER);

  strtable_free(stab);
}

/*
 * Given INTVAL, an arbitrarily long, optionally signed digit string,
 * separate groups of three consecutive digits in the usual way, putting
 * the output at the front of DS; e.g., "-1234" becomes "-1,234".
 * In addition, a leading '$' is treated the same as a signed digit string.
 *
 * The syntax of INTVAL is not validated in any way, so it could actually
 * be anything.
 *
 * The result is NOT null-terminated.
 */
static Ds *
comma_ize(Ds *ds, const char *intval, Strcomma_flag flag)
{
  int n;
  char sign_ch;
  const char *e, *s;

  sign_ch = 0;
  s = intval;
  if (flag == STRCOMMA_DEFAULT) {
	if (*intval == '-' || *intval == '+' || *intval == '$') {
	  sign_ch = *intval;
	  s++;
	}
  }
  else if (flag == STRCOMMA_WITHSIGN) {
	if (*intval == '-' || *intval == '$') {
	  sign_ch = *intval;
	  s++;
	}
	else
	  sign_ch = '+';
  }
  else if (flag == STRCOMMA_WITHOUTSIGN) {
	if (*intval == '-' || *intval == '+' || *intval == '$')
	  s++;
  }

  n = 0;
  for (e = s + strlen(s) - 1; e >= s; e--) {
    if (n && (n % 3) == 0)
      ds_prependc(ds, (int) STRCOMMA_CHAR);
    ds_prependc(ds, (int) *e);
    n++;
  }

  if (sign_ch)
	ds_prependc(ds, (int) sign_ch);

  return(ds);
}

/*
 * This is a helper to "comma-ize" INTVAL, a (possibly signed) string of digits.
 * A new, null-terminated string is allocated.  The string argument is
 * not modified.
 */
static char *
strcomma_int(const char *intval, Strcomma_flag flag)
{
  Ds d;

  if (intval == NULL || *intval == '\0')
	return(NULL);

  ds_init(&d);
  comma_ize(&d, intval, flag);
  ds_appendc(&d, (int) '\0');

  return(ds_buf(&d));
}

/*
 * "comma-ize" a sequence of digits.
 * The TYPE identifies the type of ARG, which is either a known binary number
 * type or a string that is (assumed to be) a sequence of digits.
 * For example, the STRNUM_I (a signed integer) value "1234" would produce
 * "1,234", and the STRNUM_STR string "+1234" would produce "+1,234".
 *
 * No syntactical type validation of any kind is done.
 *
 * The main pain of using strcomma() is discovered when the value being
 * converted is an expression.  The value has to be temporarily assigned
 * to a variable of the appropriate type so that its address can be passed.
 *
 * XXX a negative integer type will have a '-' prepended, but a positive
 * integer type does not have a '+' prepended.
 * XXX support STRNUM_F, STRNUM_D, STRNUM_LD
 */
char *
strcomma(Strnum type, void *arg, Strcomma_flag flag)
{
  char buf[128], *cstr, *strarg;
  Strnum num_type;

  if (arg == NULL)
	return(NULL);

  num_type = type;
  if (type == STRNUM_UI32_T) {
    /* 32-bit unsigned int. */
    if (sizeof(STRNUM_UI32_T_TYPE) == sizeof(STRNUM_UI_TYPE))
      num_type = STRNUM_UI;
    else
      return(NULL);
  }
  else if (type == STRNUM_SIZE_T) {
    /* An unsigned int. */
    if (sizeof(STRNUM_SIZE_T_TYPE) == sizeof(STRNUM_UI_TYPE))
      num_type = STRNUM_UI;
    else if (sizeof(STRNUM_SIZE_T_TYPE) == sizeof(STRNUM_UL_TYPE))
      num_type = STRNUM_UL;
    else if (sizeof(STRNUM_SIZE_T_TYPE) == sizeof(STRNUM_ULL_TYPE))
      num_type = STRNUM_ULL;
    else
      return(NULL);
  }
  else if (type == STRNUM_SSIZE_T) {
    /* A signed int. */
    if (sizeof(STRNUM_SSIZE_T_TYPE) == sizeof(STRNUM_I_TYPE))
      num_type = STRNUM_UI;
    else if (sizeof(STRNUM_SSIZE_T_TYPE) == sizeof(STRNUM_L_TYPE))
      num_type = STRNUM_UL;
    else if (sizeof(STRNUM_SSIZE_T_TYPE) == sizeof(STRNUM_LL_TYPE))
      num_type = STRNUM_ULL;
    else
      return(NULL);
  }
  else if (type == STRNUM_UI64_T) { 
    /* 64-bit unsigned int. */
    if (sizeof(STRNUM_UI64_T_TYPE) == sizeof(STRNUM_UI_TYPE))
      num_type = STRNUM_UI;
    else if (sizeof(STRNUM_UI64_T_TYPE) == sizeof(STRNUM_UL_TYPE))
      num_type = STRNUM_UL;
    else if (sizeof(STRNUM_UI64_T_TYPE) == sizeof(STRNUM_ULL_TYPE))
      num_type = STRNUM_ULL;
    else
      return(NULL);
  }
  else if (type == STRNUM_I32_T) {
    /* 32-bit int. */
    if (sizeof(STRNUM_I32_T_TYPE) == sizeof(STRNUM_I_TYPE))
      num_type = STRNUM_I;
    else
      return(NULL);
  }
  else if (type == STRNUM_I64_T) { 
    /* 64-bit int. */
    if (sizeof(STRNUM_I64_T_TYPE) == sizeof(STRNUM_I_TYPE))
      num_type = STRNUM_I;
    else if (sizeof(STRNUM_I64_T_TYPE) == sizeof(STRNUM_L_TYPE))
      num_type = STRNUM_L;
    else if (sizeof(STRNUM_I64_T_TYPE) == sizeof(STRNUM_LL_TYPE))
      num_type = STRNUM_LL;
    else
      return(NULL);
  }

  switch (num_type) {
  case STRNUM_STR:
	strarg = (char *) arg;
	break;

  case STRNUM_I:
	if (snprintf(buf, sizeof(buf), "%d", *(int *) arg) < 0)
	  return(NULL);
	strarg = buf;
	break;

  case STRNUM_L:
	if (snprintf(buf, sizeof(buf), "%ld", *(long *) arg) < 0)
		return(NULL);
	  strarg = buf;
	break;

  case STRNUM_LL:
	if (snprintf(buf, sizeof(buf), "%lld", *(long long *) arg) < 0)
	  return(NULL);
	strarg = buf;
	break;

  case STRNUM_UI:
	if (snprintf(buf, sizeof(buf), "%u", *(unsigned int *) arg) < 0)
	  return(NULL);
	strarg = buf;
	break;

  case STRNUM_UL:
	if (snprintf(buf, sizeof(buf), "%lu", *(unsigned long *) arg) < 0)
	  return(NULL);
	strarg = buf;
	break;

  case STRNUM_ULL:
	if (snprintf(buf, sizeof(buf), "%llu", *(unsigned long long *) arg) < 0)
	  return(NULL);
	strarg = buf;
	break;

  default:
	return(NULL);
	/*NOTREACHED*/
  }

  cstr = strcomma_int(strarg, flag);

  return(cstr);
}

/*
 * "Fast Hashing of Variable-Length Text Strings", Peter K. Pearson,
 * Computing Practices, CACM, Vol. 33, No. 6, (June, 1990), pp. 677-680.
 * See also RFC 3074
 */

static unsigned char *hash_pearson_T = NULL;

/*
 * As a test, the following permutation table should yield these hash values:
 * "a"   = 1, "for"  =  9, "in"   = 17, "the"   = 25
 * "and" = 2, "from" = 10, "is"   = 18, "this"  = 26
 * "are" = 3, "had"  = 11, "it"   = 19, "to"    = 27
 * "as"  = 4, "have" = 12, "not"  = 20, "was"   = 28
 * "at"  = 5, "he"   = 13, "of"   = 21, "which" = 29
 * "be"  = 6, "her"  = 14, "on"   = 22, "with"  = 30
 * "but" = 7, "his"  = 15, "or"   = 23, "you"   = 31
 * "by"  = 8, "i"    = 16, "that" = 24
 */
static unsigned char hash_pearson_test_T[256] = {
 39, 159, 180, 252,  71,   6,  13, 164, 232,  35, 226, 155,  98, 120, 154,  69,
157,  24, 137,  29, 147,  78, 121,  85, 112,   8, 248, 130,  55, 117, 190, 160,
176, 131, 228,  64, 211, 106,  38,  27, 140,  30,  88, 210, 227, 104,  84,  77,
 75, 107, 169, 138, 195, 184,  70,  90,  61, 166,   7, 244, 165, 108, 219,  51,
  9, 139, 209,  40,  31, 202,  58, 179, 116,  33, 207, 146,  76,  60, 242, 124,
254, 197,  80, 167, 153, 145, 129, 233, 132,  48, 246,  86, 156, 177,  36, 187,
 45,   1,  96,  18,  19,  62, 185, 234,  99,  16, 218,  95, 128, 224, 123, 253,
 42, 109,   4, 247,  72,   5, 151, 136,   0, 152, 148, 127, 204, 133,  17,  14,
182, 217,  54, 199, 119, 174,  82,  57, 215,  41, 114, 208, 206, 110, 239,  23,
189,  15,   3,  22, 188,  79, 113, 172,  28,   2, 222,  21, 251, 225, 237, 105,
102,  32,  56, 181, 126,  83, 230,  53, 158,  52,  59, 213, 118, 100,  67, 142,
220, 170, 144, 115, 205,  26, 125, 168, 249,  66, 175,  97, 255,  92, 229,  91,
214, 236, 178, 243,  46,  44, 201, 250, 135, 186, 150, 221, 163, 216, 162,  43,
 11, 101,  34,  37, 194,  25,  50,  12,  87, 198, 173, 240, 193, 171, 143, 231,
111, 141, 191, 103,  74, 245, 223,  20, 161, 235, 122,  63,  89, 149,  73, 238,
134,  68,  93, 183, 241,  81, 196,  49, 192,  65, 212,  94, 203,  10, 200,  47
};

#ifdef NOTDEF
static unsigned char *
hash_pearson_get(void)
{

  return(hash_pearson_T);
}
#endif

static void
hash_pearson_test_init(void)
{

  hash_pearson_T = hash_pearson_test_T;
}

static unsigned char *
hash_pearson_set(unsigned char *new_T)
{
  int i;
  unsigned char b, *r, *T;
  extern unsigned char *crypto_make_random_buffer(unsigned int len);

  if (new_T != NULL) {
	hash_pearson_T = new_T;
	return(new_T);
  }
  
  T = (unsigned char *) malloc(256);

  /* The only bad T is supposedly the unpermuted one. */
  for (i = 0; i < 256; i++)
	T[i] = i;

#ifdef _DACS_CRYPTO_H_
  r = crypto_make_random_buffer(64);
#else
  r = (unsigned char *) malloc(64);
  r[0] = getpid() & 0377;
  for (i = 1; i < 64; i++) {
	time_t now = time(NULL);

	r[i] = r[i - 1] ^ ((now + 1) & 0377);
  }	
#endif

  /* Shuffle */
  for (i = 0; i < 64; i++) {
	b = T[i];
	T[i] = T[r[i]];
	T[r[i]] = b;
  }

  hash_pearson_T = T;

  return(T);
}

static DACS_ui8
hash_pearson8(unsigned char *str, size_t slen, unsigned char first_ch)
{
  int i;
  unsigned char ch, *s, *T;
  DACS_ui8 h;
  size_t len;

  if ((T = hash_pearson_T) == NULL)
	T = hash_pearson_set(NULL);

  s = str;
  if (slen == 0)
	len = strlen((char *) str);
  else
	len = slen;

  if (len == 0)
	return(0);

  if ((ch = first_ch) == 0)
	ch = s[0];

  h = T[ch];
  for (i = 1; i < len; i++) {
	h = T[h ^ s[i]];
  }

  return(h);
}

static DACS_ui16
hash_pearson16(unsigned char *str, size_t slen)
{
  unsigned char ch, h1, h2;
  DACS_ui16 h;

  h1 = hash_pearson8(str, slen, 0);

  ch = str[0] + 1;
  h2 = hash_pearson8(str, slen, ch);

  h = (h1 << 8) | h2;

  return(h);
}

/*
 * Bernstein hash (2nd version w/ xor)
 */
DACS_ui32
hash_djb2(unsigned char *str, size_t slen)
{
  unsigned char *s;
  size_t len;
  DACS_ui32 h;

  s = str;
  if (slen == 0)
	len = strlen((char *) str);
  else
	len = slen;

  h = 0;
  while (len-- > 0) {
	/* hash = hash * 33 ^ c */
	h = ((h << 5) + h) ^ *s++;
  }

  return(h);
}

/*
 * From sdbm, an ndbm work-alike hashed database library
 * Author: Ozan Yigit <oz@nexus.yorku.ca>
 * Status: public domain.
 *
 * polynomial conversion ignoring overflows
 * [this seems to work remarkably well, in fact better
 * then the ndbm hash function. Replace at your own risk]
 * use: 65599   nice.
 *      65587   even better.
 *
 * [In one experiment, this function hashed 84165 symbols (English words
 * plus symbol table values) with no collisions. -bjb]
 * Do NOT use for cryptographic purposes.
 */
static DACS_ui32
hash_sdbm32(unsigned char *str, size_t slen, DACS_ui32 level)
{
  unsigned char *s;
  size_t len;
  DACS_ui32 n;

  s = str;
  if (slen == 0)
	len = strlen((char *) str);
  else
	len = slen;
  n = level;

#define HASHC   n = *s++ + 65587 * n

#ifdef NODUFF
  while (len--)
	n = HASHC;
#else

  if (len > 0) {
	DACS_ui32 loop;

	loop = (len + 8 - 1) >> 3;
	switch (len & (8 - 1)) {
	case 0:
	  do {
		HASHC;
	  case 7: HASHC;
	  case 6: HASHC;
	  case 5: HASHC;
	  case 4: HASHC;
	  case 3: HASHC;
	  case 2: HASHC;
	  case 1: HASHC;
	  } while (--loop);
	}
  }
#endif
  return(n);
}

/*
 * lookup8.c, by Bob Jenkins, January 4 1997
 * You can use this free for any purpose.  It has no warranty.
 * http://burtleburtle.net/bob/c/lookup8.c
 */

#define hashsize(n) ((DACS_ui64) 1 << (n))
#define hashmask(n) (hashsize(n) - 1)

/*
 * mix -- mix 3 64-bit values reversibly.
 * mix() takes 48 machine instructions, but only 24 cycles on a superscalar
 *   machine (like Intel's new MMX architecture).  It requires 4 64-bit
 *   registers for 4::2 parallelism.
 * All 1-bit deltas, all 2-bit deltas, all deltas composed of top bits of
 *   (a,b,c), and all deltas of bottom bits were tested.  All deltas were
 *   tested both on random keys and on keys that were nearly all zero.
 *   These deltas all cause every bit of c to change between 1/3 and 2/3
 *   of the time (well, only 113/400 to 287/400 of the time for some
 *   2-bit delta).  These deltas all cause at least 80 bits to change
 *   among (a,b,c) when the mix is run either forward or backward (yes it
 *   is reversible).
 * This implies that a hash using mix64 has no funnels.  There may be
 *   characteristics with 3-bit deltas or bigger, I didn't test for
 *   those.
 */

#define mix64(a, b, c) {						\
	a -= b; a -= c; a ^= (c >> 43);				\
	b -= c; b -= a; b ^= (a <<  9);				\
	c -= a; c -= b; c ^= (b >>  8);				\
	a -= b; a -= c; a ^= (c >> 38);				\
	b -= c; b -= a; b ^= (a << 23);				\
	c -= a; c -= b; c ^= (b >>  5);				\
	a -= b; a -= c; a ^= (c >> 35);				\
	b -= c; b -= a; b ^= (a << 49);				\
	c -= a; c -= b; c ^= (b >> 11);				\
	a -= b; a -= c; a ^= (c >> 12);				\
	b -= c; b -= a; b ^= (a << 18);				\
	c -= a; c -= b; c ^= (b >> 22);				\
  }

/*
 * hash() -- hash a variable-length key into a 64-bit value
 *   k     : the key (the unaligned variable-length array of bytes)
 *   len   : the length of the key, counting by bytes
 *   level : can be any 8-byte value
 * Returns a 64-bit value.  Every bit of the key affects every bit of
 * the return value.  No funnels.  Every 1-bit and 2-bit delta achieves
 * avalanche.  About 41+5len instructions.
 * 
 * The best hash table sizes are powers of 2.  There is no need to do
 * mod a prime (mod is sooo slow!).  If you need less than 64 bits,
 * use a bitmask.  For example, if you need only 10 bits, do
 *   h = (h & hashmask(10));
 * In which case, the hash table should have hashsize(10) elements.
 * 
 * If you are hashing n strings (u_int8 **)k, do it like this:
 *   for (i=0, h=0; i<n; ++i) h = hash( k[i], len[i], h);
 * 
 * By Bob Jenkins, Jan 4 1997.  bob_jenkins@compuserve.com.  You may
 * use this code any way you wish, private, educational, or commercial,
 * as long as this whole comment accompanies it.
 *
 * See:
 *    http://burtleburtle.net/bob/c/lookup8.c
 * Use for hash table lookup, or anything where one collision in 2^^64
 * is acceptable.  Do NOT use for cryptographic purposes.
 */

/*
 * The key, the length of the key, the previous hash, or an arbitrary value.
 */
static DACS_ui64
hash_bj64(unsigned char *str, size_t slen, DACS_ui64 level)
{
  unsigned char *k;
  size_t len;
  DACS_ui64 a, b, c;

  /* Set up the internal state */
  if (slen == 0)
	len = strlen((char *) str);
  else
	len = slen;
  a = b = level;                /* the previous hash value */
  c = 0x9e3779b97f4a7c13LL;		/* the golden ratio; an arbitrary value */
  k = str;

  /* Handle most of the key */
  while (len >= 24) {
    a += (k[0]
		  + ((DACS_ui64) k[ 1] <<  8)
		  + ((DACS_ui64) k[ 2] << 16) + ((DACS_ui64) k[ 3] << 24)
		  + ((DACS_ui64) k[ 4] << 32) + ((DACS_ui64) k[ 5] << 40)
		  + ((DACS_ui64) k[ 6] << 48) + ((DACS_ui64) k[ 7] << 56));

    b += (k[8]
		  + ((DACS_ui64) k[ 9] <<  8) + ((DACS_ui64) k[10] << 16)
		  + ((DACS_ui64) k[11] << 24) + ((DACS_ui64) k[12] << 32)
		  + ((DACS_ui64) k[13] << 40) + ((DACS_ui64) k[14] << 48)
		  + ((DACS_ui64) k[15] << 56));

    c += (k[16]
		  + ((DACS_ui64) k[17] <<  8) + ((DACS_ui64) k[18] << 16)
		  + ((DACS_ui64) k[19] << 24) + ((DACS_ui64) k[20] << 32)
		  + ((DACS_ui64) k[21] << 40) + ((DACS_ui64) k[22] << 48)
		  + ((DACS_ui64) k[23] << 56));

    mix64(a, b, c);
    k += 24;
	len -= 24;
  }

  /* Handle the last 63 bytes */
  c += len;
  switch(len) {
	/* All the case statements fall through... */
  case 23: c += ((DACS_ui64) k[22] << 56);
  case 22: c += ((DACS_ui64) k[21] << 48);
  case 21: c += ((DACS_ui64) k[20] << 40);
  case 20: c += ((DACS_ui64) k[19] << 32);
  case 19: c += ((DACS_ui64) k[18] << 24);
  case 18: c += ((DACS_ui64) k[17] << 16);
  case 17: c += ((DACS_ui64) k[16] <<  8);

    /* the first byte of c is reserved for the length */
  case 16: b += ((DACS_ui64) k[15] << 56);
  case 15: b += ((DACS_ui64) k[14] << 48);
  case 14: b += ((DACS_ui64) k[13] << 40);
  case 13: b += ((DACS_ui64) k[12] << 32);
  case 12: b += ((DACS_ui64) k[11] << 24);
  case 11: b += ((DACS_ui64) k[10] << 16);
  case 10: b += ((DACS_ui64) k[ 9] << 8);
  case  9: b += ((DACS_ui64) k[ 8]);
  case  8: a += ((DACS_ui64) k[ 7] << 56);
  case  7: a += ((DACS_ui64) k[ 6] << 48);
  case  6: a += ((DACS_ui64) k[ 5] << 40);
  case  5: a += ((DACS_ui64) k[ 4] << 32);
  case  4: a += ((DACS_ui64) k[ 3] << 24);
  case  3: a += ((DACS_ui64) k[ 2] << 16);
  case  2: a += ((DACS_ui64) k[ 1] <<  8);
  case  1:  a+= ((DACS_ui64) k[ 0]);
    /* case 0: nothing left to add */
  }

  mix64(a, b, c);

  return(c);
}

DACS_ui8
strhash8(unsigned char *str, size_t slen)
{

  return(hash_pearson8(str, slen, 0));
}

DACS_ui16
strhash16(unsigned char *str, size_t slen)
{

  return(hash_pearson16(str, slen));
}

DACS_ui32
strhash32(unsigned char *str, size_t slen)
{

  return(hash_sdbm32(str, slen, 0));
}

DACS_ui64
strhash64(unsigned char *str, size_t slen)
{

  return(hash_bj64(str, slen, 0));
}

/*
 * Compute the hash of STR (SLEN bytes long) using the algorithm named ALG.
 * The result is stored in the memory pointed to by VALUE, which is assumed
 * to be large enough to hold the hash value.
 * Return 0 if ok, -1 if an error occurred.
 */
int
strhash(char *alg, unsigned char *str, size_t slen, void *valuep)
{

  if (alg == NULL || str == NULL || slen == 0 || valuep == NULL)
	return(-1);

  if (strcaseeq(alg, "bj") || strcaseeq(alg, "bj_64")) {
	DACS_ui64 val;

	val = hash_bj64(str, slen, 0);
	*(DACS_ui64 *) valuep = val;
  }
  else if (strcaseeq(alg, "sdbm") || strcaseeq(alg, "sdbm_32")) {
	DACS_ui32 val;

	val = hash_sdbm32(str, slen, 0);
	*(DACS_ui32 *) valuep = val;
  }
  else if (strcaseeq(alg, "djb2") || strcaseeq(alg, "djb2_32")) {
	DACS_ui32 val;

	val = hash_djb2(str, slen);
	*(DACS_ui32 *) valuep = val;
  }
  else if (strcaseeq(alg, "pearson_16")) {
	DACS_ui16 val;

	hash_pearson_test_init();
	val = hash_pearson16(str, slen);
	*(DACS_ui16 *) valuep = val;
  }
  else if (strcaseeq(alg, "pearson") || strcaseeq(alg, "pearson_8")) {
	DACS_ui8 val;

	hash_pearson_test_init();
	val = hash_pearson8(str, slen, 0);
	*(DACS_ui8 *) valuep = val;
  }
  else
	return(-1);

  return(0);
}

/*
 * Create an inode mode descriptor string from MODE.
 * Alternate implementation of FreeBSD's strmode(3).
 * The string is always 11 bytes long (plus a null).
 * The last character is always a space.
 *
 * If OUTP is non-NULL, it is assumed to be at least 12 bytes; if OUTP is
 * NULL, allocate the 12 bytes.
 * Return a pointer to the formatted string.
 */
char *
strfilemode(mode_t mode, char *outp)
{
  char *p, *s;

  if (outp == NULL)
	s = p = (char *) malloc(12);
  else
	s = p = outp;

  /* Inode type (one character). */
  switch (mode & S_IFMT) {
  case S_IFREG:           /* Regular */
	*p++ = '-';
	break;
  case S_IFBLK:           /* Block special */
	*p++ = 'b';
	break;
  case S_IFCHR:           /* Character special */
	*p++ = 'c';
	break;
  case S_IFDIR:           /* Directory */
	*p++ = 'd';
	break;
  case S_IFLNK:           /* Symbolic link */
	*p++ = 'l';
	break;
#ifdef S_IFIFO
  case S_IFIFO:           /* FIFO */
	*p++ = 'p';
	break;
#endif
  case S_IFSOCK:          /* Socket */
	*p++ = 's';
	break;
#ifdef S_IFWHT
  case S_IFWHT:           /* Whiteout (BSD) */
	*p++ = 'w';
	break;
#endif
  default:            /* Unknown inode type */
	*p++ = '?';
	break;
  }

  /* Owner's permissions (3 characters). */
  if (mode & S_IRUSR)
	*p++ = 'r';
  else
	*p++ = '-';

  if (mode & S_IWUSR)
	*p++ = 'w';
  else
	*p++ = '-';

  switch (mode & (S_IXUSR | S_ISUID)) {
  case 0:
	*p++ = '-';
	break;
  case S_IXUSR:
	*p++ = 'x';
	break;
  case S_ISUID:
	*p++ = 'S';
	break;
  case S_IXUSR | S_ISUID:
	*p++ = 's';
	break;
  default:            /* Unknown? */
	*p++ = '?';
	break;
  }

  /* Group's permissions (3 characters). */
  if (mode & S_IRGRP)
	*p++ = 'r';
  else
	*p++ = '-';

  if (mode & S_IWGRP)
	*p++ = 'w';
  else
	*p++ = '-';

  switch (mode & (S_IXGRP | S_ISGID)) {
  case 0:
	*p++ = '-';
	break;
  case S_IXGRP:
	*p++ = 'x';
	break;
  case S_ISGID:
	*p++ = 'S';
	break;
  case S_IXGRP | S_ISGID:
	*p++ = 's';
	break;
  default:            /* Unknown? */
	*p++ = '?';
	break;
  }

  /* Other's (default) permissions (3 characters). */
  if (mode & S_IROTH)
	*p++ = 'r';
  else
	*p++ = '-';

  if (mode & S_IWOTH)
	*p++ = 'w';
  else
	*p++ = '-';

  switch (mode & (S_IXOTH | S_ISVTX)) {
  case 0:
	*p++ = '-';
	break;
  case S_IXOTH:
	*p++ = 'x';
	break;
  case S_ISVTX:
	*p++ = 'T';
	break;
  case S_IXOTH | S_ISVTX:
	*p++ = 't';
	break;
  default:            /* Unknown? */
	*p++ = '?';
	break;
  }

  *p++ = ' ';
  *p = '\0';

  return(s);
}

#ifdef PROG

typedef struct Strmatch_test {
  char *str;
  char *regex;
  int cflags;
  int expected_val;
} Strmatch_test;

static Strmatch_test strmatch1_tests[] = {
  { "pwf.pf.a", "\\(.*\\)\\.\\(PF\\)\\.\\(.\\)", REG_ICASE, 4 },
  { "pwf.pf.a", "PWF.PF.A", REG_ICASE | REG_NOSUB, 1 },
  { "pwf.pf.a", "XPWF.PF.A", REG_ICASE | REG_NOSUB, 0 },
  { "pwf.pf.a", ".*PWF.PF.A", REG_ICASE | REG_NOSUB, 1 },
  { "123 xx pwf.pf.a", "123 .* pwf.pf.a", REG_EXTENDED | REG_NOSUB, 1 },
  { NULL, NULL, 0, 0 }
};

static Strmatch_test strmatch2_tests[] = {
  { "pwf.pf.a", "\\(.*\\)\\.\\(PF\\)\\.\\(.\\)", REG_ICASE, 1 },
  { "pwf.pf.a", "PWF.PF.A", REG_ICASE | REG_NOSUB, 1 },
  { "pwf.pf.a", "XPWF.PF.A", REG_ICASE | REG_NOSUB, 0 },
  { "pwf.pf.a", ".*PWF.PF.A", REG_ICASE | REG_NOSUB, 1 },
  { "123 xx pwf.pf.a", "123 .* pwf.pf.a", REG_EXTENDED | REG_NOSUB, 1 },
  { NULL, NULL, 0, 0 }
};

static int
test_strmatch(void)
{
  int c;
  Strmatch_test *t;

  fprintf(stderr, "Testing string strregex() functions: ");
  c = 0;

  for (t = strmatch1_tests; t->str != NULL; t++) {
	int st;
	char *err;
	Dsvec *dsv;
	Strregex_conf conf;

	conf.regex = t->regex;
	conf.cflags = t->cflags;
	conf.preg_in = NULL;
	conf.nmatch = 20;
	dsv = dsvec_init(NULL, sizeof(char *));

	st = strregex_x(t->str, &conf, dsv);
	if (st == -1) {
	  fprintf(stderr, "test_strmatch test %d failed\n", c);
	  fprintf(stderr, "Regex error %d\n", conf.errcode);
	  return(-1);
	}

	if (st != t->expected_val) {
	  fprintf(stderr, "test_strmatch test %d failed\n", c);
	  fprintf(stderr, "Got %d\nExpected %d\n", st, t->expected_val);
	  return(-1);
	}
	c++;
  }

  for (t = strmatch2_tests; t->str != NULL; t++) {
	int st;
	char *err, *match;

	st = strregex(t->str, t->regex, t->cflags, &match, &err);
	if (st != t->expected_val) {
	  fprintf(stderr, "test_strmatch test %d failed\n", c);
	  fprintf(stderr, "Got %d\nExpected %d\n", st, t->expected_val);
	  return(-1);
	}
	c++;
  }

  fprintf(stderr, "%d tests ok\n", c);
  return(0);
}

typedef struct Strstr_test_word {
  char *word;
  int count;
  int expected_count;
} Strstr_test_word;

static Strstr_test_word strstr_words[] = {
  { "car", 0, -1 },  { "dog", 0, -1 }, {"cat", 0, -1 }, { "mond", 0, -1 },
  { "cing", 0, -1 }, { "diate", 0, -1 }, { "strange", 0, -1 },
  { "astrocylonic", 0, -1 }, { NULL, 0 }
};

static int
test_strstr(FILE *fp, char *search_path)
{
  int i;
  char *buf, *path, *w;
  size_t size;
  Dsvec *dsv;
  FILE *out;

  if (fp == NULL)
	out = stderr;
  else
	out = fp;

  if (search_path == NULL || *search_path == '\0') {
	fprintf(out, "No path specified, skipping this test.\n");
	return(0);
  }

  fprintf(fp, "Testing strstr() functions using search path \"%s\"\n",
		  search_path);
  if ((dsv = strsplit(search_path, ":", 0)) == NULL) {
	fprintf(out, "Invalid path specified, skipping this test.\n");
	return(0);
  }

  for (i = 0; i < dsvec_len(dsv); i++) {
	path = (char *) dsvec_ptr_index(dsv, i);
	if (load_file(path, &buf, &size) != -1)
	  break;
  }

  if (i == dsvec_len(dsv)) {
	fprintf(out, "Cannot load a file, skipping this test.\n");
	return(0);
  }
  fprintf(out, "Using \"%s\"\n", path);

  for (i = 0; strstr_words[i].word != NULL; i++)
	strstr_words[i].expected_count = -1;

#ifdef HAVE_STRSTR
  for (i = 0; strstr_words[i].word != NULL; i++) {
	w = buf;
	strstr_words[i].count = 0;
	while ((w = strstr(w, strstr_words[i].word)) != NULL) {
	  strstr_words[i].count++;
	  w++;
	}
  }

  fprintf(out, "strstr() [system implementation]:\n");
  for (i = 0; strstr_words[i].word != NULL; i++) {
	fprintf(out, "Word %d: \"%s\": %d\n",
			i, strstr_words[i].word, strstr_words[i].count);
	if (strstr_words[i].expected_count == -1)
	  strstr_words[i].expected_count = strstr_words[i].count;
	else if (strstr_words[i].expected_count != strstr_words[i].count) {
	  fprintf(out, "Inconsistent result!\n");
	  return(-1);
	}
  }
  fprintf(out, "\n");
#endif

  for (i = 0; strstr_words[i].word != NULL; i++) {
	w = buf;
	strstr_words[i].count = 0;
	while ((w = dacs_strstr(w, strstr_words[i].word)) != NULL) {
	  strstr_words[i].count++;
	  w++;
	}
  }

  fprintf(out, "dacs_strstr [DACS implementation]:\n");
  for (i = 0; strstr_words[i].word != NULL; i++) {
	fprintf(out, "Word %d: \"%s\": %d\n",
			i, strstr_words[i].word, strstr_words[i].count);
	if (strstr_words[i].expected_count == -1)
	  strstr_words[i].expected_count = strstr_words[i].count;
	else if (strstr_words[i].expected_count != strstr_words[i].count) {
	  fprintf(out, "Inconsistent result!\n");
	  return(-1);
	}
  }
  fprintf(out, "\n");

  for (i = 0; strstr_words[i].word != NULL; i++) {
	unsigned int *td;
	size_t plen;

	w = buf;
	strstr_words[i].count = 0;
	plen = strlen(strstr_words[i].word);
	td = strqs_init(NULL, strstr_words[i].word, plen, 0);

	while ((w = strqsx(w, buf + size, strstr_words[i].word, td, plen, 0))
		   != NULL) {
	  strstr_words[i].count++;
	  w++;
	}
  }

  fprintf(out, "strqs [DACS Quick Search implementation]:\n");
  for (i = 0; strstr_words[i].word != NULL; i++) {
	fprintf(out, "Word %d: \"%s\": %d\n",
			i, strstr_words[i].word, strstr_words[i].count);
	if (strstr_words[i].expected_count == -1)
	  strstr_words[i].expected_count = strstr_words[i].count;
	else if (strstr_words[i].expected_count != strstr_words[i].count) {
	  fprintf(out, "Inconsistent result!\n");
	  return(-1);
	}
  }

  return(1);
}

typedef struct Test_path {
  char *path;
  int nels;					/* Number of elements in the path. */
  char *dirname;			/* Directory components of the path. */
  char *basename;			/* Last component, with extension. */
  char *trimmed_basename;	/* Last component, without extension. */
  char *extname;			/* File extension, if any. */
} Test_path;

/* A filename extension not used in any test path below. */
#define NONMATCHING_EXT		".xxx"

static Test_path test_path[] = {
  { "foo/bar/baz", 3, "foo/bar", "baz",     "baz", NULL },
  { "/a/b/c.d",    3, "/a/b",    "c.d",     "c",   ".d" },
  { "/a/b/c.cc",   3, "/a/b",    "c.cc",    "c",   ".cc" },
  { "foo.cpp",     1, ".",       "foo.cpp", "foo", ".cpp" },
  { NULL, 0, NULL, NULL, NULL, NULL }
};

static int
test_pathfuncs(void)
{
  int c;
  Test_path *tp;

  c = 0;
  for (tp = &test_path[0]; tp->path != NULL; tp++) {
	int failed, nels;
	char *d, *b, *last, *p, *s;
	Dsvec *dsv;

	if  ((dsv = strpathparse(tp->path, NULL, NULL)) == NULL
		 || (nels = dsvec_len(dsv)) != tp->nels) {
	  fprintf(stderr, "Test %d: strpathparse failed\n", c);
	  return(-1);
	}

	if ((d = strdirname(tp->path)) == NULL || !streq(d, tp->dirname)) {
	  fprintf(stderr, "Test %d: Incorrect dirname\n", c);
	  fprintf(stderr, "Got \"%s\"\nExpected \"%s\"\n",
			  (d == NULL) ? "NULL" : d, tp->dirname);
	  return(-1);
	}

	if ((b = strbasename(tp->path, NULL)) == NULL || !streq(b, tp->basename)) {
	  fprintf(stderr, "Test %d: Incorrect basename\n", c);
	  fprintf(stderr, "Got \"%s\"\nExpected \"%s\"\n",
			  (b == NULL) ? "NULL" : b, tp->basename);
	  return(-1);
	}

	if (((s = strextname(tp->path)) == NULL && tp->extname != NULL)
		|| (s != NULL && (tp->extname == NULL || !streq(s, tp->extname)))) {
	  fprintf(stderr, "Test %d: Incorrect extname\n", c);
	  fprintf(stderr, "Got \"%s\"\nExpected \"%s\"\n",
			  (s == NULL) ? "NULL" : s, tp->extname);
	  return(-1);
	}

	/* Get the filename component with the extension removed if it matches. */
	failed = 1;
	if ((b = strbasename(tp->path, tp->extname)) != NULL
		&& streq(b, tp->trimmed_basename)) {
	  if ((b = strbasename(tp->path, NONMATCHING_EXT)) != NULL
		  && streq(b, tp->basename))
	  failed = 0;
	}

	if (failed) {
	  fprintf(stderr, "Test %d: Incorrect basename with extension\n", c);
	  fprintf(stderr, "Got \"%s\"\nExpected \"%s\"\n",
			  (b == NULL) ? "NULL" : b, tp->trimmed_basename);
	  return(-1);
	}

	c++;
  }

  return(0);
}

/*
 * A colon-separated list of paths.
 * Any large document or list of words should be ok.
 * Set to the empty string to disable the test.
 */
#define WORD_LIST_FILE_PATH	"/usr/share/dict/words:/usr/dict/words"

int
main(int argc, char **argv)
{

  fprintf(stderr, "Testing string functions...\n");
  if (test_strnum("12345") == -1) {
	fprintf(stderr, "String strnum() test failed!\n");
	exit(1);
  }

  if (test_strmatch() == -1) {
	fprintf(stderr, "String strregex() test failed!\n");
	exit(1);
  }

  if (test_strstr(stderr, WORD_LIST_FILE_PATH) == -1) {
	fprintf(stderr, "String strstr() test failed!\n");
	exit(1);
  }

  if (test_parse(stderr) == -1) {
	fprintf(stderr, "String parsing test failed!\n");
	exit(1);
  }

  if (test_strencdec(stderr) == -1) {
	fprintf(stderr, "String encoding/decoding test failed!\n");
	exit(1);
  }

  if (test_pathfuncs() == -1) {
	fprintf(stderr, "String path parsing test failed!\n");
	exit(1);
  }

  fprintf(stderr, "All string tests succeeded!\n");

  exit(0);
}
#endif
